/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Implementation of thread pool for
 *	Parallel computing.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <complib/cl_pcthreadpool.h>

static cl_status_t thread_validate_affinity(IN uint16_t * active_cores,
					    IN size_t size,
					    IN uint16_t max_threads_per_core)
{
	int i;
	cl_status_t status = CL_SUCCESS;
	pthread_t pthread = pthread_self();

	if (!max_threads_per_core)
		goto Exit;

	for (i = 0; i < size; i++) {
		cpu_set_t my_cpu;
		/* if no suitable core is found, thread keeps */
		/* running on initially allocated core */
		if (active_cores[i] < max_threads_per_core) {
			CPU_ZERO(&my_cpu);
			CPU_SET(i, &my_cpu);
			if (pthread_setaffinity_np(pthread,
						   sizeof(cpu_set_t),
						   &my_cpu) != 0) {
					status = CL_ERROR;
					goto Exit;
			}
			active_cores[i]++;
			break;
		}

	}
Exit:
	return status;
}

static void pc_parallel_function(IN void *context)
{
	cl_pc_thread_pool_t *p_pool = context;
	uint32_t my_id;
	uint32_t num_of_threads;
	CL_ASSERT(p_pool);
	num_of_threads = p_pool->p_thread_pool->running_count;

	/* allocate unique ID for each thread */
	pthread_mutex_lock(&p_pool->mutex);
	my_id = p_pool->max_id;
	p_pool->max_id++;

	thread_validate_affinity(p_pool->active_cores,
				 num_of_threads,
				 p_pool->max_threads_per_core);
	pthread_mutex_unlock(&p_pool->mutex);

	/* run the compute function */
	p_pool->pfn_callback(p_pool->context, my_id);

	pthread_mutex_lock(&p_pool->mutex);
	p_pool->finished_threads++;
	if (p_pool->finished_threads == num_of_threads){
		/* send signal when all threads complete the job */
		cl_event_signal(&p_pool->signal);
		p_pool->max_id = 0;
		p_pool->finished_threads=0;
	}
	pthread_mutex_unlock(&p_pool->mutex);
}

cl_status_t cl_pc_thread_pool_init(IN cl_pc_thread_pool_t * const p_pc_thread_pool,
				   IN unsigned count,
				   IN void (*pfn_callback) (void *, uint32_t),
				   IN uint16_t max_threads_per_core,
				   IN void *context, IN const char *const name)
{
	cl_thread_pool_t *p_thread_pool;
	cl_status_t status = CL_INSUFFICIENT_MEMORY;

	CL_ASSERT(p_pc_thread_pool);
	CL_ASSERT(pfn_callback);

	p_thread_pool = (cl_thread_pool_t *) calloc(1, sizeof(*p_thread_pool));

	if (!p_thread_pool)
		goto Exit;

	p_pc_thread_pool->pfn_callback = pfn_callback;

	pthread_mutex_init(&p_pc_thread_pool->mutex, NULL);

	cl_event_init(&p_pc_thread_pool->signal, FALSE);

	p_pc_thread_pool->p_thread_pool = p_thread_pool;

	p_pc_thread_pool->max_threads_per_core = max_threads_per_core;

	p_pc_thread_pool->max_id = 0;

	p_pc_thread_pool->finished_threads = 0;

	p_pc_thread_pool->context = context;

	status = cl_thread_pool_init(p_thread_pool, count,
				     pc_parallel_function,
				     (void *) p_pc_thread_pool,
				     name);

	if (status != CL_SUCCESS){
		free(p_thread_pool);
		pthread_mutex_destroy(&p_pc_thread_pool->mutex);
		cl_event_destroy(&p_pc_thread_pool->signal);
		goto Exit;
	}

	p_pc_thread_pool->active_cores = (uint16_t *)
				calloc(p_thread_pool->running_count,
				sizeof(*p_pc_thread_pool->active_cores));

	if (!p_pc_thread_pool->active_cores) {
		cl_pc_thread_pool_destroy(p_pc_thread_pool);
		goto Exit;
	}
	status = CL_SUCCESS;
Exit:
	return status;
}

void cl_pc_thread_pool_destroy(IN cl_pc_thread_pool_t * const p_pc_thread_pool)
{
	CL_ASSERT(p_pc_thread_pool);

	cl_thread_pool_destroy(p_pc_thread_pool->p_thread_pool);

	free(p_pc_thread_pool->p_thread_pool);

	p_pc_thread_pool->p_thread_pool = NULL;

	pthread_mutex_destroy(&p_pc_thread_pool->mutex);

	cl_event_destroy(&p_pc_thread_pool->signal);

	free(p_pc_thread_pool->active_cores);

	p_pc_thread_pool->active_cores = NULL;
}

void cl_pc_thread_pool_start(IN cl_pc_thread_pool_t * const p_pc_thread_pool)
{
	int i;
	CL_ASSERT(p_pc_thread_pool);
	/* wake up all computing threads */
	for (i=0; i < p_pc_thread_pool->p_thread_pool->running_count; i++)
		cl_thread_pool_signal(p_pc_thread_pool->p_thread_pool);
}

cl_status_t cl_pc_thread_pool_wait(IN cl_pc_thread_pool_t * const p_pc_thread_pool,
				   IN const uint32_t wait_us)
{
	cl_status_t status;
	CL_ASSERT(p_pc_thread_pool);

	if (!wait_us) {
		status = CL_ERROR;
		goto Exit;
	}
	status = cl_event_wait_on(&p_pc_thread_pool->signal, wait_us, TRUE);
Exit:
	return status;
}

