/*
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright (c) 2019 Mellanox Technologies LTD. All rights reserved.
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * FIPS pub 180-1: Secure Hash Algorithm (SHA-1)
 * based on: http://csrc.nist.gov/fips/fip180-1.txt
 * implemented by Jun-ichiro itojun Itoh <itojun@itojun.org>
 */

#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/time.h>
#include <string.h>

#include <complib/cl_sha1.h>

/* sanity check */
#if BYTE_ORDER != BIG_ENDIAN
# if BYTE_ORDER != LITTLE_ENDIAN
#  define unsupported 1
# endif
#endif

#ifndef unsupported

/* constant table */
static uint32_t _K[] = { 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xca62c1d6 };
#define	K(t)	_K[(t) / 20]

#define	F0(b, c, d)	(((b) & (c)) | ((~(b)) & (d)))
#define	F1(b, c, d)	(((b) ^ (c)) ^ (d))
#define	F2(b, c, d)	(((b) & (c)) | ((b) & (d)) | ((c) & (d)))
#define	F3(b, c, d)	(((b) ^ (c)) ^ (d))

#define	S(n, x)		(((x) << (n)) | ((x) >> (32 - n)))

#define	H(n)	(p_ctxt->h.b32[(n)])
#define	COUNT	(p_ctxt->count)
#define	BCOUNT	(p_ctxt->c.b64[0] / 8)
#define	W(n)	(p_ctxt->m.b32[(n)])

#define	PUTBYTE(x)	{ \
	p_ctxt->m.b8[(COUNT % 64)] = (x);	\
	COUNT++;				\
	COUNT %= 64;				\
	p_ctxt->c.b64[0] += 8;			\
	if (COUNT % 64 == 0)			\
		cl_sha1_step(p_ctxt);		\
     }

#define	PUTPAD(x)	{ \
	p_ctxt->m.b8[(COUNT % 64)] = (x);	\
	COUNT++;				\
	COUNT %= 64;				\
	if (COUNT % 64 == 0)			\
		cl_sha1_step(p_ctxt);		\
     }

static void cl_sha1_step(struct cl_sha1_ctxt *);

static void cl_sha1_step(struct cl_sha1_ctxt *p_ctxt)
{
	uint32_t	a, b, c, d, e;
	size_t t, s;
	uint32_t	tmp;

#if BYTE_ORDER == LITTLE_ENDIAN
	struct cl_sha1_ctxt tctxt;
	memcpy(&tctxt.m.b8[0], &p_ctxt->m.b8[0], 64);
	p_ctxt->m.b8[0] = tctxt.m.b8[3]; p_ctxt->m.b8[1] = tctxt.m.b8[2];
	p_ctxt->m.b8[2] = tctxt.m.b8[1]; p_ctxt->m.b8[3] = tctxt.m.b8[0];
	p_ctxt->m.b8[4] = tctxt.m.b8[7]; p_ctxt->m.b8[5] = tctxt.m.b8[6];
	p_ctxt->m.b8[6] = tctxt.m.b8[5]; p_ctxt->m.b8[7] = tctxt.m.b8[4];
	p_ctxt->m.b8[8] = tctxt.m.b8[11]; p_ctxt->m.b8[9] = tctxt.m.b8[10];
	p_ctxt->m.b8[10] = tctxt.m.b8[9]; p_ctxt->m.b8[11] = tctxt.m.b8[8];
	p_ctxt->m.b8[12] = tctxt.m.b8[15]; p_ctxt->m.b8[13] = tctxt.m.b8[14];
	p_ctxt->m.b8[14] = tctxt.m.b8[13]; p_ctxt->m.b8[15] = tctxt.m.b8[12];
	p_ctxt->m.b8[16] = tctxt.m.b8[19]; p_ctxt->m.b8[17] = tctxt.m.b8[18];
	p_ctxt->m.b8[18] = tctxt.m.b8[17]; p_ctxt->m.b8[19] = tctxt.m.b8[16];
	p_ctxt->m.b8[20] = tctxt.m.b8[23]; p_ctxt->m.b8[21] = tctxt.m.b8[22];
	p_ctxt->m.b8[22] = tctxt.m.b8[21]; p_ctxt->m.b8[23] = tctxt.m.b8[20];
	p_ctxt->m.b8[24] = tctxt.m.b8[27]; p_ctxt->m.b8[25] = tctxt.m.b8[26];
	p_ctxt->m.b8[26] = tctxt.m.b8[25]; p_ctxt->m.b8[27] = tctxt.m.b8[24];
	p_ctxt->m.b8[28] = tctxt.m.b8[31]; p_ctxt->m.b8[29] = tctxt.m.b8[30];
	p_ctxt->m.b8[30] = tctxt.m.b8[29]; p_ctxt->m.b8[31] = tctxt.m.b8[28];
	p_ctxt->m.b8[32] = tctxt.m.b8[35]; p_ctxt->m.b8[33] = tctxt.m.b8[34];
	p_ctxt->m.b8[34] = tctxt.m.b8[33]; p_ctxt->m.b8[35] = tctxt.m.b8[32];
	p_ctxt->m.b8[36] = tctxt.m.b8[39]; p_ctxt->m.b8[37] = tctxt.m.b8[38];
	p_ctxt->m.b8[38] = tctxt.m.b8[37]; p_ctxt->m.b8[39] = tctxt.m.b8[36];
	p_ctxt->m.b8[40] = tctxt.m.b8[43]; p_ctxt->m.b8[41] = tctxt.m.b8[42];
	p_ctxt->m.b8[42] = tctxt.m.b8[41]; p_ctxt->m.b8[43] = tctxt.m.b8[40];
	p_ctxt->m.b8[44] = tctxt.m.b8[47]; p_ctxt->m.b8[45] = tctxt.m.b8[46];
	p_ctxt->m.b8[46] = tctxt.m.b8[45]; p_ctxt->m.b8[47] = tctxt.m.b8[44];
	p_ctxt->m.b8[48] = tctxt.m.b8[51]; p_ctxt->m.b8[49] = tctxt.m.b8[50];
	p_ctxt->m.b8[50] = tctxt.m.b8[49]; p_ctxt->m.b8[51] = tctxt.m.b8[48];
	p_ctxt->m.b8[52] = tctxt.m.b8[55]; p_ctxt->m.b8[53] = tctxt.m.b8[54];
	p_ctxt->m.b8[54] = tctxt.m.b8[53]; p_ctxt->m.b8[55] = tctxt.m.b8[52];
	p_ctxt->m.b8[56] = tctxt.m.b8[59]; p_ctxt->m.b8[57] = tctxt.m.b8[58];
	p_ctxt->m.b8[58] = tctxt.m.b8[57]; p_ctxt->m.b8[59] = tctxt.m.b8[56];
	p_ctxt->m.b8[60] = tctxt.m.b8[63]; p_ctxt->m.b8[61] = tctxt.m.b8[62];
	p_ctxt->m.b8[62] = tctxt.m.b8[61]; p_ctxt->m.b8[63] = tctxt.m.b8[60];
#endif

	a = H(0); b = H(1); c = H(2); d = H(3); e = H(4);

	for (t = 0; t < 20; t++) {
		s = t & 0x0f;
		if (t >= 16) {
			W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
		}
		tmp = S(5, a) + F0(b, c, d) + e + W(s) + K(t);
		e = d; d = c; c = S(30, b); b = a; a = tmp;
	}
	for (t = 20; t < 40; t++) {
		s = t & 0x0f;
		W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
		tmp = S(5, a) + F1(b, c, d) + e + W(s) + K(t);
		e = d; d = c; c = S(30, b); b = a; a = tmp;
	}
	for (t = 40; t < 60; t++) {
		s = t & 0x0f;
		W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
		tmp = S(5, a) + F2(b, c, d) + e + W(s) + K(t);
		e = d; d = c; c = S(30, b); b = a; a = tmp;
	}
	for (t = 60; t < 80; t++) {
		s = t & 0x0f;
		W(s) = S(1, W((s+13) & 0x0f) ^ W((s+8) & 0x0f) ^ W((s+2) & 0x0f) ^ W(s));
		tmp = S(5, a) + F3(b, c, d) + e + W(s) + K(t);
		e = d; d = c; c = S(30, b); b = a; a = tmp;
	}

	H(0) = H(0) + a;
	H(1) = H(1) + b;
	H(2) = H(2) + c;
	H(3) = H(3) + d;
	H(4) = H(4) + e;

	memset(&p_ctxt->m.b8[0], 0, 64);
}

/*------------------------------------------------------------*/

void cl_sha1_init(struct cl_sha1_ctxt *p_ctxt)
{
	memset(p_ctxt, 0, sizeof(struct cl_sha1_ctxt));
	H(0) = 0x67452301;
	H(1) = 0xefcdab89;
	H(2) = 0x98badcfe;
	H(3) = 0x10325476;
	H(4) = 0xc3d2e1f0;
}

static void cl_sha1_pad(struct cl_sha1_ctxt *p_ctxt)
{
	size_t padlen;		/*pad length in bytes*/
	size_t padstart;

	PUTPAD(0x80);

	padstart = COUNT % 64;
	padlen = 64 - padstart;
	if (padlen < 8) {
		memset(&p_ctxt->m.b8[padstart], 0, padlen);
		COUNT += padlen;
		COUNT %= 64;
		cl_sha1_step(p_ctxt);
		padstart = COUNT % 64;	/* should be 0 */
		padlen = 64 - padstart;	/* should be 64 */
	}
	memset(&p_ctxt->m.b8[padstart], 0, padlen - 8);
	COUNT += (padlen - 8);
	COUNT %= 64;
#if BYTE_ORDER == BIG_ENDIAN
	PUTPAD(p_ctxt->c.b8[0]); PUTPAD(p_ctxt->c.b8[1]);
	PUTPAD(p_ctxt->c.b8[2]); PUTPAD(p_ctxt->c.b8[3]);
	PUTPAD(p_ctxt->c.b8[4]); PUTPAD(p_ctxt->c.b8[5]);
	PUTPAD(p_ctxt->c.b8[6]); PUTPAD(p_ctxt->c.b8[7]);
#else
	PUTPAD(p_ctxt->c.b8[7]); PUTPAD(p_ctxt->c.b8[6]);
	PUTPAD(p_ctxt->c.b8[5]); PUTPAD(p_ctxt->c.b8[4]);
	PUTPAD(p_ctxt->c.b8[3]); PUTPAD(p_ctxt->c.b8[2]);
	PUTPAD(p_ctxt->c.b8[1]); PUTPAD(p_ctxt->c.b8[0]);
#endif
}

void cl_sha1_update(struct cl_sha1_ctxt *p_ctxt, const uint8_t *buffer, size_t len)
{
	size_t gaplen;
	size_t gapstart;
	size_t off;
	size_t copysiz;

	off = 0;

	while (off < len) {
		gapstart = COUNT % 64;
		gaplen = 64 - gapstart;

		copysiz = (gaplen < len - off) ? gaplen : len - off;
		memcpy(&p_ctxt->m.b8[gapstart], &buffer[off], copysiz);
		COUNT += copysiz;
		COUNT %= 64;
		p_ctxt->c.b64[0] += copysiz * 8;
		if (COUNT % 64 == 0)
			cl_sha1_step(p_ctxt);
		off += copysiz;
	}
}

void cl_sha1_digest(struct cl_sha1_ctxt *p_ctxt, uint8_t result[CL_SHA1_RESULTLEN])
{
	uint8_t *digest;

	digest = (uint8_t *)result;
	cl_sha1_pad(p_ctxt);
#if BYTE_ORDER == BIG_ENDIAN
	memcpy(digest, &p_ctxt->h.b8[0], CL_SHA1_RESULTLEN);
#else
	digest[0] = p_ctxt->h.b8[3]; digest[1] = p_ctxt->h.b8[2];
	digest[2] = p_ctxt->h.b8[1]; digest[3] = p_ctxt->h.b8[0];
	digest[4] = p_ctxt->h.b8[7]; digest[5] = p_ctxt->h.b8[6];
	digest[6] = p_ctxt->h.b8[5]; digest[7] = p_ctxt->h.b8[4];
	digest[8] = p_ctxt->h.b8[11]; digest[9] = p_ctxt->h.b8[10];
	digest[10] = p_ctxt->h.b8[9]; digest[11] = p_ctxt->h.b8[8];
	digest[12] = p_ctxt->h.b8[15]; digest[13] = p_ctxt->h.b8[14];
	digest[14] = p_ctxt->h.b8[13]; digest[15] = p_ctxt->h.b8[12];
	digest[16] = p_ctxt->h.b8[19]; digest[17] = p_ctxt->h.b8[18];
	digest[18] = p_ctxt->h.b8[17]; digest[19] = p_ctxt->h.b8[16];
#endif
}

#endif /*unsupported*/
