/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_log_t.
 * This object represents the log file.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_LOG_C
#include <opensm/osm_log.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <opensm/osm_helper.h>

static int log_exit_count = 0;

#ifndef __WIN__
#include <sys/time.h>
#include <unistd.h>
#include <complib/cl_timer.h>

static const char *month_str[] = {
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec"
};

static cl_map_item_t demi_item;

#else
void OsmReportState(IN const char *p_str);
#endif				/* ndef __WIN__ */

static void truncate_log_file(osm_log_t * p_log)
{
	fprintf(stderr, "truncating log file\n");
	osm_truncate_file(p_log->out_port, &p_log->count);
}

void osm_log(IN osm_log_t * p_log, IN osm_log_level_t verbosity,
	     IN const char *p_str, ...)
{
	char buffer[LOG_ENTRY_SIZE_MAX];
	va_list args;
	int ret;
#ifdef __WIN__
	SYSTEMTIME st;
	uint32_t pid = GetCurrentThreadId();
#else
	struct timeval tv;
	pid_t pid = 0;
	time_t tim;
	struct tm result;
	uint64_t time_usecs;
	uint32_t usecs;
#endif				/* __WIN__ */

	/* If this is a call to syslog - always print it */
	if (!(verbosity & p_log->level) &&
	    !(verbosity & p_log->syslog_level)) {
		return;
	}


	va_start(args, p_str);
#ifndef __WIN__
	if (p_log->log_prefix == NULL)
		vsprintf(buffer, p_str, args);
	else {
		int n = snprintf(buffer, sizeof(buffer), "%s: ", p_log->log_prefix);
		vsprintf(buffer + n, p_str, args);
	}
#else
	if (p_log->log_prefix == NULL)
		_vsnprintf(buffer, 1024, (LPSTR)p_str, args);
	else {
		int n = snprintf(buffer, sizeof(buffer), "%s: ", p_log->log_prefix);
		_vsnprintf(buffer + n, (1024 - n), (LPSTR)p_str, args);
	}
#endif
	va_end(args);

	/* this is a call to the syslog */
	if (p_log->syslog_level & verbosity) {
		syslog(LOG_INFO, "%s\n", buffer);

		/* OSM_LOG_SYS should go to stdout too */
		if ((verbosity & OSM_LOG_SYS) && (p_log->out_port != stdout)) {
			printf("%s\n", buffer);
			fflush(stdout);
		}
#ifdef __WIN__
		OsmReportState(buffer);
#endif				/* __WIN__ */
	}

	if (!(verbosity & (OSM_LOG_SYS | p_log->level))) {
		return;
	}

	/* regular log to default out_port */
	cl_spinlock_acquire(&p_log->lock);

	if (p_log->max_size && p_log->count > p_log->max_size) {
		/* truncate here */
		fprintf(stderr,
			"osm_log: log file exceeds the limit %lu. Truncating.\n",
			p_log->max_size);
		truncate_log_file(p_log);
	}
#ifdef __WIN__
	GetLocalTime(&st);
_retry:
	ret =
	    fprintf(p_log->out_port,
		    "[%02d:%02d:%02d:%03d][%04X] 0x%02x -> %s",
		    st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
		    pid, verbosity, buffer);
#else
	gettimeofday(&tv, NULL);
	/* Convert the time of day into a microsecond timestamp. */
	time_usecs = ((uint64_t) tv.tv_sec * 1000000) + (uint64_t) tv.tv_usec;
	tim = time_usecs / 1000000;
	usecs = time_usecs % 1000000;
	localtime_r(&tim, &result);
	pid = pthread_self();
_retry:
	ret =
	    fprintf(p_log->out_port,
		    "%s %02d %02d:%02d:%02d %06d [%04X] 0x%02x -> %s",
		    (result.tm_mon <
		     12 ? month_str[result.tm_mon] : "???"),
		    result.tm_mday, result.tm_hour, result.tm_min,
		    result.tm_sec, usecs, pid, verbosity, buffer);
#endif

	/*  flush log */
	if (ret > 0 &&
	    (p_log->flush || (verbosity & (OSM_LOG_ERROR | OSM_LOG_SYS)))
	    && fflush(p_log->out_port) < 0)
		ret = -1;

	if (ret >= 0) {
		log_exit_count = 0;
		p_log->count += ret;
	} else if (log_exit_count < 3) {
		log_exit_count++;
		if (errno == ENOSPC && p_log->max_size) {
			fprintf(stderr,
				"osm_log: write failed: %s. Truncating log file.\n",
				strerror(errno));
			truncate_log_file(p_log);
			goto _retry;
		}
		fprintf(stderr, "osm_log: write failed: %s\n", strerror(errno));
	}

	cl_spinlock_release(&p_log->lock);
}

void osm_log_v2(IN osm_log_t * p_log, IN osm_log_level_t verbosity,
		IN const int file_id, IN const char *p_str, ...)
{
	char buffer[LOG_ENTRY_SIZE_MAX];
	va_list args;
	int ret;
#ifdef __WIN__
	SYSTEMTIME st;
	uint32_t pid = GetCurrentThreadId();
#else
	struct timeval tv;
	pid_t pid = 0;
	time_t tim;
	struct tm result;
	uint64_t time_usecs;
	uint32_t usecs;
#endif				/* __WIN__ */

	/* If this is a call to syslog - always print it */
	if (!(verbosity & p_log->level) &&
	    !(verbosity & p_log->syslog_level)) {
		if (!(verbosity & p_log->per_mod_log_tbl[file_id]))
			if (!osm_log_is_admin_pid(p_log))
				return;
	}

	va_start(args, p_str);
#ifndef __WIN__
	if (p_log->log_prefix == NULL)
		vsprintf(buffer, p_str, args);
	else {
		int n = snprintf(buffer, sizeof(buffer), "%s: ", p_log->log_prefix);
		vsprintf(buffer + n, p_str, args);
	}
#else
	if (p_log->log_prefix == NULL)
		_vsnprintf(buffer, 1024, (LPSTR)p_str, args);
	else {
		int n = snprintf(buffer, sizeof(buffer), "%s: ", p_log->log_prefix);
		_vsnprintf(buffer + n, (1024 - n), (LPSTR)p_str, args);
	}
#endif
	va_end(args);

	/* this is a call to the syslog */
	if (p_log->syslog_level & verbosity) {
		syslog(LOG_INFO, "%s\n", buffer);

		/* OSM_LOG_SYS should go to stdout too */
		if ((verbosity & OSM_LOG_SYS) && (p_log->out_port != stdout)) {
			printf("%s\n", buffer);
			fflush(stdout);
		}
#ifdef __WIN__
		OsmReportState(buffer);
#endif				/* __WIN__ */
	}

	/* messages with OSM_LOG_SYS will be written to the log file  */
	if (!(verbosity & (OSM_LOG_SYS | p_log->level)) &&
	      !(verbosity & (OSM_LOG_SYS | p_log->per_mod_log_tbl[file_id])))
			if (!osm_log_is_admin_pid(p_log))
				return;

	/* regular log to default out_port */
	cl_spinlock_acquire(&p_log->lock);

	if (p_log->max_size && p_log->count > p_log->max_size) {
		/* truncate here */
		fprintf(stderr,
			"osm_log: log file exceeds the limit %lu. Truncating.\n",
			p_log->max_size);
		truncate_log_file(p_log);
	}
#ifdef __WIN__
	GetLocalTime(&st);
_retry:
	ret =
	    fprintf(p_log->out_port,
		    "[%02d:%02d:%02d:%03d][%04X] 0x%02x -> %s",
		    st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
		    pid, verbosity, buffer);
#else
	gettimeofday(&tv, NULL);
	/* Convert the time of day into a microsecond timestamp. */
	time_usecs = ((uint64_t) tv.tv_sec * 1000000) + (uint64_t) tv.tv_usec;
	tim = time_usecs / 1000000;
	usecs = time_usecs % 1000000;
	localtime_r(&tim, &result);
	pid = pthread_self();
_retry:
	ret =
	    fprintf(p_log->out_port,
		    "%s %02d %02d:%02d:%02d %06d [%04X] 0x%02x -> %s",
		    (result.tm_mon <
		     12 ? month_str[result.tm_mon] : "???"),
		    result.tm_mday, result.tm_hour, result.tm_min,
		    result.tm_sec, usecs, pid, verbosity, buffer);
#endif

	/*  flush log */
	if (ret > 0 &&
	    (p_log->flush || (verbosity & (OSM_LOG_ERROR | OSM_LOG_SYS)))
	    && fflush(p_log->out_port) < 0)
		ret = -1;

	if (ret >= 0) {
		log_exit_count = 0;
		p_log->count += ret;
	} else if (log_exit_count < 3) {
		log_exit_count++;
		if (errno == ENOSPC && p_log->max_size) {
			fprintf(stderr,
				"osm_log: write failed: %s. Truncating log file.\n",
				strerror(errno));
			truncate_log_file(p_log);
			goto _retry;
		}
		fprintf(stderr, "osm_log: write failed: %s\n", strerror(errno));
	}

	cl_spinlock_release(&p_log->lock);
}

void osm_log_raw(IN osm_log_t * p_log, IN osm_log_level_t verbosity,
		 IN const char *p_buf)
{
	if (p_log->level & verbosity) {
		cl_spinlock_acquire(&p_log->lock);
		printf("%s", p_buf);
		cl_spinlock_release(&p_log->lock);

		/*
		   Flush log on errors too.
		 */
		if (p_log->flush || (verbosity & OSM_LOG_ERROR))
			fflush(stdout);
	}
}

void osm_log_msg_box(IN osm_log_t * log, osm_log_level_t level,
		     const char *func_name, const char *msg)
{
#define MSG_BOX_LENGTH 66
	char buf[MSG_BOX_LENGTH + 1];
	int i, n;

	if (!osm_log_is_active(log, level))
		return;

	n = (MSG_BOX_LENGTH - strlen(msg)) / 2 - 1;
	if (n < 0)
		n = 0;
	for (i = 0; i < n; i++)
		sprintf(buf + i, "*");
	n += snprintf(buf + n, sizeof(buf) - n, " %s ", msg);
	for (i = n; i < MSG_BOX_LENGTH; i++)
		buf[i] = '*';
	buf[i] = '\0';

	osm_log(log, level, "%s:\n\n\n"
		"*********************************************"
		"*********************\n%s\n"
		"*********************************************"
		"*********************\n\n\n", func_name, buf);
}

void osm_log_msg_box_v2(IN osm_log_t * log, osm_log_level_t level,
			const int file_id, const char *func_name,
			const char *msg)
{
#define MSG_BOX_LENGTH 66
	char buf[MSG_BOX_LENGTH + 1];
	int i, n;

	if (!osm_log_is_active_v2(log, level, file_id))
		return;

	n = (MSG_BOX_LENGTH - strlen(msg)) / 2 - 1;
	if (n < 0)
		n = 0;
	for (i = 0; i < n; i++)
		sprintf(buf + i, "*");
	n += snprintf(buf + n, sizeof(buf) - n, " %s ", msg);
	for (i = n; i < MSG_BOX_LENGTH; i++)
		buf[i] = '*';
	buf[i] = '\0';

	osm_log_v2(log, level, file_id, "%s:\n\n\n"
		   "*********************************************"
		   "*********************\n%s\n"
		   "*********************************************"
		   "*********************\n\n\n", func_name, buf);
}

boolean_t osm_is_debug(void)
{
#if defined( _DEBUG_ )
	return TRUE;
#else
	return FALSE;
#endif				/* defined( _DEBUG_ ) */
}

static int open_out_port(IN osm_log_t * p_log)
{
	struct stat st;

	if (p_log->accum_log_file)
		p_log->out_port = fopen(p_log->log_file_name, "a+");
	else
		p_log->out_port = fopen(p_log->log_file_name, "w+");

	if (!p_log->out_port) {
		syslog(LOG_CRIT, "Cannot open file \'%s\' for %s: %s\n",
		       p_log->log_file_name,
		       p_log->accum_log_file ? "appending" : "writing",
		       strerror(errno));
		fprintf(stderr, "Cannot open file \'%s\': %s\n",
			p_log->log_file_name, strerror(errno));
		return -1;
	}

	if (fstat(fileno(p_log->out_port), &st) == 0)
		p_log->count = st.st_size;

	syslog(LOG_NOTICE, "%s log file opened\n", p_log->log_file_name);

	if (p_log->daemon) {
		dup2(fileno(p_log->out_port), 0);
		dup2(fileno(p_log->out_port), 1);
		dup2(fileno(p_log->out_port), 2);
	}

	return 0;
}

int osm_log_reopen_file(osm_log_t * p_log)
{
	int ret;

	if (p_log->out_port == stdout || p_log->out_port == stderr)
		return 0;
	cl_spinlock_acquire(&p_log->lock);
	fclose(p_log->out_port);
	ret = open_out_port(p_log);
	cl_spinlock_release(&p_log->lock);
	return ret;
}

static inline osm_log_t * osm_log_assert_getset(osm_log_t * p)
{
	static osm_log_t * p_log = NULL;
	if (p)
		p_log = p;
	return p_log;
}

ib_api_status_t osm_log_init_v3(IN osm_log_t * const p_log,
				IN const boolean_t flush,
				IN const uint8_t log_flags,
				IN const uint8_t syslog_log_flags,
				IN const char *log_file,
				IN const unsigned long max_size,
				IN const boolean_t accum_log_file)
{
	p_log->level = log_flags;
	p_log->syslog_level = syslog_log_flags | OSM_LOG_SYS;
	p_log->flush = flush;
	p_log->count = 0;
	p_log->max_size = max_size << 20; /* convert size in MB to bytes */
	p_log->accum_log_file = accum_log_file;
	p_log->log_file_name = (char *)log_file;
	p_log->bypass_policy_parsed = FALSE;
	memset(p_log->per_mod_log_tbl, 0, sizeof(p_log->per_mod_log_tbl));

	openlog("OpenSM", LOG_CONS | LOG_PID, LOG_USER);

	if (log_file == NULL || !strcmp(log_file, "-") ||
	    !strcmp(log_file, "stdout"))
		p_log->out_port = stdout;
	else if (!strcmp(log_file, "stderr"))
		p_log->out_port = stderr;
	else if (open_out_port(p_log))
		return IB_ERROR;

	if (cl_spinlock_init(&p_log->lock) != CL_SUCCESS)
		return IB_ERROR;

	osm_log_assert_getset(p_log);

	return IB_SUCCESS;
}

ib_api_status_t osm_log_init_v2(IN osm_log_t * const p_log,
				IN const boolean_t flush,
				IN const uint8_t log_flags,
				IN const char *log_file,
				IN const unsigned long max_size,
				IN const boolean_t accum_log_file)
{
	return osm_log_init_v3(p_log, flush, log_flags, OSM_LOG_SYS,
			       log_file, max_size, accum_log_file);

}

ib_api_status_t osm_log_init(IN osm_log_t * p_log, IN boolean_t flush,
			     IN uint8_t log_flags, IN const char *log_file,
			     IN boolean_t accum_log_file)
{
	return osm_log_init_v3(p_log, flush, log_flags, OSM_LOG_SYS,
			       log_file, 0, accum_log_file);
}

osm_log_level_t osm_get_log_per_module(IN osm_log_t * p_log,
				       IN const int file_id)
{
	return p_log->per_mod_log_tbl[file_id];
}

void osm_set_log_per_module(IN osm_log_t * p_log, IN const int file_id,
			    IN osm_log_level_t level)
{
	p_log->per_mod_log_tbl[file_id] = level;
}

void osm_reset_log_per_module(IN osm_log_t * p_log)
{
	memset(p_log->per_mod_log_tbl, 0, sizeof(p_log->per_mod_log_tbl));
}

static void log_to_stream_v(IN FILE * out_stream,
			    IN const char *p_str, IN va_list args);

void osm_log_assert(IN const char *p_str, ...)
{
	va_list args;
	osm_log_t * p_log = osm_log_assert_getset(NULL);
	if (!p_log) {
		fprintf(stderr, "osm_log_assert: log handle is empty\n");
		return;
	}

	va_start(args, p_str);
	log_to_stream_v(p_log->out_port, p_str, args);
	va_end(args);
}

void osm_log_to_stream(IN FILE * out_stream, IN const char *p_str, ...)
{
	va_list args;
	if (!out_stream)
		return;
	va_start(args, p_str);
	log_to_stream_v(out_stream, p_str, args);
	va_end(args);
}

/*
 * Some notes about this function:
 *  - No locking on log mutex - we have no idea what state we're in,
 *    so we'd better not risk getting into deadlock with an already
 *    locked or even corrupted mutex.
 *  - No log file truncating - we'll leave this to normal logging.
 *  - No retrying if writing to log failed.
 *  - Not using log prefix.
 *  - Flushing all the open streams at the end of the function.
 */
static void log_to_stream_v(IN FILE * out_stream,
			    IN const char *p_str, IN va_list args)
{
	char *buffer;
	int ret;
#ifdef __WIN__
	SYSTEMTIME st;
	uint32_t pid = GetCurrentThreadId();
#else
	struct timeval tv;
	pid_t pid = 0;
	time_t tim;
	struct tm result;
	uint64_t time_usecs;
	uint32_t usecs;
#endif

	if (!out_stream)
		return;

	buffer = calloc(1, LOG_LARGE_BUF_SIZE_MAX);
	if (!buffer)
		return;

#ifndef __WIN__
	vsprintf(buffer, p_str, args);
#else
	_vsnprintf(buffer, 1024, (LPSTR)p_str, args);
#endif

	/*
	 * Before printing the message to the OSM log,
	 * print it to syslog and stderr
	 */
	syslog(LOG_ERR, "%s", buffer);
	if (out_stream != stderr)
		fprintf(stderr, "%s", buffer);

#ifdef __WIN__
	OsmReportState(buffer);
#endif


#ifdef __WIN__
	GetLocalTime(&st);

	ret = fprintf(out_stream,
		      "[%02d:%02d:%02d:%03d][%04X] 0x%02x -> %s",
		      st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
		      pid, 0xFF, buffer);
#else
	gettimeofday(&tv, NULL);
	/* Convert the time of day into a microsecond timestamp. */
	time_usecs = ((uint64_t) tv.tv_sec * 1000000) + (uint64_t) tv.tv_usec;
	tim = time_usecs / 1000000;
	usecs = time_usecs % 1000000;
	localtime_r(&tim, &result);
	pid = pthread_self();

	ret = fprintf(out_stream,
		      "%s %02d %02d:%02d:%02d %06d [%04X] 0x%02x -> %s",
		      (result.tm_mon <
		      12 ? month_str[result.tm_mon] : "???"),
		      result.tm_mday, result.tm_hour, result.tm_min,
		      result.tm_sec, usecs, pid, 0xFF, buffer);
#endif

	if (ret <= 0) {
		fprintf(stderr, "osm_log_to_stream: write failed: %s.\n",
			strerror(errno));
	}

	/*
	 * Flush everything, don't check the result - don't really
	 * care if it worked well. Note that fflush() is a locking
	 * function. We could use here fflush_unlocked() instead,
	 * but it is not thread-safe, and I'm not sure whether it
	 * would really be a better option.
	 */
	fflush(NULL);
	free(buffer);
}

void osm_log_register_admin_pid(osm_log_t *p_log)
{
	cl_spinlock_acquire((cl_spinlock_t * const)&p_log->lock);
	cl_qmap_insert(&p_log->admin_pid, pthread_self(), &demi_item);
	cl_spinlock_release((cl_spinlock_t * const)&p_log->lock);
}

void osm_log_unregister_admin_pid(osm_log_t *p_log)
{
	cl_spinlock_acquire((cl_spinlock_t * const)&p_log->lock);
	cl_qmap_remove(&p_log->admin_pid, pthread_self());
	cl_spinlock_release((cl_spinlock_t * const)&p_log->lock);
}

int osm_log_is_admin_pid(const osm_log_t *p_log)
{
	int res = 1;

	if (!p_log->bypass_policy_parsed) {
		res = 0;
		goto EXIT;
	}

	cl_spinlock_acquire((cl_spinlock_t * const)&p_log->lock);
	if (!cl_qmap_count(&p_log->admin_pid)) {
		res = 0;
		goto UNLOCK_AND_EXIT;
	}
	if (cl_qmap_get(&p_log->admin_pid, pthread_self()) ==
	    cl_qmap_end(&p_log->admin_pid))
		res = 0;
UNLOCK_AND_EXIT:
	cl_spinlock_release((cl_spinlock_t * const)&p_log->lock);
EXIT:
	return res;
}
