/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Event Manager
 *      This manager is responsible to handle events which are reported by the
 *      SM.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ACT_MGR_C
#include <opensm/osm_activity_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>

#define MAX_BUF_LEN 1000
#define ACTIVITY_DUMP_FILE "opensm-activity.dump"

static char *subjects[SUBJECTS_NUM] = { "mc", "prtn", "virt" , "routing", "sa_req_exceeding_counters"};

extern void remove_whitespaces(char *str);

char* activity_names[NUM_OF_EVENTS] = { "mcm_member",
					"mcg_change",
					"prtn_guid_add",
					"prtn_guid_del",
					"prtn_create",
					"prtn_delete",
					"port_virt_discover",
					"vport_state_change",
					"mcg_tree_calc",
					"routing_succeed",
					"routing_failed",
					"ucast_cache_invalidated",
					"ucast_cache_routing_done",
					"port_exceeded_max_mcm",
					"port_exceeded_max_svcr",
					"port_exceeded_max_event_subs" };

char* prtn_del_str[3] = { "Empty",
			  "Duplicate",
			  "Shutdown" };

char *group_action_str[2] = { "Create", "Delete" };

char *state_str[5] = { "",
		       "Down",
		       "Init",
		       "Armed",
		       "Active" };

typedef struct mcm_member_rec {
	ib_net16_t mlid;
	char* mgid;
	ib_net64_t port_guid;
	osm_act_mgr_mc_state_enum join_state;
} mcm_member_rec_t;

typedef struct mcg_change_rec {
	ib_net16_t mlid;
	char* mgid;
	osm_activity_mgr_change_enum change;
} mcg_change_rec_t;

typedef struct prtn_rec {
	ib_net16_t pkey;
	char *name;
	osm_activity_mgr_prtn_del_reason_enum del_reason;
} prtn_rec_t;

typedef struct prtn_member_rec {
	ib_net16_t pkey;
	ib_net64_t guid;
	uint16_t block_index;
	uint8_t pkey_index;
} prtn_member_rec_t;

typedef struct virt_rec {
	ib_net64_t port_guid;
	ib_net16_t top_index;
	ib_net64_t vport_guid;
	uint16_t vport_index;
	ib_net64_t vnode_guid;
	osm_activity_mgr_state_enum state;
} virt_rec_t;

void enable_activity_subjects(osm_activity_mgr_t *p_mgr, struct osm_sm *p_sm)
{
	char *token;
	char *context = NULL;
	boolean_t found;
	int i;
	char *subjects_list = p_sm->p_subn->opt.activity_report_subjects;

	OSM_LOG_ENTER(p_mgr->p_log);

	remove_whitespaces(subjects_list);

	for (i = 0 ; i < SUBJECTS_NUM ; i++)
		p_mgr->enabled_subjects[i] = FALSE;

	if (strcmp(subjects_list, OSM_ACTIVITY_SUBJECTS_NONE)) {
		if (strcmp(subjects_list, OSM_ACTIVITY_SUBJECTS_ALL)) {
			/* Parse list */
			token = strtok_r(subjects_list, ",", &context);
			while (token) {
				i = 0;
				found = FALSE;
				while ((i < SUBJECTS_NUM) && !found) {
					if (strcmp(token, subjects[i])) {
						i++;
					} else {
						p_mgr->enabled_subjects[i] = TRUE;
						p_mgr->enabled = TRUE;
						found = TRUE;
					}
				}
				if (!found)
					OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
						"ERR 9A1A: Unrecognized subject"
						" configured for activity "
						"manager (%s)\n", token);

				token = strtok_r(NULL, ",", &context);
			}
		} else {
			/* 'all' subjects enabled */
			p_mgr->enabled = TRUE;
			for (i = 0 ; i < SUBJECTS_NUM ; i++)
				p_mgr->enabled_subjects[i] = TRUE;
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}


static void dump_buf(osm_activity_mgr_t *p_mgr, char *buf)
{
	time_t timer;
	char time_buf[MAX_BUF_LEN + 26];
	struct tm* tm_info;

	time(&timer);
	tm_info = localtime(&timer);

	strftime(time_buf, 26, "%Y-%m-%d %H:%M:%S ", tm_info);
	fprintf(p_mgr->file, "%s %s", time_buf, buf);
	fflush(p_mgr->file);
}

static boolean_t set_event_buf(osm_activity_mgr_t *p_mgr, char *buf,
			       event_data_t *p_data)
{
	osm_activity_mgr_event_id_enum event_id = p_data->event_id;
	boolean_t res = TRUE;

	memset(buf, 0, MAX_BUF_LEN);

	switch (event_id) {
	case OSM_ACT_MGR_MCM_MEMBER:
		sprintf(buf, "%d,%s,%d,0x%" PRIx16 ",%s,0x%" PRIx64 ",%d,%s\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh16(p_data->mlid), p_data->mgid,
			cl_ntoh64(p_data->port_guid), (int)p_data->join_state,
			p_data->join_state > 0 ? "Join" : "Leave");
		break;
	case OSM_ACT_MGR_MCG_CHANGE:
		sprintf(buf, "%d,%s,%d,0x%" PRIx16 ",%s,%s\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh16(p_data->mlid), p_data->mgid,
			group_action_str[p_data->change]);
		break;
	case OSM_ACT_MGR_PRTN_ADD_GUID:
		sprintf(buf, "%d,%s,%d,0x%" PRIx64 ",0x%" PRIx16 ",%u,%u\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh64(p_data->port_guid), cl_ntoh16(p_data->pkey),
			p_data->block_index, p_data->pkey_index);
		break;
	case OSM_ACT_MGR_PRTN_DEL_GUID:
		sprintf(buf, "%d,%s,%d,0x%" PRIx64 ",0x%" PRIx16 ",%u,%u\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh64(p_data->port_guid), cl_ntoh16(p_data->pkey),
			p_data->block_index, p_data->pkey_index);
		break;
	case OSM_ACT_MGR_PRTN_CREATE:
		sprintf(buf, "%d,%s,%d,0x%" PRIx16 ",%s\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh16(p_data->pkey), p_data->name);
		break;
	case OSM_ACT_MGR_PRTN_DELETE:
		sprintf(buf, "%d,%s,%d,0x%" PRIx16 ",%s\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh16(p_data->pkey),
			prtn_del_str[p_data->del_reason]);
		break;
	case OSM_ACT_MGR_VIRT_DISCOVER:
		sprintf(buf, "%d,%s,%d,0x%" PRIx64 ",%" PRIu16 "\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh64(p_data->port_guid), cl_ntoh16(p_data->top_index));
		break;
	case OSM_ACT_MGR_VPORT_STATE_CHANGE:
		sprintf(buf, "%d,%s,%d,0x%" PRIx64 ",0x%" PRIx64 ",%" PRIu16 ",0x%"
			PRIx64 ",%s\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh64(p_data->port_guid), cl_ntoh64(p_data->vport_guid),
			p_data->vport_index, cl_ntoh64(p_data->vnode_guid),
			state_str[p_data->state]);
		break;
	case OSM_ACT_MGR_MCG_TREE_CALC:
			sprintf(buf, "%d,%s,%d,0x%" PRIx16 "\n",
				event_id, activity_names[event_id], p_mgr->pid,
				cl_ntoh16(p_data->mlid));
		break;
	case OSM_ACT_MGR_ROUTE_SUCCESS:
	case OSM_ACT_MGR_ROUTE_FAIL:
			sprintf(buf, "%d,%s,%d,%s\n",
				event_id, activity_names[event_id], p_mgr->pid,
				p_data->name);
		break;
	case OSM_ACT_MGR_UCAST_CACHE_INVAL:
	case OSM_ACT_MGR_UCAST_CACHE_USED:
			sprintf(buf, "%d,%s,%d\n",
				event_id, activity_names[event_id], p_mgr->pid);
		break;
	case OSM_ACT_MGR_MCM_EXCEEDED_MAX:
		sprintf(buf, "%d,%s,%d,0x%" PRIx16 ",%s,0x%" PRIx64 ",%u\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh16(p_data->mlid), p_data->mgid,
			cl_ntoh64(p_data->port_guid),
			p_data->sa_req_exceeding_counter);
		break;
	case OSM_ACT_MGR_SVCR_EXCEEDED_MAX:
	case OSM_ACT_MGR_EVENT_SUBS_EXCEEDED_MAX:
		sprintf(buf, "%d,%s,%d,0x%" PRIx64 ",%u\n",
			event_id, activity_names[event_id], p_mgr->pid,
			cl_ntoh64(p_data->port_guid),
			p_data->sa_req_exceeding_counter);
		break;
	default:
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A03: "
			"Unknown event id %d, Process failed\n", event_id);
		res = FALSE;
	}

	return res;
}

static void process_event(void *context, void *data)
{
	event_data_t *p_event_data = (event_data_t*)data;
	osm_sm_t *p_sm = context;
	osm_activity_mgr_t *p_mgr = &p_sm->activity_mgr;
	char buf[MAX_BUF_LEN];

	OSM_LOG_ENTER(p_mgr->p_log);

	if (set_event_buf(p_mgr, buf, p_event_data))
		dump_buf(p_mgr, buf);

	free(p_event_data);
	OSM_LOG_EXIT(p_mgr->p_log);
}

ib_api_status_t osm_activity_mgr_init(osm_activity_mgr_t *p_mgr, struct osm_sm *p_sm)
{
	ib_api_status_t status;
	char path[1024];
	int event_id;

	OSM_LOG_ENTER(p_sm->p_log);

	p_mgr->p_log = p_sm->p_log;
	p_mgr->p_sm_state = &p_sm->p_subn->sm_state;

	status = cl_disp_init(&p_mgr->disp, 1, "sm_events");
	if (status != IB_SUCCESS)
		goto Exit;

	for (event_id = 0 ; event_id < NUM_OF_EVENTS ; event_id++) {
		p_mgr->handlers[event_id] = cl_disp_register(&p_mgr->disp,
							     event_id,
							     process_event,
							     p_sm);

		if (p_mgr->handlers[event_id] == CL_DISP_INVALID_HANDLE) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A02: "
				"Failed to register handler %d to dispatcher\n",
				event_id);
			status = CL_INSUFFICIENT_RESOURCES;
			goto Exit;
		}
	}

	/* Set dump file*/
	snprintf(path, sizeof(path), "%s/%s", p_sm->p_subn->opt.dump_files_dir,
		 ACTIVITY_DUMP_FILE);
	p_mgr->file = fopen(path, "a");
	if (!p_mgr->file) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A19: "
			"Cannot create file \'%s\': %s\n",
			path, strerror(errno));
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	p_mgr->pid = getpid();
Exit:
	OSM_LOG_EXIT(p_sm->p_log);
	return status;
}

void osm_activity_mgr_destroy(osm_activity_mgr_t *p_mgr)
{
	int event_id;

	for (event_id = 0 ; event_id < NUM_OF_EVENTS ; event_id++)
		cl_disp_unregister(p_mgr->handlers[event_id]);

	cl_disp_shutdown(&p_mgr->disp);
	cl_disp_destroy(&p_mgr->disp);
	fclose(p_mgr->file);
}

static boolean_t is_subj_enabled_for_event_id(osm_activity_mgr_t *p_mgr,
					      osm_activity_mgr_event_id_enum
								event_id)
{
	boolean_t res = FALSE;

	if (osm_exit_flag || *p_mgr->p_sm_state != IB_SMINFO_STATE_MASTER)
		goto Exit;

	switch (event_id) {
	case OSM_ACT_MGR_MCM_MEMBER:
	case OSM_ACT_MGR_MCG_CHANGE:
		res = osm_act_mgr_subj_enabled(p_mgr, MC_SUBJ_ID);
		break;
	case OSM_ACT_MGR_PRTN_ADD_GUID:
	case OSM_ACT_MGR_PRTN_DEL_GUID:
	case OSM_ACT_MGR_PRTN_CREATE:
	case OSM_ACT_MGR_PRTN_DELETE:
		res = osm_act_mgr_subj_enabled(p_mgr, PRTN_SUBJ_ID);
		break;
	case OSM_ACT_MGR_VIRT_DISCOVER:
	case OSM_ACT_MGR_VPORT_STATE_CHANGE:
		res = osm_act_mgr_subj_enabled(p_mgr, VIRT_SUBJ_ID);
		break;
	case OSM_ACT_MGR_MCG_TREE_CALC:
		res = (osm_act_mgr_subj_enabled(p_mgr, MC_SUBJ_ID) ||
			osm_act_mgr_subj_enabled(p_mgr, ROUTE_SUBJ_ID));
		break;
	case OSM_ACT_MGR_ROUTE_SUCCESS:
	case OSM_ACT_MGR_ROUTE_FAIL:
	case OSM_ACT_MGR_UCAST_CACHE_INVAL:
	case OSM_ACT_MGR_UCAST_CACHE_USED:
		res = osm_act_mgr_subj_enabled(p_mgr, ROUTE_SUBJ_ID);
		break;
	case OSM_ACT_MGR_MCM_EXCEEDED_MAX:
	case OSM_ACT_MGR_SVCR_EXCEEDED_MAX:
	case OSM_ACT_MGR_EVENT_SUBS_EXCEEDED_MAX:
		res = osm_act_mgr_subj_enabled(p_mgr, EXCEEDS_MAX_SUBJ_ID);
		break;
	default:
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A04: "
			"Unknown event id %d, Post failed\n", event_id);
	}

Exit:
	return res;
}

void osm_act_mgr_post_event(osm_activity_mgr_t *p_mgr, event_data_t *p_data)
{
	event_data_t *p_event_data;
	osm_activity_mgr_event_id_enum event_id = p_data->event_id;
	cl_status_t cl_status;

	if (!is_subj_enabled_for_event_id(p_mgr, event_id))
		return;

	p_event_data = (event_data_t*)calloc(1, sizeof(event_data_t));
	if (!p_event_data) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A00: "
			"Memory allocation failure while posting event %d\n",
			event_id);
		return;
	}

	*p_event_data = *p_data;

	cl_status = cl_disp_post(p_mgr->handlers[event_id], event_id,
				 p_event_data, NULL, NULL);

	if (cl_status != CL_SUCCESS) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 9A01: "
			"Failed to post	event (event id: %d)\n", event_id);
		free(p_event_data);
	}
}
