/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Implementation of osm_advanced_routing_data_t.
 *
 *	This object is part of the OpenSM family of objects.
 */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ADVANCED_ROUTING_C
#include <opensm/osm_switch.h>
#include <opensm/osm_advanced_routing.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>

#define AR_MAX_GROUPS		4096
#define AR_INVALID_GROUP_ID	0xffff
#define AR_INVALID_GROUP	0xff
#define SUBN_TABLES_UPDATE_AR_STATUS		TRUE
#define SUBN_TABLES_DONT_UPDATE_AR_STATUS	FALSE

/* this is a maximal number of copy groups that
 * switch can copy without copy group mad to timeout,
 * this number is heuristic.
 */
#define AR_MAX_COPY_GROUPS_NUM	1000

#define AR_ROUNDUP_BLOCK(num, block) \
	(((num) + (block) - 1) - (((num) - 1) % (block)))

#define AR_SET_BITMASK(field, bit, value)	\
	do { if (value)				\
		field |= (bit);			\
	     else				\
		field &= ~(bit);		\
	} while(0)

extern int wait_for_pending_transactions(osm_stats_t * stats);

typedef void (*ar_get_next_mad_fn_t)(osm_sm_t *, osm_switch_t *, boolean_t,
				     cl_qlist_t *);

void osm_ar_data_init(osm_advanced_routing_data_t *p_ar_data)
{
	memset(p_ar_data, 0, sizeof(osm_advanced_routing_data_t));
	cl_qlist_init(&p_ar_data->cached_capabilities_list);
}

const char *get_adv_routing_name_from_type(osm_adv_routing_type_enum adv_routing_type)
{
	switch (adv_routing_type) {
	case OSM_ADV_ROUTING_TYPE_AR_TREE:
		return "ar_tree";
	case OSM_ADV_ROUTING_TYPE_AR_LAG:
		return "ar_lag";
	default:
		return "none";
	}
}

osm_adv_routing_type_enum get_adv_routing_type_from_name(const char *ar_name)
{
	if (!strcasecmp(ar_name, "ar_tree"))
		return OSM_ADV_ROUTING_TYPE_AR_TREE;
	else if (!strcasecmp(ar_name, "ar_lag"))
		return OSM_ADV_ROUTING_TYPE_AR_LAG;

	return OSM_ADV_ROUTING_TYPE_NONE;
}

static inline void ar_clear_lft_table(ib_ar_lft_entry_t *p_ar_lft, uint16_t num_lids)
{
	uint16_t lid;
	for (lid = 0; lid < num_lids; lid++) {
		p_ar_lft[lid].state_table = 0x80;
		p_ar_lft[lid].legacy_port = 0xff;
		p_ar_lft[lid].group_number = 0;
	}
}

static inline void ar_clear_table(void *p_table, uint16_t num_elements,
				  size_t element_size, uint16_t block_size,
				  uint8_t init_value)
{
	size_t size = AR_ROUNDUP_BLOCK(num_elements, block_size) * element_size;
	memset(p_table, init_value, size);
}

static int ar_reallocate_lft_table(
		ib_ar_lft_entry_t* (*pp_ar_lft) [AR_MAX_NUM_PLFT],
		uint8_t plft_id,
		uint16_t prev_num_lids,
		uint16_t num_lids, boolean_t clear_all)
{
	ib_ar_lft_entry_t *p_ar_lft;
	size_t size;

	OSM_ASSERT(plft_id < AR_MAX_NUM_PLFT);

	if (!num_lids)
		return 0;

	/* Round up to fill entire block */
	num_lids = AR_ROUNDUP_BLOCK(num_lids, IB_AR_LFT_BLOCK_SIZE);
	if (prev_num_lids < num_lids) {
		size = num_lids * sizeof(ib_ar_lft_entry_t);
		p_ar_lft = (ib_ar_lft_entry_t *)realloc((*pp_ar_lft)[plft_id], size);
		if (!p_ar_lft)
			return 1;
		(*pp_ar_lft)[plft_id] = p_ar_lft;
	}

	/*
	 * Clear the new/all entries according to parameters.
	 * num_lids can only grow in sm
	 */
	ar_clear_lft_table(&((*pp_ar_lft)[plft_id])[clear_all ? 0 : prev_num_lids],
			   num_lids - (clear_all ? 0 : prev_num_lids));

	return 0;
}

static inline int ar_reallocate_table(void **pp_table, uint16_t num_elements,
				      size_t element_size, uint16_t block_size,
				      uint8_t init_value)
{
	void *t_table = NULL;
	size_t size;

	if (!num_elements) {
		free(*pp_table);
		*pp_table = NULL;
		return 0;
	}

	size = AR_ROUNDUP_BLOCK(num_elements, block_size) * element_size;
	t_table = realloc(*pp_table, size);
	if (!t_table)
		return 1;

	memset(t_table, init_value, size);
	*pp_table = t_table;

	return 0;
}

/*
 * Clear AR/RN tables.
 */
void ar_tables_clear(osm_ar_tables_t *p_tables, uint16_t num_groups,
		     uint8_t num_group_subgroups,
		     uint16_t num_strings, uint8_t num_directions,
		     uint8_t num_plfts,
		     uint16_t num_lids, uint8_t num_ports,
		     uint8_t group_init_value)
{
	uint16_t num_subgroups = num_groups * num_group_subgroups;
	uint8_t arlft_num_plfts, plft_idx;

	if (num_subgroups)
		ar_clear_table(p_tables->ar_subgroups, num_subgroups,
			       sizeof(osm_ar_subgroup_t), 2, group_init_value);

	if (num_strings)
		ar_clear_table(p_tables->rn_rcv_strings, num_strings,
			       sizeof(ib_rn_rcv_string_entry_t),
			       IB_RN_RCV_STRING_BLOCK_SIZE,
			       0);
	if (num_directions)
		ar_clear_table(p_tables->rn_direction_string, num_directions,
			       sizeof(ib_net16_t), IB_RN_GEN_STRING_BLOCK_SIZE,
			       0);
	if (num_subgroups)
		ar_clear_table(p_tables->rn_subgroup_direction, num_subgroups,
			       sizeof(uint8_t), IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE,
			       0);
	if (num_plfts) {
		ar_clear_table(p_tables->plft_defs, num_plfts,
			       sizeof(ib_plft_def_element_t),
			       IB_PLFT_DEF_BLOCK_SIZE,
			       0);
		ar_clear_table(p_tables->port_sl_plfts, num_ports,
			       sizeof(ib_port_sl_to_plft_map_entry_t),
			       IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE, 0);
	}

	memset(p_tables->rn_subgroup_priority_flags, 0,
	       sizeof(p_tables->rn_subgroup_priority_flags));
	memset(p_tables->rn_xmit_port_mask, 0,
	       sizeof(p_tables->rn_xmit_port_mask));

	if (num_lids) {
		/* in case no plfts, still need to clear regular alfts table
		 * that sit at index 0
		 */
		arlft_num_plfts = num_plfts ? num_plfts : 1;

		OSM_ASSERT(arlft_num_plfts <= AR_MAX_NUM_PLFT);

		for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx ++)
			ar_clear_lft_table(p_tables->ar_lft[plft_idx], num_lids);
	}
}

inline static void ar_tables_destroy_ar_lft(osm_advanced_routing_data_t *p_ar_data,
					    osm_ar_tables_t *p_tables,
					    uint8_t start_plft_idx)
{
	uint8_t arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
	/*
	 * plft 0 always exists, for regular ar lfts,
	 * might want to free only additional plfts
	 */
	uint8_t plft_idx = start_plft_idx;

	for (; plft_idx < arlft_num_plfts; plft_idx ++) {
		free(p_tables->ar_lft[plft_idx]);
		p_tables->ar_lft[plft_idx] = NULL;
	}
}

/*
 * Reallocate and clear AR/RN tables.
 * This function should be called when related switch capabilities have changed.
 */
static int ar_tables_reinitialize(osm_switch_t *p_sw, osm_ar_table_type_t table_type,
				  uint8_t num_plfts,
				  uint16_t num_lids,
				  uint16_t num_groups, uint8_t num_group_subgroups,
				  uint16_t num_strings, uint8_t num_directions,
				  uint8_t groups_init_value)
{
	uint16_t num_subgroups, num_rn_subgroups;
	uint8_t plft_idx = 0, arlft_num_plfts, plft_idx_start_removal = 0;
	uint8_t num_ports = p_sw->num_ports;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_tables_t *p_tables = table_type == OSM_AR_CALC_TABLE ?
				    &p_ar_data->calc_tables : &p_ar_data->subn_tables;

	num_subgroups = num_groups * num_group_subgroups;
	num_rn_subgroups = (num_strings && num_directions) ? num_subgroups : 0;
	arlft_num_plfts = num_plfts ? num_plfts : 1;

	/* free PLFTs that are no longer needed */
	if (num_plfts < p_ar_data->num_plfts) {
		/* at index 0 regular ar lft */
		plft_idx_start_removal = num_plfts ? num_plfts : 1;
		ar_tables_destroy_ar_lft(p_ar_data, p_tables,
					 plft_idx_start_removal);
	}

	/* Initialize AR LFT tables */
	for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++) {
		if (ar_reallocate_lft_table(&p_tables->ar_lft, plft_idx,
					    0, num_lids, TRUE))
			return 1;
	}

	/* Initialize AR groups table. */
	if (ar_reallocate_table((void **)&p_tables->ar_subgroups, num_subgroups,
				sizeof(osm_ar_subgroup_t), 2, groups_init_value))
		return 1;

	/* Initialize mapping from string to ib_rn_rcv_string_entry_t objects. */
	if (ar_reallocate_table((void **)&p_tables->rn_rcv_strings,
				num_strings, sizeof(ib_rn_rcv_string_entry_t),
				IB_RN_RCV_STRING_BLOCK_SIZE, 0))
		return 1;

	/* Initialize map from direction to generated string. */
	if (ar_reallocate_table((void **)&p_tables->rn_direction_string,
				num_directions, sizeof(ib_net16_t),
				IB_RN_GEN_STRING_BLOCK_SIZE, 0))
		return 1;

	/* Initialize map from subgroup to direction */
	if (ar_reallocate_table((void **)&p_tables->rn_subgroup_direction,
				num_rn_subgroups, sizeof(uint8_t),
				IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE, 0))
		return 1;

	/* Initialize map from from subgroup priority to RN flags. */
	memset(p_tables->rn_subgroup_priority_flags, 0,
	       IB_RN_GEN_BY_SUB_GROUP_PRIORITY_BLOCK_SIZE * sizeof(ib_net32_t));

	/* plft tables */
	if (ar_reallocate_table((void **)&p_tables->plft_defs, num_plfts,
				sizeof(ib_plft_def_element_t),
				IB_PLFT_DEF_BLOCK_SIZE, 0))
		return 1;

	if (ar_reallocate_table((void **)&p_tables->port_sl_plfts, num_ports,
				sizeof(ib_port_sl_to_plft_map_entry_t),
				IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE, 0))
		return 1;

	for (plft_idx = 0; plft_idx < num_plfts; plft_idx ++)
		p_tables->plft_top[plft_idx] = 0;

	return 0;
}

static inline boolean_t ar_is_rn_supported(const ib_ar_info_t *p_ar_info)
{
	return (p_ar_info->enable_features &
		(IB_AR_INFO_FEATURE_MASK_FRN_SUPPORTED |
		 IB_AR_INFO_FEATURE_MASK_ARN_SUPPORTED));
}

static inline boolean_t ar_is_rn_supported_enabled(const ib_ar_info_t *p_ar_info)
{
	return (ar_is_rn_supported(p_ar_info) &&
		(p_ar_info->enable_features & IB_AR_INFO_FEATURE_MASK_RN_XMIT_ENABLED));
}

static inline boolean_t ar_is_ar_enabled(const ib_ar_info_t *p_ar_info)
{
	return (p_ar_info->enable_features &
		(IB_AR_INFO_FEATURE_MASK_AR_ENABLED |
		 IB_AR_INFO_FEATURE_MASK_FR_ENABLED));
}

static inline boolean_t ar_is_group_top_supported(const ib_ar_info_t *p_ar_info)
{
	return ((p_ar_info->ar_nv_version_cap & 0xf) >= 2);
}

/* the StringWidthCap field is in log2 units */
static inline int ar_get_string_width_cap(const ib_ar_info_t * const p_ar_info)
{
	return (1 << ib_ar_info_get_string_width_cap(p_ar_info));
}

/*
 * Allocate and clear routing and routing notification tables according to
 * information of the specified osm_advanced_routing_data_t object.
 *
 * This function should be called after setting switch configuration, and
 * before AR/RN table calculations.
 */
int osm_ar_prepare_sw_tables(osm_subn_t *p_subn,
			     osm_switch_t *p_sw,
			     uint16_t num_lids)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;
	osm_log_t *p_log = &p_subn->p_osm->log;
	uint16_t num_groups = 0;
	uint16_t group_id;
	uint16_t num_strings = 0;
	uint8_t num_group_subgroups = 0;
	uint8_t num_directions = 0;
	uint8_t num_plfts = 0;

	/* TODO: mark switches when allocation failed. */
	/* TODO: handle switches with disabled AR. */

	num_groups = ib_ar_info_get_group_cap_ho(p_ar_info);
	num_group_subgroups = ib_ar_info_get_sub_grps_active(p_ar_info) + 1;
	num_plfts = p_ar_data->req_num_plfts;

	if (ar_is_rn_supported(p_ar_info)) {
		num_strings = ar_get_string_width_cap(p_ar_info);
		num_directions = p_ar_info->dir_num_supp;
	}

	/* If capabilities have changed, reallocate and clear tables */
	if (num_groups != p_ar_data->num_groups ||
	    num_group_subgroups != p_ar_data->num_group_subgroups ||
	    num_strings != p_ar_data->num_strings ||
	    num_directions != p_ar_data->num_directions ||
	    num_plfts != p_ar_data->num_plfts) {
		if (ar_tables_reinitialize(p_sw, OSM_AR_CALC_TABLE, num_plfts, num_lids,
					   num_groups, num_group_subgroups,
					   num_strings, num_directions, 0) ||
		    ar_tables_reinitialize(p_sw, OSM_AR_SUBNET_TABLE, num_plfts, num_lids,
					   num_groups, num_group_subgroups,
					   num_strings, num_directions,
					   AR_INVALID_GROUP)) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7C0A: Failed to allocate memory for"
				" advanced routing capabilities\n");
			return 1;
		}

		/*
		 * Resize adaptive routing groups copy tables.
		 * Hack: The function does a memset of each byte so when
		 * passing AR_INVALID_GROUP_ID(0xffff) every 2 bytes will be equal
		 * to AR_INVALID_GROUP_ID(0xffff) because every byte is 0xff
		 */
		if (ar_reallocate_table((void **)&p_ar_data->ar_group_copy_from_table,
					num_groups,
					sizeof(*p_ar_data->ar_group_copy_from_table),
					1,
					(uint8_t) AR_INVALID_GROUP_ID)) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7C0B: Failed to allocate memory for"
				" advanced routing capabilities\n");
			return 1;
		}

		p_ar_data->num_groups = num_groups;
		p_ar_data->num_group_subgroups = num_group_subgroups;
		p_ar_data->num_strings = num_strings;
		p_ar_data->num_directions = num_directions;
		p_ar_data->num_lids = num_lids;
		p_ar_data->num_plfts = num_plfts;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"sw guid: 0x%" PRIx64 " set num groups %u sub groups %u "
			"num_strings %u num_directions %u num_lids %u num_plfts %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			p_ar_data->num_groups, p_ar_data->num_group_subgroups,
			p_ar_data->num_strings, p_ar_data->num_directions,
			p_ar_data->num_lids, p_ar_data->num_plfts);
		return 0;
	}

	/*
	 * Capabilities did not change, check if number of LIDs has changed,
	 * and change size of LFT tables accordingly.
	 */
	if (num_lids > p_ar_data->num_lids) {
		/* Extend AR LFT tables */
		if (osm_ar_reallocate_lft_tables(p_ar_data, num_lids, TRUE)) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7C0C: Failed to allocate memory for"
				" adaptive routing forwarding tables\n");
			return 1;
		}
	}

	/* Clear AR/RN tables for calculation, init groups with 0 */
	ar_tables_clear(&p_ar_data->calc_tables, num_groups,
			num_group_subgroups, num_strings,
			num_directions,
			num_plfts,
			p_ar_data->num_lids,
			p_sw->num_ports,
			0);

	/* Reset adaptive routing groups copy tables */
	for (group_id = 0; group_id < num_groups; group_id++) {
		p_ar_data->ar_group_copy_from_table[group_id] =
		    AR_INVALID_GROUP_ID;
	}

	return 0;
}

int osm_ar_prepare_tables(osm_subn_t *p_subn) {
	osm_switch_t *p_sw;
	uint16_t num_lids;

	num_lids = p_subn->max_lid + 1;

	for (p_sw = (osm_switch_t *) cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
		if (osm_ar_prepare_sw_tables(p_subn, p_sw, num_lids)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7C0D: Failed to allocate for advanced"
				" routing capabilities for switch"
				" 0x%016" PRIx64 "\n",
				cl_ntoh64(osm_node_get_node_guid
					  (p_sw->p_node)));
			return 1;
		}
	}

	return 0;
}

static void ar_tables_destroy(osm_advanced_routing_data_t *p_ar_data,
			      osm_ar_tables_t *p_tables)
{
	free(p_tables->ar_subgroups);
	/* remove all ar lfts/plfts */
	ar_tables_destroy_ar_lft(p_ar_data, p_tables, 0);
	free(p_tables->rn_rcv_strings);
	free(p_tables->rn_direction_string);
	free(p_tables->rn_subgroup_direction);
	if (p_ar_data->num_plfts) {
		free(p_tables->plft_defs);
		p_tables->plft_defs = NULL;
	}
	if (p_tables->port_sl_plfts) {
		free(p_tables->port_sl_plfts);
		p_tables->port_sl_plfts = NULL;
	}

	p_tables->ar_subgroups = NULL;
	p_tables->rn_rcv_strings = NULL;
	p_tables->rn_direction_string = NULL;
	p_tables->rn_subgroup_direction = NULL;
}

void osm_ar_data_destroy(osm_advanced_routing_data_t *p_ar_data)
{
	ar_tables_destroy(p_ar_data, &p_ar_data->subn_tables);
	ar_tables_destroy(p_ar_data, &p_ar_data->calc_tables);
	osm_ar_capabilities_clear_all(p_ar_data);
	free(p_ar_data->ar_group_copy_from_table);
	p_ar_data->ar_group_copy_from_table = NULL;
}

void osm_ar_set_required_ar_info(osm_advanced_routing_data_t *p_ar_data,
				 const ib_ar_info_t *p_ar_info)
{
	if (p_ar_info) {
		memcpy(&p_ar_data->req_ar_info, p_ar_info,
		       sizeof(ib_ar_info_t));
	} else {
		memset(&p_ar_data->req_ar_info, 0, sizeof(ib_ar_info_t));
	}
}

int osm_ar_set_required_group_top(osm_advanced_routing_data_t *p_ar_data,
				  uint16_t group_top)
{
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;

	if (group_top >= ib_ar_info_get_group_cap_ho(p_ar_info))
		return 1;

	p_ar_data->group_top = group_top;
	if (ar_is_group_top_supported(p_ar_info)) {
		ib_ar_info_set_group_top_ho(p_ar_info, group_top);
	}

	return 0;
}

int osm_ar_set_required_ar_mode(osm_advanced_routing_data_t *p_ar_data,
				boolean_t enable_ar, boolean_t enable_fr)
{
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;

	if (enable_fr &&
	    !(p_ar_info->enable_features &
	      IB_AR_INFO_FEATURE_MASK_FR_SUPPORTED))
		return 1;


	AR_SET_BITMASK(p_ar_data->req_ar_info.enable_features,
		       IB_AR_INFO_FEATURE_MASK_AR_ENABLED,
		       enable_ar || enable_fr);

	AR_SET_BITMASK(p_ar_data->req_ar_info.enable_features,
		       IB_AR_INFO_FEATURE_MASK_FR_ENABLED, enable_fr);

	return 0;
}

int osm_ar_set_required_num_subgroups(osm_advanced_routing_data_t *p_ar_data,
				      uint8_t num_subgroups)
{
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;

	// TODO: check with spec
	if (num_subgroups > p_ar_info->sub_grps_sup + 1)
		return 1;
	switch (num_subgroups) {
		case 1:
			break;
		case 2:
			if (!(p_ar_info->sub_grps_sup & 0x1))
				return 1;
			break;
		case 3:
			if (!(p_ar_info->sub_grps_sup & 0x2))
				return 1;
			break;
		case 4:
			if (!(p_ar_info->sub_grps_sup & 0x3))
				return 1;
			break;
		default:
			return 1;
	}

	ib_ar_info_set_sub_grps_active(p_ar_info, num_subgroups - 1);

	return 0;
}

int osm_ar_set_required_enabled_sl_mask(osm_advanced_routing_data_t *p_ar_data,
					uint16_t sl_mask)
{
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;

	if (sl_mask != 0xff) {
		if (!(p_ar_info->enabled_and_cap_fields &
		      IB_AR_INFO_CAP_MASK_BY_SL_CAP))
			return 1;

		AR_SET_BITMASK(p_ar_info->enabled_and_cap_fields,
		       IB_AR_INFO_CAP_MASK_BY_SL_EN, 1);
		p_ar_info->enable_by_sl_mask = cl_ntoh16(sl_mask);
	} else {
		AR_SET_BITMASK(p_ar_info->enabled_and_cap_fields,
		       IB_AR_INFO_CAP_MASK_BY_SL_EN, 0);
	}

	return 0;
}

int osm_ar_set_required_enabled_transport_mask(osm_advanced_routing_data_t *p_ar_data,
					       uint8_t enabled_transport_mask)
{
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;

	if ((enabled_transport_mask & 0xf) != 0xf &&
	    !(p_ar_info->enabled_and_cap_fields &
	      IB_AR_INFO_CAP_MASK_BY_TRANS_CAP))
		return 1;

	p_ar_info->by_trns_disable = ~enabled_transport_mask & 0xf;

	return 0;
}

void osm_ar_set_required_plft_number(osm_advanced_routing_data_t *p_ar_data,
				     uint8_t num_plfts)
{
	p_ar_data->req_num_plfts = num_plfts;
}

void osm_ar_set_required_plft_active_mode(osm_advanced_routing_data_t *p_ar_data,
					  uint8_t active_mode)
{
	p_ar_data->req_plft_info.active_mode = active_mode;
}

void osm_ar_set_required_plft_top(osm_advanced_routing_data_t *p_ar_data,
				  uint8_t plft_id, uint16_t top)
{
	if (p_ar_data->calc_tables.plft_top[plft_id] < top)
		p_ar_data->calc_tables.plft_top[plft_id] = top;
}

uint16_t osm_ar_get_plft_top(osm_advanced_routing_data_t *p_ar_data,
			     uint8_t plft_id)
{
	OSM_ASSERT(plft_id < p_ar_data->num_plfts);
	return p_ar_data->calc_tables.plft_top[plft_id];
}

static inline
uint16_t ar_group_get_offset(osm_advanced_routing_data_t *p_ar_data,
			     uint16_t group_id)
{
	return (group_id * p_ar_data->num_group_subgroups);
}

static inline
osm_ar_subgroup_t * ar_subgroup_get(osm_advanced_routing_data_t *p_ar_data,
				    uint16_t subgroup_idx)
{
	size_t offset = ar_group_get_offset(p_ar_data, 0);
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	return &p_tables->ar_subgroups[offset + subgroup_idx];
}

static inline
uint16_t ar_get_top_subgroup_idx(osm_advanced_routing_data_t *p_ar_data)
{
	return (p_ar_data->group_top + 1) * p_ar_data->num_group_subgroups - 1;
}

osm_ar_subgroup_t * osm_ar_subgroup_get(osm_advanced_routing_data_t *p_ar_data,
					uint16_t group_id, uint8_t subgroup_id)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	size_t group_offset;

	OSM_ASSERT(group_id < p_ar_data->num_groups);
	OSM_ASSERT(subgroup_id < p_ar_data->num_group_subgroups);

	group_offset = ar_group_get_offset(p_ar_data, group_id);
	return (&p_tables->ar_subgroups[group_offset + subgroup_id]);
}

void osm_ar_subgroup_add_port(osm_ar_subgroup_t *p_subgroup, uint8_t port)
{
	OSM_ASSERT(port != 0xff);
	p_subgroup->mask[31 - (port / 8)] |= 1 << (port % 8);
}

static inline boolean_t ar_is_port_in_subgroup(osm_ar_subgroup_t *p_subgroup, uint8_t port)
{
	/* the ports bitmask is presented in network order */
	return (p_subgroup->mask[31 - (port / 8)] & (1 << (port % 8)));
}

void osm_ar_subgroup_clear(osm_ar_subgroup_t *p_subgroup)
{
	memset(p_subgroup, 0, sizeof(osm_ar_subgroup_t));
}

void osm_ar_groups_clear_group(osm_advanced_routing_data_t *p_ar_data,
			       uint16_t group_id)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	size_t group_offset;

	OSM_ASSERT(group_id < p_ar_data->num_groups);

	group_offset = ar_group_get_offset(p_ar_data, group_id);
	memset(&p_tables->ar_subgroups[group_offset], 0,
	       p_ar_data->num_group_subgroups * sizeof(osm_ar_subgroup_t));
}

void osm_ar_groups_clear_tbl(osm_advanced_routing_data_t *p_ar_data)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint16_t num_subgroups;

	num_subgroups = p_ar_data->num_groups * p_ar_data->num_group_subgroups;

	memset(&p_tables->ar_subgroups, 0,
	       num_subgroups * sizeof(osm_ar_subgroup_t));
}

void osm_ar_groups_clear_all(osm_advanced_routing_data_t *p_ar_data)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint16_t num_subgroups = p_ar_data->num_groups *
				 p_ar_data->num_group_subgroups;

	memset(p_tables->ar_subgroups, 0,
	       num_subgroups * sizeof(osm_ar_subgroup_t));
}

void osm_ar_groups_copy_group(osm_advanced_routing_data_t *p_ar_data,
			      uint16_t src_group_id, uint16_t dst_group_id,
			      boolean_t copy_direction)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint16_t src_offset;
	uint16_t dst_offset;

	OSM_ASSERT(src_group_id < p_ar_data->num_groups);
	OSM_ASSERT(dst_group_id < p_ar_data->num_groups);
	OSM_ASSERT(src_group_id != dst_group_id);

	src_offset = ar_group_get_offset(p_ar_data, src_group_id);
	dst_offset = ar_group_get_offset(p_ar_data, dst_group_id);

	memcpy(&p_tables->ar_subgroups[dst_offset],
	       &p_tables->ar_subgroups[src_offset],
	       p_ar_data->num_group_subgroups * sizeof(osm_ar_subgroup_t));

	/* Copy group direction according to argument */
	if (copy_direction) {
		p_tables->rn_subgroup_direction[dst_offset] =
		    p_tables->rn_subgroup_direction[src_offset];
	}

	// TODO: consider checking if copy is enabled/supported by this switch.

	/*
	 * Ensure that source group, is not a copy (not supported in current
	 * implementation).
	 */
	OSM_ASSERT(p_ar_data->ar_group_copy_from_table[dst_group_id] ==
		   AR_INVALID_GROUP_ID);
	OSM_ASSERT(p_ar_data->ar_group_copy_from_table[src_group_id] ==
		   AR_INVALID_GROUP_ID);

	p_ar_data->ar_group_copy_from_table[dst_group_id] = src_group_id;
}

inline static void osm_ar_lft_entry_set(ib_ar_lft_entry_t *p_ar_lft_entry,
					uint8_t port, uint8_t state, uint16_t group_id)
{
	p_ar_lft_entry->legacy_port = port;
	ib_ar_lft_entry_set_state(p_ar_lft_entry, state);
	p_ar_lft_entry->group_number = cl_ntoh16(group_id);
}

void osm_ar_lft_entry_get(struct osm_switch *p_sw, uint8_t plft_id, uint16_t lid,
			  uint8_t *port, uint8_t *state, uint16_t *group_id_ho)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	ib_ar_lft_entry_t *p_ar_lft_entry = &((p_tables->ar_lft[plft_id])[lid]);

	*port = p_ar_lft_entry->legacy_port;
	*state = ib_ar_lft_entry_get_state(p_ar_lft_entry);
	*group_id_ho = cl_ntoh16(p_ar_lft_entry->group_number);
}

void osm_ar_lft_set(osm_advanced_routing_data_t *p_ar_data, uint16_t lid,
		    uint8_t plft_id,
		    uint8_t port, uint8_t state, uint16_t group_id)
{
	uint8_t arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
	uint8_t plft_idx;
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	OSM_ASSERT(lid < p_ar_data->num_lids);
	if (plft_id == ADV_ROUTING_ALL_PLFTS) {
		arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
		for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++)
			osm_ar_lft_entry_set(&((p_tables->ar_lft[plft_idx])[lid]),
					     port, state, group_id);
	} else {
		OSM_ASSERT(plft_id < AR_MAX_NUM_PLFT);
		osm_ar_lft_entry_set(&((p_tables->ar_lft[plft_id])[lid]),
				port, state, group_id);
	}
}

void osm_ar_lft_update_port(osm_advanced_routing_data_t *p_ar_data, uint16_t lid,
			    uint8_t plft_id, uint8_t port)
{
	uint8_t arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
	uint8_t plft_idx;
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	OSM_ASSERT(lid < p_ar_data->num_lids);
	if (plft_id == ADV_ROUTING_ALL_PLFTS) {
		arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
		for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++)
			ib_ar_lft_entry_set_port(&((p_tables->ar_lft[plft_idx])[lid]),
						 port);
	} else {
		OSM_ASSERT(plft_id < AR_MAX_NUM_PLFT);
		ib_ar_lft_entry_set_port(&((p_tables->ar_lft[plft_id])[lid]),
					 port);
	}
}

void osm_ar_lft_copy(osm_advanced_routing_data_t *p_ar_data, uint8_t plft_id,
		     uint16_t copy_to_lid, uint16_t copy_from_lid)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint8_t plft_idx, arlft_num_plfts;

	OSM_ASSERT(copy_from_lid < p_ar_data->num_lids);
	OSM_ASSERT(copy_to_lid < p_ar_data->num_lids);

	if (plft_id == ADV_ROUTING_ALL_PLFTS) {
		/* copy entry in all plfts */
		arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;

		for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++) {
			memcpy(&p_tables->ar_lft[plft_idx][copy_to_lid],
			       &p_tables->ar_lft[plft_idx][copy_from_lid],
			       sizeof(ib_ar_lft_entry_t));
			osm_ar_set_required_plft_top(p_ar_data, plft_idx, copy_to_lid);
		}
	} else {
		OSM_ASSERT(plft_id < AR_MAX_NUM_PLFT);

		memcpy(&p_tables->ar_lft[plft_id][copy_to_lid],
		       &p_tables->ar_lft[plft_id][copy_from_lid],
		       sizeof(ib_ar_lft_entry_t));
		osm_ar_set_required_plft_top(p_ar_data, plft_id, copy_to_lid);
	}
}

static void osm_ar_data_copy_tables(osm_switch_t *p_to_sw, osm_ar_tables_t *p_to_tables,
			            osm_switch_t *p_from_sw, osm_ar_tables_t *p_from_tables,
			            boolean_t is_rn_enabled, uint16_t num_lids)
{
	osm_advanced_routing_data_t *p_from_sw_ar_data = &p_from_sw->ar_data;
	size_t size;
	uint16_t num_subgroups, num_rn_subgroups, lid, max_lid_on_sw;
	uint8_t plft_idx = 0, arlft_num_plfts;

	num_subgroups = p_from_sw_ar_data->num_groups * p_from_sw_ar_data->num_group_subgroups;
	size = AR_ROUNDUP_BLOCK(num_subgroups, 2) * sizeof(osm_ar_subgroup_t);
	memcpy(p_to_tables->ar_subgroups, p_from_tables->ar_subgroups, size);

	arlft_num_plfts = p_from_sw_ar_data->num_plfts ? p_from_sw_ar_data->num_plfts : 1;
	max_lid_on_sw = p_from_sw->max_lid_ho > p_to_sw->max_lid_ho ?
			p_from_sw->max_lid_ho : p_to_sw->max_lid_ho;

	for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++) {
		for (lid = 0; lid <= max_lid_on_sw; lid++) {
			if ((lid <= p_from_sw->max_lid_ho) &&
			    (p_from_tables->ar_lft[plft_idx][lid]).legacy_port != OSM_NO_PATH)
				/* copy from AR table entry */
				(p_to_tables->ar_lft[plft_idx])[lid] =
					(p_from_tables->ar_lft[plft_idx])[lid];
			else
				/* create static entry based on main topology */
				osm_ar_lft_entry_set(&((p_to_tables->ar_lft[plft_idx])[lid]),
						     p_to_sw->new_lft[lid],
						     OSM_AR_IB_LID_STATE_STATIC,
						     0);
		}
	}

	if (is_rn_enabled) {
		num_rn_subgroups =
			(p_from_sw_ar_data->num_strings && p_from_sw_ar_data->num_directions) ?
			num_subgroups : 0;
		size = AR_ROUNDUP_BLOCK(num_rn_subgroups, IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE) *
			sizeof(uint8_t);
		memcpy(p_to_tables->rn_subgroup_direction, p_from_tables->rn_subgroup_direction,
		       size);

		size = AR_ROUNDUP_BLOCK(p_from_sw_ar_data->num_directions, IB_RN_GEN_STRING_BLOCK_SIZE) *
			sizeof(ib_net16_t);
		memcpy(p_to_tables->rn_direction_string, p_from_tables->rn_direction_string, size);

		size = AR_ROUNDUP_BLOCK(p_from_sw_ar_data->num_strings, IB_RN_RCV_STRING_BLOCK_SIZE) *
			sizeof(ib_rn_rcv_string_entry_t);
		memcpy(p_to_tables->rn_rcv_strings, p_from_tables->rn_rcv_strings, size);

		memcpy(p_to_tables->rn_subgroup_priority_flags,
		       p_from_tables->rn_subgroup_priority_flags,
		       sizeof(p_to_tables->rn_subgroup_priority_flags));

		memcpy(p_to_tables->rn_xmit_port_mask, p_from_tables->rn_xmit_port_mask,
		       sizeof(p_to_tables->rn_xmit_port_mask));
	}
	if (p_from_sw_ar_data->num_plfts > 1) {
		size = AR_ROUNDUP_BLOCK(p_from_sw_ar_data->num_plfts, IB_PLFT_DEF_BLOCK_SIZE) *
		       sizeof(ib_plft_def_element_t);
		memcpy(p_to_tables->plft_defs, p_from_tables->plft_defs, size);

		size = AR_ROUNDUP_BLOCK(p_from_sw->num_ports, IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE) *
			sizeof(ib_port_sl_to_plft_map_entry_t);
		memcpy(p_to_tables->port_sl_plfts, p_from_tables->port_sl_plfts, size);

		/* update top plft to the max lid in the subnet */
		if (p_to_sw->max_lid_ho > p_from_sw->max_lid_ho)
			for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++)
				p_to_tables->plft_top[plft_idx] = p_to_sw->max_lid_ho;
		else
			memcpy(p_to_tables->plft_top, p_from_tables->plft_top,
			       sizeof(p_to_tables->plft_top));
	}
}

void osm_ar_lft_set_static_state_for_dlids(osm_advanced_routing_data_t *p_ar_data,
					   uint16_t *lids, uint16_t num_lids)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint8_t plft_idx, plft_top;
	uint16_t lid_idx;
	uint16_t lid;
	boolean_t fr_enabled;

	fr_enabled = p_ar_data->req_ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_FR_ENABLED;
	plft_top = p_ar_data->num_plfts ? p_ar_data->num_plfts - 1 : 0;

	for (plft_idx = 0; plft_idx <= plft_top; plft_idx++) {
		ib_ar_lft_entry_t *ar_lft = p_tables->ar_lft[plft_idx];
		for (lid_idx = 0; lid_idx < num_lids; lid_idx++) {
			lid = lids[lid_idx];
			if (ar_lft[lid].legacy_port != OSM_NO_PATH)
				ib_ar_lft_entry_set_state(&ar_lft[lid],
						 OSM_AR_IB_LID_STATE_STATIC);

			/* if route is static and FR is disabled, set group to zero */
			if (!fr_enabled)
				ar_lft[lid].group_number = 0;
		}
	}
}

static inline
boolean_t ar_is_rn_enabled(osm_subn_t *p_subn)
{
	return (p_subn->ar_conf_params.ar_mode == OSM_ADV_ROUTING_MODE_RN ||
		p_subn->ar_conf_params.shield_mode == OSM_ADV_ROUTING_MODE_RN);
}

void osm_ar_data_copy(osm_subn_t *p_to_subn, osm_subn_t *p_from_subn,
		      osm_switch_t *p_to_sw, osm_switch_t *p_from_sw)
{
	boolean_t is_rn_enabled = ar_is_rn_enabled(p_from_subn);
	uint16_t num_lids, max_lid;

	max_lid = p_to_subn->max_lid > p_from_subn->max_lid ?
		  p_to_subn->max_lid : p_from_subn->max_lid;
	num_lids = max_lid + 1;

	p_to_sw->ar_configured = p_from_sw->ar_configured;
	p_to_sw->ar_data.status = p_from_sw->ar_data.status;
	/* MADs that were sent during discovery */
	memcpy(&p_to_sw->ar_data.ar_info, &p_from_sw->ar_data.ar_info, sizeof(ib_ar_info_t));
	memcpy(&p_to_sw->ar_data.plft_info, &p_from_sw->ar_data.plft_info, sizeof(ib_plft_info_t));

	/* MADs that were calculated during routing */
	memcpy(&p_to_sw->ar_data.req_ar_info, &p_from_sw->ar_data.req_ar_info, sizeof(ib_ar_info_t));
	memcpy(&p_to_sw->ar_data.req_plft_info, &p_from_sw->ar_data.req_plft_info, sizeof(ib_plft_info_t));

	/* need to be set prior to data structure allocation */
	p_to_sw->ar_data.req_num_plfts = p_from_sw->ar_data.req_num_plfts;
	/* group top was calculated and set during routing */
	p_to_sw->ar_data.group_top = p_from_sw->ar_data.group_top;
	if (osm_ar_prepare_sw_tables(p_to_subn, p_to_sw, num_lids)) {
		OSM_LOG(&p_from_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7C0E: Failed to allocate for advanced"
			" routing capabilities for switch"
			" 0x%016" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_to_sw->p_node)));
		return;
	}
	/* copy tables */
	p_to_sw->ar_data.num_groups = p_from_sw->ar_data.num_groups;
        p_to_sw->ar_data.num_tables = p_from_sw->ar_data.num_tables;
        p_to_sw->ar_data.num_group_subgroups = p_from_sw->ar_data.num_group_subgroups;
        p_to_sw->ar_data.num_strings = p_from_sw->ar_data.num_strings;
        p_to_sw->ar_data.num_directions = p_from_sw->ar_data.num_directions;
        p_to_sw->ar_data.group_top = p_from_sw->ar_data.group_top;

	/* no need to copy cached_capabilities_list */
	osm_ar_data_copy_tables(p_to_sw, &p_to_sw->ar_data.calc_tables,
				p_from_sw, &p_from_sw->ar_data.calc_tables,
				is_rn_enabled, num_lids);
}

void osm_ar_capabilities_clear_all(osm_advanced_routing_data_t *p_ar_data)
{
	cl_qlist_t *p_list = &p_ar_data->cached_capabilities_list;

	while (cl_qlist_count(p_list))
		free(cl_qlist_remove_head(p_list));
}

static inline
boolean_t ar_capabilities_compare(const ib_ar_info_t *p_ar_info1,
				 const ib_ar_info_t *p_ar_info2)
{

	// TODO: rewrite function, support returning true when disabling features
	// consider returning (-1,0,1)
	// -1 : if p_ar_info1 capabilities are contained in p_ar_info2
	//  0 : if capabilities match
	//  1 : if p_ar_info1 capabilities are not contained in p_ar_info2
	//  consider passing an argument for comparision type

	static const uint8_t enable_features_mask =
	    IB_AR_INFO_FEATURE_MASK_AR_ENABLED |
	    IB_AR_INFO_FEATURE_MASK_FR_ENABLED |
	    IB_AR_INFO_FEATURE_MASK_RN_XMIT_ENABLED;

	if ((p_ar_info1->enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_EN) !=
	    (p_ar_info2->enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_EN))
		return FALSE;

	if (ib_ar_info_get_sub_grps_active(p_ar_info1) !=
	    ib_ar_info_get_sub_grps_active(p_ar_info2))
		return FALSE;

	if ((p_ar_info1->enable_features & enable_features_mask) !=
	    (p_ar_info2->enable_features & enable_features_mask))
		return FALSE;

	if (p_ar_info1->enable_by_sl_mask !=
	    p_ar_info2->enable_by_sl_mask)
		return FALSE;

	if (p_ar_info1->by_trns_disable !=
	    p_ar_info2->by_trns_disable)
		return FALSE;

	return TRUE;
}

const ib_ar_info_t *
osm_ar_capabilities_search(osm_advanced_routing_data_t *p_ar_data,
			  const ib_ar_info_t *p_ar_info)
{
	cl_qlist_t *p_list = &p_ar_data->cached_capabilities_list;
	cl_list_item_t *p_list_item = cl_qlist_head(p_list);
	const osm_ar_capability_t *p_ar_cap = NULL;

	while (p_list_item != cl_qlist_end(p_list)) {
		p_ar_cap = (osm_ar_capability_t *)p_list_item;
		if (ar_capabilities_compare(&p_ar_cap->ar_info, p_ar_info))
			return &p_ar_cap->ar_info;
		p_list_item = cl_qlist_next(p_list_item);
	}

	return NULL;
}

int osm_ar_capabilities_update(osm_advanced_routing_data_t *p_ar_data,
			      const ib_ar_info_t *p_ar_info)
{
	cl_qlist_t *p_list = &p_ar_data->cached_capabilities_list;
	cl_list_item_t *p_list_item = cl_qlist_head(p_list);
	osm_ar_capability_t *p_ar_cap = NULL;

	/* Check if already have matching capabilities record */
	while (p_list_item != cl_qlist_end(p_list)) {
		p_ar_cap = (osm_ar_capability_t *)p_list_item;
		if (ar_capabilities_compare(&p_ar_cap->ar_info, p_ar_info)) {
			memcpy(&p_ar_cap->ar_info, p_ar_info,
			       sizeof(ib_ar_info_t));
			return 0;
		}
		p_list_item = cl_qlist_next(p_list_item);
	}

	/* Did not find matching capabilities information - create new record */
	p_ar_cap = (osm_ar_capability_t *)calloc(1,
						 sizeof(osm_ar_capability_t));
	if (!p_ar_cap)
		return 1;

	memcpy(&p_ar_cap->ar_info, p_ar_info, sizeof(ib_ar_info_t));
	cl_qlist_insert_tail(&p_ar_data->cached_capabilities_list,
			     &p_ar_cap->list_item);

	return 0;
}

ib_api_status_t osm_ar_get_ar_info(osm_sm_t * sm, osm_switch_t *p_sw)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.ar_mad_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_GET;

	p_sw->ar_data.pending_mads++;

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO,
			     0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C01: "
			"Request for ARInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

ib_api_status_t osm_ar_query_ar_info(osm_sm_t * sm, osm_switch_t *p_sw,
				     ib_ar_info_t *p_ar_info)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);
	p_sw->ar_data.pending_mads++;

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.ar_mad_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_DYNCALC;

	status =
	    osm_send_req(sm, p_dr_path, (uint8_t *)p_ar_info,
			 sizeof(ib_ar_info_t),
			 IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO, cl_hton32(1 << 31),
			 IB_MAD_METHOD_GET,
			 FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			 0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C02: "
			"Request for dynamically calculated  ARInfo "
			"from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

ib_api_status_t osm_ar_set_ar_info(osm_sm_t * sm, osm_switch_t *p_sw,
				   ib_ar_info_t *p_ar_info)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.ar_mad_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;

	p_sw->ar_data.pending_mads++;
	status = osm_req_set(sm, p_dr_path, (uint8_t *)p_ar_info,
			     sizeof(ib_ar_info_t),
			     IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO, 0,
			     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C03: "
			"Setting ARInfo on 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

void osm_ar_set_subgroup_direction(osm_advanced_routing_data_t *p_ar_data,
				   uint16_t group_id, uint8_t subgroup_id,
				   uint8_t direction)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	size_t subgroup_idx;

	OSM_ASSERT(group_id < p_ar_data->num_groups);
	OSM_ASSERT(subgroup_id < p_ar_data->num_group_subgroups);
	OSM_ASSERT(direction < p_ar_data->num_directions);

	subgroup_idx = ar_group_get_offset(p_ar_data, group_id) +
		       subgroup_id;

	p_tables->rn_subgroup_direction[subgroup_idx] = direction;
}

uint8_t osm_ar_subgroup_direction_get(osm_advanced_routing_data_t *p_ar_data,
				      uint16_t group_id, uint8_t subgroup_id)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	size_t subgroup_idx;

	OSM_ASSERT(group_id < p_ar_data->num_groups);
	OSM_ASSERT(subgroup_id < p_ar_data->num_group_subgroups);

	subgroup_idx = ar_group_get_offset(p_ar_data, group_id) +
		       subgroup_id;

	return p_tables->rn_subgroup_direction[subgroup_idx];
}

ib_api_status_t osm_ar_get_extended_node_info(osm_sm_t * sm, osm_switch_t *p_sw)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.gi_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.gi_context.set_method = FALSE;

	p_sw->ar_data.pending_mads++;

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_MLNX_EXTENDED_NODE_INFO,
			     0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C19: "
			"Request get for ExtendedNodeInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

ib_api_status_t osm_ar_get_plft_info(osm_sm_t * sm, osm_switch_t *p_sw)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.ar_mad_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_GET;

	p_sw->ar_data.pending_mads++;

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_PLFT_INFO,
			     0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C13: "
			"Request get for PLFTInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

ib_api_status_t osm_ar_set_plft_info(osm_sm_t * sm, osm_switch_t *p_sw,
				     ib_plft_info_t *p_plft_info)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.ar_mad_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;

	/* mark as unsupporting of AR */
	p_sw->ar_data.pending_mads++;

	status = osm_req_set(sm, p_dr_path, (uint8_t *)p_plft_info,
			     sizeof(ib_plft_info_t),
			     IB_MAD_ATTR_PLFT_INFO, 0,
			     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C14: "
			"Setting PLFTInfo on 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

ib_api_status_t osm_ar_set_extended_node_info(osm_sm_t * sm, osm_switch_t *p_sw,
					      ib_mlnx_ext_node_info_t *p_extended_node_info)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_sw->p_node;
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.gi_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.gi_context.set_method = TRUE;

	status = osm_req_set(sm, p_dr_path, (uint8_t *)p_extended_node_info,
			     sizeof(ib_mlnx_ext_node_info_t),
			     IB_MAD_ATTR_MLNX_EXTENDED_NODE_INFO, 0,
			     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7C15: "
			"Setting ExtendedNodeInfo on 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

	return status;
}

void osm_ar_subgroup_pri_rn_flags_set(osm_advanced_routing_data_t *p_ar_data,
				      uint8_t subgroup_priority,
				      uint8_t rn_flags)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	OSM_ASSERT(subgroup_priority < p_ar_data->num_group_subgroups);
	p_tables->rn_subgroup_priority_flags[subgroup_priority] = cl_hton32(rn_flags);
}

void osm_ar_rn_direction_string_set(osm_advanced_routing_data_t *p_ar_data,
				    uint8_t direction, uint16_t string)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	OSM_ASSERT(string < p_ar_data->num_strings);
	OSM_ASSERT(direction < p_ar_data->num_directions);

	p_tables->rn_direction_string[direction] = cl_hton16(string);
}

void osm_ar_plft_def_set(osm_advanced_routing_data_t *p_ar_data,
			 uint8_t plft_id, uint8_t offset,
			 uint8_t space, uint8_t bank,
			 uint8_t force_plft_sub_group0)
{
	ib_plft_def_element_t *plfts = p_ar_data->calc_tables.plft_defs;

	plfts[plft_id].force_plft_sub_group0_bank =
		(force_plft_sub_group0 << 7) | (0xf & bank);
	plfts[plft_id].plft_lid_space = space;
	plfts[plft_id].plft_lid_offset = offset;
}

void osm_ar_rn_string_handling_clear(osm_advanced_routing_data_t *p_ar_data)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	memset(p_tables->rn_rcv_strings, 0,
	       p_ar_data->num_strings * sizeof(ib_rn_rcv_string_entry_t));
}

void osm_ar_rn_string_handling_set(osm_advanced_routing_data_t *p_ar_data,
				   uint16_t string, uint8_t decision,
				   uint16_t new_string)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	OSM_ASSERT(string < p_ar_data->num_strings);

	p_tables->rn_rcv_strings[string].decision = decision;
	p_tables->rn_rcv_strings[string].string2string = cl_hton16(new_string);
}

void osm_ar_rn_port_xmit_mask_add_port(osm_advanced_routing_data_t *p_ar_data,
				       uint8_t port, uint8_t mask)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;
	uint8_t element = p_tables->rn_xmit_port_mask[port/2];

	if (port % 2 == 0) {
		/* set high 4 bits */
		element = (element & 0x0F) | (mask << 4);
	} else {
		/* set low 4 bits */
		element = (element & 0xF0) | mask;
	}

	p_tables->rn_xmit_port_mask[port/2] = element;
}

void osm_ar_rn_port_xmit_mask_clear(osm_advanced_routing_data_t *p_ar_data)
{
	osm_ar_tables_t *p_tables = &p_ar_data->calc_tables;

	memset(p_tables->rn_xmit_port_mask, 0,
	       sizeof(p_tables->rn_xmit_port_mask));
}

void osm_ar_update_status(osm_switch_t *p_sw)
{
	if (p_sw->ar_data.pending_mads > 0)
		p_sw->ar_data.status = AR_STATUS_FAILURE;
}

int ar_set_subn_tables(osm_sm_t * sm, cl_qlist_t *p_lists,
		       ar_get_next_mad_fn_t get_next_mad_fn,
		       boolean_t update_ar_status)
{
	cl_qlist_t mad_list;
	cl_map_item_t *p_map_item;
	osm_switch_t *p_sw;
	osm_subn_t *p_subn;
	boolean_t repeat;
	int num_lists = 0;
	int i;

	p_subn = sm->p_subn;
	cl_qlist_init(&mad_list);

	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		if (update_ar_status)
			osm_ar_update_status(p_sw);

		if ((p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED) ||
		    (p_sw->ar_data.status != AR_STATUS_SUCCESS))
			continue;
		get_next_mad_fn(sm, p_sw,
				p_subn->need_update || p_sw->need_update,
				&p_lists[num_lists]);

		/* Add number of MADs we send to the switch to p_sw->ar_data.pending_mads */
		p_sw->ar_data.pending_mads += cl_qlist_count(&p_lists[num_lists]);

		num_lists++;
	}

	do {
		repeat = FALSE;
		for (i = 0; i < num_lists; i++) {
			if (cl_qlist_count(&p_lists[i])) {
				cl_qlist_insert_tail(&mad_list,
						     cl_qlist_remove_head(&p_lists[i]));
				repeat = TRUE;
			}
		}
	} while (repeat);

	if (cl_qlist_count(&mad_list))
		osm_send_req_mad_batch(sm, &mad_list);

	return 0;
}

static int ar_set_subn_ar_info(osm_sm_t * sm)
{
	osm_subn_t *p_subn = sm->p_subn;
	osm_switch_t *p_sw;
	osm_advanced_routing_data_t *p_ar_data;
	cl_map_item_t *p_map_item;
	ib_ar_info_t ar_info;

	OSM_LOG_ENTER(sm->p_log);

	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		p_ar_data = &p_sw->ar_data;

		/* check if we need to force sending LFT  */
		if (p_sw->ar_configured == OSM_SW_AR_CONF_NONE ||
		    p_sw->ar_configured == OSM_SW_AR_CONF_DISABLED) {
			if (ar_is_ar_enabled(&p_ar_data->req_ar_info)) {
				/*
				 * Clean the required bit,
				 * no need to send ARInfo MAD */
				p_ar_data->req_ar_info.enable_features &=
				    ~(IB_AR_INFO_FEATURE_MASK_AR_ENABLED |
				      IB_AR_INFO_FEATURE_MASK_FR_ENABLED);
			}
			continue;
		}

		osm_ar_update_status(p_sw);
		if ((p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED) ||
		    (p_sw->ar_data.status != AR_STATUS_SUCCESS))
			continue;

		/* Prepare ARInfo SET request */
		memcpy(&ar_info, &p_ar_data->req_ar_info, sizeof(ib_ar_info_t));

		/* Set group top */
		if (ar_is_group_top_supported(&ar_info)) {
			ib_ar_info_set_group_top_ho(&ar_info, p_ar_data->group_top);
		}

		/* Avoid sending ARInfo if it did not change */
		if (!p_sw->need_update && !p_subn->need_update) {
			if (!memcmp(&ar_info, &p_ar_data->ar_info,
				    sizeof(ib_ar_info_t)))
				continue;
		}

		osm_ar_set_ar_info(sm, p_sw, &ar_info);
	}

	OSM_LOG_EXIT(sm->p_log);

	return 0;
}

static int ar_set_subn_plft_info(osm_sm_t * sm)
{
	osm_subn_t *p_subn = sm->p_subn;
	osm_switch_t *p_sw;
	osm_advanced_routing_data_t *p_ar_data;
	cl_map_item_t *p_map_item;
	ib_plft_info_t plft_info;
	boolean_t active_mode_change = FALSE;

	OSM_LOG_ENTER(sm->p_log);
	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		p_ar_data = &p_sw->ar_data;

		/*
		 * in case the active routing engine is AR routing engine, but
		 * it doesn't use PLFTs, so num_plfts is 0, we still want
		 * to send PlftInfo set, to clean previous configuration.
		 * Once plft info active is 0 in the fabric too, next time
		 * plft info set won't be sent.
		 */
		if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
			continue;

		/* Prepare PLFTInfo SET request */
		memcpy(&plft_info, &p_ar_data->req_plft_info, sizeof(ib_plft_info_t));

		/* Avoid sending PLFTInfo if it did not change */
		if (!p_sw->need_update && !p_subn->need_update) {
			if (!memcmp(&plft_info, &p_ar_data->plft_info,
				    sizeof(ib_plft_info_t)))
				continue;
		}

		/* need to go through active mode = 0 */
		if (plft_info.active_mode != p_ar_data->plft_info.active_mode) {
			plft_info.active_mode = 0;
			active_mode_change = TRUE;
		}

		osm_ar_set_plft_info(sm, p_sw, &plft_info);
	}

	/* if there was a change in some switch of the active_mode */
	if (active_mode_change) {
		if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats))
			return 1;

		/* send for the successfull switches the needed active mode */
		for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     p_map_item = cl_qmap_next(p_map_item)) {
			p_sw = (osm_switch_t *)p_map_item;
			p_ar_data = &p_sw->ar_data;

			if (!p_ar_data->num_plfts ||
			    p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
				continue;

			/* Prepare PLFTInfo SET request */
			memcpy(&plft_info, &p_ar_data->req_plft_info, sizeof(ib_plft_info_t));

			osm_ar_update_status(p_sw);

			/* send with unchanged req_plft, to set the real active_mode */
			if ((p_ar_data->status == AR_STATUS_SUCCESS) &&
			    plft_info.active_mode != p_ar_data->plft_info.active_mode)
				osm_ar_set_plft_info(sm, p_sw, &plft_info);
		}
	}
	OSM_LOG_EXIT(sm->p_log);

	return 0;
}

static int ar_set_subn_extended_node_info(osm_sm_t * sm)
{
	osm_subn_t *p_subn = sm->p_subn;
	osm_switch_t *p_sw;
	osm_node_t *p_node;
	cl_map_item_t *p_map_item;
	osm_advanced_routing_data_t *p_ar_data;

	OSM_LOG_ENTER(sm->p_log);

	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;
		p_node = p_sw->p_node;
		p_ar_data = &p_sw->ar_data;

		/*
		 * we get extended node info MAD if switch or subnet need update,
		 * so we don't miss the subnet configuration,
		 * so no need to recheck it here
		 */
		if (p_node->ext_node_info.sl_2_vl_act == p_node->new_ext_node_info.sl_2_vl_act)
			continue;

		p_ar_data->pending_mads++;

		osm_ar_set_extended_node_info(sm, p_sw, &p_node->new_ext_node_info);
	}

	OSM_LOG_EXIT(sm->p_log);

	return 0;
}

static inline
boolean_t ar_check_reset_subgroup_direction(osm_advanced_routing_data_t *p_ar_data,
					    uint16_t subgroup_indx)
{
	return (!p_ar_data->calc_tables.rn_subgroup_direction[subgroup_indx] ||
		p_ar_data->calc_tables.rn_subgroup_direction[subgroup_indx] !=
		p_ar_data->subn_tables.rn_subgroup_direction[subgroup_indx]);
}

static void get_group_table_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				 boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint8_t *p_calc_data, *p_subn_data;
	osm_ar_tables_t *p_calc_tables, *p_subn_tables;
	uint16_t block_id, max_block_id;
	uint16_t top_subgroup_id, offset, group_id1, group_id2;
	uint8_t reset_subgroup_direction = 0;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	top_subgroup_id = ar_get_top_subgroup_idx(p_ar_data);
	max_block_id = (top_subgroup_id) / 2;
	p_calc_tables = &p_ar_data->calc_tables;
	p_subn_tables = &p_ar_data->subn_tables;

	for (block_id = 0; block_id <= max_block_id; block_id++) {

		/*
		 * If groups of both entries are marked for copy, skip current
		 * block.
		 * When marked for copy, the entry will be different
		 * from 'invalid'
		 */
		/* TODO: make sure copy is supported and enabled */
		offset = block_id * 2;
		group_id1 = offset / p_ar_data->num_group_subgroups;
		group_id2 = (offset + 1) / p_ar_data->num_group_subgroups;
		if ((p_ar_data->ar_group_copy_from_table[group_id1] !=
		     AR_INVALID_GROUP_ID) &&
		    (p_ar_data->ar_group_copy_from_table[group_id2] !=
		     AR_INVALID_GROUP_ID)) {
			continue;
		}

		/* TODO: if one of the groups is marked for copy,
		 * and is being sent now, then
		 * ar_group_copy_from_table[group_id] = AR_INVALID_GROUP_ID
		 * as copy is no longer needed for it.
		 * This is possible if the 'copy direction' is false
		 * (the copy mad also copies direction).
		 */

		/* Check if block already configured */
		p_calc_data = (uint8_t*)&p_calc_tables->ar_subgroups[offset];
		p_subn_data = (uint8_t*)&p_subn_tables->ar_subgroups[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
		context.ar_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.ar_mad_context.block_id = block_id;

		/* TODO: support PLFT */
		/* TODO: support GlbGroups == 0 */

		if (ar_is_rn_supported(&p_ar_data->ar_info)) {
			if (!ar_is_rn_enabled(sm->p_subn)) {
				reset_subgroup_direction = 0x3;
			} else {
				/* reset subgroup direction if changed */
				if (ar_check_reset_subgroup_direction(p_ar_data, offset))
					reset_subgroup_direction = 0x1;
				if (offset != top_subgroup_id &&
				    ar_check_reset_subgroup_direction(p_ar_data, offset + 1))
					reset_subgroup_direction |= 0x2;
			}
		}

		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE,
				cl_hton32((uint32_t)block_id |
					(reset_subgroup_direction << 12)),
				FALSE,
				ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C04: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static osm_madw_t *
ar_prepare_ar_group_copy_mad(osm_sm_t *sm, osm_physp_t *p_physp,
			     ib_ar_group_copy_t *ib_ar_group_copy_mad,
			     uint16_t from_group)
{
	osm_madw_context_t context;
	uint32_t attr_mod = from_group & 0xfff;

	context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
	context.ar_mad_context.node_guid = osm_node_get_node_guid(p_physp->p_node);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"GUID 0x%" PRIx64 " from group %u attr mod %x\n",
		cl_ntoh64(context.ar_mad_context.node_guid),
		from_group, attr_mod);
	osm_madw_t *p_madw =
	    osm_prepare_req_set(sm, osm_physp_get_dr_path_ptr(p_physp),
				(uint8_t *)ib_ar_group_copy_mad,
				IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE_COPY,
				cl_hton32(attr_mod),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				sm->p_subn->opt.long_transaction_timeout,
				CL_DISP_MSGID_NONE, &context);

	return p_madw;
}

static inline
boolean_t group_table_is_changed(osm_advanced_routing_data_t *p_ar_data,
				 uint16_t group)
{
	uint16_t offset = group * p_ar_data->num_group_subgroups;
	return !!memcmp(&p_ar_data->calc_tables.ar_subgroups[offset],
			&p_ar_data->subn_tables.ar_subgroups[offset],
			(p_ar_data->num_group_subgroups *
			 sizeof(osm_ar_subgroup_t)));
}

static void
_get_group_table_copy_mads(osm_sm_t *sm, osm_switch_t *p_sw,
			   uint16_t from_group, uint16_t *to_group,
			   uint16_t num_groups, boolean_t force_send,
			   cl_qlist_t *p_mad_list)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	ib_ar_group_copy_t ib_ar_group_copy_mad;
	osm_madw_t *p_madw;
	uint16_t offset = 0;
	uint16_t range_start, range_end, group_num = 0;
	uint8_t entry = 0;

	memset(&ib_ar_group_copy_mad, 0, sizeof(ib_ar_group_copy_mad));

	while (offset < num_groups) {

		range_start = range_end = 0;
		/* determine start range */
		for (; offset < num_groups; offset++) {
			group_num = to_group[offset];
			if (force_send ||
			    group_table_is_changed(&p_sw->ar_data, group_num))
				break;
		}

		/* nothing to send */
		if (offset == num_groups)
			break;

		range_end = range_start = group_num;

		/* determine end range */
		for (; offset < num_groups; offset++) {
			group_num = to_group[offset];
			/* either the range is broken or we reached
			 * maximal number of copy groups
			 */
			if (range_end + 1 < group_num ||
			    (range_end - range_start + 1) >= AR_MAX_COPY_GROUPS_NUM)
				break;
			if (force_send ||
			    group_table_is_changed(&p_sw->ar_data, group_num))
				range_end = group_num;
		}

		/* add range to mad and continue */
		ib_ar_group_copy_mad.group_copy_entry[entry].first_index =
		    cl_hton16(range_start);
		ib_ar_group_copy_mad.group_copy_entry[entry].last_index =
		    cl_hton16(range_end);

		/* Check if we filled the MAD */
		entry++;
		if (entry == IB_AR_GROUP_COPY_BLOCK_SIZE ||
		    (range_end - range_start + 1) >= AR_MAX_COPY_GROUPS_NUM) {
			/* All entries were filed, send current MAD */
			p_madw = ar_prepare_ar_group_copy_mad(sm, p_physp,
							      &ib_ar_group_copy_mad,
							      from_group);
			if (!p_madw) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 7C20: Failed to allocate memory\n");
			} else {
				cl_qlist_insert_tail(p_mad_list,
						&p_madw->list_item);
			}

			/* Initialize next MAD */
			entry = 0;
			memset(&ib_ar_group_copy_mad, 0, sizeof(ib_ar_group_copy_mad));
		}
	}

	if (entry > 0) {
		/* Finalize current MAD */
		p_madw = ar_prepare_ar_group_copy_mad(sm, p_physp,
				&ib_ar_group_copy_mad,
				from_group);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 7C21: Failed to allocate memory\n");
		} else {
			cl_qlist_insert_tail(p_mad_list, &p_madw->list_item);
		}
	}

}

static void get_group_table_copy_mads(osm_sm_t * sm, osm_switch_t *p_sw,
			    	      boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	uint16_t to_count[AR_MAX_GROUPS] = { 0 };
	uint16_t to_offset[AR_MAX_GROUPS] = { 0 };
	uint16_t to_table[AR_MAX_GROUPS] = { 0 };
	uint16_t to_group, from_group;
	uint16_t offset;

	/* Check if adaptive routing group copy is supported by this switch */
	if (!ib_ar_info_get_grp_tbl_cpy_sup(&p_ar_data->req_ar_info))
		return;

	/* Count number of copies for each source group */
	for (to_group = 0; to_group <= p_ar_data->group_top; to_group++) {
		from_group = p_ar_data->ar_group_copy_from_table[to_group];
		if (from_group != AR_INVALID_GROUP_ID)
			to_count[from_group]++;
	}

	/* Determine offset in 'to_table' for each source group */
	offset = 0;
	for (to_group = 0; to_group <= p_ar_data->group_top; to_group++) {
		to_offset[to_group] = offset;
		offset += to_count[to_group];
	}

	/* If there aren't copied groups, exit the function */
	if (offset == 0)
		return;

	/* Construct 'to_table' */
	for (to_group = 0; to_group <= p_ar_data->group_top; to_group++) {
		from_group = p_ar_data->ar_group_copy_from_table[to_group];
		if (from_group != AR_INVALID_GROUP_ID) {
			to_table[to_offset[from_group]] = to_group;
			to_offset[from_group]++;
		}
	}

	/* Create AR group copy MADs */
	offset = 0;
	for (from_group = 0; from_group <= p_ar_data->group_top; from_group++) {

		if (!to_count[from_group]) {
			/* No group to copy to, skip group */
			continue;
		}

		_get_group_table_copy_mads(sm, p_sw, from_group,
					   &to_table[offset],
					   to_count[from_group],
					   force_send, p_list);
		offset += to_count[from_group];
	} /* from_group loop */
}

static void get_ar_lft_mads(osm_sm_t *sm, osm_switch_t *p_sw,
			    boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint16_t block_id, max_block_id, offset;
	uint8_t *p_calc_data, *p_subn_data, plft_idx, arlft_num_plfts;
	uint32_t attr_mod;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;

	max_block_id = (p_ar_data->num_lids - 1) / IB_AR_LFT_BLOCK_SIZE;
	for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++) {
		for (block_id = 0; block_id <= max_block_id; block_id++) {
			/* Check if block already configured */
			offset = block_id * IB_AR_LFT_BLOCK_SIZE;
			p_calc_data = (uint8_t*)&p_ar_data->calc_tables.ar_lft[plft_idx][offset];
			p_subn_data = (uint8_t*)&p_ar_data->subn_tables.ar_lft[plft_idx][offset];
			if (!force_send &&
			    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
				continue;

			context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
			context.ar_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
			context.ar_mad_context.block_id = block_id;

			attr_mod = (plft_idx << 24)  | (uint32_t)block_id;

			p_madw = osm_prepare_req_set(
					sm, osm_physp_get_dr_path_ptr(p_physp),
					p_calc_data, IB_SMP_DATA_SIZE,
					IB_MAD_ATTR_ADAPTIVE_ROUTING_LFT,
					cl_hton32(attr_mod),
					FALSE, ib_port_info_get_m_key(&p_physp->port_info),
					0, CL_DISP_MSGID_NONE, &context);

			if (!p_madw) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 7C05: Failed to allocate memory\n");
				break;
			}

			cl_qlist_insert_tail(p_list, &p_madw->list_item);
		}
	}
}

static void get_plft_map_mads(osm_sm_t *sm, osm_switch_t *p_sw,
			      boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint8_t plft_idx;
	uint16_t subn_plft_top, calc_plft_top;
	ib_plft_map_t plft_map;

	memset(&plft_map, 0, sizeof(ib_plft_map_t));

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_ar_data->num_plfts || !p_physp)
		return;

	context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
	context.ar_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);

	for (plft_idx = 0; plft_idx < p_ar_data->num_plfts; plft_idx++) {
		subn_plft_top = p_ar_data->subn_tables.plft_top[plft_idx];
		calc_plft_top = p_ar_data->calc_tables.plft_top[plft_idx];

		if (!force_send && subn_plft_top == calc_plft_top)
			continue;

		context.ar_mad_context.block_id = plft_idx;

		/* only LFT TOP enable */
		plft_map.control_map = IB_PLFT_MAP_CONTROL_MAP_LFT_TOP_ENABLE;
		plft_map.lft_top = cl_hton16(calc_plft_top);

		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				(uint8_t *)&plft_map,
				IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_PLFT_MAP,
				cl_hton32((uint32_t)plft_idx),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);

		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C18: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_rn_gen_string_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				   boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint16_t offset, block_id, max_block_id;
	uint8_t *p_calc_data, *p_subn_data;

	/* TODO: support PLFT */

	if (!ar_is_rn_supported_enabled(&p_ar_data->ar_info))
		return;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	max_block_id = (p_ar_data->num_directions - 1) / IB_RN_GEN_STRING_BLOCK_SIZE;
	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_RN_GEN_STRING_BLOCK_SIZE;
		p_calc_data =
		    (uint8_t*)&p_ar_data->calc_tables.rn_direction_string[offset];
		p_subn_data =
		    (uint8_t*)&p_ar_data->subn_tables.rn_direction_string[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.rn_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.rn_mad_context.block_id = block_id;
		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_RN_GEN_STRING_TABLE,
				cl_hton32((uint32_t)block_id << 8),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);

		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C06: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_rn_rcv_string_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				   boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint16_t offset, block_id, max_block_id;
	uint8_t *p_calc_data, *p_subn_data;

	if (!ar_is_rn_supported_enabled(&p_ar_data->ar_info))
		return;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	max_block_id = (p_ar_data->num_strings - 1) / IB_RN_RCV_STRING_BLOCK_SIZE;
	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_RN_RCV_STRING_BLOCK_SIZE;
		p_calc_data =
		    (uint8_t*)&p_ar_data->calc_tables.rn_rcv_strings[offset];
		p_subn_data =
		    (uint8_t*)&p_ar_data->subn_tables.rn_rcv_strings[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			break;

		context.rn_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.rn_mad_context.block_id = block_id;
		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_RN_RCV_STRING_TABLE,
				cl_hton32((uint32_t)block_id),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);

		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C07: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_rn_sub_group_direction_mads(osm_sm_t *sm, osm_switch_t *p_sw,
					    boolean_t force_send,
					    cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint16_t block_id, max_block_id, offset;
	uint8_t *p_calc_data, *p_subn_data;

	if (!ar_is_rn_supported_enabled(&p_ar_data->ar_info))
		return;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	max_block_id = ar_get_top_subgroup_idx(p_ar_data) / IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;
	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;
		p_calc_data =
		    (uint8_t*)&p_ar_data->calc_tables.rn_subgroup_direction[offset];
		p_subn_data =
		    (uint8_t*)&p_ar_data->subn_tables.rn_subgroup_direction[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.rn_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.rn_mad_context.block_id = block_id;
		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_RN_SUB_GROUP_DIRECTION_TABLE,
				cl_hton32((uint32_t)block_id),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);

		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C08: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_rn_xmit_port_mask_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				       boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint16_t offset, block_id, max_block_id;
	uint8_t *p_calc_data, *p_subn_data;

	/* TODO: support PLFT */

	if (!ar_is_rn_supported_enabled(&p_ar_data->ar_info))
		return;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_physp)
		return;

	max_block_id = p_sw->num_ports / IB_RN_XMIT_PORT_MASK_BLOCK_ELEMENTS_NUMBER;
	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_RN_XMIT_PORT_MASK_BLOCK_SIZE;

		p_calc_data =
		    (uint8_t*)&p_ar_data->calc_tables.rn_xmit_port_mask[offset];
		p_subn_data =
		    (uint8_t*)&p_ar_data->subn_tables.rn_xmit_port_mask[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.rn_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.rn_mad_context.block_id = block_id;
		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_RN_XMIT_PORT_MASK,
				cl_hton32((uint32_t)block_id),
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);

		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C09: Failed to allocate memory\n");
			break;
		}
		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_plft_def_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				  boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint8_t *p_calc_data, *p_subn_data;
	osm_ar_tables_t *p_calc_tables, *p_subn_tables;
	uint8_t block_id, max_block_id;
	uint16_t offset;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_ar_data->num_plfts || !p_physp)
		return;

	max_block_id = (p_ar_data->num_plfts) / IB_PLFT_DEF_BLOCK_SIZE;
	p_calc_tables = &p_ar_data->calc_tables;
	p_subn_tables = &p_ar_data->subn_tables;

	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_PLFT_DEF_BLOCK_SIZE;

		/* Check if block already configured */
		p_calc_data = (uint8_t*)&p_calc_tables->plft_defs[offset];
		p_subn_data = (uint8_t*)&p_subn_tables->plft_defs[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
		context.ar_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.ar_mad_context.block_id = block_id & 0xf;

		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_PLFT_DEF,
				cl_hton32((uint32_t)(block_id & 0xf)),
				FALSE,
				ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C16: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static void get_port_sl_to_plft_mads(osm_sm_t *sm, osm_switch_t *p_sw,
				     boolean_t force_send, cl_qlist_t *p_list)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_madw_context_t context;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	uint8_t *p_calc_data, *p_subn_data;
	osm_ar_tables_t *p_calc_tables, *p_subn_tables;
	uint8_t block_id, max_block_id;
	uint16_t top_entry, offset;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
        if (!p_ar_data->num_plfts || !p_physp)
		return;

	top_entry = p_sw->num_ports;
	max_block_id = (top_entry) / IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;
	p_calc_tables = &p_ar_data->calc_tables;
	p_subn_tables = &p_ar_data->subn_tables;

	for (block_id = 0; block_id <= max_block_id; block_id++) {
		offset = block_id * IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

		/* Check if block already configured */
		p_calc_data = (uint8_t*)&p_calc_tables->port_sl_plfts[offset];
		p_subn_data = (uint8_t*)&p_subn_tables->port_sl_plfts[offset];
		if (!force_send &&
		    !memcmp(p_calc_data, p_subn_data, IB_SMP_DATA_SIZE))
			continue;

		context.ar_mad_context.request_type = OSM_AR_REQ_TYPE_SET;
		context.ar_mad_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
		context.ar_mad_context.block_id = block_id;

		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				p_calc_data, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_PORT_SL_TO_PLFT_MAP,
				cl_hton32((uint32_t)block_id),
				FALSE,
				ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7C17: Failed to allocate memory\n");
			break;
		}

		cl_qlist_insert_tail(p_list, &p_madw->list_item);
	}
}

static int ar_set_subn_rn_gen_sub_group_priority(osm_sm_t * sm)
{
	osm_subn_t *p_subn;
	cl_qlist_t mad_list;
	cl_map_item_t *p_map_item;
	osm_switch_t *p_sw;
	osm_ar_tables_t *p_tables;
	osm_physp_t *p_physp;
	osm_madw_context_t context;
	osm_madw_t *p_madw;
	uint8_t *payload;
	boolean_t force_send;

	OSM_LOG_ENTER(sm->p_log);

	p_subn = sm->p_subn;
	cl_qlist_init(&mad_list);
	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_sw = (osm_switch_t *)p_map_item;

		if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED ||
		    !ar_is_rn_supported(&p_sw->ar_data.ar_info))
			continue;

		p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	        if (!p_physp)
			continue;

		/* Avoid sending MAD if data did not change */
		force_send = p_subn->need_update || p_sw->need_update;
		if (!force_send &&
		    !memcmp(p_sw->ar_data.calc_tables.rn_subgroup_priority_flags,
			    p_sw->ar_data.subn_tables.rn_subgroup_priority_flags,
			    IB_SMP_DATA_SIZE))
			continue;

		context.rn_mad_context.node_guid =
		    osm_node_get_node_guid(p_sw->p_node);

		p_sw->ar_data.pending_mads++;
		p_tables = &p_sw->ar_data.calc_tables;
		payload = (uint8_t*)p_tables->rn_subgroup_priority_flags;
		p_madw = osm_prepare_req_set(
				sm, osm_physp_get_dr_path_ptr(p_physp),
				payload, IB_SMP_DATA_SIZE,
				IB_MAD_ATTR_RN_GEN_BY_SUB_GROUP_PRIORITY,
				0,
				FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				0, CL_DISP_MSGID_NONE, &context);
		if (!p_madw)
			continue;
		cl_qlist_insert_tail(&mad_list, &p_madw->list_item);
	}

	if (cl_qlist_count(&mad_list))
		osm_send_req_mad_batch(sm, &mad_list);

	OSM_LOG_ENTER(sm->p_log);

	return 0;
}

/*
 * Check and fix discrepancy between device capabilities and OpenSM
 * configuration, and device configuration.
 */
static void ar_check_misconfiguration(cl_map_item_t * const p_map_item,
				      void * context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	ar_configuration_params_t *p_ar_conf = &p_subn->ar_conf_params;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	ib_ar_info_t *p_ar_info = &p_ar_data->req_ar_info;
	uint8_t enable_features = p_ar_info->enable_features, i;
	uint32_t mask = 0xFFFFFFFF;

	/*
	 * Clear information of values assigned to switches that no longer
	 * configured with AR/FR.
	 * To clear subnet group table set them with invalid values.
	 */
	if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED) {
		ar_tables_clear(&p_ar_data->subn_tables, p_ar_data->num_groups,
				p_ar_data->num_group_subgroups,
				p_ar_data->num_strings,
				p_ar_data->num_directions,
				p_ar_data->num_plfts,
				p_ar_data->num_lids,
				p_sw->num_ports,
				AR_INVALID_GROUP);
		return;
	}

	if (p_sw->ar_data.status != AR_STATUS_SUCCESS)
		return;

	/* Check that AR capabilities are enabled */
	if (!(enable_features & IB_AR_INFO_FEATURE_MASK_AR_ENABLED)) {
		p_sw->ar_configured = OSM_SW_AR_CONF_NONE;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7C11: Switch 0x%" PRIx64 " is not configured "
			" with AR enabled\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		return;
	}

	/* Enforce FRN/ARN disablement by changing RnSubGroupDirectionTable */
	if (!(enable_features & IB_AR_INFO_FEATURE_MASK_ARN_SUPPORTED)||
	    p_ar_conf->ar_mode != OSM_ADV_ROUTING_MODE_RN) {
		mask &= cl_hton32(~0x1);
	}
	if (!(enable_features & IB_AR_INFO_FEATURE_MASK_FR_ENABLED) ||
	    !(enable_features & IB_AR_INFO_FEATURE_MASK_FRN_SUPPORTED) ||
	    p_ar_conf->shield_mode != OSM_ADV_ROUTING_MODE_RN) {
		mask &= cl_hton32(~0x2);
	}
	for (i = 0; i < p_ar_data->num_group_subgroups; i++) {
		if (~mask &
		    p_ar_data->calc_tables.rn_subgroup_priority_flags[i]) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7C12: Disabling mis-configured routing"
				" notification on switch 0x%" PRIx64
				" subgroup priority %d\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
				i);
			p_ar_data->calc_tables.rn_subgroup_priority_flags[i] &=
			    mask;
		}
	}
}

int osm_ar_set_subn_ar_tables(osm_sm_t *sm)
{
	cl_qlist_t *p_lists, *p_list;
	cl_map_item_t *p_map_item;
	osm_subn_t *p_subn = sm->p_subn;
	int status = 0;

	OSM_LOG_ENTER(sm->p_log);

	p_lists = (cl_qlist_t *)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
				       sizeof(cl_qlist_t));
	if (!p_lists) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7C10: Failed to allocate memory\n");
		status = 1;
		goto exit;
	}

	p_list = p_lists;
	for (p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		cl_qlist_init(p_list++);
	}

	/*
	 * Clear information from devices that no longer configured with AR/FR,
	 * and enforce subnet configuration settings on the devices.
	 */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, ar_check_misconfiguration,
			   p_subn);

	/* Set adaptive routing configuration */
	if (ar_set_subn_ar_info(sm)) {
		status = 1;
		goto exit;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Set PLFT info */
	if (ar_set_subn_plft_info(sm)) {
		status = 1;
		goto exit_no_wait;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Set PLFT defs */
	if (ar_set_subn_tables(sm, p_lists, get_plft_def_mads,
			       SUBN_TABLES_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	/* Set ExtendedNodeInfo SL2VLAct */
	if (ar_set_subn_extended_node_info(sm)) {
		status = 1;
		goto exit_no_wait;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Set Port SL to PLFT map */
	if (ar_set_subn_tables(sm, p_lists, get_port_sl_to_plft_mads,
			       SUBN_TABLES_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Set adaptive routing groups */
	if (ar_set_subn_tables(sm, p_lists, get_group_table_mads,
			       SUBN_TABLES_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Copy adaptive routing groups */
	if (ar_set_subn_tables(sm, p_lists, get_group_table_copy_mads,
			       SUBN_TABLES_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
		goto exit_no_wait;
	}

	/* Set adaptive routing forwarding tables */
	if (ar_set_subn_tables(sm, p_lists, get_ar_lft_mads,
			       SUBN_TABLES_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	/* Set PLFTs top */
	if (ar_set_subn_tables(sm, p_lists, get_plft_map_mads,
			       SUBN_TABLES_DONT_UPDATE_AR_STATUS)) {
		status = 1;
		goto exit;
	}

	/*
	 * When sending a bunch of different mads the status can't be updated
	 * according to the pending_mads because there where no wait_for_pending_transactions
	 * between them and there is no guarantee the previous mads returned.
	 * This is why the status is updated only when there is a
	 * wait_for_pending_transactions after the previous mad.
	 */

	/* set routing notification tables */
	if (ar_is_rn_enabled(sm->p_subn)) {
		if ((ar_set_subn_tables(sm, p_lists, get_rn_gen_string_mads,
					SUBN_TABLES_DONT_UPDATE_AR_STATUS) ||
		     ar_set_subn_tables(sm, p_lists, get_rn_rcv_string_mads,
					SUBN_TABLES_DONT_UPDATE_AR_STATUS) ||
		     ar_set_subn_tables(sm, p_lists, get_rn_sub_group_direction_mads,
					SUBN_TABLES_DONT_UPDATE_AR_STATUS) ||
		     ar_set_subn_tables(sm, p_lists, get_rn_xmit_port_mask_mads,
					SUBN_TABLES_DONT_UPDATE_AR_STATUS))) {
			status = 1;
			goto exit;
		}
		/* Define which subgroup priority generates notifications */
		if (ar_set_subn_rn_gen_sub_group_priority(sm)) {
			status = 1;
			goto exit;
		}
	}

exit:
	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		status = 1;
	}

exit_no_wait:
	if (p_lists)
		free(p_lists);

	OSM_LOG_EXIT(sm->p_log);

	return status;
}

int osm_ar_groups_set_block(osm_advanced_routing_data_t *p_ar_data,
			    uint16_t block_id,
			    ib_ar_group_table_t *p_block)
{
	uint16_t offset = block_id * 2;

	if (offset > ar_get_top_subgroup_idx(p_ar_data))
		return 1;

	memcpy(&p_ar_data->subn_tables.ar_subgroups[offset], p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_lft_set_block(osm_advanced_routing_data_t *p_ar_data,
			 uint8_t plft_id,
			 uint16_t block_id,
			 ib_ar_lft_t *p_block)
{
	uint16_t offset = block_id * IB_AR_LFT_BLOCK_SIZE;

	if (offset >= p_ar_data->num_lids)
		return 1;

	memcpy(&(p_ar_data->subn_tables.ar_lft[plft_id][offset]), p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_rn_gen_string_set_block(osm_advanced_routing_data_t *p_ar_data,
				   int16_t block_id,
				   ib_rn_gen_string_table_t *p_block)
{
	uint16_t offset = block_id * IB_RN_GEN_STRING_BLOCK_SIZE;

	if (offset >= p_ar_data->num_directions)
		return 1;

	memcpy(&p_ar_data->subn_tables.rn_direction_string[offset], p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_rn_rcv_string_set_block(osm_advanced_routing_data_t *p_ar_data,
				   int16_t block_id,
				   ib_rn_rcv_string_table_t *p_block)
{
	uint16_t offset = block_id * IB_RN_RCV_STRING_BLOCK_SIZE;

	if (offset >= p_ar_data->num_strings)
		return 1;

	memcpy(&p_ar_data->subn_tables.rn_rcv_strings[offset], p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_rn_sub_group_direction_set_block(osm_advanced_routing_data_t *p_ar_data,
					    int16_t block_id,
					    ib_rn_sub_group_direction_table_t *p_block)
{
	uint16_t offset = block_id * IB_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;

	if (offset > ar_get_top_subgroup_idx(p_ar_data))
		return 1;

	memcpy(&p_ar_data->subn_tables.rn_subgroup_direction[offset], p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_rn_xmit_port_mask_set_block(osm_advanced_routing_data_t *p_ar_data,
				       int16_t block_id,
				       ib_rn_xmit_port_mask_t *p_block)
{
	uint16_t offset = block_id * IB_RN_XMIT_PORT_MASK_BLOCK_SIZE;

	memcpy(&p_ar_data->subn_tables.rn_xmit_port_mask[offset],
	       p_block,
	       IB_SMP_DATA_SIZE);

	return 0;
}

int osm_ar_rn_subgroup_priority_flags_set_block(osm_advanced_routing_data_t *p_ar_data,
						ib_rn_gen_by_sub_group_priority_t *p_block)
{
	memcpy(&p_ar_data->subn_tables.rn_subgroup_priority_flags, p_block,
	       IB_SMP_DATA_SIZE);

	return 0;
}

int osm_ar_plft_defs_set_block(osm_advanced_routing_data_t *p_ar_data,
			       uint8_t block_id,
			       ib_plft_def_t *p_block)
{
	OSM_ASSERT(block_id <= 0xF);

	uint8_t offset = block_id * IB_PLFT_DEF_BLOCK_SIZE;

	//TODO check that offset doesn't exceed the max LFT num supported by this device
	memcpy(&p_ar_data->subn_tables.plft_defs[offset], p_block,
	       IB_SMP_DATA_SIZE);
	return 0;
}

int osm_ar_port_sl_to_plft_set_block(osm_switch_t *p_sw,
				     uint8_t block_id,
				     boolean_t subnet,
			             ib_port_sl_to_plft_map_t *p_block)
{
	uint8_t offset = block_id * IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

	if (offset >= p_sw->num_ports)
		return 1;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_tables_t *p_tables = subnet ? &p_ar_data->subn_tables : &p_ar_data->calc_tables;

	memcpy(&p_tables->port_sl_plfts[offset], p_block,
	       sizeof(ib_port_sl_to_plft_map_t));
	return 0;
}

int osm_ar_plft_map_set_block(osm_advanced_routing_data_t *p_ar_data,
			      uint8_t plft_id,
			      ib_plft_map_t *p_block)
{
	OSM_ASSERT(plft_id < AR_MAX_NUM_PLFT);

	/* TODO: add support for FDB Port Bitmask */

	if (p_block->control_map & IB_PLFT_MAP_CONTROL_MAP_LFT_TOP_ENABLE)
		p_ar_data->subn_tables.plft_top[plft_id] = cl_ntoh16(p_block->lft_top);

	return 0;
}

int osm_ar_cmp_ar_lft(osm_advanced_routing_data_t *p_ar_data)
{
	size_t ar_lft_size, ar_subgroups_size;
	uint8_t arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;
	uint8_t plft_idx;
	int result = 0;

	ar_lft_size = AR_ROUNDUP_BLOCK(p_ar_data->num_lids, IB_AR_LFT_BLOCK_SIZE) *
		      sizeof(ib_ar_lft_entry_t);
	ar_subgroups_size = (ar_get_top_subgroup_idx(p_ar_data) + 1) *
			    sizeof(osm_ar_subgroup_t);

	for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx++) {
		result = memcmp(p_ar_data->subn_tables.ar_lft[plft_idx],
				p_ar_data->calc_tables.ar_lft[plft_idx],
				ar_lft_size);
		if (result)
			break;
	}

	return (result ||
		memcmp(p_ar_data->subn_tables.ar_subgroups,
		       p_ar_data->calc_tables.ar_subgroups,
		       ar_subgroups_size));
}

/* Extend AR LFT tables if needed, without clearing their content */
int osm_ar_reallocate_lft_tables(osm_advanced_routing_data_t *p_ar_data, uint16_t num_lids,
				 boolean_t clear_all_calculated)
{
	uint8_t plft_idx, arlft_num_plfts;

	arlft_num_plfts = p_ar_data->num_plfts ? p_ar_data->num_plfts : 1;

	for (plft_idx = 0; plft_idx < arlft_num_plfts; plft_idx ++) {
		if (ar_reallocate_lft_table(&p_ar_data->calc_tables.ar_lft,
					    plft_idx,
					    p_ar_data->num_lids, num_lids,
					    clear_all_calculated) ||
		    ar_reallocate_lft_table(&p_ar_data->subn_tables.ar_lft,
					    plft_idx,
					    p_ar_data->num_lids, num_lids,
					    FALSE)) {
			return -1;
		}
	}
	p_ar_data->num_lids = num_lids;
	return 0;
}

/*
 * ar_group_table_copy_obj_t describes source group to copy from.
 */
typedef struct ar_group_table_copy_obj {
	cl_fmap_item_t map_item;
	osm_ar_subgroup_t *p_key;
	uint8_t key_size;
	uint16_t copy_from_group_id;
} ar_group_table_copy_obj_t;

void osm_ar_invalidate_ar_cache_params(osm_switch_t *p_sw)
{
	/* free the data in all the tables */
	osm_ar_data_destroy(&p_sw->ar_data);
	p_sw->ar_data.ar_info_updated = FALSE;
	p_sw->ar_data.plft_info_updated = FALSE;
	/*
	 * reset ar data parameters that affect the reinitialization
	 * of ar data tables
	 */
	p_sw->ar_data.num_groups = 0;
	p_sw->ar_data.num_group_subgroups = 0;
	p_sw->ar_data.num_lids = 0;
	p_sw->ar_data.num_strings = 0;
	p_sw->ar_data.num_directions = 0;
	p_sw->ar_data.num_plfts = 0;
	p_sw->ar_data.group_top = 0;
}

void osm_ar_print_subgroups(FILE * file, struct osm_switch *p_sw)
{
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	unsigned int sub_group, group, port;
	osm_ar_subgroup_t *p_subgroup;

	fprintf(file,
		"Groups Definition:\n\n"
		"Group     Sub Group      Ports\n"
		"-------------------------------------\n");

	for (group = 0; group <= p_ar_data->group_top; group++) {
		for (sub_group = 0; sub_group < p_ar_data->num_group_subgroups; sub_group++) {
			p_subgroup = osm_ar_subgroup_get(p_ar_data, group, sub_group);
			fprintf(file, "%-15u%-10u", group, sub_group);
			/* print ports in the current calculated subgroup */
			for (port = 1; port < p_sw->num_ports; port++) {
				if (ar_is_port_in_subgroup(p_subgroup, port)) {
					fprintf(file, "%-u, ", port);
				}
			}
			fprintf(file, "\n");
		}
	}
}
