/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2018 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * 	Common an2an function implementations.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AN2AN_C
#include <opensm/osm_an2an.h>
#include <opensm/osm_db_pack.h>

static uint16_t get_sw_an_lid(osm_switch_t *p_sw)
{
	osm_physp_t *p_sw_physp, *p_an_physp;

	if (p_sw->an_port_num == OSM_NO_PATH)
		return 0;

	p_sw_physp = osm_node_get_physp_ptr(p_sw->p_node, p_sw->an_port_num);
	p_an_physp = p_sw_physp->p_remote_physp;
	if (!p_an_physp)
		return 0;

	return cl_ntoh16(osm_physp_get_base_lid(p_an_physp));
}

boolean_t osm_is_same_physical_link(osm_sm_t *p_sm,
				    osm_switch_t *p_sw1, osm_switch_t *p_sw2,
				    uint8_t port_num1, uint8_t port_num2)
{
	osm_physp_t *p_physp1, *p_physp2;

	if (port_num1 >= osm_node_get_num_physp(p_sw1->p_node) ||
	    port_num2 >= osm_node_get_num_physp(p_sw2->p_node)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": cached link (%u, %u) is not exists in current setup\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			port_num1, port_num2);
	    return FALSE;
	}

	p_physp1 = osm_node_get_physp_ptr(p_sw1->p_node, port_num1);
	p_physp2 = osm_node_get_physp_ptr(p_sw2->p_node, port_num2);

	if (!p_physp1 || !p_physp2 ||
	    !p_physp1->p_remote_physp || !p_physp1->p_remote_physp) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": cached link (%u, %u) is down\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			port_num1, port_num2);
		return FALSE;
	}

	if (p_physp1->p_remote_physp != p_physp2) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": cached link (%u, %u) not a single physical link\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			port_num1, port_num2);
		return FALSE;
	}

	return TRUE;
}

static void force_same_physical_link(osm_sm_t *p_sm,
				     osm_switch_t *p_sw1, osm_switch_t *p_sw2,
			             uint8_t port_num1, uint8_t *port_num2)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw1->p_node, port_num1);
	uint8_t new_port_num2 = p_physp->p_remote_physp->port_num;

	if (*port_num2 == new_port_num2) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": using current link (%u, %u)\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			port_num1, *port_num2);
	} else {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": fixing link (%u, %u) ==> (%u, %u)\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			port_num1, *port_num2, port_num1, new_port_num2);
		*port_num2 = new_port_num2;
	}

}

static ib_api_status_t
fixate_an2an_link(osm_sm_t *p_sm, osm_switch_t *p_sw1, osm_switch_t *p_sw2,
		  uint8_t cache_port_num1, uint8_t cache_port_num2,
		  uint8_t *port_num1, uint8_t *port_num2)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_physp_t *p_physp1;
	boolean_t is_in_cache = (cache_port_num1 != OSM_NO_PATH) &&
				(cache_port_num2 != OSM_NO_PATH);

	if (is_in_cache &&
	    osm_is_same_physical_link(p_sm, p_sw1, p_sw2,
				      cache_port_num1, cache_port_num2)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
			": using cached link (%u, %u)\n",
			cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
			cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
			cache_port_num1, cache_port_num2);
		*port_num1 = cache_port_num1;
		*port_num2 = cache_port_num2;
	} else {
		p_physp1 = osm_node_get_physp_ptr(p_sw1->p_node, *port_num1);
		if (!p_physp1 || !p_physp1->p_remote_physp) {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
				"An2an link 0x%016" PRIx64 " <==> 0x%016" PRIx64
				": link (%u, %u) is down\n",
				cl_hton64(osm_node_get_node_guid(p_sw1->p_node)),
				cl_hton64(osm_node_get_node_guid(p_sw2->p_node)),
				*port_num1, *port_num2);
			status = IB_ERROR;
			goto Exit;
		}

		force_same_physical_link(p_sm, p_sw1, p_sw2,
					 *port_num1, port_num2);
	}

Exit:
	return status;
}

static void
validate_an2an_link(osm_sm_t *p_sm, osm_switch_t *p_sw1, osm_switch_t *p_sw2)
{
	ib_net64_t sw_guid1 = osm_node_get_node_guid(p_sw1->p_node);
	ib_net64_t sw_guid2 = osm_node_get_node_guid(p_sw2->p_node);
	uint16_t an_lid1 = get_sw_an_lid(p_sw1);
	uint16_t an_lid2 = get_sw_an_lid(p_sw2);
	uint8_t port_num1 = p_sw1->new_lft[an_lid2];
	uint8_t port_num2 = p_sw2->new_lft[an_lid1];
	uint8_t cache_port_num1, cache_port_num2;
	char p_time[MAX_DATE_STR_LEN] = { 0 };

	/*
	 *  Check that there is routing from each switch to aggregation node of peer switch.
	 *  If there is no routing, assume aggregation node is not healthy / unreachable.
	 */
	if (port_num1 == OSM_NO_PATH || port_num2 == OSM_NO_PATH)
	        return;

	/* Handle each link only once */
	if (sw_guid1 > sw_guid2)
		return;

	/* Get cached link */
	if (osm_db_an2an_get(p_sm->p_subn->p_an2an, sw_guid1, sw_guid2,
			     &cache_port_num1, &cache_port_num2, p_time) == 1) {
		cache_port_num1 = OSM_NO_PATH;
		cache_port_num2 = OSM_NO_PATH;
	}

	if (fixate_an2an_link(p_sm, p_sw1, p_sw2, cache_port_num1, cache_port_num2,
			      &port_num1, &port_num2) != IB_SUCCESS)
		return;

	/* Update cache */
	if ((port_num1 != cache_port_num1) || (port_num2 != cache_port_num2)
	    || (!strcmp(p_time, "time")))
		osm_db_an2an_set(p_sm->p_subn->p_an2an, sw_guid1, sw_guid2,
				 port_num1, port_num2, FALSE);

	/* Update lfts */
	p_sw1->new_lft[an_lid2] = port_num1;
	p_sw2->new_lft[an_lid1] = port_num2;

	/* if AR is configured, update AR LFTs */
	if (p_sw1->ar_configured == OSM_SW_AR_CONF_CONFIGURED)
		osm_ar_lft_set(&p_sw1->ar_data, an_lid2, ADV_ROUTING_ALL_PLFTS,
			       port_num1, OSM_AR_IB_LID_STATE_STATIC, 0);

	if (p_sw2->ar_configured == OSM_SW_AR_CONF_CONFIGURED)
		osm_ar_lft_set(&p_sw2->ar_data, an_lid1, ADV_ROUTING_ALL_PLFTS,
			       port_num2, OSM_AR_IB_LID_STATE_STATIC, 0);
}

static void
validate_an2an_links_for_sw(cl_map_item_t * const p_map_item, void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_switch_t *p_neighbor_sw;
	osm_sm_t *p_sm = (osm_sm_t *)context;
	int i, j, num_neighbors = 0;
	int num_ports = p_sw->p_node->physp_tbl_size;
	osm_switch_t *neighbors[256];

	if (p_sw->an_port_num == OSM_NO_PATH)
		return;

	/* Get a list of unique neighbors */
	for (i = 0; i < num_ports; i++) {
		osm_physp_t *p_physp = &p_sw->p_node->physp_table[i];
		boolean_t neighbor_already_added = FALSE;

		p_neighbor_sw = p_physp->p_remote_physp ?
				p_physp->p_remote_physp->p_node->sw : NULL;

		if (!p_neighbor_sw ||
		    (p_neighbor_sw == p_sw) ||
		    (p_neighbor_sw->an_port_num == OSM_NO_PATH))
			continue;

		for (j = 0; j < num_neighbors; j++)
			if (neighbors[j] == p_neighbor_sw) {
				neighbor_already_added = TRUE;
				break;
			}

		if (!neighbor_already_added)
			neighbors[num_neighbors++] = p_neighbor_sw;
	}

	/* Validate */
	for (i = 0; i < num_neighbors; i++)
		validate_an2an_link(p_sm, p_sw, neighbors[i]);
}

static boolean_t
is_link_rewired(osm_sm_t *p_sm, uint64_t guid1, uint8_t port_num1,
		uint64_t guid2, uint8_t port_num2)
{
	osm_node_t *p_node1, *p_node2;
	osm_physp_t *p_physp1 = NULL, *p_physp2 = NULL;

	p_node1 = osm_get_node_by_guid(p_sm->p_subn, guid1);
	p_node2 = osm_get_node_by_guid(p_sm->p_subn, guid2);

	 if (p_node1 && port_num1 < osm_node_get_num_physp(p_node1))
		p_physp1 = osm_node_get_physp_ptr(p_node1, port_num1);

	 if (p_node2 && port_num2 < osm_node_get_num_physp(p_node2))
		p_physp2 = osm_node_get_physp_ptr(p_node2, port_num2);

	return (p_physp1 && p_physp1->p_remote_physp &&
	       (p_physp1->p_remote_physp != p_physp2)) ||
	       (p_physp2 && p_physp2->p_remote_physp &&
	       (p_physp2->p_remote_physp != p_physp1));
}

static
boolean_t is_valid_an2an_entry(osm_subn_t *p_subn,
			       osm_db_an2an_elem_t *p_item, osm_sm_t *p_sm)
{
	osm_switch_t *p_sw1 = osm_get_switch_by_guid(p_subn, p_item->guid1);
	osm_switch_t *p_sw2 = osm_get_switch_by_guid(p_subn, p_item->guid2);
	return (p_sw1 && p_sw2 &&
	        osm_is_same_physical_link(p_sm, p_sw1, p_sw2,
				          p_item->portnum1, p_item->portnum2));
}

static void an2an_garbage_collect(osm_sm_t *p_sm)
{
	cl_qlist_t entries;
	osm_db_an2an_elem_t *p_item;

	cl_qlist_init(&entries);

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
		"Performing an2an garbage collection\n");

	if (osm_db_an2an_get_all(p_sm->p_subn->p_an2an, &entries)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 3333: "
			"could not get an2an entry list\n");
		return;
	}

	while ((p_item =
		(osm_db_an2an_elem_t *) cl_qlist_remove_head(&entries))
	       != (osm_db_an2an_elem_t *) cl_qlist_end(&entries)) {
		if (is_link_rewired(p_sm,
				    p_item->guid1, p_item->portnum1,
				    p_item->guid2, p_item->portnum2)) {

			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
				"An2an link 0x%016" PRIx64 ":%u"
				" - 0x%016" PRIx64 ":%u rewired\n",
				cl_ntoh64(p_item->guid1), p_item->portnum1,
				cl_ntoh64(p_item->guid2), p_item->portnum2);

			if (osm_db_an2an_delete(p_sm->p_subn->p_an2an,
						p_item->guid1, p_item->guid2))
				OSM_LOG(p_sm->p_log, OSM_LOG_ERROR,
					"ERR 3336: failed to delete an2an entry"
					" (0x%016" PRIx64 ", 0x%016" PRIx64 ")\n",
					cl_ntoh64(p_item->guid1),
					cl_ntoh64(p_item->guid2));
		} else if(!is_valid_an2an_entry(p_sm->p_subn, p_item, p_sm))
				osm_db_an2an_set(p_sm->p_subn->p_an2an,
						 p_item->guid1, p_item->guid2,
						 p_item->portnum1, p_item->portnum2,
						 TRUE);
		free(p_item);
	}
	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
		"An2an garbage collection done\n");

}

void osm_validate_an2an_links(osm_sm_t *p_sm)
{
	OSM_LOG_ENTER(p_sm->p_log);

	cl_qmap_apply_func(&p_sm->p_subn->sw_guid_tbl,
			   validate_an2an_links_for_sw, p_sm);
	an2an_garbage_collect(p_sm);

	OSM_LOG_EXIT(p_sm->p_log);
}
