/*
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include "osm_ar_algorithm.hpp"
#include "osm_armgr.hpp"

int AdaptiveRoutingAlgorithm::BuildLidMapping(LidMapping &lid_mapping)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "BuildLidMapping\n");

	int rc = 0;
	osm_node_t *p_osm_node;
	cl_qmap_t *p_node_guid_tbl = &m_ar_mgr_.m_p_osm_subn->node_guid_tbl;
	u_int16_t base_lid;
	/* Go over all nodes in osm database */
	for (p_osm_node =  (osm_node_t *) cl_qmap_head(p_node_guid_tbl);
		 p_osm_node != (osm_node_t *) cl_qmap_end(p_node_guid_tbl);
		 p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

		if (p_osm_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
			base_lid = cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0));
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Map Switch LID: %u\n", base_lid);

			lid_mapping.m_lid_to_sw_lid_mapping[base_lid] = base_lid;
			lid_mapping.m_lid_to_base_lid_mapping[base_lid] = base_lid;
			continue;
		}

		for (u_int8_t port_num = 1;
			 port_num <= p_osm_node->node_info.num_ports; port_num++) {

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			osm_node_t *p_remote_node =
				osm_node_get_remote_node(p_osm_node, port_num, NULL);

			if (!p_physp || !p_remote_node || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			rc |= SetHcaLidMapping(p_physp, p_remote_node, lid_mapping);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int AdaptiveRoutingAlgorithm::SetHcaLidMapping(osm_physp_t *p_hca_physp,
					       osm_node_t *p_remote_sw_node,
					       LidMapping &lid_mapping)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	u_int16_t node_lid = cl_ntoh16(osm_physp_get_base_lid(p_hca_physp));
	if (p_remote_sw_node->node_info.node_type != IB_NODE_TYPE_SWITCH ||
		node_lid == 0) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	uint8_t lmc = osm_physp_get_lmc(p_hca_physp);

	u_int16_t remote_node_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_sw_node, 0));

	u_int16_t numLidsPerPort = (u_int16_t)((u_int16_t)1 << lmc);
	u_int16_t lid_limit = node_lid+numLidsPerPort;

	for (u_int16_t l = node_lid; l < lid_limit; l++) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Map (HCA/Router) LID: %u base LID: %u lmc:%u to SW LID: %u\n",
			   l, node_lid, lmc, remote_node_lid);

		lid_mapping.m_lid_to_sw_lid_mapping[l] = remote_node_lid;
		lid_mapping.m_lid_to_base_lid_mapping[l] = node_lid;
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}
