/*
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#if HAVE_CONFIG_H
#include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_ENGINES_CPP
#include <opensm/osm_log.h>
#include <opensm/osm_port.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_node.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_armgr.hpp>
#include <opensm/osm_an2an.h>

/* TODO: Do we need to move to header file? */
extern "C" int osm_ucast_hc_process(osm_subn_t * p_subn);

struct ar_engine_context {
	osm_ucast_mgr_t *p_ucast_mgr;
	OSMAdaptiveRoutingManager *p_armgr;
};

static void ar_update_osm_sl2vl(void *context, osm_physp_t *osm_phys_port,
				uint8_t iport_num, uint8_t oport_num,
				ib_slvl_table_t *osm_oport_sl2vl)
{
	struct ar_engine_context *ctx = (struct ar_engine_context *)context;
	OSMAdaptiveRoutingManager *p_armgr = ctx->p_armgr;
	uint8_t sl2vl[IB_NUMBER_OF_SLS] = {0}, sl, vl;

	if (p_armgr->GetSL2VL(osm_phys_port, iport_num, oport_num, &sl2vl))
		return;

	for (sl = 0; sl < IB_NUMBER_OF_SLS; sl++) {
		vl = sl2vl[sl];
		ib_slvl_table_set(osm_oport_sl2vl, sl, vl);
	}
}

static
void ar_engine_context_delete(void *context)
{
	struct ar_engine_context *ctx = (struct ar_engine_context *)context;
	OSMAdaptiveRoutingManager * p_armgr = ctx->p_armgr;
	delete p_armgr;

	free(ctx);
}

static struct ar_engine_context *ar_engine_context_create(
	osm_opensm_t *osm,
	ar_algorithm_t ar_algorithm)
{
	struct ar_engine_context *ctx;

	ctx = (struct ar_engine_context*)calloc(1, sizeof(*ctx));
	if (ctx) {
		ctx->p_ucast_mgr = &osm->sm.ucast_mgr;
		ctx->p_armgr = new OSMAdaptiveRoutingManager(osm, ar_algorithm);
		if (ctx->p_armgr)
			ctx->p_armgr->Construct();
	} else
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
				"ERR 9401: calloc: %s\n", strerror(errno));

	return ctx;
}

/*
* KDOR_HC
*/
/* TODO: should be general method: ar_build_lfts */
static int kdor_hc_build_lfts(void *context)
{
	struct ar_engine_context *ctx = (struct ar_engine_context*)context;
	osm_ucast_mgr_t *p_ucast_mgr = ctx->p_ucast_mgr;
	OSMAdaptiveRoutingManager *p_armgr = ctx->p_armgr;
	int result = 0;

	/*
	TODO: !!! call to build osm lft_tables e.g. osm_ucast_mgr_build_lft_tables
	should be algorithm specific done by Route()
	*/

	result = p_armgr->Route();
	return result;
}

static int kdor_hc_build_lid_matrices(void *context)
{
	int result;
	struct ar_engine_context *ctx = (struct ar_engine_context*)context;
	osm_ucast_mgr_t *p_ucast_mgr = ctx->p_ucast_mgr;

	result = osm_ucast_mgr_build_lid_matrices(p_ucast_mgr);
	if (result) {
		return result;
	}

	osm_ucast_hc_process(p_ucast_mgr->p_subn);
	return result;
}

extern "C" int osm_ucast_kdor_hc_setup(struct osm_routing_engine *r,
				       osm_opensm_t *osm)
{
	void *ctx;

	ctx = ar_engine_context_create(osm, AR_ALGORITHM_KDOR_HC);
	if (!ctx)
		return -1;

	r->context = ctx;
	r->ucast_build_fwd_tables = kdor_hc_build_lfts;
	r->build_lid_matrices = kdor_hc_build_lid_matrices;
	r->update_sl2vl = ar_update_osm_sl2vl;

	r->destroy = ar_engine_context_delete;
	r->type = OSM_ROUTING_ENGINE_TYPE_KDOR_HC;
	return 0;
}
