/*
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_KDOR_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>

#include "opensm/osm_ar_kdor_algorithm.hpp"
#include "osm_armgr.hpp"

#define MIN_KDOR_VL_CAP 3

extern "C" int osm_ucast_dor_build_lfts(void *context);

int ArKdorAlgorithm::Preprocess()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;

	rc = Init();
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/* TODO: check if can be done before Init */
	if (!SetCapable()) {
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	int unsupported_num = m_ar_mgr_.ARInfoGetGroupCapProcess();
	if (unsupported_num) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
			   "%d switches do not support kDOR Cap.\n" ,
			   unsupported_num);
	}

	/* Validate switch capabilities */
	rc = m_ar_mgr_.ValidateSwitchesCapabilites();
	if (rc) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9201: Failed to validate the required ar capabilities "
			   "for advanced routing configuration\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/*
	 * Calls osm_ar_prepare_tables,
	 * therefor it must be called after ValidateSwitchesCapabilites
	 */
	SetPlftConfiguration();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ArKdorAlgorithm::Preprocess Ended. rc:%d\n", rc);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::RunCycle()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc = 0;

	for (int i=0 ; i<1 ; i++) {

		if (Preprocess()) {
			rc = -1;
			break;
		}

		rc = BuildStaticRouteInfo();
		if (rc) {
			m_ar_mgr_.m_is_permanent_error = true;
			break;
		}

		rc = CalculatePortGroups();
		if (rc) {
			break;
		}

		BuildSl2VlPerOpVl(m_ar_mgr_.m_master_db.m_en_sl_mask);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArKdorAlgorithm cycle calculation Ended.\n");

		int unsupported_num = m_ar_mgr_.ARInfoSetProcess();
		if (unsupported_num) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9202: %d switches do not support AR Set.\n" ,
				   unsupported_num);
			m_ar_mgr_.m_is_temporary_error = true;
		}

		SetPlftMap();
	}

	CycleEnd(rc);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ArKdorAlgorithm::RunCycle Ended. rc:%d\n", rc);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::Init()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "Init kDOR Algorithm.\n");

	m_ar_mgr_.m_is_permanent_error = false;
	m_ar_mgr_.m_is_temporary_error = false;

	memset(m_sw_lid_to_kdor_data_, 0, sizeof(m_sw_lid_to_kdor_data_));

	uint32_t sw_idx = 0;
	uint32_t switches_number = (uint32_t)m_sw_map_.size();
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (sw_node_it->second.m_kdor_data == NULL) {
			/* Allocate KDOR Data for all switches */
			sw_node_it->second.m_kdor_data = new KdorSwData(
				sw_node_it->second, &m_ar_mgr_.m_port_groups_calculator);
			if (sw_node_it->second.m_kdor_data == NULL) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9203: Failed to allocate kDOR Data.\n");
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}

			sw_node_it->second.m_kdor_data->Resize(
				sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node->physp_tbl_size);
		}

		int rc = sw_node_it->second.m_kdor_data->Init(
			*this, sw_node_it->second, sw_idx, switches_number);

		if (rc) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9204: Failed to init kDOR Data. kDOR failed.\n");
			OSM_AR_LOG_RETURN(m_p_osm_log, rc);
		}

		m_sw_lid_to_kdor_data_[sw_node_it->second.m_general_sw_info.m_lid] =
			sw_node_it->second.m_kdor_data;
		sw_idx++;

		if (sw_node_it->second.m_osm_update_needed) {
			sw_node_it->second.m_kdor_data->m_plft_max_supported_lid_number = 0;
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/* SetCapable also sets calculated VL as SL */
bool ArKdorAlgorithm::SetCapable()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable algorithm feature: %u\n",
		   m_algorithm_feature_);

	int unsupported_num = 0;
	int disable_by_tr_err = 0;
	bool new_dev = false;
	bool algorithm_supported = true;

	if (m_ar_mgr_.m_p_osm_subn->opt.qos != TRUE) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9205: KDOR Algorithm not supported "
			   "together with disabled qos.\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}


	if (m_ar_mgr_.m_p_osm_subn->opt.max_op_vls == 0) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9206: Unsupported max_op_vls configuration parameter "
			   "value equals zero for kDOR setup.\n");

		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}
	uint8_t num_vls = (1 << (m_ar_mgr_.m_p_osm_subn->opt.max_op_vls - 1));

	if (num_vls < m_min_vl_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9207: Invalid maximal operational VLs %u < %u "
			   "configuration for kDOR setup.\n",
			   num_vls, m_min_vl_number_);

		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.m_osm_update_needed) {
			sw_node_it->second.m_support[m_algorithm_feature_] = SUPPORT_UNKNOWN;
		} else if (sw_node_it->second.m_support[m_algorithm_feature_] == NOT_SUPPORTED) {
			unsupported_num++;
			continue;
		}

		if (!m_ar_mgr_.IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip Switch GUID: 0x%" PRIx64 ", LID: %u - " \
				   "AR not supported or not enabled.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			unsupported_num++;
			continue;
		}

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (sw_node_it->second.m_general_sw_info.m_p_osm_sw->coord ==
			OSM_SW_NO_COORD) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9208: "
					   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
					   "do not have coordinate. "
					   "Cannot guarantee credit loop free.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
		}

		if (sw_node_it->second.m_support[m_algorithm_feature_] == SUPPORTED)
			continue;

		new_dev = true;

		/*
		 * All switches should support vl2vl configuration
		 * ( ExtendedNodeInfo  SL2VLCap  bit 0 is on)
		 */
		osm_switch_t  *p_osm_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_node_t *p_osm_node = p_osm_sw->p_node;
		osm_advanced_routing_data_t *p_ar_data = &p_osm_sw->ar_data;

		if (!(p_osm_node->ext_node_info.sl_2_vl_cap & 0x1)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9209: "
				   "Switch GUID 0x%" PRIx64 ", LID %u, "  \
				   "does not support VL2VL. kDOR algorithm not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_VL2VL_CAP);
			algorithm_supported = false;
		} else {
			p_osm_node->new_ext_node_info.sl_2_vl_act = 1; /* Set VL as SL */
		}

		if ((m_ar_mgr_.m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) &&
			!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920A: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
				   "does not support by_transport_disable.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			disable_by_tr_err++;
		}

		/* All switches should support adaptive Routing checked by ARInfoProcess */
		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920B: "
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "BySLCap not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_AR_BY_SL_CAP);
			unsupported_num++;
			continue;
		}

		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
			  IB_AR_INFO_CAP_MASK_GLB_GROUPS)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920C: "
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Only global AR groups supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_GLOBAL_AR_GROUP);
		}

		/*
		 * All links between switches should support at least m_min_vl_number_
		 * VLs in PortInfo VLCap
		 * We check VL cap and not operational VLs because configuring
		 * VLs comes after routing calculation
		 */
		osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
		if (!p_physp_0) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920D: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "Failed to get phys port 0.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DB_ERR);
		}

		if (p_physp_0 &&
			!(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 920E: "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "SL to VL map not supported. kDOR algorithm not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_SL_MAP_CAP);
			algorithm_supported = false;
		}

		for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
			port_num++) {

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			u_int8_t remote_node_type = osm_node_get_type(p_remote_physp->p_node);
			if (remote_node_type != IB_NODE_TYPE_SWITCH) {
				continue;
			}

			u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
			num_vls = (1 << (vl_cap - 1));

			if (num_vls < m_min_vl_number_) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 920F: "
					   "Invalid number of VLs: %u < %u on "
					   "Switch GUID: 0x%" PRIx64 ", LID: %u, port: %u\n",
					   vl_cap, m_min_vl_number_,
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num);

				sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
								    DF_VL_CAP_LESS_THAN_MIN);
				break;

			}
		}

		if (sw_node_it->second.m_support[m_algorithm_feature_] ==
		    SUPPORT_UNKNOWN) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Mark Switch GUID: 0x%" PRIx64 ", LID: %u "
				   " support algorithm feature:%u \n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   m_algorithm_feature_);

			sw_node_it->second.MarkSWSupport(m_algorithm_feature_);
		}
	}

	if (unsupported_num || disable_by_tr_err) {
		if (unsupported_num) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
				   "%d switches are not kDOR Capable.\n",
				   unsupported_num);
		}
		if (disable_by_tr_err) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
				   "%d switches do not support by_transport_disable.\n",
				   disable_by_tr_err);
		}
	}

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable returns %u\n",
		   algorithm_supported);
	OSM_AR_LOG_RETURN(m_p_osm_log, algorithm_supported);
}

void ArKdorAlgorithm::SetPlftConfiguration()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int unsupported_num = 0;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (sw_node_it->second.m_support[m_algorithm_feature_] ==
		    NOT_SUPPORTED) {
			continue;
		}

		osm_advanced_routing_data_t *p_ar_data =
			&sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data;
		int rc = 0;
		if (p_ar_data->plft_info.active_mode == 0 ||
			(GetPlftMaxSupportedLidNumber(sw_node_it->second) <
			 m_ar_mgr_.m_p_osm_subn->max_lid)) {
			rc = SetRequiredPLFTInfo(sw_node_it->second);
		} else {
			/*
			 *On every cycle required_plft_active_mode is cleared
			 *and needed to be reset to the required value.
			 */
			osm_ar_set_required_plft_active_mode(p_ar_data, p_ar_data->plft_info.active_mode);
		}

		if (rc) {
			unsupported_num++;
			sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
							    DF_PLFT_CAP_LESS_THAN_MIN); /*TODO !!! Fix names */
		}
	}

	if (unsupported_num) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
			   "Failed to set PlftConfiguration for %u switches.\n",
			   unsupported_num);
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArKdorAlgorithm::BuildStaticRouteInfo()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* build static lft tables */
	osm_ucast_dor_build_lfts(&m_ar_mgr_.m_p_osm->sm.ucast_mgr);

	SwDbEntryPrtList bfs_q;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {

		if (BuildStaticRouteInfoToSwitch(bfs_q, sw_node_it->second)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
	}

	if (m_max_vlinc_ >= m_planes_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9210: "
			   "Max required VL increase in static routing (%u) >= "
			   "max allowed kDOR planes (%u). "
			   "Cannot guarantee credit loop free.",
			   m_max_vlinc_, m_planes_number_);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int ArKdorAlgorithm::BuildStaticRouteInfoToSwitch(
	SwDbEntryPrtList &bfs_q,
	ARSWDataBaseEntry &dst_db_entry)
{
	ARSWDataBaseEntry *p_sw_db_entry;

	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Build static routing info for dest switch GUID: 0x%" PRIx64
		   ", LID: %u\n",
		   dst_db_entry.m_general_sw_info.m_guid,
		   dst_db_entry.m_general_sw_info.m_lid);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {

		sw_node_it->second.m_kdor_data->InitRouteInfo();
	}

	bfs_q.push_back(&dst_db_entry);
	while (!bfs_q.empty()) {
		p_sw_db_entry = bfs_q.front();
		bfs_q.pop_front();

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Popped from BFS Q: Switch GUID: 0x%" PRIx64 ", LID: %u \n",
			   p_sw_db_entry->m_general_sw_info.m_guid,
			   p_sw_db_entry->m_general_sw_info.m_lid);
		if (ProcessNeighborsBfs(bfs_q, p_sw_db_entry, &dst_db_entry)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, 1);
		}
		p_sw_db_entry->m_kdor_data->SetRouteStatus(BFS_DONE);
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * Adds unvisited neighbors
 */
int ArKdorAlgorithm::ProcessNeighborsBfs(SwDbEntryPrtList &bfs_q,
					 ARSWDataBaseEntry *p_sw_db_entry,
					 ARSWDataBaseEntry *p_dst_db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	KdorRouteInfo min_route_into;
	KdorRouteInfo tmp_route_info;

	SwToConnectionMap &connections = p_sw_db_entry->m_kdor_data->GetConnections();
	uint16_t sw_distance = p_sw_db_entry->m_kdor_data->GetRouteDistance();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "switch GUID: 0x%" PRIx64 ", LID %u: connections num :%u\n",
		   p_sw_db_entry->m_general_sw_info.m_guid,
		   p_sw_db_entry->m_general_sw_info.m_lid,
		   (int)connections.size());

	for (SwToConnectionMap::iterator connections_iter = connections.begin();
		connections_iter != connections.end(); ++connections_iter) {

		KdorConnection &connection = connections_iter->second;
		if (connection.m_remote_switch == NULL) {
			continue; /* CAs connection */
		}
		ARSWDataBaseEntry *p_remote_db_entry = connection.GetRemoteSwitchDbEntry();
		KdorSwData *p_remote_switch_data = p_remote_db_entry->m_kdor_data;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Processing neighbor Switch GUID: 0x%" PRIx64 ", LID: %u \n",
			   p_remote_db_entry->m_general_sw_info.m_guid,
			   p_remote_db_entry->m_general_sw_info.m_lid);

		if (p_remote_switch_data->GetRouteStatus() == BFS_INIT) {
			bfs_q.push_back(p_remote_db_entry);
			p_remote_switch_data->SetRouteStatus(BFS_QUEUED);
			p_remote_switch_data->SetRouteDistance(sw_distance + 1);
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Neighbor in BFS_INIT, adding to queue, distance: %d\n",
				   p_remote_switch_data->GetRouteDistance());
		} else if ((p_remote_switch_data->GetRouteStatus() == BFS_DONE) &&
				   (p_remote_switch_data->GetRouteDistance() + 1 == sw_distance)) {
			/* Considering shortest path candidate only */
			CalculateRouteInfo(connection,
					   p_remote_switch_data,
					   p_dst_db_entry->m_kdor_data,
					   tmp_route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Neighbor in BFS_DONE, "
				   "on shortest path to destination switch, "
				   "vl_inc: %u turn_type: %u\n",
				   tmp_route_info.m_vl_inc, tmp_route_info.m_turn_type);

			if (min_route_into.m_connection == NULL) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "No best connection so far, take this neighbor as best\n");
				min_route_into = tmp_route_info;
			} else if (tmp_route_info < min_route_into) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "This neighbor is better\n");
				min_route_into = tmp_route_info;
			}
		} else {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Neighbor in BFS_QUEUED, do nothing\n");
		}
	}


	if (min_route_into.m_connection != NULL) { /* Found the best neighbor for routing */
		KdorRouteInfo &sw_route_info =
			p_sw_db_entry->m_kdor_data->GetRouteInfo(p_dst_db_entry->m_kdor_data);
		sw_route_info = min_route_into;

		uint16_t next_hop_lid = 0;
		if ((sw_route_info.m_connection != NULL) &&
			(sw_route_info.m_connection->m_remote_switch != NULL)) {

			next_hop_lid = cl_ntoh16(osm_node_get_base_lid(
			    sw_route_info.m_connection->m_remote_switch->p_node, 0));
		}

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Set RouteInfo from "
			   "Switch GUID: 0x%" PRIx64 ", LID: %u "
			   "to switch GUID: 0x%" PRIx64 ", LID: %u next_hop_lid: %u\n",
			   p_sw_db_entry->m_general_sw_info.m_guid,
			   p_sw_db_entry->m_general_sw_info.m_lid,
			   p_dst_db_entry->m_general_sw_info.m_guid,
			   p_dst_db_entry->m_general_sw_info.m_lid,
			   next_hop_lid);

		if (m_max_vlinc_ < min_route_into.m_vl_inc) {
			m_max_vlinc_ = min_route_into.m_vl_inc;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Max vlinc of %d for path from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u \n",
				   m_max_vlinc_,
				   p_sw_db_entry->m_general_sw_info.m_guid,
				   p_sw_db_entry->m_general_sw_info.m_lid,
				   p_dst_db_entry->m_general_sw_info.m_guid,
				   p_dst_db_entry->m_general_sw_info.m_lid);
		}
		if (min_route_into.m_vl_inc >= m_planes_number_) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9211: "
				   "Max required VL increase (%u) for path from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u >= "
				   "max allowed kDOR planes (%u). \n",
				   m_max_vlinc_,
				   p_sw_db_entry->m_general_sw_info.m_guid,
				   p_sw_db_entry->m_general_sw_info.m_lid,
				   p_dst_db_entry->m_general_sw_info.m_guid,
				   p_dst_db_entry->m_general_sw_info.m_lid,
				   m_planes_number_);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 *   KDOR_TURN_TYPE_0:  +-Low   -> +-High
 *   KDOR_TURN_TYPE_1:  +-High  ->  +Low
 *   KDOR_TURN_TYPE_2:  +-High  ->  -Low
 */
kdor_turn_t ArKdorAlgorithm::GetTurnType(const KdorConnection &from_connection,
					 const KdorConnection &to_connection) const
{
	if (from_connection.m_remote_switch == NULL ||
		to_connection.m_remote_switch == NULL) {
		/* from or to non switch port */
		return KDOR_TURN_TYPE_0;
	}
	if (from_connection.m_dim_idx <= to_connection.m_dim_idx) {
		return KDOR_TURN_TYPE_0;
	}
	if (to_connection.m_dim_sign > 0) {
		return KDOR_TURN_TYPE_1;
	}
	return KDOR_TURN_TYPE_2;
}
/*
 * calculate new vl increase and turn type for given connection to sw
 *
 * Vl calculation if last turn type is T1:
 *   T1 | T2 | T1 | T2 | T1
 *   ----------------------
 *      |    |    |    | 0      total vl_inc = 0
 *   ----------------------
 *      |    |    | 1  | 2      total vl_inc = 2 (prev vl inc + 2)
 *   ----------------------
 *      |    |  0 | 1  | 2      total vl_inc = 2 (prev vl inc)
 *   ----------------------
 *      |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc + 2)
 *   ----------------------
 *    0 |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc)
 *
 * Vl calculation if last turn type is T2:
 *   T2 | T1 | T2 | T1 | T2
 *   ----------------------
 *      |    |    |    | 1      total vl_inc = 1
 *   ----------------------
 *      |    |    | 0  | 1      total vl_inc = 1 (prev vl inc)
 *   ----------------------
 *      |    |  1 | 2  | 3      total vl_inc = 2 (prev vl inc + 2)
 *   ----------------------
 *      |  0 |  1 | 2  | 3      total vl_inc = 4 (prev vl inc)
 *   ----------------------
 *    1 |  2 |  3 | 4  | 5      total vl_inc = 5 (prev vl inc + 2)
 */
void ArKdorAlgorithm::CalculateRouteInfo(KdorConnection &connection,
					 KdorRouteInfo &remote_route_info,
					 KdorRouteInfo &route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	route_info.m_connection = &connection;

	kdor_turn_t turn_type = KDOR_TURN_TYPE_0;

	if (remote_route_info.m_connection != NULL) {
		/* if remote is destination remote_route_info.m_connection == NULL */
		turn_type = GetTurnType(connection, *remote_route_info.m_connection);
	}

	if (turn_type == KDOR_TURN_TYPE_0) { /* TURN_TYPE_0 */
		route_info.m_vl_inc = remote_route_info.m_vl_inc;
		route_info.m_turn_type = remote_route_info.m_turn_type;
	} else if (turn_type == KDOR_TURN_TYPE_1) { /* TURN_TYPE_1 */
		route_info.m_turn_type = KDOR_TURN_TYPE_1;
		/*
		 * TURN_TYPE_1 starts on lowest vl (TURN_TYPE_2 VL -1) there for it
		 * does not change vl_inc
		 */
		route_info.m_vl_inc = remote_route_info.m_vl_inc;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "TURN_TYPE_1, vl_inc remains: %u\n", route_info.m_vl_inc);
	} else { /* TURN_TYPE_2 */
		route_info.m_turn_type = KDOR_TURN_TYPE_2;
		if (remote_route_info.m_turn_type == KDOR_TURN_TYPE_0) {
			route_info.m_vl_inc = remote_route_info.m_vl_inc + 1;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to TURN_TYPE_0, inc by 1 vl_inc: %u\n",
				   route_info.m_vl_inc);
		} else if (remote_route_info.m_turn_type == KDOR_TURN_TYPE_1) {
			route_info.m_vl_inc = remote_route_info.m_vl_inc + 2;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to TURN_TYPE_1, inc by 2 vl_inc: %u\n",
				   route_info.m_vl_inc);
		} else {
			route_info.m_vl_inc = remote_route_info.m_vl_inc;
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "TURN_TYPE_2 to same turn type, vl_inc remains: %u\n",
				   route_info.m_vl_inc);
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

OSMThreadPoolTask *ArKdorAlgorithm::GetCalculatePortGroupsTask(
	ARSWDataBaseEntry &db_entry)
{
	return db_entry.m_kdor_data->GetCalculatePortGroupsTask();
}

uint16_t ArKdorAlgorithm::GetPlftMaxSupportedLidNumber(
	ARSWDataBaseEntry &sw_db_entry)
{
	return sw_db_entry.m_kdor_data->m_plft_max_supported_lid_number;
}

void ArKdorAlgorithm::SetPlftMaxSupportedLidNumber(
	ARSWDataBaseEntry &sw_db_entry, u_int8_t plft_max_supported_lid_number)
{
	sw_db_entry.m_kdor_data->m_plft_max_supported_lid_number =
		plft_max_supported_lid_number;
}

void ArKdorAlgorithm::RestorePriv(ARSWDataBaseEntry &sw_db_entry)
{
	if (sw_db_entry.m_kdor_data != NULL) {
		/* set priv to point to it's origin value */
		sw_db_entry.m_general_sw_info.m_p_osm_sw->priv =
			sw_db_entry.m_kdor_data->GetSavedPriv();
	}
}

void ArKdorAlgorithm::ClearSwitchAlgotithmData(ARSWDataBaseEntry &sw_db_entry)
{
	if (sw_db_entry.m_kdor_data != NULL)
		sw_db_entry.m_kdor_data->m_plft_max_supported_lid_number = 0;
}

int ArKdorAlgorithm::CalculatePortGroups()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = m_ar_mgr_.m_port_groups_calculator.CalculatePortGroups(this);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArKdorAlgorithm::CalculateSwitchPortGroups(
	ARSWDataBaseEntry &sw_db_entry,
	LidMapping *p_lid_mapping,
	LidToConnectionArray &lid_to_connections,
	LidToPortArray &base_lid_to_port)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/*
	 * Step 1.1 -
	 *  Ensure that this switch supports the algorithm.
	 */
	if (!IsFeatureActive(sw_db_entry, m_algorithm_feature_)) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Algorithm not supported or not enabled.\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Algorithm %u not supported or not enabled. "
			   "sw_db_entry.m_support[SUPPORT_AR]: %u "
			   "sw_db_entry.m_support[m_algorithm_feature_]: %u "
			   "sw_db_entry.m_option_on: %u\n",
			   m_algorithm_feature_,
			   sw_db_entry.m_support[SUPPORT_AR],
			   sw_db_entry.m_support[m_algorithm_feature_],
			   sw_db_entry.m_option_on);

		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	KdorAlgorithmData algorithm_data;

	/*
	 * Step 2 - Calculate Ar Groups
	 */
	CalculateArGroups(sw_db_entry,
			  p_lid_mapping->m_lid_to_sw_lid_mapping,
			  algorithm_data);

	/*
	 * Step 3 - Calculate Ar PLFTs
	 */
	CalculateArPlfts(sw_db_entry,
			 p_lid_mapping,
			 algorithm_data,
			 lid_to_connections,
			 base_lid_to_port);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArGroups(ARSWDataBaseEntry &sw_db_entry,
					u_int16_t *lid_to_sw_lid_mapping,
					KdorAlgorithmData &algorithm_data)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate group bitmask for GUID: 0x%" PRIx64 ", LID: %u: \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);

	/*
	 * Step 2.1 -
	 *  Iterate through all lids and calculate group bitmask for each lid.
	 */
	osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	uint16_t max_lid = p_osm_sw->max_lid_ho;
	uint8_t **hops = p_osm_sw->hops;
	/* last plft will include only static route (no AR) */
	GroupData group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
    /* ports to neighbors with invalid dim */
    GroupData invalid_dim_group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
	int group_data_array_size = MAX_KDOR_PLFT_NUMBER - 1;
	KdorGroupData kdor_group_data;
	KdorRouteInfo route_info;
	KdorRouteInfo ingress_route_info;
	KdorConnection ingress_connection;
	ingress_connection.m_dim_idx = 0xFF; /* max index TODO: optimize to max dim */
	ingress_connection.m_remote_switch = p_osm_sw;

	uint16_t sw_lid_num;
	for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {

		if (lid_to_sw_lid_mapping[lid_num] == lid_num) {
			/* this is sw lid */
			sw_lid_num = lid_num;
		} else {
			sw_lid_num = lid_to_sw_lid_mapping[lid_num];
			if (sw_lid_num == 0) {
				continue; /* lid_num is invalid lid */
			}
		}
		uint8_t least_hops = osm_switch_get_least_hops(p_osm_sw, sw_lid_num);
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "lid_num: %u sw_lid_num: %u least_hops: %u\n",
			   lid_num , sw_lid_num, least_hops);

		if (least_hops == OSM_NO_PATH || least_hops == 0) {
			if (!hops[sw_lid_num]) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
			}
			continue;
		}

		LidToKdorGroupDataMapIter lid_to_kdor_group_iter =
			algorithm_data.m_lid_to_kdor_group_map.find(sw_lid_num);
		if (lid_to_kdor_group_iter != algorithm_data.m_lid_to_kdor_group_map.end()) {

			algorithm_data.AddLidToKdorGroupData(lid_num,
							     sw_lid_num,
							     lid_to_kdor_group_iter->second,
							     false,
							     m_p_osm_log);
			continue;
		}

		KdorSwData *p_dest_switch_data = m_sw_lid_to_kdor_data_[sw_lid_num];
		if (p_dest_switch_data == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9212: "
				   "Failed to find kDOR data for switch lid %u\n",
				   sw_lid_num);
			continue;
		}

		for (int i = 0; i < group_data_array_size; ++i) {
			group_data_array[i].m_group_bitmask.reset();
			group_data_array[i].m_group_size = 0;
			invalid_dim_group_data_array[i].m_group_bitmask.reset();
			invalid_dim_group_data_array[i].m_group_size = 0;
		}

		/*
		 *  Calculate group bitmask.
		 */
		SwToConnectionMap &connections = sw_db_entry.m_kdor_data->GetConnections();

		for (SwToConnectionMap::iterator connections_iter = connections.begin();
			connections_iter != connections.end(); ++connections_iter) {

			KdorConnection &connection = connections_iter->second;

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate AR group on Switch GUID: 0x%" PRIx64
					   ", LID: %u to destination switch LID: %u %s\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   sw_lid_num,
					   connection.ToString().c_str());
			}

			if (connection.m_remote_switch == NULL) {
				continue; /* CAs connection */
			}
			/* first port */
			u_int8_t port_num = (*connection.m_ports.begin())->port_num;
			if (osm_switch_get_hop_count(p_osm_sw, sw_lid_num, port_num) !=
			    least_hops) {
				continue; /* not shortest path */
			}

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "shortest path to LID: %u using port: %u \n",
				   sw_lid_num, port_num);

			ARSWDataBaseEntry *p_remote_db_entry =
				connection.GetRemoteSwitchDbEntry();
			KdorSwData *p_remote_switch_data = p_remote_db_entry->m_kdor_data;

			KdorRouteInfo &remote_route_info =
				p_remote_switch_data->GetRouteInfo(p_dest_switch_data);

			/* calculate route info from curr sw */
			CalculateRouteInfo(connection,
					   remote_route_info,
					   route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from remote SW LID: %u to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   sw_lid_num,
				   route_info.m_vl_inc,
				   route_info.m_turn_type);

			/* calculate worst case route info from ingress to destination */
			CalculateRouteInfo(ingress_connection,
							   route_info,
							   ingress_route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from local SW LID: %u to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   sw_db_entry.m_general_sw_info.m_lid,
				   sw_lid_num,
				   ingress_route_info.m_vl_inc,
				   ingress_route_info.m_turn_type);
			/*
			 *            -----------        ------------         -------------
			 * ---------> | curr sw | -----> | neighbor | ------> |destination|
			 *  ingress   |   (U)   |        |    (V)   |  static |    (D)    |
			 *            -----------        ------------         -------------
			 *
			 *
			 *
			 * The rout out of this port is valid if:
			 * ingress plane + (is invalid turn from ingress to neighbor) +
			 * min static vl increase to destination +
			 * is invalid turn from current sw to neighbor static egress <
			 * m_max_planes_
			 * there for:
			 * plft + ingress_route_info.m_vl_inc < m_max_planes_
			 *
			 * The following table show the final VL/plft for given
			 * in VL/plft (row) and ingress_route vl increase (column)
			 *
			 * vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
			 * -----------|        |        |        |        |        |
			 * in plft/vl |        |        |        |        |        |
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     0      |   0    |   1    |   2    |   3    |   4    |   5
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     1      |   2    |   1    |   4    |   3    |   6    |   5
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     2      |   2    |   3    |   4    |   5    |   6    |   7
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     3      |   4    |   3    |   6    |   5    |   8    |   7
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     4      |   4    |   5    |   6    |   7    |   8    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     5      |   6    |   5    |   8    |   7    |   -    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *
			 * From the table:
			 *
			 * For ingress_route_info.m_turn_type == T1:
			 * final_vl = (in_vl + 1) / 2 * 2 + vl_inc <= planes_number - 1
			 * max_in_vl is even -> in_vl + vl_inc <= planes_number - 1
			 *  -> max_in_vl <= planes_number - vl_inc - 1
			 * max_in_vl is even - round doun to even number
			 * max_in_vl <= (planes_number - vl_inc - 1) / 2 * 2
			 *
			 * For ingress_route_info.m_turn_type == T2:
			 * final_vl = (in_vl / 2) * 2 + vl_inc <= planes_number - 1
			 * max_in_vl is odd -> in_vl - 1 + vl_inc <= planes_number - 1
			 * in_vl <= planes_number - vl_inc
			 * max_in_vl is odd - convert down to odd number
			 * max_in_vl <= (planes_number - vl_inc + 1) / 2 * 2 - 1;
			 *
			 * The folowing teble provide max_plft for given vl_inc and max planes
			 * vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
			 * -----------|        |        |        |        |        |
			 * max planes |        |        |        |        |        |
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     4      |   2    |   3    |    0   |   1    |   -    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     5      |   4    |   3    |    2   |   1    |   0    |   -
			 * -----------+--------+--------+--------+--------+--------+--------
			 *     6      |   4    |   5    |    2   |   3    |   0    |  1
			 * -----------+--------+--------+--------+--------+--------+--------
			 */

			if (ingress_route_info.m_vl_inc >= m_planes_number_) {
				continue; /* Round doun works only for positive number */
			}

			GroupData *p_group_data_array = group_data_array;
			if (remote_route_info.m_connection != NULL && /* Not the destination */
				(remote_route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
				 route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX)) {
				p_group_data_array = invalid_dim_group_data_array;
			}

			int max_plft;
			if (ingress_route_info.m_turn_type == KDOR_TURN_TYPE_2) {
				max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc + 1) /
				    2 * 2 - 1;
			} else {
				max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc - 1) /
				    2 * 2;
			}

			max_plft = std::min(max_plft, group_data_array_size - 1);
			for (int plft = 0; plft <= max_plft; ++plft) {
				p_group_data_array[plft].m_group_bitmask |=
					connection.m_ports_bitset;
				p_group_data_array[plft].m_group_size +=
					connection.m_ports.size();

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Add ports %s to group_data of plft_id: %u "
					   "total size: %u is_valid_dim: %u\n",
					   connection.m_ports_bitset.to_string().c_str(),
					   plft,
					   p_group_data_array[plft].m_group_size,
					   p_group_data_array == group_data_array);
			}
		}

		/*
		 *If no route through neighbors with valid coordinate
		 *use route through neighbors without valid coordinate
		 */
		if (invalid_dim_group_data_array[0].m_group_size) {

			for (int plft = 0; plft < group_data_array_size; ++plft) {
				if ((group_data_array[plft].m_group_size == 0) &&
					(invalid_dim_group_data_array[plft].m_group_size != 0)) {

					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "AR group on Switch GUID: 0x%" PRIx64 ", LID: %u: "
						   "to destination switch LID: %u plft_id:%u "
						   "uses neighbors without valid coordinate.\n",
						   sw_db_entry.m_general_sw_info.m_guid,
						   sw_db_entry.m_general_sw_info.m_lid,
						   sw_lid_num,
						   plft);

					group_data_array[plft].m_group_bitmask =
						invalid_dim_group_data_array[plft].m_group_bitmask;
					group_data_array[plft].m_group_size =
						invalid_dim_group_data_array[plft].m_group_size;
				}
			}
		}

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			for (int plft = 0; plft < group_data_array_size; ++plft) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "AR group on Switch GUID: 0x%" PRIx64 ", LID: %u "
					   "to destination switch LID: %u plft_id:%u "
					   "is %s size: %u\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   sw_lid_num,
					   plft,
					   group_data_array[plft].m_group_bitmask.to_string().c_str(),
					   group_data_array[plft].m_group_size);
			}
		}

		/*
		 * Build kdor_group_data with discovered group_data for each plft
		 * group for plf == 0 is the largest groupt
		 */
		if (group_data_array[0].m_group_size > 1) {

			for (int plft = 0; plft < group_data_array_size; ++plft) {
				if (group_data_array[plft].m_group_size > 1) {
					if (m_p_osm_log->level & OSM_LOG_DEBUG) {
						OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
							   "Insert group bitmask: %s to "
							   "algorithm_data groups_map\n",
							   group_data_array[plft].m_group_bitmask.
							   to_string().c_str());
					}

					GroupBitmaskToDataMapInsertRes res =
						algorithm_data.m_groups_map.insert(
							make_pair(group_data_array[plft].m_group_bitmask,
								  group_data_array[plft]));
					kdor_group_data.group_data_array[plft] = &(res.first->second);
				} else {
					kdor_group_data.group_data_array[plft] = NULL;
				}
			}

			/* Save discovered kdor_group_data for new destination */
			algorithm_data.AddLidToKdorGroupData(lid_num,
							     sw_lid_num,
							     kdor_group_data,
							     true,
							     m_p_osm_log);
		}
	}

	/*
	 * Step 2.2 -
	 *  try to assign group number for all groups.
	 */
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 2.2 AssignPortGroups \n");

	if (!algorithm_data.m_groups_map.empty()) {
		int rc = m_ar_mgr_.AssignPortGroups(sw_db_entry, algorithm_data);
		if (rc) {
			/* Sanity check, should not reach here */
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9213: Failed to assign port groups.\n");
		}
	}

	/*
	 * Step 3 -
	 *  Update calculated ar_group_table:
	 *  At this point we have all assign groups in algorithm_data.m_assign_groups list
	 */
	unsigned calculated_max_pg_id = 0;
	osm_ar_subgroup_t *p_subgroup;
	osm_advanced_routing_data_t *p_ar_data = &p_osm_sw->ar_data;

	if (!algorithm_data.m_assign_groups.empty()) {

		GroupsListIter list_iter = algorithm_data.m_assign_groups.begin();
		for (;
			list_iter != algorithm_data.m_assign_groups.end();
			++list_iter) {

			p_subgroup = osm_ar_subgroup_get(p_ar_data, (*list_iter)->m_group_number, 0);
			m_ar_mgr_.UpdateCalculatedArGroupTable(
				p_subgroup,
				(*list_iter)->m_group_number,
				(*list_iter)->m_group_bitmask,
				calculated_max_pg_id);
		}
	}

	/*
	 * Step 4 -
	 *  Update AR Group Top
	 */
	m_ar_mgr_.ARUpdateSWGroupTableTop(sw_db_entry, calculated_max_pg_id+1);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlfts(ARSWDataBaseEntry &sw_db_entry,
				       LidMapping *p_lid_mapping,
				       KdorAlgorithmData &algorithm_data,
				       LidToConnectionArray &lid_to_connections,
				       LidToPortArray &base_lid_to_port)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 4 configure AR PLFTs \n");
	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	memset(lid_to_connections, 0, sizeof(lid_to_connections));
	BuildDstSwLidToConnection(sw_db_entry, lid_to_connections);

	/* save calculated port for base lid (in order to select same port to all vlids) */
	memset(base_lid_to_port, 0, sizeof(base_lid_to_port));

	uint8_t lid_state;
	uint16_t ar_group_id;
	for (unsigned dest_lid = 1;
		 dest_lid <= m_ar_mgr_.m_p_osm_subn->max_lid;
		 ++dest_lid) {

		u_int8_t ucast_lft_port = GetStaticUcastLftPort(sw_db_entry,
								p_lid_mapping,
								lid_to_connections,
								base_lid_to_port,
								dest_lid);

		if (ucast_lft_port == OSM_NO_PATH)	/* No route to this lid_num from this switch */
			continue;

		LidToKdorGroupDataMapConstIter ar_kdor_groups_iter =
			algorithm_data.m_lid_to_kdor_group_map.find(dest_lid);

		for (int plft_id = 0; plft_id < m_planes_number_; ++plft_id) {
			CalculateArPlft(sw_db_entry,
					algorithm_data,
					ucast_lft_port,
					dest_lid,
					plft_id,
					ar_kdor_groups_iter,
					lid_state,
					ar_group_id);

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u plft id %u dest_lid: %u - "
					   "GroupNumber: %u LidState: %u Static Port %u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   plft_id,
					   dest_lid,
					   ar_group_id, lid_state, ucast_lft_port);
			}

			osm_ar_set_required_plft_top(p_ar_data, plft_id,
						     m_ar_mgr_.m_p_osm_subn->max_lid);

		} /* End of plft_id loop */
	} /* End of loop destination lids */

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlft(ARSWDataBaseEntry &sw_db_entry,
				      KdorAlgorithmData &algorithm_data,
				      u_int8_t ucast_lft_port,
				      unsigned dest_lid,
				      int plft_id,
				      LidToKdorGroupDataMapConstIter &ar_kdor_groups_iter,
				      OUT uint8_t &lid_state,
				      OUT uint16_t &ar_group_id)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "CalculateArPlft LID %u dest_lid:%u static_port: %u plft: %u\n",
		   sw_db_entry.m_general_sw_info.m_lid,
		   dest_lid,
		   ucast_lft_port,
		   plft_id);

	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	if (plft_id == (m_planes_number_ - 1) || /* last plane is static only */
		ar_kdor_groups_iter == algorithm_data.m_lid_to_kdor_group_map.end() ||
		ar_kdor_groups_iter->second.group_data_array[plft_id] == NULL) {

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "No AR group - Set lid sate Static Group: 0\n");

		lid_state = (uint8_t)OSM_AR_IB_LID_STATE_STATIC;
		ar_group_id = 0;
		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);

		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	GroupData* p_group_data =
		ar_kdor_groups_iter->second.group_data_array[plft_id];

	if (p_group_data->m_group_table_number == ARMGR_MAX_GROUP_TABLES) {

		lid_state = (uint8_t)OSM_AR_IB_LID_STATE_STATIC;
		ar_group_id = 0;
		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "No AR group because group bitmask:%s was not mapped.\n",
				   p_group_data->m_group_bitmask.to_string().c_str());
		}

	} else {
		lid_state = (uint8_t)m_ar_mgr_.m_master_db.m_ar_mode;
		ar_group_id = p_group_data->m_group_number;
		/*
		 * Currently RN Not implemented
		 */
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - "
			   "Set lid state %u Group:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid, dest_lid,
			   lid_state,
			   ar_group_id);

		osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port,
			       lid_state, ar_group_id);
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

uint8_t ArKdorAlgorithm::GetStaticUcastLftPort(ARSWDataBaseEntry &sw_db_entry,
					       LidMapping *p_lid_mapping,
					       KdorConnection *dst_sw_lid_to_conection[],
					       uint8_t base_lid_to_port[],
					       uint16_t dest_lid)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	uint8_t port_number;
	KdorConnection *connection;
	if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == 0) {
		OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
	}

	if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == dest_lid) {
		/* sw lid */
		connection = dst_sw_lid_to_conection[dest_lid];
		if (connection == NULL) {
			if (dest_lid == sw_db_entry.m_general_sw_info.m_lid) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "GetStaticUcastLftPort for local switch lid: %u "
					   "return: %u\n",
					   dest_lid, 0);
				OSM_AR_LOG_RETURN(m_p_osm_log, 0);
			}
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9214: Failed to find static connection to "
				   "destination lid: %u\n",
				   dest_lid);
			OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
		}
		/* No load balancing to switch lid */
		port_number = (*connection->m_next_port_for_routing)->port_num;
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Use static port %u for destination switch lid: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_number, dest_lid);

		OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
	}
	/* CA lid */
	uint16_t base_lid = p_lid_mapping->m_lid_to_base_lid_mapping[dest_lid];
	if (base_lid_to_port[base_lid] != 0) {
		port_number = base_lid_to_port[base_lid];

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Use static port %u for dest_lid: %u base_lid: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_number, dest_lid, base_lid);

		OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
	}

	uint16_t sw_lid = p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid];
	connection = dst_sw_lid_to_conection[sw_lid];
	if (connection == NULL) {
		if (sw_lid == sw_db_entry.m_general_sw_info.m_lid) {
			port_number = OsmSwitchGetPortByLidUseNewLft(
			    sw_db_entry.m_general_sw_info.m_p_osm_sw, base_lid);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - "
				   "Use static port %u for dest_lid: %u base_lid: %u"
				   "on local switch\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   port_number, dest_lid, base_lid);

			OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
		}
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9215: Failed to find static connection to "
			   "switch lid:%u of destination lid: %u\n",
			   sw_lid, dest_lid);
		OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
	}
	port_number = (*connection->m_next_port_for_routing)->port_num;
	base_lid_to_port[base_lid] = port_number;
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetStaticUcastLftPort for base_lid: %u "
		   "of destination lid: %u set and return: %u\n",
		   base_lid, dest_lid, port_number);

	/* Increment next_port_for_routing */
	++connection->m_next_port_for_routing;
	if (connection->m_next_port_for_routing == connection->m_ports.end()) {
		connection->m_next_port_for_routing = connection->m_ports.begin();
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
}

void ArKdorAlgorithm::BuildDstSwLidToConnection(
	ARSWDataBaseEntry &sw_db_entry,
	KdorConnection *dst_sw_lid_to_conection[])
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	KdorConnection *connection;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (&sw_node_it->second == &sw_db_entry) {
			continue;
		}
		ARSWDataBaseEntry &dst_sw_db_entry = sw_node_it->second;
		KdorRouteInfo &route_info_to_dst =
			sw_db_entry.m_kdor_data->GetRouteInfo(dst_sw_db_entry.m_kdor_data);
		connection = route_info_to_dst.m_connection;
		if (connection == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9216: -A- Failed to get RouteInfo from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_guid,
				   dst_sw_db_entry.m_general_sw_info.m_lid);
			continue;
		}
		dst_sw_lid_to_conection[dst_sw_db_entry.m_general_sw_info.m_lid] =
			connection;
		connection->m_next_port_for_routing = connection->m_ports.begin();
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorVl2VlPerOpVl()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/*
	 * For turn type 1 increase vl if vl%2 == 1
	 * For turn type 2 increase vl if vl%2 == 0
	 */
	m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_0] = m_vl2vl_per_op_vls_;
	m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_1] = m_turn_1_vl2vl_per_op_vls_;
	m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_2] = m_turn_2_vl2vl_per_op_vls_;



	/* Currently support only one vl per plane, there for: */
	uint8_t vls_per_stream = m_planes_number_;

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildVl2Vl t1,t2 PerOpVl.\n");
	u_int8_t t1_vl2vl_array[IB_NUMBER_OF_SLS]; /* turn_type_1 */
	u_int8_t t2_vl2vl_array[IB_NUMBER_OF_SLS]; /* turn_type_2 */
	u_int8_t num_vls;
	u_int8_t max_vl;
	u_int8_t qos_max_vl;     /* max vl for the current qos */
	u_int8_t t1_out_vl;	     /* turn_type_1 */
	u_int8_t t2_out_vl;	     /* turn_type_2 */

	/*
	 * Turn type 1 is valid on even vls ((in_vl % 2) == 0)
	 * Turn type 2 is valid on odd vls ((in_vl % 2) != 0)
	 */

	memset(m_turn_1_vl2vl_per_op_vls_, 0, sizeof(m_turn_1_vl2vl_per_op_vls_));
	memset(m_turn_2_vl2vl_per_op_vls_, 0, sizeof(m_turn_2_vl2vl_per_op_vls_));

	for (uint8_t op_vls = 2; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));
		max_vl = min((uint8_t)(num_vls - 1), (uint8_t)(IB_DROP_VL - 1));

		if (num_vls < vls_per_stream)
			continue; /* Not Supported */

		/* On invalid turn that my lead to loop we set to qos_max_vl */
		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			/* vl < num_vls */
			u_int8_t vl = in_vl % num_vls;

			/*
			 * The qos groups number of given vl [1..last] is:
			 * (vl / vls_per_stream) + 1
			 *
			 * In each qos group there are vls_per_stream vls so for vl in specific
			 * qos group the max vl is:
			 */
			qos_max_vl = ((vl / vls_per_stream) + 1 ) * vls_per_stream - 1;
			qos_max_vl = min(qos_max_vl, max_vl);

			if ((in_vl % 2) == 0) {
				t1_out_vl = min(qos_max_vl, vl);
				t2_out_vl  = min(qos_max_vl, (uint8_t)(vl + 1));
			} else {
				t1_out_vl = min(qos_max_vl, (uint8_t)(vl + 1));
				t2_out_vl  = min(qos_max_vl, vl);
			}

			m_turn_1_vl2vl_per_op_vls_[op_vls][in_vl] = t1_out_vl;
			m_turn_2_vl2vl_per_op_vls_[op_vls][in_vl] = t2_out_vl;
		}
	} /* op_vls loop */

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated kdor turns vl to vl mapping, vls_per_stream: %u\n",
			   vls_per_stream);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u turn type 1 vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_turn_1_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u turn type 2 vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(
					   m_turn_2_vl2vl_per_op_vls_[op_vls]).c_str());
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorPlftMap()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildKdorPlftMap\n");
	ib_port_sl_to_plft_map_entry_t vl_to_plft_map;
	u_int8_t plft_id;

	memset(&vl_to_plft_map, 0, sizeof(vl_to_plft_map));

	plft_id = 0;
	for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
		if ((in_vl % m_planes_number_) == 0) {
			plft_id = 0;
		}

		vl_to_plft_map.plft_port_sl[in_vl] = plft_id;
		plft_id++;
	}

	for (int i = 0; i < IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE; ++i) {
		memcpy(&m_plft_map_.port[i], &vl_to_plft_map,
			   sizeof(vl_to_plft_map));
	}

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated VL to plft mapping, planes_number: %u\n",
			   m_planes_number_);

		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "map VL: %u to plft id:%u\n",
				   in_vl, vl_to_plft_map.plft_port_sl[in_vl]);
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArKdorAlgorithm::GetSwitchSL2VL(osm_physp_t *osm_phys_port,
				    uint8_t iport_num, uint8_t oport_num,
				    osm_node_t *p_node,
				    uint64_t guid,
				    uint8_t (*sl2vl)[IB_NUMBER_OF_SLS]) const
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	uint8_t op_vls;

	GuidToSWDataBaseEntryIter sw_it = m_sw_map_.find(guid);

	if (sw_it == m_sw_map_.end())
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);

	rc = m_ar_mgr_.GetOpVlForVL2VL(
		sw_it->second, osm_phys_port, oport_num, MIN_KDOR_VL_CAP, op_vls);

	if (rc)
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);

	if (oport_num == 0 || iport_num == 0) {

		bool is_enhanced_port0 =
			ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate Vl2Vl Mappingn on switch GUID: 0x%" PRIx64
			   " from port %u to port: %u is_enhanced_port0: %u\n",
			   guid, iport_num, oport_num, is_enhanced_port0);

		if (!is_enhanced_port0)
			OSM_AR_LOG_RETURN(m_p_osm_log, -1);

		memcpy(sl2vl, m_sl2vl_per_op_vls_[op_vls], sizeof(*sl2vl));

		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	const KdorConnection *from_connection =
		sw_it->second.m_kdor_data->GetPortConnection(iport_num);
	const KdorConnection *to_connection =
		sw_it->second.m_kdor_data->GetPortConnection(oport_num);

	if (from_connection == NULL || to_connection == NULL) {
		/* From inactive port or to inactive port */
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	kdor_turn_t turn_type = GetTurnType(*from_connection, *to_connection);

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate Vl2Vl Mappingn on "
			   "switch GUID: 0x%" PRIx64 ", LID: %u "
			   "from port %u to port: %u turn_type: %u\n",
			   sw_it->second.m_general_sw_info.m_guid,
			   sw_it->second.m_general_sw_info.m_lid,
			   iport_num, oport_num, turn_type);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "op_vls: %u turn type %u vl to vl mapping %s\n",
			   op_vls, turn_type,
			   SLToVLMappingToStr(
				   m_turn_type_to_vl2vl_per_op_vls_[turn_type][op_vls]).c_str());

	}

	memcpy(sl2vl, m_turn_type_to_vl2vl_per_op_vls_[turn_type][op_vls],
	       sizeof(*sl2vl));
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArKdorAlgorithm::SetPlftMap()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (!IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID: 0x%" PRIx64 ", LID: %u - " \
				   "Algorithm (%u) not supported or not enabled, "
				   "Set PlftMap skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   m_algorithm_feature_);

			/* clear AR configuration from sw - SM should configure this sw */
			sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
			continue;
		}

		SetPlftMap(sw_node_it->second);
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::SetPlftMap(ARSWDataBaseEntry &sw_db_entry)
{
	uint8_t max_block = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node->
	    node_info.num_ports / IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

	for (uint8_t port_block = 0; port_block <= max_block; ++port_block) {
		osm_ar_port_sl_to_plft_set_block(sw_db_entry.m_general_sw_info.m_p_osm_sw,
						 port_block,
						 FALSE, /* calculated */
						 &m_plft_map_);
	}
}

/*
 * KdorAlgorithmData
 */

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
					      uint16_t sw_lid_num,
					      KdorGroupData &kdor_group_data,
					      bool is_new_group,
					      osm_log_t *p_osm_log)
{
	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "Add lid_num:%u sw_lid_num: %u to KdorGroupData is_new_group: %u\n",
		   lid_num, sw_lid_num, is_new_group);

	if (is_new_group) {
		AddLidToKdorGroupData(sw_lid_num, kdor_group_data);
	}

	if (lid_num != sw_lid_num) {
		AddLidToKdorGroupData(lid_num, kdor_group_data);
	}
}

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
					      KdorGroupData &kdor_group_data)
{
	static int num_groups = MAX_KDOR_PLFT_NUMBER - 1;

	for (int plft = 0; plft < num_groups; ++plft) {
		/* might be NULL incase of group size <= 1 */
		if (kdor_group_data.group_data_array[plft] != NULL) {
			kdor_group_data.group_data_array[plft]->
			m_lids_list.push_back(lid_num);
		}
	}
	m_lid_to_kdor_group_map.insert(make_pair(lid_num, kdor_group_data));
}

/*
 * ArHcAlgorithm
 */

/*
 * Update connection::direction dimensions index and sign for given
 * src and dst switches
 * Valid connection direction is connection cange of one and only one dimension.
 * If no chane found (same connection for src and dst)
 * or more than one dimension changed, return non zero value
 */
int ArHcAlgorithm::SetDirection(osm_switch_t* src,
				osm_switch_t* dst,
				KdorConnection &connection)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if (src->coord == OSM_SW_NO_COORD ||
	    dst->coord == OSM_SW_NO_COORD) {
	    connection.m_dim_idx = INVALID_DIM_INDEX;
	    connection.m_dim_sign = (dim_sign_t)(1);

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
			u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
			u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
			u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Invalid coordinate on set direction from "
				   "switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
				   "to switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
				   "dim_idx: %u  dim_sign: %d\n",
				   src_node_guid, src_node_lid, src->coord,
				   dst_node_guid, dst_node_lid, dst->coord,
				   connection.m_dim_idx, connection.m_dim_sign);
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	bool diff_found = false;
	dim_index_t dim_idx = 0;
	uint16_t coord_diff = src->coord ^ dst->coord;

	while (coord_diff) {
		if (coord_diff & 0x1) {
			if (coord_diff == 0x1) {
				diff_found = true;
			}
			break;
		}
		coord_diff = coord_diff >> 1;
		dim_idx++;
	}

	if (!diff_found) {
		u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
		u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9217: Switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x"
			   "is not valid HC member of "
			   "switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x\n",
			   src_node_guid, src_node_lid, src->coord,
			   dst_node_guid, dst_node_lid, dst->coord);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	connection.m_dim_idx = dim_idx;
	if (src->coord >> dim_idx & 0x1) {
		connection.m_dim_sign = (dim_sign_t)(-1);
	} else {
		connection.m_dim_sign = (dim_sign_t)(1);
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		u_int64_t src_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t src_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));
		u_int64_t dst_node_guid = cl_ntoh64(osm_node_get_node_guid(src->p_node));
		u_int16_t dst_node_lid = cl_ntoh16(osm_node_get_base_lid(src->p_node, 0));

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Direction from switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
			   "to switch GUID: 0x%" PRIx64 ", LID: %u coord: 0x%x "
			   "dim_idx: %u  dim_sign: %d\n",
			   src_node_guid, src_node_lid, src->coord,
			   dst_node_guid, dst_node_lid, dst->coord,
			   connection.m_dim_idx, connection.m_dim_sign);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/*
 * KdorSwData
 */

int KdorSwData::Init(ArKdorAlgorithm &algorithm,
		     ARSWDataBaseEntry &db_entry,
		     uint32_t sw_idx,
		     uint32_t switches_number)
{
	OSM_AR_LOG_ENTER(algorithm.GetLog());

	OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
		   "Init KdorSwData of "
		   "switch GUID: 0x%" PRIx64 ", LID: %u sw_idx: %u\n",
		   db_entry.m_general_sw_info.m_guid,
		   db_entry.m_general_sw_info.m_lid,
		   sw_idx);

	osm_physp_t* p_physp;
	osm_node_t *p_remote_node;
	osm_switch_t  *p_osm_sw = db_entry.m_general_sw_info.m_p_osm_sw;
	osm_node_t *p_osm_node = p_osm_sw->p_node;
	KdorConnection init_connection;

	m_calculate_port_groups_task_.Init(&db_entry, &algorithm);
	m_sw_idx_ = sw_idx;
	if (switches_number > (uint32_t)m_route_info_.size()) {
		m_route_info_.resize((uint32_t)(switches_number * 1.05));
	}

	/* Set priv to point to current ARSWDataBaseEntry */
	m_saved_priv_ = p_osm_sw->priv;
	p_osm_sw->priv = &(db_entry);

	for (KdorRouteInfoVec::iterator iter = m_route_info_.begin();
		iter != m_route_info_.end(); ++iter) {
		iter->Init();
	}

	/* TODO: Optimization reset connection only if one sw connection changed */
	m_connections_.clear();
	m_port_num_to_connection_tbl_.assign(m_port_num_to_connection_tbl_.size(),
					     NULL);
	osm_switch_t  *p_remote_sw;
	/* Build connections */
	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ;
		port_num++) {
		p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp) {
			continue;
		}
		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp) {
			OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
				   "Unconnected port: %d\n", port_num);
			continue;
		}
		p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
		if (!p_remote_node) {
			OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_ERROR,
				   "ERR 9218: "
				   "-A- Faile to get remote node connected to "
				   "switch GUID: 0x%" PRIx64 ", LID: %u port: %u\n",
				   db_entry.m_general_sw_info.m_guid,
				   db_entry.m_general_sw_info.m_lid,
				   port_num);
			continue;
		}

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);
		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			/* Create one connection to all hosts with remote_sw key == NULL */
			p_remote_sw = NULL;
		} else {
			/* remote is switch */
			p_remote_sw = p_remote_node->sw;
		}

		SwToConnectionMapInsertRes insert_res = m_connections_.insert(
		    make_pair(p_remote_sw, init_connection));
		KdorConnection &connection = insert_res.first->second;

		if (insert_res.second) { /* new connection was inserted */

			connection.m_remote_switch = p_remote_sw;
			if (p_remote_sw != NULL &&
				algorithm.SetDirection(p_osm_sw, p_remote_sw, connection)) {
				OSM_AR_LOG_RETURN(algorithm.GetLog(), 1);
			}
		}

		m_port_num_to_connection_tbl_[port_num] = &connection;
		connection.m_ports.push_back(p_physp);
		connection.m_ports_bitset.set(port_num);

		OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
			   "Add port:%u to connection to %p "
			   "Total connections number:%u \n",
			   port_num, p_remote_sw,
			   (int)m_connections_.size());
	}
	OSM_AR_LOG_RETURN(algorithm.GetLog(), 0);
}

void KdorSwData::Resize(uint32_t ports_num) {
	m_port_num_to_connection_tbl_.resize(ports_num, NULL);
}

const KdorConnection *KdorSwData::GetPortConnection(uint8_t port_num) const
{
	if (port_num >= m_port_num_to_connection_tbl_.size()) {
		return NULL;
	}
	return m_port_num_to_connection_tbl_[port_num];
}

string KdorConnection::ToString() const
{
	std::stringstream stream;

	if (m_remote_switch == NULL) {

		stream << "Connection to " << (int) m_ports.size() <<
		    " locat CAs";
	} else {

		uint16_t next_hop_lid = cl_ntoh16(osm_node_get_base_lid(
		    m_remote_switch->p_node, 0));

		stream << "Connection to " << (int)m_ports.size() <<
		    " on dim_index " << (int)m_dim_idx <<
		    " dim_sign " << (int)m_dim_sign <<
		    " links to remote switch LID: " << next_hop_lid;
	}

	return stream.str();
}

/* Select turn type using the following priority: T0, T2, T1 */
uint8_t KdorRouteInfo::turn_type_wait[KDOR_TURN_TYPE_LAST] = {
	0, /* T0 */
	2, /* T1 */
	1  /* T2 */
};

bool KdorRouteInfo::operator<(const KdorRouteInfo& __rhs) const
{
	if ((__rhs.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
	     m_connection->m_dim_idx == INVALID_DIM_INDEX) &&
	    m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
		/* Prefer destination with valid coordinate */
		return m_connection->m_dim_idx != INVALID_DIM_INDEX;
	}
	if (m_vl_inc != __rhs.m_vl_inc) {
		return m_vl_inc < __rhs.m_vl_inc;
	}
	if (m_turn_type != __rhs.m_turn_type) {
		/* Select by order T0, T2, T1 */
		return turn_type_wait[m_turn_type] < turn_type_wait[__rhs.m_turn_type];
	}
	/* Among top candidate choose one laying in the maximum dimension */
	if (m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
		return m_connection->m_dim_idx > __rhs.m_connection->m_dim_idx;
	}
	if (m_connection->m_dim_sign != __rhs.m_connection->m_dim_sign) {
		if (m_turn_type == KDOR_TURN_TYPE_2) {
			/*
			 * This will guarantee that prev hop will not have KDOR_TURN_TYPE_1
			 * and there for will not increase vl
			 */
			return m_connection->m_dim_sign < __rhs.m_connection->m_dim_sign;
		} else {
			/*
			 * This will guarantee that prev hop will not have KDOR_TURN_TYPE_2
			 * and there for will not increase vl
			 */
			return m_connection->m_dim_sign > __rhs.m_connection->m_dim_sign;
		}
	}
	return false;
}
