/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_MAD_RCV_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
/*
 * Abstract:
 */

static void process_ar_info(osm_sm_t *sm, osm_node_t *p_node,
			    osm_ar_mad_context_t *p_context,
			    ib_ar_info_t *p_ar_info)
{
	osm_switch_t *p_sw = p_node->sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	OSM_LOG_ENTER(sm->p_log);

	switch (p_context->request_type) {
		case OSM_AR_REQ_TYPE_GET:
			/*
			 * In case the SET request failed the cached ar_info
			 * matches the one on the switch meaning there is need
			 * to mark the ar_info_updated flag only in GET request.
			 */
			p_ar_data->ar_info_updated = TRUE;
			/* Fall through */
		case OSM_AR_REQ_TYPE_SET:
			memcpy(&p_ar_data->ar_info, p_ar_info,
			       sizeof(p_ar_data->ar_info));
			/* Fall through */
		case OSM_AR_REQ_TYPE_DYNCALC:
			if (osm_ar_capabilities_update(p_ar_data, p_ar_info)) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8300:"
					" Failed to update advanced routing"
					" capabilities of switch"
					" 0x%" PRIx64 "\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)));
			}
			break;
	}

	OSM_LOG_EXIT(sm->p_log);
}


static void process_ar_lft(osm_sm_t *sm, osm_node_t *p_node,
			   osm_ar_mad_context_t *p_context,
			   ib_ar_lft_t *p_ar_lft,
			   uint8_t plft_id)
{
	osm_advanced_routing_data_t *p_ar_data = &p_node->sw->ar_data;

	OSM_LOG_ENTER(sm->p_log);

	if(osm_ar_lft_set_block(p_ar_data, plft_id, p_context->block_id, p_ar_lft))
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 8301: Failed to store ARLft"
			" response for switch 0x%" PRIx64
			" (block id %u) (plft id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			p_context->block_id,
			plft_id);

	OSM_LOG_EXIT(sm->p_log);
}


static void process_ar_group_table(osm_sm_t *sm, osm_node_t *p_node,
				   osm_ar_mad_context_t *p_context,
				   ib_ar_group_table_t *p_ar_group_table)
{
	osm_advanced_routing_data_t *p_ar_data = &p_node->sw->ar_data;

	OSM_LOG_ENTER(sm->p_log);

	if (osm_ar_groups_set_block(p_ar_data, p_context->block_id,
				    p_ar_group_table))
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 8302: Failed to store ARGroup"
			" response for switch 0x%" PRIx64 " (block id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			p_context->block_id);

	OSM_LOG_EXIT(sm->p_log);
}

static void process_ar_group_table_copy(osm_sm_t *sm, osm_node_t *p_node,
					osm_ar_mad_context_t *p_context,
					ib_ar_group_copy_t *p_ar_group_copy_table,
					ib_net32_t attr_mod)
{
	OSM_LOG_ENTER(sm->p_log);
	osm_advanced_routing_data_t *p_ar_data = &p_node->sw->ar_data;
	/* source group to copy */
	uint16_t src_grp = cl_ntoh32(attr_mod) &
		OSM_AR_GROUP_COPY_ATTR_MOD_SRC_GROUP_MASK;
	uint8_t num_group_subgroups = p_ar_data->num_group_subgroups, entry;
	uint16_t src_grp_offset = src_grp * num_group_subgroups, dest_grp_offset;
	size_t group_size  = sizeof(osm_ar_subgroup_t) * num_group_subgroups;
	ib_ar_group_copy_entry_t *curr_entry;
	uint16_t start, end, grp_i;

	for (entry = 0; entry < IB_AR_GROUP_COPY_BLOCK_SIZE; entry++) {
		curr_entry = &p_ar_group_copy_table->group_copy_entry[entry];
		start = cl_ntoh16(curr_entry->first_index);
		end = cl_ntoh16(curr_entry->last_index);
		/*
		 * If First or Last index of the AdaptiveRoutingGroupTable
		 * to be copied to, indicate Group 0 then no copy is performed
		 * and subsequent block elements are ignored in terms of copying.
		 */
		if (start == 0 || end == 0)
			break;

		for (grp_i = start; grp_i <= end; grp_i++) {
			dest_grp_offset = grp_i * num_group_subgroups;
			memcpy(&p_ar_data->subn_tables.ar_subgroups[dest_grp_offset],
			       &p_ar_data->subn_tables.ar_subgroups[src_grp_offset],
			       group_size);
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"GUID 0x%" PRIx64 " dest_grp_offset %u, src_grp_offset %u\n",
				osm_node_get_node_guid(p_node), dest_grp_offset, src_grp_offset);
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}


void osm_ar_mad_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	void *p_payload;
	osm_ar_mad_context_t *p_context;
	boolean_t verbose_bypass = FALSE;
	uint8_t plft_id;

	osm_node_t *p_node;
	ib_net64_t node_guid;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_ar_mad_context_ptr(p_madw);
	p_payload = ib_smp_get_payload_ptr(p_smp);

	node_guid = p_context->node_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_node = osm_get_node_by_guid(sm->ucast_mgr.p_subn, node_guid);
	if (!p_node || !p_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7D01: "
			"Received ARInfo for nonexistent switch node "
			"0x%" PRIx64 "\n", cl_ntoh64(node_guid));
		goto Lock_release;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		p_node->sw->ar_configured = OSM_SW_AR_CONF_DISABLED;
		goto Lock_release;
	}

	/* Mark that the switch received the successful response */
	p_node->sw->ar_data.pending_mads--;

	switch (p_smp->attr_id) {
		case IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO:
			verbose_bypass = osm_verbose_bypass_handle_ar_info(sm, p_node);
			process_ar_info(sm, p_node, p_context,
					(ib_ar_info_t *)p_payload);
			break;
		case IB_MAD_ATTR_ADAPTIVE_ROUTING_LFT:
			verbose_bypass = osm_verbose_bypass_handle_ar_lft(sm, p_node);
			plft_id = (uint8_t)(
				((cl_ntoh32(p_smp->attr_mod) & 0xFF000000)) >> 24);
			process_ar_lft(sm, p_node, p_context,
				       (ib_ar_lft_t *)p_payload, plft_id);
			break;
		case IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE:
			verbose_bypass = osm_verbose_bypass_handle_ar_group_table(sm, p_node);
			process_ar_group_table(sm, p_node, p_context,
					       (ib_ar_group_table_t *)p_payload);
			break;
		case IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE_COPY:
			verbose_bypass = osm_verbose_bypass_handle_ar_group_table_copy(sm, p_node);
			process_ar_group_table_copy(sm, p_node, p_context,
						    (ib_ar_group_copy_t *)p_payload,
						    p_smp->attr_mod);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7D02: "
				"Received unsupported MAD Attribute 0x%x(%s) "
				" for switch node 0x%" PRIx64 "\n",
				cl_ntoh16(p_smp->attr_id),
				ib_get_sm_attr_str(p_smp->attr_id),
				cl_ntoh64(node_guid));
			goto Lock_release;
	}

	osm_verbose_bypass_end(verbose_bypass, sm->p_log);

Lock_release:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
	return;
}
