/*
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_PLFT_BASED_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include "osm_ar_plft_based_algorithm.hpp"
#include "osm_armgr.hpp"

int PlftBasedArAlgorithm::SetRequiredPLFTInfo(ARSWDataBaseEntry &db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_advanced_routing_data_t *p_ar_data =
	&db_entry.m_general_sw_info.m_p_osm_sw->ar_data;
	ib_plft_info_t *p_plft_info = &p_ar_data->plft_info;

	u_int8_t ModeCap = p_plft_info->mode_cap;
	if (ModeCap == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9101: Switch GUID 0x%" PRIx64 ", LID %u, "
			   "Do not have sufficient PrivateLFT cap: (num tables , size).\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
		db_entry.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_ZERO);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, 1);
	}

	ib_plft_bank_entry_t *p_block_element = NULL;
	u_int8_t Mode = 0;

	uint16_t plft_max_supported_lid_number = 0;
	uint8_t min_lftable_size = (m_ar_mgr_.m_p_osm_subn->max_lid / 1024) + 1;
	uint8_t max_lftable_size = 0;
	uint8_t lftable_size;
	uint8_t plfts_per_table;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "SetRequiredPLFTInfo: Switch GUID: 0x%" PRIx64 ", LID: %u, "
		   "ModeCap: %u, max_lid: %u, min_lftable_size: %u planes_number: %u.\n",
		   db_entry.m_general_sw_info.m_guid,
		   db_entry.m_general_sw_info.m_lid,
		   ModeCap,
		   m_ar_mgr_.m_p_osm_subn->max_lid,
		   min_lftable_size,
		   m_planes_number_);

	for (int i = 0 ; i < ModeCap ; i++) {
		p_block_element = &p_plft_info->modes[i];

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID: 0x%" PRIx64 ", LID: %u, "
			   "ModeCap: %u, Mode: %u, size: %u, num_banks: %u.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   ModeCap, i,
			   p_block_element->size,
			   p_block_element->num_banks);

		if ((p_block_element->size / min_lftable_size) *
		    p_block_element->num_banks < m_planes_number_) {
		    continue;
		}

		if (p_block_element->num_banks < m_planes_number_) {
			plfts_per_table = ((m_planes_number_ - 1) /
					   p_block_element->num_banks) + 1;
		} else {
			plfts_per_table = 1;
		}

		lftable_size = p_block_element->size / plfts_per_table;

		if (lftable_size > max_lftable_size) {
			max_lftable_size  = lftable_size;
			Mode = i;
			if (max_lftable_size > min_lftable_size) {
				break; /* It is enough to have extra of 1K lids */
			}
		}
	}

	if (max_lftable_size == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9102: Switch GUID: 0x%" PRIx64 ", LID: %u, "
			   "Do not have sufficient PrivateLFT cap:"
			   "(num tables < %u or max_lid_number < %u). \n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   m_planes_number_,
			   m_ar_mgr_.m_p_osm_subn->max_lid);

		db_entry.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_LESS_THAN_MIN);

		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	osm_ar_set_required_plft_active_mode(p_ar_data, Mode + 1);
	osm_ar_set_required_plft_number(p_ar_data, m_planes_number_);

	/* max lid number is restricted by plft table size */
	plft_max_supported_lid_number = max_lftable_size * 1024 - 1;
	SetPlftMaxSupportedLidNumber(db_entry, plft_max_supported_lid_number);

	int rc = osm_ar_prepare_sw_tables(m_ar_mgr_.m_p_osm_subn,
					  db_entry.m_general_sw_info.m_p_osm_sw,
					  m_ar_mgr_.m_p_osm_subn->max_lid + 1);

	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9103: Failed to allocate for advanced "
			   "routing capabilities for switch GUID: 0x%" PRIx64 ", LID: %u\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "PLFT Info Switch GUID: 0x%" PRIx64 ", LID: %u"
		   " desired mode:%d Active_Mode:%d plft_max_supported_lid_number: %u.\n",
		   db_entry.m_general_sw_info.m_guid,
		   db_entry.m_general_sw_info.m_lid,
		   Mode + 1,
		   p_plft_info->active_mode,
		   plft_max_supported_lid_number);

	rc = SetPlftDef(db_entry,
			p_plft_info->modes[Mode].size,
			p_plft_info->modes[Mode].num_banks,
			max_lftable_size, m_planes_number_);

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int PlftBasedArAlgorithm::SetPlftDef(ARSWDataBaseEntry &db_entry,
				     uint8_t bank_size_kb,
				     uint8_t banks_number,
				     uint8_t plft_size_kb,
				     uint8_t plfts_number)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_advanced_routing_data_t *p_ar_data =
		&db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	uint8_t offset = 0;
	uint8_t bank = 0;
	uint8_t plft_id = 0;
	for (; plft_id < plfts_number; ++plft_id) {
		if ((offset + plft_size_kb) > bank_size_kb) {
			bank++;
			offset = 0;
		}
		if (bank >= banks_number) {
			break;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetPlftDef GUID: 0x%" PRIx64 ", LID: %u "
			   "plft_id: %u ofset: %u plft_size_kb: %u bank: %u.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   plft_id, offset, plft_size_kb, bank);

		osm_ar_plft_def_set(p_ar_data, plft_id, offset, plft_size_kb, bank, 0);

		offset += plft_size_kb;
	}

	if (plft_id != plfts_number) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9104: SetPlftDef failed on GUID: 0x%" PRIx64 ", LID: %u. "
			   "Invalid Arguments -  "
			   "bank_size_kb: %u banks_number: %u plft_size_kb: %u "
			   "plfts_number: %u plft_id: %u\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid,
			   bank_size_kb, banks_number, plft_size_kb, plfts_number, plft_id);
		OSM_AR_LOG_RETURN(m_p_osm_log, 1);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int PlftBasedArAlgorithm::GetSL2VL(osm_physp_t *osm_phys_port,
				   uint8_t iport_num, uint8_t oport_num,
				   uint8_t (*sl2vl)[IB_NUMBER_OF_SLS])
{
	int rc = 0;

	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_node_t *p_node = osm_physp_get_node_ptr(osm_phys_port);
	uint64_t guid = cl_ntoh64(osm_node_get_node_guid(p_node));

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetSLVL GUID: 0x%" PRIx64 " iport: %u oport: %u\n",
		   guid, iport_num, oport_num);

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {

		rc = GetSwitchSL2VL(osm_phys_port, iport_num, oport_num, p_node, guid,
				    sl2vl);

		if (OSM_LOG_DEBUG & m_p_osm_log->level) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "GetSLVL switch GUID: 0x%" PRIx64
				   " iport: %u oport: %u rc: %d sl2vl: %s\n",
				   guid, iport_num, oport_num, rc,
				   SLToVLMappingToStr(*sl2vl).c_str());
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	uint8_t op_vls = ib_port_info_get_op_vls(&osm_phys_port->port_info);
	memcpy(sl2vl, m_sl2vl_per_op_vls_[op_vls], sizeof(*sl2vl));
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void PlftBasedArAlgorithm::BuildSl2VlPerOpVl(uint16_t en_sl_mask)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* Currently support only one vl per plane, there for: */
	uint8_t vls_per_stream = m_planes_number_;

	if (en_sl_mask == m_en_sl_mask_) {
		return;
	}
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildSl2VlPerOpVl.\n");

	m_en_sl_mask_ = en_sl_mask;
	m_update_sl2vl_ = true;
	u_int8_t num_vls;
	u_int8_t ar_enabled_vl;
	u_int8_t ar_first_enabled_vl;
	u_int8_t max_ar_enabled_vl;

	memset(&m_sl2vl_per_op_vls_, 0, sizeof(m_sl2vl_per_op_vls_));

	for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));

		if (num_vls < vls_per_stream) { /* map all SLs to VL0 */
			continue;
		} else {
			max_ar_enabled_vl = num_vls - vls_per_stream;
			if (m_en_sl_mask_ == OSM_ALL_SL_VL_AR_ENABLED) {
				/*
				 * SLs will be mapped (round robin) to:
				 * VL(0) = 0, VL(n) = VL(n-1) + vls_per_stream, .. max op vl
				 */
				ar_enabled_vl = 0;
				ar_first_enabled_vl = 0;
			} else {
				if (num_vls == vls_per_stream) /* map all SLs to VL0 */
					continue;
				/*
				 * Disabled SLs will be mapped to VL 0
				 * Enabled SLs will be mapped (round robin) to
				 * vls_per_stream, .. max_ar_enabled_vl
				 */
				ar_enabled_vl = vls_per_stream;
				ar_first_enabled_vl = vls_per_stream;
			}

			for (u_int8_t sl = 0; sl < IB_NUMBER_OF_SLS; ++sl) {
				if ((1 << sl) & m_en_sl_mask_) {

					m_sl2vl_per_op_vls_[op_vls][sl] = ar_enabled_vl;
					ar_enabled_vl += vls_per_stream;
					if (ar_enabled_vl > max_ar_enabled_vl) {
						ar_enabled_vl = ar_first_enabled_vl;
					}
				} else {
					m_sl2vl_per_op_vls_[op_vls][sl] = 0;
				}
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
					   "slvl_mapping[%u][%u]=%u\n",
					   sl, op_vls, m_sl2vl_per_op_vls_[op_vls][sl]);
			}
		}
	} /* op_vls loop */

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated sl to vl mapping "
			   "vls_per_stream: %u en_sl_mask: 0x%x\n",
			   vls_per_stream, m_en_sl_mask_);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u slvl_mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_sl2vl_per_op_vls_[op_vls]).c_str());
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void PlftBasedArAlgorithm::BuildVl2VlPerOpVl()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* there is one vl per plane, there for: */
	uint8_t vls_per_stream = m_planes_number_;

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildVl2VlPerOpVl.\n");
	u_int8_t num_vls;
	u_int8_t max_vl;
	u_int8_t qos_max_vl; /* max vl for the current qos */

	memset(m_vl2vl_per_op_vls_, 0, sizeof(m_vl2vl_per_op_vls_));
	memset(m_inc_vl2vl_per_op_vls_, 0, sizeof(m_inc_vl2vl_per_op_vls_));

	for (uint8_t op_vls = 2; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

		num_vls = ( 1 << (op_vls-1));
		max_vl = min((uint8_t)(num_vls - 1), (uint8_t)(IB_DROP_VL - 1));

		if (num_vls < vls_per_stream)
			continue; /* Not Supported */

		for (u_int8_t in_vl = 0; in_vl < IB_NUMBER_OF_SLS; ++in_vl) {
			/* vl < num_vls */
			u_int8_t vl = in_vl % num_vls;

			/*
			 * The qos groups number of given vl [1..last] is:
			 * (vl / vls_per_stream) + 1
			 *
			 * In each qos group there are vls_per_stream vls so for vl in specific
			 * qos group the max vl is:
			 */
			qos_max_vl = ((vl / vls_per_stream) + 1 ) * vls_per_stream - 1;
			qos_max_vl = min(qos_max_vl, max_vl);

			m_vl2vl_per_op_vls_[op_vls][in_vl] = min(qos_max_vl, vl);
			m_inc_vl2vl_per_op_vls_[op_vls][in_vl] = min(qos_max_vl, (uint8_t)(vl + 1));
		}
	} /* op_vls loop */

	if (OSM_LOG_DEBUG & m_p_osm_log->level) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculated vl to vl mapping vls_per_stream: %u\n",
			   vls_per_stream);

		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_vl2vl_per_op_vls_[op_vls]).c_str());
		}
		for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "op_vls: %u increase vl to vl mapping %s\n",
				   op_vls,
				   SLToVLMappingToStr(m_inc_vl2vl_per_op_vls_[op_vls]).c_str());
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/* TODO remove doplicated code: AdaptiveRoutingManager::ARDragonFlyCycleEnd() */
void PlftBasedArAlgorithm::CycleEnd(int rc)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "CycleEnd.\n");

	m_update_sl2vl_ = false;

	if (m_ar_mgr_.m_is_permanent_error)
		m_ar_mgr_.m_is_temporary_error = false;
	else if (rc && !m_ar_mgr_.m_is_temporary_error)
		m_ar_mgr_.m_is_permanent_error = true;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (sw_node_it = m_sw_map_.begin();
		sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		RestorePriv(sw_node_it->second);
	}

	if (m_ar_mgr_.m_is_permanent_error) {
		ClearAlgorithmConfiguration();
	} else {
		map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
		for (sw_node_it = m_sw_map_.begin();
			sw_node_it != m_sw_map_.end(); ++sw_node_it) {
			if (IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
				sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;
			}
		}
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void PlftBasedArAlgorithm::ClearAlgorithmConfiguration()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (sw_node_it = m_sw_map_.begin();
		sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		ClearSwitchAlgotithmData(sw_node_it->second);
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

string PlftBasedArAlgorithm::SLToVLMappingToStr(const sl2vl_t &sl2vl) const
{
	string str = "";
	char buff[1024];
	sprintf(buff,"%u, %u, %u, %u, %u, %u, %u, %u, "
		"%u, %u, %u, %u, %u, %u, %u, %u",
		sl2vl[0],
		sl2vl[1],
		sl2vl[2],
		sl2vl[3],
		sl2vl[4],
		sl2vl[5],
		sl2vl[6],
		sl2vl[7],
		sl2vl[8],
		sl2vl[9],
		sl2vl[10],
		sl2vl[11],
		sl2vl[12],
		sl2vl[13],
		sl2vl[14],
		sl2vl[15]);
	str += buff;
	return(str);
}
