/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Capability manager
 *
 *      The capability manager manage decisions as for which capability is
 *      enabled for which vendor id and device id.
 *
 *      The manager process expressions which define what capabilities are
 *      enabled for which vendor and device and it responds to queries whether
 *      a given combination of cap_id + vendor_id + device_id is enabled.
 *
 *	The capability manager process expressions with the below format:
 *
 *	<DEVICES_LIST> = Comma separated list of device id's or device id's range
 *
 *			Devices list Examples:
 *
 *			DEV_ID_1, DEV_ID_2, DEV_ID_3
 *			DEV_ID_1, DEV_ID_2 - DEV_ID_3, DEV_ID_4
 *
 *	<VENDOR_LIST> = A tupple of vendor id and <DEVICES_LIST> related to it
 *
 *			The vendor id is optional, if <DEVICES_LIST> is
 *			specified without vendor id the capability is enabled
 *			for all vendors
 *
 *			Vendor list Examples:
 *
 *			<vendor_id> : <DEVICES_LIST>
 *			<DEVICES_LIST>
 *
 *	the capability manager expects expressions which are built out of
 *	one or more <VENDOR_LISTS>'s, if more than one list is specified
 *	each list shall be located within [ ], If only one list is specified
 *	it can either be with or without [ ]
 *
 *			Examples for valid expressions:
 *
 *			<DEVICES_LIST>
 *			[<DEVICES_LIST>]
 *			<VENDOR_LIST>
 *			[<VENDOR_LIST>]
 *			[<VENDOR_LIST>][<VENDOR_LIST>][<VENDOR_LIST>]
 *			[<VENDOR_LIST>][<DEVICES_LIST>][<VENDOR_LIST>]
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CAPABILITY_MGR_C
#include <opensm/osm_capability_mgr.h>
#include <opensm/osm_log.h>
#include <opensm/osm_opensm.h>

const char *SUBLIST_END = "]";
const int SUBLIST_START = '[';
const int VEND_DEV_DELIM = ':';
const char *DEV_DELIM = ",";
const int RANGE_DELIM = '-';

/* Max dev id should be '0x' + 4 hex digits */
#define MAX_DEV_ID_LEN_HEX 6

#define GI_SUPPORTING_DEVICES  "[0x119f:0x1b83,0x1b93,0x1b94,0x1bb4,0x1bb5,  \
					0x1bc4-0x1bc9,0x1ba0,0x1bd0-0x1bd5,  \
					0x1bf0]  \
				[0x1011,0x1013-0x101b,0xcb20,0xcf08,0xcf09, \
				 0xd2f0, 0xa2d2]"

#define MEPI_FORCE_DEVICES "[0x119f:0x1b33,0x1b73,0x1b40,0x1b41,0x1b60,0x1b61, \
				    0x1b02,0x1b50]        \
			    [0x1003-0x1007,0xc738-0xc73b]"

extern void remove_whitespaces(char *str);
extern boolean_t validate_hex(const char* str);

/* We use host order so the key will be readable while looking at logs */
static uint64_t build_key(osm_capability_id_t cap_id, uint32_t vendor_id,
			  uint16_t device_id)
{
	uint64_t capability = (uint64_t)cap_id;
	uint64_t device = (uint64_t)device_id;

	return (capability << 48) | (device << 32) | vendor_id;
}

/* Enable capability for specific vendor and device */
static void osm_cap_mgr_enable(osm_capability_mgr_t *p_cap_mgr,
			       osm_log_t *p_log, osm_capability_id_t cap_id,
			       uint32_t vendor_id, uint16_t device_id)
{
	uint64_t key = build_key(cap_id, vendor_id, device_id);

	OSM_LOG_ENTER(p_log);

	if (cl_map_get(&p_cap_mgr->capabilities, key)) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "CapMgr: Ignoring request to "
			"enable new capability: cap id = %d, vendor id = 0x%"
			PRIx32 ", device id = 0x%" PRIx16 " - already enabled\n",
			cap_id, vendor_id, device_id);
		goto Exit;
	}

	OSM_LOG(p_log, OSM_LOG_DEBUG, "CapMgr: Enabling new capability: "
		"cap id = %d, vendor id = 0x%" PRIx32 ", device id = 0x%" PRIx16
		"\n", cap_id, vendor_id, device_id);

	/*
	 * We just need to add any pointer under this key, it does not matter
	 * where does it pointing to.
	 */
	cl_map_insert(&p_cap_mgr->capabilities, key, p_cap_mgr);

Exit:
	OSM_LOG_EXIT(p_log);
}

boolean_t osm_cap_mgr_is_enabled(osm_capability_mgr_t *p_cap_mgr,
				 osm_capability_id_t cap_id,
				 struct osm_node *p_node)
{
	ib_net32_t vendor_id = ib_node_info_get_vendor_id(&p_node->node_info);
	ib_net16_t device_id = p_node->node_info.device_id;
	uint64_t key = build_key(cap_id, cl_ntoh32(vendor_id),
				 cl_ntoh16(device_id));

	if (cl_map_get(&p_cap_mgr->capabilities, key) != NULL)
		return TRUE;

	/*
	 *  The cap is not enabled for this dev-id and vendor id but maybe it is
	 *  enabled for ANY_VENDOR
	 */
	key = build_key(cap_id, ANY_VENDOR, cl_ntoh16(device_id));

	return (cl_map_get(&p_cap_mgr->capabilities, key) != NULL);
}

static boolean_t validate_hex_min_max(char *min, char *max, osm_log_t *p_log,
				      uint16_t *min_dev_id, uint16_t *max_dev_id)
{
	boolean_t res = FALSE;
	char *e;

	if (!validate_hex(min)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A06: "
			"CapMgr: Parsed Ilegal min dev id value (%s)\n", min);
		goto Exit;
	}
	if (!validate_hex(max)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A07: "
			"CapMgr: Parsed Ilegal max dev id value (%s)\n", max);
		goto Exit;
	}

	*min_dev_id = strtoul(min, &e, 0);
	if (!*min_dev_id) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A08: "
			"CapMgr: Parsed Ilegal min dev id value (%s)\n", min);
		goto Exit;
	}

	*max_dev_id = strtoul(max, &e, 0);
	if (!*max_dev_id) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A09: "
			"CapMgr: Parsed Ilegal max dev id value (%s)\n", max);
		goto Exit;
	}

	if (*min_dev_id > *max_dev_id) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A0A: "
			"CapMgr: Parsed min val (%s) higher than max val (%s)\n",
			min, max);
		goto Exit;
	}

	res = TRUE;
Exit:
	return res;
}

/* Process devices expression for a given vendor */
static void process_vendor(osm_capability_mgr_t *p_cap_mgr, osm_log_t *p_log,
			   osm_capability_id_t cap_id, uint32_t vendor_id,
			   char *devices)
{
	char *token;
	char *context = NULL;
	char *max, *e, *hyphen;
	uint16_t dev_id, dev_id_min, dev_id_max;
	char min[MAX_DEV_ID_LEN_HEX + 1];

	OSM_LOG_ENTER(p_log);

	token = strtok_r(devices, DEV_DELIM, &context);

	while (token) {
		hyphen = strchr(token, RANGE_DELIM);
		if (hyphen) {
			if ((hyphen - token) > MAX_DEV_ID_LEN_HEX) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A05: "
					"CapMgr: Parsed Ilegal device id (%s)\n",
					token);
				goto Exit;
			}
			memset(min, 0, MAX_DEV_ID_LEN_HEX + 1);
			memcpy(min, token, hyphen - token);
			max = hyphen + 1;
			if (validate_hex_min_max(min, max, p_log,
                                      &dev_id_min, &dev_id_max)) {
				for (dev_id = dev_id_min ;
					dev_id <= dev_id_max ; dev_id++)
					osm_cap_mgr_enable(p_cap_mgr, p_log, cap_id,
							   vendor_id, dev_id);
			}
		} else {
			if (!validate_hex(token)) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A04: "
					"CapMgr: Parsed Ilegal device id (%s)\n",
					token);
				goto Exit;
			}
			dev_id = strtoul(token, &e, 0);
			if (dev_id)
				osm_cap_mgr_enable(p_cap_mgr, p_log, cap_id,
						   vendor_id, dev_id);
		}
		token = strtok_r(NULL, DEV_DELIM, &context);
	}
Exit:
	OSM_LOG_EXIT(p_log);
}

/* Process a single vendor expression */
static void process_single_vendor_exp(osm_capability_mgr_t *p_cap_mgr,
				      osm_log_t *p_log,
				      osm_capability_id_t cap_id,
				      char *exp)
{
	char *vend_str = NULL, *dev_delim, *e;
	uint32_t vendor_id = 0;

	OSM_LOG_ENTER(p_log);

	dev_delim = strchr(exp, VEND_DEV_DELIM);

	if (dev_delim) {
		vend_str = calloc(1, (dev_delim - exp) + 1);
		if (!vend_str) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A02: "
				"CapMgr: Memory allocation failed\n");
			goto Exit;
		}
		memcpy(vend_str, exp, (dev_delim - exp));
		if (validate_hex(vend_str))
			vendor_id = strtoul(vend_str, &e, 0);

		if (!vendor_id) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A03: "
				"CapMgr: Invalid vendor id configured (%s)\n",
				vend_str);
			goto Exit;
		}
		process_vendor(p_cap_mgr, p_log, cap_id, vendor_id, dev_delim + 1);
	} else
		process_vendor(p_cap_mgr, p_log, cap_id, ANY_VENDOR, exp);

Exit:
	free(vend_str);
	OSM_LOG_EXIT(p_log);
}

/* Process expression for a given capability */
static void process_cap_exp(osm_capability_mgr_t *p_cap_mgr, osm_log_t *p_log,
				 osm_capability_id_t cap_id,
				 const char *exp)
{
	char *token;
	char *context = NULL;
	char *exp_copy = NULL, *open_bracket;

	OSM_LOG_ENTER(p_log);

	if (!exp || !strlen(exp))
		goto Exit;

	/* We do not chance the input expression so we work with a copy */
	exp_copy = strdup(exp);
	if (!exp_copy) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A01: "
				"CapMgr: Memory allocation failed\n");
		goto Exit;
	}

	remove_whitespaces(exp_copy);

	/* Handle the specific case of a single exp without any brackets */
	if (!strchr(exp_copy, SUBLIST_START)) {
		process_single_vendor_exp(p_cap_mgr, p_log, cap_id, exp_copy);
		goto Exit;
	}

	token = strtok_r(exp_copy, SUBLIST_END, &context);

	while (token) {
		open_bracket = strchr(token, SUBLIST_START);
		if (open_bracket) {
			if (open_bracket == token)
				process_single_vendor_exp(p_cap_mgr, p_log,
							  cap_id,
							  open_bracket + 1);
			else
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A0B: "
					"CapMgr: Devices expression parsing for cap "
					"id %d failed |%s|\n", cap_id, token);
		} else {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 6A00: "
				"CapMgr: Devices expression parsing for cap "
				"id %d failed |%s|\n", cap_id, token);
			break;
		}
		token = strtok_r(NULL, SUBLIST_END, &context);
	}
Exit:
	free(exp_copy);
	OSM_LOG_EXIT(p_log);
}

void osm_cap_mgr_process(IN osm_capability_mgr_t *p_cap_mgr, IN struct osm_sm *p_sm)
{
	OSM_LOG_ENTER(p_sm->p_log);

	cl_map_remove_all(&p_cap_mgr->capabilities);

	if (!p_sm->p_subn->opt.disable_default_gi_support) {
		process_cap_exp(p_cap_mgr, p_sm->p_log, GI, GI_SUPPORTING_DEVICES);
	}

	process_cap_exp(p_cap_mgr, p_sm->p_log, MEPI, MEPI_FORCE_DEVICES);
	process_cap_exp(p_cap_mgr, p_sm->p_log, GI,
			p_sm->p_subn->opt.additional_gi_supporting_devices);
	process_cap_exp(p_cap_mgr, p_sm->p_log, MEPI,
			p_sm->p_subn->opt.additional_mepi_force_devices);

	OSM_LOG_EXIT(p_sm->p_log);
}

ib_api_status_t osm_cap_mgr_init(IN osm_capability_mgr_t *p_cap_mgr,
				 IN struct osm_sm *p_sm)
{
	cl_map_construct(&p_cap_mgr->capabilities);
	cl_map_init(&p_cap_mgr->capabilities, MIN_CAPABILITIES_TO_KEEP);
	return IB_SUCCESS;
}
