/*
 * Copyright (c) 2006-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2012 Lawrence Livermore National Lab.  All rights reserved.
 * Copyright (c) 2014-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    OSM Congestion Control configuration implementation
 *
 * Author:
 *    Albert Chu, LLNL
 *    Alex Minchiu, Mellanox
 *    Daniel Klein, Mellanox
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CONGESTION_CONTROL_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_congestion_control.h>

#define CONGESTION_CONTROL_INITIAL_TID_VALUE 0x7A93

extern int parse_congestion_control_policy(struct osm_opensm *osm_ptr);

int cc_compare_profiles(IN const void * p_profiles_name_1, IN const void * p_profiles_name_2)
{
	return strcmp((const char *)p_profiles_name_1, (const char *)p_profiles_name_2);
}


void osm_cc_mgr_construct(IN osm_cc_mgr_t * p_mgr)
{
	cl_fmap_init(&p_mgr->cc_profiles_map, cc_compare_profiles);
	cl_qlist_init(&p_mgr->sl_profiles_conf);
	cl_qlist_init(&p_mgr->vl_profiles_conf);
	cl_qlist_init(&p_mgr->sw_conf);
	cl_qlist_init(&p_mgr->ca_conf);
}

void osm_cc_mgr_destroy(IN osm_cc_mgr_t * p_mgr)
{
	cl_fmap_item_t *p_item, *p_next_item;
	osm_cc_sl_conf_t *p_sl_conf = NULL;
	osm_cc_vl_conf_t *p_vl_conf = NULL;
	osm_cc_ca_conf_t *p_ca_conf = NULL;
	osm_cc_sw_conf_t *p_sw_conf = NULL;

	p_next_item = cl_fmap_head(&p_mgr->cc_profiles_map);
	while (p_next_item != cl_fmap_end(&p_mgr->cc_profiles_map)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&p_mgr->cc_profiles_map, p_item);
		free((osm_cc_profiles_conf_t *) p_item);
	}

	while (cl_qlist_count(&p_mgr->sl_profiles_conf) > 0) {
		p_sl_conf =
		    (osm_cc_sl_conf_t *) cl_qlist_remove_head(&p_mgr->sl_profiles_conf);
		free(p_sl_conf);
	}

	while (cl_qlist_count(&p_mgr->vl_profiles_conf) > 0) {
		p_vl_conf =
		    (osm_cc_vl_conf_t *) cl_qlist_remove_head(&p_mgr->vl_profiles_conf);
		free(p_vl_conf);
	}

	while (cl_qlist_count(&p_mgr->sw_conf) > 0) {
		p_sw_conf =
		    (osm_cc_sw_conf_t *) cl_qlist_remove_head(&p_mgr->sw_conf);
		free(p_sw_conf);
	}

	while (cl_qlist_count(&p_mgr->ca_conf) > 0) {
		p_ca_conf =
		    (osm_cc_ca_conf_t *) cl_qlist_remove_head(&p_mgr->ca_conf);
		free(p_ca_conf);
	}
}

/* Clear CA NP data and mark for resend */
static void cc_clear_ca_subn_np_params(osm_port_t *p_port)
{
	p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_NP_PARAMETERS;
	memset(&p_port->cc_data.subn_mads.hca_np_params, 0,
	       sizeof(p_port->cc_data.subn_mads.hca_np_params));
}

/* Clear CA RP data and mark for resend */
static void cc_clear_ca_subn_rp_params(osm_port_t *p_port)
{
	p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_RP_PARAMETERS;
	memset(&p_port->cc_data.subn_mads.hca_rp_params, 0,
	       sizeof(p_port->cc_data.subn_mads.hca_rp_params));
}

static void cc_clear_ca_subn_data(osm_port_t *p_port)
{
	/*
	 * Mark CA for re-query capabilities and resend NP and NR
	 * configurations.
	 */
	p_port->cc_data.resend_mask |= OSM_CC_RESEND_ALL;
	p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_NP_PARAMETERS;
	p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_RP_PARAMETERS;

	/* Clear data */
	memset(&p_port->cc_data.subn_mads, 0, sizeof(p_port->cc_data.subn_mads));
}

static void cc_clear_sw_subn_ports_data(osm_switch_t *p_switch)
{
	uint8_t port;
	osm_node_t *p_node = p_switch->p_node;

	/* Mark switch to resend SL mapping configurations */
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SW_SL_MAPPING;

	/* Clear data from ports */
	for (port = 0; port < p_switch->num_ports; port++) {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port);
		memset(&p_physp->cc_data.subn_mads, 0,
		       sizeof(p_physp->cc_data.subn_mads));
	}
}

static void cc_clear_sw_subn_data(osm_switch_t *p_switch)
{
	/* Mark switch for re-query capabilities and resend configurations */
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_ALL;
	p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SW_SL_MAPPING;

	cc_clear_sw_subn_ports_data(p_switch);
}

/* Check if CA is marked for full update */
static boolean_t cc_check_ca_need_update(const osm_port_t *p_port, uint8_t mask)
{
	return p_port->cc_data.resend_mask & (OSM_CC_RESEND_ALL | mask);
}

/* Check if switch is marked for full update */
static boolean_t cc_check_sw_need_update(const osm_switch_t *p_switch, uint8_t mask)
{
	return p_switch->cc_data.resend_mask & (OSM_CC_RESEND_ALL | mask);
}

/* Disable CA need update flags */
static inline void cc_diable_ca_need_update(osm_port_t *p_port, uint8_t mask)
{
	p_port->cc_data.resend_mask &= ~mask;
}

/* Disable switch need update flags */
static inline void cc_diable_sw_need_update(osm_switch_t *p_switch, uint8_t mask)
{
	p_switch->cc_data.resend_mask &= ~mask;
}

/*
 * Mark devices that needs to be updated and reset information
 * of configuration that is set on the subnet.
 */
static void cc_mark_need_update(struct osm_opensm * p_osm)
{
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *)p_next;
		osm_node_t *p_node =  p_port->p_node;
		osm_switch_t *p_switch = p_node->sw;
		p_next = cl_qmap_next(p_next);

		/* Check if device is marked as needs update */
		if (!p_osm->subn.need_update &&
		    !osm_port_get_physp(p_port)->need_update &&
		    !(p_switch && p_switch->need_update)) {
			continue;
		}

		if (p_switch) {
			cc_clear_sw_subn_data(p_switch);
		} else if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA) {
			cc_clear_ca_subn_data(p_port);
		}
	}
}

osm_cc_group_type_enum osm_cc_get_group_type(IN const char *str)
{
	if (!str || !strcasecmp(str, "null"))
		return OSM_CC_GROUP_NOT_DEFINED;
	else if (!strcasecmp(str, "ALL_SWITCHES"))
		return OSM_CC_GROUP_ALL_SWITCHES;
	else if (!strcasecmp(str, "SWITCH_PORTS"))
		return OSM_CC_GROUP_SWITCH_PORTS;
	else
		return OSM_CC_GROUP_NOT_DEFINED;
}

static
void print_profiles_conf(struct osm_opensm * p_osm, osm_cc_profiles_conf_t *p_profiles_conf)
{
	int i;
	osm_cc_profile_t *p_profile;

	for(i = 0; i < 3; i++) {
		p_profile = &p_profiles_conf->profiles[i];
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"profile_name:%s (mode:%u), "
			"index:%u percent:%u, min:%lu, max:%lu\n",
			p_profiles_conf->name, p_profiles_conf->mode, i,
			p_profile->percent, p_profile->min, p_profile->max);
	}
}

static
void print_profiles_map(struct osm_opensm * p_osm, cl_fmap_t *p_profiles)
{
	cl_fmap_item_t *p_item;

	for (p_item = cl_fmap_head(p_profiles);
	     p_item != cl_fmap_end(p_profiles);
	     p_item = cl_fmap_next(p_item)) {
		print_profiles_conf(p_osm, (osm_cc_profiles_conf_t *)p_item);
	}
}

static
void print_sl_conf(struct osm_opensm * p_osm, cl_qlist_t *p_sl_conf_list)
{
	cl_list_item_t *p_item;
	osm_cc_sl_conf_t *p_sl_conf;

	p_item = cl_qlist_head(p_sl_conf_list);
	while (p_item != cl_qlist_end(p_sl_conf_list)) {
		p_sl_conf = (osm_cc_sl_conf_t *)p_item;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"CC group %d SL to profile: "
			"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
			p_sl_conf->group_type,
			p_sl_conf->sl2profile_num[0], p_sl_conf->sl2profile_num[1],
			p_sl_conf->sl2profile_num[2], p_sl_conf->sl2profile_num[3],
			p_sl_conf->sl2profile_num[4], p_sl_conf->sl2profile_num[5],
			p_sl_conf->sl2profile_num[6], p_sl_conf->sl2profile_num[7],
			p_sl_conf->sl2profile_num[8], p_sl_conf->sl2profile_num[9],
			p_sl_conf->sl2profile_num[10], p_sl_conf->sl2profile_num[11],
			p_sl_conf->sl2profile_num[12], p_sl_conf->sl2profile_num[13],
			p_sl_conf->sl2profile_num[14], p_sl_conf->sl2profile_num[15]);
		p_item = cl_qlist_next(p_item);
	}
}

static
void print_vl_conf(struct osm_opensm * p_osm, cl_qlist_t *p_vl_conf_list)
{
	cl_list_item_t *p_item;
	osm_cc_vl_conf_t *p_vl_conf;
	osm_cc_profiles_conf_t **p_vl_to_profiles;

	p_item = cl_qlist_head(p_vl_conf_list);
	while (p_item != cl_qlist_end(p_vl_conf_list)) {
		p_vl_conf = (osm_cc_vl_conf_t *)p_item;
		p_vl_to_profiles = p_vl_conf->vl_to_profiles;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"CC group %d VL to profile: "
			"%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
			p_vl_conf->group_type,
			p_vl_to_profiles[0]->name, p_vl_to_profiles[1]->name,
			p_vl_to_profiles[2]->name, p_vl_to_profiles[3]->name,
			p_vl_to_profiles[4]->name, p_vl_to_profiles[5]->name,
			p_vl_to_profiles[6]->name, p_vl_to_profiles[7]->name,
			p_vl_to_profiles[8]->name, p_vl_to_profiles[9]->name,
			p_vl_to_profiles[10]->name, p_vl_to_profiles[11]->name,
			p_vl_to_profiles[12]->name, p_vl_to_profiles[13]->name,
			p_vl_to_profiles[14]->name, p_vl_to_profiles[15]->name);
		p_item = cl_qlist_next(p_item);
	}
}

static void cc_mad_post(osm_congestion_control_t *p_cc, osm_madw_t *p_madw,
			osm_node_t *p_node, osm_physp_t *p_physp,
			ib_net16_t attr_id, ib_net32_t attr_mod,
			uint8_t method_id)
{
	osm_subn_opt_t *p_opt = &p_cc->subn->opt;
	ib_cc_mad_t *p_cc_mad;
	uint8_t port;

	OSM_LOG_ENTER(p_cc->log);

	port = osm_physp_get_port_num(p_physp);

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_cc_mad->header.base_ver = 1;
	p_cc_mad->header.mgmt_class = IB_MCLASS_CC;
	p_cc_mad->header.class_ver = 2;
	p_cc_mad->header.method = method_id;
	p_cc_mad->header.status = 0;
	p_cc_mad->header.class_spec = 0;
	p_cc_mad->header.trans_id =
		cl_hton64((uint64_t) cl_atomic_inc(&p_cc->trans_id) &
			  (uint64_t) (0xFFFFFFFF));
	if (p_cc_mad->header.trans_id == 0)
		p_cc_mad->header.trans_id =
			cl_hton64((uint64_t) cl_atomic_inc(&p_cc->trans_id) &
				  (uint64_t) (0xFFFFFFFF));
	p_cc_mad->header.attr_id = attr_id;
	p_cc_mad->header.resv = 0;
	p_cc_mad->header.attr_mod = attr_mod;

	p_cc_mad->cc_key = p_opt->cc_key;

	memset(p_cc_mad->log_data, '\0', IB_CC_LOG_DATA_SIZE);

	p_madw->mad_addr.dest_lid = osm_node_get_base_lid(p_node, port);
	p_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	p_madw->mad_addr.addr_type.gsi.remote_qkey =
		cl_hton32(IB_QP1_WELL_KNOWN_Q_KEY);
	p_madw->resp_expected = TRUE;
	p_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_madw->context.cc_context.node_guid = osm_node_get_node_guid(p_node);
	p_madw->context.cc_context.port_guid = osm_physp_get_port_guid(p_physp);
	p_madw->context.cc_context.port = port;
	p_madw->context.cc_context.mad_method = method_id;
	p_madw->context.cc_context.attr_mod = attr_mod;

	cl_spinlock_acquire(&p_cc->mad_queue_lock);
	cl_atomic_inc(&p_cc->outstanding_mads);
	cl_qlist_insert_tail(&p_cc->mad_queue, &p_madw->list_item);
	cl_spinlock_release(&p_cc->mad_queue_lock);

	cl_event_signal(&p_cc->cc_poller_wakeup);

	OSM_LOG_EXIT(p_cc->log);
}


static ib_api_status_t
cc_send_hca_general_settings(osm_sm_t * p_sm,
			     osm_node_t *p_node,
			     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_general_settings_t *p_ca_general_setting = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, 0) &&
	    !memcmp(&p_port->cc_data.calc_mads.hca_general_settings,
		    &p_port->cc_data.subn_mads.hca_general_settings,
		    sizeof(p_port->cc_data.calc_mads.hca_general_settings)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_ca_general_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_ca_general_setting,
	       &p_port->cc_data.calc_mads.hca_general_settings,
	       sizeof(p_port->cc_data.calc_mads.hca_general_settings));

	cc_mad_post(p_cc, p_madw, p_node, p_physp,
		    IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_hca_rp_params(osm_sm_t * p_sm,
					     osm_node_t *p_node,
					     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_rp_parameters_t *p_hca_rp_params = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_CA_RP_PARAMETERS) &&
	    !memcmp(&p_port->cc_data.calc_mads.hca_rp_params,
		    &p_port->cc_data.subn_mads.hca_rp_params,
		    sizeof(p_port->cc_data.subn_mads.hca_rp_params)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_hca_rp_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_hca_rp_params,
	       &p_port->cc_data.calc_mads.hca_rp_params,
	       sizeof(p_port->cc_data.calc_mads.hca_rp_params));

	cc_mad_post(p_cc, p_madw, p_node, p_physp,
		    IB_MAD_ATTR_CC_HCA_RP_PARAMETERS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_hca_np_params(osm_sm_t * p_sm,
					     osm_node_t *p_node,
					     osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_hca_np_parameters_t *p_hca_np_params = NULL;
	osm_physp_t *p_physp = p_port->p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!cc_check_ca_need_update(p_port, OSM_CC_RESEND_CA_NP_PARAMETERS) &&
	    !memcmp(&p_port->cc_data.calc_mads.hca_np_params,
		    &p_port->cc_data.subn_mads.hca_np_params,
		    sizeof(p_port->cc_data.subn_mads.hca_np_params)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);
	p_hca_np_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_hca_np_params,
	       &p_port->cc_data.calc_mads.hca_np_params,
	       sizeof(p_port->cc_data.calc_mads.hca_np_params));

	cc_mad_post(p_cc, p_madw, p_node, p_physp,
		    IB_MAD_ATTR_CC_HCA_NP_PARAMETERS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_cc_enh_info(osm_sm_t * p_sm,
					   osm_port_t *p_port)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		if (!cc_check_sw_need_update(p_node->sw, 0))
			return IB_SUCCESS;
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	} else {
		if (!cc_check_ca_need_update(p_port, 0))
			return IB_SUCCESS;
		p_physp = p_port->p_physp;
	}

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cc_mad_post(p_cc, p_madw, p_node, p_physp,
		    IB_MAD_ATTR_CC_ENH_INFO, 0, IB_MAD_METHOD_GET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t
cc_send_port_profile_settings(osm_sm_t * p_sm,
			      osm_node_t *p_node,
			      osm_physp_t *p_physp,
			      ib_net32_t attr_mod)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	uint8_t force_update;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	osm_physp_t *p_physp0 = NULL;
	ib_cc_port_profile_settings_t *p_port_profile_setting = NULL;
	boolean_t set_all_vls = TRUE;
	uint8_t vl;

	OSM_LOG_ENTER(p_sm->p_log);

	OSM_ASSERT(p_node->sw);

	p_physp0 = osm_node_get_physp_ptr(p_node, 0);

	force_update = cc_check_sw_need_update(p_node->sw, 0);

	/*
	 * HACK: check if can set all VLs profiles using single MAD, by
	 * comparing all VLs profiles to VL0 profile.
	 */
	for (vl = 1; vl < IB_DROP_VL; vl++) {
		if (memcmp(&p_physp->cc_data.calc_mads.vl_to_profiles[0],
			   &p_physp->cc_data.calc_mads.vl_to_profiles[vl],
			   sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl]))) {
			set_all_vls = FALSE;
			break;
		}
	}

	for (vl = 0; vl < IB_DROP_VL/*&p_sm->p_subn->opt.max_op_vls*/; vl++) {
		if (!force_update &&
		    !memcmp(&p_physp->cc_data.calc_mads.vl_to_profiles[vl],
			    &p_physp->cc_data.subn_mads.vl_to_profiles[vl],
			    sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl])))
			continue;

		p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
		if (p_madw == NULL) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
				"failed to allocate mad\n");
			return IB_INSUFFICIENT_MEMORY;
		}

		p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

		p_port_profile_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

		memcpy(p_port_profile_setting,
		       &p_physp->cc_data.calc_mads.vl_to_profiles[vl],
		       sizeof(p_physp->cc_data.calc_mads.vl_to_profiles[vl]));

		/* Set VL mask of outgoing MAD */
		if (set_all_vls)
			p_port_profile_setting->vl_mask = 0xffff;
		else
			p_port_profile_setting->vl_mask = cl_hton16(1 << vl);

		cc_mad_post(p_cc, p_madw, p_node, p_physp0,
			    IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS,
			    attr_mod, IB_MAD_METHOD_SET);

		if (set_all_vls)
			/*
			 * We sent single set request for all vls, no need to
			 * continue
			 */
			break;
	}

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t
cc_send_sl_mapping_settings(osm_sm_t * p_sm,
			    osm_node_t *p_node,
			    osm_physp_t *p_physp,
			    ib_net32_t attr_mod)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	osm_physp_t *p_physp0 = NULL;
	ib_cc_sl_mapping_settings_t *p_cc_sl_setting = NULL;

	OSM_LOG_ENTER(p_sm->p_log);
	OSM_ASSERT(p_node->sw);

	p_physp0 = osm_node_get_physp_ptr(p_node, 0);

	if (!cc_check_sw_need_update(p_node->sw, OSM_CC_RESEND_SW_SL_MAPPING) &&
	    !memcmp(&p_physp->cc_data.calc_mads.sl_mapping,
		    &p_physp->cc_data.subn_mads.sl_mapping,
		    sizeof(p_physp->cc_data.calc_mads.sl_mapping)))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C110: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_cc_sl_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_cc_sl_setting,
	       &p_physp->cc_data.calc_mads.sl_mapping,
	       sizeof(p_physp->cc_data.calc_mads.sl_mapping));

	cc_mad_post(p_cc, p_madw, p_node, p_physp0,
		    IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS,
		    attr_mod, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_sw_general_settings(osm_sm_t * p_sm,
						   osm_switch_t *p_switch)
{
	osm_congestion_control_t *p_cc = &p_sm->p_subn->p_osm->cc;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw = NULL;
	ib_cc_mad_t *p_cc_mad = NULL;
	ib_cc_sw_general_settings_t *p_sw_general_settings = NULL;
	ib_cc_sw_general_settings_t *p_calc_sw_general_settings;
	ib_cc_sw_general_settings_t *p_subn_sw_general_settings;
	osm_node_t *p_node = p_switch->p_node;

	OSM_LOG_ENTER(p_sm->p_log);

	p_physp = osm_node_get_physp_ptr(p_node, 0);

	p_calc_sw_general_settings = &p_switch->cc_data.calc_sw_general_settings;
	p_subn_sw_general_settings = &p_switch->cc_data.subn_sw_general_settings;

	/* Send MAD if forced update or non-read only fields were changed */
	if (!cc_check_sw_need_update(p_switch, 0) &&
	    (p_calc_sw_general_settings->enable == p_subn_sw_general_settings->enable &&
	     p_calc_sw_general_settings->aqs_weight == p_subn_sw_general_settings->aqs_weight &&
	     p_calc_sw_general_settings->aqs_time == p_subn_sw_general_settings->aqs_time))
		return IB_SUCCESS;

	p_madw = osm_mad_pool_get(p_cc->mad_pool, p_cc->bind_handle,
				  MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR C101: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	p_sw_general_settings = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);

	memcpy(p_sw_general_settings,
	       &p_switch->cc_data.calc_sw_general_settings,
	       sizeof(p_switch->cc_data.calc_sw_general_settings));

	cc_mad_post(p_cc, p_madw, p_node, p_physp,
		    IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS, 0, IB_MAD_METHOD_SET);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static ib_api_status_t cc_send_to_ext_ports(osm_sm_t * p_sm, osm_node_t *p_node,
					    ib_api_status_t (send)(osm_sm_t *, osm_node_t *,
								   osm_physp_t *, ib_net32_t attr_mod))
{
	int ret = 0;
	osm_physp_t *p_ext_port_physp;
	uint8_t	ext_port_num;
	ib_api_status_t status;
	uint8_t num_ports = osm_node_get_num_physp(p_node);

	for (ext_port_num = 1; ext_port_num < num_ports; ext_port_num++) {
		p_ext_port_physp = osm_node_get_physp_ptr(p_node, ext_port_num);
		/* skip special ports */
		if (ib_mlnx_epi_get_is_special_port_bit(&p_ext_port_physp->ext_port_info))
			continue;

		status = send(p_sm, p_node, p_ext_port_physp,
			      cl_hton32((uint32_t)ext_port_num));
		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_profiles(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_item;
	ib_api_status_t status;
	osm_physp_t *p_physp;

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;
		p_physp = osm_node_get_physp_ptr(p_node, 0);

		if (!ib_cc_get_enh_info_ver0(&p_node->cc_enh_info))
			continue;

		/* Skip device when congestion control is disabled on it */
		if (!ib_cc_sw_get_eanble_cc(&p_switch->cc_data.calc_sw_general_settings)) {
			/* Mark SW for future update */
			cc_clear_sw_subn_ports_data(p_switch);
			continue;
		}

		if (p_switch->cc_data.is_vl_conf_for_all_ports)
			status = cc_send_port_profile_settings(&p_osm->sm, p_node, p_physp,
							       cl_hton32((uint32_t)0xff));
		else
			status = cc_send_to_ext_ports(&p_osm->sm, p_node,
						      cc_send_port_profile_settings);

		if (status != IB_SUCCESS)
			ret = -1;

		if (p_switch->cc_data.is_sl_conf_for_all_ports)
			status = cc_send_sl_mapping_settings(&p_osm->sm, p_node, p_physp,
							     cl_hton32((uint32_t)0xff));
		else
			status = cc_send_to_ext_ports(&p_osm->sm, p_node,
						      cc_send_sl_mapping_settings);

		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_switches(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.sw_guid_tbl)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_next;
		ib_api_status_t status;

		p_next = cl_qmap_next(p_next);

		if (!ib_cc_get_enh_info_ver0(&p_switch->p_node->cc_enh_info))
			continue;

		status = cc_send_sw_general_settings(&p_osm->sm, p_switch);
		if (status != IB_SUCCESS)
			ret = -1;
	}

	return ret;
}

static ib_api_status_t configure_hcas(struct osm_opensm *p_osm)
{
	int ret = 0;
	cl_map_item_t *p_next;
	ib_cc_hca_general_settings_t *p_cc_hca_gs;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *) p_next;
		osm_node_t *p_node = p_port->p_node;
		ib_api_status_t status;

		p_next = cl_qmap_next(p_next);

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		if (!ib_cc_get_enh_info_ver0(&p_node->cc_enh_info))
			continue;

		status = cc_send_hca_general_settings(&p_osm->sm, p_node, p_port);
		if (status != IB_SUCCESS)
			ret = -1;

		p_cc_hca_gs = &p_port->cc_data.calc_mads.hca_general_settings;

		/* Skip RP configuration if device not configured for reaction */
		if (ib_cc_hca_get_en_react(p_cc_hca_gs)) {
			status = cc_send_hca_rp_params(&p_osm->sm, p_node, p_port);
			if (status != IB_SUCCESS)
				ret = -1;
		} else {
			/* Mark RP for future update */
			cc_clear_ca_subn_rp_params(p_port);
		}

		/* Skip NP configuration if device not configured for notifications */
		if (ib_cc_hca_get_en_notify(p_cc_hca_gs)) {
			status = cc_send_hca_np_params(&p_osm->sm, p_node, p_port);
			if (status != IB_SUCCESS)
				ret = -1;
		} else {
			/* Mark NP for future update */
			cc_clear_ca_subn_np_params(p_port);
		}

	}

	return ret;
}

static void calculate_vl_profiles_mad_data(struct osm_opensm *p_osm,
					   cl_fmap_t *cc_profiles_map,
					   osm_cc_vl_conf_t *p_vl_conf)
{
	cl_map_item_t *p_item;
	osm_cc_profiles_conf_t *p_profiles_conf;
	ib_cc_port_profile_settings_t vl_to_profiles[IB_MAX_NUM_VLS];
	osm_physp_t *p_physp;
	int vl, profile;
	uint8_t num_ports, port_num;
	boolean_t all_switch_ports = FALSE;

	memset(vl_to_profiles, 0, sizeof(ib_cc_port_profile_settings_t) * IB_MAX_NUM_VLS);

	if (p_vl_conf->group_type == OSM_CC_GROUP_SWITCH_PORTS)
		all_switch_ports = TRUE;

	for (vl = 0; vl < IB_MAX_NUM_VLS; vl++) {
		p_profiles_conf = p_vl_conf->vl_to_profiles[vl];

		ib_cc_port_profile_set_mode(&vl_to_profiles[vl], p_profiles_conf->mode);
		for (profile = 0; profile < OSM_CC_MAX_PROFILE_NUM; profile++) {
			ib_cc_port_profiles_set_profile(&vl_to_profiles[vl], profile,
							p_profiles_conf->profiles[profile].min,
							p_profiles_conf->profiles[profile].max,
							p_profiles_conf->profiles[profile].percent);
		}
	}

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;

		num_ports = osm_node_get_num_physp(p_node);
		p_switch->cc_data.is_vl_conf_for_all_ports = all_switch_ports;

		/* If the VL configuration is the same for all the ports save it on port 0 */
		port_num = p_switch->cc_data.is_vl_conf_for_all_ports ? 0 : 1;
		for ( ; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			memcpy(p_physp->cc_data.calc_mads.vl_to_profiles,
			       vl_to_profiles, sizeof(ib_cc_port_profile_settings_t) * IB_MAX_NUM_VLS);
		}
	}
}

static void calculate_sl_profiles_mad_data(struct osm_opensm *p_osm,
					   osm_cc_sl_conf_t *p_sl_conf)
{
	cl_map_item_t *p_item;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, sl_num;
	ib_cc_sl_mapping_settings_t cc_sl_mapping;
	boolean_t all_switch_ports = FALSE;

	memset(&cc_sl_mapping, 0, sizeof(ib_cc_sl_mapping_settings_t));

	if (p_sl_conf->group_type == OSM_CC_GROUP_SWITCH_PORTS)
		all_switch_ports = TRUE;

	for (sl_num = 0; sl_num < IB_NUMBER_OF_SLS; sl_num++) {
		ib_cc_sl_set_profile_num(&cc_sl_mapping, sl_num,
					 p_sl_conf->sl2profile_num[sl_num]);
	}

	for (p_item = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_item;
		osm_node_t *p_node = p_switch->p_node;

		num_ports = osm_node_get_num_physp(p_node);
		p_switch->cc_data.is_sl_conf_for_all_ports = all_switch_ports;

		/* If the SL configuration is the same for all the ports save it on port 0 */
		port_num = p_switch->cc_data.is_sl_conf_for_all_ports ? 0 : 1;
		for ( ; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			memcpy(&p_physp->cc_data.calc_mads.sl_mapping,
			       &cc_sl_mapping, sizeof(cc_sl_mapping));
		}
	}
}

static void calculate_sw_mad_data(struct osm_opensm *p_osm,
				  osm_cc_sw_conf_t *p_sw_conf)
{
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.sw_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.sw_guid_tbl)) {
		osm_switch_t *p_switch = (osm_switch_t *) p_next;

		if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_DISABLED) {
			ib_cc_sw_set_enable_cc(&p_switch->cc_data.calc_sw_general_settings, 0);
			continue;
		}

		ib_cc_sw_set_enable_cc(&p_switch->cc_data.calc_sw_general_settings,
				       p_sw_conf->enable);

		p_switch->cc_data.calc_sw_general_settings.aqs_weight = p_sw_conf->aqs_weight;
		p_switch->cc_data.calc_sw_general_settings.aqs_time = p_sw_conf->aqs_time;

		p_next = cl_qmap_next(p_next);
	}
}


static void calculate_ca_mad_data(struct osm_opensm *p_osm,
				  osm_cc_ca_conf_t *p_ca_conf)
{
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *) p_next;
		osm_node_t *p_node = p_port->p_node;

		p_next = cl_qmap_next(p_next);

		if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA)
			continue;

		if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_DISABLED) {
			ib_cc_hca_set_en_notify(&p_port->cc_data.calc_mads.hca_general_settings, 0);
			ib_cc_hca_set_en_react(&p_port->cc_data.calc_mads.hca_general_settings, 0);
			continue;
		}

		//TODO: check if port in group
		ib_cc_hca_set_en_notify(&p_port->cc_data.calc_mads.hca_general_settings,
					p_ca_conf->enable_notify);
		ib_cc_hca_set_en_react(&p_port->cc_data.calc_mads.hca_general_settings,
				       p_ca_conf->enable_react);

		ib_cc_hca_rp_params_set_tgt_rate(&p_port->cc_data.calc_mads.hca_rp_params,
					      p_ca_conf->hca_rp_params.enable_clamp_tgt_rate);
		ib_cc_hca_rp_params_set_tgt_rate_at(&p_port->cc_data.calc_mads.hca_rp_params,
						 p_ca_conf->hca_rp_params.enable_clamp_tgt_rate_after_time_inc);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_time_reset = cl_hton32(p_ca_conf->hca_rp_params.rpg_time_reset);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_byte_reset = cl_hton32(p_ca_conf->hca_rp_params.rpg_byte_reset);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_threshold = p_ca_conf->hca_rp_params.rpg_threshold;
		p_port->cc_data.calc_mads.hca_rp_params.rpg_max_rate = cl_hton32(p_ca_conf->hca_rp_params.rpg_max_rate);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_ai_rate = cl_hton16(p_ca_conf->hca_rp_params.rpg_ai_rate);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_hai_rate = cl_hton16(p_ca_conf->hca_rp_params.rpg_hai_rate);
		p_port->cc_data.calc_mads.hca_rp_params.rpg_gd = p_ca_conf->hca_rp_params.rpg_gd;
		p_port->cc_data.calc_mads.hca_rp_params.rpg_min_dec_fac = p_ca_conf->hca_rp_params.rpg_min_dec_fac;
		p_port->cc_data.calc_mads.hca_rp_params.rpg_min_rate = cl_hton32(p_ca_conf->hca_rp_params.rpg_min_rate);
		p_port->cc_data.calc_mads.hca_rp_params.rate_to_set_on_first_cnp = cl_hton32(p_ca_conf->hca_rp_params.rate_to_set_on_first_cnp);
		p_port->cc_data.calc_mads.hca_rp_params.dce_tcp_g = cl_hton16(p_ca_conf->hca_rp_params.dce_tcp_g);
		p_port->cc_data.calc_mads.hca_rp_params.dce_tcp_rtt = cl_hton32(p_ca_conf->hca_rp_params.dce_tcp_rtt);
		p_port->cc_data.calc_mads.hca_rp_params.rate_reduce_monitor_period = cl_hton32(p_ca_conf->hca_rp_params.rate_reduce_monitor_period);
		p_port->cc_data.calc_mads.hca_rp_params.initial_alpha_value = cl_hton16(p_ca_conf->hca_rp_params.initial_alpha_value);

		p_port->cc_data.calc_mads.hca_np_params.min_time_between_cnps = cl_hton16(p_ca_conf->hca_np_params.min_time_between_cnps);
		p_port->cc_data.calc_mads.hca_np_params.cnp_sl_mode = p_ca_conf->hca_np_params.cnp_sl_mode;
		p_port->cc_data.calc_mads.hca_np_params.cnp_sl = p_ca_conf->hca_np_params.cnp_sl;
	}
}


static void calculate_mad_data(struct osm_opensm *p_osm, osm_cc_mgr_t *p_mgr)
{
	cl_list_item_t *p_item;

	p_item = cl_qlist_head(&p_mgr->ca_conf);
	while (p_item != cl_qlist_end(&p_mgr->ca_conf)) {
		calculate_ca_mad_data(p_osm, (osm_cc_ca_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->sw_conf);
	while (p_item != cl_qlist_end(&p_mgr->sw_conf)) {
		calculate_sw_mad_data(p_osm, (osm_cc_sw_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->vl_profiles_conf);
	while (p_item != cl_qlist_end(&p_mgr->vl_profiles_conf)) {
		calculate_vl_profiles_mad_data(p_osm, &p_mgr->cc_profiles_map,
					       (osm_cc_vl_conf_t *)p_item);
		p_item = cl_qlist_next(p_item);
	}

	p_item = cl_qlist_head(&p_mgr->sl_profiles_conf);
	while (p_item != cl_qlist_end(&p_mgr->sl_profiles_conf)) {

		calculate_sl_profiles_mad_data(p_osm, (osm_cc_sl_conf_t *)p_item);

		p_item = cl_qlist_next(p_item);
	}
}


static int get_enh_cc_info(struct osm_opensm *p_osm)
{
	ib_api_status_t status;
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		osm_port_t *p_port = (osm_port_t *) p_next;
		osm_node_t *p_node = p_port->p_node;
		p_next = cl_qmap_next(p_next);

		if (!(p_node->general_info.payload.cap_mask[1] & IB_GENERAL_INFO_IS_CC_SUPPORTED_MASK)) {
			/* Mark that device does not support MLNX congestion control */
			memset(&p_node->cc_enh_info, 0, sizeof(ib_cc_enh_info_t));
			continue;
		}

		status = cc_send_cc_enh_info(&p_osm->sm, p_port);
		if (status != IB_SUCCESS)
			return -1;
	}

	return IB_SUCCESS;
}


static void cc_unmark_devices(struct osm_opensm *p_osm)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_switch_t *p_switch;
	unsigned int cc_sw_counter = 0;
	unsigned int cc_ca_counter = 0;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;

		if (p_node->sw) {
			cc_diable_sw_need_update(p_node->sw, OSM_CC_RESEND_ALL);
			p_switch = p_node->sw;
			if (p_switch->cc_data.subn_sw_general_settings.enable)
				cc_sw_counter++;
			continue;
		}

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
			cc_diable_ca_need_update(p_port, OSM_CC_RESEND_ALL);
			if (p_port->cc_data.calc_mads.hca_general_settings.enable_mask & 0xC0)
				cc_ca_counter++;
			continue;
		}
	}

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"congestion control configured on %u switches %u CAs\n",
		cc_sw_counter, cc_ca_counter);
}

static int cc_configure_subnet(struct osm_opensm *p_osm)
{
	int ret = 0;

	cl_plock_acquire(&p_osm->lock);

	if (get_enh_cc_info(p_osm)) {
		ret = -1;
		goto Exit;
	}

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	calculate_mad_data(p_osm, &p_osm->sm.cc_mgr);

	/*
	 * Configuration order on switches should be:
	 * 1. CongestionSwitchGeneralSetting
	 * 2. CongestionPortProfileSetting
	 * 3. CongestionSLMappingSetting?
	 */

	if (configure_switches(p_osm))
		ret = -1;

	cl_plock_release(&p_osm->lock);
	osm_congestion_control_wait_pending_transactions(p_osm);
	cl_plock_acquire(&p_osm->lock);

	if (configure_profiles(p_osm))
		ret = -1;

	if (configure_hcas(p_osm))
		ret = -1;;

Exit:
	cl_plock_release(&p_osm->lock);

	osm_congestion_control_wait_pending_transactions(p_osm);

	return ret;
}

int osm_congestion_control_setup(struct osm_opensm *p_osm)
{
	osm_cc_mgr_t *p_mgr = &p_osm->sm.cc_mgr;
	int ret = 0;

	OSM_LOG_ENTER(&p_osm->log);

	osm_cc_mgr_destroy(p_mgr);

	cc_mark_need_update(p_osm);

	/*
	 * Do nothing unless the most recent routing attempt was successful.
	 */
	if (!p_osm->subn.routing_engine_used)
		goto exit;

	if (p_osm->subn.opt.mlnx_congestion_control == OSM_CAPABILITY_ENABLED) {
		if (parse_congestion_control_policy(p_osm)) {
			ret = 1;
			goto exit;
		}

		if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_DEBUG)) {
			print_profiles_map(p_osm, &p_mgr->cc_profiles_map);
			print_sl_conf(p_osm, &p_mgr->sl_profiles_conf);
			print_vl_conf(p_osm, &p_mgr->vl_profiles_conf);
		}
	}


	ret = cc_configure_subnet(p_osm);

	cc_unmark_devices(p_osm);
exit:
	if (ret)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR C100: Errors during congestion control configuration\n");

	OSM_LOG_EXIT(&p_osm->log);

	return ret;
}

int osm_congestion_control_wait_pending_transactions(struct osm_opensm *p_osm)
{
	osm_congestion_control_t *cc = &p_osm->sm.p_subn->p_osm->cc;


	while (1) {
		unsigned count = cc->outstanding_mads;
		if (!count || osm_exit_flag)
			break;
		cl_event_wait_on(&cc->outstanding_mads_done_event,
				 EVENT_NO_TIMEOUT,
				 TRUE);
	}

	return osm_exit_flag;
}

static inline void decrement_outstanding_mads(osm_congestion_control_t *p_cc)
{
	uint32_t outstanding;

	outstanding = cl_atomic_dec(&p_cc->outstanding_mads);
	if (!outstanding)
		cl_event_signal(&p_cc->outstanding_mads_done_event);

	cl_atomic_dec(&p_cc->outstanding_mads_on_wire);
	cl_event_signal(&p_cc->sig_mads_on_wire_continue);
}

const char *cc_get_cc_attr_str(IN ib_net16_t attr)
{
	switch (attr) {
	case IB_MAD_ATTR_CC_ENH_INFO:
		return "EnhancedCongestionInfo";
	case IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS:
		return "CongestionSwitchGeneralSettings";
	case IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS:
		return "CongestionPortProfileSettings";
	case IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS:
		return "CongestionSLMappingSettings";
	case IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS:
		return "CongestionHCAGeneralSettings";
	case IB_MAD_ATTR_CC_HCA_RP_PARAMETERS:
		return "CongestionHCARPParameters";
	case IB_MAD_ATTR_CC_HCA_NP_PARAMETERS:
		return "CongestionHCANPParameters";
	}

	return "UNKNOWN";
}

static void dump_cc_mad(ib_cc_mad_t *p_cc_mad, uint32_t buf_size, osm_log_t *p_log)
{
		char buf[BUF_SIZE];
		buf[0] = '\0';
		unsigned n;
		int i;
		if (!p_cc_mad)
			return;

		n = sprintf(buf,
			    "CC Mad dump:\n"
			    "\t\t\t\tbase_ver................0x%X\n"
			    "\t\t\t\tmgmt_class..............0x%X\n"
			    "\t\t\t\tclass_ver...............0x%X\n"
			    "\t\t\t\tmethod..................0x%X ()\n",
			    p_cc_mad->header.base_ver, p_cc_mad->header.mgmt_class,
			    p_cc_mad->header.class_ver, p_cc_mad->header.method);
			    //ib_get_sm_method_str(p_cc_mad->header.method));

		n += snprintf(buf + n, buf_size - n,
			      "\t\t\t\tstatus..................0x%X\n",
			      cl_ntoh16(p_cc_mad->header.status));

		n += snprintf(buf + n, buf_size - n,
			      "\t\t\t\ttrans_id................0x%" PRIx64 "\n"
			      "\t\t\t\tattr_id.................0x%X (%s)\n"
			      "\t\t\t\tresv....................0x%X\n"
			      "\t\t\t\tattr_mod................0x%X\n"
			      "\t\t\t\tcc_key...................0x%016" PRIx64
			      "\n", cl_ntoh64(p_cc_mad->header.trans_id),
			      cl_ntoh16(p_cc_mad->header.attr_id),
			      cc_get_cc_attr_str(p_cc_mad->header.attr_id),
			      cl_ntoh16(p_cc_mad->header.resv),
			      cl_ntoh32(p_cc_mad->header.attr_mod),
			      cl_ntoh64(p_cc_mad->cc_key));

		for (i = 0; i < 64; i += 16) {
			n += snprintf(buf + n, buf_size - n,
				      "\n\t\t\t\t%02X %02X %02X %02X "
				      "%02X %02X %02X %02X"
				      "   %02X %02X %02X %02X %02X %02X %02X %02X\n",
				      p_cc_mad->mgt_data[i],
				      p_cc_mad->mgt_data[i + 1],
				      p_cc_mad->mgt_data[i + 2],
				      p_cc_mad->mgt_data[i + 3],
				      p_cc_mad->mgt_data[i + 4],
				      p_cc_mad->mgt_data[i + 5],
				      p_cc_mad->mgt_data[i + 6],
				      p_cc_mad->mgt_data[i + 7],
				      p_cc_mad->mgt_data[i + 8],
				      p_cc_mad->mgt_data[i + 9],
				      p_cc_mad->mgt_data[i + 10],
				      p_cc_mad->mgt_data[i + 11],
				      p_cc_mad->mgt_data[i + 12],
				      p_cc_mad->mgt_data[i + 13],
				      p_cc_mad->mgt_data[i + 14],
				      p_cc_mad->mgt_data[i + 15]);
		}

	osm_log(p_log, OSM_LOG_DEBUG, "%s\n", buf);
}

static void cc_rcv_mad_port_profile_settings(osm_node_t *p_node, uint8_t port_num,
					     ib_cc_port_profile_settings_t *p_cc_port_profiles)
{
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint16_t vl_mask;
	uint8_t port, port_end, vl;

	OSM_ASSERT(p_node && p_node->sw);
	p_sw = p_node->sw;
	OSM_ASSERT(port_num == 0xff || (port_num && port_num < p_sw->num_ports));

	vl_mask = cl_ntoh16(p_cc_port_profiles->vl_mask);

	/*
	 * Store MAD without VL mask in order to all comparing profiles
	 * configuration of different VLs.
	 */
	p_cc_port_profiles->vl_mask = 0;

	if (port_num == 0xff) {
		port = 0;
		port_end = p_sw->num_ports;
	} else {
		port = port_num;
		port_end = port_num + 1;
	}

	for (; port < port_end; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		for (vl = 0; vl < IB_MAX_NUM_VLS; vl++) {
			if ((vl_mask & (1 << vl)) == 0)
				continue;
			memcpy(&p_physp->cc_data.subn_mads.vl_to_profiles[vl],
			       p_cc_port_profiles,
			       sizeof(ib_cc_port_profile_settings_t));
		}
	}
}

static void cc_rcv_mad_sl_settings(osm_node_t *p_node, uint8_t port_num,
				   ib_cc_sl_mapping_settings_t *p_cc_sl_setting)
{
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint8_t port, port_end, sl;

	OSM_ASSERT(p_node && p_node->sw);
	p_sw = p_node->sw;
	OSM_ASSERT(port_num == 0xff || (port_num && port_num < p_sw->num_ports));

	if (port_num == 0xff) {
		port = 0;
		port_end = p_sw->num_ports;
	} else {
		port = port_num;
		port_end = port_num + 1;
	}

	for (; port < port_end; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		for (sl = 0; sl < IB_NUMBER_OF_SLS; sl++) {
			memcpy(&p_physp->cc_data.subn_mads.sl_mapping,
			       p_cc_sl_setting,
			       sizeof(ib_cc_sl_mapping_settings_t));
		}
	}

	/* HACK - we set the mask once per switch, instead of per port */
	cc_diable_sw_need_update(p_sw, OSM_CC_RESEND_SW_SL_MAPPING);
}

static void cc_rcv_mad(void *context, void *data)
{
	osm_congestion_control_t *p_cc = context;
	osm_opensm_t *p_osm = p_cc->osm;
	osm_madw_t *p_madw = data;
	ib_cc_mad_t *p_cc_mad;
	osm_madw_context_t *p_mad_context = &p_madw->context;
	ib_mad_t *p_mad = osm_madw_get_mad_ptr(p_madw);
	ib_net64_t node_guid = p_mad_context->cc_context.node_guid;
	ib_net64_t port_guid = p_mad_context->cc_context.port_guid;
	uint8_t port = p_mad_context->cc_context.port;
	osm_port_t *p_port;

	OSM_LOG_ENTER(p_cc->log);

	OSM_LOG(p_cc->log, OSM_LOG_VERBOSE,
		"Processing received MAD status 0x%x for "
		"attr ID 0x%x mod 0x%x node 0x%" PRIx64 " port %u\n",
		cl_ntoh16(p_mad->status), cl_ntoh16(p_mad->attr_id),
		cl_ntoh32(p_mad_context->cc_context.attr_mod),
		cl_ntoh64(node_guid), port);

	p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

	cl_plock_acquire(&p_osm->lock);

	p_port = osm_get_port_by_guid(p_cc->subn, port_guid);
	if (!p_port) {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C109: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	p_port->cc_timeout_count = 0;

	if (p_cc_mad->header.status) {
		if (p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_CLASS_VER
		    || p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD
		    || p_cc_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD_ATTR)
			p_port->cc_unavailable_flag = TRUE;
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C10A:"
			" Received invalid status from port GUID 0x%" PRIx64 ","
			" status:0x%x, attr_id:0x%x\n",
			cl_ntoh64(port_guid),
			cl_ntoh16(p_cc_mad->header.status),
			cl_ntoh16(p_cc_mad->header.attr_id));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	} else {
		p_port->cc_unavailable_flag = FALSE;
	}

	if (OSM_LOG_IS_ACTIVE_V2(p_cc->log, OSM_LOG_FRAMES))
		dump_cc_mad(p_cc_mad, BUF_SIZE, p_cc->log);

	if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_ENH_INFO) {
		ib_cc_enh_info_t *p_cc_enh_info = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->p_node->cc_enh_info, p_cc_enh_info, sizeof(ib_cc_enh_info_t));
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_GENERAL_SETTINGS) {
		ib_cc_hca_general_settings_t *p_ca_general_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->cc_data.subn_mads.hca_general_settings, p_ca_general_setting,
		       sizeof(ib_cc_hca_general_settings_t));
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_PORT_SW_PROFILE_SETTINGS) {
		ib_cc_port_profile_settings_t *p_cc_port_profiles = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		ib_net32_t attr_mod = p_mad_context->cc_context.attr_mod;
		uint8_t port_num = cl_ntoh32(attr_mod) & 0xff;
		cc_rcv_mad_port_profile_settings(p_port->p_node, port_num, p_cc_port_profiles);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_SL_MAPPING_SETTINGS) {
		ib_cc_sl_mapping_settings_t *p_cc_sl_setting = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		ib_net32_t attr_mod = p_mad_context->cc_context.attr_mod;
		uint8_t port_num = cl_ntoh32(attr_mod) & 0xff;
		cc_rcv_mad_sl_settings(p_port->p_node, port_num, p_cc_sl_setting);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_SW_GENERAL_SETTINGS) {
		ib_cc_sw_general_settings_t *p_sw_general_settings = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->p_node->sw->cc_data.subn_sw_general_settings, p_sw_general_settings,
		       sizeof(ib_cc_sw_general_settings_t));
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_RP_PARAMETERS) {
		ib_cc_hca_rp_parameters_t *p_hca_rp_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->cc_data.subn_mads.hca_rp_params, p_hca_rp_params,
		       sizeof(p_port->cc_data.subn_mads.hca_rp_params));
		cc_diable_ca_need_update(p_port, OSM_CC_RESEND_CA_RP_PARAMETERS);
	} else if (p_cc_mad->header.attr_id == IB_MAD_ATTR_CC_HCA_NP_PARAMETERS) {
		ib_cc_hca_rp_parameters_t *p_hca_np_params = ib_cc_mad_get_mgt_data_ptr(p_cc_mad);
		memcpy(&p_port->cc_data.subn_mads.hca_np_params, p_hca_np_params,
		       sizeof(p_port->cc_data.subn_mads.hca_np_params));
		cc_diable_ca_need_update(p_port, OSM_CC_RESEND_CA_NP_PARAMETERS);
	} else {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C10A: "
			"Unexpected MAD attribute ID %u received\n",
			cl_ntoh16(p_cc_mad->header.attr_id));
	}

	cl_plock_release(&p_osm->lock);

Exit:
	decrement_outstanding_mads(p_cc);
	osm_mad_pool_put(p_cc->mad_pool, p_madw);
	OSM_LOG_EXIT(p_cc->log);
}

static void cc_poller_send(osm_congestion_control_t *p_cc,
			   osm_madw_t *p_madw)
{
	osm_subn_opt_t *p_opt = &p_cc->subn->opt;
	cl_status_t sts;
	ib_api_status_t status;
	osm_madw_context_t mad_context = p_madw->context;

	status = osm_vendor_send(p_cc->bind_handle, p_madw, TRUE);
	if (status == IB_SUCCESS) {
		cl_atomic_inc(&p_cc->outstanding_mads_on_wire);
		while (p_cc->outstanding_mads_on_wire >
		       (int32_t)p_opt->cc_max_outstanding_mads) {
wait:
			sts = cl_event_wait_on(&p_cc->sig_mads_on_wire_continue,
					       EVENT_NO_TIMEOUT, TRUE);
			if (sts != CL_SUCCESS)
				goto wait;
		}
	} else
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C104: "
			"send failed to node 0x%" PRIx64 "port %u\n",
			cl_ntoh64(mad_context.cc_context.node_guid),
			mad_context.cc_context.port);
}



static void cc_poller(void *p_ptr)
{
	osm_congestion_control_t *p_cc = p_ptr;
	osm_madw_t *p_madw;
	ib_cc_mad_t *p_cc_mad;

	OSM_LOG_ENTER(p_cc->log);

	if (p_cc->thread_state == OSM_THREAD_STATE_NONE)
		p_cc->thread_state = OSM_THREAD_STATE_RUN;

	while (p_cc->thread_state == OSM_THREAD_STATE_RUN) {
		cl_spinlock_acquire(&p_cc->mad_queue_lock);

		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_cc->mad_queue);
		p_cc_mad = osm_madw_get_cc_mad_ptr(p_madw);

		cl_spinlock_release(&p_cc->mad_queue_lock);
		if (p_madw != (osm_madw_t *) cl_qlist_end(&p_cc->mad_queue)) {
			if (OSM_LOG_IS_ACTIVE_V2(p_cc->log, OSM_LOG_FRAMES))
				dump_cc_mad(p_cc_mad, BUF_SIZE, p_cc->log);
			cc_poller_send(p_cc, p_madw);
		} else {
			cl_event_wait_on(&p_cc->cc_poller_wakeup,
					 EVENT_NO_TIMEOUT, TRUE);
		}
	}

	OSM_LOG_EXIT(p_cc->log);
}

ib_api_status_t osm_congestion_control_init(osm_congestion_control_t * p_cc,
					    struct osm_opensm *p_osm,
					    const osm_subn_opt_t * p_opt)
{
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	memset(p_cc, 0, sizeof(*p_cc));

	p_cc->osm = p_osm;
	p_cc->subn = &p_osm->subn;
	p_cc->sm = &p_osm->sm;
	p_cc->log = &p_osm->log;
	p_cc->mad_pool = &p_osm->mad_pool;
	p_cc->trans_id = CONGESTION_CONTROL_INITIAL_TID_VALUE;
	p_cc->vendor = p_osm->p_gmp_vendor;

	p_cc->cc_disp_h = cl_disp_register(&p_osm->gmp_disp, OSM_MSG_MAD_CC,
					   cc_rcv_mad, p_cc);
	if (p_cc->cc_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	cl_qlist_init(&p_cc->mad_queue);

	status = cl_spinlock_init(&p_cc->mad_queue_lock);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->cc_poller_wakeup);
	status = cl_event_init(&p_cc->cc_poller_wakeup, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->outstanding_mads_done_event);
	status = cl_event_init(&p_cc->outstanding_mads_done_event, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_cc->sig_mads_on_wire_continue);
	status = cl_event_init(&p_cc->sig_mads_on_wire_continue, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	p_cc->thread_state = OSM_THREAD_STATE_NONE;

	status = cl_thread_init(&p_cc->cc_poller, cc_poller, p_cc,
				"cc poller");
	if (status != IB_SUCCESS)
		goto Exit;

	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(p_cc->log);
	return status;
}

static void cc_mad_recv_callback(osm_madw_t * p_madw, void *bind_context,
				 osm_madw_t * p_req_madw)
{
	osm_congestion_control_t *p_cc = bind_context;

	OSM_LOG_ENTER(p_cc->log);

	OSM_ASSERT(p_madw);

	/* HACK - should be extended to support CC traps */
	OSM_ASSERT(p_req_madw != NULL);

	osm_madw_copy_context(p_madw, p_req_madw);
	osm_mad_pool_put(p_cc->mad_pool, p_req_madw);

	/* Do not decrement outstanding mads here, do it in the dispatcher */

	if (cl_disp_post(p_cc->cc_disp_h, OSM_MSG_MAD_CC,
			 p_madw, NULL, NULL) != CL_SUCCESS) {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C105: "
			"Congestion Control Dispatcher post failed\n");
		osm_mad_pool_put(p_cc->mad_pool, p_madw);
	}

	OSM_LOG_EXIT(p_cc->log);
}

static void cc_mad_send_err_callback(void *bind_context,
				     osm_madw_t * p_madw)
{
	osm_congestion_control_t *p_cc = bind_context;
	osm_madw_context_t *p_madw_context = &p_madw->context;
	osm_opensm_t *p_osm = p_cc->osm;
	uint64_t node_guid = p_madw_context->cc_context.node_guid;
	uint64_t port_guid = p_madw_context->cc_context.port_guid;
	uint8_t port = p_madw_context->cc_context.port;
	osm_port_t *p_port;
	int log_flag = 1;

	OSM_LOG_ENTER(p_cc->log);

	cl_plock_acquire(&p_osm->lock);

	p_port = osm_get_port_by_guid(p_cc->subn, port_guid);
	if (!p_port) {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C10B: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	/* If timed out before, don't bothering logging again
	 * we assume no CC support
	 */
	if (p_madw->status == IB_TIMEOUT
	    && p_port->cc_timeout_count)
		log_flag = 0;

	if (log_flag)
		OSM_LOG(p_cc->log, OSM_LOG_ERROR, "ERR C106: MAD Error (%s): "
			"attr id = %u LID %u GUID 0x%016" PRIx64 " port %u "
			"TID 0x%" PRIx64 "\n",
			ib_get_err_str(p_madw->status),
			p_madw->p_mad->attr_id,
			cl_ntoh16(p_madw->mad_addr.dest_lid),
			cl_ntoh64(node_guid),
			port,
			cl_ntoh64(p_madw->p_mad->trans_id));

	if (p_madw->status == IB_TIMEOUT) {
		p_port->cc_timeout_count++;
		if (p_port->cc_timeout_count > OSM_CC_TIMEOUT_COUNT_THRESHOLD
		    && !p_port->cc_unavailable_flag) {
			p_port->cc_unavailable_flag = TRUE;
			p_port->cc_timeout_count = 0;
		}
	} else {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR,
			"ERR C10F: Subnet initialization error\n");
		p_cc->subn->subnet_initialization_error = TRUE;
	}

	cl_plock_release(&p_osm->lock);

Exit:
	osm_mad_pool_put(p_cc->mad_pool, p_madw);

	decrement_outstanding_mads(p_cc);

	OSM_LOG_EXIT(p_cc->log);
}

ib_api_status_t osm_congestion_control_bind(osm_congestion_control_t * p_cc,
					    ib_net64_t port_guid)
{
	osm_bind_info_t bind_info;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(p_cc->log);

	bind_info.port_guid = p_cc->port_guid = port_guid;
	bind_info.mad_class = IB_MCLASS_CC;
	bind_info.class_version = 2;
	bind_info.is_responder = FALSE;
	bind_info.is_report_processor = FALSE;
	bind_info.is_trap_processor = FALSE;
	bind_info.recv_q_size = OSM_SM_DEFAULT_QP1_RCV_SIZE;
	bind_info.send_q_size = OSM_SM_DEFAULT_QP1_SEND_SIZE;
	bind_info.timeout = p_cc->subn->opt.transaction_timeout;
	bind_info.retries = p_cc->subn->opt.transaction_retries;

	OSM_LOG(p_cc->log, OSM_LOG_VERBOSE,
		"Binding to port GUID 0x%" PRIx64 "\n", cl_ntoh64(port_guid));

	p_cc->bind_handle = osm_vendor_bind(p_cc->vendor, &bind_info,
					    p_cc->mad_pool,
					    cc_mad_recv_callback,
					    cc_mad_send_err_callback, p_cc);

	if (p_cc->bind_handle == OSM_BIND_INVALID_HANDLE) {
		status = IB_ERROR;
		OSM_LOG(p_cc->log, OSM_LOG_ERROR,
			"ERR C107: Vendor specific bind failed (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_cc->log);
	return status;
}

void osm_congestion_control_shutdown(osm_congestion_control_t * p_cc)
{
	OSM_LOG_ENTER(p_cc->log);
	if (p_cc->bind_handle == OSM_BIND_INVALID_HANDLE) {
		OSM_LOG(p_cc->log, OSM_LOG_ERROR,
			"ERR C108: No previous bind\n");
		goto Exit;
	}
	cl_disp_unregister(p_cc->cc_disp_h);
Exit:
	OSM_LOG_EXIT(p_cc->log);
}

void osm_congestion_control_destroy(osm_congestion_control_t * p_cc)
{
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_cc->log);

	p_cc->thread_state = OSM_THREAD_STATE_EXIT;

	cl_event_signal(&p_cc->sig_mads_on_wire_continue);
	cl_event_signal(&p_cc->cc_poller_wakeup);

	cl_thread_destroy(&p_cc->cc_poller);

	cl_spinlock_acquire(&p_cc->mad_queue_lock);

	while (!cl_is_qlist_empty(&p_cc->mad_queue)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_cc->mad_queue);
		osm_mad_pool_put(p_cc->mad_pool, p_madw);
	}

	cl_spinlock_release(&p_cc->mad_queue_lock);

	cl_spinlock_destroy(&p_cc->mad_queue_lock);

	cl_event_destroy(&p_cc->cc_poller_wakeup);
	cl_event_destroy(&p_cc->outstanding_mads_done_event);
	cl_event_destroy(&p_cc->sig_mads_on_wire_continue);

	OSM_LOG_EXIT(p_cc->log);
}

void osm_congestion_control_invalidate(osm_congestion_control_t * p_cc)
{
	osm_opensm_t *p_osm = p_cc->osm;
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_switch_t *p_switch;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
			p_port->cc_data.resend_mask |= OSM_CC_RESEND_ALL;
			p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_NP_PARAMETERS;
			p_port->cc_data.resend_mask |= OSM_CC_RESEND_CA_RP_PARAMETERS;
			cc_clear_ca_subn_data(p_port);
			continue;
		}

		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			p_switch = p_node->sw;
			p_switch->cc_data.resend_mask |= OSM_CC_RESEND_ALL;
			p_switch->cc_data.resend_mask |= OSM_CC_RESEND_SW_SL_MAPPING;
			cc_clear_sw_subn_data(p_switch);
			continue;
		}
	}
}
