/*
 * Copyright (c) 2010 Gilad Ben-Yossef https://github.com/gby/libcrash.
 * Copyright (c) 2012 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdlib.h>
#include <errno.h>
#include <execinfo.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/wait.h>
#include <opensm/osm_crashd.h>
#include <complib/cl_spinlock.h>

static cl_spinlock_t crash_lock;

/* Global static crash message buffer */
static osm_crash_msg_t crash_msg;

/* Pipe file descriptor to crashd */
static int write_pipe_fd = -1;

/* crash daemon PID */
static int crashd_pid = 0;

/********************************************************************
 ********************************************************************/

/*
 * The fault handler function. The rules are as follows:
 * 1. Can't use any function that relies on dynamic memory allocation,
 *    as the malloc arena may be corrupt.
 * 2. Can only use a the POSIX.1-2003 list of async-safe functions.
 * 3. Some of the functions on the list are not always safe
 *    (like fork when atfork() is used), so need to avoid these also.
 * 4. No locking allowed. We don't know in what state the process/thread
 *    was when the exception occured.
 *
 * A nice list of the functions that can be safely called from the
 * signal processing function:
 *   http://pubs.opengroup.org/onlinepubs/009695399/functions/xsh_chap02_04.html
 */
void fault_handler(int signal, siginfo_t * siginfo, void *context)
{
	int res;

	/* Handling signal. We'd better not call any logging function here. */

	res = cl_spinlock_try(&crash_lock);
	if (res != CL_SUCCESS) {
		/*
		 * Block indefinitely and wait for nothing.
		 * Effectively, yield and give control to other thread.
		 */
		sigset_t smask;
		sigemptyset(&smask);
		pselect(0, NULL, NULL, NULL, NULL, &smask);
	}

	crash_msg.num_backtrace_frames =
		backtrace(crash_msg.backtrace, CRASHD_MAX_BACKTRACE_DEPTH);

	/* Grab the kernel thread ID. Because signal handlers are shared
	 * between all threads of the same process, this can only be done
	 * at fault time.
	 * Note that glibc doesn't expose any interface to get thread ID
	 * directly, so need to use syscall().
	 */
	crash_msg.thread_id = syscall(SYS_gettid);
	crash_msg.signal_number = signal;
	crash_msg.signal_code = siginfo->si_code;
	crash_msg.fault_address = siginfo->si_addr;
	crash_msg.signal_errno = siginfo->si_errno;
	crash_msg.handler_errno = errno;

_write:

	res = write(write_pipe_fd, &crash_msg, sizeof(crash_msg));

	/* If we got interrupted by a signal, retry the write.
	 * This shouldn't happen since we mask all signals
	 * during the handler run via sigaction sa_mask field,
	 * but who knows...
	 *
	 * It's useless to test for any other condition since we
	 * can't do anything if we fail
	 */
	if (res && errno == EINTR)
		goto _write;

	/*
	 * Use backtrace_symbols_fd rather then backtrace_symbols
	 * since the latter uses malloc.
	 */
	backtrace_symbols_fd(crash_msg.backtrace,
			     crash_msg.num_backtrace_frames,
			     write_pipe_fd);

	close(write_pipe_fd);

	waitpid(crashd_pid, NULL, 0);

	/* Produce a core dump for post morteum debugging */
	abort();

	/* not reached */
	CL_ASSERT(0);
}

/********************************************************************
 ********************************************************************/

/*
 * Set FD_CLOEXEC flag of file descriptor 'fd',
 * Return 0 on success, or -1 on error with errno set.
 */
static int set_cloexec(int fd)
{
	int flags = fcntl(fd, F_GETFD);

	if (flags < 0)
		return flags;

	flags |= FD_CLOEXEC;
	return fcntl(fd, F_SETFD, flags);
}

/********************************************************************
 ********************************************************************/

/*
 * Registration function.
 * Needs to be called once by each process (not thread).
 */
int register_crash_handler(int argc, char *argv[], const char *log)
{
	struct sigaction act;	/* Signal handler register struct */
	int res;
	int pfd[2];		/* Pipe file descriptor array */

	/* See ahead about these two */
	void *dummy_trace_array[1];
	unsigned int dummy_trace_size;

	CL_ASSERT(sizeof(crash_msg) <= CRASHD_MAX_MSG_SIZE);

	if (cl_spinlock_init(&crash_lock) != CL_SUCCESS)
		return -1;
	/*
	 * Check that this is the first call. If called again
	 * (perhaps after a fork), the pipe would already be opened.
	 */
	if (write_pipe_fd == -1) {

		res = pipe(pfd);
		if (res == -1)
			return errno;

		/* use the write handle */
		write_pipe_fd = pfd[1];

		/* If the caller program execs, we want the pipe to close,
		 * because it's not likely a random program will have the
		 * right signal handler set to use the crash daemon. */
		if (set_cloexec(write_pipe_fd))
			return errno;

		/* Set our daemon up */
		res = fork();
		if (res) {
			/*
			 * Father - close the read handle and
			 * get back to usual business.
			 */
			close(pfd[0]);
			crashd_pid = res;
		} else {
			/*
			 * Son - close the writing end and launch main
			 * daemon function that does NOT return
			 */
			close(pfd[1]);
			crashd_main(argc, argv, log, pfd[0]);
			CL_ASSERT(0 && "should not happen");
		}

	}

	/*
	 * In theory, neither backtrace nor backtrace_symbold_fd call malloc and
	 * friends so we are able to use them in a an exception handler safely.
	 *
	 * In practice recent glibc versions put these function in a separate
	 * shared library called libgcc_s.so, that gets loaded on-demand,
	 * and the dynamic linker uses malloc to get some internal buffer.
	 *
	 * We therefore give these a dummy call here during registration, to
	 * assure that the library gets loaded where it's safe to malloc.
	 */

	dummy_trace_size = backtrace(dummy_trace_array, 1);
	backtrace_symbols_fd(dummy_trace_array, dummy_trace_size, -1);

	/* Fill in the data that is known before the crash */
	crash_msg.magic = CRASHD_MSG_MAGIC;
	crash_msg.process_id = getpid();

	/*
	 * Prepare a sigaction struct for exception handler registrations.
	 * Disable signals during handler run.
	 */

	memset(&act, 0, sizeof(act));
	act.sa_sigaction = fault_handler;
	sigfillset(&act.sa_mask);
	/*
	 * We want the 3 parameter form of the handler
	 * with the siginfo_t addtional data.
	 */
	act.sa_flags = SA_SIGINFO;

	/* Register the handler for all exception signals. */
	res  = sigaction(SIGSEGV, &act, NULL);
	res |= sigaction(SIGILL,  &act, NULL);
	res |= sigaction(SIGFPE,  &act, NULL);
	res |= sigaction(SIGBUS,  &act, NULL);
	res |= sigaction(SIGQUIT, &act, NULL);
	res |= sigaction(SIGSYS, &act, NULL);

	return res;
}
