/*
 * Copyright (c) 2010 Gilad Ben-Yossef https://github.com/gby/libcrash.
 * Copyright (c) 2012 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <complib/cl_types.h>
#include <opensm/osm_crashd.h>

static char osm_binary_name[CRASHD_MAX_FILENAME];
static char crash_msg_buf[CRASHD_MAX_MSG_SIZE];
static osm_crash_msg_t *crash_msg = (osm_crash_msg_t *)&crash_msg_buf;

static volatile char terminate_flag = 0;
static volatile char signal_flag = 0;
static volatile char timeout_flag = 0;

/********************************************************************
 ********************************************************************/

static inline char *code2str(int code, int signal)
{

	switch (code) {
		case SI_USER  : return "kill, sigsend or raise ";
		case SI_KERNEL: return "kernel";
		case SI_QUEUE : return "sigqueue";
	}

	if (SIGILL == signal) {
		switch (code) {
			case ILL_ILLOPC: return "illegal opcode";
			case ILL_ILLOPN: return "illegal operand";
			case ILL_ILLADR: return "illegal addressing mode";
			case ILL_ILLTRP: return "illegal trap";
			case ILL_PRVOPC: return "privileged register";
			case ILL_COPROC: return "coprocessor error";
			case ILL_BADSTK: return "internal stack error";
		}
	}
	else if (SIGFPE == signal) {
		switch (code) {
			case FPE_INTDIV: return "integer divide by zero";
			case FPE_INTOVF: return "integer overflow";
			case FPE_FLTDIV: return "floating point divide by zero";
			case FPE_FLTOVF: return "floating point overflow";
			case FPE_FLTUND: return "floating point underflow";
			case FPE_FLTRES: return "floating point inexact result";
			case FPE_FLTINV: return "floating point invalid operation";
			case FPE_FLTSUB: return "subscript out of range";
		}
	}
	else if (SIGSEGV == signal) {
		switch (code) {
			case SEGV_MAPERR: return "address not mapped to object";
			case SEGV_ACCERR: return "invalid permissions for mapped object";
		}
	}
	else if (SIGBUS == signal) {
		switch (code) {
			case BUS_ADRALN: return "invalid address alignment";
			case BUS_ADRERR: return "non-existent physical address";
			case BUS_OBJERR: return "object specific hardware error";
		}
	}
	else if (SIGTRAP == signal) {
		switch (code) {
			case TRAP_BRKPT: return "process breakpoint";
			case TRAP_TRACE: return "process trace trap";
		}
	}
	return "Unhandled signal handler";
}

/********************************************************************
 ********************************************************************/

static int run_command(const char * command, char * command_out)
{
	char *tmp;
	int n = 0;
	FILE * command_pipe;

	tmp = calloc(1, LOG_LARGE_BUF_SIZE_MAX);
	if (!tmp)
		goto EXIT;

	command_pipe = popen(command, "r");
	if (!command_pipe)
		goto FREE_AND_EXIT;

	command_out[0] = '\0';
	while (fgets(tmp, sizeof(tmp)-1, command_pipe))
		n += snprintf(command_out + n, LOG_LARGE_BUF_SIZE_MAX - n,
			      "%s", tmp);
	pclose(command_pipe);

FREE_AND_EXIT:
	free(tmp);
EXIT:
	return n;
}

/********************************************************************
 ********************************************************************/

#if defined(HAVE_ADDR2LINE)
static char * __strip_white(char * str)
{
	char *p;
	while (isspace(*str)) str++;
	if (!*str) return str;
	p = str + strlen(str) - 1;
	while (isspace(*p)) *p-- = '\0';
	return str;
}

static void format_addr2line(char *str, char **func, char **fileline)
{
	int i;
	str = __strip_white(str);
	*func = str;
	*fileline = NULL;
	for (i = 0; str[i]; i++) {
		if (str[i] == '\n') {
			str[i] = '\0';
			*fileline = &str[i+1];
			return;
		}
	}
}

static void crashd_resolve_address(FILE * stream, void * addr)
{
/*
 * Try to resolve the line number and function name in
 * the source file using "addr2line" Linux utility.
 */
	char command[CRASHD_MAX_FILENAME];

	char command_out[CRASHD_MAX_FILENAME];
	char * command_out_function;
	char * command_out_file_and_line;

	if (!osm_binary_name[0])
		return;

	snprintf(command, CRASHD_MAX_FILENAME-1,
		 "addr2line %p -i -C -f -e %s",
		 addr, osm_binary_name);

	if (!run_command(command, command_out))
		return;

	/* Print line number only if it has been resolved */
	format_addr2line(command_out,
			 &command_out_function, &command_out_file_and_line);

	/*
	 * Note that if line resolution failed, you'll see
	 * something like this in the backtrace: "??:0 - ??()"
	 */
	osm_log_to_stream(stream, " [CRASHD] =       %s - %s()\n",
			  command_out_file_and_line, command_out_function);
}

#else
#define crashd_resolve_address(var1,var2)
#endif

/********************************************************************
 ********************************************************************/

#if defined(HAVE_GSTACK)
static void crashd_dump_with_gstack(FILE * stream)
{
	char command[CRASHD_MAX_FILENAME];
	char *command_out;

	command_out = calloc(1, LOG_LARGE_BUF_SIZE_MAX);
	if (!command_out)
		return;

	if (!osm_binary_name[0])
		goto EXIT;

	snprintf(command, sizeof(command) - 1,
		 "gstack %d", crash_msg->process_id);

	/* print stack only if we got something */
	if (run_command(command, command_out))
		osm_log_to_stream(stream,
				  " [CRASHD] = backtrace obtained with gstack:\n"
		                  " ==== [CRASHD] gstack ==== \n"
				  "%s"
				  " ==== [CRASHD] gstack ==== \n\n",
				  command_out);
EXIT:
	free(command_out);
}
#else
#define crashd_dump_with_gstack(var)
#endif

/********************************************************************
 ********************************************************************/

extern const char *const sys_siglist[];

static void print_to_stream(FILE * stream)
{
	int i;

	osm_log_to_stream(stream, " [CRASHD]\n");
	osm_log_to_stream(stream, " [CRASHD] =================================\n");
	osm_log_to_stream(stream, " [CRASHD] =      OpenSM crash daemon      =\n");
	osm_log_to_stream(stream, " [CRASHD] =================================\n");
	osm_log_to_stream(stream, " [CRASHD]\n");

	osm_log_to_stream(stream, " [CRASHD] = OpenSM binary: %s\n",
			  osm_binary_name);
	osm_log_to_stream(stream, " [CRASHD] = OpenSM process ID: %d\n",
			  crash_msg->process_id);
	osm_log_to_stream(stream, " [CRASHD] = OpenSM thread ID: %d\n",
			  crash_msg->thread_id);
	osm_log_to_stream(stream, " [CRASHD] = Exception: %s\n",
			  sys_siglist[crash_msg->signal_number]);
	osm_log_to_stream(stream, " [CRASHD] = Reason: %s\n",
			  code2str(crash_msg->signal_code,
			  crash_msg->signal_number));
	osm_log_to_stream(stream, " [CRASHD] = Fault address: %p\n",
			  crash_msg->fault_address);
	osm_log_to_stream(stream, " [CRASHD] = Signal error: %s\n",
			  strerror(crash_msg->signal_errno));
	osm_log_to_stream(stream, " [CRASHD] = Last error: %s\n",
			  strerror(crash_msg->handler_errno));

	osm_log_to_stream(stream, " [CRASHD] = Stack trace addresses:\n");
	for (i = 0; i < crash_msg->num_backtrace_frames; ++i) {
		osm_log_to_stream(stream, " [CRASHD] = [#%02d] %p\n", i,
				  crash_msg->backtrace[i]);
		crashd_resolve_address(stream, crash_msg->backtrace[i]);
	}

	osm_log_to_stream(stream, " [CRASHD] = Backtrace symbols:\n"
		                  " ==== [CRASHD] ==== \n"
				  "%s"
				  " ==== [CRASHD] ==== \n\n",
			  crash_msg->symbols);

	crashd_dump_with_gstack(stream);

	fflush(NULL);
}

/********************************************************************
 ********************************************************************/

static void crashd_log_message(const char *filename)
{
	if (filename && strlen(filename)) {
		FILE *log_file = fopen(filename, "a");
		if (log_file) {
			print_to_stream(log_file);
			fclose(log_file);
			return;
		}
	}
	/*
	 * In any other case, just print the
	 * whole thing to stderr and syslog.
	 */
	print_to_stream(stderr);
}

/********************************************************************
 ********************************************************************/

static void crashd_term_sig_handler(int signal)
{

	terminate_flag = 1;
}

static void crashd_alarm_sig_handler(int signal)
{

	timeout_flag = 1;
}

static void crashd_fault_sig_handler(int signal)
{
	signal_flag = 1;
}

/* Utility function to register a simple signal handler with no flags
 * (as opposed to signal(2))
 */
static int crashd_register_signal(int signo, void *handler)
{
	struct sigaction act;

	memset(&act, 0, sizeof(act));
	act.sa_handler = handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;

	return sigaction(signo, &act, NULL);
}

/********************************************************************
 ********************************************************************/

static void crashd_set_daemon_name(int argc, char *argv[])
{
	int len = 0, i, j;
	for (i = argc - 1; i >= 0; i--) {
		len = strlen(argv[i]);
		for (j = len - 1; j >= 0; j--)
			argv[i][j] = '\0';
	}
	snprintf(argv[0], len, CRASHD_NAME);
}

/********************************************************************
 ********************************************************************/

void crashd_main(int argc, char *argv[], const char *log_file_name, int read_fd)
{
	int ret;
	char *p = crash_msg_buf;
	int remaining_bytes = CRASHD_MAX_MSG_SIZE;
	fd_set rfds;

	/*********************************************************
	 *                                                       *
	 * From this point on, we're in context of crash daemon. *
	 *                                                       *
	 *********************************************************/

	/* This forks again, closing stdin/out/err and loose our TTY, if asked to. */
	/*
	if(daemonise_flag && (daemon(0, 1) == -1)) goto _exit;
	*/

	memset(osm_binary_name, 0, sizeof(osm_binary_name));

	/* get real full path of the OpenSM executable */
	if (readlink("/proc/self/exe", osm_binary_name, CRASHD_MAX_FILENAME-1) < 0)
		goto _exit;

	crashd_set_daemon_name(argc, argv);

	/* Register all signal handlers for timeout, kill and fault */
	if (crashd_register_signal(SIGTERM, crashd_term_sig_handler)  ||
	    crashd_register_signal(SIGINT,  crashd_term_sig_handler)  ||
	    crashd_register_signal(SIGALRM, crashd_alarm_sig_handler) ||
	    crashd_register_signal(SIGSEGV, crashd_fault_sig_handler) ||
	    crashd_register_signal(SIGILL,  crashd_fault_sig_handler) ||
	    crashd_register_signal(SIGFPE,  crashd_fault_sig_handler) ||
	    crashd_register_signal(SIGBUS,  crashd_fault_sig_handler)) {
		goto _exit;
	}

	/* OK, wait for someone to tickle us */

	FD_ZERO(&rfds);
	FD_SET(read_fd, &rfds);

	do {
		ret = select(read_fd + 1, &rfds, NULL, NULL, NULL);
		/*mb();*/
		if (terminate_flag || signal_flag)
			goto _exit;
	/*
	 * Deal correctly with random harmless signals.
	 * Especially useful for when we run under debugger.
	 */
	} while (-1 == ret && EINTR == errno);

	if (-1 == ret)
		goto _exit;

	/* OK, we have action. First thing arm the timer */
	(void)alarm(CRASHD_READ_TIMEOUT);

	/*
	 * Reads the message from the pipe.
	 * Continue till done or till the timer expiration.
	 */

	do {

		ret = read(read_fd, p, remaining_bytes);

		if (0 == ret) {
			//osm_log_to_stream(stderr, " [OpenSM CRASHD] EOF received - pipe closed");
			break;
		}

		if (signal_flag) {
			osm_log_to_stream(stderr, " [OpenSM CRASHD] terminating due to signal");
			goto _exit;
		}

		if (terminate_flag) {
			osm_log_to_stream(stderr, " [OpenSM CRASHD] terminating");
			goto _exit;
		}

		if (timeout_flag) {
			osm_log_to_stream(stderr, " [OpenSM CRASHD] timeout");
			goto _exit;
		}

		/* Handle random signals nicely */
		if (-1 == ret && EINTR == errno) {
			//osm_log_to_stream(stderr, " [OpenSM CRASHD] read interrupted - reading again");
			continue;
		}

		/* Read error. Nothing we can do here... */
		if (-1 == ret) {
			osm_log_to_stream(stderr, " [OpenSM CRASHD] read error");
			goto _exit;
		}

		p += ret;
		remaining_bytes -= ret;

	} while (ret && (remaining_bytes > 0));

	/*
	 * Make sure that the message was properly received
	 */
	if (CRASHD_MSG_MAGIC != crash_msg->magic)
		goto _exit; /* No info received from OpenSM - nothing to do */

	crashd_log_message(log_file_name);

  _exit:

#if (0)
  bail_out:
	/*
	 * The crash daemon itself crashed.
	 * No need for backtrace - we have only one function in crash daemon
	 */

	crash_msg->process_id = getpid();
	crash_msg->thread_id = getpid();
	crash_msg->handler_errno = errno;
	crash_msg->num_backtrace_frames =
	    backtrace(crash_msg->backtrace, CRASHD_MAX_BACKTRACE_DEPTH);
	snprintf(crash_msg->auxiliary_data, CRASHD_AUXILIARY_DATA_SIZE - 1,
		 "crashd bailing out due to %s\n", strerror(errno));

	crashd_log_message(log_file_name);
#endif

	close(read_fd);
	exit(0);
}
