/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2006 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DB_PACK_C
#include <opensm/osm_db_pack.h>
#include <opensm/osm_log.h>
#include <opensm/osm_helper.h>
#include <sys/time.h>

#define LIST_INIT_MIN_ITEMS 10
#define PORT_PAIR_MAX_LENGTH 10
#define GUID_PAIR_MAX_LENGTH 40

static inline void pack_guid(uint64_t guid, char *p_guid_str)
{
	sprintf(p_guid_str, "0x%016" PRIx64, guid);
}

static inline uint64_t unpack_guid(char *p_guid_str)
{
	return strtoull(p_guid_str, NULL, 0);
}

static inline uint8_t unpack_port(char *p_num)
{
	unsigned long tmp_port = strtoul(p_num, NULL, 0);
	OSM_ASSERT(tmp_port < 0x100);
	return (uint8_t)tmp_port;
}

static inline void pack_lids(uint16_t min_lid, uint16_t max_lid, char *lid_str)
{
	sprintf(lid_str, "0x%04x 0x%04x", min_lid, max_lid);
}

static inline int unpack_lids(IN char *p_lid_str, OUT uint16_t * p_min_lid,
			      OUT uint16_t * p_max_lid)
{
	unsigned long tmp;
	char *p_next;
	char *p_num;
	char lids_str[24];

	strncpy(lids_str, p_lid_str, 23);
	lids_str[23] = '\0';
	p_num = strtok_r(lids_str, " \t", &p_next);
	if (!p_num)
		return 1;
	tmp = strtoul(p_num, NULL, 0);
	if (tmp >= 0xC000)
		return 1;

	*p_min_lid = (uint16_t) tmp;

	p_num = strtok_r(NULL, " \t", &p_next);
	if (!p_num)
		return 1;
	tmp = strtoul(p_num, NULL, 0);
	if (tmp >= 0xC000)
		return 1;

	*p_max_lid = (uint16_t) tmp;

	return 0;
}

static inline void pack_mkey(uint64_t mkey, char *p_mkey_str)
{
	sprintf(p_mkey_str, "0x%016" PRIx64, mkey);
}

static inline uint64_t unpack_mkey(char *p_mkey_str)
{
	return strtoull(p_mkey_str, NULL, 0);
}

static inline void pack_neighbor(uint64_t guid, uint8_t portnum, char *p_str)
{
	sprintf(p_str, "0x%016" PRIx64 ":%u", guid, portnum);
}

static inline int unpack_neighbor(char *p_str, uint64_t *guid,
				  uint8_t *portnum)
{
	char tmp_str[24];
	char *p_num, *p_next;
	unsigned long tmp_port;

	strncpy(tmp_str, p_str, 23);
	tmp_str[23] = '\0';
	p_num = strtok_r(tmp_str, ":", &p_next);
	if (!p_num)
		return 1;
	if (guid)
		*guid = strtoull(p_num, NULL, 0);

	p_num = strtok_r(NULL, ":", &p_next);
	if (!p_num)
		return 1;
	if (portnum) {
		tmp_port = strtoul(p_num, NULL, 0);
		OSM_ASSERT(tmp_port < 0x100);
		*portnum = (uint8_t) tmp_port;
	}

	return 0;
}

/*
 * Split a string into 2 tokens using a given delimiter. If the delimiter exists
 * more than once, then the splitting is done by the 1st occurrence. If the
 * delimiter is not found then 1 is returned, and the the result tokens are
 * undefined.
 * The caller should pass a mutable buffer at least as long as the input str.
 */
static
inline int split_str_to_2(char *p_str, char *buf, int buf_size,
			  char* delimiter, char **p_token1, char **p_token2)
{
	char *p_next;

	strncpy(buf, p_str, buf_size - 1);
	*p_token1 = strtok_r(buf, delimiter, &p_next);
	if (!*p_token1)
		return 1;

	*p_token2 = strtok_r(NULL, "", &p_next);
	if (!*p_token2)
		return 1;
	return 0;
}

static inline void pack_guid_pair(uint64_t guid1, uint64_t guid2, char *p_str)
{
	sprintf(p_str, "0x%016" PRIx64 "-0x%016" PRIx64,
		cl_hton64(guid1), cl_hton64(guid2));
}

static inline void pack_time_ports_pair(char* time, char* ports, char *p_str)
{
	sprintf(p_str, "%s-%s", time, ports);
}

static inline void pack_port_pair(uint8_t port1, uint8_t port2, char *p_str)
{
	sprintf(p_str, "%u-%u", port1, port2);
}

static inline
int unpack_guid_pair(char *p_str, uint64_t *guid1, uint64_t *guid2)
{
	char buf[GUID_PAIR_MAX_LENGTH], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	*guid1 = cl_ntoh64(unpack_guid(p_token1));
	*guid2 = cl_ntoh64(unpack_guid(p_token2));
	return 0;
}

static inline
int unpack_port_pair(char *p_str, uint8_t *port1, uint8_t *port2)
{
	char buf[PORT_PAIR_MAX_LENGTH], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	*port1 = unpack_port(p_token1);
	*port2 = unpack_port(p_token2);
	return 0;
}

static inline
int unpack_time_ports_pair(char *p_str, uint8_t *port1, uint8_t *port2, char *time)
{
	char buf[PORT_PAIR_MAX_LENGTH + MAX_DATE_STR_LEN + 1], *p_token1, *p_token2;

	if (split_str_to_2(p_str, buf, sizeof(buf), "-",
			   &p_token1, &p_token2) == 1)
		return 1;

	strncpy(time, p_token1, MAX_DATE_STR_LEN);
	if(unpack_port_pair(p_token2, port1, port2))
		return 1;
	return 0;
}

/******************************************************************************/
/*                                guid2lid                                    */
/******************************************************************************/

int osm_db_guid2lid_guids(IN osm_db_domain_t * p_g2l,
			  OUT cl_qlist_t * p_guid_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_guid_elem_t *p_guid_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_g2l, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_guid_elem =
		    (osm_db_guid_elem_t *)calloc(1, sizeof(osm_db_guid_elem_t));
		OSM_ASSERT(p_guid_elem != NULL);

		p_guid_elem->guid = unpack_guid(p_key);
		cl_qlist_insert_head(p_guid_list, &p_guid_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_guid2lid_get(IN osm_db_domain_t * p_g2l, IN uint64_t guid,
			OUT uint16_t * p_min_lid, OUT uint16_t * p_max_lid)
{
	char guid_str[20];
	char *p_lid_str;
	uint16_t min_lid, max_lid;

	pack_guid(guid, guid_str);
	p_lid_str = osm_db_lookup(p_g2l, guid_str);
	if (!p_lid_str)
		return 1;
	if (unpack_lids(p_lid_str, &min_lid, &max_lid))
		return 1;

	if (p_min_lid)
		*p_min_lid = min_lid;
	if (p_max_lid)
		*p_max_lid = max_lid;

	return 0;
}

int osm_db_guid2lid_set(IN osm_db_domain_t * p_g2l, IN uint64_t guid,
			IN uint16_t min_lid, IN uint16_t max_lid)
{
	char guid_str[20];
	char lid_str[16];

	pack_guid(guid, guid_str);
	pack_lids(min_lid, max_lid, lid_str);

	return osm_db_update(p_g2l, guid_str, lid_str);
}

int osm_db_guid2lid_delete(IN osm_db_domain_t * p_g2l, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);
	return osm_db_delete(p_g2l, guid_str);
}

/******************************************************************************/
/*                                guid2mkey                                   */
/******************************************************************************/

int osm_db_guid2mkey_guids(IN osm_db_domain_t * p_g2m,
			   OUT cl_qlist_t * p_guid_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_guid_elem_t *p_guid_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_g2m, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_guid_elem =
		    (osm_db_guid_elem_t *)calloc(1, sizeof(osm_db_guid_elem_t));
		OSM_ASSERT(p_guid_elem != NULL);

		p_guid_elem->guid = unpack_guid(p_key);
		cl_qlist_insert_head(p_guid_list, &p_guid_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_guid2mkey_get(IN osm_db_domain_t * p_g2m, IN uint64_t guid,
			 OUT uint64_t * p_mkey)
{
	char guid_str[20];
	char *p_mkey_str;

	pack_guid(guid, guid_str);
	p_mkey_str = osm_db_lookup(p_g2m, guid_str);
	if (!p_mkey_str)
		return 1;

	if (p_mkey)
		*p_mkey = unpack_mkey(p_mkey_str);

	return 0;
}

int osm_db_guid2mkey_set(IN osm_db_domain_t * p_g2m, IN uint64_t guid,
			 IN uint64_t mkey)
{
	char guid_str[20];
	char mkey_str[20];

	pack_guid(guid, guid_str);
	pack_mkey(mkey, mkey_str);

	return osm_db_update(p_g2m, guid_str, mkey_str);
}

int osm_db_guid2mkey_delete(IN osm_db_domain_t * p_g2m, IN uint64_t guid)
{
	char guid_str[20];
	pack_guid(guid, guid_str);
	return osm_db_delete(p_g2m, guid_str);
}

/******************************************************************************/
/*                                neighbor                                    */
/******************************************************************************/

int osm_db_neighbor_guids(IN osm_db_domain_t * p_neighbor,
			  OUT cl_qlist_t * p_neighbor_list)
{
	char *p_key;
	cl_list_t keys;
	osm_db_neighbor_elem_t *p_neighbor_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_neighbor, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_neighbor_elem =
		    (osm_db_neighbor_elem_t *)calloc(1, sizeof(osm_db_neighbor_elem_t));
		OSM_ASSERT(p_neighbor_elem != NULL);

		unpack_neighbor(p_key, &p_neighbor_elem->guid,
				&p_neighbor_elem->portnum);
		cl_qlist_insert_head(p_neighbor_list, &p_neighbor_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_neighbor_get(IN osm_db_domain_t * p_neighbor, IN uint64_t guid1,
			IN uint8_t portnum1, OUT uint64_t * p_guid2,
			OUT uint8_t * p_portnum2)
{
	char neighbor_str[24];
	char *p_other_str;
	uint64_t temp_guid;
	uint8_t temp_portnum;

	pack_neighbor(guid1, portnum1, neighbor_str);
	p_other_str = osm_db_lookup(p_neighbor, neighbor_str);
	if (!p_other_str)
		return 1;
	if (unpack_neighbor(p_other_str, &temp_guid, &temp_portnum))
		return 1;

	if (p_guid2)
		*p_guid2 = temp_guid;
	if (p_portnum2)
		*p_portnum2 = temp_portnum;

	return 0;
}

int osm_db_neighbor_set(IN osm_db_domain_t * p_neighbor, IN uint64_t guid1,
			IN uint8_t portnum1, IN uint64_t guid2,
			IN uint8_t portnum2)
{
	char n1_str[24], n2_str[24];

	pack_neighbor(guid1, portnum1, n1_str);
	pack_neighbor(guid2, portnum2, n2_str);

	return osm_db_update(p_neighbor, n1_str, n2_str);
}

int osm_db_neighbor_delete(IN osm_db_domain_t * p_neighbor, IN uint64_t guid,
			   IN uint8_t portnum)
{
	char n_str[24];

	pack_neighbor(guid, portnum, n_str);
	return osm_db_delete(p_neighbor, n_str);
}

/******************************************************************************/
/*                                  an2an                                     */
/******************************************************************************/

int osm_db_an2an_get_all(IN osm_db_domain_t * p_domain,
			 OUT cl_qlist_t * p_elements)
{
	char *p_key, *p_value;
	cl_list_t keys;
	osm_db_an2an_elem_t *p_elem;

	cl_list_construct(&keys);
	cl_list_init(&keys, LIST_INIT_MIN_ITEMS);

	if (osm_db_keys(p_domain, &keys))
		return 1;

	while ((p_key = cl_list_remove_head(&keys)) != NULL) {
		p_elem =
		    (osm_db_an2an_elem_t *)calloc(1, sizeof(osm_db_an2an_elem_t));
		OSM_ASSERT(p_elem != NULL);
		p_value = osm_db_lookup(p_domain, p_key);

		if (unpack_guid_pair(p_key, &p_elem->guid1, &p_elem->guid2) ||
		    unpack_time_ports_pair(p_value, &p_elem->portnum1,
					   &p_elem->portnum2, p_elem->timestamp_str))
			free(p_elem);
		else
			cl_qlist_insert_head(p_elements, &p_elem->item);
	}

	cl_list_destroy(&keys);
	return 0;
}

int osm_db_an2an_get(IN osm_db_domain_t * p_domain,
		     IN uint64_t sw_guid1, IN uint64_t sw_guid2,
		     OUT uint8_t * port1, OUT uint8_t * port2, OUT char *time)
{
	char p_key[40];
	char *p_value;

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	p_value = osm_db_lookup(p_domain, p_key);
	if (!p_value)
		return 1;
	return unpack_time_ports_pair(p_value, port1, port2, time);
}

static inline void update_timestamp(char *time)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	sprint_time_str(&tv, time, MAX_DATE_STR_LEN, 1);
}

int osm_db_an2an_set(IN osm_db_domain_t * p_domain,
		     IN uint64_t sw_guid1, IN uint64_t sw_guid2,
		     IN uint8_t port1, IN uint8_t port2,
		     IN boolean_t zero_time)
{
	char p_key[GUID_PAIR_MAX_LENGTH], p_ports[PORT_PAIR_MAX_LENGTH];
	char p_time[MAX_DATE_STR_LEN] = { 0 };
	char p_value[MAX_DATE_STR_LEN + 1 + PORT_PAIR_MAX_LENGTH];

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	pack_port_pair(port1, port2, p_ports);
	if(zero_time)
		sprintf(p_time, "%s", "time");
	else
		update_timestamp(p_time);

	pack_time_ports_pair(p_time, p_ports, p_value);
	return osm_db_update(p_domain, p_key, p_value);
}

int osm_db_an2an_delete(IN osm_db_domain_t * p_domain,
			IN uint64_t sw_guid1, IN uint64_t sw_guid2)
{
	char p_key[40];

	pack_guid_pair(sw_guid1, sw_guid2, p_key);
	return osm_db_delete(p_domain, p_key);
}
