/*
 * Copyright (c) 2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DFP_CPP
#include <opensm/osm_log.h>
#include <opensm/osm_port.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_node.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_armgr.hpp>

struct dfp_context {
	osm_ucast_mgr_t *p_ucast_mgr;
	OSMAdaptiveRoutingManager *p_armgr;
};

static int dfp_build_lfts(void *context)
{
	struct dfp_context *ctx = (struct dfp_context*)context;
	osm_ucast_mgr_t *p_ucast_mgr = ctx->p_ucast_mgr;
	OSMAdaptiveRoutingManager *p_armgr = ctx->p_armgr;
	int result = 0;

	if ((result = osm_ucast_mgr_build_lft_tables(p_ucast_mgr)))
		return result;

	result = p_armgr->Route();
	return result;
}

static void dfp_update_osm_sl2vl(void *context, osm_physp_t *osm_phys_port,
				 uint8_t iport_num, uint8_t oport_num,
				 ib_slvl_table_t *osm_oport_sl2vl)
{
	struct dfp_context *ctx = (struct dfp_context *)context;
	OSMAdaptiveRoutingManager *p_armgr = ctx->p_armgr;
	uint8_t sl2vl[IB_NUMBER_OF_SLS] = {0}, sl, vl;

	if (p_armgr->GetSL2VL(osm_phys_port, iport_num, oport_num, &sl2vl))
		return;

	for (sl = 0; sl < IB_NUMBER_OF_SLS; sl++) {
		vl = sl2vl[sl];
		ib_slvl_table_set(osm_oport_sl2vl, sl, vl);
	}
}

static
void dfp_context_delete(void *context)
{
	struct dfp_context *ctx = (struct dfp_context *)context;
	OSMAdaptiveRoutingManager * p_armgr = ctx->p_armgr;
	delete p_armgr;

	free(ctx);
}

static struct dfp_context *dfp_context_create(osm_opensm_t *osm)
{
	struct dfp_context *ctx;

	ctx = (struct dfp_context*)calloc(1, sizeof(*ctx));
	if (ctx) {
		ctx->p_ucast_mgr = &osm->sm.ucast_mgr;
		ctx->p_armgr = new OSMAdaptiveRoutingManager(osm, AR_ALGORITHM_DF_PLUS);
		if (ctx->p_armgr)
			ctx->p_armgr->Construct();
	} else
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
			"ERR 8A01: calloc: %s\n", strerror(errno));

	return ctx;
}

extern "C" {
int osm_ucast_dfp_setup(struct osm_routing_engine *r,
			osm_opensm_t *osm)
{
	void *ctx;

	ctx = dfp_context_create(osm);
	if (!ctx)
		return -1;

	r->context = ctx;
	/* build default minhop lid matrice */
	r->ucast_build_fwd_tables = dfp_build_lfts;
	r->update_sl2vl = dfp_update_osm_sl2vl;

	r->destroy = dfp_context_delete;

	r->ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->shield_mode_auto = OSM_ADV_ROUTING_MODE_DISABLED;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_NONE;

	return 0;
}
} //extern
