/*
 * Copyright (c) 2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of OpenSM enhanced QoS manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <complib/cl_crc.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ENH_QOS_MGR_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_enhanced_qos_mgr.h>

#define BW_NAMES "BW_NAMES"
#define BW_RULES "BW_RULES"
#define VPORT_BW_RULES "VPORT_BW_RULES"

#define LINE_LEN 1024
#define EQUAL "="
#define SEMICOLON ":"
#define DEFAULT_RULE "default"
#define MAX_RATE_LIMIT 0x000FFFFF
#define UNLIMITED_RATE_LIMIT 0
#define DEFAULT_RATE_LIMIT UNLIMITED_RATE_LIMIT
#define PHYS_RULES 1
#define VPORT_RULES 2

extern void remove_whitespaces(char *str);
extern boolean_t atouint32(char *buf, uint32_t *val);
extern boolean_t str_to_guid(const char* str, uint64_t *guid);
extern const char *osm_get_node_type_str_fixed_width(IN uint8_t node_type);

/*
 * This function expects a buf which represents bandwidth layout.
 *
 * The function parse the buf and set bandwidth layout attributes in given
 * osm_qos_config_t struct
 *
 * Currently rate limit is the only attribute supported in qos config so
 * expected buf is a number which represents rate limit value
 */
static boolean_t str_to_qos_config(osm_log_t *p_log, char *buf,
				   osm_qos_config_t *p_qos_config)
{
	uint32_t rate_limit;
	boolean_t res = TRUE;

	if (!atouint32(buf, &rate_limit) || (rate_limit > MAX_RATE_LIMIT)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0706: "
			"%s is not a valid rate limit\n", buf);
		res = FALSE;
		goto Exit;
	}

	p_qos_config->rate_limit = rate_limit;

Exit:
	return res;
}

/*
 * This function expects a buf in a "<key>=<value>" format.
 *
 * If buf is in expected format the function does two things:
 *	1) set p_val to point on value
 *	2) return the length of the key
 *
 * If buf is NOT in expected format the function return zero.
 */
static size_t analyze_key_value(char *buf, char **p_val, char *delim, uint16_t max_len)
{
	size_t key_len = strcspn(buf, delim);

	if (key_len == strlen(buf) || key_len >= max_len)
		/* delimiter not found or key too long */
		key_len = 0;
	else
		*p_val = buf + key_len + 1;
	return key_len;
}

/*
 * Validate BW NAME by making sure this name is not already used
 *
 * If Valid -  return the crc to use as key for this name
 * If Not Valid -  return 0
 */
static uint32_t validate_name(osm_log_t *p_log, cl_qmap_t *p_bw_names,
			      char *name)
{
	uint32_t crc = cl_calc_crc32((void *)name, strlen(name), 0);

	if (cl_qmap_get(p_bw_names, crc) != cl_qmap_end(p_bw_names)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0702: "
			"Duplicate BW NAME '%s'\n", name);
		crc = 0;
	}
	return crc;
}

static void save_bw_name(osm_log_t *p_log, cl_qmap_t *p_bw_names, uint32_t key,
			 osm_qos_config_t *p_qos_config)
{
	osm_qos_config_t *p_new;

	p_new = calloc(1, sizeof(osm_qos_config_t));
	if (!p_new) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0703: "
			"Failed to allocate memory for qos config\n");
		return;
	}
	p_new->rate_limit = p_qos_config->rate_limit;

	cl_qmap_insert(p_bw_names, (uint64_t)key, (cl_map_item_t*)p_new);
}

/* This function returns indication whether process was successful or not */
static boolean_t process_name(osm_enhanced_qos_mgr_t *p_mgr, osm_log_t *p_log,
			      cl_qmap_t *p_bw_names, int line, char *buf)
{
	char *p_val;
	size_t key_len = analyze_key_value(buf, &p_val, EQUAL, LINE_LEN);
	char key[LINE_LEN];
	osm_qos_config_t qos_config;
	boolean_t res = FALSE;
	uint32_t crc;

	if (key_len && (key_len < LINE_LEN)) {
		strncpy(key, buf, key_len);
		key[key_len] = '\0';
		crc = validate_name(p_log, p_bw_names, key);
		if (!crc)
			goto Exit;

		res = str_to_qos_config(p_log, p_val, &qos_config);
	}

	if (res) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Enhanced qos, line %d, parsed BW NAME '%s': "
			"rate limit = %u\n",
			line, key, qos_config.rate_limit);
		save_bw_name(p_log, p_bw_names, crc, &qos_config);
	} else
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0701: "
			"Enhanced qos, line %d, failed to parse BW NAME '%s'\n",
			line, buf);

Exit:
	return res;
}

/*
 * Receive a buf in expected format <sl>:<BW NAME>
 * Perform validation
 *
 * If valid translate values to given pointer of qos config
 *
 * Return indication whether validation pass
 */
static boolean_t analyze_sl_qos_config(osm_log_t *p_log, int line,
				       cl_qmap_t *p_bw_names,
				       char *buf, uint8_t *sl,
				       osm_qos_config_t *p_qos_config)
{
	char *p_val;
	size_t key_len = analyze_key_value(buf, &p_val, SEMICOLON, LINE_LEN);
	char key[LINE_LEN];
	uint32_t crc, sl_val;
	cl_map_item_t *p_item;
	boolean_t res = TRUE;

	if (key_len && (key_len < LINE_LEN)) {
		strncpy(key, buf, key_len);
		key[key_len] = '\0';

		/* Validate SL */
		if (key_len > 2 && !strncasecmp(key, "all", 3)) {
			*sl = IB_NUMBER_OF_SLS;
		} else {
			if (!atouint32(key, &sl_val) || (sl_val >= IB_NUMBER_OF_SLS)) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0705: "
					"Enhanced qos, line %d, %s is not a valid sl\n",
					line, key);
				res = FALSE;
				goto Exit;
			}
			*sl = (uint8_t)sl_val;
		}

		/* Validate BS NAME */
		crc = cl_calc_crc32((void *)p_val, strlen(p_val), 0);
		p_item = cl_qmap_get(p_bw_names, (uint64_t)crc);
		if (p_item == cl_qmap_end(p_bw_names)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0704: "
				"Enhanced qos, line %d, BW NAME %s not found\n",
				line, p_val);
			res = FALSE;
			goto Exit;
		}
		*p_qos_config = *((osm_qos_config_t *)p_item);
	} else {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0710: "
			"Enhanced qos, line %d, buffer \"%.1024s\" is not "
			"in expected format.\n", line, buf);
		res = FALSE;
	}
Exit:
	return res;
}

/*
 * This function expects a buf in qos config set format.
 *
 * qos config set format is as follow:
 *
 * <sl id>:<BM NAME>,<sl id><BW_NAME>
 *
 * If buf is in qos config set format this function will translate each token
 * to qos_config_t struct and set it in given set pointer in appropriate sl
 * index.
 *
 * This function also set default qos_config_t to every sl id which was not
 * specified in the buf
 *
 * This function return boolean indication whether buf was in expected format
 */
static boolean_t str_to_qos_config_set(osm_log_t *p_log, int line,
				       cl_qmap_t *p_bw_names,
				       char *buf,
				       osm_all_sls_qos_config_set_t *p_sls_set,
				       boolean_t is_manual_rule)
{
	boolean_t res = TRUE;
	char *token;
	char *context = NULL;
	osm_qos_config_t qos_config;
	uint8_t sl;
	uint8_t last_sl;

	memset(&qos_config.map_item, 0, sizeof(cl_map_item_t));
	token = strtok_r(buf, ",", &context);
	while (token) {
		res = analyze_sl_qos_config(p_log, line, p_bw_names, token, &sl,
					    &qos_config);
		if (res) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Enhanced qos, line %d, setting rate limit %u "
				"for SL %x\n", line, qos_config.rate_limit,
					       (sl == IB_NUMBER_OF_SLS)? 0xFFFF : sl);
			last_sl = sl + 1;

			if (sl == IB_NUMBER_OF_SLS) {
				sl = 0;
				last_sl = IB_NUMBER_OF_SLS;
			}
			qos_config.is_manual_rule = is_manual_rule;
			for (;sl < last_sl; sl++) {
				p_sls_set->qos_configs[sl] = qos_config;
			}
		} else
			break;
		token = strtok_r(NULL, ",", &context);
	}

	return res;
}

typedef struct pgrp_name {
	cl_fmap_item_t map_item;
	int line_num;
	char *name;
} pgrp_name_t;

int ib_compare_pgrp_names(IN const void *const p_name1,
			  IN const void *const p_name2)
{
        return strcmp((const char *)p_name1, (const char *)p_name2);
}

static boolean_t
save_pg_name(osm_log_t *p_log, cl_fmap_t *p_pg_names, char* key, int line)
{
	pgrp_name_t *p_new;

	p_new = calloc(1, sizeof(pgrp_name_t));
	if (!p_new) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 070D: "
			"Enhanced qos, line %d, failed to allocate memory for "
			"enhanced qos port group entry", line);
		return FALSE;
	}
	p_new->name = strdup(key);
	if (p_new->name == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 070E: "
			"Enhanced qos, line %d, failed to allocate memory for "
			"enhanced qos port group name %s", line, key);
		free(p_new);
		return FALSE;
	}

	p_new->line_num = line;

	cl_fmap_insert(p_pg_names, p_new->name, &p_new->map_item);
	return TRUE;
}

static boolean_t
is_valid_rule_pgrp_name(osm_log_t *p_log, int line, char *key,
			osm_port_group_t **pgrp, cl_fmap_t *p_pg_names)
{
	boolean_t res = TRUE;
	pgrp_name_t *p_pgrp_name;

	p_pgrp_name = (pgrp_name_t *)cl_fmap_match(p_pg_names, key,
						   ib_compare_pgrp_names);

	/* such port group already appeared in rules */
	if (p_pgrp_name !=
	    (pgrp_name_t *)cl_fmap_end(p_pg_names)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 070C: "
			"Enhanced qos, line %d, rule already exists for "
			"port group '%s' at line %d\n",
			line, key, p_pgrp_name->line_num);
		res = FALSE;
	} else
		/* keep for next time */
		res = save_pg_name(p_log, p_pg_names, key, line);

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Enhanced qos, line %d, port group name "
		"%s is %s.\n", line, key, res ? "valid" : "invalid");

	return res;
}

static boolean_t
is_valid_rule_key(osm_log_t *p_log, int line, char *key,
		  cl_qmap_t *sl_qos_config_map, ib_net64_t *guid,
		  osm_port_group_t **p_pgrp, struct osm_opensm *p_osm,
		  cl_fmap_t *p_pg_names,
		  boolean_t *to_ignore)
{
	uint64_t host_guid = 0;
	boolean_t res = TRUE;

	*to_ignore = FALSE;
	/* First validate guid key */
	if (!str_to_guid(key, &host_guid)) {
		if (strcmp(key, DEFAULT_RULE)) {
			*p_pgrp = osm_get_port_grp(p_osm, key);
			if (*p_pgrp)
				/*
				 * check that the rule with such port group
				 * name, didn't appear earlier in the file
				 */
				res = is_valid_rule_pgrp_name(p_log, line, key,
							      p_pgrp, p_pg_names);
			else {
				/*
				 * give a warning but don't fail the parsing
				 * just ignore the line
				 */
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Enhanced qos, line %d, '%s' "
					"is neither a valid guid, "
					"nor an existing "
					"port group name \n",
					line, key);
				*to_ignore = TRUE;
			}
		} else /* Default rule */
			*guid = 0;
	} else
		*guid = cl_hton64(host_guid);

	return res;
}

static boolean_t
process_port_group(osm_log_t *p_log, int line, osm_port_group_t *p_pgrp,
		   cl_qmap_t *p_sl_qos_config_map,
		   osm_all_sls_qos_config_set_t *p_set)
{
	cl_map_item_t *p_item;
	osm_port_group_member_t *p_member;
	ib_net64_t guid;
	osm_all_sls_qos_config_set_t *p_new_set, *p_set_check;

	for (p_item = cl_qmap_head(&p_pgrp->ports);
	     p_item != cl_qmap_end(&p_pgrp->ports);
	     p_item = cl_qmap_next(p_item))
	{
		p_member = (osm_port_group_member_t*)p_item;
		guid = osm_port_get_guid(p_member->port);
		p_new_set = calloc(1, sizeof(osm_all_sls_qos_config_set_t));
		if (!p_new_set) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0709: "
				"Enhanced qos, line %d, failed to allocate "
				"memory for sl set\n", line);
			return FALSE;
		}
		memcpy(p_new_set->qos_configs, p_set->qos_configs,
		       sizeof(p_new_set->qos_configs));
		p_new_set->is_guid_rule = FALSE;
		p_set_check =
		(osm_all_sls_qos_config_set_t *)cl_qmap_insert(
				p_sl_qos_config_map, guid, &p_new_set->map_item);
		/* such guid already exists */
		if (p_set_check != p_new_set) {
			/*
			 * override only guids that are set by
			 * port group rules.
			 */
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Enhanced qos, line %d, "
				"rule overrides exisiting guid 0x%" PRIx64
				" that was set by %s rule.\n",
				line, cl_ntoh64(guid),
				p_set_check->is_guid_rule ?
					"guid" : "port group");
			if (!p_set_check->is_guid_rule)
				memcpy(&p_set_check->qos_configs,
				       &p_new_set->qos_configs, sizeof(p_set_check->qos_configs));
			/* free anyway */
			free(p_new_set);
		}
	}
	return TRUE;
}

static boolean_t
process_guid(osm_log_t *p_log, int line, cl_qmap_t *p_sl_qos_config_map,
	     ib_net64_t guid, osm_all_sls_qos_config_set_t *p_set)
{
	boolean_t res = TRUE;
	osm_all_sls_qos_config_set_t *p_new_set, *p_set_check;

	p_new_set = calloc(1, sizeof(osm_all_sls_qos_config_set_t));
	if (!p_new_set) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 070F: "
			"Enhanced qos, line %d, failed to allocate memory "
			"for sl set\n",	line);
		res = FALSE;
		goto Exit;
	}

	memcpy(p_new_set->qos_configs, p_set->qos_configs,
	       sizeof(p_new_set->qos_configs));
	/* mark that this guid came from guid rule */
	p_new_set->is_guid_rule = TRUE;

	p_set_check =
	(osm_all_sls_qos_config_set_t *)cl_qmap_insert(p_sl_qos_config_map,
						guid, &p_new_set->map_item);
	if (p_set_check != p_new_set) {
		if (p_set_check->is_guid_rule) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 070A: "
				"Enhanced qos, line %d, rule already exists "
				"for guid 0x%" PRIx64 "\n", line,
				cl_ntoh64(guid));
			res = FALSE;
		} else
			/* entry exists from port group, just override it */
			memcpy(p_set_check->qos_configs, p_new_set->qos_configs,
			       sizeof(p_set_check->qos_configs));

		/* either way release memory */
		free(p_new_set);
	}
Exit:
	return res;
}

static boolean_t process_rule(osm_enhanced_qos_mgr_t *p_mgr, osm_log_t *p_log,
			      cl_qmap_t *p_bw_names, cl_fmap_t *p_pg_names,
			      int line, char *buf, boolean_t is_vport_section)
{
	char *p_val;
	size_t key_len = analyze_key_value(buf, &p_val, EQUAL, LINE_LEN);
	char key[LINE_LEN];
	osm_all_sls_qos_config_set_t sls_set;
	boolean_t res = FALSE;
	ib_net64_t guid = 0;
	osm_port_group_t *p_pgrp = NULL;
	cl_qmap_t *p_map = is_vport_section ? &p_mgr->sl_qos_vport_config :
					      &p_mgr->sl_qos_config;

	osm_all_sls_qos_config_set_t *p_default_sl_set = is_vport_section ?
					&p_mgr->default_vport_sl_set :
					&p_mgr->default_sl_set;
	boolean_t ignore_line = FALSE;

	if (key_len && (key_len < LINE_LEN)) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "Processing rule '%s'\n", buf);
		strncpy(key, buf, key_len);
		key[key_len] = '\0';

		res = is_valid_rule_key(p_log, line, key, p_map,
					&guid, &p_pgrp, p_mgr->p_subn->p_osm,
					p_pg_names, &ignore_line);
		if (!res || ignore_line)
			goto Exit;
		/* Initialize with default set */
		memcpy(&sls_set, p_default_sl_set, sizeof(sls_set));

		/* Now validate the sl qos config set */
		res = str_to_qos_config_set(p_log, line, p_bw_names, p_val,
					    &sls_set, (guid != 0));
	} else
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0708: Enhanced qos, "
			"line %d, Invalid Rule '%s'\n", line, buf);

	if (res) {
		if (p_pgrp)
			/* the last port group is the first in the list */
			res = process_port_group(p_log, line, p_pgrp, p_map,
						 &sls_set);
		else if (guid == 0)
			/* Update the default set with new default */
			memcpy(p_default_sl_set, &sls_set, sizeof(sls_set));
		else
			res = process_guid(p_log, line, p_map, guid, &sls_set);
	}

Exit:
	OSM_LOG(p_log, OSM_LOG_DEBUG, "Enhanced qos, line %d, "
		"rule processed with %s and was %signored\n",
		line, res ? "SUCCESS" : "FAILURE", ignore_line ? "" : "NOT ");
	return res;
}

static void
osm_enhanced_qos_mgr_empty_config(IN osm_enhanced_qos_mgr_t * p_mgr)
{
	cl_qmap_t *p_map;
	cl_map_item_t *p_item;

	OSM_LOG_ENTER(&p_mgr->p_subn->p_osm->log);

	p_map = &p_mgr->sl_qos_config;
	while ((p_item = cl_qmap_head(p_map)) != cl_qmap_end(p_map)) {
		cl_qmap_remove_item(p_map, p_item);
		free(p_item);
        }
	p_map = &p_mgr->sl_qos_vport_config;
	while ((p_item = cl_qmap_head(p_map)) != cl_qmap_end(p_map)) {
		cl_qmap_remove_item(p_map, p_item);
		free(p_item);
        }
	p_mgr->is_configured = FALSE;

	OSM_LOG_EXIT(&p_mgr->p_subn->p_osm->log);
}

void osm_enhanced_qos_mgr_parse_config(OUT osm_enhanced_qos_mgr_t *p_mgr,
				       IN osm_sm_t *p_sm)
{
	FILE *file;
	char *file_name = p_sm->p_subn->opt.enhanced_qos_policy_file;
	char buf[LINE_LEN];
	int line = 0;
	int reading_rules = 0;
	boolean_t reading_names = FALSE;
	boolean_t reading_phys_rules = FALSE, reading_vport_rules = FALSE;
	boolean_t res = TRUE;
	cl_qmap_t bw_names;
	cl_fmap_t pg_names;
	cl_map_item_t *p_item;
	cl_fmap_item_t *p_fitem;
	pgrp_name_t *p_pgrp_name;

	OSM_LOG_ENTER(p_sm->p_log);

	cl_qmap_init(&bw_names);
	cl_fmap_init(&pg_names, ib_compare_pgrp_names);

	/* empty previous configuration database. */
	osm_enhanced_qos_mgr_empty_config(p_mgr);

	if (!p_sm->p_subn->opt.qos || !file_name) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Enahnced Qos support "
			"is disabled, skipping enhanced qos policy "
			"file parsing\n");
		goto Exit;
	}

	file = fopen(file_name, "r");
	if (!file) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 0700: "
			"Can't open enhanced qos policy file \'%s\'\n",
			file_name);
		goto Exit;
	}

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Parsing Enhanced QoS Policy "
		"File\n");
	while (fgets(buf, sizeof buf, file) != NULL) {
		line++;
		remove_whitespaces(buf);

		if (!strlen(buf) || buf[0] == '#')
			continue;

		reading_phys_rules = !strcmp(buf, BW_RULES);
		if (!reading_phys_rules)
			reading_vport_rules = !strcmp(buf, VPORT_BW_RULES);

		if ((reading_phys_rules || reading_vport_rules) && !reading_names) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 0713: "
				"BW_NAMES should be defined first \n");
			res = FALSE;
			break;
		}

		if (reading_phys_rules) {
			/* Clean up settings from previous cycle */
			memcpy(&p_mgr->default_sl_set, &p_mgr->unlimited_sl_set,
			       sizeof(p_mgr->unlimited_sl_set));

			reading_rules = PHYS_RULES;
			continue;
		}

		if (reading_vport_rules) {
			/* Clean up settings from previous cycle */
			memcpy(&p_mgr->default_vport_sl_set,
			       &p_mgr->unlimited_sl_set,
			       sizeof(p_mgr->unlimited_sl_set));

			reading_rules = VPORT_RULES;
			continue;
		}
		if (reading_rules > 0) {
			if (!process_rule(p_mgr, p_sm->p_log,
					  &bw_names, &pg_names,
					  line, buf,
					  (reading_rules == VPORT_RULES))) {
				res = FALSE;
				break;
			}
		}
		else if (reading_names &&
			 !process_name(p_mgr, p_sm->p_log, &bw_names,
				       line, buf)) {
				res = FALSE;
				break;
		} else {
			if (!strcmp(buf, BW_NAMES))
				reading_names = TRUE;
		}
	}

	if (res && reading_names && (reading_rules > 0)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Enhanced QoS Policy File "
			"Parsed Successfully\n");
		p_mgr->is_configured = TRUE;
	} else
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 0707: "
			"Enhanced QoS Policy File Parsing Failed\n");
	fclose(file);
Exit:
	while (cl_qmap_count(&bw_names)) {
		p_item = cl_qmap_head(&bw_names);
		cl_qmap_remove_item(&bw_names, p_item);
		free(p_item);
	}

        while(!cl_is_fmap_empty(&pg_names)) {
                p_fitem = cl_fmap_head(&pg_names);
                cl_fmap_remove_item(&pg_names, p_fitem);
                p_pgrp_name = (pgrp_name_t *)p_fitem;
		free(p_pgrp_name->name);
                free(p_pgrp_name);
        }

	OSM_LOG_EXIT(p_sm->p_log);
}

ib_api_status_t
osm_enhanced_qos_mgr_init(OUT osm_enhanced_qos_mgr_t *p_mgr,
			  IN struct osm_sm *p_sm)
{
	int i;
	OSM_LOG_ENTER(p_sm->p_log);
	cl_qmap_init(&p_mgr->sl_qos_config);
	cl_qmap_init(&p_mgr->sl_qos_vport_config);

	/* Currently the default value is set to zero's */
	memset(&p_mgr->unlimited_sl_set, 0,
	       sizeof(osm_all_sls_qos_config_set_t));

	for (i = 0; i < IB_NUMBER_OF_SLS; i++)
		p_mgr->unlimited_sl_set.qos_configs[i].rate_limit =
				DEFAULT_RATE_LIMIT;

	memcpy(&p_mgr->default_sl_set, &p_mgr->unlimited_sl_set,
	       sizeof(p_mgr->default_sl_set));

	memcpy(&p_mgr->default_vport_sl_set, &p_mgr->unlimited_sl_set,
	       sizeof(p_mgr->default_vport_sl_set));

	p_mgr->p_subn = p_sm->p_subn;

	p_mgr->is_configured = FALSE;

	OSM_LOG_EXIT(p_sm->p_log);
	return IB_SUCCESS;
}

void osm_enhanced_qos_mgr_destroy(IN osm_enhanced_qos_mgr_t * p_mgr)
{
	OSM_LOG_ENTER(&p_mgr->p_subn->p_osm->log);

	osm_enhanced_qos_mgr_empty_config(p_mgr);

	OSM_LOG_EXIT(&p_mgr->p_subn->p_osm->log);
	return;
}

static osm_all_sls_qos_config_set_t *get_sl_set(osm_log_t *p_log,
						osm_enhanced_qos_mgr_t *p_mgr,
						ib_net64_t guid,
						osm_vport_t *p_vport)
{
	cl_qmap_t *p_map = p_vport ? &p_mgr->sl_qos_vport_config :
				      &p_mgr->sl_qos_config;
	cl_map_item_t *p_item;
	osm_all_sls_qos_config_set_t *p_sl_set;
	osm_all_sls_qos_config_set_t *p_default_sl_set = p_vport ?
		&p_mgr->default_vport_sl_set :
		&p_mgr->default_sl_set;

	p_item = cl_qmap_get(p_map, guid);
	if (p_item == cl_qmap_end(p_map)) {
		if (p_vport && p_vport->index == 0 &&
		    p_mgr->p_subn->opt.enhanced_qos_vport0_unlimit_default_rl == TRUE)
			p_sl_set = &p_mgr->unlimited_sl_set;
		else
			p_sl_set = p_default_sl_set;

		/* Returned configuration default */
		OSM_LOG(p_log, OSM_LOG_DEBUG, "Getting QosConfigSl for "
			"guid 0x%" PRIx64 " based on default matching "
			"rule\n", cl_ntoh64(guid));
	} else {
		int i;
		/* Set configured guid value */
		OSM_LOG(p_log, OSM_LOG_DEBUG, "Getting QosConfigSl for guid 0x%"
			PRIx64 " based on guid matching rule\n",
			cl_ntoh64(guid));
		p_sl_set = (osm_all_sls_qos_config_set_t*)p_item;

		if (!p_vport || p_vport->index ||
		    !p_mgr->p_subn->opt.enhanced_qos_vport0_unlimit_default_rl)
			goto Exit;
		/* Special handling of vport0 SLs not mentioned in the rule */
		/* When opt.enhanced_qos_vport0_unlimit_default_rl = TRUE */
		for (i = 0; i < IB_NUMBER_OF_SLS; i++) {
			if (!p_sl_set->qos_configs[i].is_manual_rule)
				p_sl_set->qos_configs[i].rate_limit = UNLIMITED_RATE_LIMIT;
		}
	}
Exit:
	return p_sl_set;
}

/* get wire speed in 1 Mbs units */
static uint32_t get_wire_speed_val(uint8_t wire_speed_rate) {
	uint32_t wire_speed;

	switch (wire_speed_rate) {
		case IB_PATH_RECORD_RATE_2_5_GBS:
			wire_speed = 2500;
			break;
		case IB_PATH_RECORD_RATE_10_GBS:
			wire_speed = 10000;
			break;
		case IB_PATH_RECORD_RATE_30_GBS:
			wire_speed = 30000;
			break;
		case IB_PATH_RECORD_RATE_5_GBS:
			wire_speed = 5000;
			break;
		case IB_PATH_RECORD_RATE_20_GBS:
			wire_speed = 20000;
			break;
		case IB_PATH_RECORD_RATE_40_GBS:
			wire_speed = 40000;
			break;
		case IB_PATH_RECORD_RATE_60_GBS:
			wire_speed = 60000;
			break;
		case IB_PATH_RECORD_RATE_80_GBS:
			wire_speed = 80000;
			break;
		case IB_PATH_RECORD_RATE_120_GBS:
			wire_speed = 120000;
			break;
		case IB_PATH_RECORD_RATE_14_GBS:
			wire_speed = 14000;
			break;
		case IB_PATH_RECORD_RATE_56_GBS:
			wire_speed = 56000;
			break;
		case IB_PATH_RECORD_RATE_112_GBS:
			wire_speed = 112000;
			break;
		case IB_PATH_RECORD_RATE_168_GBS:
			wire_speed = 168000;
			break;
		case IB_PATH_RECORD_RATE_25_GBS:
			wire_speed = 25000;
			break;
		case IB_PATH_RECORD_RATE_100_GBS:
			wire_speed = 100000;
			break;
		case IB_PATH_RECORD_RATE_200_GBS:
			wire_speed = 200000;
			break;
		case IB_PATH_RECORD_RATE_300_GBS:
			wire_speed = 300000;
			break;
		case IB_PATH_RECORD_RATE_28_GBS:
			wire_speed = 28000;
			break;
		case IB_PATH_RECORD_RATE_50_GBS:
			wire_speed = 50000;
			break;
		case IB_PATH_RECORD_RATE_400_GBS:
			wire_speed = 400000;
			break;
		case IB_PATH_RECORD_RATE_600_GBS:
			wire_speed = 600000;
			break;
		default:
			wire_speed = 0;
	}

	return wire_speed;
}

static uint32_t get_wire_speed(osm_enhanced_qos_mgr_t *p_mgr, ib_net64_t guid)
{
	cl_qmap_t *p_ports_tbl = &p_mgr->p_subn->port_guid_tbl;
	osm_port_t *p_port;
	ib_port_info_t *p_pi;
	int extended;
	uint8_t wire_speed_rate;
	uint32_t wire_speed = MAX_RATE_LIMIT;

	p_port = (osm_port_t *)cl_qmap_get(p_ports_tbl, guid);
	if (p_port != (osm_port_t *)cl_qmap_end(p_ports_tbl)) {
		p_pi = &p_port->p_physp->port_info;
		extended = p_pi->capability_mask &
			   IB_PORT_CAP_HAS_EXT_SPEEDS;
		wire_speed_rate =
			ib_port_info_compute_rate(p_pi,
						  extended);
		wire_speed = get_wire_speed_val(wire_speed_rate);
	}

	return wire_speed;
}


static void
osm_enhanced_qos_mgr_get_guid_sl(IN osm_log_t *p_log,
				 IN osm_enhanced_qos_mgr_t *p_mgr,
				 IN ib_net64_t guid,
				 IN ib_net64_t check_wire_speed_guid,
				 IN osm_vport_t  *p_vport,
				 OUT qos_config_sl_t *p_qos_conf)
{
	int sl;
	ib_bw_layout_t *p_bw_layout;
	osm_all_sls_qos_config_set_t *p_sl_set = get_sl_set(p_log, p_mgr,
							    guid, p_vport);
	uint32_t rate_limit;
	uint32_t wire_speed = get_wire_speed(p_mgr, check_wire_speed_guid);

	for (sl = 0 ; sl < IB_NUMBER_OF_SLS ; sl++) {
		p_bw_layout = &p_qos_conf->sl_bandwidth[sl];
		rate_limit = p_sl_set->qos_configs[sl].rate_limit;
		/* check that we got known wire_speed and compare to it */
		if (wire_speed && (rate_limit > wire_speed)) {
			OSM_LOG(p_log, OSM_LOG_VERBOSE,
				"QosConfigSL for guid (0x%" PRIx64 "): "
				"sl %d, rate limit %u Mbs "
				"is above port's (0x%" PRIx64 ") "
				"link rate limit %u Mbs, "
				"setting to link rate limit.\n",
				cl_ntoh64(guid), sl, rate_limit,
				cl_ntoh64(check_wire_speed_guid), wire_speed);
			rate_limit = wire_speed;
		}
		ib_bw_layout_set_rate_limit(p_bw_layout, 0, rate_limit);
	}
}

static boolean_t
osm_enhanced_qos_mgr_set_required(IN osm_sm_t * sm,
				  IN qos_config_sl_t *p_prev_qos_conf,
				  IN qos_config_sl_t *p_qos_conf,
				  IN osm_physp_t *p_physp,
				  IN ib_net64_t guid,
				  IN boolean_t is_vport)
{
	boolean_t res = TRUE;
	unsigned force_update = p_physp->need_update ||	sm->p_subn->need_update;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Qos Config SL configuration for "
		"guid 0x%" PRIx64 ", physp need update %u, "
		"subnet need update %u, force update %u\n",
		cl_ntoh64(guid), p_physp->need_update,
		sm->p_subn->need_update, force_update);

	if (!force_update &&
	    !memcmp(p_prev_qos_conf, p_qos_conf, sizeof(*p_prev_qos_conf))) {
		res = FALSE;
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Qos Config SL configuration "
			"matches stored payload for %s guid 0x%" PRIx64	"\n",
			is_vport ? "vport" : "port", cl_ntoh64(guid));
	}

	return res;
}

static void
osm_enhanced_qos_mgr_process_port(IN osm_sm_t * sm,
				  IN osm_port_t *p_port,
				  OUT cl_qlist_t *mad_list)
{
	osm_enhanced_qos_mgr_t *p_mgr = &sm->enhanced_qos_mgr;
	qos_config_sl_t qos_conf;
	osm_madw_t *p_madw;
	ib_net64_t m_key;
	osm_physp_t *p;
	osm_madw_context_t context;
	osm_node_t *p_node = p_port->p_node;
	uint8_t node_type = osm_node_get_type(p_node);
	ib_general_info_smp_t *p_gi = &p_node->general_info;
	ib_net64_t port_guid = osm_port_get_guid(p_port);

	OSM_LOG_ENTER(sm->p_log);

	/* currently supported for CAs only */
	if (node_type != IB_NODE_TYPE_CA ||
	    !ib_general_info_is_cap_enabled(p_gi,
		IB_GENERAL_INFO_QOS_CONFIG_SL_RATE_LIMIT_SUPPORTED_MASK)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Enhanced qos is not supported on port 0x%" PRIx64
			". Node is of type %s\n",
			cl_ntoh64(port_guid),
			osm_get_node_type_str_fixed_width(node_type));
		goto Exit;
	}

	memset(&qos_conf, 0, sizeof(qos_conf));

	osm_enhanced_qos_mgr_get_guid_sl(sm->p_log,
					 p_mgr,
					 port_guid,
					 port_guid,
					 NULL,
					 &qos_conf);
	p = p_port->p_physp;
	if (osm_enhanced_qos_mgr_set_required(sm,
					      &p->qos_config_sl, &qos_conf,
					      p, port_guid, FALSE)) {
		m_key = ib_port_info_get_m_key(&p->port_info);
		context.qos_config_sl_context.port_guid = port_guid;
		context.qos_config_sl_context.set_method = TRUE;
		/* add mad to list */
		p_madw = osm_prepare_req_set(sm,
					     osm_physp_get_dr_path_ptr(p),
					     (uint8_t *) &qos_conf,
					     sizeof(qos_conf),
					     IB_MAD_ATTR_QOS_CONFIG_SL,
					     0, FALSE, m_key,
					     0, CL_DISP_MSGID_NONE, &context);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0714: "
				"Failed to prepare QoS Config SL MAD for "
				"port guid 0x%" PRIx64 "\n",
				cl_ntoh64(port_guid));
			goto Exit;
		}

		cl_qlist_insert_tail(mad_list, &p_madw->list_item);

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Added QosConfigSL MAD set "
			"for port guid 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid));
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_enhanced_qos_mgr_process_virt(IN osm_sm_t * sm,
				       OUT cl_qlist_t *batch_mad_list)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	qos_mad_list_t *p_list, *p_list_next;
	cl_qlist_t qos_mad_list;
	osm_madw_t *p_port_mad;

	OSM_LOG_ENTER(sm->p_log);

	cl_qlist_init(&qos_mad_list);

	p_item = cl_qmap_head(&sm->p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&sm->p_subn->port_guid_tbl)) {
		p_port = (osm_port_t *)p_item;
		p_item = cl_qmap_next(p_item);
		p_list = (qos_mad_list_t *) malloc(sizeof(*p_list));
		if (!p_list) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0712: "
				"Failed to allocate plist for "
				"port guid 0x%" PRIx64 " \n",
				cl_ntoh64(osm_port_get_guid(p_port)));
			goto Exit;
		}
		memset(p_list, 0, sizeof(*p_list));

		cl_qlist_init(&p_list->port_mad_list);

		/* add vports MADs of this port */
		osm_enhanced_qos_mgr_process_vports(sm, p_port,
						    &p_list->port_mad_list);
		if (cl_qlist_count(&p_list->port_mad_list)) {
			cl_qlist_insert_tail(&qos_mad_list, &p_list->list_item);
		} else
			free(p_list);
	}

	while (cl_qlist_count(&qos_mad_list)) {
		p_list_next = (qos_mad_list_t *) cl_qlist_head(&qos_mad_list);
		while (p_list_next !=
			(qos_mad_list_t *) cl_qlist_end(&qos_mad_list)) {
			p_list = p_list_next;
			p_list_next = (qos_mad_list_t *)
				cl_qlist_next(&p_list->list_item);
			/* next MAD to send*/
			p_port_mad = (osm_madw_t *)
				     cl_qlist_remove_head(&p_list->port_mad_list);
			cl_qlist_insert_tail(batch_mad_list, &p_port_mad->list_item);
			/* remove the QoS MAD from global MAD list */
			if (cl_qlist_count(&p_list->port_mad_list) == 0) {
				cl_qlist_remove_item(&qos_mad_list, &p_list->list_item);
				free(p_list);
			}
		}
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_enhanced_qos_mgr_process(IN struct osm_sm *p_sm,
				  OUT cl_qlist_t *batch_mad_list)
{
	cl_qmap_t *p_tbl;
	cl_map_item_t *p_next;
	osm_port_t *p_port;
	osm_enhanced_qos_mgr_t *p_mgr = &p_sm->enhanced_qos_mgr;

	OSM_LOG_ENTER(p_sm->p_log);
	osm_enhanced_qos_mgr_parse_config(
		&p_sm->p_subn->p_osm->sm.enhanced_qos_mgr,
		&p_sm->p_subn->p_osm->sm);

	if (!p_mgr->is_configured) {
		OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE,
			"enhanced qos database is not configured.\n");
		goto Exit;
	}

	p_tbl = &p_sm->p_subn->port_guid_tbl;
	p_next = cl_qmap_head(p_tbl);
	while (p_next != cl_qmap_end(p_tbl)) {
		p_port = (osm_port_t *) p_next;
		p_next = cl_qmap_next(p_next);
		/* check if port is eligible and add its mads to list. */
		osm_enhanced_qos_mgr_process_port(p_sm,
						  p_port,
						  batch_mad_list);
	}

	/* take care of exisiting virtual ports */
	osm_enhanced_qos_mgr_process_virt(p_sm,
					  batch_mad_list);

Exit:
	OSM_LOG_EXIT(p_sm->p_log);
}

void
osm_enhanced_qos_mgr_process_vports(IN osm_sm_t * p_sm,
				    IN struct osm_port *p_port,
				    OUT cl_qlist_t *port_mad_list)
{
	osm_enhanced_qos_mgr_t *p_mgr = &p_sm->enhanced_qos_mgr;
	qos_config_sl_t qos_conf;
	ib_net64_t m_key;
	osm_physp_t *p;
	osm_madw_context_t context;
	osm_node_t *p_node = p_port->p_node;
	uint8_t node_type = osm_node_get_type(p_node);
	ib_general_info_smp_t *p_gi = &p_node->general_info;
	osm_vport_t *p_vport;
	uint16_t vport_index_top, vport_index;
	osm_dr_path_t path;
	ib_api_status_t status;
	uint32_t attr_mod_ho;
	ib_net64_t port_guid = osm_port_get_guid(p_port), vport_guid;
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_sm->p_log);

	if (!p_mgr->is_configured)
		goto Exit;

	if (!osm_port_is_virt_supported(p_port) ||
	    node_type != IB_NODE_TYPE_CA ||
	    !ib_general_info_is_cap_enabled(p_gi,
	    IB_GENERAL_INFO_QOS_CONFIG_SL_VPORT_RATE_LIMIT_SUPPORTED_MASK)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"Enhanced qos for vports is not supported on port 0x%" PRIx64
			". Node is of type %s\n",
			cl_ntoh64(port_guid),
			osm_get_node_type_str_fixed_width(node_type));
		goto Exit;
	}

	if (!p_port->virt_data.vports) {
		OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE,
			"Did not find virtual ports on port 0x%" PRIx64 ", "
			"skipping port\n",
			cl_ntoh64(osm_port_get_guid(p_port)));
		goto Exit;
	}

	if (!(vport_index_top = osm_port_get_vport_top_index(p_port)))
		goto Exit;

	p = p_port->p_physp;
	path = *osm_physp_get_dr_path_ptr(p);
	m_key = ib_port_info_get_m_key(&p->port_info);

	context.qos_config_sl_context.port_guid = port_guid;
	context.qos_config_sl_context.set_method = TRUE;

	for (vport_index = 0; vport_index <= vport_index_top; vport_index++) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport)
			continue;

		vport_guid = osm_vport_get_guid(p_vport);
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Setting enhanced qos data "
			"for virtual port guid 0x%" PRIx64 " index %u "
			"of port 0x%" PRIx64 "\n", cl_ntoh64(vport_guid),
			vport_index, cl_ntoh64(port_guid));

		memset(&qos_conf, 0, sizeof(qos_conf));
		osm_enhanced_qos_mgr_get_guid_sl(p_sm->p_log, p_mgr,
						 vport_guid, port_guid,
						 p_vport, &qos_conf);
		if (osm_enhanced_qos_mgr_set_required(p_sm,
						      &p_vport->qos_config_sl,
						      &qos_conf, p, vport_guid,
						      TRUE)) {
			attr_mod_ho = vport_index |
				      VPORT_QOS_CONFIG_SL_ATTR_MOD_HO;
			if (port_mad_list) {
				p_madw = osm_prepare_req_set(p_sm,
							     osm_physp_get_dr_path_ptr(p),
							     (uint8_t *) &qos_conf,
							     sizeof(qos_conf),
							     IB_MAD_ATTR_QOS_CONFIG_SL,
							     cl_hton32(attr_mod_ho), FALSE, m_key,
							     0, CL_DISP_MSGID_NONE, &context);
				if (!p_madw) {
					OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 0711: "
						"Failed to prepare QoS Config SL MAD for "
						"port guid 0x%" PRIx64 ", vport index %u, "
						"vport guid 0x%" PRIx64 "\n",
						cl_ntoh64(port_guid), vport_index,
						cl_ntoh64(vport_guid));
					goto Exit;
				}

				cl_qlist_insert_tail(port_mad_list, &p_madw->list_item);
				osm_port_increase_virt_transactions(p_sm, p_port);

				OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Added QosConfigSL MAD set "
					"for port guid 0x%" PRIx64 ", vport index %u, "
					"vport guid 0x%" PRIx64 "\n",
					cl_ntoh64(port_guid), vport_index,
					cl_ntoh64(vport_guid));
			} else {
				/* send the MAD immediately */
				status = osm_req_set(p_sm, &path, (uint8_t *) &qos_conf,
						     sizeof(qos_conf),
						     IB_MAD_ATTR_QOS_CONFIG_SL,
						     cl_hton32(attr_mod_ho), FALSE, m_key,
						     0, CL_DISP_MSGID_NONE, &context);
				if (status == IB_SUCCESS) {
					osm_port_increase_virt_transactions(p_sm, p_port);
					OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
					"Sent qos config sl set request to VPort "
					"0x%" PRIx64 ", VPort index %u "
					"of port 0x%" PRIx64 " (%s)\n",
					cl_ntoh64(osm_vport_get_guid(p_vport)),
					vport_index, cl_ntoh64(port_guid),
					ib_get_err_str(status));
				} else
					OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 070B: "
					"Failure initiating qos config sl set request "
					"to VPort 0x%" PRIx64 ", VPort index %u of port"
					" 0x%" PRIx64 " (%s)\n",
					cl_ntoh64(vport_guid), vport_index,
					cl_ntoh64(port_guid), ib_get_err_str(status));
			}
		}
	}
Exit:
	OSM_LOG_EXIT(p_sm->p_log);
}
