/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_general_info_rcv_t.
 * This object represents the GeneralInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_GENERAL_INFO_RCV_C
#include <opensm/osm_opensm.h>

extern void ni_rcv_complete_process_new_node(IN osm_sm_t *sm,
					     IN osm_node_t *p_node,
					     IN const ib_smp_t *p_smp,
					     IN osm_ni_context_t *ni_context,
					     IN ib_node_info_t *p_ni);


void osm_general_info_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm;
	sm = context;
	osm_madw_t *p_madw = data;
	ib_general_info_smp_t *p_general_info;
	ib_smp_t *p_smp;
	osm_node_t *p_node;
	osm_general_info_context_t *p_context;
	ib_net64_t node_guid;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);
	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_general_info_context_ptr(p_madw);
	p_general_info = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_GENERAL_INFO);

	if (p_smp->attr_mod != IB_GENERAL_INFO_CAP_ATTRIBUTE_MODIFIER){
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4000: Got MAD with "
			"unknown attribute modifier 0x%08" PRIx32 "\n",
			cl_ntoh32(p_smp->attr_mod));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "ERR 4002: "
			"GeneralInfo MAD with status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
	}

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	node_guid = p_context->node_guid;
	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);

	if (!p_node) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4001: Got GeneralInfo "
			"from unknown node 0x%" PRIx64 "\n",
			cl_ntoh64(node_guid));
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received General Info response from "
		"node 0x%" PRIx64 "\n", cl_ntoh64(node_guid));

	verbose_bypass = osm_verbose_bypass_handle_geni(sm, p_node);

	memcpy(&p_node->general_info ,p_general_info,
	       sizeof(p_node->general_info));

	p_node->gi_updated = TRUE;

	/*
	 * Now that we got the general info of the node we can continue to
	 * process it
	 */
	ni_rcv_complete_process_new_node(sm, p_node, p_smp,
					 &p_context->ni_context,
					 &p_context->node_info);

	CL_PLOCK_RELEASE(sm->p_lock);
Exit:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	return;
}
