/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_gi_rcv_t.
 * This object represents the GUIDInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_GUID_INFO_RCV_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_remote_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_guid.h>
#include <opensm/osm_unhealthy_ports.h>

void osm_gi_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_guid_info_t *p_gi;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_gi_context_t *p_context;
	ib_net64_t port_guid, node_guid;
	uint8_t block_num;
	osm_guidinfo_work_obj_t *p_obj;
	cl_list_item_t *p_list_item, *p_list_item_next;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_gi_context_ptr(p_madw);
	p_gi = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_GUID_INFO);

	block_num = (uint8_t) cl_ntoh32(p_smp->attr_mod);

	port_guid = p_context->port_guid;
	node_guid = p_context->node_guid;

	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	verbose_bypass = osm_verbose_bypass_handle_gi(sm, p_port);

	osm_dump_guid_info_v2(sm->p_log, node_guid, port_guid, block_num, p_gi,
			      FILE_ID, OSM_LOG_DEBUG);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4701: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid), cl_ntoh64(node_guid),
			cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		if (p_context->set_method)
			osm_hm_set_by_node(sm, osm_port_get_node(p_port),
					   OSM_HM_COND_SET_ERR);
		goto Exit;
	}

	p_list_item_next = cl_qlist_head(&sm->p_subn->alias_guid_list);
	while (p_list_item_next != cl_qlist_end(&sm->p_subn->alias_guid_list)) {
		p_obj = (osm_guidinfo_work_obj_t *) p_list_item_next;
		if (p_obj->p_port != p_port) {
			p_list_item_next = cl_qlist_next(p_list_item_next);
			continue;
		}

		if (p_obj->block_num != block_num) {
			p_list_item_next = cl_qlist_next(p_list_item_next);
			continue;
		}

		if (p_obj->req_dirty)
			break;

		p_list_item = p_list_item_next;
		p_list_item_next = cl_qlist_next(p_list_item);

		cl_qlist_remove_item(&sm->p_subn->alias_guid_list, p_list_item);
		osm_guid_work_obj_delete(p_obj);

	}

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	osm_verbose_bypass_end(verbose_bypass, sm->p_log);

	OSM_LOG_EXIT(sm->p_log);
}
