/*
 * Copyright (c) 2006-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_guid_mgr_t.
 * This object implements the GUID manager object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_GUID_MGR_C
#include <opensm/osm_port.h>
#include <opensm/osm_node.h>
#include <opensm/osm_guid.h>
#include <opensm/osm_opensm.h>

static void guidinfo_set(IN osm_sa_t *sa, IN osm_port_t *p_port,
			 IN uint8_t block_num)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	osm_madw_context_t context;
	ib_api_status_t status;

	memcpy(payload,
	       &((*p_port->p_physp->p_guids)[block_num * GUID_TABLE_MAX_ENTRIES]),
	       sizeof(ib_guid_info_t));

	context.gi_context.node_guid = osm_node_get_node_guid(p_port->p_node);
	context.gi_context.port_guid = osm_physp_get_port_guid(p_port->p_physp);
	context.gi_context.set_method = TRUE;
	context.gi_context.port_num = osm_physp_get_port_num(p_port->p_physp);

	status = osm_req_set(sa->sm, osm_physp_get_dr_path_ptr(p_port->p_physp),
			     payload, sizeof(payload), IB_MAD_ATTR_GUID_INFO,
			     cl_hton32((uint32_t)block_num), FALSE,
			     ib_port_info_get_m_key(&p_port->p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);
	if (status != IB_SUCCESS)
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 5109: "
			"Failure initiating GUIDInfo request (%s)\n",
			ib_get_err_str(status));
}

osm_guidinfo_work_obj_t *osm_guid_work_obj_new(IN osm_port_t * p_port,
					       IN uint8_t block_num, boolean_t force_client_rereg)
{
	osm_guidinfo_work_obj_t *p_obj;

	/*
	   clean allocated memory to avoid assertion when trying to insert to
	   qlist.
	   see cl_qlist_insert_tail(): OSM_ASSERT(p_list_item->p_list != p_list)
	*/
	p_obj = calloc(1, sizeof(*p_obj));
	if (p_obj) {
		p_obj->p_port = p_port;
		p_obj->block_num = block_num;
		p_obj->force_client_rereg = force_client_rereg;
		p_obj->req_dirty = FALSE;
	}

	return p_obj;
}

void osm_guid_work_obj_delete(IN osm_guidinfo_work_obj_t * p_wobj)
{
	free(p_wobj);
}

int osm_queue_guidinfo(IN osm_sa_t *sa, IN osm_port_t *p_port,
		      IN uint8_t block_num, boolean_t force_client_rereg)
{
	osm_guidinfo_work_obj_t *p_obj;
	cl_list_item_t *p_list_item;
	int status = 1;

	p_list_item = cl_qlist_head(&sa->p_subn->alias_guid_list);
	while (p_list_item != cl_qlist_end(&sa->p_subn->alias_guid_list)) {
		p_obj = (osm_guidinfo_work_obj_t *) p_list_item;
		if (p_obj->p_port != p_port) {
			p_list_item = cl_qlist_next(p_list_item);
			continue;
		}

		if (p_obj->block_num != block_num) {
			p_list_item = cl_qlist_next(p_list_item);
			continue;
		}

		p_obj->req_dirty = TRUE;
		p_obj->force_client_rereg |= force_client_rereg;
		return 0;
	}

	p_obj = osm_guid_work_obj_new(p_port, block_num, force_client_rereg);
	if (p_obj)
		cl_qlist_insert_tail(&sa->p_subn->alias_guid_list,
				     &p_obj->list_item);
	else {
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 510F: "
			"Memory allocation of guid work object failed\n");
		status = 0;
	}

	return status;
}

void osm_guid_mgr_process(IN osm_sm_t * sm)
{
	osm_guidinfo_work_obj_t *p_obj;
	cl_list_item_t          *p_list_item;
	cl_map_item_t           *p_map_item;
	osm_port_t              *p_port;
	u_int64_t               r_aguid;
	uint32_t                max_block;
	osm_alias_guid_t       *p_alias_guid, *p_alias_guid_check;
	osm_rtr_aguid_mode_t rguid_mode = sm->p_subn->opt.rtr_aguid_enable;
	cl_qmap_t *aguid_tbl = &sm->p_subn->alias_port_guid_tbl;
	ib_net64_t cur_rguid;

	OSM_LOG_ENTER(sm->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Processing alias guid list\n");

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	/* configure routing alias guids for router support */
	if (rguid_mode != OSM_RTR_AGUID_IGNORE) {
		for (p_map_item = cl_qmap_head(&sm->p_subn->port_guid_tbl);
		     p_map_item != cl_qmap_end(&sm->p_subn->port_guid_tbl);
		     p_map_item = cl_qmap_next(p_map_item)) {

			p_port = (osm_port_t *)p_map_item;

			/* check that there is room for routing alias guid */
			if (p_port->p_physp->port_info.guid_cap < 2)
				continue;
			if (!p_port->p_physp->p_guids) {
				max_block =
					(p_port->p_physp->port_info.guid_cap +
					  GUID_TABLE_MAX_ENTRIES - 1) /
					GUID_TABLE_MAX_ENTRIES;
				p_port->p_physp->p_guids =
					calloc(max_block *
					       GUID_TABLE_MAX_ENTRIES,
					       sizeof(ib_net64_t));
				if (!p_port->p_physp->p_guids) {
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 5135: cannot allocate "
						"GUID table for port GUID 0x%"
						PRIx64 "\n",
						cl_ntoh64(
						p_port->p_physp->port_guid));
					/*goto _out;*/
					break;
				}
			}

			cur_rguid = (*p_port->p_physp->p_guids)[1];

			/* clear routing alias guids,
			 * if it's not already clear */
			if (rguid_mode == OSM_RTR_AGUID_DISABLE) {

				/* skip if already configured this a_guid */
				if (!cur_rguid &&
				    !sm->p_subn->first_time_master_sweep &&
				    !p_port->is_new)
					continue;

				r_aguid = 0;

				if (!cur_rguid)
					goto Set_aguid;

				p_alias_guid_check = (osm_alias_guid_t *)
					cl_qmap_remove(aguid_tbl, cur_rguid);

				if (p_alias_guid_check) {
					OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
						"Clearing router Aguid 0x%"
						PRIx64 " for port 0x%" PRIx64
						"\n", cl_ntoh64(cur_rguid),
						cl_ntoh64(p_port->p_physp->port_guid));
					free(p_alias_guid_check);
				} else
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 5138: Failed to remove "
						" alias guid 0x%" PRIx64 " from "
						"alias guid table (not found)\n",
						cl_ntoh64(cur_rguid));
			} else {
				/* OUI | index | lid */
				r_aguid = cl_ntoh64(
				    ((u_int64_t)OSM_VENDOR_ID_OPENIB << 40) |
				    cl_ntoh16(p_port->lid));
				/* skip if already configured this a_guid */
				if (cur_rguid == r_aguid)
					continue;

				p_alias_guid = osm_alias_guid_new(r_aguid,
							  p_port, 0);
				if (!p_alias_guid) {
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 5136: Alias guid at index"
						" 1, memory allocation failed "
						"for port GUID 0x%" PRIx64 "\n",
						cl_ntoh64(p_port->p_physp->port_guid));
					break;
				}
				p_alias_guid_check =
					(osm_alias_guid_t *) cl_qmap_insert(
						aguid_tbl,
						p_alias_guid->alias_guid,
						&p_alias_guid->map_item);
				if (p_alias_guid_check != p_alias_guid) {
					if (p_alias_guid_check->dirty) {
						p_alias_guid_check->dirty = FALSE;
						p_alias_guid_check->p_base_port = p_alias_guid->p_base_port;
						p_alias_guid_check->base_port_guid = p_alias_guid->base_port_guid;
						p_alias_guid_check->vport_index = p_alias_guid->vport_index;
						osm_alias_guid_delete(&p_alias_guid);
					} else {
						/* alias GUID is a duplicate */
						OSM_LOG(sm->p_log, OSM_LOG_ERROR,
							"ERR 5137: Duplicate alias"
							" port GUID 0x%" PRIx64
							" at index 1,"
							" base port GUID 0x%" PRIx64
							"\n",
							cl_ntoh64(p_alias_guid->alias_guid),
							cl_ntoh64(p_alias_guid->p_base_port->guid));
						osm_alias_guid_delete(&p_alias_guid);
						continue;
					}
				}
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Creating "
					"router Aguid 0x%" PRIx64 " for port "
					"0x%" PRIx64 "\n", cl_ntoh64(r_aguid),
					cl_ntoh64(p_port->p_physp->port_guid));
			}
Set_aguid:
			/* setup base port guid in index 0 */
	                (*p_port->p_physp->p_guids)[0] = p_port->p_physp->port_guid;
			(*p_port->p_physp->p_guids)[1] = r_aguid;
			osm_queue_guidinfo(&sm->p_subn->p_osm->sa, p_port, 0, FALSE);
		}
	}
	p_list_item = cl_qlist_head(&sm->p_subn->alias_guid_list);
	while (p_list_item != cl_qlist_end(&sm->p_subn->alias_guid_list)) {
		p_obj = (osm_guidinfo_work_obj_t *) p_list_item;
		if (p_obj->force_client_rereg) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Setting client reregister on the subnet\n");
			sm->p_subn->force_client_rereg = TRUE;
			p_obj->force_client_rereg = FALSE;
		}
		guidinfo_set(&sm->p_subn->p_osm->sa, p_obj->p_port,
			     p_obj->block_num);
		p_obj->req_dirty = FALSE;
		p_list_item = cl_qlist_next(p_list_item);
	}

	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
}
