/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2006 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of inform record functions.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_INFORM_C
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_inform_mgr.h>
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_pkey.h>
#include <opensm/osm_sa.h>
#include <opensm/osm_opensm.h>

typedef struct osm_infr_match_ctxt {
	cl_list_t *p_remove_infr_list;
	ib_mad_notice_attr_t *p_ntc;
} osm_infr_match_ctxt_t;

void osm_infr_delete(IN osm_infr_t * p_infr)
{
	free(p_infr);
}

osm_infr_t *osm_infr_new(IN const osm_infr_t * p_infr_rec)
{
	osm_infr_t *p_infr;

	OSM_ASSERT(p_infr_rec);

	p_infr = (osm_infr_t *) malloc(sizeof(osm_infr_t));
	if (p_infr)
		memcpy(p_infr, p_infr_rec, sizeof(osm_infr_t));

	return p_infr;
}

static void dump_all_informs(IN const osm_subn_t * p_subn, IN osm_log_t * p_log)
{
	cl_list_item_t *p_list_item;

	if (!OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_DEBUG))
		return;

	p_list_item = cl_qlist_head(&p_subn->sa_infr_list);
	while (p_list_item != cl_qlist_end(&p_subn->sa_infr_list)) {
		osm_dump_inform_info_v2(p_log,
				        &((osm_infr_t *) p_list_item)->
				        inform_record.inform_info, FILE_ID, OSM_LOG_DEBUG);
		p_list_item = cl_qlist_next(p_list_item);
	}
}

/**********************************************************************
 * Match an infr by the InformInfo and Address vector
 **********************************************************************/
static cl_status_t match_inf_rec(IN const cl_list_item_t * p_list_item,
				 IN void *context)
{
	osm_infr_t *p_infr_rec = (osm_infr_t *) context;
	osm_infr_t *p_infr = (osm_infr_t *) p_list_item;
	ib_inform_info_t *p_ii_rec = &p_infr_rec->inform_record.inform_info;
	ib_inform_info_t *p_ii = &p_infr->inform_record.inform_info;
	osm_log_t *p_log = p_infr_rec->sa->p_log;
	cl_status_t status = CL_NOT_FOUND;

	OSM_LOG_ENTER(p_log);

	if (memcmp(&p_infr->report_addr, &p_infr_rec->report_addr,
		   sizeof(p_infr_rec->report_addr))) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "Differ by Address\n");
		goto Exit;
	}

	/* if inform_info.gid is not zero, ignore lid range */
	if (ib_gid_is_notzero(&p_ii_rec->gid)) {
		if (memcmp(&p_ii->gid, &p_ii_rec->gid, sizeof(p_ii->gid))) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.gid\n");
			goto Exit;
		}
	} else {
		if ((p_ii->lid_range_begin != p_ii_rec->lid_range_begin) ||
		    (p_ii->lid_range_end != p_ii_rec->lid_range_end)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.LIDRange\n");
			goto Exit;
		}
	}

	if (p_ii->trap_type != p_ii_rec->trap_type) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Differ by InformInfo.TrapType\n");
		goto Exit;
	}

	if (p_ii->is_generic != p_ii_rec->is_generic) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Differ by InformInfo.IsGeneric\n");
		goto Exit;
	}

	if (p_ii->is_generic) {
		if (p_ii->g_or_v.generic.trap_num !=
		    p_ii_rec->g_or_v.generic.trap_num)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Generic.TrapNumber\n");
		else if (p_ii->g_or_v.generic.qpn_resp_time_val !=
			 p_ii_rec->g_or_v.generic.qpn_resp_time_val)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Generic.QPNRespTimeVal\n");
		else if (p_ii->g_or_v.generic.node_type_msb !=
			 p_ii_rec->g_or_v.generic.node_type_msb)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Generic.NodeTypeMSB\n");
		else if (p_ii->g_or_v.generic.node_type_lsb !=
			 p_ii_rec->g_or_v.generic.node_type_lsb)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Generic.NodeTypeLSB\n");
		else
			status = CL_SUCCESS;
	} else {
		if (p_ii->g_or_v.vend.dev_id != p_ii_rec->g_or_v.vend.dev_id)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Vendor.DeviceID\n");
		else if (p_ii->g_or_v.vend.qpn_resp_time_val !=
			 p_ii_rec->g_or_v.vend.qpn_resp_time_val)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Vendor.QPNRespTimeVal\n");
		else if (p_ii->g_or_v.vend.vendor_id_msb !=
			 p_ii_rec->g_or_v.vend.vendor_id_msb)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Vendor.VendorIdMSB\n");
		else if (p_ii->g_or_v.vend.vendor_id_lsb !=
			 p_ii_rec->g_or_v.vend.vendor_id_lsb)
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by InformInfo.Vendor.VendorIdLSB\n");
		else
			status = CL_SUCCESS;
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}


static boolean_t match_notice(IN osm_log_t * p_log,
			      IN osm_mad_addr_t *p_mad_addr,
			      IN ib_mad_notice_attr_t *p_ntci,
			      IN osm_infr_t *p_infr)
{
	ib_inform_info_t *p_ii = &p_infr->inform_record.inform_info;

	/*
	 * Matching rules:
	 * InformInfo   Notice
	 * GID          IssuerGID    if non zero, and trap number is 66/67
	 *			     GID must match notice details.
	 *			     if non zero, and trap number is not 66/67
	 *			     GID must match IssuerGID.
	 * LIDRange     IssuerLID    apply only if GID=0
	 * IsGeneric    IsGeneric    is compulsory and must match the trap
	 * Type         Type         if not 0xFFFF must match
	 * TrapNumber   TrapNumber   if not 0xFFFF must match
	 * DeviceId     DeviceID     if not 0xFFFF must match
	 * QPN dont care
	 * ProducerType ProducerType match or 0xFFFFFF
	 * VendorID     VendorID     match or 0xFFFFFF
	 *
	 * In addition, If address is given (p_mad_addr), verify it matches
	 * the InformInfo subscriber.
	 */

	if (p_mad_addr) {
		if (memcmp(&p_infr->report_addr, p_mad_addr,
			   sizeof(p_infr->report_addr))) {
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Differ by Address\n");
			goto Exit;
		}
	}

	if (p_ii->is_generic != (ib_notice_is_generic(p_ntci) != 0)) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "Differ by IsGeneric/Vendor\n");
		goto Exit;
	}

	/* if inform_info.gid is not zero, ignore lid range */
	if (ib_gid_is_notzero(&p_ii->gid)) {
		if (!ib_gid_is_multicast(&p_ii->gid)) {
			/* unicast subscription */
			if (memcmp(&p_ii->gid, &p_ntci->issuer_gid,
				   sizeof(p_ii->gid))) {
				OSM_LOG(p_log, OSM_LOG_DEBUG, "Differ by GID\n");
				goto Exit;
			}
		} else {
			/* multicast subscription */
			if (ib_notice_is_generic(p_ntci) &&
			    (p_ntci->g_or_v.generic.trap_num == CL_HTON16(66) ||
			     p_ntci->g_or_v.generic.trap_num == CL_HTON16(67))) {
				if (memcmp(&p_ii->gid,
					   &p_ntci->data_details.ntc_64_67.gid,
					   sizeof(p_ii->gid))) {
					OSM_LOG(p_log, OSM_LOG_DEBUG, "Differ by MGID\n");
					goto Exit;
				}
			} else {
				OSM_LOG(p_log, OSM_LOG_DEBUG, "Not multicast notice\n");
				goto Exit;
			}
		}
	} else if (p_ii->lid_range_begin != 0xFFFF) {
		if (p_ii->lid_range_end) {
			if ((cl_ntoh16(p_ii->lid_range_begin) >
			     cl_ntoh16(p_ntci->issuer_lid)) ||
			    (cl_ntoh16(p_ii->lid_range_end) <
			     cl_ntoh16(p_ntci->issuer_lid))) {
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Differ by LIDRange "
					"(II:%u-%u Trap:%u)\n",
					cl_ntoh16(p_ii->lid_range_begin),
					cl_ntoh16(p_ii->lid_range_end),
					cl_ntoh16(p_ntci->issuer_lid));
				goto Exit;
			}
		} else if (p_ii->lid_range_begin != p_ntci->issuer_lid) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Differ by LIDRangeBegin "
					"(II:%u Trap:%u)\n",
					cl_ntoh16(p_ii->lid_range_begin),
					cl_ntoh16(p_ntci->issuer_lid));
			goto Exit;
		}
	}

	if (p_ii->trap_type != 0xFFFF &&
	    cl_ntoh16(p_ii->trap_type) !=
	    ib_notice_get_type(p_ntci)) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Differ by TrapType (II:0x%X Trap:0x%X)\n",
			cl_ntoh16(p_ii->trap_type),
			ib_notice_get_type(p_ntci));
		goto Exit;
	}

	if (p_ii->is_generic) {
		if (p_ii->g_or_v.generic.trap_num != 0xFFFF &&
		    p_ii->g_or_v.generic.trap_num !=
		    p_ntci->g_or_v.generic.trap_num) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				 "Differ by Generic.TrapNumber (II:%u Trap:%u)\n",
				cl_ntoh16(p_ii->g_or_v.generic.trap_num),
				cl_ntoh16(p_ntci->g_or_v.generic.trap_num));
			goto Exit;
		}
		if (ib_inform_info_get_prod_type(p_ii) != cl_hton32(0xFFFFFF) &&
		    ib_inform_info_get_prod_type(p_ii) !=
		    ib_notice_get_prod_type(p_ntci)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by Generic.Type (II:0x%X Trap:0x%X)\n",
				cl_ntoh32(ib_inform_info_get_prod_type(p_ii)),
				cl_ntoh32(ib_notice_get_prod_type(p_ntci)));
			goto Exit;
		}
	} else {
		if (ib_inform_info_get_vend_id(p_ii) !=
		    ib_notice_get_vend_id(p_ntci)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by Vendor.VendorId (II:0x%X Trap:0x%X)\n",
				cl_ntoh32(ib_inform_info_get_vend_id(p_ii)),
				cl_ntoh32(ib_notice_get_vend_id(p_ntci)));
			goto Exit;
		}
		if (p_ii->g_or_v.vend.dev_id != 0xFFFF &&
		    p_ii->g_or_v.vend.dev_id != p_ntci->g_or_v.vend.dev_id) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Differ by Vendor.DeviceID (II:0x%X Trap:0x%X)\n",
				cl_ntoh16(p_ii->g_or_v.vend.dev_id),
				cl_ntoh16(p_ntci->g_or_v.vend.dev_id));
			goto Exit;
		}
	}

	return TRUE;

Exit:
	return FALSE;
}

int osm_infr_match_notice_to_db(IN osm_subn_t *p_subn,
				IN osm_log_t * p_log,
				IN osm_mad_addr_t *p_mad_addr,
				IN ib_mad_notice_attr_t *p_ntci,
				IN int remove)
{
	cl_list_item_t *p_list_item, *p_next_item;
	osm_infr_t *p_infr;
	int status = 1;

	OSM_LOG_ENTER(p_log);

	dump_all_informs(p_subn, p_log);

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Looking for Inform Records matching Notice\n");
	osm_dump_notice_v2(p_log, p_ntci, FILE_ID, OSM_LOG_DEBUG);
	OSM_LOG(p_log, OSM_LOG_DEBUG, "InformInfo list size %d\n",
		cl_qlist_count(&p_subn->sa_infr_list));

	p_next_item = cl_qlist_head(&p_subn->sa_infr_list);
	while (p_next_item != cl_qlist_end(&p_subn->sa_infr_list)) {
		p_list_item = p_next_item;
		p_next_item = cl_qlist_next(p_list_item);
		p_infr = (osm_infr_t *)p_list_item;

		if (match_notice(p_log, p_mad_addr, p_ntci, p_infr)) {
			if (remove)
				osm_infr_remove_from_db(p_subn, p_log, p_infr);
			else
				goto Exit;
		}
	}
	status = 0;

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}

osm_infr_t *osm_infr_get_by_rec(IN osm_subn_t const *p_subn,
				IN osm_log_t * p_log,
				IN osm_infr_t * p_infr_rec)
{
	cl_list_item_t *p_list_item;

	OSM_LOG_ENTER(p_log);

	dump_all_informs(p_subn, p_log);

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Looking for Inform Record\n");
	osm_dump_inform_info_v2(p_log, &(p_infr_rec->inform_record.inform_info),
			        FILE_ID, OSM_LOG_DEBUG);
	OSM_LOG(p_log, OSM_LOG_DEBUG, "InformInfo list size %d\n",
		cl_qlist_count(&p_subn->sa_infr_list));

	p_list_item = cl_qlist_find_from_head(&p_subn->sa_infr_list,
					      match_inf_rec, p_infr_rec);

	if (p_list_item == cl_qlist_end(&p_subn->sa_infr_list))
		p_list_item = NULL;

	OSM_LOG_EXIT(p_log);
	return (osm_infr_t *) p_list_item;
}

void osm_infr_insert_to_db(IN osm_subn_t * p_subn, IN osm_log_t * p_log,
			   IN osm_infr_t * p_infr)
{
	OSM_LOG_ENTER(p_log);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Inserting new InformInfo Record into Database\n");
	OSM_LOG(p_log, OSM_LOG_DEBUG, "Dump before insertion (size %d)\n",
		cl_qlist_count(&p_subn->sa_infr_list));
	dump_all_informs(p_subn, p_log);

#if 0
	osm_dump_inform_info_v2(p_log,
			        &(p_infr->inform_record.inform_info),
			        FILE_ID, OSM_LOG_DEBUG);
#endif

	if (p_subn->opt.sa_enhanced_trust_model) {
		osm_alias_guid_t *p_report_aguid =
			(osm_alias_guid_t*)cl_qmap_get(&p_subn->alias_port_guid_tbl,
						       ib_gid_get_guid(&p_infr->inform_record.subscriber_gid));

		OSM_ASSERT(p_report_aguid != (osm_alias_guid_t*)cl_qmap_end(&p_subn->alias_port_guid_tbl));
		/* Increase the counter of service subscriptions for the alias guid */
		p_report_aguid->event_subs_counter++;
	}

	cl_qlist_insert_head(&p_subn->sa_infr_list, &p_infr->list_item);
	p_subn->p_osm->sa.dirty = TRUE;

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Dump after insertion (size %d)\n",
		cl_qlist_count(&p_subn->sa_infr_list));
	dump_all_informs(p_subn, p_log);
	OSM_LOG_EXIT(p_log);
}

void osm_infr_remove_from_db(IN osm_subn_t * p_subn, IN osm_log_t * p_log,
			     IN osm_infr_t * p_infr)
{
	char gid_str[INET6_ADDRSTRLEN];

	OSM_LOG_ENTER(p_log);

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Removing InformInfo Subscribing GID:%s"
		" Enum:0x%X from Database\n",
		inet_ntop(AF_INET6, p_infr->inform_record.subscriber_gid.raw,
			  gid_str, sizeof gid_str),
		p_infr->inform_record.subscriber_enum);

	osm_dump_inform_info_v2(p_log, &(p_infr->inform_record.inform_info),
			        FILE_ID, OSM_LOG_DEBUG);

	if (p_subn->opt.sa_enhanced_trust_model) {
		osm_alias_guid_t *p_report_aguid =
			(osm_alias_guid_t*)cl_qmap_get(&p_subn->alias_port_guid_tbl,
						       ib_gid_get_guid(&p_infr->inform_record.subscriber_gid));

		if (p_report_aguid != (osm_alias_guid_t*)cl_qmap_end(&p_subn->alias_port_guid_tbl)) {
			OSM_ASSERT(p_report_aguid->event_subs_counter > 0);
			/* Decrease the counter of service subsciptions for the alias guid */
			p_report_aguid->event_subs_counter--;
		}
	}

	cl_qlist_remove_item(&p_subn->sa_infr_list, &p_infr->list_item);
	p_subn->p_osm->sa.dirty = TRUE;

	osm_infr_delete(p_infr);

	OSM_LOG_EXIT(p_log);
}

/**********************************************************************
 * Send a report:
 * Given a target address to send to and the notice.
 * We need to send SubnAdmReport
 **********************************************************************/
static ib_api_status_t send_report(IN osm_infr_t * p_infr_rec,	/* the informinfo */
				   IN ib_mad_notice_attr_t * p_ntc	/* notice to send */
    )
{
	osm_madw_t *p_report_madw;
	ib_mad_notice_attr_t *p_report_ntc;
	ib_mad_t *p_mad;
	ib_sa_mad_t *p_sa_mad;
	static atomic32_t trap_fwd_trans_id = 0x02DAB000;
	ib_api_status_t status = IB_SUCCESS;
	osm_log_t *p_log = p_infr_rec->sa->p_log;
	ib_net64_t tid;

	OSM_LOG_ENTER(p_log);

	/* HACK: who switches or uses the src and dest GIDs in the grh_info ?? */

	/* advance trap trans id (cant simply ++ on some systems inside ntoh) */
	tid = cl_hton64((uint64_t) cl_atomic_inc(&trap_fwd_trans_id) &
			(uint64_t) (0xFFFFFFFF));
	if (trap_fwd_trans_id == 0)
		tid = cl_hton64((uint64_t) cl_atomic_inc(&trap_fwd_trans_id) &
				(uint64_t) (0xFFFFFFFF));

	if (OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_DEBUG)) {
		if (p_infr_rec->report_addr.addr_type.gsi.global_route) {
			char gid_str[INET6_ADDRSTRLEN];
			ib_gid_t *p_gid =
			     &p_infr_rec->report_addr.addr_type.gsi.grh_info.dest_gid;

			OSM_LOG(p_log, OSM_LOG_DEBUG, "Forwarding Notice Event "
				"from LID:%u to InformInfo LID:%u DGID:%s "
				"TID:0x%X\n",
				cl_ntoh16(p_ntc->issuer_lid),
				cl_ntoh16(p_infr_rec->report_addr.dest_lid),
				inet_ntop(AF_INET6, p_gid,
					  gid_str, sizeof(gid_str)),
				trap_fwd_trans_id);
		} else {
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Forwarding Notice Event "
				"from LID:%u to InformInfo LID:%u TID:0x%X\n",
				cl_ntoh16(p_ntc->issuer_lid),
				cl_ntoh16(p_infr_rec->report_addr.dest_lid),
				trap_fwd_trans_id);
		}
	}

	/* get the MAD to send */
	p_report_madw = osm_mad_pool_get(p_infr_rec->sa->p_mad_pool,
					 p_infr_rec->h_bind, MAD_BLOCK_SIZE,
					 &(p_infr_rec->report_addr));

	if (!p_report_madw) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0203: "
			"osm_mad_pool_get failed\n");
		status = IB_ERROR;
		goto Exit;
	}

	p_report_madw->resp_expected = TRUE;

	p_mad = osm_madw_get_mad_ptr(p_report_madw);
	ib_mad_init_new(p_mad, IB_MCLASS_SUBN_ADM, 2, IB_MAD_METHOD_REPORT,
			tid, IB_MAD_ATTR_NOTICE, 0);

	p_sa_mad = osm_madw_get_sa_mad_ptr(p_report_madw);

	p_report_ntc = (ib_mad_notice_attr_t *) & (p_sa_mad->data);

	/* copy the notice */
	*p_report_ntc = *p_ntc;

	/* Queue report */
	osm_queue_report(p_infr_rec->sa, p_report_madw);

	/* Notify SA report sending thread */
	osm_inform_mgr_send_signal(p_infr_rec->sa);

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}

static int is_access_permitted(osm_infr_t *p_infr_rec,
			       osm_infr_match_ctxt_t *p_infr_match )
{
	cl_list_t *p_infr_to_remove_list = p_infr_match->p_remove_infr_list;
	ib_inform_info_t *p_ii = &(p_infr_rec->inform_record.inform_info);
	ib_mad_notice_attr_t *p_ntc = p_infr_match->p_ntc;
	uint16_t trap_num = cl_ntoh16(p_ntc->g_or_v.generic.trap_num);
	osm_subn_t *p_subn = p_infr_rec->sa->p_subn;
	osm_log_t *p_log = p_infr_rec->sa->p_log;
	osm_mgrp_t *p_mgrp;
	ib_gid_t source_gid;
	osm_alias_guid_t *p_src_aguid;
	osm_alias_guid_t *p_dest_aguid;

	/* In case of GID_IN(64) or GID_OUT(65) traps the source gid
	   comparison should be done on the trap source (saved as the gid in the
	   data details field).
	   For traps MC_CREATE(66) or MC_DELETE(67) the data details gid is
	   the MGID. We need to check whether the subscriber has a compatible
	   pkey with MC group.
	   In all other cases the issuer gid is the trap source.
	*/
	if (trap_num >= 64 && trap_num <= 67 )
		/* The issuer of these traps is the SM so source_gid
		   is the gid saved on the data details */
		source_gid = p_ntc->data_details.ntc_64_67.gid;
	else
		source_gid = p_ntc->issuer_gid;

	p_dest_aguid = osm_get_alias_guid_by_mad_addr(p_log, p_subn,
						      &p_infr_rec->report_addr);
	if (!p_dest_aguid) {
		OSM_LOG(p_log, OSM_LOG_INFO, "Cannot find destination port\n");
		goto Exit;
	}

	/* Check if there is a pkey match. o13-17.1.1 */
	switch (trap_num) {
		case 65:
		case 67:
			break;
		case 66:
			p_mgrp = osm_get_mgrp_by_mgid(p_subn, &source_gid);
			if (!p_mgrp) {
				char gid_str[INET6_ADDRSTRLEN];
				OSM_LOG(p_log, OSM_LOG_INFO,
					"Cannot find MGID %s\n",
					inet_ntop(AF_INET6, source_gid.raw, gid_str, sizeof gid_str));
				goto Exit;
			}

			if (!osm_pkey_tbl_has_pkey(p_log,
						   p_mgrp->mcmember_rec.pkey,
						   osm_alias_guid_get_pkey_tbl(p_dest_aguid),
						   p_subn->opt.allow_both_pkeys)) {
				char gid_str[INET6_ADDRSTRLEN];
				OSM_LOG(p_log, OSM_LOG_INFO,
					"MGID %s and port GUID:0x%016" PRIx64 " do not share same pkey\n",
					inet_ntop(AF_INET6, source_gid.raw, gid_str, sizeof gid_str),
					cl_ntoh64(p_dest_aguid->alias_guid));
				goto Exit;
			}
			break;

		default:
			p_src_aguid =
			    osm_get_alias_guid_by_guid(p_subn, source_gid.unicast.interface_id);
			if (!p_src_aguid) {
				OSM_LOG(p_log, OSM_LOG_INFO,
					"Cannot find source port with GUID:0x%016" PRIx64 "\n",
					cl_ntoh64(source_gid.unicast.interface_id));
				goto Exit;
			}

                       if (osm_pkey_tbl_share_pkey(p_log,
						   osm_alias_guid_get_pkey_tbl(p_src_aguid),
						   osm_alias_guid_get_pkey_tbl(p_dest_aguid),
						   p_subn->opt.allow_both_pkeys) == FALSE) {
				OSM_LOG(p_log, OSM_LOG_DEBUG, "Mismatch by Pkey\n");
				/* According to o13-17.1.2 - If this informInfo
				   does not have lid_range_begin of 0xFFFF,
				   then this informInfo request should be
				   removed from database */
				if (p_ii->lid_range_begin != 0xFFFF) {
					OSM_LOG(p_log, OSM_LOG_VERBOSE,
						"Pkey mismatch on lid_range_begin != 0xFFFF. "
						"Need to remove this informInfo from db\n");
					/* add the informInfo record to the remove_infr list */
					cl_list_insert_tail(p_infr_to_remove_list, p_infr_rec);
				}
				goto Exit;
			}
			break;
	}

	return 1;
Exit:
	return 0;
}


/**********************************************************************
 * This routine compares a given Notice and a ListItem of InformInfo type.
 * PREREQUISITE:
 * The Notice.GID should be pre-filled with the trap generator GID
 **********************************************************************/
static void match_notice_to_inf_rec(IN cl_list_item_t * p_list_item,
				    IN void *context)
{
	osm_infr_match_ctxt_t *p_infr_match = (osm_infr_match_ctxt_t *) context;
	ib_mad_notice_attr_t *p_ntc = p_infr_match->p_ntc;
	osm_infr_t *p_infr_rec = (osm_infr_t *) p_list_item;
	osm_log_t *p_log = p_infr_rec->sa->p_log;

	OSM_LOG_ENTER(p_log);

	if (!match_notice(p_log, NULL, p_ntc, p_infr_rec))
		goto Exit;

	if (!is_access_permitted(p_infr_rec, p_infr_match))
		goto Exit;

	/* send the report to the address provided in the inform record */
	OSM_LOG(p_log, OSM_LOG_DEBUG, "MATCH! Sending Report...\n");
	send_report(p_infr_rec, p_ntc);

Exit:
	OSM_LOG_EXIT(p_log);
}

/**********************************************************************
 * Once a Trap was received by osm_trap_rcv, or a Trap sourced by
 * the SM was sent (Traps 64-67), this routine is called with a copy of
 * the notice data.
 * Given a notice attribute - compare and see if it matches the InformInfo
 * element and if it does - call the Report(Notice) for the
 * target QP registered by the address stored in the InformInfo element
 **********************************************************************/
static void log_notice(osm_log_t * log, osm_log_level_t level,
		       ib_mad_notice_attr_t * ntc)
{
	char gid_str[INET6_ADDRSTRLEN];
	ib_gid_t *gid;

	/* an official Event information log */
	if (ib_notice_is_generic(ntc)) {
		if ((ntc->g_or_v.generic.trap_num == CL_HTON16(64)) ||
		    (ntc->g_or_v.generic.trap_num == CL_HTON16(65)) ||
		    (ntc->g_or_v.generic.trap_num == CL_HTON16(66)) ||
		    (ntc->g_or_v.generic.trap_num == CL_HTON16(67)))
			gid = &ntc->data_details.ntc_64_67.gid;
		else
			gid = &ntc->issuer_gid;
		OSM_LOG(log, level,
			"Reporting Generic Notice type:%u num:%u (%s)"
			" from LID:%u GID:%s\n",
			ib_notice_get_type(ntc),
			cl_ntoh16(ntc->g_or_v.generic.trap_num),
			ib_get_trap_str(ntc->g_or_v.generic.trap_num),
			cl_ntoh16(ntc->issuer_lid),
			inet_ntop(AF_INET6, gid->raw, gid_str, sizeof gid_str));
	} else
		OSM_LOG(log, level,
			"Reporting Vendor Notice type:%u vend:%u dev:%u"
			" from LID:%u GID:%s\n",
			ib_notice_get_type(ntc),
			cl_ntoh32(ib_notice_get_vend_id(ntc)),
			cl_ntoh16(ntc->g_or_v.vend.dev_id),
			cl_ntoh16(ntc->issuer_lid),
			inet_ntop(AF_INET6, ntc->issuer_gid.raw, gid_str,
				  sizeof gid_str));
}

ib_api_status_t osm_send_report(IN osm_subn_t * p_subn,
				IN ib_mad_notice_attr_t * p_ntc)
{
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_infr_match_ctxt_t context;
	cl_list_t infr_to_remove_list;
	osm_infr_t *p_infr_rec;
	osm_infr_t *p_next_infr_rec;
	boolean_t is_skip_report = TRUE;

	OSM_LOG_ENTER(p_log);

	/* Skip Notice reports if asked by the user */
	if (p_subn->opt.reports == 2 ||
	    (p_subn->opt.reports == 1 && ib_notice_is_generic(p_ntc) &&
	    ((p_ntc->g_or_v.generic.trap_num == CL_HTON16(64)) ||
	    (p_ntc->g_or_v.generic.trap_num == CL_HTON16(65)) ||
	    (p_ntc->g_or_v.generic.trap_num == CL_HTON16(66)) ||
	    (p_ntc->g_or_v.generic.trap_num == CL_HTON16(67)) ||
	    (p_ntc->g_or_v.generic.trap_num == CL_HTON16(144)))))
		is_skip_report = FALSE;

	if (is_skip_report)
		goto skip_report;

	/* Create a list that will hold all the infr records that should
	   be removed due to violation. o13-17.1.2 */
	cl_list_construct(&infr_to_remove_list);
	cl_list_init(&infr_to_remove_list, 5);
	context.p_remove_infr_list = &infr_to_remove_list;
	context.p_ntc = p_ntc;

	/* go over all inform info available at the subnet */
	/* try match to the given notice and send if match */
	cl_qlist_apply_func(&p_subn->sa_infr_list, match_notice_to_inf_rec,
			    &context);

	/* If we inserted items into the infr_to_remove_list - we need to
	   remove them */
	p_infr_rec = (osm_infr_t *) cl_list_remove_head(&infr_to_remove_list);
	while (p_infr_rec != NULL) {
		p_next_infr_rec =
		    (osm_infr_t *) cl_list_remove_head(&infr_to_remove_list);
		osm_infr_remove_from_db(p_subn, p_log, p_infr_rec);
		p_infr_rec = p_next_infr_rec;
	}
	cl_list_destroy(&infr_to_remove_list);

skip_report:
	OSM_LOG_EXIT(p_log);

	return IB_SUCCESS;
}

ib_api_status_t osm_report_notice(IN osm_log_t * p_log, IN osm_subn_t * p_subn,
				  IN ib_mad_notice_attr_t * p_ntc,
				  IN uint64_t tid, IN uint64_t trap_tid)
{
	OSM_LOG_ENTER(p_log);

	/*
	 * we must make sure we are ready for this...
	 * note that the trap receivers might be initialized before
	 * the osm_infr_init call is performed.
	 */
	if (p_subn->rcvd_traps_tbl.state != CL_INITIALIZED) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Ignoring Traps since received traps table is not initialized yet!\n");
		return IB_ERROR;
	}

	if (OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_INFO))
		log_notice(p_log, OSM_LOG_INFO, p_ntc);

	if (osm_queue_trap(&p_subn->p_osm->sa, p_ntc, tid, trap_tid))
		osm_inform_mgr_queue_signal(&p_subn->p_osm->sm);

	/* report IB traps to plugin */
	osm_opensm_report_event(p_subn->p_osm, OSM_EVENT_ID_TRAP, p_ntc);

	OSM_LOG_EXIT(p_log);

	return IB_SUCCESS;
}

ib_api_status_t osm_report_notice_64_67(osm_subn_t * p_subn, ib_gid_t * p_gid,
					uint16_t trap_num)
{

	ib_mad_notice_attr_t notice;
	ib_api_status_t status;

	OSM_ASSERT(64 <= trap_num && trap_num <= 67);

	/* details of the notice */
	notice.generic_type = 0x83; /* is generic subn mgt type */
	ib_notice_set_prod_type_ho(&notice, 4); /* A class manager generator */
	/* endport ceases to be reachable */
	notice.g_or_v.generic.trap_num = CL_HTON16(trap_num);
	/* The sm_base_lid is saved in network order already. */
	notice.issuer_lid = p_subn->sm_base_lid;

	/* following C14-72.1.2 and table 119 p725 */
	/* we need to provide the GID */
	memcpy(&(notice.data_details.ntc_64_67.gid),
	       p_gid, sizeof(ib_gid_t));

	/*
	 * According to page 653 - the issuer gid in this case of trap
	 * is the SM gid, since the SM is the initiator of this trap.
	 */
	notice.issuer_gid.unicast.prefix = p_subn->opt.subnet_prefix;
	notice.issuer_gid.unicast.interface_id = p_subn->sm_port_guid;

	status = osm_report_notice(&p_subn->p_osm->log, p_subn, &notice,
				   (uint64_t)cl_atomic_inc(&p_subn->tid) &
				   (uint64_t)(0xFFFFFFFF),
				   0);
	if (status != IB_SUCCESS) {
		char gid_str[INET6_ADDRSTRLEN];
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 0204: "
			"Error sending trap reports (%s), trap %d gid: %s\n",
			ib_get_err_str(status),
			trap_num,
			inet_ntop(AF_INET6, p_gid->raw,
				  gid_str, sizeof(gid_str)));
	}

	return status;
}

ib_api_status_t osm_report_sa_security_notice(osm_subn_t *p_subn,
					      struct _vendor_data_details *notice_data)
{
	ib_mad_notice_attr_t notice = {0};
	ib_api_status_t status;

	OSM_ASSERT(SA_VS_TRAP_NUM_MIN <= notice_data->trap_num &&
		   notice_data->trap_num <= SA_VS_TRAP_NUM_MAX);

	/* details of the notice */
	notice.generic_type = 0x02; /* vendor specific security type */

	/* The sm_base_lid is saved in network order already. */
	notice.issuer_lid = p_subn->sm_base_lid;

	/* Set Mellanox OUI VendorID */
	ib_notice_set_vend_id(&notice, CL_HTON32(OSM_VENDOR_ID_MELLANOX));

	memcpy(&notice.data_details.vendor_data_details, notice_data,
	       sizeof(struct _vendor_data_details));

	/*
	 * According to page 653 - the issuer gid in this case of trap
	 * is the SM gid, since the SM is the initiator of this trap.
	 */
	notice.issuer_gid.unicast.prefix = p_subn->opt.subnet_prefix;
	notice.issuer_gid.unicast.interface_id = p_subn->sm_port_guid;

	status = osm_report_notice(&p_subn->p_osm->log, p_subn, &notice,
				   (uint64_t)cl_atomic_inc(&p_subn->tid) &
				   (uint64_t)(0xFFFFFFFF),
				   0);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 0205: "
			"Error sending notice reports (%s), trap %d\n",
			ib_get_err_str(status),
			notice_data->trap_num);
	}

	return status;
}


void osm_report_notices_1303_1305(osm_subn_t *p_subn, osm_alias_guid_t *p_req_aguid,
				  osm_alias_guid_t *p_target_aguid, uint16_t trap_num)
{
	struct _vendor_data_details ntc_data;

	ntc_data.v_data_details.ntc_1303_1305.sgid.unicast.prefix = p_subn->opt.subnet_prefix;
	ntc_data.v_data_details.ntc_1303_1305.sgid.unicast.interface_id = p_req_aguid->alias_guid;
	ntc_data.v_data_details.ntc_1303_1305.physp_sguid = p_req_aguid->base_port_guid;
	ntc_data.v_data_details.ntc_1303_1305.target_guid = p_target_aguid->alias_guid;
	ntc_data.v_data_details.ntc_1303_1305.physp_target_guid = p_target_aguid->base_port_guid;
	ntc_data.trap_num = trap_num;

	osm_report_sa_security_notice(p_subn, &ntc_data);
}

