/*
 * Copyright (c) 2012 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_inform_mgr_t.
 * This object implements the Inform manager object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_INFORM_MGR_C
#include <opensm/osm_port.h>
#include <opensm/osm_node.h>
#include <opensm/osm_inform_mgr.h>
#include <opensm/osm_opensm.h>

/* Return value of 1 means found and 0 means not found */
typedef int (* find_match_t) (const osm_sa_t * p_sa,
			      const ib_mad_notice_attr_t *p_ntc,
			      const osm_inform_work_obj_t *p_map_obj,
			      const uint64_t trap_tid);

static int compare_issuerlid(const void *t1, const void *t2)
{
	osm_inform_work_obj_t *trap1, *trap2;

	trap1 = (* (osm_inform_work_obj_t **) t1);
	trap2 = (* (osm_inform_work_obj_t **) t2);

        /* Compare issuer LIDs */
	if (cl_ntoh16(trap1->ntc.issuer_lid) < cl_hton16(trap2->ntc.issuer_lid))
		return -1;
	if (cl_ntoh16(trap1->ntc.issuer_lid) > cl_hton16(trap2->ntc.issuer_lid))
		return 1;
	return 0;
}

static cl_fmap_item_t *find_trap(IN osm_sa_t * p_sa,
				 IN ib_mad_notice_attr_t *p_ntc,
				 IN osm_inform_work_obj_t *p_obj,
				 IN find_match_t find_match_func)
{
	cl_fmap_item_t *p_map_rem = NULL, *p_map_rem_orig = NULL;
	osm_inform_work_obj_t *p_map_obj;
	int direction = 0;

	while (p_map_rem != cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl)) {
		if (p_map_rem == NULL) {
			p_map_rem = cl_fmap_match(&p_sa->p_subn->rcvd_traps_tbl,
						  (void *) &p_obj->p_obj,
						  compare_issuerlid);
			if (p_map_rem == cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl))
				return p_map_rem;
			p_map_rem_orig = p_map_rem;
			direction = 0;		/* go backwards first */
		} else {
			if (direction) {	/* forward */
				p_map_rem = cl_fmap_next(p_map_rem);
				if (p_map_rem == cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl))
					return p_map_rem;
			} else {		/* backwards */
				p_map_rem = cl_fmap_prev(p_map_rem);
				if (p_map_rem == cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl)) {
					/* Now move forward from original item that matched */
					direction = 1;
					p_map_rem = p_map_rem_orig;
					continue;
				}
			}

			/* same issuer LID ? */
			p_map_obj = (osm_inform_work_obj_t *)p_map_rem;
			if (p_map_obj->ntc.issuer_lid != p_ntc->issuer_lid) {
				/* No need to continue past first non match of issuer LID */
				if (direction) {	/* forward */
					p_map_rem = (cl_fmap_item_t *)cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl);
					return p_map_rem;
				} else {		/* backwards */
					/* Now move forward from original item that matched */
					direction = 1;
					p_map_rem = p_map_rem_orig;
					continue;
				}
			}
		}

		/* Now, invoke find_match function */
		p_map_obj = (osm_inform_work_obj_t *)p_map_rem;
		if (find_match_func(p_sa, p_ntc, p_map_obj, p_obj->trap_tid))
			break;
	}

	return p_map_rem;
}

static int find_trap_num_match(const osm_sa_t * p_sa,
			       const ib_mad_notice_attr_t *p_ntc,
			       const osm_inform_work_obj_t *p_map_obj,
			       const uint64_t trap_tid)
{
	ib_net16_t trap_num;

	/* Check trap number and trap TID for match */
	trap_num = p_ntc->g_or_v.generic.trap_num;
	if (p_map_obj->ntc.g_or_v.generic.trap_num == trap_num &&
	    p_map_obj->trap_tid == trap_tid)
		return 1;
	return 0;
}

static int find_trap_128_match(const osm_sa_t * p_sa,
			       const ib_mad_notice_attr_t *p_ntc,
			       const osm_inform_work_obj_t *p_map_obj,
			       const uint64_t trap_tid)
{
	ib_net16_t trap_num;

	/* Just check trap number for match */
	trap_num = p_ntc->g_or_v.generic.trap_num;
	if (p_map_obj->ntc.g_or_v.generic.trap_num == trap_num)
		return 1;
	return 0;
}

static int find_trap_129_131_match(const osm_sa_t * p_sa,
				   const ib_mad_notice_attr_t *p_ntc,
				   const osm_inform_work_obj_t *p_map_obj,
				   const uint64_t trap_tid)
{
	ib_net16_t trap_num;
	uint8_t port_num;

	/* Check trap and port numbers for match */
	trap_num = p_ntc->g_or_v.generic.trap_num;
	port_num = p_ntc->data_details.ntc_129_131.port_num;
	if (p_map_obj->ntc.g_or_v.generic.trap_num == trap_num &&
	    p_map_obj->ntc.data_details.ntc_129_131.port_num == port_num)
		return 1;
	return 0;
}

static cl_fmap_item_t *find_trap_num(IN osm_sa_t * p_sa,
				     IN ib_mad_notice_attr_t *p_ntc,
				     IN osm_inform_work_obj_t *p_obj)
{
	return find_trap(p_sa, p_ntc, p_obj, &find_trap_num_match);
}

static cl_fmap_item_t *find_trap_128(IN osm_sa_t * p_sa,
				     IN ib_mad_notice_attr_t *p_ntc,
				     IN osm_inform_work_obj_t *p_obj)
{
	return find_trap(p_sa, p_ntc, p_obj, &find_trap_128_match);
}

static cl_fmap_item_t *find_trap_129_131(IN osm_sa_t * p_sa,
					 IN ib_mad_notice_attr_t *p_ntc,
					 IN osm_inform_work_obj_t *p_obj)
{
	return find_trap(p_sa, p_ntc, p_obj, &find_trap_129_131_match);
}

static cl_fmap_item_t *coalesce_traps(IN osm_sa_t * p_sa,
				      IN ib_mad_notice_attr_t *p_ntc,
				      IN osm_inform_work_obj_t *p_obj)
{
	cl_fmap_item_t *p_rem = (cl_fmap_item_t *) cl_fmap_end(&p_sa->p_subn->rcvd_traps_tbl);
	uint16_t trap_num;

	OSM_LOG_ENTER(p_sa->p_log);

	if (ib_notice_is_generic(p_ntc) == 0)
		goto Exit;

	trap_num = cl_ntoh16(p_ntc->g_or_v.generic.trap_num);

	/* Subnet Manager generated traps */
	if (trap_num < 128)
		goto Exit;

	/* For now, only coalesce "hardware" based traps */
	/* For others, make sure not duplicates */
	/* Note that currently traps 129-131 are not "reported" */
	switch (trap_num) {
	case 128:
		p_rem = find_trap_128(p_sa, p_ntc, p_obj);
		break;
	case 129:
	case 130:
	case 131:
		p_rem = find_trap_129_131(p_sa, p_ntc, p_obj);
		break;
	default:
		p_rem = find_trap_num(p_sa, p_ntc, p_obj);
		break;
	}

Exit:
	OSM_LOG_EXIT(p_sa->p_log);

	return p_rem;
}

osm_inform_work_obj_t *osm_inform_work_obj_new(IN ib_mad_notice_attr_t *p_ntc,
					       IN uint64_t tid,
					       IN uint64_t trap_tid)
{
	osm_inform_work_obj_t *p_obj;

	/*
	   clean allocated memory to avoid assertion when trying to insert to
	   qlist.
	   see cl_qlist_insert_tail(): OSM_ASSERT(p_list_item->p_list != p_list)
	*/
	p_obj = calloc(1, sizeof(*p_obj));
	if (p_obj) {
		p_obj->p_obj = p_obj;
		p_obj->ntc = *p_ntc;
		p_obj->tid = tid;
		p_obj->trap_tid = trap_tid;
	}

	return p_obj;
}

void osm_inform_work_obj_delete(IN osm_inform_work_obj_t * p_wobj)
{
	free(p_wobj);
}

int osm_queue_trap(IN osm_sa_t *sa, IN ib_mad_notice_attr_t *p_ntc,
		   IN uint64_t tid, IN uint64_t trap_tid)
{
	osm_inform_work_obj_t *p_obj, *p_obj_check;
	cl_fmap_item_t *p_item;
	int status = 1;

	/* Only queue trap if it matches some subscription */
	if (!osm_infr_match_notice_to_db(sa->p_subn, sa->p_log, NULL,
					 p_ntc, 0)) {
		status = 0;
		goto Exit;
	}

	p_obj = osm_inform_work_obj_new(p_ntc, tid, trap_tid);
	if (!p_obj) {
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 520F: "
			"Memory allocation of inform work object failed\n");
		status = 0;
		goto Exit;
	}

	cl_spinlock_acquire(&sa->p_inform_mgr->traps_lock);
	p_item = coalesce_traps(sa, p_ntc, p_obj);
	/* Already found in fmap ? */
	if (p_item != cl_fmap_end(&sa->p_subn->rcvd_traps_tbl)) {
		cl_spinlock_release(&sa->p_inform_mgr->traps_lock);
		osm_inform_work_obj_delete(p_obj);
		status = 0;
		goto Exit;
	}

	p_obj_check =
		(osm_inform_work_obj_t *) cl_fmap_insert(&sa->p_subn->rcvd_traps_tbl,
							(void *) &p_obj->p_obj,
							 &p_obj->map_item);
	if (p_obj_check != p_obj) {
		/* duplicate */
		cl_spinlock_release(&sa->p_inform_mgr->traps_lock);
		OSM_LOG(sa->p_log, OSM_LOG_DEBUG,
			"Trap is already in table\n");
		osm_inform_work_obj_delete(p_obj);
		status = 0;
	} else {
		cl_qlist_insert_tail(&sa->p_subn->rcvd_traps_list,
				     &p_obj->list_item);
		cl_spinlock_release(&sa->p_inform_mgr->traps_lock);
	}

Exit:
	return status;
}

void osm_flush_traps(IN osm_subn_t *p_subn)
{
	osm_sa_t *sa = &p_subn->p_osm->sa;
	osm_inform_work_obj_t *p_obj;
	cl_list_item_t * item;

	cl_spinlock_acquire(&sa->p_inform_mgr->traps_lock);

	while (cl_qlist_count(&p_subn->rcvd_traps_list)) {
		item = cl_qlist_remove_head(&p_subn->rcvd_traps_list);
		p_obj = cl_item_obj(item, p_obj, list_item);
		cl_fmap_remove_item(&p_subn->rcvd_traps_tbl,

				    &p_obj->map_item);
		osm_inform_work_obj_delete(p_obj);
	}

	cl_spinlock_release(&sa->p_inform_mgr->traps_lock);
}

void osm_queue_report(IN osm_sa_t *sa, IN osm_madw_t *p_report_madw)
{
	cl_spinlock_acquire(&sa->p_inform_mgr->reports_lock);
	cl_atomic_inc(&sa->p_subn->p_osm->stats.sa_reports_queued);
	cl_qlist_insert_tail(&sa->p_subn->sa_reports_list,
			     &p_report_madw->list_item);
	cl_spinlock_release(&sa->p_inform_mgr->reports_lock);
}

void osm_flush_reports(IN osm_subn_t *p_subn)
{
	osm_sa_t *sa = &p_subn->p_osm->sa;
	osm_madw_t *p_madw;
	int32_t count = 0;

	cl_spinlock_acquire(&sa->p_inform_mgr->reports_lock);

	while (cl_qlist_count(&p_subn->sa_reports_list)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_subn->sa_reports_list);
		osm_mad_pool_put(sa->p_mad_pool, p_madw);
		count++;
	}

	if (count)
		cl_atomic_sub(&sa->p_subn->p_osm->stats.sa_reports_queued, count);

	cl_spinlock_release(&sa->p_inform_mgr->reports_lock);
}

void osm_inform_mgr_send_signal(IN osm_sa_t *sa)
{
	if (!sa->p_subn->opt.max_sa_reports_on_wire ||
	    sa->p_subn->p_osm->stats.sa_mads_outstanding <
	    (int32_t) sa->p_subn->opt.max_sa_reports_on_wire)
		cl_event_signal(&sa->p_inform_mgr->signal);
}

void osm_inform_mgr_queue_signal(IN osm_sm_t *sm)
{
	if (!sm->p_subn->opt.max_sa_reports_queued ||
	    sm->p_subn->p_osm->stats.sa_reports_queued <
	    (int32_t) sm->p_subn->opt.max_sa_reports_queued)
		osm_sm_signal(sm, OSM_SIGNAL_INFORM_PROCESS_REQUEST);
}

void osm_inform_mgr_process(IN osm_sm_t * sm)
{
	cl_list_item_t *item, *next_item;
	osm_inform_work_obj_t *p_obj;
	struct osm_inform_mgr *p_inform_mgr = sm->p_subn->p_osm->sa.p_inform_mgr;

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_ACQUIRE(sm->p_lock);
	cl_spinlock_acquire(&p_inform_mgr->traps_lock);

	next_item = cl_qlist_head(&sm->p_subn->rcvd_traps_list);
	while (next_item != cl_qlist_end(&sm->p_subn->rcvd_traps_list)) {

		if (osm_exit_flag)
			break;

		/* Only proceed if under SA report limit */
		if (sm->p_subn->opt.max_sa_reports_queued &&
		    sm->p_subn->p_osm->stats.sa_reports_queued >=
		    (int32_t) sm->p_subn->opt.max_sa_reports_queued)
			break;

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Processing trap list/table\n");

		item = next_item;
		/* First remove from qlist */
		next_item = cl_qlist_next(item);
		p_obj = cl_item_obj(item, p_obj, list_item);
		cl_qlist_remove_item(&sm->p_subn->rcvd_traps_list, item);
		/* Now remove from fmap */
		cl_fmap_remove_item(&sm->p_subn->rcvd_traps_tbl,
				    &p_obj->map_item);
		osm_send_report(sm->p_subn, &p_obj->ntc);
		osm_inform_work_obj_delete(p_obj);
	}

	cl_spinlock_release(&p_inform_mgr->traps_lock);
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}

static void report_mgr_process(IN void *p_ptr)
{
	osm_sm_t *sm = p_ptr;
	osm_sa_t *sa = &sm->p_subn->p_osm->sa;
	cl_list_item_t *item, *next_item;
	osm_madw_t *p_madw;
	int32_t count;
	cl_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	while (1) {

		status = cl_event_wait_on(&sa->p_inform_mgr->signal,
					  EVENT_NO_TIMEOUT, TRUE);
		if (status != CL_SUCCESS) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 520E: Event wait failed (%s)\n",
				CL_STATUS_MSG(status));
			if (osm_exit_flag)
				break;
			continue;
		}

		if (osm_exit_flag)
			break;

		count = 0;

		/* In heavy sweep or SM port down ? */
		if (sm->p_subn->in_heavy_sweep ||
		    sm->p_subn->sm_state != IB_SMINFO_STATE_MASTER)
			continue;

		/* Make sure not spurious wakeup */
		if (sm->p_subn->opt.max_sa_reports_on_wire &&
		    sm->p_subn->p_osm->stats.sa_mads_outstanding >=
		    (int32_t) sm->p_subn->opt.max_sa_reports_on_wire)
			continue;

		CL_PLOCK_ACQUIRE(sm->p_lock);
		cl_spinlock_acquire(&sa->p_inform_mgr->reports_lock);

		next_item = cl_qlist_head(&sm->p_subn->sa_reports_list);
		while (next_item != cl_qlist_end(&sm->p_subn->sa_reports_list)) {
			if (osm_exit_flag)
				break;

			if (sm->p_subn->in_heavy_sweep ||
			    sm->p_subn->sm_state != IB_SMINFO_STATE_MASTER)
				break;

			/* Enforce SA report limit outstanding */
			if (sm->p_subn->opt.max_sa_reports_on_wire &&
			    sm->p_subn->p_osm->stats.sa_mads_outstanding >=
			    (int32_t) sm->p_subn->opt.max_sa_reports_on_wire)
				break;

			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Processing SA report list\n");

			item = next_item;
			next_item = cl_qlist_next(item);
			p_madw = cl_item_obj(item, p_madw, list_item);
			cl_qlist_remove_item(&sm->p_subn->sa_reports_list, item);
			count++;

			/* The TRUE is for: response is expected */
			osm_sa_send(sa, p_madw, TRUE);
		}

		if (count) {
			cl_atomic_sub(&sm->p_subn->p_osm->stats.sa_reports_queued,
				      count);
			osm_inform_mgr_queue_signal(sm);
		}

		cl_spinlock_release(&sa->p_inform_mgr->reports_lock);
		CL_PLOCK_RELEASE(sm->p_lock);

	}

	OSM_LOG_EXIT(sm->p_log);
}

struct osm_inform_mgr *osm_inform_mgr_construct()
{
	struct osm_inform_mgr *p_inform_mgr;

	p_inform_mgr = calloc(1, sizeof(*p_inform_mgr));
	if (p_inform_mgr) {
		cl_event_construct(&p_inform_mgr->signal);
		cl_spinlock_construct(&p_inform_mgr->traps_lock);
		cl_spinlock_construct(&p_inform_mgr->reports_lock);
		p_inform_mgr->state = OSM_INFORM_MGR_STATE_INIT;
		p_inform_mgr->thread_state = OSM_THREAD_STATE_NONE;
		cl_thread_construct(&p_inform_mgr->poller);
	}
	return p_inform_mgr;
}

ib_api_status_t osm_inform_mgr_init(IN struct osm_inform_mgr *p_inform_mgr,
				    IN osm_sa_t *p_sa, IN osm_log_t *p_log)
{
	cl_status_t status;

	OSM_LOG_ENTER(p_log);

	p_inform_mgr->p_sa = p_sa;
	p_inform_mgr->p_log = p_log;

	status = cl_event_init(&p_inform_mgr->signal, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_spinlock_init(&p_inform_mgr->traps_lock);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_spinlock_init(&p_inform_mgr->reports_lock);
	if (status != CL_SUCCESS)
		goto Exit;

	p_inform_mgr->state = OSM_INFORM_MGR_STATE_READY;

	/*
	   Initialize the thread after all other dependent objects
	   have been initialized.
	 */
	status = cl_thread_init(&p_inform_mgr->poller, report_mgr_process,
				p_sa->sm, "SA reporter");

Exit:
	OSM_LOG_EXIT(p_log);
	return (status == CL_SUCCESS) ? IB_SUCCESS : IB_UNKNOWN_ERROR;
}

void osm_inform_mgr_destroy(IN struct osm_inform_mgr *p_inform_mgr)
{
	OSM_LOG_ENTER(p_inform_mgr->p_log);

	/* Signal our thread that we're leaving */
	p_inform_mgr->thread_state = OSM_THREAD_STATE_EXIT;

	/*
	   Don't trigger unless event has been initialized.
	   Destroy the thread before we tear down the other objects.
	 */
	if (p_inform_mgr->state != OSM_INFORM_MGR_STATE_INIT)
		cl_event_signal(&p_inform_mgr->signal);

	cl_thread_destroy(&p_inform_mgr->poller);

	cl_event_destroy(&p_inform_mgr->signal);
	p_inform_mgr->state = OSM_INFORM_MGR_STATE_INIT;
	cl_spinlock_destroy(&p_inform_mgr->reports_lock);
	cl_spinlock_destroy(&p_inform_mgr->traps_lock);

	OSM_LOG_EXIT(p_inform_mgr->p_log);
}

void osm_inform_mgr_shutdown(IN struct osm_inform_mgr *p_inform_mgr,
			     IN osm_subn_t *p_subn)
{
	OSM_LOG_ENTER(p_inform_mgr->p_log);

	/* we only should get here after the inform mgr was initialized */
	OSM_ASSERT(p_inform_mgr->state == OSM_INFORM_MGR_STATE_READY);

	osm_flush_reports(p_subn);
	osm_flush_traps(p_subn);

	OSM_LOG_EXIT(p_inform_mgr->p_log);
}
