/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2011 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2008 Xsigo Systems Inc.  All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2010 HNR Consulting. All rights reserved.
 * Copyright (C) 2012-2013 Tokyo Institute of Technology. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_mcast_mgr_t.
 * This file implements the Multicast Manager object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <arpa/inet.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MCAST_MGR_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_multicast.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_mcast_mgr.h>

extern int wait_for_pending_transactions(osm_stats_t * stats);

typedef struct list_array_obj {
	cl_qlist_t array_obj;
	uint8_t index;
} list_array_obj_t;

static osm_mcast_work_obj_t *mcast_work_obj_new(IN osm_port_t * p_port)
{
	osm_mcast_work_obj_t *p_obj;

	/*
	   clean allocated memory to avoid assertion when trying to insert to
	   qlist.
	   see cl_qlist_insert_tail(): OSM_ASSERT(p_list_item->p_list != p_list)
	 */
	p_obj = malloc(sizeof(*p_obj));
	if (p_obj) {
		memset(p_obj, 0, sizeof(*p_obj));
		p_obj->p_port = p_port;
	}

	return p_obj;
}

static void mcast_work_obj_delete(IN osm_mcast_work_obj_t * p_wobj)
{
	free(p_wobj);
}

void osm_mcast_drop_port_list(cl_qlist_t * list)
{
	while (cl_qlist_count(list))
		mcast_work_obj_delete((osm_mcast_work_obj_t *)
				      cl_qlist_remove_head(list));
}

int osm_mcast_make_port_list_and_map(cl_qlist_t * list, cl_qmap_t * map,
				     osm_mgrp_box_t * mbox)
{
	cl_map_item_t *map_item, *wobj_map_item;
	cl_list_item_t *list_item;
	osm_mgrp_t *mgrp;
	osm_mcm_port_t *mcm_port;
	osm_mcast_work_obj_t *wobj;
	boolean_t has_receivers = FALSE;

	cl_qmap_init(map);
	cl_qlist_init(list);

	for (list_item = cl_qlist_head(&mbox->mgrp_list);
	     list_item != cl_qlist_end(&mbox->mgrp_list);
	     list_item = cl_qlist_next(list_item)) {
		mgrp = cl_item_obj(list_item, mgrp, list_item);
		for (map_item = cl_qmap_head(&mgrp->mcm_port_tbl);
		     map_item != cl_qmap_end(&mgrp->mcm_port_tbl);
		     map_item = cl_qmap_next(map_item)) {
			/* Acquire the port object for this port guid, then
			   create the new worker object to build the list. */
			mcm_port = cl_item_obj(map_item, mcm_port, map_item);
			wobj_map_item = cl_qmap_get(map, mcm_port->port->guid);
			if (wobj_map_item != cl_qmap_end(map)) {
				wobj = cl_item_obj(wobj_map_item, wobj,
						   map_item);
			} else {
				wobj = mcast_work_obj_new(mcm_port->port);
				if (!wobj)
					return -1;
				cl_qlist_insert_tail(list, &wobj->list_item);
				cl_qmap_insert(map, mcm_port->port->guid,
					       &wobj->map_item);
			}

			if (mcm_port->num_receivers) {
				wobj->is_receiver = 1;
				has_receivers = TRUE;
			}
		}
	}

	/* If there are no receivers, we clear the list */
	if (!has_receivers)
		osm_mcast_drop_port_list(list);

	return 0;
}

void osm_purge_mtree(osm_subn_t * p_subn, IN osm_mgrp_box_t * mbox)
{
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	if (mbox->root)
		osm_mtree_destroy(mbox->root);
	mbox->root = NULL;

	OSM_LOG_EXIT(p_log);
}

static void create_mgrp_switch_map(cl_qmap_t * m, cl_qlist_t * port_list)
{
	osm_mcast_work_obj_t *wobj;
	osm_port_t *port;
	osm_switch_t *sw;
	ib_net64_t guid;
	cl_list_item_t *i;

	cl_qmap_init(m);
	for (i = cl_qlist_head(port_list); i != cl_qlist_end(port_list);
	     i = cl_qlist_next(i)) {
		wobj = cl_item_obj(i, wobj, list_item);
		port = wobj->p_port;
		if (port->p_node->sw) {
			sw = port->p_node->sw;
			sw->is_mc_member = 1;
		} else if (port->p_physp->p_remote_physp) {
				sw = port->p_physp->p_remote_physp->p_node->sw;
				sw->num_of_mcm++;
		} else
			continue;
		guid = osm_node_get_node_guid(sw->p_node);
		if (cl_qmap_get(m, guid) == cl_qmap_end(m))
			cl_qmap_insert(m, guid, &sw->mgrp_item);
	}
}

static void destroy_mgrp_switch_map(cl_qmap_t * m)
{
	osm_switch_t *sw;
	cl_map_item_t *i;

	for (i = cl_qmap_head(m); i != cl_qmap_end(m); i = cl_qmap_next(i)) {
		sw = cl_item_obj(i, sw, mgrp_item);
		sw->num_of_mcm = 0;
		sw->is_mc_member = 0;
	}
	cl_qmap_remove_all(m);
}

/**********************************************************************
 Calculate the maximal "min hops" from the given switch to any
 of the group HCAs
 **********************************************************************/
#ifdef OSM_VENDOR_INTF_ANAFA
static uint32_t mcast_mgr_compute_avg_hops(osm_subn_t * p_subn, cl_qmap_t * m,
					const osm_switch_t * this_sw)
{
	uint32_t avg_hops = 0;
	uint32_t hops = 0;
	uint32_t num_ports = 0;
	uint16_t lid;
	uint32_t least_hops;
	cl_map_item_t *i;
	osm_switch_t *sw;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	for (i = cl_qmap_head(m); i != cl_qmap_end(m); i = cl_qmap_next(i)) {
		sw = cl_item_obj(i, sw, mcast_item);
		lid = cl_ntoh16(osm_node_get_base_lid(sw->p_node, 0));
		least_hops = osm_switch_get_least_hops(this_sw, lid);
		/* for all host that are MC members and attached to the switch,
		   we should add the (least_hops + 1) * number_of_such_hosts.
		   If switch itself is in the MC, we should add the least_hops only */
		hops += (least_hops + 1) * sw->num_of_mcm +
		    least_hops * sw->is_mc_member;
		num_ports += sw->num_of_mcm + sw->is_mc_member;
	}

	/* We shouldn't be here if there aren't any ports in the group. */
	OSM_ASSERT(num_ports);

	avg_hops = (hops * 1000 / num_ports);

	OSM_LOG_EXIT(p_log);
	return avg_hops;
}
#else
static uint32_t mcast_mgr_compute_max_hops(osm_subn_t * p_subn, cl_qmap_t * m,
					const osm_switch_t * this_sw)
{
	uint32_t max_hops = 0, hops;
	uint16_t lid;
	cl_map_item_t *i;
	osm_switch_t *sw;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/*
	   For each member of the multicast group, compute the
	   number of hops to its base LID.
	 */
	for (i = cl_qmap_head(m); i != cl_qmap_end(m); i = cl_qmap_next(i)) {
		sw = cl_item_obj(i, sw, mgrp_item);
		lid = cl_ntoh16(osm_node_get_base_lid(sw->p_node, 0));
		hops = osm_switch_get_least_hops(this_sw, lid);
		if (!sw->is_mc_member)
			hops += 1;
		if (hops > max_hops)
			max_hops = hops;
	}

	/* Note that at this point we might get (max_hops == 0),
	   which means that there's only one member in the mcast
	   group, and it's the current switch */

	OSM_LOG_EXIT(p_log);
	return max_hops;
}
#endif

/**********************************************************************
   This function attempts to locate the optimal switch for the
   center of the spanning tree.  The current algorithm chooses
   a switch with the lowest average hop count to the members
   of the multicast group.
**********************************************************************/
static osm_switch_t *mcast_mgr_find_optimal_switch(osm_subn_t * p_subn,
						   cl_qmap_t * p_mgrp_sw_map,
						   osm_mgrp_box_t * mbox)
{
	cl_qmap_t *p_sw_tbl;
	osm_switch_t *p_sw, *p_best_sw = NULL;
	uint32_t hops = 0;
	uint32_t best_hops = 10000;	/* any big number will do */
	uint32_t max_mlid_root_num = 0xFFFFFFFF;
	osm_switch_t *p_root_sw;
	uint32_t root_hops = 0;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	p_root_sw = osm_get_switch_by_guid(p_subn, cl_hton64(mbox->root_guid));

	p_sw_tbl = &p_subn->sw_guid_tbl;

	for (p_sw = (osm_switch_t *) cl_qmap_head(p_sw_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(p_sw_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
		if (!osm_switch_supports_mcast(p_sw))
			continue;

#ifdef OSM_VENDOR_INTF_ANAFA
		hops = mcast_mgr_compute_avg_hops(p_subn, p_mgrp_sw_map, p_sw);
#else
		hops = mcast_mgr_compute_max_hops(p_subn, p_mgrp_sw_map, p_sw);
#endif

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Switch 0x%016" PRIx64 " mlid=0x%x hops=%u, MC root count %d \n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			mbox->mlid, hops, p_sw->mlid_root_num);

		if (p_sw == p_root_sw)
			root_hops = hops;

		if (hops < best_hops ||
				((hops == best_hops) && (p_sw->mlid_root_num < max_mlid_root_num))) {
			p_best_sw = p_sw;
			best_hops = hops;
			max_mlid_root_num = p_sw->mlid_root_num;
		}
	}

	/* keep old root switch if number of hops is the same as new best switch */
	if (p_root_sw && (root_hops == best_hops))
		p_best_sw = p_root_sw;

	if (p_best_sw) {
		mbox->root_guid = cl_ntoh64(osm_node_get_node_guid(p_best_sw->p_node));
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Switch 0x%016" PRIx64 " (%s) mlid = 0x%x hops = %u, MC root count %d \n",
			mbox->root_guid, p_best_sw->p_node->print_desc,
			mbox->mlid, best_hops, p_best_sw->mlid_root_num);
	}
	else
		OSM_LOG(p_log, OSM_LOG_INFO,
			"No multicast capable switches detected for mlid 0x%x\n", mbox->mlid);

	OSM_LOG_EXIT(p_log);
	return p_best_sw;
}

/**********************************************************************
 *   The function checks whether the existing mtree root is still valid
 *     by checking that all the MC group members are reachable from it.
 **********************************************************************/
static boolean_t mcast_mgr_is_valid_root(cl_qmap_t * p_mgrp_sw_map,
                                         osm_switch_t * p_sw)
{
	cl_map_item_t *i;
	osm_switch_t *sw;
	uint32_t hops;
	uint16_t lid;
	boolean_t result = TRUE;

	/* Checking if all MC group members are reachable from the p_sw */
	for (i = cl_qmap_head(p_mgrp_sw_map); i != cl_qmap_end(p_mgrp_sw_map);
	     i = cl_qmap_next(i)) {
		sw = cl_item_obj(i, sw, mgrp_item);
		lid = cl_ntoh16(osm_node_get_base_lid(sw->p_node, 0));
		hops = osm_switch_get_least_hops(p_sw, lid);

		if (hops >= OSM_NO_PATH) {
			result = FALSE;
			goto Exit;
		}
	}

Exit:
	return result;
}

/**********************************************************************
  This function checks whether a switch is a valid mtree root by
  checking that every member has a LID route toward the switch.
**********************************************************************/
static boolean_t mcast_mgr_is_valid_root_by_lid(cl_qmap_t * p_mgrp_sw_map,
						osm_switch_t * p_sw)
{
	cl_map_item_t *i;
	osm_switch_t *sw;
	uint16_t lid = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));

	for (i = cl_qmap_head(p_mgrp_sw_map);
	     i != cl_qmap_end(p_mgrp_sw_map);
	     i = cl_qmap_next(i)) {
		sw = cl_item_obj(i, sw, mgrp_item);
		if (osm_switch_get_port_by_lid(sw, lid, OSM_NEW_LFT) ==
		    OSM_NO_PATH)
			return FALSE;
	}

	return TRUE;
}

/* Find valid root from MC roots table */
static osm_switch_t * mcast_mgr_find_mc_root_by_mgid(osm_subn_t * p_subn,
						     ib_gid_t * p_mgid,
						     cl_qmap_t * p_mgrp_sw_map)
{
	osm_mc_roots_list_t *p_mc_roots_list;
	cl_fmap_item_t *p_item;
	cl_list_iterator_t list_iter;
	osm_switch_t *p_sw;

	p_item = cl_fmap_get(&p_subn->mc_roots_tbl, p_mgid);
	if (p_item != cl_fmap_end(&p_subn->mc_roots_tbl)) {
		p_mc_roots_list = (osm_mc_roots_list_t *)p_item;
		for (list_iter = cl_list_head(&p_mc_roots_list->switch_list);
		     list_iter != cl_list_end(&p_mc_roots_list->switch_list);
		     list_iter = cl_list_next(list_iter)) {
			p_sw = (osm_switch_t *)cl_list_obj(list_iter);
		    	if (mcast_mgr_is_valid_root(p_mgrp_sw_map, p_sw))
				return p_sw;
		}
	}

	return NULL;
}

/*
 * This function searches MC roots table for root switch for the tree.
 */
osm_switch_t * mcast_mgr_find_root_mc_roots_tbl(osm_subn_t * p_subn,
						cl_qmap_t * p_mgrp_sw_map,
						osm_mgrp_box_t *mbox)
{
	cl_list_item_t *p_item;
	osm_mgrp_t *p_mgrp;
	osm_switch_t *p_sw = NULL;
	osm_log_t *p_log = &p_subn->p_osm->log;
	static ib_gid_t mgid0 = {{ 0 }};
	char str[INET6_ADDRSTRLEN];

	/*
 	 * Check if there is a valid root switch that was defined as prefered
 	 * root for one of groups in this mbox.
 	 */
	for (p_item = cl_qlist_head(&mbox->mgrp_list);
	     p_item != cl_qlist_end(&mbox->mgrp_list);
	     p_item = cl_qlist_next(p_item)) {
		p_mgrp = cl_item_obj(p_item, p_mgrp, list_item);
		p_sw = mcast_mgr_find_mc_root_by_mgid(p_subn,
						      &p_mgrp->mcmember_rec.mgid,
						      p_mgrp_sw_map);
		if (p_sw) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Using MC root switch 0x%" PRIx64 " of MGID %s "
				"as root for MLID 0x%X\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
                        	inet_ntop(AF_INET6,
					  &p_mgrp->mcmember_rec.mgid.raw, str,
					  sizeof(str)),
				mbox->mlid);
			goto Exit;
		}
	}

	/*
	 * Check if there is a valid root switch that was defined as default
	 * root for fabric.
	 */
	p_sw = mcast_mgr_find_mc_root_by_mgid(p_subn, &mgid0, p_mgrp_sw_map);
	if (p_sw) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Using default MC root switch 0x%" PRIx64 " as root "
			"for MLID 0x%X\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			mbox->mlid);
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return p_sw;
}

/**********************************************************************
   This function returns the existing or optimal root switch for the tree.
**********************************************************************/
osm_switch_t * osm_mcast_mgr_find_root_switch_map(osm_subn_t * p_subn,
						  cl_qmap_t * p_mgrp_sw_map,
						  osm_mgrp_box_t *mbox)
{
	osm_switch_t *p_sw = NULL;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/*
	 * Check if there is a valid root switch that was defined as prefered
	 * root for one of groups in this mbox.
	 */
	p_sw = mcast_mgr_find_root_mc_roots_tbl(p_subn, p_mgrp_sw_map, mbox);

	if (!p_sw)
		/*
		 * We always look for the best multicast tree root switch.
		 * Otherwise since we always start with a a single join
		 * the root will be always on the first switch attached to it.
		 * - Very bad ...
		 */
		p_sw = mcast_mgr_find_optimal_switch(p_subn, p_mgrp_sw_map,
						     mbox);

	OSM_LOG_EXIT(p_log);
	return p_sw;
}

osm_switch_t * osm_mcast_mgr_find_root_switch(osm_subn_t * p_subn,
					      cl_qlist_t * list,
					      osm_mgrp_box_t *mbox)
{
	osm_switch_t *p_sw;
	cl_qmap_t mgrp_sw_map;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	create_mgrp_switch_map(&mgrp_sw_map, list);
	p_sw = osm_mcast_mgr_find_root_switch_map(p_subn, &mgrp_sw_map, mbox);
	destroy_mgrp_switch_map(&mgrp_sw_map);

	OSM_LOG_EXIT(p_log);
	return p_sw;
}

static int mcast_mgr_set_mft_block(IN osm_sm_t * sm,
				   IN osm_switch_t * p_sw,
				   IN uint32_t block_num,
				   IN uint32_t position,
				   IN boolean_t clear_removed_entries,
				   OUT osm_madw_t ** p_madw)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	osm_madw_context_t context;
	uint32_t block_id_ho;
	osm_mcast_tbl_t *p_tbl;
	ib_net16_t block[IB_MCAST_BLOCK_SIZE];
	int res = 0;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_sw);

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_path = osm_physp_get_dr_path_ptr(p_physp);

	/*
	   Send multicast forwarding table blocks to the switch
	   as long as the switch indicates it has blocks needing
	   configuration.
	 */

	context.mft_context.node_guid = osm_node_get_node_guid(p_node);
	context.mft_context.set_method = TRUE;

	*p_madw = NULL;
	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);


	if (!osm_mcast_tbl_get_block(p_tbl, (uint16_t) block_num,
				     (uint8_t) position,
				     (clear_removed_entries ?
				      OSM_MCAST_TBL_UNCHANGED_ENTRIES:
				      OSM_MCAST_TBL_ALL_ENTRIES),
				     block))
		goto exit;

	/*
	  Send MFT block only if one of the following persists:
	  1) There is a connectiviry change in the switch.
	  2) This is the first time, the block is sent.
	  3) New block is different than the previous time the
	     block was sent.
	 */
	if ((p_sw->need_update != 0 && !clear_removed_entries) ||
	    (p_tbl->need_update_mask[block_num] & 1 << position) ||
	    !osm_mcast_tbl_cmpr_block(p_tbl, (uint16_t) block_num,
				      (uint8_t) position, block)) {


		p_tbl->need_update_mask[block_num] |= (1 << position);

		block_id_ho = block_num + (position << 28);

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Writing MFT block %u position %u to switch 0x%" PRIx64
			"\n", block_num, position,
			cl_ntoh64(context.mft_context.node_guid));

		*p_madw = osm_prepare_req_set(sm, p_path, (void *)block, sizeof(block),
				     IB_MAD_ATTR_MCAST_FWD_TBL,
				     cl_hton32(block_id_ho), FALSE,
				     ib_port_info_get_m_key(&p_physp->port_info),
				     0, CL_DISP_MSGID_NONE, &context);
		if (*p_madw == NULL) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A02: "
				"Out of memory\n");
			res = -1;
		}
	}

exit:
	OSM_LOG_EXIT(sm->p_log);
	return res;
}

/**********************************************************************
  This is part of the recursive function to compute the paths in the
  spanning tree that emanate from this switch.  On input, the p_list
  contains the group members that must be routed from this switch.
**********************************************************************/
static void mcast_mgr_subdivide(osm_subn_t * p_subn, osm_mgrp_box_t * mbox,
				osm_switch_t * p_sw, cl_qlist_t * p_list,
				cl_qlist_t * list_array, uint8_t array_size)
{
	uint8_t port_num;
	boolean_t ignore_existing;
	osm_mcast_work_obj_t *p_wobj;
	osm_log_t *p_log = &p_subn->p_osm->log;
	uint16_t mlid_ho = mbox->mlid;

	OSM_LOG_ENTER(p_log);

	/*
	   For Multicast Groups, we don't want to count on previous
	   configurations - since we can easily generate a storm
	   by loops.
	 */
	ignore_existing = TRUE;

	/*
	   Subdivide the set of ports into non-overlapping subsets
	   that will be routed to other switches.
	 */
	while ((p_wobj =
		(osm_mcast_work_obj_t *) cl_qlist_remove_head(p_list)) !=
	       (osm_mcast_work_obj_t *) cl_qlist_end(p_list)) {
		port_num =
		    osm_switch_recommend_mcast_path(p_sw, p_wobj->p_port,
						    mlid_ho, ignore_existing);
		if (port_num == OSM_NO_PATH) {
			/*
			   This typically occurs if the switch does not support
			   multicast and the multicast tree must branch at this
			   switch.
			 */
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A03: "
				"Error routing MLID 0x%X through switch 0x%"
				PRIx64 " %s\n"
				"\t\t\t\tNo multicast paths from this switch "
				"for port with LID %u\n", mlid_ho,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
				p_sw->p_node->print_desc,
				cl_ntoh16(osm_port_get_base_lid
					  (p_wobj->p_port)));
			mcast_work_obj_delete(p_wobj);
			mbox->num_failed_routes++;
			continue;
		}

		if (port_num >= array_size) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A04: "
				"Error routing MLID 0x%X through switch 0x%"
				PRIx64 " %s\n"
				"\t\t\t\tNo multicast paths from this switch "
				"to port with LID %u\n", mlid_ho,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
				p_sw->p_node->print_desc,
				cl_ntoh16(osm_port_get_base_lid
					  (p_wobj->p_port)));
			mcast_work_obj_delete(p_wobj);
			mbox->num_failed_routes++;
			continue;
		}

		cl_qlist_insert_tail(&list_array[port_num], &p_wobj->list_item);
	}

	OSM_LOG_EXIT(p_log);
}

static void mcast_mgr_purge_list(osm_log_t * p_log, osm_mgrp_box_t * mbox,
				 cl_qlist_t * list)
{
	if (OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_ERROR)) {
		osm_mcast_work_obj_t *wobj;
		cl_list_item_t *i;
		for (i = cl_qlist_head(list); i != cl_qlist_end(list);
		     i = cl_qlist_next(i)) {
			wobj = cl_item_obj(i, wobj, list_item);
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A06: "
				"Unable to route MLID 0x%x to port 0x%" PRIx64
				" \n",
				mbox->mlid,
				cl_ntoh64(osm_port_get_guid(wobj->p_port)));
		}
	}

	mbox->num_failed_routes += cl_qlist_count(list);
	osm_mcast_drop_port_list(list);
}

static void mcast_mgr_subdivide_const(osm_subn_t * p_subn, osm_mgrp_box_t * mbox,
				osm_mtree_node_t * p_mtn, cl_qlist_t * p_list,
				list_array_obj_t * list_array, uint8_t array_size)
{
	uint8_t used_port_index = 0;
	uint8_t index;
	uint8_t port_num;
	uint8_t * new_port_order;
	boolean_t ignore_existing;
	osm_mcast_work_obj_t *p_wobj;
	osm_switch_t *p_sw = p_mtn->p_sw;
	osm_log_t *p_log = &p_subn->p_osm->log;
	uint16_t mlid_ho = mbox->mlid;

	OSM_LOG_ENTER(p_log);

	/* Init new port_order arrat */
	new_port_order = (uint8_t*)malloc(array_size * sizeof(uint8_t));

	if (!new_port_order) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A16: "
                        "Unable to allocate ports array\n");
		mcast_mgr_purge_list(p_log, mbox, p_list);
		return;
	}

	for (index = 0; index<array_size; index++)
		new_port_order[index] = OSM_NO_PATH;

	/*
	   For Multicast Groups, we don't want to count on previous
	   configurations - since we can easily generate a storm
	   by loops.
	 */
	ignore_existing = TRUE;

	/*
	   Subdivide the set of ports into non-overlapping subsets
	   that will be routed to other switches.
	 */
	while ((p_wobj =
		(osm_mcast_work_obj_t *) cl_qlist_remove_head(p_list)) !=
	       (osm_mcast_work_obj_t *) cl_qlist_end(p_list)) {
		port_num =
		    osm_switch_recommend_mcast_path_const(p_mtn, p_wobj->p_port,
						    mlid_ho, ignore_existing);
		if (port_num == OSM_NO_PATH) {
			/*
			   This typically occurs if the switch does not support
			   multicast and the multicast tree must branch at this
			   switch.
			 */
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A03: "
				"Error routing MLID 0x%X through switch 0x%"
				PRIx64 " %s\n"
				"\t\t\t\tNo multicast paths from this switch "
				"for port with LID %u\n", mlid_ho,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
				p_sw->p_node->print_desc,
				cl_ntoh16(osm_port_get_base_lid
					  (p_wobj->p_port)));
			mbox->num_failed_routes++;
			mcast_work_obj_delete(p_wobj);
			continue;
		}

		if (port_num > array_size) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A04: "
				"Error routing MLID 0x%X through switch 0x%"
				PRIx64 " %s\n"
				"\t\t\t\tNo multicast paths from this switch "
				"to port with LID %u\n", mlid_ho,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
				p_sw->p_node->print_desc,
				cl_ntoh16(osm_port_get_base_lid
					  (p_wobj->p_port)));
			mbox->num_failed_routes++;
			mcast_work_obj_delete(p_wobj);
			/* This is means OpenSM has a bug. */
			OSM_ASSERT(FALSE);
			continue;
		}

		if (list_array[port_num].index == OSM_NO_PATH) {
		/* If this port is used first time for multicast routing */
			for (index = 0; index < array_size; index++) {

				if (p_mtn->port_order[index] == port_num)
				/* Routing already known port */
					break;

				if (p_mtn->port_order[index] == OSM_NO_PATH)
				/* Routing new port */
					break;
			}
			/* Update the discovered port order */
			OSM_ASSERT(index < array_size);
			if (p_mtn->port_order)
				p_mtn->port_order[index] = port_num;
			list_array[port_num].index = index;
		}
		cl_qlist_insert_tail(&list_array[port_num].array_obj, &p_wobj->list_item);
	}
	/* Go over all ports and create new port_order list of only used ports */
	for (index = 0; index < array_size; index++) {
		int port_number = p_mtn->port_order[index];
		if (port_number == OSM_NO_PATH)
		/* End of populated port list */
			break;
		if (list_array[port_number].index == OSM_NO_PATH) {
		/* No work object on the port. Destroying the sub-tree */
			OSM_ASSERT(cl_is_qlist_empty(&list_array[port_number].array_obj));

			if (p_mtn->child_array[port_number] != OSM_MTREE_LEAF &&
			    p_mtn->child_array[port_number] != OSM_MTREE_SENDER)
				osm_mtree_destroy(p_mtn->child_array[port_number]);
			p_mtn->child_array[port_number] = NULL;
		} else {
		/* Selected port, populate new port_oder list */
			new_port_order[used_port_index] = port_number;
			used_port_index++;
		}
	}
	free(p_mtn->port_order);
	p_mtn->port_order = new_port_order;
	OSM_LOG_EXIT(p_log);
}

/**********************************************************************
  This is the recursive function to compute the paths in the spanning
  tree that emanate from this switch.  On input, the p_list contains
  the group members that must be routed from this switch.

  The function returns the newly created mtree node element.
**********************************************************************/
static osm_mtree_node_t *mcast_mgr_branch(osm_subn_t * p_subn,
					  osm_mgrp_box_t * mbox,
					  osm_switch_t * p_sw,
					  cl_qlist_t * p_list, uint8_t depth,
					  uint8_t upstream_port,
					  uint8_t * p_max_depth)
{
	uint8_t max_children;
	osm_mtree_node_t *p_mtn = NULL;
	cl_qlist_t *list_array = NULL;
	uint8_t i;
	ib_net64_t node_guid;
	osm_mcast_work_obj_t *p_wobj;
	cl_qlist_t *p_port_list;
	size_t count;
	osm_mcast_tbl_t *p_tbl;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	uint16_t mlid_ho = mbox->mlid;

	OSM_LOG_ENTER(p_log);

	OSM_ASSERT(p_sw);
	OSM_ASSERT(p_list);
	OSM_ASSERT(p_max_depth);

	node_guid = osm_node_get_node_guid(p_sw->p_node);

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Routing MLID 0x%X through switch 0x%" PRIx64
		" %s, %u nodes at depth %u\n",
		mlid_ho, cl_ntoh64(node_guid), p_sw->p_node->print_desc,
		cl_qlist_count(p_list), depth);

	OSM_ASSERT(cl_qlist_count(p_list) > 0);

	depth++;

	if (depth >= 64) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A21: "
			"Maximal hops number is reached for MLID 0x%x."
			" Break processing.\n", mlid_ho);
		mcast_mgr_purge_list(p_log, mbox, p_list);
		goto Exit;
	}

	if (depth > *p_max_depth) {
		OSM_ASSERT(depth == *p_max_depth + 1);
		*p_max_depth = depth;
	}

	if (osm_switch_supports_mcast(p_sw) == FALSE) {
		/*
		   This switch doesn't do multicast.  Clean-up.
		 */
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A14: "
			"Switch 0x%" PRIx64 " %s does not support multicast\n",
			cl_ntoh64(node_guid), p_sw->p_node->print_desc);

		/*
		   Deallocate all the work objects on this branch of the tree.
		 */
		mcast_mgr_purge_list(p_log, mbox, p_list);
		goto Exit;
	}

	p_mtn = osm_mtree_node_new(p_sw);
	if (p_mtn == NULL) {
		/*
		   We are unable to continue routing down this
		   leg of the tree.  Clean-up.
		 */
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A15: "
			"Insufficient memory to build multicast tree\n");

		/*
		   Deallocate all the work objects on this branch of the tree.
		 */
		mcast_mgr_purge_list(p_log, mbox, p_list);
		goto Exit;
	}

	max_children = osm_mtree_node_get_max_children(p_mtn);

	OSM_ASSERT(max_children > 1);

	/*
	   Prepare an empty list for each port in the switch.
	   TO DO - this list array could probably be moved
	   inside the switch element to save on malloc thrashing.
	 */
	list_array = malloc(sizeof(cl_qlist_t) * max_children);
	if (list_array == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A16: "
			"Unable to allocate list array\n");
		mcast_mgr_purge_list(p_log, mbox, p_list);
		osm_mtree_destroy(p_mtn);
		p_mtn = NULL;
		goto Exit;
	}

	memset(list_array, 0, sizeof(cl_qlist_t) * max_children);

	for (i = 0; i < max_children; i++)
		cl_qlist_init(&list_array[i]);

	mcast_mgr_subdivide(p_subn, mbox, p_sw, p_list, list_array,
			    max_children);

	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);

	/*
	   Add the upstream port to the forwarding table unless
	   we're at the root of the spanning tree.
	 */
	if (depth > 1) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Adding upstream port %u\n", upstream_port);

		OSM_ASSERT(upstream_port);
		osm_mcast_tbl_set(p_tbl, mlid_ho, upstream_port);
		p_mtn->up_port = upstream_port;
	}

	/*
	   For each port that was allocated some routes,
	   recurse into this function to continue building the tree
	   if the node on the other end of that port is another switch.
	   Otherwise, the node is an endpoint, and we've found a leaf
	   of the tree.  Mark leaves with our special pointer value.
	 */

	for (i = 0; i < max_children; i++) {
		const osm_physp_t *p_physp;
		const osm_physp_t *p_remote_physp;
		osm_node_t *p_node;
		const osm_node_t *p_remote_node;

		p_port_list = &list_array[i];

		count = cl_qlist_count(p_port_list);

		/*
		   There should be no children routed through the upstream port!
		 */
		OSM_ASSERT(upstream_port == 0 || i != upstream_port ||
			  (i == upstream_port && count == 0));

		if (count == 0)
			continue;	/* No routes down this port. */

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Routing %zu destinations via switch port %u\n",
			count, i);

		if (i == 0) {
			/* This means we are adding the switch to the MC group.
			   We do not need to continue looking at the remote
			   port, just needed to add the port to the table */
			OSM_ASSERT(count == 1);

			osm_mcast_tbl_set(p_tbl, mlid_ho, i);

			p_wobj = (osm_mcast_work_obj_t *)
			    cl_qlist_remove_head(p_port_list);

			/* Update number of children that receive traffic */
			if (p_wobj->is_receiver) {
				p_mtn->child_array[i] = OSM_MTREE_LEAF;
				p_mtn->num_receivers++;
			} else
				p_mtn->child_array[i] = OSM_MTREE_SENDER;

			mcast_work_obj_delete(p_wobj);
			continue;
		}

		p_node = p_sw->p_node;
		p_remote_node = osm_node_get_remote_node(p_node, i, NULL);
		if (!p_remote_node) {
			/*
			 * If we reached here, it means the minhop table has
			 * invalid entries that leads to disconnected ports.
			 *
			 * A possible reason for the code to reach here is
			 * that ucast cache is enabled, and a leaf switch that
			 * is used as a non-leaf switch in a multicast has been
			 * removed from the fabric.
			 *
			 * When it happens, we should invalidate the cache
			 * and force rerouting of the fabric.
			 */

			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 0A1E: Tried to route MLID 0x%X through "
				"disconnected switch 0x%" PRIx64 " port %d\n",
				mlid_ho, cl_ntoh64(node_guid), i);

			/* Free memory */
			mcast_mgr_purge_list(p_log, mbox, p_port_list);

			/* Invalidate ucast cache */
			if (p_sm->ucast_mgr.p_subn->opt.use_ucast_cache &&
			    p_sm->ucast_mgr.cache_valid) {
				OSM_LOG(p_log, OSM_LOG_INFO,
					"Unicast Cache will be invalidated due "
					"to multicast routing errors\n");
				p_sm->ucast_mgr.variables_invalidate_cache =
				    TRUE;
				p_sm->p_subn->force_heavy_sweep = TRUE;
			}

			continue;
		}

		/*
		   This port routes frames for this mcast group.  Therefore,
		   set the appropriate bit in the multicast forwarding
		   table for this switch.
		 */
		osm_mcast_tbl_set(p_tbl, mlid_ho, i);

		if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH) {
			/*
			   Acquire a pointer to the remote switch then recurse.
			 */
			OSM_ASSERT(p_remote_node->sw);

			p_physp = osm_node_get_physp_ptr(p_node, i);
			OSM_ASSERT(p_physp);

			p_remote_physp = osm_physp_get_remote(p_physp);
			OSM_ASSERT(p_remote_physp);

			p_mtn->child_array[i] =
			    mcast_mgr_branch(p_subn, mbox, p_remote_node->sw,
					     p_port_list, depth,
					     osm_physp_get_port_num
					     (p_remote_physp), p_max_depth);

			if (p_mtn->child_array[i]) {
				/* Update number of children that receive traffic */
				if (p_mtn->child_array[i]->num_receivers)
					p_mtn->num_receivers++;
				p_mtn->child_array[i]->p_up = p_mtn;
			}
		} else {
			/*
			   The neighbor node is not a switch, so this
			   must be a leaf.
			 */
			OSM_ASSERT(count == 1);

			p_wobj = (osm_mcast_work_obj_t *)
			    cl_qlist_remove_head(p_port_list);

			/* Update number of children that receive traffic */
			if (p_wobj->is_receiver) {
				p_mtn->child_array[i] = OSM_MTREE_LEAF;
				p_mtn->num_receivers++;
			} else
				p_mtn->child_array[i] = OSM_MTREE_SENDER;

			OSM_ASSERT(cl_is_qlist_empty(p_port_list));

			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Found leaf for port 0x%016" PRIx64
				" on switch port %u\n",
				cl_ntoh64(osm_port_get_guid(p_wobj->p_port)),
				i);
			mcast_work_obj_delete(p_wobj);
		}
	}

	free(list_array);
Exit:
	OSM_LOG_EXIT(p_log);
	return p_mtn;
}

static osm_mtree_node_t *mcast_mgr_branch_const(osm_subn_t * p_subn,
					  osm_mgrp_box_t * mbox,
					  osm_switch_t * p_sw,
					  osm_mtree_node_t *p_mtn,
					  cl_qlist_t * p_list, uint8_t depth,
					  uint8_t upstream_port,
					  uint8_t * p_max_depth)
{
	uint8_t max_children;
	list_array_obj_t *list_array = NULL;
	uint8_t i;
	ib_net64_t node_guid;
	osm_mcast_work_obj_t *p_wobj;
	cl_qlist_t *p_port_list;
	size_t count;
	osm_mcast_tbl_t *p_tbl;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	uint16_t mlid_ho = mbox->mlid;

	OSM_LOG_ENTER(p_log);

	OSM_ASSERT(p_sw);
	OSM_ASSERT(p_list);
	OSM_ASSERT(p_max_depth);

	/* Find node guid of the switch */
	node_guid = osm_node_get_node_guid(p_sw->p_node);

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Routing MLID 0x%X through switch 0x%" PRIx64
		" %s, %u nodes at depth %u\n",
		mlid_ho, cl_ntoh64(node_guid), p_sw->p_node->print_desc,
		cl_qlist_count(p_list), depth);

	OSM_ASSERT(cl_qlist_count(p_list) > 0);

	depth++;

	if (depth >= 64) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A23: "
			"Maximal hops number is reached for MLID 0x%x."
			" Break processing.", mlid_ho);
		mcast_mgr_purge_list(p_log, mbox, p_list);
		goto Exit;
	}

	if (depth > *p_max_depth) {
		OSM_ASSERT(depth == *p_max_depth + 1);
		*p_max_depth = depth;
	}

	if (osm_switch_supports_mcast(p_sw) == FALSE) {
		/*
		   This switch doesn't do multicast.  Clean-up.
		 */
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A14: "
			"Switch 0x%" PRIx64 " %s does not support multicast\n",
			cl_ntoh64(node_guid), p_sw->p_node->print_desc);

		/*
		   Deallocate all the work objects on this branch of the tree.
		 */
		mcast_mgr_purge_list(p_log, mbox, p_list);
		goto Exit;
	}

	if (p_mtn && p_mtn->port_order[0] == OSM_NO_PATH) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Previous tree does not support constant tree\n");
		osm_mtree_destroy(p_mtn);
		p_mtn = NULL;
	}

	if (p_mtn == NULL) {
	/* New mtree node */
		p_mtn = osm_mtree_node_new(p_sw);
		if (p_mtn == NULL) {
			/*
			   We are unable to continue routing down this
			   leg of the tree.  Clean-up.
			 */
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A15: "
				"Insufficient memory to build multicast tree\n");

			/*
			   Deallocate all the work objects on this branch of the tree.
			 */
			mcast_mgr_purge_list(p_log, mbox, p_list);
			goto Exit;
		}
	} else {
		p_mtn->p_sw = p_sw;

		/* Reset number of children that receive traffic */
		p_mtn->num_receivers = 0;

	}

	max_children = osm_mtree_node_get_max_children(p_mtn);

	OSM_ASSERT(max_children > 1);

	/*
	   Prepare an empty list for each port in the switch.
	   TO DO - this list array could probably be moved
	   inside the switch element to save on malloc thrashing.
	 */
	list_array = malloc(sizeof(list_array_obj_t) * max_children);
	if (list_array == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A16: "
			"Unable to allocate list array\n");
		mcast_mgr_purge_list(p_log, mbox, p_list);
		osm_mtree_destroy(p_mtn);
		p_mtn = NULL;
		goto Exit;
	}

	/* Init list_array */
	for (i = 0; i < max_children; i++) {
		memset(&list_array[i].array_obj, 0, sizeof(cl_qlist_t));
		cl_qlist_init(&list_array[i].array_obj);
		list_array[i].index = OSM_NO_PATH;
	}

	mcast_mgr_subdivide_const(p_subn, mbox, p_mtn, p_list, list_array,
				  max_children);

	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);

	/*
	   Add the upstream port to the forwarding table unless
	   we're at the root of the spanning tree.
	 */
	if (depth > 1) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Adding upstream port %u\n", upstream_port);

		OSM_ASSERT(upstream_port);
		osm_mcast_tbl_set(p_tbl, mlid_ho, upstream_port);
		p_mtn->up_port = upstream_port;
	}

	/*
	   For each port that was allocated some routes,
	   recurse into this function to continue building the tree
	   if the node on the other end of that port is another switch.
	   Otherwise, the node is an endpoint, and we've found a leaf
	   of the tree.  Mark leaves with our special pointer value.
	 */

	for (i = 0; i < max_children; i++) {
		const osm_physp_t *p_physp;
		const osm_physp_t *p_remote_physp;
		osm_node_t *p_node;
		const osm_node_t *p_remote_node;

		p_port_list = &list_array[i].array_obj;

		count = cl_qlist_count(p_port_list);

		/*
		   There should be no children routed through the upstream port!
		 */
		OSM_ASSERT(upstream_port == 0 || i != upstream_port ||
			  (i == upstream_port && count == 0));

		if (count == 0)
			continue;	/* No routes down this port. */

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Routing %zu destinations via switch port %u\n",
			count, i);

		if (i == 0) {
			/* This means we are adding the switch to the MC group.
			   We do not need to continue looking at the remote
			   port, just needed to add the port to the table */
			OSM_ASSERT(count == 1);

			osm_mcast_tbl_set(p_tbl, mlid_ho, i);

			p_wobj = (osm_mcast_work_obj_t *)
			    cl_qlist_remove_head(p_port_list);

			/* Update number of children that receive traffic */
			if (p_wobj->is_receiver) {
				p_mtn->child_array[i] = OSM_MTREE_LEAF;
				p_mtn->num_receivers++;
			} else
				p_mtn->child_array[i] = OSM_MTREE_SENDER;

			mcast_work_obj_delete(p_wobj);
			continue;
		}

		p_node = p_sw->p_node;
		p_remote_node = osm_node_get_remote_node(p_node, i, NULL);
		if (!p_remote_node) {
			/*
			 * If we reached here, it means the minhop table has
			 * invalid entries that leads to disconnected ports.
			 *
			 * A possible reason for the code to reach here is
			 * that ucast cache is enabled, and a leaf switch that
			 * is used as a non-leaf switch in a multicast has been
			 * removed from the fabric.
			 *
			 * When it happens, we should invalidate the cache
			 * and force rerouting of the fabric.
			 */

			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 0A1F: Tried to route MLID 0x%X through "
				"disconnected switch 0x%" PRIx64 " port %d\n",
				mlid_ho, cl_ntoh64(node_guid), i);

			/* Free memory */
			mcast_mgr_purge_list(p_log, mbox, p_port_list);

			/* Invalidate ucast cache */
			if (p_sm->ucast_mgr.p_subn->opt.use_ucast_cache &&
			    p_sm->ucast_mgr.cache_valid) {
				OSM_LOG(p_log, OSM_LOG_INFO,
					"Unicast Cache will be invalidated due "
					"to multicast routing errors\n");
				p_sm->ucast_mgr.variables_invalidate_cache =
				    TRUE;
				p_sm->p_subn->force_heavy_sweep = TRUE;
			}

			continue;
		}

		/*
		   This port routes frames for this mcast group.  Therefore,
		   set the appropriate bit in the multicast forwarding
		   table for this switch.
		 */
		osm_mcast_tbl_set(p_tbl, mlid_ho, i);

		if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH) {
			/*
			   Acquire a pointer to the remote switch then recurse.
			 */
			OSM_ASSERT(p_remote_node->sw);

			p_physp = osm_node_get_physp_ptr(p_node, i);
			OSM_ASSERT(p_physp);

			p_remote_physp = osm_physp_get_remote(p_physp);
			OSM_ASSERT(p_remote_physp);
			/* Check if p_mtn is pointing to a correct switch */
			if (p_mtn->child_array[i] &&
			    ((p_mtn->child_array[i]->node_guid != osm_physp_get_port_guid(p_remote_physp)) ||
			     (p_remote_node->sw != p_mtn->child_array[i]->p_sw))) {
				osm_mtree_destroy(p_mtn->child_array[i]);
				p_mtn->child_array[i] = NULL;
			}
			p_mtn->child_array[i] =
			    mcast_mgr_branch_const(p_subn, mbox, p_remote_node->sw,
					     p_mtn->child_array[i],
					     p_port_list, depth,
					     osm_physp_get_port_num
					     (p_remote_physp), p_max_depth);

			if (p_mtn->child_array[i]) {
				/* Update number of children that receive traffic */
				if (p_mtn->child_array[i]->num_receivers)
					p_mtn->num_receivers++;
				p_mtn->child_array[i]->p_up = p_mtn;
			}
		} else {
			/*
			   The neighbor node is not a switch, so this
			   must be a leaf.
			 */
			OSM_ASSERT(count == 1);

			p_wobj = (osm_mcast_work_obj_t *)
			    cl_qlist_remove_head(p_port_list);

			/* Update number of children that receive traffic */
			if (p_wobj->is_receiver) {
				p_mtn->child_array[i] = OSM_MTREE_LEAF;
				p_mtn->num_receivers++;
			} else
				p_mtn->child_array[i] = OSM_MTREE_SENDER;

			OSM_ASSERT(cl_is_qlist_empty(p_port_list));

			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Found leaf for port 0x%016" PRIx64
				" on switch port %u\n",
				cl_ntoh64(osm_port_get_guid(p_wobj->p_port)),
				i);
			mcast_work_obj_delete(p_wobj);
		}
	}

	free(list_array);
Exit:
	OSM_LOG_EXIT(p_log);
	return p_mtn;
}

static void compact_mtree(osm_subn_t * p_subn, osm_mgrp_box_t * mbox)
{
	uint8_t port_num;
	osm_mtree_node_t *p_mtn;
	osm_mcast_tbl_t *p_tbl;
	uint8_t parent_port_num;
	osm_node_t *p_rem_node;
	uint16_t mlid_ho = mbox->mlid;
	uint8_t max_port = osm_mtree_node_get_max_children(mbox->root);
	int16_t child_port_num = -1;
	osm_switch_t * p_sw = mbox->root->p_sw;

	for (port_num = 0; port_num < max_port; port_num++)
	{
		if (mbox->root->child_array[port_num]) {
			if (child_port_num == -1)
			/* First connected port found */
				child_port_num = port_num;
			else
			/* If there are more than one port connected to the switch */
				return;
		}
	}
	if (child_port_num != -1 &&
	    (mbox->root->child_array[child_port_num] != OSM_MTREE_LEAF &&
	     mbox->root->child_array[child_port_num] != OSM_MTREE_SENDER)) {
	/* If only one port connected to a switch, move the root to this switch */
		p_mtn = mbox->root->child_array[child_port_num];
		mbox->root->child_array[child_port_num] = NULL;

		p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
		osm_mcast_tbl_clear(p_tbl, mlid_ho, child_port_num);

		p_rem_node = osm_node_get_remote_node(p_sw->p_node,
						      child_port_num,
						      &parent_port_num);
		if (p_rem_node && p_rem_node->sw) {
			p_tbl = osm_switch_get_mcast_tbl_ptr(p_rem_node->sw);
			osm_mcast_tbl_clear(p_tbl, mlid_ho, parent_port_num);
		}

		osm_mtree_destroy(mbox->root);
		mbox->root = p_mtn;
		p_mtn->up_port = 0;
		p_mtn->p_up = NULL;
		compact_mtree(p_subn, mbox);
	}

}

ib_api_status_t osm_mcast_mgr_build_spanning_tree(osm_subn_t * p_subn,
					      osm_mgrp_box_t * mbox)
{
	cl_qlist_t port_list;
	cl_qmap_t port_map;
	uint32_t num_ports;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t max_depth = 0;
	osm_switch_t *p_root_sw;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/*
	   TO DO - for now, just blow away the old tree.
	   In the future we'll need to construct the tree based
	   on multicast forwarding table information if the user wants to
	   preserve existing multicast routes.
	 */
	osm_purge_mtree(p_subn, mbox);

	/* build the first "subset" containing all member ports */
	if (osm_mcast_make_port_list_and_map(&port_list, &port_map, mbox)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A10: "
			"Insufficient memory to make port list\n");
		status = IB_ERROR;
		goto Exit;
	}

	num_ports = cl_qlist_count(&port_list);
	if (num_ports < 2) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"MLID 0x%X has %u members - nothing to do\n",
			mbox->mlid, num_ports);
		goto Exit;
	}

	/*
	   This function builds the single spanning tree recursively.
	   At each stage, the ports to be reached are divided into
	   non-overlapping subsets of member ports that can be reached through
	   a given switch port.  Construction then moves down each
	   branch, and the process starts again with each branch computing
	   for its own subset of the member ports.

	   The maximum recursion depth is at worst the maximum hop count in the
	   subnet, which is spec limited to 64.
	 */

	/*
	   Locate the switch around which to create the spanning
	   tree for this multicast group.
	 */

	/*
	   If the global multicast root is already configured, then check it is
	   validity and use it. If global multicast root is defined but not
	   valid, multicast routing will fail.
	 */
	p_root_sw = osm_mcast_mgr_find_root_switch(p_subn, &port_list, mbox);
        if (p_root_sw == NULL) {
                OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A08: "
			"Unable to locate a suitable switch for group 0x%X\n",
			mbox->mlid);
                status = IB_ERROR;
                goto Exit;
        }

	mbox->root = mcast_mgr_branch(p_subn, mbox, p_root_sw, &port_list,
				      0, 0, &max_depth);

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Configured MLID 0x%X for %u ports, max tree depth = %u\n",
		mbox->mlid, num_ports, max_depth);

Exit:
	if (status == IB_SUCCESS)
		mbox->imr_state = OSM_MGRP_BOX_IMR_SUPPORTED;
	else
		mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;

	osm_mcast_drop_port_list(&port_list);
	OSM_LOG_EXIT(p_log);
	return status;
}

ib_api_status_t osm_mcast_mgr_build_spanning_tree_const(osm_subn_t * p_subn,
							osm_mgrp_box_t * mbox)
{
	cl_qlist_t port_list;
	cl_qmap_t mgrp_sw_map;
	cl_qmap_t port_map;
	uint32_t num_ports;
	osm_node_t *p_node;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t max_depth = 0;
	osm_switch_t *p_root_sw;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/* build the first "subset" containing all member ports */
	if (osm_mcast_make_port_list_and_map(&port_list, &port_map, mbox)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A20: "
			"Insufficient memory to make port list\n");
		status = IB_ERROR;
		goto Exit;
	}

	num_ports = cl_qlist_count(&port_list);
	if (num_ports < 2) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"MLID 0x%04X has %d members - nothing to do\n",
			mbox->mlid, num_ports);
		osm_purge_mtree(p_subn, mbox);
		goto Exit;
	}

	create_mgrp_switch_map(&mgrp_sw_map, &port_list);

	/*
	   This function builds the single spanning tree recursively.
	   At each stage, the ports to be reached are divided into
	   non-overlapping subsets of member ports that can be reached through
	   a given switch port.  Construction then moves down each
	   branch, and the process starts again with each branch computing
	   for its own subset of the member ports.

	   The maximum recursion depth is at worst the maximum hop count in the
	   subnet, which is spec limited to 64.
	 */

	/*
	   Locate the switch around which to create the spanning
	   tree for this multicast group.
	   Try to use preiusly found root switch.
	 */

	p_root_sw = mcast_mgr_find_root_mc_roots_tbl(p_subn, &mgrp_sw_map,
						     mbox);
	if (p_root_sw) {
		if (mbox->root && mbox->root->p_sw != p_root_sw)
			/*
			 * root switch from mc roots file is different than
			 * current root
			 */
			osm_purge_mtree(p_subn, mbox);
	} else if (mbox->root) {
	/* If root switch is already defined */
		p_node = osm_get_node_by_guid(p_subn ,mbox->root->node_guid);
		if (!p_node || mbox->root->p_sw != p_node->sw ||
		    !mcast_mgr_is_valid_root(&mgrp_sw_map, mbox->root->p_sw)) {
		/*
		   If the switch is no longer persist in the fabric or
		   pointer to the switch is changed
		*/
			p_subn->global_mc_root_guid = 0;
			osm_purge_mtree(p_subn, mbox);
			p_root_sw =
			    osm_mcast_mgr_find_root_switch_map(p_subn,
							       &mgrp_sw_map,
							       mbox);
		} else
			p_root_sw = mbox->root->p_sw;
	} else
		/* If no root switch is defined to the group yet */
		p_root_sw = osm_mcast_mgr_find_root_switch_map(p_subn,
							       &mgrp_sw_map,
							       mbox);
	destroy_mgrp_switch_map(&mgrp_sw_map);

	if (p_root_sw == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A08: "
			"Unable to locate a suitable switch for group 0x%X\n",
			mbox->mlid);
		status = IB_ERROR;
		goto Exit;
	}

	mbox->root = mcast_mgr_branch_const(p_subn, mbox, p_root_sw,
					    mbox->root, &port_list, 0, 0,
					    &max_depth);

	/* Remove switches with only one connected port	to the other switches */
	compact_mtree(p_subn, mbox);
	mbox->root_guid =
	    cl_ntoh64(osm_node_get_node_guid(mbox->root->p_sw->p_node));

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Configured MLID 0x%X for %u ports, max tree depth = %u\n",
		mbox->mlid, num_ports, max_depth);

Exit:

	if (status == IB_SUCCESS)
		mbox->imr_state = OSM_MGRP_BOX_IMR_SUPPORTED;
	else
		mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;

	osm_mcast_drop_port_list(&port_list);
	OSM_LOG_EXIT(p_log);
	return status;
}

#if 0
/* unused */
void osm_mcast_mgr_set_table(osm_sm_t * sm, IN const osm_mgrp_t * p_mgrp,
			     IN const osm_mtree_node_t * p_mtn)
{
	uint8_t i;
	uint8_t max_children;
	osm_mtree_node_t *p_child_mtn;
	uint16_t mlid_ho;
	osm_mcast_tbl_t *p_tbl;
	osm_switch_t *p_sw;

	OSM_LOG_ENTER(sm->p_log);

	mlid_ho = cl_ntoh16(osm_mgrp_get_mlid(p_mgrp));
	p_sw = osm_mtree_node_get_switch_ptr(p_mtn);

	OSM_ASSERT(p_sw);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Configuring MLID 0x%X on switch 0x%" PRIx64 "\n",
		mlid_ho, osm_node_get_node_guid(p_sw->p_node));

	/*
	   For every child of this tree node, set the corresponding
	   bit in the switch's mcast table.
	 */
	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
	max_children = osm_mtree_node_get_max_children(p_mtn);

	OSM_ASSERT(max_children <= osm_switch_get_num_ports(p_sw));

	osm_mcast_tbl_clear_mlid(p_tbl, mlid_ho);

	for (i = 0; i < max_children; i++) {
		p_child_mtn = osm_mtree_node_get_child(p_mtn, i);
		if (p_child_mtn == NULL)
			continue;

		osm_mcast_tbl_set(p_tbl, mlid_ho, i);
	}

	OSM_LOG_EXIT(sm->p_log);
}
#endif

void osm_mcast_mgr_clear_mft(osm_subn_t * p_subn, uint16_t mlid)
{
	osm_switch_t *p_sw;
	cl_qmap_t *p_sw_tbl;
	osm_mcast_tbl_t *p_mcast_tbl;
	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/* Walk the switches and clear the routing entries for this MLID. */
	p_sw_tbl = &p_subn->sw_guid_tbl;
	p_sw = (osm_switch_t *) cl_qmap_head(p_sw_tbl);
	while (p_sw != (osm_switch_t *) cl_qmap_end(p_sw_tbl)) {
		p_mcast_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
		osm_mcast_tbl_clear_mlid(p_mcast_tbl, mlid);
		p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item);
	}

	OSM_LOG_EXIT(p_log);
}

#if 0
/* TO DO - make this real -- at least update spanning tree */
/**********************************************************************
   Lock must be held on entry.
**********************************************************************/
ib_api_status_t osm_mcast_mgr_process_single(osm_sm_t * sm,
					     IN ib_net16_t const mlid,
					     IN ib_net64_t const port_guid,
					     IN uint8_t const join_state)
{
	uint8_t port_num;
	uint16_t mlid_ho;
	ib_net64_t sw_guid;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_physp_t *p_remote_physp;
	osm_node_t *p_remote_node;
	osm_mcast_tbl_t *p_mcast_tbl;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(mlid);
	OSM_ASSERT(port_guid);

	mlid_ho = cl_ntoh16(mlid);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Attempting to add port 0x%" PRIx64 " to MLID 0x%X, "
		"\n\t\t\t\tjoin state = 0x%X\n",
		cl_ntoh64(port_guid), mlid_ho, join_state);

	/*
	   Acquire the Port object.
	 */
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A01: "
			"Unable to acquire port object for 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid));
		status = IB_ERROR;
		goto Exit;
	}

	p_physp = p_port->p_physp;
	if (p_physp == NULL) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A05: "
			"Unable to acquire phsyical port object for 0x%" PRIx64
			"\n", cl_ntoh64(port_guid));
		status = IB_ERROR;
		goto Exit;
	}

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp == NULL) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A11: "
			"Unable to acquire remote phsyical port object "
			"for 0x%" PRIx64 "\n", cl_ntoh64(port_guid));
		status = IB_ERROR;
		goto Exit;
	}

	p_remote_node = osm_physp_get_node_ptr(p_remote_physp);

	OSM_ASSERT(p_remote_node);

	sw_guid = osm_node_get_node_guid(p_remote_node);

	if (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A22: "
			"Remote node not a switch node 0x%" PRIx64 "\n",
			cl_ntoh64(sw_guid));
		status = IB_ERROR;
		goto Exit;
	}

	if (!p_remote_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A12: "
			"No switch object 0x%" PRIx64 "\n", cl_ntoh64(sw_guid));
		status = IB_ERROR;
		goto Exit;
	}

	if (osm_switch_is_in_mcast_tree(p_remote_node->sw, mlid_ho)) {
		/*
		   We're in luck. The switch attached to this port
		   is already in the multicast group, so we can just
		   add the specified port as a new leaf of the tree.
		 */
		if (join_state & (IB_JOIN_STATE_FULL | IB_JOIN_STATE_NON)) {
			/*
			   This node wants to receive multicast frames.
			   Get the switch port number to which the new member port
			   is attached, then configure this single mcast table.
			 */
			port_num = osm_physp_get_port_num(p_remote_physp);
			OSM_ASSERT(port_num);

			p_mcast_tbl =
			    osm_switch_get_mcast_tbl_ptr(p_remote_node->sw);
			osm_mcast_tbl_set(p_mcast_tbl, mlid_ho, port_num);
		} else {
			if (join_state & IB_JOIN_STATE_SEND_ONLY)
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Success.  Nothing to do for send"
					"only member\n");
			else {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A13: "
					"Unknown join state 0x%X\n",
					join_state);
				status = IB_ERROR;
				goto Exit;
			}
		}
	} else
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Unable to add port\n");

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}
#endif

static void mcast_mgr_redirect_tree_dfs(osm_mtree_node_t * p_mtn, uint16_t mlid,
					boolean_t forward_up)
{
	uint8_t i;
	osm_mtree_node_t *p_child;
	osm_mcast_tbl_t *p_tbl;

	/*
	 * Redirect multicast tree according to the following rules:
	 * 1. If child of current node has receivers, forward traffic to child.
	 * 2. If there are other receivers other than this node (and it's
	 *    decendents) forward trafic to parent.
	 */

	OSM_ASSERT(p_mtn && p_mtn->p_sw);
	p_tbl = osm_switch_get_mcast_tbl_ptr(p_mtn->p_sw);
	osm_mcast_tbl_clear_mlid(p_tbl, mlid);

	if (p_mtn->p_up) {
		/*
		 * If current node does not have any receivers, we forward
		 * traffic up only if one of his ancestors does have receivers.
		 */
		forward_up |= (p_mtn->num_receivers == 0 &&
			       p_mtn->p_up->num_receivers);
		if (forward_up)
			osm_mcast_tbl_set(p_tbl, mlid, p_mtn->up_port);
	}

	/*
	 * If current node has more than one receiving branches, all branches
	 * should forward traffic to this node.
	 */
	forward_up |= p_mtn->num_receivers > 1;

	for (i = 0; i < p_mtn->max_children; i++) {
		p_child = p_mtn->child_array[i];
		if (!p_child)
			continue;

		/* Child is receiver */
		if (p_child == OSM_MTREE_LEAF) {
			osm_mcast_tbl_set(p_tbl, mlid, i);
			continue;
		}

		/* Child is SendOnlyNonMember */
		if (p_child == OSM_MTREE_SENDER)
			continue;

		/* Child has receivers, hence should receive traffic */
		if (p_child->num_receivers)
			osm_mcast_tbl_set(p_tbl, mlid, i);

		/* Handle sub-tree */
		mcast_mgr_redirect_tree_dfs(p_child, mlid, forward_up);
	}
}

/*
 * Travers multicast tree structure and set multicast forwarding tables
 * of the switches according to the information that was set by the routing
 * algorithm to each osm_mtree_node_t object.
 */
static void mcast_mgr_redirect_tree(osm_sm_t *sm, osm_mgrp_box_t * mbox)
{
	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(mbox);
	mcast_mgr_redirect_tree_dfs(mbox->root, mbox->mlid, FALSE);

	OSM_LOG_EXIT(sm->p_log);
}

static void mcast_mgr_increase_weights_dfs(osm_mtree_node_t * p_mtn)
{
	uint8_t i;
	osm_mtree_node_t *p_child;

	OSM_ASSERT(p_mtn && p_mtn->p_sw);

	for (i = 1; i < p_mtn->max_children; i++) {
		p_child = p_mtn->child_array[i];
		if (p_child && p_child != OSM_MTREE_LEAF &&
		    p_child != OSM_MTREE_SENDER) {
			mcast_mgr_increase_weights_dfs(p_child);
			osm_node_increase_mcast_weight(p_mtn->p_sw->p_node, i);
		}
	}
}

/* Traverse multicast tree and increase mcast weigths on the links */
void osm_mcast_mgr_increase_weights(IN osm_subn_t * p_subn,
				    IN osm_mgrp_box_t * mbox)
{
	OSM_LOG_ENTER(&p_subn->p_osm->log);

	OSM_ASSERT(mbox);
	if (mbox->root)
		mcast_mgr_increase_weights_dfs(mbox->root);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void mcast_mgr_decrease_weights_dfs(osm_subn_t * p_subn,
					   osm_mtree_node_t * p_mtn)
{
	uint8_t i;
	osm_node_t *p_node;
	osm_mtree_node_t *p_child;

	OSM_ASSERT(p_mtn && p_mtn->p_sw);

	p_node = osm_get_node_by_guid(p_subn ,p_mtn->node_guid);
	if (p_node && p_node->sw != p_mtn->p_sw)
		/* p_mtn->p_sw no longer valid */
		p_node = NULL;

	for (i = 1; i < p_mtn->max_children; i++) {
		p_child = p_mtn->child_array[i];
		if (p_child && p_child != OSM_MTREE_LEAF &&
		    p_child != OSM_MTREE_SENDER) {
			mcast_mgr_decrease_weights_dfs(p_subn, p_child);
			if (p_node)
				osm_node_decrease_mcast_weight(p_node, i);
		}
	}
}

/* Decrease multicast weights of the links of a tree */
void osm_mcast_mgr_decrease_weights(IN osm_subn_t * p_subn,
				    IN osm_mgrp_box_t * mbox)
{
	OSM_LOG_ENTER(&p_subn->p_osm->log);

	OSM_ASSERT(mbox);
	if (mbox->root)
		mcast_mgr_decrease_weights_dfs(p_subn, mbox->root);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static osm_mtree_node_t *
mcast_mgr_incremental_multicast_dfs(osm_sm_t * sm, osm_mgrp_box_t * mbox,
				    osm_mtree_node_t * p_mtn)
{
	osm_mtree_node_t *p_child;
	osm_mcast_tbl_t *p_tbl;
	uint8_t port;
	uint8_t num_children = 0;

	p_mtn->num_receivers = 0;

	for (port = 0; port < p_mtn->max_children; port++) {
		p_child = p_mtn->child_array[port];

		if (p_child == OSM_MTREE_LEAF) {
			p_mtn->num_receivers++;
			num_children++;
		} else if (p_child == OSM_MTREE_SENDER) {
			num_children++;
		} else if (p_child) {
			/* Handle sub-tree */
			p_child = mcast_mgr_incremental_multicast_dfs(sm, mbox,
								      p_child);
			if (p_child) {
				num_children++;
				if (p_child->num_receivers)
					p_mtn->num_receivers++;
			} else
				/* child tree was pruned */
				osm_node_decrease_mcast_weight(p_mtn->p_sw->p_node,
							       port);

			p_mtn->child_array[port] = p_child;
		}
	}

	/*
	 * If current node does not have any children, we remove it from the
	 * tree and delete MFT table of the switch.
	 *
	 * NOTE: we do not need to remove MFT entry toward this node.
	 * MFT will be assigned by mcast_mgr_redirect_tree.
	 */
	if (!num_children) {
		OSM_ASSERT(p_mtn->p_sw);
		p_tbl = osm_switch_get_mcast_tbl_ptr(p_mtn->p_sw);
		osm_mcast_tbl_clear_mlid(p_tbl, mbox->mlid);
		cl_qmap_remove_item(&mbox->mtree_nodes_tbl,
				    (cl_map_item_t *)p_mtn);
		osm_mtree_destroy(p_mtn);

		p_mtn = NULL;
	}

	return p_mtn;
}

static void mcast_mgr_incremental_multicast(osm_sm_t * sm,
					    osm_mgrp_box_t * mbox)
{
	OSM_LOG_ENTER(sm->p_log);

	if (!mbox->imr_tree_changed_flag) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Multicast tree of mlid 0x%X not changed\n",
			mbox->mlid);
	} else {
		if (mbox->root)
			mbox->root =
			    mcast_mgr_incremental_multicast_dfs(sm, mbox,
								mbox->root);
		if (mbox->root)
			mcast_mgr_redirect_tree(sm, mbox);

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Multicast tree of mlid 0x%X calculated incrementally\n",
			mbox->mlid);
	}

	mbox->imr_tree_changed_flag = FALSE;

	OSM_LOG_EXIT(sm->p_log);
}

/*
 * Add sub tree to mtree_nodes_tbl table.
 */
static void mcast_mgr_create_mtree_nodes_tbl_dfs(cl_qmap_t * p_map,
						 osm_mtree_node_t * p_mtn)
{
	osm_mtree_node_t *p_child;
	uint8_t port;

	OSM_ASSERT(p_mtn->p_sw);

	cl_qmap_insert(p_map, osm_node_get_node_guid(p_mtn->p_sw->p_node),
	               &p_mtn->map_item);

	for (port = 1; port < p_mtn->max_children; port++) {
		p_child = p_mtn->child_array[port];
		if (p_child && p_child != OSM_MTREE_LEAF &&
		    p_child != OSM_MTREE_SENDER)
			mcast_mgr_create_mtree_nodes_tbl_dfs(p_map, p_child);
	}
}

/*
 * Create mapping from switch guid to osm_mtree_node_t objects of the multicast
 * tree.
 */
static void mcast_mgr_create_mtree_nodes_tbl(osm_sm_t * sm,
					     osm_mgrp_box_t * mbox)
{
	OSM_LOG_ENTER(sm->p_log);

	cl_qmap_remove_all(&mbox->mtree_nodes_tbl);
	if (mbox->root)
		mcast_mgr_create_mtree_nodes_tbl_dfs(&mbox->mtree_nodes_tbl,
						     mbox->root);

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 Process the entire group.
 NOTE : The lock should be held externally!
 **********************************************************************/
static ib_api_status_t mcast_mgr_process_mlid(osm_sm_t * sm, uint16_t mlid)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_subn_t *p_subn = sm->p_subn;
	struct osm_routing_engine *re = p_subn->routing_engine_used;
	osm_mgrp_box_t *mbox;
	event_data_t event_data;

	OSM_LOG_ENTER(sm->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Processing multicast group with mlid 0x%X\n", mlid);

	mbox = osm_get_mbox_by_mlid(p_subn, cl_hton16(mlid));

	if (mbox &&
	    !p_subn->force_multicast_rerouting &&
	    (mbox->imr_state == OSM_MGRP_BOX_IMR_SUPPORTED ||
	    (mbox->imr_state == OSM_MGRP_BOX_IMR_ON_IDLE &&
	     !p_subn->in_heavy_sweep))) {
		mcast_mgr_incremental_multicast(sm, mbox);
		goto Exit;
	}

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Recalculating multicast tree of mlid 0x%X\n", mlid);

	memset(&event_data, 0, sizeof(event_data_t));
	event_data.event_id = OSM_ACT_MGR_MCG_TREE_CALC ;
	event_data.mlid = cl_ntoh16(mlid);
	osm_act_mgr_post_event(&sm->activity_mgr, &event_data);

	/* Clear the multicast tables to start clean, then build
	   the spanning tree which sets the mcast table bits for each
	   port in the group. */
	osm_mcast_mgr_clear_mft(p_subn, mlid);

	if (mbox) {
		mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
		mbox->num_failed_routes = 0;

		/* Update weights */
		mcast_mgr_decrease_mtree_root_weight(p_subn, mbox);
		osm_mcast_mgr_decrease_weights(p_subn, mbox);

		if (re && re->mcast_build_stree)
			status = re->mcast_build_stree(re->context, mbox);
		else if (p_subn->opt.const_multicast)
			status = osm_mcast_mgr_build_spanning_tree_const(p_subn,
									 mbox);
		else
			status = osm_mcast_mgr_build_spanning_tree(p_subn, mbox);

		if (status == IB_SUCCESS) {
			if (mbox->num_failed_routes) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A28: "
					"Failed to route %d destinations in "
					"multicast tree for MLID 0x%X\n",
					mbox->num_failed_routes, mlid);
				/*
				 * Since it is possible that some members could
				 * not be routed due to problems with routing,
				 * we disable IMR support in order to force full
				 * multicast rerouting for this group.
				 */
				if (mbox->imr_state ==
				    OSM_MGRP_BOX_IMR_SUPPORTED)
					mbox->imr_state =
					    OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
			}

			/* Update weights */
			mcast_mgr_increase_mtree_root_weight(p_subn, mbox);
			osm_mcast_mgr_increase_weights(p_subn, mbox);

			/* Create directed multicast tree */
			if (mbox->root && mbox->root->num_receivers)
				mcast_mgr_redirect_tree(sm, mbox);

			mcast_mgr_create_mtree_nodes_tbl(sm, mbox);
			mbox->imr_tree_changed_flag = FALSE;

		} else {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A17: "
				"Unable to create spanning tree (%s) for mlid "
				"0x%x\n", ib_get_err_str(status), mlid);
		}
	} else {
		/* Handle deleted mbox in routing chain manager */
		osm_rch_rbox_map_destroy(&sm->routing_chain_mgr, mlid);
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

static void mcast_mgr_set_mfttop(IN osm_sm_t * sm, IN osm_switch_t * p_sw)
{
	osm_node_t *p_node;
	osm_dr_path_t *p_path;
	osm_physp_t *p_physp;
	osm_mcast_tbl_t *p_tbl;
	osm_madw_context_t context;
	ib_api_status_t status;
	ib_switch_info_t si;
	ib_net16_t mcast_top;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_sw);

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_path = osm_physp_get_dr_path_ptr(p_physp);
	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);

	if (sm->p_subn->opt.use_mfttop &&
	    p_physp->port_info.capability_mask & IB_PORT_CAP_HAS_MCAST_FDB_TOP) {
		/*
		   Set the top of the multicast forwarding table.
		 */
		si = p_sw->switch_info;
		if (sm->p_subn->first_time_master_sweep == TRUE)
			mcast_top = cl_hton16(sm->mlids_init_max);
		else {
			if (p_tbl->max_block_in_use == -1)
				mcast_top = cl_hton16(IB_LID_MCAST_START_HO - 1);
			else
				mcast_top = cl_hton16(IB_LID_MCAST_START_HO +
						      (p_tbl->max_block_in_use + 1) * IB_MCAST_BLOCK_SIZE - 1);
		}
		if (mcast_top == si.mcast_top)
			return;

		si.mcast_top = mcast_top;

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Setting switch MFT top to MLID 0x%x\n",
			cl_ntoh16(si.mcast_top));

		context.si_context.light_sweep = FALSE;
		context.si_context.node_guid = osm_node_get_node_guid(p_node);
		context.si_context.set_method = TRUE;
		context.si_context.lft_top_change = FALSE;

		status = osm_req_set(sm, p_path, (uint8_t *) & si,
				     sizeof(si), IB_MAD_ATTR_SWITCH_INFO,
				     0, FALSE,
				     ib_port_info_get_m_key(&p_physp->port_info),
				     0, CL_DISP_MSGID_NONE, &context);

		if (status != IB_SUCCESS)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0A1B: "
				"Sending SwitchInfo attribute failed (%s)\n",
				ib_get_err_str(status));
	}
}

static int mcast_mgr_set_mftables(osm_sm_t * sm,
				  boolean_t clear_removed_entries)
{
	cl_qmap_t *p_sw_tbl = &sm->p_subn->sw_guid_tbl;
	osm_switch_t *p_sw;
	osm_mcast_tbl_t *p_tbl;
	int block_notdone, ret = 0;
	int16_t block_num, max_block = -1;
	osm_madw_t *p_madw = NULL;
	cl_qlist_t mad_list;

	cl_qlist_init(&mad_list);

	p_sw = (osm_switch_t *) cl_qmap_head(p_sw_tbl);
	while (p_sw != (osm_switch_t *) cl_qmap_end(p_sw_tbl)) {
		p_sw->mft_block_num = 0;
		p_sw->mft_position = 0;
		p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
		if (osm_mcast_tbl_get_max_block_in_use(p_tbl) > max_block)
			max_block = osm_mcast_tbl_get_max_block_in_use(p_tbl);
		mcast_mgr_set_mfttop(sm, p_sw);
		p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item);
	}

	/* Stripe the MFT blocks across the switches */
	for (block_num = 0; block_num <= max_block; block_num++) {
		block_notdone = 1;
		while (block_notdone) {
			block_notdone = 0;
			p_sw = (osm_switch_t *) cl_qmap_head(p_sw_tbl);
			while (p_sw != (osm_switch_t *) cl_qmap_end(p_sw_tbl)) {
				if (p_sw->mft_block_num == block_num) {
					block_notdone = 1;
					if (mcast_mgr_set_mft_block(sm, p_sw,
								    p_sw->mft_block_num,
								    p_sw->mft_position,
								    clear_removed_entries,
								    &p_madw))
						ret = -1;
					if (p_madw) {
						cl_qlist_insert_tail(&mad_list,
								 &p_madw->list_item);
					}
					p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
					if (++p_sw->mft_position > p_tbl->max_position) {
						p_sw->mft_position = 0;
						p_sw->mft_block_num++;
					}
				}
				p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item);
			}
		}
	}

	osm_send_req_mad_batch(sm, &mad_list);
	return ret;
}

ib_api_status_t osm_mcast_mgr_alloc_mfts(osm_subn_t * p_subn, uint16_t max_mlid)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	unsigned mlid_offset;

	if (max_mlid < IB_LID_MCAST_START_HO)
		return IB_SUCCESS;

	mlid_offset = max_mlid - IB_LID_MCAST_START_HO;

	/* Now, walk switches and (re)allocate multicast tables */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		if (osm_mcast_tbl_realloc(&p_sw->mcast_tbl, mlid_offset))
			return IB_ERROR;;
	}
	return IB_SUCCESS;
}

/**********************************************************************
  Checks if a switch has "down" path to each port in the list
**********************************************************************/
static boolean_t mcast_mgr_is_valid_topdown_root(osm_switch_t *p_root_sw,
						 cl_qmap_t * p_map)
{
	cl_map_item_t *item;
	osm_node_t *p_node;
	osm_switch_t *p_switch;
	uint16_t lid_ho;

	for (item = cl_qmap_head(p_map);
	     item != cl_qmap_end(p_map);
	     item = cl_qmap_next(item)) {

		p_switch = cl_item_obj(item, p_switch, mgrp_item);
		p_node = p_switch->p_node;

		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_node, 0));

		if (p_switch->rank - p_root_sw->rank !=
		    osm_switch_get_least_hops(p_root_sw, lid_ho))
			return FALSE;
	}

	return TRUE;
}

/**********************************************************************
  This function searches for the optimal root of a tree without down-up
  turns by searching for a switch with only "down" paths to every member
  of the multicast gropus
**********************************************************************/
static osm_switch_t *
mcast_mgr_find_optimal_topdown_root(osm_subn_t *p_subn, cl_qmap_t * p_map,
				    osm_mgrp_box_t * mbox,
				    osm_switch_t * p_prev_root)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	osm_switch_t *p_root_sw = p_prev_root;
	int max_root_rank = 0xff;
	boolean_t found_root_rank_switch = FALSE;

	/* Find the minimal member rank in order to rule out candidates
	 * faster */
	for (item = cl_qmap_head(p_map);
	     item != cl_qmap_end(p_map);
	     item = cl_qmap_next(item)) {
		p_sw = cl_item_obj(item, p_sw, mgrp_item);
		if (p_sw->rank < max_root_rank) {
			max_root_rank = p_sw->rank;
			found_root_rank_switch = TRUE;
		} else if (p_sw->rank == max_root_rank) {
			/*
			 * When there are two switches in the same rank, the
			 * root must be of a lower rank.
			 */
			if (found_root_rank_switch)
				max_root_rank--;
			found_root_rank_switch = !found_root_rank_switch;
		}
	}

	if (max_root_rank < 0)
		return NULL;

	/* Iterate over all swithces and search for a better candidate */
	for (p_sw = (osm_switch_t *) cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {

		if (p_sw->rank > max_root_rank)
			continue;

		if (!p_root_sw ||
		    p_root_sw->rank < p_sw->rank ||
		    (p_root_sw->rank == p_sw->rank &&
		     p_root_sw != p_prev_root &&
		     p_root_sw->mlid_root_num > p_sw->mlid_root_num)) {

			if (mcast_mgr_is_valid_topdown_root(p_sw, p_map))
				p_root_sw = p_sw;
		}
	}

	return p_root_sw;
}

/*
 * Destroy osm_mtree_node_t objects allocated by
 * mcast_mgr_build_global_root_stree.
 */
static void mcast_mgr_global_root_destroy_mtree_nodes(osm_subn_t * p_subn)
{
	osm_switch_t *p_sw;
	osm_mtree_node_t *p_mtn;
	cl_map_item_t *map_item;

	/* Remove references from non-root switches to tree nodes */
	for (map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     map_item = cl_qmap_next(map_item)) {
		p_sw = (osm_switch_t *)map_item;
		p_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;
		if (p_mtn && p_mtn->p_up)
			p_sw->multicast_priv = NULL;
	}

	/* Destroy all trees from the root */
	for (map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     map_item = cl_qmap_next(map_item)) {
		p_sw = (osm_switch_t *)map_item;
		p_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;
		if (p_mtn) {
			osm_mtree_destroy(p_mtn);
			p_sw->multicast_priv = NULL;
		}
	}
}

/**********************************************************************
  This function creates multicast tree based on unicast routes from
  members toward the global root.
**********************************************************************/
static osm_mtree_node_t * mcast_mgr_build_global_root_stree(osm_subn_t * p_subn,
						cl_qlist_t * p_list,
						osm_switch_t * p_root_sw,
						osm_mgrp_box_t * mbox)
{
	cl_list_item_t *list_item;
	cl_map_item_t *map_item;
	osm_mcast_work_obj_t *p_wobj;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_switch_t *p_sw;
	uint8_t ds_port;
	uint8_t us_port;
	uint8_t port;
	osm_mtree_node_t *p_mtn, *p_up_mtn;
	osm_mcast_tbl_t *p_tbl;
	osm_log_t *p_log = &p_subn->p_osm->log;
	boolean_t update_num_receivers;
	uint16_t mlid_ho = mbox->mlid;
	uint16_t root_switch_base_lid =
	    cl_ntoh16(osm_node_get_base_lid(p_root_sw->p_node, 0));

	for (map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     map_item = cl_qmap_next(map_item))
		((osm_switch_t *)map_item)->multicast_priv = NULL;

	/* Create osm_mtree_node_t object for every switch with members */
	for (list_item = cl_qlist_head(p_list);
	     list_item != cl_qlist_end(p_list);
	     list_item = cl_qlist_next(list_item)) {

		/* Get object of the switch the member is connected to */
		p_wobj = (osm_mcast_work_obj_t *)list_item;
		p_port = p_wobj->p_port;
		if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) {
			p_sw = p_port->p_node->sw;
			ds_port = 0;
		} else {
			p_node =
			    osm_node_get_remote_node(p_port->p_node,
						     p_port->p_physp->port_num,
						     &ds_port);
			OSM_ASSERT(p_node && p_node->sw);
			p_sw = p_node->sw;
		}

		/* Get or create osm_mtree_node object of the switch */
		if (!p_sw->multicast_priv) {
			p_mtn = osm_mtree_node_new(p_sw);
			if (!p_mtn) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A23: "
					"Insufficient memory to build "
					"multicast tree\n");
				mcast_mgr_global_root_destroy_mtree_nodes(p_subn);
				goto exit;
			}
			p_sw->multicast_priv = p_mtn;
		} else {
			p_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;
			OSM_ASSERT(p_mtn->p_sw == p_sw);
		}

		if (p_wobj->is_receiver) {
			p_mtn->child_array[ds_port] = OSM_MTREE_LEAF;
			p_mtn->num_receivers++;
		} else
			 p_mtn->child_array[ds_port] = OSM_MTREE_SENDER;
	}

	/* From every switch create route toward the root */
	for (map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     map_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     map_item = cl_qmap_next(map_item)) {
		p_sw = (osm_switch_t *)map_item;
		p_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;
		if (!p_mtn)
			continue;

		p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
                for (port = 0; port < p_sw->num_ports; port++) {
                        if (p_mtn->child_array[port] == OSM_MTREE_LEAF ||
                            p_mtn->child_array[port] == OSM_MTREE_SENDER)
                                osm_mcast_tbl_set(p_tbl, mlid_ho, port);
                }

		update_num_receivers = FALSE;
		while (p_sw != p_root_sw && !p_mtn->p_up) {
			 /* Create osm_mtree_node object for parent switch */
			us_port =
			    osm_switch_get_port_by_lid(p_sw,
						       root_switch_base_lid,
						       OSM_NEW_LFT);
			if (us_port == 0 || us_port >= p_sw->num_ports) {
				p_node = p_sw->p_node;
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 0A26: Switch 0x%016" PRIx64 " LFT "
					"for multicast root lid 0x%x leads to "
					"invalid port (%d)\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					root_switch_base_lid, us_port);
				mcast_mgr_global_root_destroy_mtree_nodes(p_subn);
				p_mtn = NULL;
				goto exit;
			}
			p_node = osm_node_get_remote_node(p_sw->p_node, us_port,
							  &ds_port);
			if (!p_node || !p_node->sw) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 0A27: Switch 0x%016" PRIx64 " LFT "
					"for multicast root lid 0x%x leads to "
					"disconnected port (%d)\n",
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
					root_switch_base_lid, us_port);
				mcast_mgr_global_root_destroy_mtree_nodes(p_subn);
				p_mtn = NULL;
				goto exit;
			}

			p_sw = p_node->sw;
			if (!p_sw->multicast_priv) {
				p_up_mtn = osm_mtree_node_new(p_sw);
				if (!p_up_mtn) {
					OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A29: "
						"Insufficient memory to build "
						"multicast tree\n");
					mcast_mgr_global_root_destroy_mtree_nodes(p_subn);
					p_mtn = NULL;
					goto exit;
				} else
					p_sw->multicast_priv = p_up_mtn;
			} else
				p_up_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;

			/* Connect the swtich to its parent node */
			p_up_mtn->child_array[ds_port] = p_mtn;
			osm_mcast_tbl_set(p_tbl, mlid_ho, us_port);
			p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
			osm_mcast_tbl_set(p_tbl, mlid_ho, ds_port);

			p_mtn->up_port = us_port;
			p_mtn->p_up = p_up_mtn;

			/*
			 * Update num_receivers of parent node, and decide
			 * whether we should continue to update num_receivers of
			 * all the way to the root.
			 */
                        if (p_mtn->num_receivers)
                                update_num_receivers = !p_up_mtn->num_receivers++;

                        p_mtn = p_up_mtn;
		}

		/* Update num_receivers all the way to the root */
		while (update_num_receivers && p_mtn->p_up) {
			p_mtn = p_mtn->p_up;
			update_num_receivers = !p_mtn->num_receivers++;
		}
	}

	p_mtn = (osm_mtree_node_t *)p_root_sw->multicast_priv;
exit:
	if (!p_mtn) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A24: "
			"Faild building  multicast tree from global "
			"root for MLID 0x%x\n",
			mlid_ho);
		osm_mcast_mgr_clear_mft(p_subn, mlid_ho);
		return NULL;
	}

	mbox->root = p_mtn;
	compact_mtree(p_subn, mbox);

	return mbox->root;
}

ib_api_status_t
mcast_mgr_build_default_stree(osm_subn_t * p_subn, osm_mgrp_box_t * mbox,
			      cl_qmap_t * p_mgrp_sw_map,
			      cl_qlist_t * p_port_list,
			      osm_switch_t * p_root_sw)
{
	uint8_t max_depth = 0;
	ib_api_status_t status = IB_SUCCESS;
	boolean_t const_multicast = p_subn->opt.const_multicast;

	if (const_multicast) {
		if (mbox->root && mbox->root->p_sw != p_root_sw)
			osm_purge_mtree(p_subn, mbox);
		mbox->root = mcast_mgr_branch_const(p_subn, mbox, p_root_sw,
						    mbox->root, p_port_list,
						    0, 0, &max_depth);
	} else
		mbox->root = mcast_mgr_branch(p_subn, mbox, p_root_sw,
					      p_port_list, 0, 0, &max_depth);

	if (mbox->root) {
		if (const_multicast)
			compact_mtree(p_subn, mbox);
		mbox->root_guid =
		    cl_ntoh64(osm_node_get_node_guid(mbox->root->p_sw->p_node));
	} else
		status = IB_ERROR;

	return status;
}

ib_api_status_t
osm_mcast_mgr_build_clf_stree(osm_subn_t * p_subn, osm_mgrp_box_t * mbox) {

	cl_qlist_t port_list;
	cl_qmap_t port_map;
	cl_qmap_t mgrp_switch_map;
	uint32_t num_ports;
	ib_api_status_t status = IB_SUCCESS;
	osm_switch_t *p_new_root;
	osm_switch_t *p_prev_root;
	boolean_t const_multicast = p_subn->opt.const_multicast;
	osm_log_t *p_log = &p_subn->p_osm->log;
	boolean_t force_rerouting = TRUE;

	OSM_LOG_ENTER(p_log);

	p_prev_root =
	    osm_get_switch_by_guid(p_subn, cl_hton64(mbox->root_guid));

	if (!const_multicast)
		osm_purge_mtree(p_subn, mbox);
	else if (p_prev_root && mbox->root && mbox->root->p_sw != p_prev_root)
		osm_purge_mtree(p_subn, mbox);

	/* build the first "subset" containing all member ports */
	if (osm_mcast_make_port_list_and_map(&port_list, &port_map, mbox)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0A25: "
			"Insufficient memory to make port list\n");
		if (const_multicast)
			osm_purge_mtree(p_subn, mbox);
		status = IB_ERROR;
		goto Exit;
	}
	create_mgrp_switch_map(&mgrp_switch_map, &port_list);

	num_ports = cl_qlist_count(&port_list);
	if (num_ports < 2) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"MLID 0x%X has %u members - nothing to do\n",
			mbox->mlid, num_ports);
		if (const_multicast)
			osm_purge_mtree(p_subn, mbox);
		force_rerouting = FALSE;
		goto ExitRelease;
	}

	p_new_root = mcast_mgr_find_root_mc_roots_tbl(p_subn, &mgrp_switch_map,
						      mbox);
	if (p_new_root) {
		/* Use root switch from MC roots table */
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Using switch 0x%016" PRIx64 " as predefined root for "
			"multicast tree for MLID 0x%x\n",
			cl_ntoh64(osm_node_get_node_guid(p_new_root->p_node)),
			mbox->mlid);
		status = mcast_mgr_build_default_stree(p_subn, mbox,
						       &mgrp_switch_map,
						       &port_list, p_new_root);
		force_rerouting = FALSE;
		goto ExitRelease;
	}

	/*
	 * In order to avoid credit-loops on tree based topologies, we use two
	 * methods to build the tree:
	 *
	 * 1. Create a top-down tree without any down/up turns. the root of such
         *    tree is a switch with direct down paths to every group member.
         *
	 * 2. Build the tree based on the unicast routes from the members
	 *    towards the global hub switch, which confines all down/up turns
	 *    in same sub tree of the topology.
	 *
	 * If there is no top-down root nor there is a valid hub switch, the
	 * default multicast algorithm is used, and credits-loop free routing
	 * is not guaranteed.
	 */

	/* If previous root is a valid top-down tree root, use it. */
	if (!p_prev_root ||
	    !mcast_mgr_is_valid_topdown_root(p_prev_root, &mgrp_switch_map))
		p_new_root = mcast_mgr_find_optimal_topdown_root(p_subn,
								 &mgrp_switch_map,
								 mbox, NULL);
	else if (!const_multicast)
		p_new_root = mcast_mgr_find_optimal_topdown_root(p_subn,
								 &mgrp_switch_map,
								 mbox, p_prev_root);
	else
		p_new_root = p_prev_root;

	if (p_new_root) {
		/* We found valid top-down tree root switch */
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Using switch 0x%016" PRIx64 " as root for top-down "
			"multicast tree for MLID 0x%x\n",
			cl_ntoh64(osm_node_get_node_guid(p_new_root->p_node)),
			mbox->mlid);
		status = mcast_mgr_build_default_stree(p_subn, mbox,
						       &mgrp_switch_map,
						       &port_list, p_new_root);
		force_rerouting = FALSE;
		goto ExitRelease;
	}

	p_new_root = osm_subn_get_hub_switch(p_subn);
	if (p_new_root && p_subn->opt.calc_missing_routes &&
	    mcast_mgr_is_valid_root_by_lid(&mgrp_switch_map, p_new_root)) {
		/* Use hub switch as root */
		OSM_LOG(p_log, OSM_LOG_INFO,
			"Using hub switch 0x%016" PRIx64 " as root for "
			"multicast tree for MLID 0x%x\n",
			cl_ntoh64(osm_node_get_node_guid(p_new_root->p_node)),
			mbox->mlid);

		if (const_multicast && mbox->root)
			osm_purge_mtree(p_subn, mbox);

		mbox->root =
		    mcast_mgr_build_global_root_stree(p_subn, &port_list,
						      p_new_root, mbox);
		if (mbox->root) {
			p_new_root = mbox->root->p_sw;
			mbox->root_guid =
			    cl_ntoh64(osm_node_get_node_guid(p_new_root->p_node));
		} else
			status = IB_ERROR;

		goto ExitRelease;
	}


	/* Use default multicast routing */
	if (p_prev_root && const_multicast &&
	    mcast_mgr_is_valid_root(&mgrp_switch_map, p_prev_root))
		p_new_root = p_prev_root;
	else
		p_new_root = mcast_mgr_find_optimal_switch(p_subn, &mgrp_switch_map, mbox);

	if (p_new_root) {
		OSM_LOG(p_log, OSM_LOG_INFO, "Using default multicast routing "
			"algorithm for MLID 0x%x\n",
			mbox->mlid);
		status = mcast_mgr_build_default_stree(p_subn, mbox,
						       &mgrp_switch_map,
						       &port_list, p_new_root);
	} else
		status = IB_ERROR;

ExitRelease:
	destroy_mgrp_switch_map(&mgrp_switch_map);
	osm_mcast_drop_port_list(&port_list);

Exit:
	if (status == IB_SUCCESS) {
		mbox->imr_state = force_rerouting ?
				  OSM_MGRP_BOX_IMR_ON_IDLE :
				  OSM_MGRP_BOX_IMR_SUPPORTED;
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Created multicast tree for MLID 0x%x\n",
			mbox->mlid);
	} else {
		mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"Failed to create multicast tree for MLID 0x%x\n",
			mbox->mlid);
	}
	OSM_LOG_EXIT(p_log);
	return status;
}

/**********************************************************************
  This is the function that is invoked during idle time and sweep to
  handle the process request for mcast groups where join/leave/delete
  was required.
 **********************************************************************/
int osm_mcast_mgr_process(osm_sm_t * sm, boolean_t config_all)
{
	int ret = 0;
	unsigned i;
	unsigned max_mlid;
	uint16_t subn_max_mlid;

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	/* If there are no switches in the subnet we have nothing to do. */
	if (cl_qmap_count(&sm->p_subn->sw_guid_tbl) == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"No switches in subnet. Nothing to do\n");
		goto exit;
	}

	subn_max_mlid = osm_subn_get_max_mlid(sm->ucast_mgr.p_subn);
	if (osm_mcast_mgr_alloc_mfts(sm->p_subn, subn_max_mlid) != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0A09: alloc_mfts failed\n");
		ret = -1;
		goto exit;
	}

	if (config_all)
		sm->p_subn->reroute_all_mlids = FALSE;

	max_mlid = config_all ? sm->p_subn->max_mcast_lid_ho
			- IB_LID_MCAST_START_HO : sm->mlids_req_max;
	for (i = 0; i <= max_mlid; i++) {
		if (sm->mlids_req[i] ||
		    (config_all && sm->p_subn->mboxes[i])) {
			sm->mlids_req[i] = 0;
			mcast_mgr_process_mlid(sm, i + IB_LID_MCAST_START_HO);
		}
	}

	sm->mlids_req_max = 0;

	/*
	 * Send MFT set request to clear removed MFT entries.
	 * This is required in order to prevent temporary multicast loops
	 * that might appear during MFT tables configuration.
	 */
	if (!sm->p_subn->first_time_master_sweep) {
		ret = mcast_mgr_set_mftables(sm, TRUE);
		if (ret) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 0A1C: Failed to clear removed MFT entries\n");
			goto dump_and_exit;
		}
		/* Wait for fabric configuration */
		CL_PLOCK_RELEASE(sm->p_lock);
		wait_for_pending_transactions(&sm->p_subn->p_osm->stats);
		CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	}


	/*
	 *  Set final MFT tables to the switches.
	 */
	ret = mcast_mgr_set_mftables(sm, FALSE);
	if (ret) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0A1D: Failed to set MFT tables\n");
	}

dump_and_exit:
	osm_dump_mcast_routes(sm->p_subn->p_osm);

exit:
	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
	return ret;
}

