/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_mcast_tbl_t.
 * This object represents a multicast forwarding table.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <complib/cl_math.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MCAST_TBL_C
#include <opensm/osm_mcast_tbl.h>
#include <opensm/osm_log.h>


void osm_mcast_tbl_reset_mask(IN osm_mcast_tbl_t * p_tbl)
{
	memset(p_tbl->need_update_mask, 0xff, sizeof(p_tbl->need_update_mask));
}

void osm_mcast_tbl_init(IN osm_mcast_tbl_t * p_tbl, IN uint8_t num_ports,
			IN uint16_t capacity)
{
	OSM_ASSERT(p_tbl);
	OSM_ASSERT(num_ports);

	memset(p_tbl, 0, sizeof(*p_tbl));

	p_tbl->max_block_in_use = -1;

	if (capacity == 0) {
		/*
		   This switch apparently doesn't support multicast.
		   Everything is initialized to zero already, so return.
		 */
		return;
	}

	p_tbl->num_entries = capacity;
	p_tbl->num_ports = num_ports;
	p_tbl->max_position =
	    (uint8_t) ((ROUNDUP(num_ports, IB_MCAST_MASK_SIZE) /
			IB_MCAST_MASK_SIZE) - 1);

	p_tbl->max_block = (uint16_t) ((ROUNDUP(p_tbl->num_entries,
						IB_MCAST_BLOCK_SIZE) /
					IB_MCAST_BLOCK_SIZE) - 1);

	osm_mcast_tbl_reset_mask(p_tbl);
}

void osm_mcast_tbl_destroy(IN osm_mcast_tbl_t * p_tbl)
{
	free(p_tbl->p_mask_tbl);
	free(p_tbl->p_curr_mask_tbl);
}

void osm_mcast_tbl_set(IN osm_mcast_tbl_t * p_tbl, IN uint16_t mlid_ho,
		       IN uint8_t port)
{
	unsigned mlid_offset, mask_offset, bit_mask;
	uint16_t block_num;

	OSM_ASSERT(p_tbl && p_tbl->p_mask_tbl);
	OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);
	OSM_ASSERT(mlid_ho <= p_tbl->max_mlid_ho);

	mlid_offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;
	mask_offset = mlid_offset + ( port / IB_MCAST_MASK_SIZE ) * IB_MCAST_BLOCK_SIZE;
	block_num = (uint16_t) ((mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE);
	bit_mask = cl_ntoh16((uint16_t) (1 << (port % IB_MCAST_MASK_SIZE)));
	(*p_tbl->p_mask_tbl)[block_num][mask_offset] |= bit_mask;

	if (block_num > p_tbl->max_block_in_use)
		p_tbl->max_block_in_use = (uint16_t) block_num;
}

void osm_mcast_tbl_clear(IN osm_mcast_tbl_t * p_tbl, IN uint16_t mlid_ho,
		         IN uint8_t port)
{
	unsigned mlid_offset, mask_offset, bit_mask;
	uint16_t block_num;

	OSM_ASSERT(p_tbl && p_tbl->p_mask_tbl);
	OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);
	OSM_ASSERT(mlid_ho <= p_tbl->max_mlid_ho);

	mlid_offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;
	mask_offset = mlid_offset + ( port / IB_MCAST_MASK_SIZE ) * IB_MCAST_BLOCK_SIZE;
	block_num = (uint16_t) ((mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE);
	bit_mask = ~cl_ntoh16((uint16_t) (1 << (port % IB_MCAST_MASK_SIZE)));
	(*p_tbl->p_mask_tbl)[block_num][mask_offset] &= bit_mask;
}

int osm_mcast_tbl_realloc(IN osm_mcast_tbl_t * p_tbl, IN unsigned mlid_offset)
{
	size_t mft_depth, size, old_size;
	uint16_t (*p_mask_tbl)[][IB_MCAST_BLOCK_SIZE*(IB_MCAST_POSITION_MAX + 1)];

	if (mlid_offset < p_tbl->mft_depth)
		goto done;

	/*
	   The number of bytes needed in the mask table is:
	   The (maximum bit mask 'position' + 1) times the
	   number of bytes in each bit mask times the
	   number of MLIDs supported by the table.

	   We must always allocate the array with the maximum position
	   since it is (and must be) defined that way the table structure
	   in order to create a pointer to a two dimensional array.
	 */
	mft_depth = (mlid_offset / IB_MCAST_BLOCK_SIZE + 1) * IB_MCAST_BLOCK_SIZE;
	size = mft_depth * (IB_MCAST_POSITION_MAX + 1) * IB_MCAST_MASK_SIZE / 8;
	old_size =  p_tbl->mft_depth * (IB_MCAST_POSITION_MAX + 1) * IB_MCAST_MASK_SIZE / 8;
	p_mask_tbl = realloc(p_tbl->p_mask_tbl, size);
	if (!p_mask_tbl)
		return -1;
	memset((uint8_t *)p_mask_tbl + old_size,
	       0, size - old_size);
	p_tbl->p_mask_tbl = p_mask_tbl;

	p_mask_tbl = realloc(p_tbl->p_curr_mask_tbl, size);
	if (!p_mask_tbl)
		return -1;
	memset((uint8_t *)p_mask_tbl + old_size,
		0, size - old_size);
	p_tbl->p_curr_mask_tbl = p_mask_tbl;

	p_tbl->mft_depth = mft_depth;
done:
	p_tbl->max_mlid_ho = mlid_offset + IB_LID_MCAST_START_HO;
	return 0;
}

boolean_t osm_mcast_tbl_is_port(IN const osm_mcast_tbl_t * p_tbl,
				IN uint16_t mlid_ho, IN uint8_t port_num)
{
	unsigned mlid_offset, mask_offset, bit_mask;
	uint16_t block_num;

	OSM_ASSERT(p_tbl);

	if (p_tbl->p_mask_tbl) {
		OSM_ASSERT(port_num <=
			  (p_tbl->max_position + 1) * IB_MCAST_MASK_SIZE);
		OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);
		OSM_ASSERT(mlid_ho <= p_tbl->max_mlid_ho);

		mlid_offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;
		mask_offset = mlid_offset + ( port_num / IB_MCAST_MASK_SIZE ) * IB_MCAST_BLOCK_SIZE;
		block_num = (uint16_t) ((mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE);
		bit_mask = cl_ntoh16((uint16_t)
				     (1 << (port_num % IB_MCAST_MASK_SIZE)));
		return (((*p_tbl->
			  p_mask_tbl)[block_num][mask_offset] & bit_mask) ==
			bit_mask);
	}

	return FALSE;
}

int osm_mcast_tbl_get_sole_active_port(IN const osm_mcast_tbl_t * p_tbl,
				       IN uint16_t mlid_ho)
{
	unsigned block_num;
	uint8_t position;
	unsigned offset = 0;
	uint16_t value;
	int ports_offset = 0;
	int result = 0;

	OSM_ASSERT(p_tbl && p_tbl->p_mask_tbl);
	OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);
	OSM_ASSERT(mlid_ho <= p_tbl->max_mlid_ho);

	block_num = (mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE;
	offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;

	for (position = 0; position <= p_tbl->max_position; position++) {
		value = cl_ntoh16((*p_tbl->p_mask_tbl)[block_num][offset]);
		if (value) {
			if (result || (value & (value - 1)))
				return -1;
			result = ffs(value) + ports_offset;
		}
		offset += IB_MCAST_BLOCK_SIZE;
		ports_offset += IB_MCAST_MASK_SIZE;
	}

	return result - 1;
}

boolean_t osm_mcast_tbl_is_any_port(IN const osm_mcast_tbl_t * p_tbl,
				    IN uint16_t mlid_ho)
{
	unsigned mlid_offset;
	uint8_t position;
	uint16_t result = 0;
	uint16_t block_num;

	OSM_ASSERT(p_tbl);

	if (p_tbl->p_mask_tbl) {
		OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);
		OSM_ASSERT(mlid_ho <= p_tbl->max_mlid_ho);

		mlid_offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;
		block_num = (uint16_t) ((mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE);

		for (position = 0; position <= p_tbl->max_position; position++)
			result |= (*p_tbl->p_mask_tbl)[block_num][mlid_offset + position * IB_MCAST_BLOCK_SIZE];
	}

	return (result != 0);
}

ib_api_status_t osm_mcast_tbl_set_block(IN osm_mcast_tbl_t * p_tbl,
					IN const ib_net16_t * p_block,
					IN int16_t block_num,
					IN uint8_t position)
{
	uint16_t mlid_start_ho;

	OSM_ASSERT(p_tbl);
	OSM_ASSERT(p_block);

	if (block_num > p_tbl->max_block)
		return IB_INVALID_PARAMETER;

	if (position > p_tbl->max_position)
		return IB_INVALID_PARAMETER;

	mlid_start_ho = (uint16_t) (block_num * IB_MCAST_BLOCK_SIZE);

	if (mlid_start_ho + IB_MCAST_BLOCK_SIZE - 1 > p_tbl->mft_depth)
		return IB_INVALID_PARAMETER;

	memcpy(&(*p_tbl->p_curr_mask_tbl)[block_num][position * IB_MCAST_BLOCK_SIZE],p_block,IB_MCAST_BLOCK_SIZE * IB_MCAST_MASK_SIZE / 8);
	if (block_num > p_tbl->max_block_in_use)
		p_tbl->max_block_in_use = (uint16_t) block_num;

	/* Blocked received, cleanup the relevant bit in need_update_mask */
	p_tbl->need_update_mask[block_num] &= ~(1 << position);
	return IB_SUCCESS;
}

void osm_mcast_tbl_clear_mlid(IN osm_mcast_tbl_t * p_tbl, IN uint16_t mlid_ho)
{
	unsigned mlid_offset;
	uint16_t block_num, mft_depth;
	uint32_t i;

	OSM_ASSERT(p_tbl);
	OSM_ASSERT(mlid_ho >= IB_LID_MCAST_START_HO);

	mft_depth = mlid_ho - IB_LID_MCAST_START_HO;
	mlid_offset = (mlid_ho - IB_LID_MCAST_START_HO) % IB_MCAST_BLOCK_SIZE;
	block_num = (uint16_t) ((mlid_ho - IB_LID_MCAST_START_HO) / IB_MCAST_BLOCK_SIZE);

	if (p_tbl->p_mask_tbl && mft_depth < p_tbl->mft_depth)
		for (i=0;i<IB_MCAST_POSITION_MAX + 1; i++)
			memset(&(*p_tbl->p_mask_tbl)[block_num][i * IB_MCAST_BLOCK_SIZE + mlid_offset],
			       0, IB_MCAST_MASK_SIZE / 8);
}

boolean_t osm_mcast_tbl_cmpr_block(IN osm_mcast_tbl_t * p_tbl,
				   IN int16_t block_num, IN uint8_t position,
				   ib_net16_t *p_block)
{
	OSM_ASSERT(p_tbl);
	OSM_ASSERT(p_block);
	OSM_ASSERT(block_num * IB_MCAST_BLOCK_SIZE <= p_tbl->mft_depth);
	if (position > p_tbl->max_position)
		return FALSE;

	if (memcmp(p_block,
		   &(*p_tbl->p_curr_mask_tbl)[block_num][position * IB_MCAST_BLOCK_SIZE],
		   IB_MCAST_BLOCK_SIZE * IB_MCAST_MASK_SIZE / 8))
		return FALSE;

	return TRUE;
}

boolean_t osm_mcast_tbl_get_block(IN osm_mcast_tbl_t * p_tbl,
				  IN int16_t block_num, IN uint8_t position,
				  IN osm_mcast_tbl_block_modifier_enum modifier,
				  OUT ib_net16_t * p_block)
{
	uint16_t * p_curr_block;
	int i;

	OSM_ASSERT(p_tbl);
	OSM_ASSERT(p_block);

	if (block_num > p_tbl->max_block_in_use)
		return FALSE;

	if (position > p_tbl->max_position) {
		/*
		   Caller shouldn't do this for efficiency's sake...
		 */
		memset(p_block, 0, IB_SMP_DATA_SIZE);
		return TRUE;
	}

	OSM_ASSERT(block_num * IB_MCAST_BLOCK_SIZE <= p_tbl->mft_depth);

	memcpy(p_block,&(*p_tbl->p_mask_tbl)[block_num][position*IB_MCAST_BLOCK_SIZE],IB_MCAST_BLOCK_SIZE * IB_MCAST_MASK_SIZE / 8);

	if (modifier == OSM_MCAST_TBL_UNCHANGED_ENTRIES) {
		/*
		 * Retrieve only the entries that are equal in the
		 * corresponding block in  p_tbl->p_mask_tbl and
		 * p_tbl->p_curr_mask_tbl.
		 *
		 * This is used by mcast_mgr to clear removed MFT entries before
		 * adding new ones in order to prevent multicast loops that may
		 * occur during MFT assignment.
		 */
		p_curr_block =
		    &(*p_tbl->p_curr_mask_tbl)[block_num][position * IB_MCAST_BLOCK_SIZE];
		for (i = 0; i < IB_MCAST_BLOCK_SIZE; i++) {
			p_block[i] &= p_curr_block[i];
		}
	}

	return TRUE;
}
