/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of multicast functions.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MULTICAST_C
#define INCREASE_MCM TRUE
#define DECREASE_MCM FALSE
#include <opensm/osm_multicast.h>
#include <opensm/osm_mcm_port.h>
#include <opensm/osm_mtree.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_partition.h>

static osm_mgrp_box_t *mgrp_box_new(uint16_t mlid)
{
	osm_mgrp_box_t *mbox = malloc(sizeof(*mbox));
	if (!mbox)
		return NULL;

	memset(mbox, 0, sizeof(*mbox));

	cl_qmap_init(&mbox->mtree_nodes_tbl);
	mbox->mlid = mlid;
	cl_qlist_init(&mbox->mgrp_list);
	mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;

	return mbox;
}

void osm_mgrp_box_clear_mtree_nodes_tbl(osm_mgrp_box_t * mbox)
{
	cl_map_item_t *p_item;
	while (cl_qmap_count(&mbox->mtree_nodes_tbl)) {
		p_item = cl_qmap_head(&mbox->mtree_nodes_tbl);
		cl_qmap_remove_item(&mbox->mtree_nodes_tbl,p_item);
		free(p_item);
	}
}

void osm_mgrp_box_clear_mtree(osm_subn_t * p_subn, osm_mgrp_box_t * mbox)
{
	mcast_mgr_decrease_mtree_root_weight(p_subn, mbox);
	cl_qmap_remove_all(&mbox->mtree_nodes_tbl);
	osm_mcast_mgr_decrease_weights(p_subn, mbox);
	osm_mtree_destroy(mbox->root);
	mbox->root = NULL;
	mbox->root_guid = 0;
}

void mgrp_box_delete(osm_subn_t * subn, osm_mgrp_box_t * mbox)
{
	osm_mgrp_box_clear_mtree(subn, mbox);
	free(mbox);
}

static void mgrp_handle_mcm_counter(osm_mcm_alias_guid_t *p_mcm_aguid,
				    osm_subn_t *subn, boolean_t is_increase)
{
	osm_alias_guid_t *p_aguid;

	/* mcm_counter is relevent only if sa_enhanced_trust_model is enabled */
	if (!subn->opt.sa_enhanced_trust_model)
		return;

	p_aguid = (osm_alias_guid_t *)
		  cl_qmap_get(&subn->alias_port_guid_tbl,
			      p_mcm_aguid->alias_guid);

	/* If there is no p_aguid matching the alias guid we do nothing */
	if (p_aguid ==
	    (osm_alias_guid_t *)cl_qmap_end(&subn->alias_port_guid_tbl))
		return;

	if (is_increase) {
		/* Increase the counter of multicast groups for the alias guid */
		p_aguid->mcm_counter++;
	} else {
		/* Decrease the counter of multicast groups for the alias guid */
		OSM_ASSERT(p_aguid->mcm_counter > 0);
		p_aguid->mcm_counter--;
	}
}

void mgrp_delete(IN osm_mgrp_t * p_mgrp, osm_subn_t * subn)
{
	osm_mcm_alias_guid_t *p_mcm_alias_guid;
	osm_mcm_port_t *p_mcm_port;

	OSM_ASSERT(p_mgrp);

	while (cl_qmap_count(&p_mgrp->mcm_alias_port_tbl)) {
		p_mcm_alias_guid = (osm_mcm_alias_guid_t *)
				   cl_qmap_head(&p_mgrp->mcm_alias_port_tbl);
		mgrp_handle_mcm_counter(p_mcm_alias_guid, subn, DECREASE_MCM);
		cl_qmap_remove_item(&p_mgrp->mcm_alias_port_tbl,
				    &p_mcm_alias_guid->map_item);
		osm_mcm_alias_guid_delete(&p_mcm_alias_guid);
	}

	while (cl_qmap_count(&p_mgrp->mcm_port_tbl)) {
		p_mcm_port =
		    (osm_mcm_port_t *)cl_qmap_head(&p_mgrp->mcm_port_tbl);
		cl_qmap_remove_item(&p_mgrp->mcm_port_tbl,
				    &p_mcm_port->map_item);
		cl_qlist_remove_item(&p_mcm_port->port->mcm_list,
				     &p_mcm_port->list_item);
		osm_mcm_port_delete(p_mcm_port);
	}

	free(p_mgrp);
}

void osm_mgrp_box_delete(osm_subn_t * subn, osm_mgrp_box_t *mbox)
{
	osm_mgrp_t *mgrp;

	while (cl_qlist_count(&mbox->mgrp_list)) {
		mgrp = cl_item_obj(cl_qlist_remove_head(&mbox->mgrp_list),
				   mgrp, list_item);
		mgrp_delete(mgrp, subn);
	}
	mgrp_box_delete(subn, mbox);
}

static void post_mgc_change_event(osm_sm_t *p_sm, ib_net16_t mlid,
				  osm_activity_mgr_change_enum change,
				  char *gid_str)
{
	event_data_t event_data;

	memset(&event_data, 0, sizeof(event_data_t));

	event_data.event_id = OSM_ACT_MGR_MCG_CHANGE;
	event_data.mlid = mlid;
	event_data.change = change;
	memcpy(&event_data.mgid, gid_str, strlen(gid_str));
	osm_act_mgr_post_event(&p_sm->activity_mgr, &event_data);
}

static void post_mcm_member_event(osm_sm_t *p_sm, ib_net16_t mlid,
				  ib_net64_t port_guid, char *gid_str,
				  osm_act_mgr_mc_state_enum join_state)
{
	event_data_t event_data;

	memset(&event_data, 0, sizeof(event_data_t));

	event_data.event_id = OSM_ACT_MGR_MCM_MEMBER;
	event_data.mlid = mlid;
	event_data.port_guid = port_guid;
	event_data.join_state = join_state;
	memcpy(&event_data.mgid, gid_str, strlen(gid_str));
	osm_act_mgr_post_event(&p_sm->activity_mgr, &event_data);
}

osm_mgrp_t *osm_mgrp_new(IN osm_subn_t * subn, IN ib_net16_t mlid,
			 IN ib_member_rec_t * mcmr)
{
	osm_mgrp_t *p_mgrp;
	osm_mgrp_box_t *mbox;
	osm_prtn_t *p_prtn;
	char gid_str[INET6_ADDRSTRLEN];

	p_prtn = osm_prtn_find_by_pkey(subn, mcmr->pkey);
	if (!p_prtn) {
		OSM_LOG(&subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7603: Failed to find partition object "
			"for pkey 0x%04x\n",
			cl_ntoh16(mcmr->pkey));
		return NULL;
	}

	p_mgrp = (osm_mgrp_t *) malloc(sizeof(*p_mgrp));
	if (!p_mgrp)
		return NULL;

	memset(p_mgrp, 0, sizeof(*p_mgrp));
	cl_qmap_init(&p_mgrp->mcm_port_tbl);
	cl_qmap_init(&p_mgrp->mcm_alias_port_tbl);
	p_mgrp->mlid = mlid;
	p_mgrp->mcmember_rec = *mcmr;

	mbox = osm_get_mbox_by_mlid(subn, p_mgrp->mlid);
	if (!mbox && !(mbox = mgrp_box_new(cl_ntoh16(p_mgrp->mlid)))) {
		free(p_mgrp);
		return NULL;
	}

	osm_prtn_link_mgrp(p_prtn, p_mgrp);

	OSM_LOG(&subn->p_osm->log, OSM_LOG_VERBOSE,
		"Add MGID 0x%016" PRIx64 ":0x%016" PRIx64 " to MLID 0x%04X\n",
		cl_ntoh64(p_mgrp->mcmember_rec.mgid.unicast.prefix),
		cl_ntoh64(p_mgrp->mcmember_rec.mgid.unicast.interface_id),
		cl_ntoh16(p_mgrp->mlid));
	cl_qlist_insert_tail(&mbox->mgrp_list, &p_mgrp->list_item);
	osm_subn_set_mbox(subn, mbox);

	inet_ntop(AF_INET6, p_mgrp->mcmember_rec.mgid.raw, gid_str,
		  sizeof(gid_str));

	post_mgc_change_event(&subn->p_osm->sm, mlid, OSM_ACT_MGR_CREATE,
			      gid_str);

	cl_fmap_insert(&subn->mgrp_mgid_tbl, &p_mgrp->mcmember_rec.mgid,
		       &p_mgrp->map_item);


	subn->p_osm->sa.dirty = TRUE;
	return p_mgrp;
}

static void mgrp_send_notice(osm_subn_t * subn, osm_log_t * log,
			     osm_mgrp_t * mgrp, unsigned num)
{
	ib_mad_notice_attr_t notice;
	ib_api_status_t status;

	notice.generic_type = 0x83;	/* generic SubnMgt type */
	ib_notice_set_prod_type_ho(&notice, 4);	/* A Class Manager generator */
	notice.g_or_v.generic.trap_num = CL_HTON16(num);
	/* The sm_base_lid is saved in network order already. */
	notice.issuer_lid = subn->sm_base_lid;
	/* following o14-12.1.11 and table 120 p726 */
	/* we need to provide the MGID */
	memcpy(&notice.data_details.ntc_64_67.gid,
	       &mgrp->mcmember_rec.mgid, sizeof(ib_gid_t));

	/* According to page 653 - the issuer gid in this case of trap
	   is the SM gid, since the SM is the initiator of this trap. */
	notice.issuer_gid.unicast.prefix = subn->opt.subnet_prefix;
	notice.issuer_gid.unicast.interface_id = subn->sm_port_guid;

	if ((status = osm_report_notice(log, subn, &notice,
					(uint64_t) cl_atomic_inc(&subn->tid) & (uint64_t)(0xFFFFFFFF),
					0)))
		OSM_LOG(log, OSM_LOG_ERROR, "ERR 7601: "
			"Error sending trap reports (%s)\n",
			ib_get_err_str(status));
}

void osm_mgrp_cleanup(osm_subn_t * subn, osm_mgrp_t * mgrp)
{
	osm_mgrp_box_t *mbox;
	osm_mcm_alias_guid_t *mcm_alias_guid;
	osm_mcm_port_t *mcm_port;
	char gid_str[INET6_ADDRSTRLEN];

	if (mgrp->full_members && !osm_exit_flag) {
		mgrp_send_notice(subn, &subn->p_osm->log, mgrp, 67);
	}

	if (mgrp->well_known)
		return;

	mgrp->full_members = 0;

	while (cl_qmap_count(&mgrp->mcm_alias_port_tbl)) {
		mcm_alias_guid = (osm_mcm_alias_guid_t *) cl_qmap_head(&mgrp->mcm_alias_port_tbl);
		mgrp_handle_mcm_counter(mcm_alias_guid, subn, DECREASE_MCM);
		cl_qmap_remove_item(&mgrp->mcm_alias_port_tbl, &mcm_alias_guid->map_item);
		osm_mcm_alias_guid_delete(&mcm_alias_guid);
	}

	while (cl_qmap_count(&mgrp->mcm_port_tbl)) {
		mcm_port = (osm_mcm_port_t *) cl_qmap_head(&mgrp->mcm_port_tbl);
		cl_qmap_remove_item(&mgrp->mcm_port_tbl, &mcm_port->map_item);
		cl_qlist_remove_item(&mcm_port->port->mcm_list,
				     &mcm_port->list_item);
		osm_mcm_port_delete(mcm_port);
	}

	cl_fmap_remove_item(&subn->mgrp_mgid_tbl, &mgrp->map_item);

	OSM_LOG(&subn->p_osm->log, OSM_LOG_VERBOSE,
		"Remove MGID 0x%016" PRIx64 ":0x%016" PRIx64 " to MLID 0x%04X\n",
		cl_ntoh64(mgrp->mcmember_rec.mgid.unicast.prefix),
		cl_ntoh64(mgrp->mcmember_rec.mgid.unicast.interface_id),
		cl_ntoh16(mgrp->mlid));

	inet_ntop(AF_INET6, mgrp->mcmember_rec.mgid.raw, gid_str,
		  sizeof(gid_str));

	post_mgc_change_event(&subn->p_osm->sm, mgrp->mlid, OSM_ACT_MGR_DELETE,
			      gid_str);

	mbox = osm_get_mbox_by_mlid(subn, mgrp->mlid);
	cl_qlist_remove_item(&mbox->mgrp_list, &mgrp->list_item);
	if (cl_is_qlist_empty(&mbox->mgrp_list)) {
		osm_subn_remove_mbox(subn, mbox->mlid);
		mgrp_box_delete(subn, mbox);
	}

	osm_prtn_unlink_mgrp(mgrp->p_prtn, mgrp);

	free(mgrp);

	subn->p_osm->sa.dirty = TRUE;
}

static boolean_t is_qmap_empty_for_port(IN const cl_qmap_t * const p_map,
					IN const osm_port_t *port)
{
	size_t count = 0;
	cl_map_item_t *item;
	osm_mcm_alias_guid_t *mcm_alias_guid;

	for (item = cl_qmap_head(p_map); item != cl_qmap_end(p_map);
	     item = cl_qmap_next(item)) {
		mcm_alias_guid = (osm_mcm_alias_guid_t *) item;
		if (mcm_alias_guid->p_base_mcm_port->port == port) {
			count++;
			break;
		}
	}

	return (count == 0);
}

static boolean_t is_qmap_empty_for_mcm_port(IN const cl_qmap_t * const p_map,
					    IN const osm_mcm_port_t *mcm_port)
{
	size_t count = 0;
	cl_map_item_t *item;
	osm_mcm_alias_guid_t *mcm_alias_guid;

	for (item = cl_qmap_head(p_map); item != cl_qmap_end(p_map);
	     item = cl_qmap_next(item)) {
		mcm_alias_guid = (osm_mcm_alias_guid_t *) item;
		if (mcm_alias_guid->p_base_mcm_port == mcm_port) {
			count++;
			break;
		}
	}

	return (count == 0);
}
static osm_mcm_alias_guid_t *insert_alias_guid(IN osm_mgrp_t * mgrp,
					       IN osm_mcm_alias_guid_t * p_mcm_alias_guid)
{
	osm_mcm_alias_guid_t *p_mcm_alias_guid_check;

	/* insert into mcm alias guid table */
	p_mcm_alias_guid_check =
		(osm_mcm_alias_guid_t *) cl_qmap_insert(&mgrp->mcm_alias_port_tbl,
							p_mcm_alias_guid->alias_guid,
							&p_mcm_alias_guid->map_item);
	if (p_mcm_alias_guid_check != (osm_mcm_alias_guid_t *) &p_mcm_alias_guid->map_item) {
		/* alias GUID is a duplicate */
		osm_mcm_alias_guid_delete(&p_mcm_alias_guid);
		return p_mcm_alias_guid_check;
	}
	return NULL;
}

static void osm_mgrp_update_change(IN osm_subn_t * p_subn,
				   IN osm_mgrp_box_t * mbox,
				   IN osm_port_t * p_port)
{
	cl_list_item_t *p_list_item;
	cl_map_item_t *p_map_item;
	osm_node_t *p_node;
	osm_node_t *p_remote_node;
	osm_mgrp_t * p_mgrp;
	osm_mcm_port_t *p_mcm_port;
	osm_mtree_node_t *p_mtn, *p_child, *p_child2;
	ib_net64_t port_guid;
	ib_net64_t remote_node_guid;
	uint8_t port_num;

	if (mbox->imr_state == OSM_MGRP_BOX_IMR_NOT_SUPPORTED ||
	    p_subn->force_multicast_rerouting)
		return;

	/* Find the GUID of the switch connected to the port */
	p_node = p_port->p_node;
	if (p_port->p_node->sw) {
		port_num = 0;
		p_remote_node = p_node;
	} else {
		port_num = osm_physp_get_port_num(osm_port_get_physp(p_port));
		p_remote_node = osm_node_get_remote_node(p_node, port_num,
							 &port_num);
	}

	/* node does not exists or already removed by drop manager */
	if (!p_remote_node)
		return;

	remote_node_guid = osm_node_get_node_guid(p_remote_node);

	/* Find mtree_node_t object of the switch of the port */
	p_map_item = cl_qmap_get(&mbox->mtree_nodes_tbl, remote_node_guid);
	if (p_map_item == cl_qmap_end(&mbox->mtree_nodes_tbl)) {
		mbox->imr_state = OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
		return;
	}

	p_mtn = (osm_mtree_node_t *)p_map_item;
	p_child = p_mtn->child_array[port_num];
	OSM_ASSERT(!p_child || p_child == OSM_MTREE_LEAF ||
		   p_child == OSM_MTREE_SENDER);

	/*
	 * Determine change operation to be done by checking osm_mcm_port_t
	 * objects corresponding with the port, and mbox objects.
	 *
	 * For performance, if mbox has more than one multicast group, iterate
	 * over osm_mcm_port_t objects from osm_port_t->mcm_list.
	 */
	p_child2 = NULL;
	if (cl_qlist_count(&mbox->mgrp_list) == 1) {
		port_guid = osm_port_get_guid(p_port);
		for (p_list_item = cl_qlist_head(&mbox->mgrp_list);
		     p_list_item != cl_qlist_end(&mbox->mgrp_list);
		     p_list_item = cl_qlist_next(p_list_item)) {
			p_mgrp = (osm_mgrp_t *)cl_item_obj(p_list_item, p_mgrp,
							   list_item);
			p_map_item = cl_qmap_get(&p_mgrp->mcm_port_tbl, port_guid);
			if (p_map_item != cl_qmap_end(&p_mgrp->mcm_port_tbl)) {
				if (((osm_mcm_port_t *)p_map_item)->num_receivers) {
					p_child2 = OSM_MTREE_LEAF;
					break;
				} else {
					p_child2 = OSM_MTREE_SENDER;
				}
			}
		}
	} else {
		for (p_list_item = cl_qlist_head(&p_port->mcm_list);
		     p_list_item != cl_qlist_end(&p_port->mcm_list);
		     p_list_item = cl_qlist_next(p_list_item)) {
	                p_mcm_port = cl_item_obj(p_list_item, p_mcm_port, list_item);
			if (p_mcm_port->mgrp->mlid != mbox->mlid)
				continue;

			if (p_mcm_port->num_receivers) {
				p_child2 = OSM_MTREE_LEAF;
				break;
			} else {
				p_child2 = OSM_MTREE_SENDER;
			}
		}
	}

	if (p_child != p_child2) {
		mbox->imr_tree_changed_flag = TRUE;
		p_mtn->child_array[port_num] = p_child2;
	}
}

osm_mcm_port_t *osm_mgrp_add_port(IN osm_subn_t * subn, osm_log_t * log,
				  IN osm_mgrp_t * mgrp, osm_port_t *port,
				  IN ib_member_rec_t *mcmr, IN boolean_t proxy)
{
	osm_mcm_port_t *mcm_port;
	osm_mcm_alias_guid_t *p_mcm_alias_guid, *p_mcm_alias_guid_check;
	cl_map_item_t *prev_item;
	osm_mgrp_box_t *mbox;
	uint8_t prev_join_state = 0, join_state = mcmr->scope_state;
	uint8_t prev_scope, full_state;
	boolean_t need_reroute = FALSE;
	char gid_str[INET6_ADDRSTRLEN];

	inet_ntop(AF_INET6, mgrp->mcmember_rec.mgid.raw, gid_str,
		  sizeof(gid_str));

	if (OSM_LOG_IS_ACTIVE_V2(log, OSM_LOG_VERBOSE))
		OSM_LOG(log, OSM_LOG_VERBOSE, "GUID 0x%016" PRIx64
			" Port 0x%016" PRIx64 " joining "
			"MC group %s (mlid 0x%x)\n",
			cl_ntoh64(mcmr->port_gid.unicast.interface_id),
			cl_ntoh64(port->guid), gid_str, cl_ntoh16(mgrp->mlid));

	post_mcm_member_event(&subn->p_osm->sm, mgrp->mlid, port->guid, gid_str,
			      OSM_ACT_MGR_MC_JOIN);

	mcm_port = osm_mcm_port_new(port, mgrp);
	if (!mcm_port)
		return NULL;

	p_mcm_alias_guid = osm_mcm_alias_guid_new(mcm_port, mcmr, proxy);
	if (!p_mcm_alias_guid) {
		osm_mcm_port_delete(mcm_port);
		return NULL;
	}

	/*
	   prev_item = cl_qmap_insert(...)
	   Pointer to the item in the map with the specified key.  If insertion
	   was successful, this is the pointer to the item.  If an item with the
	   specified key already exists in the map, the pointer to that item is
	   returned.
	 */
	prev_item = cl_qmap_insert(&mgrp->mcm_port_tbl, port->guid,
				   &mcm_port->map_item);

	if (prev_item != &mcm_port->map_item) {	/* mcm port already exists */
		osm_mcm_port_delete(mcm_port);
		mcm_port = (osm_mcm_port_t *) prev_item;

		p_mcm_alias_guid->p_base_mcm_port = (osm_mcm_port_t *) prev_item;
		p_mcm_alias_guid_check = insert_alias_guid(mgrp, p_mcm_alias_guid);
		if (p_mcm_alias_guid_check) {	/* alias GUID already exists */
			p_mcm_alias_guid = p_mcm_alias_guid_check;
			ib_member_get_scope_state(p_mcm_alias_guid->scope_state,
						  &prev_scope, &prev_join_state);
			p_mcm_alias_guid->scope_state =
			    ib_member_set_scope_state(prev_scope,
						      prev_join_state | join_state);
		} else {
			mgrp_handle_mcm_counter(p_mcm_alias_guid, subn, INCREASE_MCM);
		}
	} else {
	   /*
	      Check if the p_mcm_alias_guid has been freed in the case of alias
	      GUID is duplicate.
		*/
		p_mcm_alias_guid_check = insert_alias_guid(mgrp, p_mcm_alias_guid);
		if (p_mcm_alias_guid_check)
			p_mcm_alias_guid = p_mcm_alias_guid_check;

		cl_qlist_insert_tail(&port->mcm_list, &mcm_port->list_item);

		need_reroute = TRUE;

		mgrp_handle_mcm_counter(p_mcm_alias_guid, subn, INCREASE_MCM);
	}

	/* o15.0.1.11: copy the join state */
	mcmr->scope_state = p_mcm_alias_guid->scope_state;

	/* Update number of receivers on mcm_port */
	if ((join_state & (IB_JOIN_STATE_FULL | IB_JOIN_STATE_NON)) &&
	    !(prev_join_state & (IB_JOIN_STATE_FULL | IB_JOIN_STATE_NON)))
		/*
		 * Increase number of receiving members, and decide whether to
		 * reroute multicast tree.
		 */
		if (++mcm_port->num_receivers == 1)
			need_reroute = TRUE;

	/* update mgrp_box_t changes tbl */
	mbox = osm_get_mbox_by_mlid(subn, mgrp->mlid);
	osm_mgrp_update_change(subn, mbox, mcm_port->port);

	full_state = IB_JOIN_STATE_FULL | IB_JOIN_STATE_SEND_ONLY_FULL;

	if ((join_state & full_state) && !(prev_join_state & full_state) &&
	    ++mgrp->full_members == 1) {
		mgrp_send_notice(subn, log, mgrp, 66);
		need_reroute = TRUE;
	}

	if (need_reroute)
		osm_sm_reroute_mlid(&subn->p_osm->sm, mgrp->mlid);

	subn->p_osm->sa.dirty = TRUE;
	return mcm_port;
}

boolean_t osm_mgrp_remove_port(osm_subn_t * subn, osm_log_t * log, osm_mgrp_t * mgrp,
			  osm_mcm_alias_guid_t * mcm_alias_guid,
			  ib_member_rec_t *mcmr)
{
	uint8_t join_state = mcmr->scope_state & 0xf;
	uint8_t port_join_state, new_join_state, full_state;
	osm_mgrp_box_t *mbox;
	osm_port_t *p_port = mcm_alias_guid->p_base_mcm_port->port;
	boolean_t mgrp_deleted = FALSE;
	ib_net16_t mlid = mgrp->mlid;
	boolean_t need_reroute = FALSE;
	char gid_str[INET6_ADDRSTRLEN];

	/*
	 * according to the same o15-0.1.14 we get the stored
	 * JoinState and the request JoinState and they must be
	 * opposite to leave - otherwise just update it
	 */
	port_join_state = mcm_alias_guid->scope_state & 0x0F;
	new_join_state = port_join_state & ~join_state;

	inet_ntop(AF_INET6, mgrp->mcmember_rec.mgid.raw,
				  gid_str, sizeof(gid_str));

	if (OSM_LOG_IS_ACTIVE_V2(log, OSM_LOG_VERBOSE))
		OSM_LOG(log, OSM_LOG_VERBOSE,
			"GUID 0x%" PRIx64 " Port 0x%" PRIx64
			" leaving MC group %s (mlid 0x%x)\n",
			cl_ntoh64(mcm_alias_guid->alias_guid),
			cl_ntoh64(mcm_alias_guid->p_base_mcm_port->port->guid),
			gid_str, cl_ntoh16(mlid));

	post_mcm_member_event(&subn->p_osm->sm, mlid, mcm_alias_guid->alias_guid,
			      gid_str, OSM_ACT_MGR_MC_LEAVE);

	/* Update number of receivers on mcm_port */
	if ((port_join_state & (IB_JOIN_STATE_FULL | IB_JOIN_STATE_NON)) &&
	    !(new_join_state & (IB_JOIN_STATE_FULL | IB_JOIN_STATE_NON))) {
		OSM_ASSERT(mcm_alias_guid->p_base_mcm_port);
		/*
		 * Decrease number of receiving members, and decide whether to
		 * reroute multicast tree.
		 */
		if (--mcm_alias_guid->p_base_mcm_port->num_receivers == 0)
			need_reroute = TRUE;
	}

	if (new_join_state & IB_JOIN_STATE_FULL ||
	    (new_join_state &&
	     (mgrp->full_members > (port_join_state & IB_JOIN_STATE_FULL) ? 1 : 0))) {
		mcm_alias_guid->scope_state =
		    new_join_state | (mcm_alias_guid->scope_state & 0xf0);
		OSM_LOG(log, OSM_LOG_DEBUG,
			"updating GUID 0x%" PRIx64 " port 0x%" PRIx64
			" JoinState 0x%x -> 0x%x\n",
			cl_ntoh64(mcm_alias_guid->alias_guid),
			cl_ntoh64(mcm_alias_guid->p_base_mcm_port->port->guid),
			port_join_state, new_join_state);
		mcmr->scope_state = mcm_alias_guid->scope_state;
	} else {
		mcmr->scope_state = mcm_alias_guid->scope_state & 0xf0;
		OSM_LOG(log, OSM_LOG_DEBUG, "removing alias GUID 0x%" PRIx64 "\n",
			cl_ntoh64(mcm_alias_guid->alias_guid));
		cl_qmap_remove_item(&mgrp->mcm_alias_port_tbl,
				    &mcm_alias_guid->map_item);
		mgrp_handle_mcm_counter(mcm_alias_guid, subn, DECREASE_MCM);

		if (is_qmap_empty_for_port(&mgrp->mcm_alias_port_tbl,
					   mcm_alias_guid->p_base_mcm_port->port)) { /* last alias in mcast group for this port */
			OSM_LOG(log, OSM_LOG_DEBUG, "removing port 0x%" PRIx64 "\n",
				cl_ntoh64(mcm_alias_guid->p_base_mcm_port->port->guid));
			cl_qmap_remove_item(&mgrp->mcm_port_tbl,
					    &mcm_alias_guid->p_base_mcm_port->map_item);
			cl_qlist_remove_item(&mcm_alias_guid->p_base_mcm_port->port->mcm_list,
					     &mcm_alias_guid->p_base_mcm_port->list_item);
			if (is_qmap_empty_for_mcm_port(&mgrp->mcm_alias_port_tbl,
						       mcm_alias_guid->p_base_mcm_port)) {
				/* last alias in mcast group for this mcm port */
				osm_mcm_port_delete(mcm_alias_guid->p_base_mcm_port);
			}

			need_reroute = TRUE;
		}
		osm_mcm_alias_guid_delete(&mcm_alias_guid);
	}

	/* update mgrp_box_t changes tbl */
	mbox = osm_get_mbox_by_mlid(subn, mgrp->mlid);
	osm_mgrp_update_change(subn, mbox, p_port);

	full_state = IB_JOIN_STATE_FULL | IB_JOIN_STATE_SEND_ONLY_FULL;

	/* no more full members so the group will be deleted after re-route
	   but only if it is not a well known group */
	if ((port_join_state & full_state) && !(new_join_state & full_state) &&
	    --mgrp->full_members == 0) {
		mgrp_send_notice(subn, log, mgrp, 67);
		osm_mgrp_cleanup(subn, mgrp);
		mgrp_deleted = TRUE;
		need_reroute = TRUE;
	}

	subn->p_osm->sa.dirty = TRUE;

	if (need_reroute)
		osm_sm_reroute_mlid(&subn->p_osm->sm, mlid);

	return (mgrp_deleted);
}

boolean_t osm_mgrp_remove_mcm_alias_guid(IN osm_subn_t * subn, osm_log_t * log,
					 IN osm_mgrp_t * mgrp,
					 IN osm_mcm_alias_guid_t * mcm_alias_guid)
{
	ib_member_rec_t mcmrec;

	mcmrec.scope_state = 0xf;
	return (osm_mgrp_remove_port(subn, log, mgrp, mcm_alias_guid, &mcmrec));
}

void osm_mgrp_delete_port(osm_subn_t * subn, osm_log_t * log, osm_mgrp_t * mgrp,
			  osm_port_t * port)
{
	osm_mcm_alias_guid_t *mcm_alias_guid, *next_mcm_alias_guid;
	boolean_t mgrp_deleted = FALSE;

	next_mcm_alias_guid = (osm_mcm_alias_guid_t *) cl_qmap_head(&mgrp->mcm_alias_port_tbl);
	while (!mgrp_deleted &&
	       next_mcm_alias_guid != (osm_mcm_alias_guid_t *) cl_qmap_end(&mgrp->mcm_alias_port_tbl)) {
		mcm_alias_guid = next_mcm_alias_guid;
		next_mcm_alias_guid = (osm_mcm_alias_guid_t *) cl_qmap_next(&next_mcm_alias_guid->map_item);
		if (mcm_alias_guid->p_base_mcm_port->port == port)
			mgrp_deleted =
			    osm_mgrp_remove_mcm_alias_guid(subn, log, mgrp,
							   mcm_alias_guid);
	}
}

osm_mcm_port_t *osm_mgrp_get_mcm_port(IN const osm_mgrp_t * p_mgrp,
				      IN ib_net64_t port_guid)
{
	cl_map_item_t *item = cl_qmap_get(&p_mgrp->mcm_port_tbl, port_guid);
	if (item != cl_qmap_end(&p_mgrp->mcm_port_tbl))
		return (osm_mcm_port_t *) item;
	return NULL;
}

osm_mcm_alias_guid_t *osm_mgrp_get_mcm_alias_guid(IN const osm_mgrp_t * p_mgrp,
						  IN ib_net64_t port_guid)
{
	cl_map_item_t *item = cl_qmap_get(&p_mgrp->mcm_alias_port_tbl,
					  port_guid);
	if (item != cl_qmap_end(&p_mgrp->mcm_alias_port_tbl))
		return (osm_mcm_alias_guid_t *) item;
	return NULL;
}
