/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_ni_rcv_t.
 * This object represents the NodeInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NODE_INFO_RCV_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_router.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_capability_mgr.h>

static void report_duplicated_guid(IN osm_sm_t * sm, osm_physp_t * p_physp,
				   osm_node_t * p_neighbor_node,
				   const uint8_t port_num)
{
	osm_physp_t *p_old, *p_new;
	osm_dr_path_t path;

	p_old = p_physp->p_remote_physp;
	p_new = osm_node_get_physp_ptr(p_neighbor_node, port_num);

	OSM_LOG(sm->p_log, OSM_LOG_SYS | OSM_LOG_ERROR, "ERR 0D01: "
		"Found duplicated node GUID.\n"
		"Node 0x%" PRIx64 " port %u is reachable from remote node "
		"0x%" PRIx64 " port %u and remote node 0x%" PRIx64 " port %u.\n"
		"Paths are:\n",
		cl_ntoh64(p_physp->p_node->node_info.node_guid),
		p_physp->port_num,
		p_old ? cl_ntoh64(p_old->p_node->node_info.node_guid) : 0,
		p_old ? p_old->port_num : 0,
		p_new ? cl_ntoh64(p_new->p_node->node_info.node_guid) : 0,
		p_new ? p_new->port_num : 0);

	osm_dump_dr_path_v2(sm->p_log, osm_physp_get_dr_path_ptr(p_physp),
			    FILE_ID, OSM_LOG_ERROR);

	path = *osm_physp_get_dr_path_ptr(p_new);
	if (osm_dr_path_extend(&path, port_num))
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D05: "
			"DR path with hop count %d couldn't be extended\n",
			path.hop_count);
	osm_dump_dr_path_v2(sm->p_log, &path, FILE_ID, OSM_LOG_ERROR);
}

static void requery_dup_node_info(IN osm_sm_t * sm, osm_physp_t * p_physp,
				  unsigned count)
{
	osm_madw_context_t context;
	osm_dr_path_t path;
	cl_status_t status;

	if (!p_physp->p_remote_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D0D: "
			"DR path couldn't be extended\n");
		return;
	}
	path = *osm_physp_get_dr_path_ptr(p_physp->p_remote_physp);
	if (osm_dr_path_extend(&path, p_physp->p_remote_physp->port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D08: "
			"DR path with hop count %d couldn't be extended\n",
			path.hop_count);
		return;
	}
	if (!osm_is_dr_path_valid(sm, &path)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D28: "
			"Probably outdated DR path \n");
		osm_dump_dr_path_v2(sm->p_log, &path, FILE_ID, OSM_LOG_ERROR);
		return;
	}

	context.ni_context.node_guid =
	    p_physp->p_remote_physp->p_node->node_info.port_guid;
	context.ni_context.port_num = p_physp->p_remote_physp->port_num;
	context.ni_context.dup_node_guid = p_physp->p_node->node_info.node_guid;
	context.ni_context.dup_port_num = p_physp->port_num;
	context.ni_context.dup_count = count;

	status = osm_req_get(sm, &path, IB_MAD_ATTR_NODE_INFO, 0,
			     TRUE, 0, 0, CL_DISP_MSGID_NONE, &context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D02: "
			"Failure initiating NodeInfo request (%s)\n",
			ib_get_err_str(status));
}

static ib_api_status_t validate_node_info_existing(osm_sm_t *sm,
						   ib_node_info_t *p_ni,
						   osm_node_t *p_node)
{
	if (PF(p_ni->num_ports != p_node->node_info.num_ports)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1F: "
			"Got Illegal num_ports %d different than stored %d\n",
			p_ni->num_ports,
			p_node->node_info.num_ports);
		return IB_ERROR;
	}

	if (PF(p_ni->node_type != p_node->node_info.node_type)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D20: "
			"Got Illegal node_type %d different than stored %d\n",
			p_ni->node_type,
			p_node->node_info.node_type);
		return IB_ERROR;
	}

	if (PF(p_ni->device_id != p_node->node_info.device_id)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D21: "
			"Got Illegal device_id 0x%x different than stored 0x%x\n",
			cl_ntoh16(p_ni->device_id),
			cl_ntoh16(p_node->node_info.device_id));
		return IB_ERROR;
	}

	if (PF(p_ni->partition_cap != p_node->node_info.partition_cap)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D22: "
			"Got Illegal partition_cap %d different than stored %d\n",
			cl_ntoh16(p_ni->partition_cap),
			cl_ntoh16(p_node->node_info.partition_cap));
		return IB_ERROR;
	}

	if (PF(p_ni->revision != p_node->node_info.revision)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D23: "
			"Got Illegal revision %d different than stored %d\n",
			cl_ntoh32(p_ni->revision),
			cl_ntoh32(p_node->node_info.revision));
		return IB_ERROR;
	}

	if (PF((p_ni->port_num_vendor_id & IB_NODE_INFO_VEND_ID_MASK) !=
	       (p_node->node_info.port_num_vendor_id & IB_NODE_INFO_VEND_ID_MASK))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D24: "
			"Got Illegal vendor_id %d different than stored %d\n",
			cl_ntoh32(p_ni->port_num_vendor_id & IB_NODE_INFO_VEND_ID_MASK),
			cl_ntoh32(p_node->node_info.port_num_vendor_id & IB_NODE_INFO_VEND_ID_MASK));
		return IB_ERROR;
	}

	return IB_SUCCESS;
}

static ib_api_status_t validate_node_info(osm_sm_t *sm, ib_node_info_t *p_ni,
					  ib_smp_t *p_smp, ib_net16_t dlid)
{
	char buf[BUF_SIZE];

	if (PF(p_ni->node_guid == 0)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D16: "
			"Got Zero Node GUID, %s\n",
			osm_get_smp_path_str(p_smp, buf, BUF_SIZE, dlid));
		return IB_ERROR;
	}

	if (PF(p_ni->port_guid == 0)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D17: "
			"Got Zero Port GUID, %s\n",
			osm_get_smp_path_str(p_smp, buf, BUF_SIZE, dlid));
		return IB_ERROR;
	}

	if (PF(p_ni->node_type != IB_NODE_TYPE_CA &&
	       p_ni->node_type != IB_NODE_TYPE_SWITCH &&
	       p_ni->node_type != IB_NODE_TYPE_ROUTER)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D19: "
			"Got Illegal NodeType %d, %s\n",
			p_ni->node_type,
			osm_get_smp_path_str(p_smp, buf, BUF_SIZE, dlid));
		return IB_ERROR;
	}

	if (PF(p_ni->num_ports == IB_INVALID_PORT_NUM)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D25: "
			"Got Illegal num_ports %d, %s\n",
			ib_node_info_get_local_port_num(p_ni),
			osm_get_smp_path_str(p_smp, buf, BUF_SIZE, dlid));
		return IB_ERROR;
	}

	if (PF(ib_node_info_get_local_port_num(p_ni) > p_ni->num_ports)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1E: "
			"Got Illegal port_num %d > num_ports %d, %s\n",
			ib_node_info_get_local_port_num(p_ni),
			p_ni->num_ports,
			osm_get_smp_path_str(p_smp, buf, BUF_SIZE, dlid));
		return IB_ERROR;
	}

	return IB_SUCCESS;
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_set_links(IN osm_sm_t * sm, osm_node_t * p_node,
			     const uint8_t port_num,
			     const osm_ni_context_t * p_ni_context)
{
	osm_node_t *p_neighbor_node;
	osm_physp_t *p_physp, *p_remote_physp;

	OSM_LOG_ENTER(sm->p_log);

	/*
	   A special case exists in which the node we're trying to
	   link is our own node.  In this case, the guid value in
	   the ni_context will be zero.
	 */
	if (p_ni_context->node_guid == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Nothing to link for our own node 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		goto _exit;
	}

	p_neighbor_node = osm_get_node_by_guid(sm->p_subn,
					       p_ni_context->node_guid);
	if (PF(!p_neighbor_node)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D10: "
			"Unexpected removal of neighbor node 0x%" PRIx64 "\n",
			cl_ntoh64(p_ni_context->node_guid));
		goto _exit;
	}

	/* When setting the link, ports on both
	   sides of the link should be initialized */
	if (!osm_node_link_has_valid_ports(p_node, port_num,
					   p_neighbor_node,
					   p_ni_context->port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1D: "
			"No physical port found for one(or both) "
			"of the nodes GUID 0x%" PRIx64 " port %u; "
			"GUID 0x%" PRIx64 " port %u. Might be "
			"duplicate port GUID\n",
			cl_ntoh64(p_node->node_info.node_guid),
			port_num,
			cl_ntoh64(p_neighbor_node->node_info.node_guid),
			p_ni_context->port_num);
		goto _exit;
	}

	if (osm_node_link_exists(p_node, port_num,
				 p_neighbor_node, p_ni_context->port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Link already exists\n");
		goto _exit;
	}

	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	if (PF(!p_physp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD06: "
			"Failed to find physp for port %d of Node GUID 0x%"
			PRIx64 "\n", port_num,
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		goto _exit;
	}

	/*
	 * If the link went UP, after we already discovered it, we shouldn't
	 * set the link between the ports and resweep.
	 */
	if (osm_physp_get_port_state(p_physp) == IB_LINK_DOWN &&
	    p_node->physp_discovered[port_num]) {
		/* Link down on another side. Don't create a link*/
		p_node->physp_discovered[port_num] = 0;
		sm->p_subn->force_heavy_sweep = TRUE;
		goto _exit;
	}

	if (osm_node_has_any_link(p_node, port_num) &&
	    sm->p_subn->force_heavy_sweep == FALSE &&
	    (!p_ni_context->dup_count ||
	     (p_ni_context->dup_node_guid == osm_node_get_node_guid(p_node) &&
	      p_ni_context->dup_port_num == port_num))) {
		/*
		   Uh oh...
		   This could be reconnected ports, but also duplicated GUID
		   (2 nodes have the same guid) or a 12x link with lane reversal
		   that is not configured correctly.
		   We will try to recover by querying NodeInfo again.
		   In order to catch even fast port moving to new location(s)
		   and back we will count up to 5.
		   Some crazy reconnections (newly created switch loop right
		   before targeted CA) will not be catched this way. So in worst
		   case - report GUID duplication and request new discovery.
		   When switch node is targeted NodeInfo querying will be done
		   in opposite order, this is much stronger check, unfortunately
		   it is impossible with CAs.
		 */
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (PF(!p_physp)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD07: "
				"Failed to find physp for port %d of Node GUID 0x%"
				PRIx64 "\n", port_num,
				cl_ntoh64(osm_node_get_node_guid(p_node)));
			goto _exit;
		}

		if (p_ni_context->dup_count > 5) {
			report_duplicated_guid(sm, p_physp, p_neighbor_node,
					       p_ni_context->port_num);
			sm->p_subn->force_heavy_sweep = TRUE;
		} else if (p_node->sw)
			requery_dup_node_info(sm, p_physp->p_remote_physp,
					      p_ni_context->dup_count + 1);
		else
			requery_dup_node_info(sm, p_physp,
					      p_ni_context->dup_count + 1);
	}

	/*
	   When there are only two nodes with exact same guids (connected back
	   to back) - the previous check for duplicated guid will not catch
	   them. But the link will be from the port to itself...
	   Enhanced Port 0 is an exception to this
	 */
	if (osm_node_get_node_guid(p_node) == p_ni_context->node_guid &&
	    port_num == p_ni_context->port_num &&
	    port_num != 0 && cl_qmap_count(&sm->p_subn->sw_guid_tbl) == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Duplicate GUID found by link from a port to itself:"
			"node 0x%" PRIx64 ", port number %u\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num);
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (PF(!p_physp)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD08: "
				"Failed to find physp for port %d of Node GUID 0x%"
				PRIx64 "\n", port_num,
				cl_ntoh64(osm_node_get_node_guid(p_node)));
			goto _exit;
		}

		osm_dump_dr_path_v2(sm->p_log, osm_physp_get_dr_path_ptr(p_physp),
				    FILE_ID, OSM_LOG_VERBOSE);

		if (sm->p_subn->opt.exit_on_fatal == TRUE) {
			osm_log_v2(sm->p_log, OSM_LOG_SYS, FILE_ID,
				   "Errors on subnet. Duplicate GUID found "
				   "by link from a port to itself. "
				   "See verbose opensm.log for more details\n");
			exit(1);
		}
	}

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Creating new link between:\n\t\t\t\tnode 0x%" PRIx64
		", port number %u and\n\t\t\t\tnode 0x%" PRIx64
		", port number %u\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)), port_num,
		cl_ntoh64(p_ni_context->node_guid), p_ni_context->port_num);

	if (sm->ucast_mgr.cache_valid)
		osm_ucast_cache_check_new_link(&sm->ucast_mgr, p_node, port_num,
					       p_neighbor_node,
					       p_ni_context->port_num);

	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	p_remote_physp = osm_node_get_physp_ptr(p_neighbor_node,
						p_ni_context->port_num);
	if (!p_physp || !p_remote_physp)
		goto _exit;

	osm_node_link(p_node, port_num, p_neighbor_node, p_ni_context->port_num);

	osm_db_neighbor_set(sm->p_subn->p_neighbor,
			    cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			    port_num,
			    cl_ntoh64(osm_physp_get_port_guid(p_remote_physp)),
			    p_ni_context->port_num);
	osm_db_neighbor_set(sm->p_subn->p_neighbor,
			    cl_ntoh64(osm_physp_get_port_guid(p_remote_physp)),
			    p_ni_context->port_num,
			    cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			    port_num);

_exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void update_ignored_status(osm_subn_t *p_subn, uint8_t port_num,
				  ib_net64_t port_guid)
{
	osm_port_t *p_port;

	p_port = osm_get_port_by_guid(p_subn, port_guid);
	if (p_port) {
		p_port->disc_through = OSM_DISC_IGNORED;
		p_port->client_rereg = TRUE;
		if (!cl_is_object_in_list(&p_subn->ignored_nodes_list, p_port)) {
			if (cl_list_insert_tail(&p_subn->ignored_nodes_list,
						 p_port) != CL_SUCCESS)
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 0D30: Out of memeory\n");
		}
	}
}

static osm_ignore_port_t get_ignore_type(IN osm_sm_t *sm, IN osm_node_t *p_node,
					 IN const uint8_t port_num)
{
	if (lookup_held_back_switch(sm->p_subn, p_node))
		return OSM_IGNORE_PORT_HELDBACK;
	else if (lookup_isolated_port(sm, p_node, port_num))
		return OSM_IGNORE_PORT_ISOLATE;
	else if (lookup_no_discover_port(sm, p_node, port_num))
		return OSM_IGNORE_PORT_NO_DISCOVER;
	else
		return OSM_IGNORE_PORT_NONE;
}

static void update_remote_ignore_type(IN const osm_ni_context_t *p_ni_context,
				      IN osm_subn_t *p_subn,
				      IN osm_ignore_port_t ignore_type)
{
	osm_node_t * p_remote_node;

	if (p_subn->in_sweep_hop_0)
		return;
	p_remote_node = osm_get_node_by_guid(p_subn, p_ni_context->node_guid);
	OSM_ASSERT(p_remote_node);
	p_remote_node->remote_ignore_type[p_ni_context->port_num] =
		ignore_type;
}
static void ni_rcv_get_port_info(IN osm_sm_t * sm, IN osm_node_t * node,
				 IN ib_node_info_t *ni)
{
	osm_madw_context_t context;
	osm_physp_t *physp;
	unsigned port;
	ib_api_status_t status;
	osm_ignore_port_t ignore_type;

	port = ib_node_info_get_local_port_num(ni);

	ignore_type = get_ignore_type(sm, node, port);
	/*
	 * We stop the discovery when we find an isolated port.
	 * We'll scan all the isolated ports during the sweep_ignored stage.
	 */
	if (ignore_type == OSM_IGNORE_PORT_ISOLATE) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Skipping discovery of ignored node 0x%" PRIx64
			" port %d \n", cl_ntoh64(ni->node_guid), port);
		update_ignored_status(sm->p_subn, port, ni->port_guid);
		return;
	}

	physp = osm_node_get_physp_ptr(node, port);
	if (PF(!physp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD03: "
			"Failed to find physp for port %d of Node GUID 0x%"
			PRIx64 "\n", port,
			cl_ntoh64(osm_node_get_node_guid(node)));
		return;
	}

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(node);
	context.pi_context.port_guid = osm_physp_get_port_guid(physp);
	context.pi_context.set_method = FALSE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;
	context.pi_context.client_rereg = FALSE;
	context.pi_context.port_num = 0;

	status = osm_req_get_pi(sm, physp, port, TRUE, 0, &context, FALSE);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD02: "
			"Failure initiating PortInfo request (%s)\n",
			ib_get_err_str(status));
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
void osm_req_get_node_desc(IN osm_sm_t * sm, osm_physp_t * p_physp)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_madw_context_t context;

	OSM_LOG_ENTER(sm->p_log);

	context.nd_context.node_guid =
	    osm_node_get_node_guid(osm_physp_get_node_ptr(p_physp));

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(p_physp),
			     IB_MAD_ATTR_NODE_DESC, 0, TRUE, 0,
			     0, CL_DISP_MSGID_NONE, &context);
	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D03: "
			"Failure initiating NodeDescription request (%s)\n",
			ib_get_err_str(status));

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_get_node_desc(IN osm_sm_t * sm, IN osm_node_t * p_node,
				 IN ib_node_info_t *p_ni)
{
	uint8_t port_num;
	osm_physp_t *p_physp = NULL;

	OSM_LOG_ENTER(sm->p_log);

	port_num = ib_node_info_get_local_port_num(p_ni);

	/*
	   Request PortInfo & NodeDescription attributes for the port
	   that responded to the NodeInfo attribute.
	   Because this is a channel adapter or router, we are
	   not allowed to request PortInfo for the other ports.
	   Set the context union properly, so the recipient
	   knows which node & port are relevant.
	 */
	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	if (PF(!p_physp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR OD04: "
			"Failed to find physp for port %d of Node GUID 0x%"
			PRIx64 "\n", port_num,
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		return;
	}

	osm_req_get_node_desc(sm, p_physp);

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_process_new_ca_or_router(IN osm_sm_t * sm,
					    IN osm_node_t * p_node,
					    IN ib_node_info_t *p_ni)
{
	OSM_LOG_ENTER(sm->p_log);

	ni_rcv_get_port_info(sm, p_node, p_ni);

	/*
	   A node guid of 0 is the corner case that indicates
	   we discovered our own node.  Initialize the subnet
	   object with the SM's own port guid.
	 */
	if (p_ni->node_guid == 0)
		sm->p_subn->sm_port_guid = p_node->node_info.port_guid;

	OSM_LOG_EXIT(sm->p_log);
}

static void ni_send_general_info_to_node(IN osm_sm_t *sm, IN osm_node_t *p_node,
					 IN const osm_ni_context_t *ni_context,
					 IN const ib_node_info_t *p_ni)
{
	OSM_LOG_ENTER(sm->p_log);

	osm_madw_context_t context;
	ib_api_status_t status;
	osm_physp_t *p_physp;

	context.general_info_context.node_guid = osm_node_get_node_guid(p_node);
	context.general_info_context.ni_context = *ni_context;
	context.general_info_context.node_info = *p_ni;

	p_physp = osm_node_get_physp_ptr(p_node, ib_node_info_get_local_port_num(&p_node->node_info));

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(p_physp),
			     IB_MAD_ATTR_GENERAL_INFO,
			     IB_GENERAL_INFO_CAP_ATTRIBUTE_MODIFIER,
			     TRUE, 0, 0, CL_DISP_MSGID_NONE, &context);

	if (status == IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Sending GeneralInfo request to node 0x%016" PRIx64 "\n",
			cl_ntoh64(context.general_info_context.node_guid));
	else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0D27: Failed to send GeneralInfo request "
			"(%s) to node 0x%016" PRIx64 "\n", ib_get_err_str(status),
			cl_ntoh64(context.general_info_context.node_guid));

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_process_existing_ca_or_router(IN osm_sm_t * sm,
						 IN osm_node_t * p_node,
						 IN const osm_madw_t * p_madw)
{
	ib_node_info_t *p_ni;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_port_t *p_port_check;
	uint8_t port_num;
	osm_dr_path_t *p_dr_path;
	osm_alias_guid_t *p_alias_guid, *p_alias_guid_check;
	osm_ni_context_t *ni_context;

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ni = ib_smp_get_payload_ptr(p_smp);
	port_num = ib_node_info_get_local_port_num(p_ni);

	/*
	   Determine if we have encountered this node through a
	   previously undiscovered port.  If so, build the new
	   port object.
	 */
	p_port = osm_get_port_by_guid(sm->p_subn, p_ni->port_guid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Creating new port object with GUID 0x%" PRIx64 "\n",
			cl_ntoh64(p_ni->port_guid));

		osm_node_init_physp(p_node, port_num, p_madw);

		p_port = osm_port_new(p_ni, p_node);
		if (PF(p_port == NULL)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D04: "
				"Unable to create new port object\n");
			goto Exit;
		}

		/*
		   Add the new port object to the database.
		 */
		p_port_check =
		    (osm_port_t *) cl_qmap_insert(&sm->p_subn->port_guid_tbl,
						  p_ni->port_guid,
						  &p_port->map_item);
		if (p_port_check != p_port) {
			/*
			   We should never be here!
			   Somehow, this port GUID already exists in the table.
			 */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D12: "
				"Port 0x%" PRIx64 " already in the database!\n",
				cl_ntoh64(p_ni->port_guid));

			osm_port_delete(sm, &p_port);
			goto Exit;
		}

		p_alias_guid = osm_alias_guid_new(p_ni->port_guid,
						  p_port, 0);
		if (PF(!p_alias_guid)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D11: "
				"alias guid memory allocation failed"
				" for port GUID 0x%" PRIx64 "\n",
				cl_ntoh64(p_ni->port_guid));
			goto alias_done;
		}

		/* insert into alias guid table */
		p_alias_guid_check =
			(osm_alias_guid_t *) cl_qmap_insert(&sm->p_subn->alias_port_guid_tbl,
							    p_alias_guid->alias_guid,
							    &p_alias_guid->map_item);
		if (p_alias_guid_check != p_alias_guid) {
			if (!p_alias_guid_check->dirty) {
				/* alias GUID is a duplicate */
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D13: "
					"Duplicate alias port GUID 0x%" PRIx64 "\n",
					cl_ntoh64(p_ni->port_guid));
				osm_port_delete(sm, &p_port);
				goto Exit;
			} else {
				p_alias_guid_check->dirty = FALSE;
				p_alias_guid_check->p_base_port = p_alias_guid->p_base_port;
				p_alias_guid_check->base_port_guid = p_alias_guid->base_port_guid;
			}
			osm_alias_guid_delete(&p_alias_guid);
		}

alias_done:
		/* If we are a master, then this means the port is new on the subnet.
		   Mark it as new - need to send trap 64 for these ports.
		   The condition that we are master is true, since if we are in discovering
		   state (meaning we woke up from standby or we are just initializing),
		   then these ports may be new to us, but are not new on the subnet.
		   If we are master, then the subnet as we know it is the updated one,
		   and any new ports we encounter should cause trap 64. C14-72.1.1 */
		if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER)
			p_port->is_new = 1;

	} else {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);

		if (PF(p_physp == NULL)) {
			osm_port_t *p_port_existing;
			osm_physp_t *p_physp_existing;
			osm_dr_path_t *p_dr_path_existing;
			osm_dr_path_t dr_path;

			/* Dump the DR path to the duplicate GUID port */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1C: "
				"No physical port found for node GUID 0x%"
				PRIx64 " port %u. Might be duplicate port GUID. "
				"DR path to duplicate GUID:\n",
				cl_ntoh64(p_node->node_info.node_guid),
				port_num);

			osm_dr_path_init(&dr_path, p_smp->hop_count,
					 p_smp->initial_path);
			osm_dump_dr_path_v2(sm->p_log, &dr_path, FILE_ID, OSM_LOG_ERROR);

			/* Dump the DR path to the duplicate GUID port */
			p_port_existing =
			   osm_get_port_by_guid(sm->p_subn,
						p_node->node_info.node_guid);
			if (!p_port_existing)
				goto Exit;

			p_physp_existing = p_port->p_physp;
			if (!p_physp_existing)
				goto Exit;

			p_dr_path_existing =
			    osm_physp_get_dr_path_ptr(p_physp_existing);

			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1C: "
				"No physical port found for node GUID 0x%"
				PRIx64 ". Might be duplicate port GUID. "
				"DR path to existing GUID:\n",
				cl_ntoh64(p_node->node_info.node_guid));

			osm_dump_dr_path_v2(sm->p_log, p_dr_path_existing,
					    FILE_ID, OSM_LOG_ERROR);
			goto Exit;
		}

		/*
		   Update the DR Path to the port,
		   in case the old one is no longer available.
		 */
		p_dr_path = osm_physp_get_dr_path_ptr(p_physp);

		osm_dr_path_init(p_dr_path, p_smp->hop_count,
				 p_smp->initial_path);
	}

	if (osm_cap_mgr_is_enabled(&sm->cap_mgr, GI, p_node)) {
		if (p_node->gi_updated) {
			ni_rcv_get_port_info(sm, p_node, p_ni);
		} else {
			ni_context = osm_madw_get_ni_context_ptr(p_madw);
			ni_send_general_info_to_node(sm, p_node, ni_context,
						     p_ni);
		}
	} else
		ni_rcv_get_port_info(sm, p_node, p_ni);

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void ni_rcv_process_switch(IN osm_sm_t * sm, IN osm_node_t * p_node,
				  const ib_smp_t *p_smp,
				  const osm_ni_context_t *ni_context,
				  const ib_node_info_t *p_ni)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_physp_t *p_physp;
	osm_madw_context_t context;
	osm_dr_path_t *path;
	unsigned port;
	osm_ignore_port_t ignore_type;

	OSM_LOG_ENTER(sm->p_log);

	port = ib_node_info_get_local_port_num(p_ni);

	ignore_type = get_ignore_type(sm, p_node, port);
	/*
	 * If this switch was added to the held back list, we need to check
	 * which multicast trees used the switch, and mark them for
	 * recalculation.
	 * We check that the switch is new in the held back list by checking
	 * if port 0 ignore status.
	 */
	if ((ignore_type == OSM_IGNORE_PORT_HELDBACK) && p_node->sw &&
	    p_node->physp_ignored[0] == OSM_PHYSP_NORMAL)
		osm_subn_mark_mbox_by_sw(sm->p_subn, p_node->sw, 0);

	/*
	 * We stop the discovery when we find an isolated port.
	 * We'll scan all the isolated ports during the sweep_ignored stage
	 * and mark all nodes that reachable through isolated ports as
	 * ignored.
	 */
	if (ignore_type) {

		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Found ignored switch node 0x%" PRIx64
			" port %d \n", cl_ntoh64(p_ni->node_guid), port);
		update_ignored_status(sm->p_subn, port, p_ni->port_guid);
		/*
		 * Update the ignore type in the remote port which originated
		 * the node info request
		 */
		update_remote_ignore_type(ni_context, sm->p_subn, ignore_type);
		goto Exit;
	}

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	/* update DR path of already initialized switch port 0 */
	path = osm_physp_get_dr_path_ptr(p_physp);
	osm_dr_path_init(path, p_smp->hop_count, p_smp->initial_path);

	context.si_context.node_guid = osm_node_get_node_guid(p_node);
	context.si_context.set_method = FALSE;
	context.si_context.light_sweep = FALSE;
	context.si_context.lft_top_change = FALSE;

	/* Request a SwitchInfo attribute */
	status = osm_req_get(sm, path, IB_MAD_ATTR_SWITCH_INFO, 0, TRUE, 0,
			     0, CL_DISP_MSGID_NONE, &context);
	if (status != IB_SUCCESS)
		/* continue despite error */
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D06: "
			"Failure initiating SwitchInfo request (%s)\n",
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_process_existing_switch(IN osm_sm_t * sm,
					   IN osm_node_t * p_node,
					   IN const osm_madw_t * p_madw)
{
	osm_ni_context_t *p_ni_context;
	ib_node_info_t *p_ni;
	ib_smp_t *p_smp;

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ni = ib_smp_get_payload_ptr(p_smp);

	/*
	   If this switch has already been probed during this sweep,
	   then don't bother reprobing it.
	 */
	if (p_node->discovery_count == 1) {
		if (osm_cap_mgr_is_enabled(&sm->cap_mgr, GI, p_node) &&
		    !p_node->gi_updated) {
			p_ni_context = osm_madw_get_ni_context_ptr(p_madw);
			ni_send_general_info_to_node(sm, p_node,p_ni_context,
						     p_ni);
		} else
			ni_rcv_process_switch(sm, p_node, osm_madw_get_smp_ptr(p_madw),
					      osm_madw_get_ni_context_ptr(p_madw), p_ni);
	} else {
		unsigned port = ib_node_info_get_local_port_num(p_ni);

		osm_ignore_port_t ignore_type = get_ignore_type(sm, p_node,
								port);
		if (ignore_type) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Found ignored switch node 0x%" PRIx64
				" port %d \n", cl_ntoh64(p_ni->node_guid), port);
			/*
			 * Update the ignore type in the remote port which
			 * originated the node info request
			 */
			update_remote_ignore_type(osm_madw_get_ni_context_ptr(p_madw),
						  sm->p_subn, ignore_type);
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_process_new_switch(IN osm_sm_t * sm, IN osm_node_t * p_node,
				      IN const ib_smp_t *p_smp,
				      IN const osm_ni_context_t *ni_context,
				      IN const ib_node_info_t *p_ni)
{
	OSM_LOG_ENTER(sm->p_log);

	ni_rcv_process_switch(sm, p_node, p_smp, ni_context, p_ni);

	/*
	   A node guid of 0 is the corner case that indicates
	   we discovered our own node.  Initialize the subnet
	   object with the SM's own port guid.
	 */
	if (p_ni->node_guid == 0)
		sm->p_subn->sm_port_guid = p_node->node_info.port_guid;

	OSM_LOG_EXIT(sm->p_log);
}

static void update_client_rereg(osm_port_t * p_port,
				disc_state_t context_disc_through)
{
	/*
	 * When we are not discovery of self node, we want to know
	 * whether we discovered this port through ignored port or not.
	 * If this is the first time the port discovered through ignored
	 * or normal port or if it was earlier discovered through
	 * ignored port and discovered through normal port, we want to
	 * reset the discovery_count, so we will continue the discovery.
	 */
	if (context_disc_through > p_port->disc_through)
		p_port->disc_through = context_disc_through;

	/* If we discovered this port through ignored port,
	 * we want to send client_rereg to it when it won't be
	 * ignored.
	 */
	if (p_port->disc_through == OSM_DISC_IGNORED)
		p_port->client_rereg = TRUE;
}

void ni_rcv_complete_process_new_node(IN osm_sm_t *sm, IN osm_node_t *p_node,
				      IN const ib_smp_t *p_smp,
				      IN const osm_ni_context_t *ni_context,
				      IN ib_node_info_t *p_ni)
{
	OSM_LOG_ENTER(sm->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Complete NodeInfo handling for node "
		"0x%" PRIx64 "\n", cl_ntoh64(osm_node_get_node_guid(p_node)));

	switch (p_ni->node_type) {
	case IB_NODE_TYPE_CA:
	case IB_NODE_TYPE_ROUTER:
		ni_rcv_process_new_ca_or_router(sm, p_node, p_ni);
		break;
	case IB_NODE_TYPE_SWITCH:
		ni_rcv_process_new_switch(sm, p_node, p_smp, ni_context, p_ni);
		break;
	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D16: "
			"Unknown node type %u with GUID 0x%" PRIx64 "\n",
			p_ni->node_type, cl_ntoh64(p_ni->node_guid));
		break;
	}

	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must NOT be held before calling this function.
**********************************************************************/
static void ni_rcv_process_new(IN osm_sm_t * sm, IN const osm_madw_t * p_madw)
{
	osm_node_t *p_node;
	osm_node_t *p_node_check;
	osm_port_t *p_port;
	osm_port_t *p_port_check;
	osm_router_t *p_rtr = NULL;
	osm_router_t *p_rtr_check;
	cl_qmap_t *p_rtr_guid_tbl;
	ib_node_info_t *p_ni;
	ib_smp_t *p_smp;
	osm_ni_context_t *p_ni_context;
	osm_alias_guid_t *p_alias_guid, *p_alias_guid_check;
	uint8_t port_num;

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ni = ib_smp_get_payload_ptr(p_smp);
	p_ni_context = osm_madw_get_ni_context_ptr(p_madw);
	port_num = ib_node_info_get_local_port_num(p_ni);

	osm_dump_smp_dr_path_v2(sm->p_log, p_smp, FILE_ID, OSM_LOG_VERBOSE);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Discovered new %s node,"
		"\n\t\t\t\tGUID 0x%" PRIx64 ", TID 0x%" PRIx64 "\n",
		ib_get_node_type_str(p_ni->node_type),
		cl_ntoh64(p_ni->node_guid), cl_ntoh64(p_smp->trans_id));

	p_node = osm_node_new(p_madw);
	if (PF(p_node == NULL)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D07: "
			"Unable to create new node object\n");
		goto Exit;
	}


	/*
	   Create a new port object to represent this node's physical
	   ports in the port table.
	 */
	p_port = osm_port_new(p_ni, p_node);
	if (PF(p_port == NULL)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D14: "
			"Unable to create new port object\n");
		osm_node_delete(&p_node);
		goto Exit;
	}

	/*
	   Add the new port object to the database.
	 */
	p_port_check =
	    (osm_port_t *) cl_qmap_insert(&sm->p_subn->port_guid_tbl,
					  p_ni->port_guid, &p_port->map_item);
	if (PF(p_port_check != p_port)) {
		/*
		   We should never be here!
		   Somehow, this port GUID already exists in the table.
		 */
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D15: "
			"Duplicate Port GUID 0x%" PRIx64
			"! Found by the two directed routes:\n",
			cl_ntoh64(p_ni->port_guid));
		osm_dump_dr_path_v2(sm->p_log,
				    osm_physp_get_dr_path_ptr(p_port->p_physp),
				    FILE_ID, OSM_LOG_ERROR);
		osm_dump_dr_path_v2(sm->p_log,
				    osm_physp_get_dr_path_ptr(p_port_check->
							   p_physp),
				    FILE_ID, OSM_LOG_ERROR);
		osm_port_delete(sm, &p_port);
		osm_node_delete(&p_node);
		goto Exit;
	}

	p_alias_guid = osm_alias_guid_new(p_ni->port_guid,
					  p_port, 0);
	if (PF(!p_alias_guid)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D18: "
			"alias guid memory allocation failed"
			" for port GUID 0x%" PRIx64 "\n",
			cl_ntoh64(p_ni->port_guid));
		goto alias_done2;
	}

	/* insert into alias guid table */
	p_alias_guid_check =
		(osm_alias_guid_t *) cl_qmap_insert(&sm->p_subn->alias_port_guid_tbl,
						    p_alias_guid->alias_guid,
						    &p_alias_guid->map_item);
	if (p_alias_guid_check != p_alias_guid) {
		/* alias GUID is a duplicate */
		if (!p_alias_guid_check->dirty) {
			/* alias GUID is a duplicate */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D13: "
				"Duplicate alias port GUID 0x%" PRIx64 "\n",
				cl_ntoh64(p_ni->port_guid));
		} else {
			p_alias_guid_check->dirty = FALSE;
			p_alias_guid_check->p_base_port = p_alias_guid->p_base_port;
			p_alias_guid_check->base_port_guid = p_alias_guid->base_port_guid;
		}
		osm_alias_guid_delete(&p_alias_guid);
	}

alias_done2:
	/* If we are a master, then this means the port is new on the subnet.
	   Mark it as new - need to send trap 64 on these ports.
	   The condition that we are master is true, since if we are in discovering
	   state (meaning we woke up from standby or we are just initializing),
	   then these ports may be new to us, but are not new on the subnet.
	   If we are master, then the subnet as we know it is the updated one,
	   and any new ports we encounter should cause trap 64. C14-72.1.1 */
	if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER)
		p_port->is_new = 1;

	/* If there were RouterInfo or other router attribute,
	   this would be elsewhere */
	if (p_ni->node_type == IB_NODE_TYPE_ROUTER) {
		if (PF((p_rtr = osm_router_new(p_port)) == NULL))
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1A: "
				"Unable to create new router object\n");
		else {
			p_rtr_guid_tbl = &sm->p_subn->rtr_guid_tbl;
			p_rtr_check =
			    (osm_router_t *) cl_qmap_insert(p_rtr_guid_tbl,
							    p_ni->port_guid,
							    &p_rtr->map_item);
			if (PF(p_rtr_check != p_rtr))
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D1B: "
					"Unable to add port GUID:0x%016" PRIx64
					" to router table\n",
					cl_ntoh64(p_ni->port_guid));
			else
				p_node->router = p_rtr;
		}
	}

	p_node_check =
	    (osm_node_t *) cl_qmap_insert(&sm->p_subn->node_guid_tbl,
					  p_ni->node_guid, &p_node->map_item);
	if (PF(p_node_check != p_node)) {
		/*
		   This node must have been inserted by another thread.
		   This is unexpected, but is not an error.
		   We can simply clean-up, since the other thread will
		   see this processing through to completion.
		 */
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Discovery race detected at node 0x%" PRIx64 "\n",
			cl_ntoh64(p_ni->node_guid));
		osm_node_delete(&p_node);
		p_node = p_node_check;
		ni_rcv_set_links(sm, p_node, port_num, p_ni_context);
		goto Exit;
	} else
		ni_rcv_set_links(sm, p_node, port_num, p_ni_context);

	if (sm->p_subn->in_sweep_hop_0 == FALSE)
		update_client_rereg(p_port, p_ni_context->disc_through);
	else
		p_port->disc_through = OSM_DISC_NORMAL;

	p_node->discovery_count++;
	ni_rcv_get_node_desc(sm, p_node, p_ni);

	if (osm_cap_mgr_is_enabled(&sm->cap_mgr, GI, p_node))
		ni_send_general_info_to_node(sm, p_node, p_ni_context, p_ni);
	else
		ni_rcv_complete_process_new_node(sm, p_node, p_smp,
						 p_ni_context, p_ni);

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void ni_rcv_process_existing(IN osm_sm_t * sm, IN osm_node_t * p_node,
				    IN const osm_madw_t * p_madw)
{
	ib_node_info_t *p_ni;
	ib_smp_t *p_smp;
	osm_ni_context_t *p_ni_context;
	osm_port_t *p_port;
	uint8_t port_num;

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ni = ib_smp_get_payload_ptr(p_smp);
	p_ni_context = osm_madw_get_ni_context_ptr(p_madw);
	port_num = ib_node_info_get_local_port_num(p_ni);

	if (validate_node_info_existing(sm, p_ni, p_node) != IB_SUCCESS) {
		sm->p_subn->subnet_initialization_error = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0D26: Failed to validate NodeInfo for 0x%"
			PRIx64 "\n", cl_ntoh64(p_ni->node_guid));
		goto Exit;
	}

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Rediscovered %s node 0x%" PRIx64 " TID 0x%" PRIx64
		", discovered %u times already\n",
		ib_get_node_type_str(p_ni->node_type),
		cl_ntoh64(p_ni->node_guid),
		cl_ntoh64(p_smp->trans_id), p_node->discovery_count);

	/*
	   If we haven't already encountered this existing node
	   on this particular sweep, then process further.
	 */
	p_node->discovery_count++;

	switch (p_ni->node_type) {
	case IB_NODE_TYPE_CA:
	case IB_NODE_TYPE_ROUTER:
		ni_rcv_process_existing_ca_or_router(sm, p_node, p_madw);
		p_port = osm_get_port_by_guid(sm->p_subn, p_ni->port_guid);
		OSM_ASSERT(p_port);

		if (sm->p_subn->in_sweep_hop_0 == FALSE)
			update_client_rereg(p_port, p_ni_context->disc_through);
		else
			p_port->disc_through = OSM_DISC_NORMAL;

		break;

	case IB_NODE_TYPE_SWITCH:
		p_port = osm_get_port_by_guid(sm->p_subn, p_ni->port_guid);
		OSM_ASSERT(p_port);

		/*
		 * If previously we found this node through ignored port,
		 * and through a normal port, we want to reset the discovery_count
		 * inorder to discover its' ports.
		 */
		if (p_ni_context->disc_through > p_port->disc_through)
			p_node->discovery_count = 1;

		if (sm->p_subn->in_sweep_hop_0 == FALSE)
			update_client_rereg(p_port, p_ni_context->disc_through);
		else
			p_port->disc_through = OSM_DISC_NORMAL;
		ni_rcv_process_existing_switch(sm, p_node, p_madw);
		break;

	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0D09: "
			"Unknown node type %u with GUID 0x%" PRIx64 "\n",
			p_ni->node_type, cl_ntoh64(p_ni->node_guid));
		break;
	}

	if ( p_ni->sys_guid != p_node->node_info.sys_guid) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Updated SysImageGUID: 0x%"
			PRIx64 " for node 0x%" PRIx64 "\n",
			cl_ntoh64(p_ni->sys_guid),
			cl_ntoh64(p_ni->node_guid));
	}
	ni_rcv_set_links(sm, p_node, port_num, p_ni_context);
	p_node->node_info = *p_ni;

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_ni_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_node_info_t *p_ni;
	ib_smp_t *p_smp;
	osm_node_t *p_node;
	osm_mad_addr_t *p_mad_addr;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ni = ib_smp_get_payload_ptr(p_smp);
	p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_NODE_INFO);

	/*
	   Determine if this node has already been discovered,
	   and process accordingly.
	   During processing of this node, hold the shared lock.
	 */

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_node = osm_get_node_by_guid(sm->p_subn, p_ni->node_guid);

	verbose_bypass = osm_verbose_bypass_handle_ni(sm, p_ni);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	osm_dump_node_info_v2(sm->p_log, p_ni, FILE_ID, OSM_LOG_DEBUG);

	if (validate_node_info(sm, p_ni, p_smp, p_mad_addr->dest_lid) != IB_SUCCESS) {
		osm_hm_set_by_dr_path(sm, p_smp->hop_count,
				      p_smp->initial_path,
				      OSM_HM_COND_ILLEGAL);
		goto Exit;
	}

	if (!p_node)
		ni_rcv_process_new(sm, p_madw);
	else
		ni_rcv_process_existing(sm, p_node, p_madw);

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	osm_verbose_bypass_end(verbose_bypass, sm->p_log);

	OSM_LOG_EXIT(sm->p_log);
}
