/*
 * Copyright (c) 2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_plft_mad_rcv_process.
 * This object represents the PrivateLFT MADs Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PLFT_MAD_RCV_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>

static void process_plft_info(osm_sm_t *sm, osm_node_t *p_node,
			      osm_ar_mad_context_t *p_context,
			      ib_plft_info_t *p_plft_info)
{
	OSM_LOG_ENTER(sm->p_log);

	osm_switch_t *p_sw = p_node->sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	switch (p_context->request_type) {
		case OSM_AR_REQ_TYPE_GET:
			/*
			 * In case the SET request failed the cached plft_info
			 * matches the one on the switch meaning there is need
			 * to mark the plft_info_updated flag only in GET request.
			 */
			p_ar_data->plft_info_updated = TRUE;
			/* Fall through */
		case OSM_AR_REQ_TYPE_SET:
			memcpy(&p_ar_data->plft_info, p_plft_info,
			       sizeof(p_ar_data->plft_info));
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Node 0x%" PRIx64 " received PLFTInfo "
				"with non supported request type %u\n",
				osm_node_get_node_guid(p_node),
				p_context->request_type);
			break;
	}

	OSM_LOG_EXIT(sm->p_log);
}

static void process_plft_def(osm_sm_t *sm, osm_node_t *p_node,
			     osm_ar_mad_context_t *p_context,
			     ib_plft_def_t *p_plft_def)
{
	OSM_LOG_ENTER(sm->p_log);
	osm_switch_t *p_sw = p_node->sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	switch (p_context->request_type) {
		case OSM_AR_REQ_TYPE_SET:
			if (osm_ar_plft_defs_set_block(p_ar_data,
						       p_context->block_id,
						       p_plft_def))
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 8803: Failed to store PLFTDef"
					" response for switch 0x%" PRIx64 " (block id %u)\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_context->block_id);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Node 0x%" PRIx64 " received PLFTDef "
				"with non supported request type %u\n",
				osm_node_get_node_guid(p_node),
				p_context->request_type);
			break;
	}

	OSM_LOG_EXIT(sm->p_log);
}

static void process_port_sl_plft_map(osm_sm_t *sm, osm_node_t *p_node,
				     osm_ar_mad_context_t *p_context,
				     ib_port_sl_to_plft_map_t *p_port_sl_plft_map)
{
	OSM_LOG_ENTER(sm->p_log);
	osm_switch_t *p_sw = p_node->sw;

	switch (p_context->request_type) {
		case OSM_AR_REQ_TYPE_SET:
			if (osm_ar_port_sl_to_plft_set_block(p_sw,
							     p_context->block_id,
							     TRUE,
							     p_port_sl_plft_map))
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 8804: Failed to store PortSLToPlftMap"
					" response for switch 0x%" PRIx64 " (block id %u)\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_context->block_id);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Node 0x%" PRIx64 " received PortSLToPlftMap "
				"with non supported request type %u\n",
				osm_node_get_node_guid(p_node),
				p_context->request_type);
			break;
	}

	OSM_LOG_EXIT(sm->p_log);
}

static void process_plft_map(osm_sm_t *sm, osm_node_t *p_node,
			     osm_ar_mad_context_t *p_context,
			     ib_plft_map_t *p_plft_map)
{
	OSM_LOG_ENTER(sm->p_log);
	osm_switch_t *p_sw = p_node->sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	switch (p_context->request_type) {
		case OSM_AR_REQ_TYPE_SET:
			if (osm_ar_plft_map_set_block(p_ar_data,
						      p_context->block_id,
						      p_plft_map))
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 8805: Failed to store PlftMap"
					" response for switch 0x%" PRIx64 " (plft id %u)\n",
					cl_ntoh64(osm_node_get_node_guid(p_node)),
					p_context->block_id);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Node 0x%" PRIx64 " received PlftMap "
				"with non supported request type %u\n",
				osm_node_get_node_guid(p_node),
				p_context->request_type);
			break;
	}

	OSM_LOG_EXIT(sm->p_log);
}

void osm_plft_mad_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	void *p_payload;
	osm_ar_mad_context_t *p_context;

	osm_node_t *p_node;
	ib_net64_t node_guid;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_ar_mad_context_ptr(p_madw);
	p_payload = ib_smp_get_payload_ptr(p_smp);

	node_guid = p_context->node_guid;
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || !p_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8801: "
			"Received PLFTInfo for nonexistent switch node "
			"0x%" PRIx64 "\n", cl_ntoh64(node_guid));
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	/* mark as successfull */
	p_node->sw->ar_data.pending_mads--;

	switch (p_smp->attr_id) {
		case IB_MAD_ATTR_PLFT_INFO:
			process_plft_info(sm, p_node, p_context,
					  (ib_plft_info_t *)p_payload);
			break;
		case IB_MAD_ATTR_PLFT_DEF:
			process_plft_def(sm, p_node, p_context,
					 (ib_plft_def_t *)p_payload);
			break;
		case IB_MAD_ATTR_PORT_SL_TO_PLFT_MAP:
			process_port_sl_plft_map(sm, p_node, p_context,
						 (ib_port_sl_to_plft_map_t *)p_payload);
			break;
		case IB_MAD_ATTR_PLFT_MAP:
			process_plft_map(sm, p_node, p_context,
					 (ib_plft_map_t *)p_payload);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8802: "
				"Received unsupported MAD Attribute 0x%x(%s) "
				" for switch node 0x%" PRIx64 "\n",
				cl_ntoh16(p_smp->attr_id),
				ib_get_sm_attr_str(p_smp->attr_id),
				cl_ntoh64(node_guid));
	}


	CL_PLOCK_RELEASE(sm->p_lock);

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return;
}

