/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2010 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (C) 2012-2017 Tokyo Institute of Technology. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_physp_t.
 * This object represents an Infiniband Port.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <complib/cl_debug.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PORT_C
#include <opensm/osm_port.h>
#include <opensm/osm_node.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_vport.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_activity_mgr.h>

#define VPORTS_BLOCK_BYTES (OSM_VPS_BLOCK_SIZE * sizeof(struct osm_vport*))

/*
 * This global is defined in order to avoid the need to allocate and later
 * on free a default_pkey value for each vport pkey table since for the beta
 * we hard coded plant default pkey on each vport table
 * When we will come to implement support in pkeys for vports we shall remove
 * this global parameter
 *
 * TODO: Remove this on virtualization GA version
 */
const ib_net16_t def_pkey = IB_DEFAULT_PKEY;

void osm_physp_construct(IN osm_physp_t * p_physp)
{
	memset(p_physp, 0, sizeof(*p_physp));
	osm_dr_path_construct(&p_physp->dr_path);
	cl_ptr_vector_construct(&p_physp->slvl_by_port);
	osm_pkey_tbl_construct(&p_physp->pkeys);
	p_physp->topo_id = NO_TOPO_ID;
	p_physp->expected_lmc = OSM_PORT_LMC_NOT_SET;
}

void osm_physp_destroy(IN osm_physp_t * p_physp)
{
	size_t num_slvl, i;

	/* the physp might be uninitialized */
	if (p_physp->port_guid) {
		if (p_physp->p_guids)
			free(p_physp->p_guids);

		/* free the SL2VL Tables */
		num_slvl = cl_ptr_vector_get_size(&p_physp->slvl_by_port);
		for (i = 0; i < num_slvl; i++)
			free(cl_ptr_vector_get(&p_physp->slvl_by_port, i));
		cl_ptr_vector_destroy(&p_physp->slvl_by_port);

		/* free the P_Key Tables */
		osm_pkey_tbl_destroy(&p_physp->pkeys);

		memset(p_physp, 0, sizeof(*p_physp));
		osm_dr_path_construct(&p_physp->dr_path);	/* clear dr_path */
	}
}

static inline void osm_physp_update_state_timestamp(IN osm_physp_t *p_physp)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	sprint_time_str(&tv, p_physp->state_timestamp_str,
			OSM_PORT_TIMESTAMP_STR_LEN, 1);
}

void osm_physp_init(IN osm_physp_t * p_physp, IN ib_net64_t port_guid,
		    IN uint8_t port_num, IN const struct osm_node *p_node,
		    IN osm_bind_handle_t h_bind, IN uint8_t hop_count,
		    IN const uint8_t * p_initial_path)
{
	uint16_t num_slvl, i;
	ib_slvl_table_t *p_slvl;

	OSM_ASSERT(p_node);

	osm_physp_construct(p_physp);
	p_physp->port_guid = port_guid;
	p_physp->port_num = port_num;
	p_physp->healthy = TRUE;
	p_physp->num_dr_paths_tried = 0;
	p_physp->need_update = 1;
	p_physp->p_node = (struct osm_node *)p_node;

	osm_dr_path_init(&p_physp->dr_path, hop_count, p_initial_path);

	/* allocate enough SL2VL tables */
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		/* we need node num ports + 1 SL2VL tables */
		num_slvl = osm_node_get_num_physp(p_node) + 1;
	else
		/* An end node - we need only one SL2VL */
		num_slvl = 1;

	cl_ptr_vector_init(&p_physp->slvl_by_port, num_slvl, 1);
	for (i = 0; i < num_slvl; i++) {
		p_slvl = (ib_slvl_table_t *) malloc(sizeof(ib_slvl_table_t));
		if (!p_slvl)
			break;
		memset(p_slvl, 0, sizeof(ib_slvl_table_t));
		cl_ptr_vector_set(&p_physp->slvl_by_port, i, p_slvl);
	}

	/* initialize the pkey table */
	osm_pkey_tbl_init(&p_physp->pkeys);

	/* Set port to disabled/down */
	ib_port_info_set_port_state(&p_physp->port_info, IB_LINK_DOWN);
	ib_port_info_set_port_phys_state(IB_PORT_PHYS_STATE_DISABLED, &p_physp->port_info);

	/* Initialize port's update timestamp */
	osm_physp_update_state_timestamp(p_physp);
}

void osm_port_delete_virtualization(IN struct osm_sm *p_sm,
				    IN OUT osm_port_t *p_port)
{
	uint16_t i, top;

	osm_virt_mgr_dequeue_port(p_sm, p_port);
	if (p_port->virt_data.allocated_blocks) {
		top = cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);
		for (i = 0 ; i <= top ; i++)
			osm_port_delete_vport(p_sm, p_port, i);
		free(p_port->virt_data.vports);
		p_port->virt_data.vports = NULL;
		p_port->virt_data.allocated_blocks = 0;
	}

	memset(&p_port->virt_data.virtualization_info, 0,
	       sizeof(ib_virtualization_info_t));
}

void osm_port_delete(IN struct osm_sm *p_sm, IN OUT osm_port_t ** pp_port)
{
	cl_list_t *p_list = &(*pp_port)->topology_list;
	cl_list_remove_all(p_list);
	cl_list_destroy(p_list);
	osm_port_delete_virtualization(p_sm, *pp_port);
	free(*pp_port);
	*pp_port = NULL;
}

osm_port_t *osm_port_new(IN const ib_node_info_t * p_ni,
			 IN osm_node_t * p_parent_node)
{
	osm_port_t *p_port;
	ib_net64_t port_guid;
	osm_physp_t *p_physp;
	uint8_t port_num;

	p_port = malloc(sizeof(*p_port));
	if (!p_port)
		return NULL;

	memset(p_port, 0, sizeof(*p_port));
	cl_qlist_init(&p_port->mcm_list);
	p_port->p_node = (struct osm_node *)p_parent_node;
	port_guid = p_ni->port_guid;
	p_port->guid = port_guid;
	p_port->client_rereg = FALSE;
	port_num = p_ni->node_type == IB_NODE_TYPE_SWITCH ?
	    0 : ib_node_info_get_local_port_num(p_ni);

	/*
	   Get the pointers to the physical node objects "owned" by this
	   logical port GUID.
	   For switches, port '0' is owned; for HCA's and routers,
	   only the singular part that has this GUID is owned.
	 */
	p_physp = osm_node_get_physp_ptr(p_parent_node, port_num);
	if (PF(!p_physp)) {
		free(p_port);
		return NULL;
	}

	OSM_ASSERT(port_guid == osm_physp_get_port_guid(p_physp));
	p_port->p_physp = p_physp;

	cl_list_construct(&p_port->topology_list);
	if (cl_list_init(&p_port->topology_list, 1) != CL_SUCCESS) {
		free(p_port);
		return NULL;
	}

	p_port->virt_data.p_port = p_port;
	p_port->virt_data.first_virt_discovery = TRUE;

	return p_port;
}

void osm_port_get_lid_range_ho(IN const osm_port_t * p_port,
			       IN uint16_t * p_min_lid, IN uint16_t * p_max_lid)
{
	uint8_t lmc;

	*p_min_lid = cl_ntoh16(osm_port_get_base_lid(p_port));
	lmc = ib_port_info_get_lmc(&p_port->p_physp->port_info);
	*p_max_lid = (uint16_t) (*p_min_lid + (1 << lmc) - 1);
}

uint8_t osm_physp_calc_link_mtu(IN osm_log_t * p_log,
				IN const osm_physp_t * p_physp,
				IN uint8_t current_mtu)
{
	const osm_physp_t *p_remote_physp;
	uint8_t mtu;
	uint8_t remote_mtu;

	OSM_LOG_ENTER(p_log);

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp) {
		/* use the available MTU */
		mtu = ib_port_info_get_mtu_cap(&p_physp->port_info);

		remote_mtu =
		    ib_port_info_get_mtu_cap(&p_remote_physp->port_info);

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Remote port 0x%016" PRIx64 " port = %u : "
			"MTU = %u. This Port MTU: %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_remote_physp)),
			osm_physp_get_port_num(p_remote_physp),
			remote_mtu, mtu);

		if (mtu != remote_mtu) {
			if (mtu > remote_mtu)
				mtu = remote_mtu;
			if (mtu != current_mtu)
				OSM_LOG(p_log, OSM_LOG_VERBOSE,
					"MTU mismatch between ports."
					"\n\t\t\t\tPort 0x%016" PRIx64 ", port %u"
					" and port 0x%016" PRIx64 ", port %u."
					"\n\t\t\t\tUsing lower MTU of %u\n",
					cl_ntoh64(osm_physp_get_port_guid(p_physp)),
					osm_physp_get_port_num(p_physp),
					cl_ntoh64(osm_physp_get_port_guid
						  (p_remote_physp)),
					osm_physp_get_port_num(p_remote_physp), mtu);
		}
	} else
		mtu = ib_port_info_get_neighbor_mtu(&p_physp->port_info);

	if (mtu == 0) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "ERR 4101: "
			"Invalid MTU = 0. Forcing correction to 256\n");
		mtu = 1;
	}

	OSM_LOG_EXIT(p_log);
	return mtu;
}

uint8_t osm_physp_calc_link_op_vls(IN osm_log_t * p_log,
				   IN const osm_subn_t * p_subn,
				   IN const osm_physp_t * p_physp,
				   IN uint8_t current_op_vls)
{
	const osm_physp_t *p_remote_physp;
	uint8_t op_vls;
	uint8_t remote_op_vls;

	OSM_LOG_ENTER(p_log);

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp) {
		/* use the available VLCap */
		op_vls = ib_port_info_get_vl_cap(&p_physp->port_info);

		remote_op_vls =
		    ib_port_info_get_vl_cap(&p_remote_physp->port_info);

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Remote port 0x%016" PRIx64 " port = 0x%X : "
			"VL_CAP = %u. This port VL_CAP = %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_remote_physp)),
			osm_physp_get_port_num(p_remote_physp),
			remote_op_vls, op_vls);

		if (op_vls != remote_op_vls) {
			if (op_vls > remote_op_vls)
				op_vls = remote_op_vls;
			if (op_vls != current_op_vls)
				OSM_LOG(p_log, OSM_LOG_VERBOSE,
					"OP_VLS mismatch between ports."
					"\n\t\t\t\tPort 0x%016" PRIx64 ", port 0x%X"
					" and port 0x%016" PRIx64 ", port 0x%X."
					"\n\t\t\t\tUsing lower OP_VLS of %u\n",
					cl_ntoh64(osm_physp_get_port_guid(p_physp)),
					osm_physp_get_port_num(p_physp),
					cl_ntoh64(osm_physp_get_port_guid
						  (p_remote_physp)),
					osm_physp_get_port_num(p_remote_physp), op_vls);
		}
	} else
		op_vls = ib_port_info_get_op_vls(&p_physp->port_info);

	if (op_vls == 0) {
		/* for non compliant implementations */
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Invalid OP_VLS = 0. Forcing correction to 1 (VL0)\n");
		op_vls = 1;
	}

	/* support user limitation of max_op_vls */
	if (op_vls > p_subn->opt.max_op_vls)
		op_vls = p_subn->opt.max_op_vls;

	OSM_LOG_EXIT(p_log);
	return op_vls;
}

static inline uint64_t ptr_to_key(void const *p)
{
	uint64_t k = 0;

	memcpy(&k, &p, sizeof(void *));
	return k;
}

static inline void *key_to_ptr(uint64_t k)
{
	void *p = 0;

	memcpy(&p, &k, sizeof(void *));
	return p;
}

/**********************************************************************
 Traverse the fabric from the SM node following the DR path given and
 add every out phys port traversed to the map. Avoid tracking the first and
 last phys ports (going into the first switch and into the target port).
 **********************************************************************/
static cl_status_t physp_get_dr_out_physp_set(IN osm_log_t * p_log,
					      IN osm_subn_t const *p_subn,
					      IN osm_dr_path_t const *p_path,
					      IN osm_physp_t * p_dest_physp,
					      OUT cl_map_t * p_physp_map)
{
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_node_t *p_node, *p_dest_node;
	uint8_t hop, cnt_hops;
	cl_status_t status = CL_SUCCESS;

	OSM_ASSERT(p_dest_physp);

	OSM_LOG_ENTER(p_log);

	/* find the OSM node */
	p_port = osm_get_port_by_guid(p_subn, p_subn->sm_port_guid);
	if (!p_port) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 4103: "
			"Failed to find the SM own port by guid\n");
		status = CL_ERROR;
		goto Exit;
	}

	/* get the node of the SM */
	p_node = p_port->p_node;

	p_dest_node = osm_physp_get_node_ptr(p_dest_physp);
	cnt_hops = p_path->hop_count;

	/*
	 * In case of destinaion physical port residing on a switch the
	 * last physp in the DR will be inserted into the map.
	 */
        if (osm_node_get_type(p_dest_node) != IB_NODE_TYPE_SWITCH)
                cnt_hops--;

	/*
	  traverse the path adding the out physp to the table
	  start after the first dummy hop and stop just before the
	  last one. Note the hop_count does not count entry 0.
	*/
	for (hop = 1; hop <= cnt_hops; hop++) {
		/* go out using the phys port of the path */
		p_physp = osm_node_get_physp_ptr(p_node, p_path->path[hop]);

		/* make sure we got a valid port and it has a remote port */
		if (!p_physp) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 4104: "
				"DR Traversal stopped on invalid port at hop:%u\n",
				hop);
			status = CL_ERROR;
			goto Exit;
		}

		/* 
		 * In case of SM residing on a switch the first physp is
		 * tracked as forbidden for use in alternative DR path.
		 */
		if (hop == 1 && osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
			cl_map_insert(p_physp_map, ptr_to_key(p_physp), p_physp);

		/* we track the ports we go out along the path */
		if (hop > 1)
			cl_map_insert(p_physp_map, ptr_to_key(p_physp), p_physp);

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Traversed through node: 0x%016" PRIx64
			" port:%u\n",
			cl_ntoh64(p_node->node_info.node_guid),
			p_path->path[hop]);

		if (!(p_physp = osm_physp_get_remote(p_physp))) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 4106: "
				"DR Traversal stopped on missing remote physp at hop:%u\n",
				hop);
			status = CL_ERROR;
			goto Exit;
		}

		p_node = osm_physp_get_node_ptr(p_physp);
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}

/*
 * BFS through the given physp to the target physp not going through
 * previously visited physp.
 * Return TRUE if found the path_num path and set the given p_new_dr_path.
 * Note that since the last step should NOT be marked we need to
 * check success before we move through the physp
 */
static boolean_t bfs_physp2physp_4new_path(IN osm_log_t *p_log,
					   IN osm_subn_t  const *p_subn,
					   IN osm_port_t  const *p_port,
					   IN osm_physp_t const *p_dest_physp,
					   IN cl_map_t *visited_out_physp_map,
					   IN uint8_t path_num,
					   OUT osm_dr_path_t *p_new_dr_path)
{
	osm_node_t  *p_node, *p_rem_node;
	osm_physp_t *p_physp, *p_in_physp, *p_out_physp, *p_rem_physp, *p_port_physp;
	uint8_t pn, hops, hop;
	uint8_t rev_order_path[IB_SUBNET_PATH_HOPS_MAX];
	cl_list_t out_physp_bfs_q;
	cl_map_t prev_out_physp_map; /* collect BFS steps */
	boolean_t found = FALSE;
	uint8_t num_found = 0;

	OSM_LOG_ENTER(p_log);

	cl_map_construct(&prev_out_physp_map);
	cl_map_init(&prev_out_physp_map, 4);

	cl_list_construct(&out_physp_bfs_q);
	cl_list_init(&out_physp_bfs_q, 36);

	/* insert to the bfs q all out physp of the starting port */
	p_node = osm_port_get_node(p_port);
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		for (pn = 1; pn < osm_node_get_num_physp(p_node); pn++) {
			p_physp = osm_node_get_physp_ptr(p_node, pn);

			/* was the switch out physp already visited ? */
			if (cl_map_get(visited_out_physp_map, ptr_to_key(p_physp)))
				continue;

			if (p_physp && (p_rem_physp = osm_physp_get_remote(p_physp))) {
				p_rem_node = osm_physp_get_node_ptr(p_rem_physp);
				if (p_node->physp_discovered[pn]) {
					if (osm_node_get_type(p_rem_node) == IB_NODE_TYPE_SWITCH) {
						cl_list_insert_head(&out_physp_bfs_q, p_physp);
						cl_map_insert(visited_out_physp_map, ptr_to_key(p_physp),
							      p_rem_physp);
					}
				}
			}
		}
	} else {
		p_physp = p_port->p_physp;
		OSM_ASSERT(p_physp);
		if (p_physp && (p_rem_physp = osm_physp_get_remote(p_physp))) {
			p_rem_node = osm_physp_get_node_ptr(p_rem_physp);
			if (p_node->physp_discovered[p_physp->port_num]) {
				if (osm_node_get_type(p_rem_node) == IB_NODE_TYPE_SWITCH) {
					cl_list_insert_head(&out_physp_bfs_q, p_physp);
					cl_map_insert(visited_out_physp_map, ptr_to_key(p_physp),
						      p_rem_physp);
				}
			}
		}
	}

	/* bfs until no more out pyhsp */
	while (!cl_is_list_empty(&out_physp_bfs_q)) {
		p_physp = cl_list_remove_head(&out_physp_bfs_q);
		p_in_physp = osm_physp_get_remote(p_physp);
		p_node = osm_physp_get_node_ptr(p_in_physp);
		OSM_ASSERT(osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH);

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"at node: 0x%016" PRIx64 " in-port:%u\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			osm_physp_get_port_num(p_in_physp));

		/* try all non visited physp leaving this node */
		for (pn = 1; pn < osm_node_get_num_physp(p_node); pn++) {
			p_out_physp = osm_node_get_physp_ptr(p_node, pn);
			if (!p_out_physp || (p_out_physp == p_in_physp))
				continue;

			p_rem_physp = osm_physp_get_remote(p_out_physp);
			if (!p_rem_physp)
				continue;

			/* was the out physp already visited ? */
			if (cl_map_get(visited_out_physp_map, ptr_to_key(p_out_physp)))
				continue;

			p_rem_node = osm_physp_get_node_ptr(p_rem_physp);
			if (osm_node_get_type(p_rem_node) == IB_NODE_TYPE_SWITCH)
				p_port_physp = osm_node_get_physp_ptr(p_rem_node, 0);
			else
				p_port_physp = p_rem_physp;

			if (p_port_physp == p_dest_physp) {
				num_found++;
				if (num_found >= path_num) {
					char buf[512];
					/* build the DR path by going backward from current p_physp */
					hops = 0;
					rev_order_path[hops++] = pn;
					do {
						rev_order_path[hops++] = osm_physp_get_port_num(p_physp);
						p_physp = cl_map_get(&prev_out_physp_map,
								     ptr_to_key(p_physp));
					} while (p_physp != NULL);

					p_new_dr_path->hop_count = hops;
					hop = 0;
					while (hops > 0)
						p_new_dr_path->path[hops--] = rev_order_path[hop++];

					osm_dump_dr_path_as_buf(sizeof(buf), p_new_dr_path, buf);
					OSM_LOG(p_log, OSM_LOG_DEBUG,
						"Reached destination through DR path:%s \n", buf);
					found = TRUE;
					goto Exit;
				} else
					OSM_LOG(p_log, OSM_LOG_DEBUG,
						"Found DR path %d out of the %d required\n",
						num_found, path_num);
			} else {
				/* did not reach destination */
				if (p_node->physp_discovered[pn]) {
					if (osm_node_get_type(p_rem_node) == IB_NODE_TYPE_SWITCH) {
						cl_map_insert(visited_out_physp_map, ptr_to_key(p_out_physp),
							      p_rem_physp);
						cl_map_insert(&prev_out_physp_map, ptr_to_key(p_out_physp),
							      p_physp);
						cl_list_insert_tail(&out_physp_bfs_q, p_out_physp);
					}
				}
			}
		}
	}
Exit:
	cl_list_remove_all(&out_physp_bfs_q);
	cl_map_remove_all(&prev_out_physp_map);

	cl_list_destroy(&out_physp_bfs_q);
	cl_map_destroy(&prev_out_physp_map);

	OSM_LOG_EXIT(p_log);

	return(found);
}

boolean_t osm_physp_replace_dr_path(IN osm_log_t *p_log,
				    IN osm_subn_t const *p_subn,
				    IN osm_physp_t *p_dest_physp)
{
	cl_map_t visited_out_physp_map;
	osm_dr_path_t *p_dr_path;
	osm_dr_path_t new_dr_path;
	ib_net64_t port_guid;
	osm_port_t *p_port;
	boolean_t success = FALSE;
	unsigned prev_trials;
	char oldDR[256], newDR[256];

	OSM_LOG_ENTER(p_log);

	prev_trials = osm_physp_get_num_dr_tried(p_dest_physp);
	if (prev_trials >= p_subn->opt.max_alt_dr_path_retries)
		goto Exit2;

	/*
	  initialize the map of all out physp participating in
	  current dr path not including first and last switches
	*/
	cl_map_construct(&visited_out_physp_map);
	cl_map_init(&visited_out_physp_map, 4);

	p_dr_path = osm_physp_get_dr_path_ptr(p_dest_physp);
	physp_get_dr_out_physp_set(p_log, p_subn, p_dr_path, p_dest_physp,
				   &visited_out_physp_map);

	osm_dr_path_construct(&new_dr_path);

	/* get the alternate path if possible */
	port_guid = p_subn->sm_port_guid;

	OSM_ASSERT(port_guid);

	p_port = osm_get_port_by_guid(p_subn, port_guid);
	if (!p_port) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 4105: No SM port object\n");
		goto Exit;
	}

	if (bfs_physp2physp_4new_path(p_log, p_subn, p_port, p_dest_physp,
				      &visited_out_physp_map, ++prev_trials,
				      &new_dr_path)) {

		osm_dump_dr_path_as_buf(sizeof(oldDR), p_dr_path, oldDR);
		osm_dump_dr_path_as_buf(sizeof(newDR), &new_dr_path, newDR);
		osm_dr_path_init(p_dr_path, new_dr_path.hop_count, new_dr_path.path);
		OSM_LOG(p_log, OSM_LOG_INFO, "Updated DR path of port 0x%016"
			PRIx64 " from: %s to: %s\n",
			cl_ntoh64(osm_physp_get_port_guid(p_dest_physp)),
			oldDR, newDR);
		success = TRUE;
		osm_physp_set_num_dr_tried(p_dest_physp, prev_trials);
	}

Exit:
	cl_map_remove_all(&visited_out_physp_map);
	cl_map_destroy(&visited_out_physp_map);

Exit2:
	OSM_LOG_EXIT(p_log);
	return(success);
}

boolean_t osm_link_is_healthy(IN const osm_physp_t * p_physp)
{
	osm_physp_t *p_remote_physp;

	OSM_ASSERT(p_physp);
	p_remote_physp = p_physp->p_remote_physp;
	if (p_remote_physp != NULL)
		return ((p_physp->healthy) & (p_remote_physp->healthy));
	/* the other side is not known - consider the link as healthy */
	return TRUE;
}

boolean_t osm_link_is_throttled(IN osm_physp_t * p_physp,
				IN const boolean_t subn_has_fdr10_enabled)
{
	osm_physp_t *p_remote;
	uint8_t speed_physp, speed_remote, width_physp, width_remote;
	uint8_t highest_speed, highest_width;
	boolean_t physp_has_extended_speeds_capability;
	boolean_t remote_has_extended_speeds_capability;
	ib_port_info_t *p_physp_info, *p_remote_info;

	CL_ASSERT(p_physp);
	p_remote = p_physp->p_remote_physp;

	/* the other side is not known - consider the link as unthrottled */
	if (!p_remote)
		return FALSE;

	/* only SP0 (and not Sw Ext.) have a valid CapabilityMask */
	if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH)
		p_physp_info =
		    &(osm_node_get_physp_ptr(p_physp->p_node, 0)->port_info);
	else
		p_physp_info = &p_physp->port_info;
	if (osm_node_get_type(p_remote->p_node) == IB_NODE_TYPE_SWITCH)
		p_remote_info =
		    &(osm_node_get_physp_ptr(p_remote->p_node, 0)->port_info);
	else
		p_remote_info = &p_remote->port_info;

	physp_has_extended_speeds_capability =
	    p_physp_info->capability_mask & IB_PORT_CAP_HAS_EXT_SPEEDS;
	remote_has_extended_speeds_capability =
	    p_remote_info->capability_mask & IB_PORT_CAP_HAS_EXT_SPEEDS;

	/* reset again to the original port_info */
	p_physp_info = &p_physp->port_info;
	p_remote_info = &p_remote->port_info;

	/* first determine the enabled link speed/width of both sides */
	speed_physp =
	    (physp_has_extended_speeds_capability ?
	     ib_port_info_get_link_speed_ext_enabled(p_physp_info) << 4 : 0)
	    + (subn_has_fdr10_enabled ?
	       (p_physp->ext_port_info.link_speed_enabled & FDR10) << 3 : 0)
	    + ib_port_info_get_link_speed_enabled(p_physp_info);
	width_physp = p_physp_info->link_width_enabled;

	speed_remote =
	    (remote_has_extended_speeds_capability ?
	     ib_port_info_get_link_speed_ext_enabled(p_remote_info) << 4 : 0)
	    + (subn_has_fdr10_enabled ?
	       (p_remote->ext_port_info.link_speed_enabled & FDR10) << 3 : 0)
	    + ib_port_info_get_link_speed_enabled(p_remote_info);
	width_remote = p_remote_info->link_width_enabled;

	highest_speed = ib_get_highest_link_speed(speed_physp & speed_remote);
	highest_width = ib_get_highest_link_width(width_physp & width_remote);

	/* and now determine the currently active link speed/width */
	speed_physp =
	    (physp_has_extended_speeds_capability ?
	     ib_port_info_get_link_speed_ext_active(p_physp_info) << 4 : 0)
	    + (subn_has_fdr10_enabled ?
	       (p_physp->ext_port_info.link_speed_active & FDR10) << 3 : 0)
	    + ib_port_info_get_link_speed_active(p_physp_info);
	speed_physp = ib_get_highest_link_speed(speed_physp);
	width_physp = p_physp_info->link_width_active;

	speed_remote =
	    (remote_has_extended_speeds_capability ?
	     ib_port_info_get_link_speed_ext_active(p_remote_info) << 4 : 0)
	    + (subn_has_fdr10_enabled ?
	       (p_remote->ext_port_info.link_speed_active & FDR10) << 3 : 0)
	    + ib_port_info_get_link_speed_active(p_remote_info);
	speed_remote = ib_get_highest_link_speed(speed_remote);
	width_remote = p_remote_info->link_width_active;

	/* check if the link supports same speed in both directions
	   and whether or not it runs at maximum speed/width which is
	   enabled by both ends (if not then its considered 'throttled')
	 */
	if (speed_physp != speed_remote || speed_physp != highest_speed ||
	    width_physp != width_remote || width_physp != highest_width)
		return TRUE;

	return FALSE;
}

void osm_physp_set_pkey_tbl(IN osm_log_t * p_log, IN const osm_subn_t * p_subn,
			    IN osm_physp_t * p_physp,
			    IN ib_pkey_table_t * p_pkey_tbl,
			    IN uint16_t block_num,
			    IN boolean_t is_set)
{
	uint16_t max_blocks;

	OSM_ASSERT(p_pkey_tbl);
	/*
	   (14.2.5.7) - the block number valid values are 0-2047, and are
	   further limited by the size of the P_Key table specified by
	   the PartitionCap on the node.
	 */
	if (!p_physp->p_node->sw || p_physp->port_num == 0)
		/*
		   The maximum blocks is defined in the node info: partition cap
		   for CA, router, and switch management ports.
		 */
		max_blocks =
		    (cl_ntoh16(p_physp->p_node->node_info.partition_cap) +
		     IB_NUM_PKEY_ELEMENTS_IN_BLOCK - 1)
		    / IB_NUM_PKEY_ELEMENTS_IN_BLOCK;
	else
		/*
		   This is a switch, and not a management port. The maximum
		   blocks is defined in the switch info: partition enforcement
		   cap.
		 */
		max_blocks =
		    (cl_ntoh16(p_physp->p_node->sw->switch_info.enforce_cap) +
		     IB_NUM_PKEY_ELEMENTS_IN_BLOCK -
		     1) / IB_NUM_PKEY_ELEMENTS_IN_BLOCK;

	if (block_num >= max_blocks) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 4108: "
			"Got illegal update for block number:%u max:%u "
			"for GUID: %" PRIx64 " port number:%u\n",
			block_num, max_blocks,
			cl_ntoh64(p_physp->p_node->node_info.node_guid),
			p_physp->port_num);
		return;
	}

	/* decrement block received counter */
	if(!is_set)
		p_physp->pkeys.rcv_blocks_cnt--;
	osm_pkey_tbl_set(&p_physp->pkeys, block_num, p_pkey_tbl,
			 p_subn->opt.allow_both_pkeys);
}

osm_alias_guid_t *osm_alias_guid_new(IN const ib_net64_t alias_guid,
				     IN osm_port_t *p_base_port,
				     IN uint16_t vport_index)
{
	osm_alias_guid_t *p_alias_guid;

	p_alias_guid = calloc(1, sizeof(*p_alias_guid));
	if (p_alias_guid) {
		p_alias_guid->alias_guid = alias_guid;
		p_alias_guid->p_base_port = p_base_port;
		p_alias_guid->dirty = FALSE;
		p_alias_guid->base_port_guid = osm_port_get_guid(p_base_port);
		p_alias_guid->vport_index = vport_index;
	}
	return p_alias_guid;
}

void osm_alias_guid_delete(IN OUT osm_alias_guid_t ** pp_alias_guid)
{
	free(*pp_alias_guid);
	*pp_alias_guid = NULL;
}


const osm_pkey_tbl_t *osm_alias_guid_get_pkey_tbl(IN const osm_alias_guid_t *p_alias_guid)
{
	const osm_pkey_tbl_t *pkey_tbl = NULL;
	osm_physp_t *p_physp;
	osm_vport_t *p_vport;

	if (p_alias_guid->dirty)
		goto EXIT;;

	if (p_alias_guid->vport_index) {
		p_vport = osm_port_get_vport(p_alias_guid->p_base_port,
					     p_alias_guid->vport_index);
		pkey_tbl = &p_vport->pkeys;
	} else {
		p_physp = p_alias_guid->p_base_port->p_physp;
		pkey_tbl = osm_physp_get_pkey_tbl(p_physp);
	}
EXIT:
	return pkey_tbl;
}
void osm_physp_set_port_info(IN osm_physp_t * p_physp,
			     IN const ib_port_info_t * p_pi,
			     IN const struct osm_sm * p_sm,
			     IN boolean_t set_timestamp)
{
	OSM_ASSERT(p_pi);
	OSM_ASSERT(p_physp);
	OSM_ASSERT(osm_physp_is_valid(p_physp));

	/* If port state change, update port's timestamp */
	if (set_timestamp ||
	    (ib_port_info_get_port_state(p_pi) !=
	     ib_port_info_get_port_state(&p_physp->port_info)))
		osm_physp_update_state_timestamp(p_physp);

	if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN) {
		/* If PortState is down, only copy PortState */
		/* and PortPhysicalState per C14-24-2.1 */
		ib_port_info_set_port_state(&p_physp->port_info, IB_LINK_DOWN);
		ib_port_info_set_port_phys_state
		    (ib_port_info_get_port_phys_state(p_pi),
		     &p_physp->port_info);
		p_physp->port_info.mtu_cap = IB_MTU_LEN_4096;
		p_physp->port_info.mtu_smsl = 0x44;
                p_physp->port_info.base_lid = p_pi->base_lid;
                p_physp->port_info.mkey_lmc = p_pi->mkey_lmc;
	} else {
		p_physp->port_info = *p_pi;

		/* The MKey in p_pi can only be considered valid if it's
		 * for a HCA/router or switch port 0, and it's either
		 * non-zero or the MKeyProtect bits are also zero.
		 */
		if ((osm_node_get_type(p_physp->p_node) !=
		     IB_NODE_TYPE_SWITCH || p_physp->port_num == 0) &&
		    (p_pi->m_key != 0 || ib_port_info_get_mpb(p_pi) == 0))
			osm_db_guid2mkey_set(p_sm->p_subn->p_g2m,
					     cl_ntoh64(p_physp->port_guid),
					     cl_ntoh64(p_pi->m_key));
	}
}

ib_net64_t osm_physp_calc_mkey(IN osm_subn_t *p_subn, IN osm_physp_t *p_physp)
{
	if (p_subn->opt.m_key_per_port) {
		/* Check if cached M_KEY is valid */
		if (p_physp->m_key_seed != p_subn->opt.m_key) {
			p_physp->m_key =
			    osm_subn_calc_mkey(p_subn, osm_physp_get_port_guid(p_physp));
			p_physp->m_key_seed = p_subn->opt.m_key;
		}
	} else {
		p_physp->m_key = p_subn->opt.m_key;
		p_physp->m_key_seed = 0;
	}

	return p_physp->m_key;
}

static void alias_guid_remove_mc_subscriptions(osm_sm_t *p_sm,
					       osm_port_t *p_port,
					       osm_alias_guid_t *p_alias_guid)
{
	cl_list_item_t *p_list_item;
	osm_mcm_port_t *p_mcm_port;
	osm_mcm_alias_guid_t *p_mcm_alias_guid;
	ib_net64_t guid = p_alias_guid->alias_guid;;

	p_list_item = cl_qlist_head(&p_port->mcm_list);
	while (p_list_item != cl_qlist_end(&p_port->mcm_list)) {
		p_mcm_port = cl_item_obj(p_list_item, p_mcm_port, list_item);
		p_list_item = cl_qlist_next(p_list_item);
		p_mcm_alias_guid = osm_mgrp_get_mcm_alias_guid(p_mcm_port->mgrp,
							       guid);
		if (p_mcm_alias_guid)
			osm_mgrp_remove_mcm_alias_guid(p_sm->p_subn,
						       p_sm->p_log,
						       p_mcm_port->mgrp,
						       p_mcm_alias_guid);
	}
}

static void vport_remove_ii_subscriptions(osm_sm_t *p_sm, osm_vport_t *p_vport)
{
	cl_list_item_t *p_item, *p_next_item;
	osm_infr_t *p_infr;
	ib_net64_t guid;

	guid = p_vport->vport_info.port_guid;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Removing Vport 0x%" PRIx64 " from"
		"InformInfo subscriptions list\n",
		cl_hton64(guid));

	p_next_item = cl_qlist_head(&p_sm->p_subn->sa_infr_list);
	while (p_next_item != cl_qlist_end(&p_sm->p_subn->sa_infr_list)) {
		p_item = p_next_item;
		p_next_item = cl_qlist_next(p_next_item);
		p_infr = (osm_infr_t *)p_item;
		if ((p_infr->inform_record.subscriber_gid.unicast.interface_id ==
		     guid) &&
		    (p_infr->inform_record.subscriber_gid.unicast.prefix ==
		     p_sm->p_subn->opt.subnet_prefix))
			osm_infr_remove_from_db(p_sm->p_subn, p_sm->p_log,
						p_infr);
	}
}


static void post_vport_changed_event(osm_sm_t *p_sm, ib_net64_t port_guid,
				     ib_net64_t vport_guid, uint16_t vport_index,
				     ib_net64_t vnode_guid,
				     osm_activity_mgr_state_enum state)
{
	event_data_t event_data;

	memset(&event_data, 0, sizeof(event_data_t));
	event_data.event_id = OSM_ACT_MGR_VPORT_STATE_CHANGE;
	event_data.port_guid = port_guid;
	event_data.vport_guid = vport_guid;
	event_data.vport_index = vport_index;
	event_data.vnode_guid = vnode_guid;
	event_data.state = state;
	osm_act_mgr_post_event(&p_sm->activity_mgr, &event_data);
}

/*
 * This method does not verify that index is not higher than top
 * This is up to the user of this method to perform this verification
 */
void osm_port_delete_vport(IN osm_sm_t *p_sm, osm_port_t *p_port,
			   IN uint16_t index)
{
	struct osm_vport **pp_vport;
	ib_net64_t vport_guid;
	osm_alias_guid_t *p_alias_guid;
	cl_qmap_t  *alias_guid_tbl;
	ib_gid_t vport_gid;
	uint8_t vport_state;
	osm_vnode_t *p_vnode;

	//TODO check if there is an entry in the map with the alias guid and the list length is 1

	pp_vport = p_port->virt_data.vports + index;

	if (!(*pp_vport))
		return;

	vport_guid = (*pp_vport)->vport_info.port_guid;

	/*
	 * pkey table not created and guid not added to alias guid tbl for
	 * index 0 so nothing to do if this is index 0
	 */
	if (!index)
		goto DELETE;

	alias_guid_tbl = &p_sm->p_subn->alias_port_guid_tbl;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Deleting vport %u of Port 0x%"
		PRIx64 "\n", index, cl_ntoh64(p_port->guid));

	if (p_sm->p_subn->opt.drop_event_subscriptions)
		vport_remove_ii_subscriptions(p_sm, *pp_vport);

	p_alias_guid =(osm_alias_guid_t*)cl_qmap_get(alias_guid_tbl, vport_guid);
	if (p_alias_guid == (osm_alias_guid_t*)cl_qmap_end(alias_guid_tbl))
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 410E: "
			"Vport guid 0x%" PRIx64 " not found in "
			"alias_port_guid_tbl (index %u of port 0x%" PRIx64
			")\n", cl_ntoh64(vport_guid), (*pp_vport)->index,
			cl_ntoh64(p_port->guid));
	else {
		vport_state =
		    ib_vport_info_get_vport_state(&(*pp_vport)->vport_info);
		if (p_sm->p_subn->opt.aguid_inout_notice &&
		    vport_state == IB_LINK_ACTIVE &&
		    !osm_exit_flag) {
			vport_gid.unicast.prefix = p_sm->p_subn->opt.subnet_prefix;
			vport_gid.unicast.interface_id = vport_guid;
			osm_report_notice_64_67(p_sm->p_subn, &vport_gid, 65);
		}

		cl_qmap_remove_item(alias_guid_tbl, &p_alias_guid->map_item);
		alias_guid_remove_mc_subscriptions(p_sm, p_port, p_alias_guid);
		osm_alias_guid_delete(&p_alias_guid);
	}

	osm_pkey_tbl_destroy(&(*pp_vport)->pkeys);

	if (osm_vport_is_lid_required(*pp_vport)) {
		osm_lid_mgr_clear_vport_lid(&p_sm->lid_mgr, p_port, *pp_vport,
					    FALSE);
		osm_ucast_cache_add_vport(&p_sm->ucast_mgr, p_port, index);
	}

DELETE:

	osm_virt_mgr_unregister_from_vnode(p_sm, (*pp_vport)->vnode_guid,
					   *pp_vport);

	post_vport_changed_event(p_sm, p_port->guid, vport_guid, index,
				 (*pp_vport)->vnode_guid, OSM_ACT_MGR_STATE_DOWN);

	p_vnode = osm_virt_mgr_get_vnode(p_sm, (*pp_vport)->vnode_guid);
	if (!p_vnode)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 4112: "
			"Could not find vnode 0x%" PRIx64 " of vport 0x%"
			PRIx64 "\n", cl_ntoh64((*pp_vport)->vnode_guid),
			cl_ntoh64(vport_guid));

	free(*pp_vport);
	*pp_vport = NULL;
	osm_virt_mgr_set_dump_required(&p_sm->virt_mgr, TRUE);
}

/*
 * At the end of this method we can be sure that:
 *
 * 1) vports array has enough vports blocks to support new top index value
 * 2) Any index heigher than the new top is initialized to zero
 *   (till end of the block)
 */
boolean_t osm_port_set_virt_info(IN struct osm_sm *p_sm, osm_port_t *p_port,
				 IN const ib_virtualization_info_t *p_vi)
{
	uint16_t required_blocks, new_blocks;
	uint16_t old_top, new_top, index;
	boolean_t res = TRUE;
	void *p_tmp, *new_blocks_ptr;
	uint32_t new_array_size;

	OSM_LOG_ENTER(p_sm->p_log);
	old_top = cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);
	new_top = cl_ntoh16(p_vi->vport_index_top);


	/*
	 * Handle the case that port discovered for the first time with top = 0
	 * Need to allocate one block for index 0
	 */
	if (!p_port->virt_data.vports && (new_top == 0)) {
		p_port->virt_data.vports  = calloc(1, VPORTS_BLOCK_BYTES);
		if (!p_port->virt_data.vports) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 410F:"
				"No memory to reallocate vports array\n");
			res = FALSE;
			goto EXIT;
		}
		p_port->virt_data.allocated_blocks = 1;
		goto COPY_AND_EXIT;
	}

	/*
	 * The top is the max number of vports to support, if it haven't
	 * changed then there is nothing to do other than copy the mad data
	 */
	if (new_top == old_top)
		goto COPY_AND_EXIT;

	/*
	 * If the top became smaller we need to delete the vports which are no
	 * longer supported (vports with index higher than new top)
	 */
	if (new_top < old_top) {
		for (index = (new_top + 1) ; index <= old_top ; index++)
			osm_port_delete_vport(p_sm, p_port, index);
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64 " (Node 0x%"
			PRIx64 ") - Vports top reduced from %u to %u, Deleting"
			" vports with index %u or higher\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			old_top, new_top, (new_top + 1));
		goto COPY_AND_EXIT;
	}

	/*
	 * Number of vports has grown (new top > current top)
	 * We need to check if new vports top can fit into the current
	 * allocated blocks of vports, if it does than we have nothing
	 * to do, if not we need to reallocate vports array
	 */
	required_blocks = new_top/OSM_VPS_BLOCK_SIZE + 1;

	if (required_blocks > p_port->virt_data.allocated_blocks) {
		/* Need to reallocate vports array to fit new top */
		new_blocks = required_blocks - p_port->virt_data.allocated_blocks;
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64 " (Node 0x%"
			PRIx64 ") - Vports top increased from %u to %u, "
			"Allocating %u blocks to fit new top\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			old_top, new_top, new_blocks);
		new_array_size = required_blocks * VPORTS_BLOCK_BYTES;
		p_tmp = realloc(p_port->virt_data.vports, new_array_size);
		if (!p_tmp) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 4109:"
				"No memory to reallocate vports array\n");
			res = FALSE;
			goto EXIT;
		}
		p_port->virt_data.vports = p_tmp;
		/* Initialize the new blocks we allocated to zero */
		new_blocks_ptr = p_port->virt_data.vports +
			p_port->virt_data.allocated_blocks * OSM_VPS_BLOCK_SIZE;
		memset(new_blocks_ptr , 0, new_blocks * VPORTS_BLOCK_BYTES);
		p_port->virt_data.allocated_blocks = required_blocks;
	} else
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64 " (Node "
			"0x%" PRIx64 ") - Vports top increased from %u to %u, "
			"No need to allocate new blocks\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			old_top, new_top);

COPY_AND_EXIT:
	p_port->virt_data.virtualization_info = *p_vi;
EXIT:
	OSM_LOG_EXIT(p_sm->p_log);
	return res;
}

osm_vport_t *osm_port_create_vport(IN struct osm_sm *p_sm, osm_port_t * p_port,
				   IN uint16_t index, IN ib_vport_info_t *p_vpi,
				   IN osm_vpi_context_t *p_vpi_context)
{
	uint8_t state;
	osm_vport_t *p_vport = NULL;
	osm_alias_guid_t *p_alias_guid, *p_alias_guid_check;
	ib_net64_t orig_port_guid;

	OSM_LOG_ENTER(p_sm->p_log);

	state = ib_vport_info_get_vport_state(p_vpi);
	if (state == IB_LINK_DOWN) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 4111: "
			"Port 0x%" PRIx64 " got request to create "
			"vport with state DOWN at index %u\n",
			cl_ntoh64(p_port->guid), index);
		goto EXIT;
	}


	p_vport = osm_port_get_vport(p_port, index);
	if (p_vport) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 410A: "
			"Vport %u already exist in the port - Abandoning vport "
			"creation\n", index);
		goto EXIT;
	}

	p_vport = calloc(1, sizeof(osm_vport_t));
	if (!p_vport) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 410B: "
			"Failed allocating memory for Vport %u\n", index);
		goto EXIT;
	}

	p_vport->vport_info = *p_vpi;
	p_vport->index = index;
	p_vport->p_port = p_port;
	p_vport->report_new_vport = TRUE;
	p_vport->vnode_guid = p_vpi_context->vnode_guid;
	p_vport->vnode_port = p_vpi_context->vnode_port;

	if (!osm_virt_mgr_register_to_vnode(p_sm, p_vport->vnode_guid, p_vport)) {
		free(p_vport);
		p_vport = NULL;
		goto EXIT;
	}

	/*
	 * Vport 0 represents the physical port so it should not be added to
	 * the alias guids table (this guid is already in the table) and it
	 * doesn't own pkey table of it's own (uses pkey table of the physp)
	 * so we can jump over these stages for index 0
	 */
	if (!index)
		goto SET_PTR;

	/* insert into alias guid table */
	p_alias_guid = osm_alias_guid_new(p_vpi->port_guid, p_port, index);
	if (!p_alias_guid) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 410C: "
			"vport alias guid memory allocation failed"
			" for port GUID 0x%" PRIx64 " vport index %u\n",
			cl_ntoh64(p_vpi->port_guid), index);
		osm_virt_mgr_unregister_from_vnode(p_sm, p_vport->vnode_guid,
						   p_vport);
		free(p_vport);
		p_vport= NULL;
		goto EXIT;
	}

	p_alias_guid_check =
		(osm_alias_guid_t *) cl_qmap_insert(&p_sm->p_subn->alias_port_guid_tbl,
						    p_alias_guid->alias_guid,
						    &p_alias_guid->map_item);
	if (p_alias_guid_check != p_alias_guid) {
		if (p_alias_guid_check->dirty) {
			p_alias_guid_check->dirty = FALSE;
			p_alias_guid_check->p_base_port = p_alias_guid->p_base_port;
			p_alias_guid_check->base_port_guid = p_alias_guid->base_port_guid;
			p_alias_guid_check->vport_index = p_alias_guid->vport_index;
			osm_alias_guid_delete(&p_alias_guid);
		} else {
			/* alias GUID is a duplicate
			 * This vport already registered to its vnode,
			 * this means that this vport already exist on a
			 * different port under a different vnode guid, so
			 * this is not a migration but an error, need to delete
			 * the vport
			 */

			orig_port_guid =
				    osm_port_get_guid(p_alias_guid_check->p_base_port);

			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Vport 0x%" PRIx64
				" already exist on port 0x%" PRIx64 " under a"
				" different VNode, this is illegal, deleting "
				" vport\n", cl_ntoh64(p_vpi->port_guid),
				cl_ntoh64(orig_port_guid));

			osm_alias_guid_delete(&p_alias_guid);
			osm_virt_mgr_unregister_from_vnode(p_sm,
							   p_vport->vnode_guid,
							   p_vport);
			free(p_vport);
			p_vport = NULL;
			goto EXIT;
		}
	}

	osm_pkey_tbl_construct(&p_vport->pkeys);
	osm_pkey_tbl_init(&p_vport->pkeys);

	/*
	 * In beta stage we do not support pkeys per vports, each vport shall
	 * only the default pkey so for now we manually plant the default pkey
	 * in each vport (and we do not need to set the table in the vport)
	 * this shall be updated on GA
	 */
	if (p_sm->p_subn->opt.allow_both_pkeys)
		cl_map_insert(&p_vport->pkeys.keys, def_pkey, &def_pkey);
	else
		cl_map_insert(&p_vport->pkeys.keys, ib_pkey_get_base(def_pkey),
			      &def_pkey);

SET_PTR:
	*(p_port->virt_data.vports + index) = p_vport;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64 " (Node 0x%"
		PRIx64 ") - Vport %u created successfully (guid = 0x%" PRIx64
		", addr = %p)\n", cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)), index,
		cl_ntoh64(p_vpi->port_guid), p_vport);

	post_vport_changed_event(p_sm, p_port->guid, p_vpi->port_guid, index,
				 p_vport->vnode_guid,
				 (osm_activity_mgr_state_enum)state);

	osm_virt_mgr_set_dump_required(&p_sm->virt_mgr, TRUE);
EXIT:
	OSM_LOG_EXIT(p_sm->p_log);
	return p_vport;
}

void osm_port_update_vport(IN struct osm_sm *p_sm, osm_port_t *p_port,
			   IN uint16_t index, IN ib_vport_info_t *p_vpi,
			   IN osm_vpi_context_t *p_vpi_context)
{
	uint8_t old_state, new_state;
	osm_vport_t *p_vport;
	ib_net64_t vnode_guid = p_vpi_context->vnode_guid;
	uint8_t vnode_port = p_vpi_context->vnode_port;

	p_vport = osm_port_get_vport(p_port, index);
	if (!p_vport) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 4110: "
			"Port 0x%" PRIx64 " (Node 0x%" PRIx64 ") - Request "
			"to update vport index %u which does not exist\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			index);
		return;
	}

	if (p_vport->vnode_guid != vnode_guid) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 4111: "
			"Port 0x%" PRIx64 " (Node 0x%" PRIx64 ") - "
			"Vport %u (guid = 0x%" PRIx64 ", addr = %p) VNode "
			"changed from 0x%" PRIx64 " to 0x%" PRIx64
			" - Ignoring VPI Message\n", cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)), index,
			cl_ntoh64(p_vpi->port_guid), p_vport,
			cl_ntoh64(p_vport->vnode_guid), cl_ntoh64(vnode_guid));
		return;
	}

	if (p_vport->vnode_port != vnode_port) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64
			" (Node 0x%" PRIx64 ") - Vport %u (guid = 0x%" PRIx64
			", addr = %p) VNode 0x%" PRIx64 ", port changed from "
			"%u to %u\n", cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
			index, cl_ntoh64(p_vpi->port_guid), p_vport,
			cl_ntoh64(vnode_guid), p_vport->vnode_port, vnode_port);
		p_vport->vnode_port = vnode_port;
	}

	if (memcmp(&p_vport->vport_info, p_vpi, sizeof(ib_vport_info_t))) {
		osm_virt_mgr_set_dump_required(&p_sm->virt_mgr, TRUE);

		old_state = ib_vport_info_get_vport_state(&p_vport->vport_info);
		new_state = ib_vport_info_get_vport_state(p_vpi);

		p_vport->vport_info = *p_vpi;

		if (old_state != new_state) {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64
				" Vport %u state changed from %u to %u\n",
				cl_ntoh64(p_port->guid), index, old_state,
				new_state);

			post_vport_changed_event(p_sm, p_port->guid,
						 p_vpi->port_guid, index,
						 p_vport->vnode_guid,
						 (osm_activity_mgr_state_enum)new_state);
		}

		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64
			" (Node 0x%" PRIx64 ") - Vport %u (guid = 0x%" PRIx64
			", addr = %p) updated (state = %u)\n",
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)), index,
			cl_ntoh64(p_vpi->port_guid), p_vport, new_state);
	}

}

void osm_port_decrease_virt_transactions(IN struct osm_sm *p_sm,
					 osm_port_t *p_port)
{
	OSM_ASSERT(p_port->virt_data.transactions);

	p_port->virt_data.transactions--;
	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64
		" (Node 0x%" PRIx64 ") decreased transactions value to "
		"%u\n", cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		p_port->virt_data.transactions);
}

void osm_port_increase_virt_transactions(IN struct osm_sm *p_sm,
					 osm_port_t *p_port)
{
	p_port->virt_data.transactions++;
	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Port 0x%" PRIx64
		" (Node 0x%" PRIx64 ") increased transactions value to "
		"%u\n", cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		p_port->virt_data.transactions);
}

void osm_alias_guid_desc(IN const osm_alias_guid_t *p_alias_guid, OUT char *desc)
{
	if (p_alias_guid->vport_index)
		sprintf(desc, "VPort 0x%" PRIx64 ", VPort index %u of port "
			"0x%" PRIx64,
			cl_ntoh64(p_alias_guid->alias_guid),
			p_alias_guid->vport_index,
			cl_ntoh64(p_alias_guid->base_port_guid));
	else
		sprintf(desc, "Port 0x%" PRIx64 ", Port %u of %s",
			cl_ntoh64(p_alias_guid->alias_guid),
			p_alias_guid->p_base_port->p_physp->port_num,
			p_alias_guid->p_base_port->p_node->print_desc);
}

boolean_t osm_port_is_vport_lid_assigned(osm_port_t *p_port,
					 uint16_t vport_index)
{
	osm_vport_t *p_vport = osm_port_get_vport(p_port, vport_index);

	if (p_vport && osm_vport_get_lid(p_vport))
		return TRUE;

	return FALSE;
}

osm_vport_t *osm_port_get_vport_by_lid(const osm_port_t *p_port,
				       ib_net16_t lid)
{
	osm_vport_t *p_vport;
	boolean_t found = FALSE;
	uint16_t i, top;

	i = 1;
	top = cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);
	while ((i <= top) && !found) {
		p_vport = osm_port_get_vport(p_port, i);
		if (p_vport && (lid == osm_vport_get_lid(p_vport)))
			found = TRUE;
		i++;
	}

	return found ? p_vport : NULL;
}

boolean_t osm_port_is_virt_supported(IN osm_port_t *p_port)
{
	return ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
				IB_GENERAL_INFO_VIRTUALIZATION_SUPPORTED_MASK);
}

void  osm_port_update_vports_mcast_membership(IN struct osm_sm *sm,
					      IN osm_port_t *p_port,
					      IN osm_mgrp_t *mgrp)

{
	osm_vport_t *p_vport;
	uint16_t vport_index_top;
	uint16_t vport_index;
	ib_net64_t guid;
	osm_mcm_alias_guid_t *p_mcm_alias_guid;
	boolean_t mgrp_deleted = FALSE;
	ib_net16_t pkey;

	OSM_ASSERT(p_port);
	OSM_ASSERT(mgrp);

	OSM_LOG_ENTER(sm->p_log);

	if (!(vport_index_top = osm_port_get_vport_top_index(p_port)))
		goto Exit;

	for (vport_index = 1 ;
	     !mgrp_deleted && vport_index <= vport_index_top;
	     vport_index++) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport)
			continue;

		pkey = mgrp->mcmember_rec.pkey;
		/* skip MCG with default pkey */
		if (ib_pkey_get_base(pkey) == IB_DEFAULT_PARTIAL_PKEY)
				continue;

		if (osm_pkey_tbl_has_pkey(sm->p_log,
					  pkey,
					  &p_vport->pkeys,
					  sm->p_subn->opt.allow_both_pkeys))
			continue;

		guid = osm_vport_get_guid(p_vport);
		p_mcm_alias_guid = osm_mgrp_get_mcm_alias_guid(mgrp, guid);
		if (p_mcm_alias_guid)
			mgrp_deleted =
				osm_mgrp_remove_mcm_alias_guid(sm->p_subn,
							       sm->p_log,
							       mgrp,
							       p_mcm_alias_guid);
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
}

boolean_t osm_is_dr_path_valid(IN osm_sm_t * sm,
			       IN const osm_dr_path_t * p_path)
{
	osm_node_t *p_node;
	osm_port_t *p_sm_port;
	osm_physp_t *p_physp = NULL;
	uint8_t hop;
	boolean_t status = TRUE;

	OSM_LOG_ENTER(sm->p_log);

	/* hop_count == 0: destination port guid is SM */
	if (p_path->hop_count == 0) {
		status = TRUE;
		goto Exit;
	}

	p_sm_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
	if (!p_sm_port) {
		status = FALSE;
		goto Exit;
	}
	p_node = osm_port_get_node(p_sm_port);
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, p_path->path[1]);
	else
		p_physp = p_sm_port->p_physp;

	/* hop_count == 1: outgoing physp is SM physp */
	for (hop = 2; p_physp && hop <= p_path->hop_count; hop++) {
		p_physp = p_physp->p_remote_physp;
		if (!p_physp)
			break;
		p_node = p_physp->p_node;

		if (!p_node || p_path->path[hop] >= p_node->physp_tbl_size)
			p_physp = NULL;
		else
			p_physp = osm_node_get_physp_ptr(p_node, p_path->path[hop]);
	}

	status = (p_physp != NULL);
Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}
