/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Port Group manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PORT_GRP_MGR_C
#include <opensm/osm_switch.h>
#include <opensm/osm_log.h>
#include <opensm/osm_port_group.h>
#include <opensm/osm_opensm.h>

extern int parse_port_groups_policy(struct osm_opensm *p_osm);
extern char *trim_whitespace(char *str);

#define HOSTNAME "hostname"
#define PORT "port"
#define HCA_ID "hca_idx"
#define PORTS_DELIM "/"
#define PARAM_DELIM ";"
#define EQUAL "="
#define MAX_PORT_STR_LEN 4
#define MAX_PORTS 256

uint32_t osm_ports_grp_count(IN struct osm_opensm *p_osm)
{
	return (cl_qlist_count(&p_osm->sm.port_group_mgr.port_groups));
}

static cl_status_t match_group_name(IN const cl_list_item_t * p_list_item,
				    IN void *context)
{
	char *name = (char *)context;
	osm_port_group_t *p_group = (osm_port_group_t *) p_list_item;
	if (!strcmp(name, p_group->name))
		return CL_SUCCESS;
	else
		return CL_NOT_FOUND;
}

osm_port_group_t *osm_get_port_grp(IN struct osm_opensm * p_osm,
				   IN const char *group_name)
{
	cl_list_item_t *p_item;
	osm_port_group_mgr_t *p_mgr = &p_osm->sm.port_group_mgr;
	p_item = cl_qlist_find_from_head(&p_mgr->port_groups, match_group_name,
					 group_name);
	if (p_item == cl_qlist_end(&p_mgr->port_groups))
		return NULL;
	else
		return (osm_port_group_t *) p_item;
}

static void set_all_physp(osm_port_group_member_t * p_member, boolean_t val)
{
	uint32_t i;

	p_member->physp_members[0] = FALSE;
	for (i = 1; i < p_member->port->p_node->physp_tbl_size; i++)
		p_member->physp_members[i] = val;
}

/* Check if member has at least one physp set in it */
static boolean_t has_physp(osm_port_group_member_t * p_member)
{
	uint32_t i = 1;
	boolean_t found = FALSE;

	while ((i < p_member->port->p_node->physp_tbl_size) && (!found)) {
		found = p_member->physp_members[i];
		i++;
	}
	return found;
}

static boolean_t is_legal_port(osm_port_t * p_port, uint8_t port_num)
{
	uint32_t tbl_size;

	tbl_size = p_port->p_node->physp_tbl_size;
	return ((port_num > 0) && (port_num < tbl_size));
}

static inline void set_physp(osm_port_group_member_t * p_member,
			     uint8_t port_num, boolean_t val)
{
	p_member->physp_members[port_num] = val;
}

boolean_t osm_set_physp_reliable(struct osm_opensm * p_osm,
			         osm_port_group_member_t * p_member,
				 uint8_t port_num, boolean_t val)
{
	if (is_legal_port(p_member->port, port_num)) {
		set_physp(p_member, port_num, val);
		return TRUE;
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "physp %d not legal for"
			" setting - Ignored\n", port_num);
		return FALSE;
	}
}

static boolean_t get_physp_val(osm_port_group_member_t * p_member,
			       uint8_t port_num)
{
	uint32_t tbl_size;

	tbl_size = p_member->port->p_node->physp_tbl_size;
	if ((port_num > 0) && (port_num < tbl_size))
		return p_member->physp_members[port_num];
	else
		return FALSE;
}

static boolean_t create_physp_tbl(osm_opensm_t * p_osm,
				  osm_port_group_member_t * p_mem,
				  boolean_t def_val)
{
	uint32_t tbl_size;

	tbl_size = p_mem->port->p_node->physp_tbl_size;
	p_mem->physp_members = malloc((tbl_size) * sizeof(boolean_t));
	if (!p_mem->physp_members)
		return FALSE;

	set_all_physp(p_mem, def_val);
	return TRUE;
}

osm_port_group_member_t *osm_create_port_grp_member(struct osm_opensm * p_osm,
						    struct osm_port * p_port,
						    boolean_t
						    physp_def_val)
{
	osm_port_group_member_t *p_member;

	p_member = malloc(sizeof(osm_port_group_member_t));

	if (!p_member) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3000: Failed to "
			"allocate Port Group member");
		return NULL;
	}

	p_member->port = p_port;

	if (create_physp_tbl(p_osm, p_member, physp_def_val) == FALSE) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3009: Failed to "
			 "create physp table\n");
		free(p_member);
		return NULL;
	}

	return p_member;
}

void osm_destroy_port_grp_member(osm_port_group_member_t * grp_mem)
{
	free(grp_mem->physp_members);
	free(grp_mem);
}

void osm_destroy_pgrp_rule(osm_pgrp_rule_t * p_rule)
{
	if (!p_rule)
		return;

	free(p_rule->list);
	free(p_rule->regex_str);
	free(p_rule->hostname);
	free(p_rule->src_port_grp_1);
	free(p_rule->src_port_grp_2);
	free(p_rule);
}

void osm_destroy_pgrp_rules_list(cl_qlist_t *p_list)
{
	cl_list_item_t *p_item;

	if (!p_list)
		return;

	p_item = cl_qlist_remove_head(p_list);
	while (p_item != cl_qlist_end(p_list)) {
		osm_destroy_pgrp_rule((osm_pgrp_rule_t*)p_item);
		p_item = cl_qlist_remove_head(p_list);
	}
	free (p_list);
}

void osm_destroy_port_grp(osm_port_group_t * port_group)
{
	cl_map_item_t *p_item;

	if (!port_group)
		return;
	free(port_group->name);
	free(port_group->use);

	while (cl_qmap_count(&port_group->ports) > 0) {
		p_item = cl_qmap_head(&port_group->ports);
		cl_qmap_remove_item(&port_group->ports, p_item);
		osm_destroy_port_grp_member((osm_port_group_member_t *) p_item);
	}

	osm_destroy_pgrp_rules_list(port_group->rules);
	free(port_group);
}

void osm_port_grp_mgr_construct(IN osm_port_group_mgr_t * p_mgr) {

	cl_qlist_init(&p_mgr->port_groups);
}

void osm_port_grp_mgr_destroy(IN osm_port_group_mgr_t * p_mgr)
{
	osm_port_group_t *p_item;

	while (cl_qlist_count(&p_mgr->port_groups) > 0) {
		p_item =
		    (osm_port_group_t *) cl_qlist_remove_head(&p_mgr->
							      port_groups);
		osm_destroy_port_grp(p_item);
	}
}

static void print_port_group(osm_opensm_t * p_osm,
			     osm_port_group_t * p_port_group)
{
	uint32_t i = 0;
	cl_map_item_t *p_item;
	cl_list_item_t *p_list_item;
	osm_port_group_member_t *grp_mem;
	uint32_t ports_num;
	char *ports_str;
	uint8_t count;
	char p_list[MAX_PORT_STR_LEN * MAX_PORTS];
	char port_num_str[MAX_PORT_STR_LEN];
	char *use;

	use = p_port_group->use ? p_port_group->use : "";
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Port group: %s, "
		"Use: %s, Ports Count: %d\n",
		p_port_group->name, use ,
		cl_qmap_count(&p_port_group->ports));

	p_list_item = cl_qlist_head(p_port_group->rules);

	while (p_list_item != cl_qlist_end(p_port_group->rules)) {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Rule %u: Type %u\n",
			++i, ((osm_pgrp_rule_t *)p_list_item)->rule_type);
		p_list_item = cl_qlist_next(p_list_item);
	}

	p_item = cl_qmap_head(&p_port_group->ports);
	while (p_item != cl_qmap_end(&p_port_group->ports)) {
		grp_mem = (osm_port_group_member_t *) p_item;
		if (grp_mem->physp_members) {
			memset(p_list, 0, (MAX_PORT_STR_LEN * MAX_PORTS));
			count = 0;
			ports_num = grp_mem->port->p_node->physp_tbl_size;
			for (i = 1; i < ports_num; i++) {
				if (grp_mem->physp_members[i]) {
					snprintf(port_num_str, MAX_PORT_STR_LEN,
						 ",%d", i);
					strcat(p_list,port_num_str);
					count++;
				}
			}
			if (count)
				if (count == (i - 1))
					ports_str = "All";
				else
					ports_str = p_list + 1;
			else
				ports_str = "None";;
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "0x%" PRIx64
				", Physp Ports: %s\n",
				cl_ntoh64(grp_mem->port->guid), ports_str);
		} else
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "0x%" PRIx64
				" has no Physp Ports\n",
				cl_ntoh64(grp_mem->port->guid));
		p_item = cl_qmap_next(p_item);
	}
}

static boolean_t add_member_to_grp(osm_opensm_t * p_osm,
				   osm_port_group_t * p_port_group,
				   osm_port_group_member_t * p_member)
{
	cl_map_item_t *p_item = cl_qmap_insert(&p_port_group->ports,
					       p_member->port->guid,
					       (cl_map_item_t *) p_member);
	/* If p_item not equal to p_member it mean that a member with
	 * the same key was already in the map */
	return (p_item == (cl_map_item_t *) p_member);
}

static osm_port_group_member_t *create_new_member_in_grp(osm_opensm_t * p_osm,
							 osm_port_group_t *
							 p_port_group,
							 osm_port_t * p_port,
							 boolean_t
							 physp_def_val)
{
	osm_port_group_member_t *p_member;

	p_member = osm_create_port_grp_member(p_osm, p_port, physp_def_val);
	if (p_member) {
		if (add_member_to_grp(p_osm, p_port_group, p_member))
			return p_member;
		else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 3003: Can't create new member in group"
				" - Member already found\n");
			free(p_member);
			return NULL;
		}
	} else
		return NULL;
}

void osm_remove_port_from_grp(IN osm_port_group_t * p_grp,
			      IN osm_port_t * p_port)
{
	cl_map_item_t *p_item;

	p_item = cl_qmap_get(&p_grp->ports, p_port->guid);

	if (p_item != cl_qmap_end(&p_grp->ports)) {
		cl_qmap_remove_item(&p_grp->ports, p_item);
		osm_destroy_port_grp_member((osm_port_group_member_t *) p_item);
	}
}

static osm_port_group_member_t *find_member(osm_port_group_t * p_grp,
					    ib_net64_t guid)
{
	cl_map_item_t *p_item;

	p_item = cl_qmap_get(&p_grp->ports, guid);

	if (p_item == cl_qmap_end(&p_grp->ports))
		return NULL;
	else
		return (osm_port_group_member_t *) p_item;
}

/* check if port already found in the group and do one of the below:
 *
 *	1) If not found and port num is zero - add it to the group together with
 *		all its physp's
 *
 *	2) If not found and port num is not zero - add it to the group only with
 *		the required port number
 *
 *	3) If found and port num is zero - add all physp's (set to TRUE)
 *
 *	4) If found and port num is not zero - add the required port num
 *		(set to TRUE)
 **/
static boolean_t set_port_in_grp(osm_opensm_t * p_osm,
				 osm_port_group_t * p_port_group,
				 osm_port_t * p_port, uint8_t port_num,
				 boolean_t val)
{
	osm_port_group_member_t *p_member;
	boolean_t res = TRUE;

	p_member = find_member(p_port_group, p_port->guid);
	if (p_member) {
		if (port_num)
			res = osm_set_physp_reliable(p_osm, p_member, port_num, val);
		else
			set_all_physp(p_member, val);
	} else {
		/* If request is to remove port/physp (negative val) and port
		 * doesn't exist there is nothing to do */
		if (!val)
			goto EXIT;

		/* We know for sure that this member not exist in the group */
		if (port_num) {
			if (is_legal_port(p_port, port_num)) {
				p_member = osm_create_port_grp_member(p_osm,
								      p_port,
								      FALSE);
				if (!p_member) {
					res = FALSE;
					goto EXIT;
				}
				set_physp(p_member, port_num, TRUE);
				add_member_to_grp(p_osm, p_port_group,
						  p_member);
			} else {
				res = FALSE;
				goto EXIT;
			}
		} else
		    if (!create_new_member_in_grp(p_osm, p_port_group,
						  p_port, TRUE))
			res = FALSE;
	}
EXIT:
	return res;
}

static inline uint64_t convert_guid_str(char *guid_str)
{
	char *e = NULL;

	return strtoull(guid_str, &e, 0);
}

static boolean_t
add_port_to_grp_by_guid(osm_opensm_t * p_osm, osm_port_group_t * p_grp,
			uint64_t guid, uint8_t port_num)
{
	boolean_t res = TRUE;

	osm_port_t *p_port = osm_get_port_by_guid(&p_osm->subn, guid);
	if (p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "found port 0x%" PRIx64
			" - Adding to group\n", cl_ntoh64(guid));
		res = set_port_in_grp(p_osm, p_grp, p_port, port_num, TRUE);
	} else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"port GUID 0x%" PRIx64 " NOT FOUND\n", cl_ntoh64(guid));

	return res;
}

static boolean_t create_predefined_port_groups(osm_opensm_t * p_osm,
					       osm_port_group_mgr_t * p_mgr)
{
	osm_port_group_t *all_group;
	osm_port_group_t *all_ca_group;
	osm_port_group_t *all_sw_group;
	osm_port_group_t *all_rtr_group;
	osm_port_t *p_item;

	all_group = osm_create_port_grp(p_osm, PG_ALL, "All Ports", NULL, TRUE);
	if (!all_group)
		return FALSE;

	all_sw_group = osm_create_port_grp(p_osm, PG_ALL_SW, "All SWITCHES", NULL, TRUE);
	if (!all_sw_group)
		return FALSE;

	all_ca_group = osm_create_port_grp(p_osm, PG_ALL_CA, "All CAS", NULL, TRUE);
	if (!all_ca_group)
		return FALSE;

	all_rtr_group = osm_create_port_grp(p_osm, PG_ALL_RTR, "All ROUTERSS", NULL, TRUE);
	if (!all_rtr_group)
		return FALSE;

	p_item = (osm_port_t *) cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_item != (osm_port_t *) cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		if (!set_port_in_grp(p_osm, all_group, p_item, 0, TRUE))
			return FALSE;
		if (osm_node_get_type(p_item->p_node) == IB_NODE_TYPE_SWITCH)
			if (!set_port_in_grp
			    (p_osm, all_sw_group, p_item, 0, TRUE))
				return FALSE;
		if (osm_node_get_type(p_item->p_node) == IB_NODE_TYPE_CA)
			if (!set_port_in_grp
			    (p_osm, all_ca_group, p_item, 0, TRUE))
				return FALSE;
		if (osm_node_get_type(p_item->p_node) == IB_NODE_TYPE_ROUTER)
			if (!set_port_in_grp
			    (p_osm, all_rtr_group, p_item, 0, TRUE))
				return FALSE;
		p_item = (osm_port_t *) cl_qmap_next((cl_map_item_t *) p_item);
	}
	all_group->applied = TRUE;
	all_ca_group->applied = TRUE;
	all_sw_group->applied = TRUE;
	all_rtr_group->applied = TRUE;

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"[%s: %d members, %s: %d members, " "%s: %d members, "
		"%s: %d members]\n",
		PG_ALL, cl_qmap_count(&all_group->ports),
		PG_ALL_SW, cl_qmap_count(&all_sw_group->ports),
		PG_ALL_CA, cl_qmap_count(&all_ca_group->ports),
		PG_ALL_RTR, cl_qmap_count(&all_rtr_group->ports));
	return TRUE;
}

static boolean_t merge_member_to_grp(osm_opensm_t * p_osm,
				     osm_port_group_member_t * p_member,
				     osm_port_group_t * p_grp)
{
	osm_port_t *p_port;
	osm_port_group_member_t *target_mem;
	uint32_t tbl_size;
	uint32_t i;

	if (!p_member)
		return FALSE;

	p_port = p_member->port;
	target_mem = find_member(p_grp, p_port->guid);

	if (!target_mem) {
		target_mem = create_new_member_in_grp(p_osm, p_grp,
						      p_port, FALSE);
		if (!target_mem)
			return FALSE;
	}

	tbl_size = p_member->port->p_node->physp_tbl_size;
	for (i = 0; i < tbl_size; i++)
		if (p_member->physp_members[i])
			set_physp(target_mem, i, TRUE);
	return TRUE;
}

static boolean_t copy_port_group_map(osm_opensm_t * p_osm,
				     osm_port_group_t * p_src_grp,
				     osm_port_group_t * p_dest_grp)
{
	cl_map_item_t *p_item;
	osm_port_group_member_t *p_member;
	boolean_t res = TRUE;

	p_item = cl_qmap_head(&p_src_grp->ports);
	while (p_item != cl_qmap_end(&p_src_grp->ports) && res) {
		p_member = (osm_port_group_member_t *) p_item;
		res = merge_member_to_grp(p_osm, p_member, p_dest_grp);
		p_item = cl_qmap_next(p_item);
	}
	return res;
}

boolean_t osm_physp_in_grp(IN osm_port_group_t * p_grp, IN ib_net64_t guid,
			   IN uint8_t port_num)
{
	cl_map_item_t *p_member;

	p_member = cl_qmap_get(&p_grp->ports, guid);
	if (p_member == cl_qmap_end(&p_grp->ports))
		return FALSE;
	else
		return get_physp_val((osm_port_group_member_t *) p_member,
				     port_num);
}

static boolean_t unite_groups(osm_opensm_t * p_osm, osm_port_group_t * p_grp1,
			      osm_port_group_t * p_grp2,
			      osm_port_group_t * p_dest)
{
	if (!copy_port_group_map(p_osm, p_grp1, p_dest) ||
	    !copy_port_group_map(p_osm, p_grp2, p_dest))
		return FALSE;

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"United port groups '%s' and '%s'\n",
		p_grp1->name, p_grp2->name);
	return TRUE;
}

static boolean_t subtract_groups(osm_opensm_t * p_osm,
				 osm_port_group_t * p_from_grp,
				 osm_port_group_t * p_grp_to_subtract,
				 osm_port_group_t * p_dest)
{
	osm_port_group_member_t *p_member;
	osm_port_group_member_t *p_dest_mem;
	osm_port_group_member_t *p_subtract;
	cl_map_item_t *p_item;
	osm_port_group_member_t *new_mem;
	uint32_t tbl_size;
	uint32_t i;
	boolean_t empty = TRUE;
	ib_net64_t guid;

	p_item = cl_qmap_head(&p_from_grp->ports);
	while (p_item != cl_qmap_end(&p_from_grp->ports)) {
		p_member = (osm_port_group_member_t *) p_item;
		p_item = cl_qmap_next(p_item);
		guid = p_member->port->guid;
		p_subtract = find_member(p_grp_to_subtract, guid);
		if (p_subtract) {
			tbl_size = p_member->port->p_node->physp_tbl_size;
			p_dest_mem = find_member(p_dest, guid);
			/* member exit in both groups */
			if (p_dest_mem) {
				/* Member exist in dest group, just need to set
				 * required ports
				 */
				for (i = 0; i < tbl_size; i++)
					if (p_member->physp_members[i] &&
					    !p_subtract->physp_members[i])
						set_physp(p_dest_mem, i, TRUE);
			} else {
				/* No such member in dest group:
				 * 1. Create new member
				 * 2. Set subtracted ports
				 * 3. If any port was set, add to dest group
				 * 4. If no port was set, delete member
				 */
				new_mem = osm_create_port_grp_member(p_osm,
								     p_member->port,
								     FALSE);
				if (!new_mem)
					return FALSE;

				for (i = 0; i < tbl_size; i++)
					if (p_member->physp_members[i] &&
					    !p_subtract->physp_members[i]) {
						set_physp(new_mem, i, TRUE);
						empty = FALSE;
					}

				if (empty)
					osm_destroy_port_grp_member(new_mem);
				else
					add_member_to_grp(p_osm, p_dest, new_mem);
			}
		} else {
			/* No member in 'subtract' group, just merge to dest */
			merge_member_to_grp(p_osm, p_member, p_dest);
		}
	}

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Subtracted port group '%s' from '%s'\n",
		p_grp_to_subtract->name, p_from_grp->name);

	return TRUE;
}

static boolean_t apply_range_rule(osm_opensm_t * p_osm,
				   osm_port_group_t * p_grp,
				   uint64_t min_guid, uint64_t max_guid)
{
	uint64_t guid;
	cl_map_item_t *p_item;
	boolean_t res = TRUE;

	p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		guid = cl_ntoh64(((osm_port_t *) p_item)->guid);
		if ((guid >= min_guid) && (guid <= max_guid)) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Adding port"
				" 0x%" PRIx64 "\n", guid);
			if (!set_port_in_grp
			    (p_osm, p_grp, (osm_port_t *) p_item, 0, TRUE)) {
				res = FALSE;
				goto EXIT;
			}
		}
		p_item = cl_qmap_next(p_item);
	}
EXIT:
	return res;
}

static boolean_t apply_regex_rule(osm_opensm_t * p_osm,
				  osm_port_group_t * p_port_group,
				  osm_pgrp_rule_t *p_rule)
{
	cl_map_item_t *p_item;
	char *desc;
	osm_node_t *p_node;
	int res;
	regex_t *regex;
	char err_buf[100];
	boolean_t apply_res = TRUE;
	osm_port_t *p_port;
	uint8_t max_port, i;

	regex = &p_rule->regex;
	if (regcomp(regex, p_rule->regex_str, REG_EXTENDED)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 300B: Could not "
			"compile regex %.200s\n", p_rule->regex_str);
		apply_res = FALSE;
		goto EXIT;
	}

	p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		p_node = ((osm_port_t *) p_item)->p_node;
		desc = strdup(p_node->print_desc);

		if (desc == NULL) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3010: "
				"Out of memory for apply_regex_group\n");
			apply_res = FALSE;
			goto EXIT;
		}
		res = regexec(regex, desc, 0, NULL, 0);
		if (res)
			if (res == REG_NOMATCH)
				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"No match for desc: %s\n", desc);
			else {
				regerror(res, regex, err_buf, sizeof(err_buf));
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
					"ERR 3005: Regex Match Failed for desc:"
					" %s, error: %s\n", desc, err_buf);
		} else {
			p_port = (osm_port_t *) p_item;
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Found match for desc: %s\n", desc);

			/* First handle the case that no ports specified */
			if (p_rule->min_port_num == 0) {
				if (set_port_in_grp(p_osm, p_port_group, p_port,
						    0, TRUE))
					goto NEXT;
				else {
					apply_res = FALSE;
					free(desc);
					goto EXIT;
				}
			}

			/* Handle the case that specific ports specified */
			if (!is_legal_port(p_port, p_rule->min_port_num)) {
				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"No matching ports for found node "
					"- Ignoring\n");
				goto NEXT;
			}
			if (is_legal_port(p_port, p_rule->max_port_num))
				max_port = p_rule->max_port_num;
			else {
				max_port = p_port->p_node->physp_tbl_size - 1;
				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"Max port in rule higher than max "
					"port of matched node (%u), "
					"overriding\n", max_port);
			}
			for (i = p_rule->min_port_num ; i <= max_port ; i++) {
				if (!set_port_in_grp(p_osm, p_port_group, p_port,
						     i, TRUE)) {
					apply_res = FALSE;
					free(desc);
					goto EXIT;
				}
			}
		}
NEXT:
		free(desc);
		p_item = cl_qmap_next(p_item);
	}

EXIT:
	regfree(regex);
	return apply_res;
}

boolean_t match_pattern(const char *pre, const char *str)
{
	size_t lenpre = strlen(pre), lenstr = strlen(str);

	return lenstr < lenpre ? FALSE : strncmp(pre, str, lenpre) == 0;
}

static boolean_t set_physp_in_grp(osm_opensm_t *p_osm, osm_port_group_t *p_grp,
				  osm_node_t *p_node, uint32_t port_num)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
	if (!p_physp)
		return TRUE;
	osm_port_t *p_port = osm_get_port_by_guid(&p_osm->subn,
						  p_physp->port_guid);
	return set_port_in_grp(p_osm, p_grp, p_port, port_num, TRUE);
}

static boolean_t apply_hostname_group(osm_opensm_t * p_osm,
				      osm_port_group_t * p_grp, char *hostname,
				      uint32_t port_num, uint8_t card_num)
{
	static const char *hca_str = " HCA-";
	cl_map_item_t *p_item;
	char pattern[IB_NODE_DESCRIPTION_SIZE + 1];
	osm_node_t *p_node;
	boolean_t res = TRUE;
	size_t matches = 0;
	uint32_t i;

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Apply hostname: name=%s, "
		"hca_idx=%d, port_num=%d\n", hostname, card_num, port_num);

	if (card_num)
		snprintf(pattern, sizeof(pattern), "%s%s%d", hostname, hca_str,
			 card_num);
	else
		snprintf(pattern, sizeof(pattern), "%s%s", hostname, hca_str);

	p_item = cl_qmap_head(&p_osm->subn.node_guid_tbl);
	while ((p_item != cl_qmap_end(&p_osm->subn.node_guid_tbl)) && res) {
		p_node = (osm_node_t *)p_item;
		p_item = cl_qmap_next(p_item);
		/* hostname is relevant for HCA's only */
		if (p_node->sw)
			continue;

		if (strcmp(p_node->print_desc, hostname) &&
		    !match_pattern(pattern, p_node->print_desc))
			continue;

		matches++;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
			"Found match for desc: %s\n", p_node->print_desc);
		if (port_num) {
			res = set_physp_in_grp(p_osm, p_grp, p_node,
					       port_num);
		} else {
			/* port num 0 mean all ports */
			for (i = 1 ; i < p_node->physp_tbl_size && res; i++)
				res = set_physp_in_grp(p_osm, p_grp, p_node, i);
		}
	}

	if (res && (matches > 1))
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Notice: More than one HCA"
			" which match the node description found\n");
	return res;
}

/*
 * Handle input of the following format:
 *
 * <guid> - Add all ports of guid to ports table
 * <guid>@<port number> - Add specified port to ports table
 * <guid>@<ports range> - Add specified range of ports to ports table
 * 			  Ports range is in the format: <min>-<max>
 */
static boolean_t process_guid_input(osm_opensm_t * p_osm, char *input,
				    osm_port_group_t * p_grp)
{
	char *token;
	char *context = NULL;
	char *ports_spec;
	char *hyphen;
	uint64_t guid;
	size_t guid_len;
	uint8_t min;
	uint8_t max;
	uint8_t i;
	size_t min_len;
	char guid_str[MAX_GUID_LEN];
	char min_str[MAX_PORT_LEN + 1];
	osm_port_t *p_port;
	osm_port_group_member_t *p_member;
	boolean_t new_member = FALSE;
	boolean_t res = TRUE;

	ports_spec = strchr(input, '@');
	if (ports_spec) {
		guid_len = ports_spec - input;
		if (guid_len >= MAX_GUID_LEN)
			guid_len = MAX_GUID_LEN -1;

		memset(guid_str, 0, MAX_GUID_LEN);
		strncpy(guid_str, input, guid_len);
		guid = cl_hton64(convert_guid_str(guid_str));
		p_port = osm_get_port_by_guid(&p_osm->subn, guid);
		if (p_port)
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "found port 0x%"
				PRIx64 " - Adding to group\n", cl_ntoh64(guid));

		else {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"port GUID 0x%" PRIx64 " NOT FOUND\n",
				cl_ntoh64(guid));
			goto EXIT;
		}
		p_member = find_member(p_grp, guid);
		if (!p_member) {
			new_member = TRUE;
			p_member = osm_create_port_grp_member(p_osm, p_port,
							      FALSE);
			if (!p_member) {
				res = FALSE;
				goto EXIT;
			}
		}
		token = strtok_r(ports_spec + 1, PORTS_DELIM, &context);
		while (token) {
			hyphen = strchr(token, '-');
			if (hyphen) {
				min_len = hyphen - token;
				if (min_len > MAX_PORT_LEN)
					min_len = MAX_PORT_LEN;

				memset(min_str, 0, MAX_PORT_LEN + 1);
				strncpy(min_str, token, min_len);
				min_str[MAX_PORT_LEN] = '\0';
				min = (uint8_t) atoi(min_str);
				max = (uint8_t) atoi(hyphen + 1);
				if ((min > 0) && (min < max)) {
					for (i = min; i <= max; i++)
						osm_set_physp_reliable(p_osm,
								   p_member, i,
								   TRUE);
				} else {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
						"ERR 300E: Illegal min/max, "
						"value (%d-%d)!!\n", min, max);
					res = FALSE;
					goto LOOP_END;
				}
			} else
				osm_set_physp_reliable(p_osm, p_member,
						   (uint8_t) atoi(token), TRUE);
			token = strtok_r(NULL, PORTS_DELIM, &context);
		}
LOOP_END:
		if (new_member) {
			if (has_physp(p_member) && res)
				/* New member so no need to check result */
				add_member_to_grp(p_osm, p_grp, p_member);
			else
				osm_destroy_port_grp_member(p_member);
		}
	} else {
		guid = cl_hton64(convert_guid_str(input));
		res = add_port_to_grp_by_guid(p_osm, p_grp, guid, 0);
	}

EXIT:
	return res;
}

static boolean_t process_hostname_input(osm_opensm_t * p_osm, char *input,
					osm_port_group_t * p_grp)
{
	char *token;
	char *context = NULL;
	char *sub_context = NULL;
	char *sub_token;
	char *param;
	uint32_t port_num;
	uint8_t card_num;
	char *hostname;
	char *dup_str;
	boolean_t res;
	boolean_t host_def = FALSE;
	boolean_t card_def = FALSE;
	boolean_t port_def = FALSE;
	char *trimed_param;
	char *trimed_val;

	port_num = 0;
	card_num = 0;
	hostname = "";
	token = strtok_r(input, PARAM_DELIM, &context);
	while (token) {
		sub_token = strtok_r(token, EQUAL, &sub_context);
		if (sub_token) {
			param = strdup(sub_token);
			trimed_param = trim_whitespace(param);
			sub_token = strtok_r(NULL, EQUAL, &sub_context);
			if (sub_token) {
				trimed_val = trim_whitespace(sub_token);
				if (!strcmp(trimed_param, HOSTNAME)) {
					if (host_def) {
						dup_str = HOSTNAME;
						goto DUPLICATE_ERR;
					} else {
						hostname = strdup(trimed_val);
						host_def = TRUE;
						free(param);
						continue;
					}
				}
				if (!strcmp(trimed_param, PORT)) {
					if (port_def) {
						dup_str = PORT;
						goto DUPLICATE_ERR;
					} else {
						port_num =
						    (uint32_t) atoi(trimed_val);
						port_def = TRUE;
						free(param);
						continue;
					}
				}
				if (!strcmp(trimed_param, HCA_ID)) {
					if (card_def) {
						dup_str = HCA_ID;
						goto DUPLICATE_ERR;
					} else {
						card_num =
						    (uint8_t) atoi(trimed_val);
						card_def = TRUE;
						free(param);
						continue;
					}
				}
				/* If we got here - Unknown Parameter */
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
					"ERR 3013: Unknown parameter in "
					"hostname rule [%.200s]!\n", param);
				free(param);
				if (host_def)
					free(hostname);
				return FALSE;
			}
			free(param);
		}
		token = strtok_r(NULL, PARAM_DELIM, &context);
	}
	if (host_def) {
		res = apply_hostname_group(p_osm, p_grp, hostname, port_num,
					   card_num);
		free(hostname);
		return res;
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3011: hostname not"
			" defined in hostname rule!\n");
		return FALSE;
	}

DUPLICATE_ERR:
	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3012: duplicate definition"
		" in hostname rule [%.200s]!\n", dup_str);
	free(param);
	if (host_def)
		free(hostname);
	return FALSE;
}

/* This method apply the port group rule, meaning analyze the 'main' ports table
 * and copy to the port group map the pointers to the ports which answer to the
 * rule of the port group
 */
static boolean_t apply_pgrp_rule(osm_opensm_t * p_osm, osm_port_group_t * p_grp,
				 osm_pgrp_rule_t *p_rule)
{
	char *token;
	char *context = NULL;
	osm_port_group_t *p_grp1;
	osm_port_group_t *p_grp2;
	osm_port_group_t *exist_g;
	boolean_t res = TRUE;

	switch (p_rule->rule_type) {
	case OSM_PORT_GROUP_GUID_RULE:
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying GUID"
			" list for port group %s [%.200s]\n",
			p_grp->name, p_rule->list);
		token = strtok(p_rule->list, LIST_DELIMITER);
		while (token) {
			if (!process_guid_input(p_osm, token, p_grp)) {
				res = FALSE;
				goto EXIT;
			}
			token = strtok(NULL, LIST_DELIMITER);
		}
		p_rule->applied = TRUE;
		break;
	case OSM_PORT_GROUP_GUID_RANGE_RULE:
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying GUID"
			" range for port group %s [0x%" PRIx64 "-0x%"
			PRIx64 "]\n", p_grp->name, p_rule->min_guid,
			p_rule->max_guid);
		if (!apply_range_rule(p_osm, p_grp, p_rule->min_guid,
				       p_rule->max_guid)) {
			res = FALSE;
			goto EXIT;
		}
		p_rule->applied = TRUE;
		break;
	case OSM_PORT_GROUP_GUID_NAME_RULE:
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying"
			" name for port group %s [%.200s]\n",
			p_grp->name, p_rule->hostname);
		token = strtok_r(p_rule->hostname, LIST_DELIMITER, &context);
		while (token) {
			if (!process_hostname_input(p_osm, token, p_grp)) {
				res = FALSE;
				goto EXIT;
			}
			token = strtok_r(NULL, LIST_DELIMITER, &context);
		}
		p_rule->applied = TRUE;
		break;

	case OSM_PORT_GROUP_GUID_REGEX_RULE:
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying GUID"
			" RegEx for port group %s [%.200s], min port num |%u|,"
			" max port num |%u|\n", p_grp->name, p_rule->regex_str,
			p_rule->min_port_num, p_rule->max_port_num);
		if (!apply_regex_rule(p_osm, p_grp, p_rule)) {
			res = FALSE;
			goto EXIT;
		}
		p_rule->applied = TRUE;
		break;

	case OSM_PORT_GROUP_GUID_UNION_RULE:
		p_grp1 = osm_get_port_grp(p_osm, p_rule->src_port_grp_1);
		p_grp2 = osm_get_port_grp(p_osm, p_rule->src_port_grp_2);
		/* If not both group exist this group is not valid */
		if (!p_grp1 && !p_grp2) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Port group %s not valid (both"
				" port groups it unites do not"
				" exist)\n", p_grp->name);
			res = FALSE;
			goto EXIT;
		}

		if (!p_grp1 || !p_grp2) {
			if (p_grp1)
				exist_g = p_grp1;
			else
				exist_g = p_grp2;
			if (exist_g->applied) {
				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"Applying GUID union '%s' and"
					" '%s' for port group '%s' (Only"
					" '%s' exist - copying it)\n",
					p_rule->src_port_grp_1,
					p_rule->src_port_grp_2,
					p_grp->name, exist_g->name);
				if (!copy_port_group_map(p_osm, exist_g, p_grp)) {
					res = FALSE;
					goto EXIT;
				}
				break;
			} else
				goto EXIT;
		}

		/* If not both group applied this group can't be applied yet */
		if (!p_grp1->applied || !p_grp2->applied) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Port group %s "
			"unifies groups that were not yet applied, will try "
			"later\n", p_grp->name);
			goto EXIT;
		}
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying GUID"
			" union for port group %s ['%s' and '%s']\n",
			p_grp->name, p_rule->src_port_grp_1,
			p_rule->src_port_grp_2);
		if (!unite_groups(p_osm, p_grp1, p_grp2, p_grp)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 300D: Union Failed, "
				"Removing Group!!\n");
			res = FALSE;
			goto EXIT;
		}
		p_rule->applied = TRUE;
		break;

	case OSM_PORT_GROUP_GUID_SUBTRACT_RULE:
		p_grp1 = osm_get_port_grp(p_osm, p_rule->src_port_grp_1);
		p_grp2 = osm_get_port_grp(p_osm, p_rule->src_port_grp_2);
		/* If not both group exist this group is not valid */
		if (!p_grp1) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Port group '%s' not valid (port group"
				" to subtract from do not exist)\n",
				p_grp->name);
			res = FALSE;
			goto EXIT;
		}

		if (!p_grp2) {
			if (p_grp1->applied) {
				OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
					"Applying subtract group %s, "
					"group '%s' not found, copying "
					"group '%s'\n", p_grp->name,
					p_rule->src_port_grp_2,
					p_rule->src_port_grp_1);
				if (!copy_port_group_map(p_osm, p_grp1, p_grp)) {
					res = FALSE;
					goto EXIT;
				}
				break;
			} else
				goto EXIT;
		}

		/* If not both group applied this group can't be applied yet */
		if (!p_grp1->applied || !p_grp2->applied)
			goto EXIT;
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Applying"
			" subtract for port group '%s' ['%s' - '%s']\n",
			p_grp->name, p_rule->src_port_grp_1,
			p_rule->src_port_grp_2);
		if (!subtract_groups(p_osm, p_grp1, p_grp2, p_grp)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 300C: Subtraction Failed, "
				"Removing Group!!\n");
			res = FALSE;
			goto EXIT;
		}
		p_rule->applied = TRUE;
		break;
	default:
		res = FALSE;
		goto EXIT;
	}

	if (osm_log_is_active_v2(&p_osm->log, OSM_LOG_DEBUG,
				 OSM_FILE_PORT_GRP_MGR_C))
		print_port_group(p_osm, p_grp);
EXIT:
	return res;
}

/*
 * Apply port group by iterating all rules which define the group and add to
 * the group ports table all ports which answer any of the rules
 *
 * Return value is number of rules in the group that were applied on this run
 */
static size_t apply_port_group(osm_opensm_t * p_osm, osm_port_group_t * p_grp,
			       boolean_t *valid_grp)
{
	cl_list_item_t *p_item;
	size_t applied_total = 0;
	size_t applied_now = 0;
	osm_pgrp_rule_t *p_rule;

	*valid_grp = TRUE;
	if (p_grp->applied)
		goto EXIT;

	p_item = cl_qlist_head(p_grp->rules);

	while (p_item != cl_qlist_end(p_grp->rules)) {
		p_rule = (osm_pgrp_rule_t*)p_item;
		if (p_rule->applied) {
			applied_total++;
			p_item = cl_qlist_next(p_item);
			continue;
		}
		if (apply_pgrp_rule(p_osm, p_grp, p_rule)) {
			if (p_rule->applied) {
				applied_now++;
				applied_total++;
			}
			p_item = cl_qlist_next(p_item);
		} else {
			*valid_grp = FALSE;
			goto EXIT;
		}
	}

	if (applied_total == cl_qlist_count(p_grp->rules))
		p_grp->applied = TRUE;
EXIT:
	return applied_now;
}

static boolean_t port_groups_mgr_apply(struct osm_opensm *p_osm)
{
	cl_list_item_t *p_item;
	cl_list_item_t *p_item_copy;
	size_t applied_groups, applied_rules;
	boolean_t error;
	osm_port_group_mgr_t *p_mgr = &p_osm->sm.port_group_mgr;
	boolean_t valid_grp = TRUE;
	osm_port_group_t *p_pgrp;

	applied_groups = 0;
	applied_rules = 0;
	error = FALSE;

	while (applied_groups < cl_qlist_count(&p_mgr->port_groups) && !error) {
		p_item = cl_qlist_head(&p_mgr->port_groups);
		applied_groups = 0;
		applied_rules = 0;
		while (p_item != cl_qlist_end(&p_mgr->port_groups)) {
			p_pgrp = (osm_port_group_t *)p_item;
			if (p_pgrp->applied) {
				applied_groups++;
				p_item = cl_qlist_next(p_item);
				continue;
			}
			applied_rules += apply_port_group(p_osm, p_pgrp,
							  &valid_grp);
			if (valid_grp) {
				if (p_pgrp->applied)
					applied_groups++;
				p_item = cl_qlist_next(p_item);
			} else {
				OSM_LOG(&p_osm->log, OSM_LOG_INFO,
					"NOTE: Group %s indicated as invalid,"
					" Removing Group\n", p_pgrp->name);
				p_item_copy = p_item;
				p_item = cl_qlist_next(p_item);
				cl_qlist_remove_item(&p_mgr->port_groups,
						     p_item_copy);
				osm_destroy_port_grp((osm_port_group_t *)
						     p_item_copy);
				error = TRUE;
			}
		}
		/* safety measurment to avoid endless loop, at each loop at
		 * least one rule shall be applied */
		if (!applied_rules)
			error = TRUE;
	}
	if (error) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3006: Failed to "
			"apply all port groups!!!\n");
		osm_port_grp_mgr_destroy(p_mgr);
		return FALSE;
	}
	return TRUE;
}

boolean_t osm_port_grp_mgr_create(IN struct osm_opensm * p_osm)
{
	osm_port_group_mgr_t *p_mgr = &p_osm->sm.port_group_mgr;

	osm_port_grp_mgr_destroy(p_mgr);

	if (parse_port_groups_policy(p_osm))
		return FALSE;

	if (!create_predefined_port_groups(p_osm, p_mgr)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 300F: Failed to "
			"create predefined port groups\n");
		osm_port_grp_mgr_destroy(p_mgr);
		return FALSE;
	}

	if (!port_groups_mgr_apply(p_osm))
		return FALSE;

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Port Group Manager Created (%d groups found)\n",
		cl_qlist_count(&p_mgr->port_groups));
	return TRUE;
}

osm_port_group_member_t *osm_pgrp_get_member(IN osm_port_group_t * p_group,
					     IN ib_net64_t port_guid)
{
	cl_map_item_t *p_member;

	p_member = cl_qmap_get(&p_group->ports, port_guid);

	if (p_member == cl_qmap_end(&p_group->ports))
		return NULL;
	else
		return (osm_port_group_member_t *) p_member;
}

struct osm_port *osm_pgrp_get_port(IN osm_port_group_t * p_group,
				   IN ib_net64_t port_guid)
{
	osm_port_group_member_t *p_member;

	p_member = osm_pgrp_get_member(p_group, port_guid);

	if (p_member)
		return p_member->port;
	else
		return NULL;
}

/* Implement API To Be Used By Parser */

osm_port_group_t*
osm_create_port_grp(IN struct osm_opensm *p_osm, IN char *name,
		    IN char *use, IN cl_qlist_t *rules, IN boolean_t keep)
{
	osm_port_group_t *new_group = NULL;
	size_t rules_cnt = 0;

	if (!name) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3015: Ilegal request "
			"to create port group without name\n");
		goto EXIT;
	}

	new_group = calloc(1, sizeof(osm_port_group_t));
	if (!new_group) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3001: Failed to "
			"allocate memory new port group: %s\n", name);
		goto EXIT;
	}

	cl_qmap_init(&new_group->ports);
	new_group->name = strdup(name);
	if (use)
		new_group->use = strdup(use);
	new_group->applied = FALSE;
	new_group->rules = rules;

	if (keep) {
		if (rules)
			rules_cnt = cl_qlist_count(rules);
		cl_qlist_insert_tail(&p_osm->sm.port_group_mgr.port_groups,
				     (cl_list_item_t *) new_group);
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "New Port Group Added: %s, "
			" %zd rules in group\n",
			new_group->name, rules_cnt);
	}
EXIT:
	return new_group;
}

/* API for other modules who would like to add ports to port groups */

boolean_t osm_set_port_in_grp(IN struct osm_subn *p_subn,
			      IN osm_port_group_t * p_grp,
			      IN struct osm_port *p_port, IN boolean_t val)
{
	return set_port_in_grp(p_subn->p_osm, p_grp, p_port, 0, val);
}

boolean_t osm_set_physp_in_grp_by_num(IN struct osm_subn * p_subn,
				      IN osm_port_group_t * p_grp,
				      IN struct osm_port * p_port,
				      IN uint8_t port_num, IN boolean_t val)
{
	return set_port_in_grp(p_subn->p_osm, p_grp, p_port, port_num, val);
}

boolean_t osm_set_physp_in_grp(IN struct osm_subn * p_subn,
			       IN osm_port_group_t * p_grp,
			       IN struct osm_physp * p_physp, IN boolean_t val)
{
	osm_port_t *p_port = osm_get_port_by_guid(p_subn, p_physp->port_guid);
	if (p_port)
		return set_port_in_grp(p_subn->p_osm, p_grp, p_port,
				       p_physp->port_num, val);
	else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 3008: "
			"port GUID 0x%" PRIx64 " NOT FOUND\n",
			cl_ntoh64(p_physp->port_guid));
		return FALSE;
	}
}
/*
 * Compare port_group_member ports as follows:
 * HCA ports under the same switch are sorted by remote port number
 * in ascending order
 * HCA ports under different switches are sorted by switch NodeGUID
 * in ascending order
 * Switch port0 are compare to the HCA remote switches according to
 * the switch NodeGUID in ascending order
 *
*/
static int compar_by_sw(const void *s1, const void *s2)
{
	osm_port_group_member_t **p1 = (osm_port_group_member_t **)s1;
	osm_port_group_member_t **p2 = (osm_port_group_member_t **)s2;
	uint64_t sw_guid1;
	uint64_t sw_guid2;
	uint8_t port1 = 0;
	uint8_t port2 = 0;
	osm_physp_t *p_physp_remote;

	if ((*p1)->port->p_node->sw)
		sw_guid1 = cl_ntoh64(osm_node_get_node_guid((*p1)->port->p_node));
	else {
		p_physp_remote = osm_physp_get_remote(osm_port_get_physp((*p1)->port));

		sw_guid1 = cl_ntoh64(osm_node_get_node_guid(
				osm_physp_get_node_ptr(p_physp_remote)));
		port1 = p_physp_remote->port_num;
	}

	if ((*p2)->port->p_node->sw)
		sw_guid2 = cl_ntoh64(osm_node_get_node_guid((*p2)->port->p_node));
	else {
		p_physp_remote = osm_physp_get_remote(osm_port_get_physp((*p2)->port));
		sw_guid2 = cl_ntoh64(osm_node_get_node_guid(
				osm_physp_get_node_ptr(p_physp_remote)));
		port2 = p_physp_remote->port_num;
	}

	if (sw_guid1 < sw_guid2) return -1;

	if (sw_guid1 > sw_guid2) return 1;

	if (port1 < port2) return -1;

	if (port1 > port2) return 1;

	return 0;

}

int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
				       void *cxt, char *p, void *sm, boolean_t go_to_node)
{
	osm_port_group_t *p_pgrp;
	cl_map_item_t *p_map_item;
	osm_port_group_member_t *p_pgrp_member;
	cl_qmap_t *p_ports_map;
	osm_port_group_member_t **pgrp_member_arr = NULL;
	int i, pgrp_member_arr_len;

	char *e;
	int ret = 0;
	osm_opensm_t *p_osm = (osm_opensm_t *)sm;
	uint64_t guid = strtoull(pgrp_name, &e, 0);

	if (*e == '\0' && e != pgrp_name)
		return pfn_func(cxt, guid, p);

	p_pgrp = osm_get_port_grp(p_osm, pgrp_name);
	if (!p_pgrp){
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 3014: "
			"port group: %s not found\n", pgrp_name);
		return 1;
	}
	p_ports_map = &p_pgrp->ports;

	/* Note that context can have any arbitrary value. */
	OSM_ASSERT(p_ports_map);
	OSM_ASSERT(p_ports_map->state == CL_INITIALIZED);
	OSM_ASSERT(pfn_func);

	pgrp_member_arr_len = cl_qmap_count(p_ports_map);
	pgrp_member_arr = calloc(1, sizeof(osm_port_group_member_t *) * pgrp_member_arr_len);

	if (!pgrp_member_arr)
		return 1;

	i = 0;

	p_map_item = cl_qmap_head(p_ports_map);
	while (p_map_item != cl_qmap_end(p_ports_map)) {
		p_pgrp_member = (osm_port_group_member_t *)p_map_item;
		pgrp_member_arr[i++] = p_pgrp_member;
		p_map_item = cl_qmap_next(p_map_item);
	}

	qsort(&pgrp_member_arr[0], pgrp_member_arr_len, sizeof(osm_port_group_member_t *), compar_by_sw);

	for (i = 0; i < pgrp_member_arr_len; i++) {
		p_pgrp_member = pgrp_member_arr[i];
		guid = go_to_node ? cl_ntoh64(osm_node_get_node_guid(p_pgrp_member->port->p_node)) :
			      cl_ntoh64(osm_port_get_guid(p_pgrp_member->port));
		ret = ret || pfn_func(cxt, guid, p);
	}
	free(pgrp_member_arr);
	return ret;
}
