/*
 * Copyright (c) 2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_qos_config_sl_rcv_t.
 * This object represents the QosConfig Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_QOS_CONFIG_SL_RCV_C
#include <opensm/osm_opensm.h>

#define VQOS_CONFIG_SL_INDEX_MASK               0xff

static uint16_t osm_qos_config_sl_get_vport_index(IN const ib_mad_t *p_mad) {
        return (uint16_t) (cl_ntoh32(p_mad->attr_mod) &
			   VQOS_CONFIG_SL_INDEX_MASK);
}

static boolean_t osm_qos_config_sl_is_vport(IN const ib_mad_t *p_mad) {
	return cl_ntoh32(p_mad->attr_mod) & VPORT_QOS_CONFIG_SL_ATTR_MOD_HO;
}

void osm_qos_config_sl_rcv_process(IN void *context, IN void *data) {
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	qos_config_sl_t *p_qos_config_sl;
	osm_qos_config_sl_context_t *p_context;
	ib_net64_t port_guid;
	boolean_t verbose_bypass = FALSE, is_vport = FALSE;
	uint16_t vport_index;
	osm_vport_t *p_vport;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_QOS_CONFIG_SL);

	p_context = osm_madw_get_qos_config_sl_context_ptr(p_madw);
	port_guid = p_context->port_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	verbose_bypass = osm_verbose_bypass_handle_qcsl(sm, p_port);

	is_vport = osm_qos_config_sl_is_vport(p_madw->p_mad);
	vport_index = osm_qos_config_sl_get_vport_index(p_madw->p_mad);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received QosConfigSL response "
		"for port GUID 0x%" PRIx64 ", TID 0x%" PRIx64
		", %s method, vport %s (idx = %u)\n",
		cl_ntoh64(port_guid), cl_ntoh64(p_smp->trans_id),
		p_context->set_method ? "SET" : "GET", is_vport ? "YES" : "NO",
		vport_index);

	if (PF(!p_port)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1602: "
			"No port object for port with GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid),
			cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	if (is_vport)
		osm_port_decrease_virt_transactions(sm, p_port);

	p_qos_config_sl = ib_smp_get_payload_ptr(p_smp);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1603: "
			"QosConfigSL MAD with status 0x%x received, "
			"for port GUID 0x%" PRIx64 "\n",
			cl_ntoh16(ib_smp_get_status(p_smp)),
			cl_ntoh64(port_guid));
		goto Exit;
	}

	/* remember what we got, in order not to send the same twice */
	if (is_vport) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1601: "
				"VPort with vport index %u of port 0x%" PRIx64
				" not found - Ignoring response\n",
				vport_index, cl_ntoh64(port_guid));
			goto Exit;
		}
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received QoS Config SL for "
			"vport 0x%" PRIx64 " (vport index %u of port 0x%"
			PRIx64 ")\n",
                        cl_ntoh64(osm_vport_get_guid(p_vport)), vport_index,
                        cl_ntoh64(port_guid));

		p_vport->qos_config_sl = *p_qos_config_sl;
	} else
		p_port->p_physp->qos_config_sl = *p_qos_config_sl;

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	osm_verbose_bypass_end(verbose_bypass, sm->p_log);

	OSM_LOG_EXIT(sm->p_log);
}
