%{
/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_RCH_PARSER_Y_Y
#include <opensm/osm_routing_chain.h>
#include <opensm/osm_port_group.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_topology_mgr.h>

#define YYSTYPE char*
extern FILE * __rch_parser_in;
static int in_step;
static size_t id_val;
static uint8_t path_bit_val;
static uint64_t topo_id_val;
static struct osm_subn  *p_subn_val;
static size_t fallback_to_val;
static char *use_val;
static char *config_val;
static char *engine_val;
static struct osm_opensm *p_osm;
static size_t lowest_eng_id;
static size_t highest_eng_id;
char *__rch_parser_yylval;
uint8_t engines;
uint8_t fallbacks;
static size_t used_topo_id[MAX_ENGINE_ID + 1];
static size_t t2q_eng_id;

char *trim(char *str)
{
	char *end;

	/* Trim leading space */
	while(isspace(*str)) str++;

	if(*str == 0)  // All spaces?
		return str;

	/* Trim trailing space */
	end = str + strlen(str) - 1;
	while(end > str && isspace(*end)) end--;

	/*  Write new null terminator */
	*(end+1) = 0;

	return str;
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void clear_step()
{
	clear_val(&use_val);
	clear_val(&config_val);
	clear_val(&engine_val);
	clear_val(&__rch_parser_yylval);
	fallback_to_val = 0;
	id_val = 0;
	path_bit_val = OSM_SUBNET_DEFAULT_PATH_BITS;
	p_subn_val = NULL;
	topo_id_val = 0;
}

static void print_step()
{
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Unicast step: Id=|%zd|,"
		"Engine=|%.200s|, Use=|%.200s|, Topo_id=|%" PRIu64 "|,"
		"Config=|%.200s|, Path_bit=|%d|, Fallback_to=|%zd|\n", id_val,
		engine_val, use_val, topo_id_val, config_val, path_bit_val,
		fallback_to_val);
}

static void inform_illegal_config(char *code, char *param, char *val, char* reason)
{
	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "RCH parser, ERR %.200s: illegal value "
		"param=|%.200s|, Value=|%.200s|, Reason=|%.200s| - Aborting Parser!\n",
		code, param, val, reason);
}

/* Cehck if first engine (lowest id) is valid */
static boolean_t first_engine_valid()
{
	boolean_t res = TRUE;
	uint8_t path_bit = 0;
	osm_routing_chain_t *p_engine = osm_rch_get_engine(p_osm, lowest_eng_id);

	if (!p_engine) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 901E: Engine with lowest"
			" id (%zd) not found - Aborting Parser!\n",
			lowest_eng_id);
		res = FALSE;
		goto EXIT;
	}

	if (p_engine->topo_id != MAIN_TOPO_ID) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9013: Illegal topology"
			" id (%" PRIu64 ") on first engine (%zd) - must use"
			" main topology - Aborting Parser!\n",
			p_engine->topo_id, lowest_eng_id);
		res = FALSE;
		goto EXIT;
	}

	path_bit = p_engine->path_bit;
	if ((path_bit != 0) && (path_bit != OSM_SUBNET_DEFAULT_PATH_BITS)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "RCH parser, ERR 900E: "
			"illegal value param=|path-bit|, Value=|%d|, Reason"
			"=|First engine must include path bit 0| "
			"- Aborting Parser!\n", path_bit);
		res = FALSE;
	}
EXIT:
	return res;
}

static boolean_t topo_id_legal()
{
	boolean_t result = TRUE;

	if (used_topo_id[topo_id_val]) {
		/* Only a fallback engine is allowed to define same topology
		 * as the engine it is fallback to
		 */
		if (used_topo_id[topo_id_val] != fallback_to_val) {
			inform_illegal_config("901B", "topology",
					      __rch_parser_yylval,
					     "topology id already used"
					     " in this RCH");
			result = FALSE;
		}
	}

	return result;
}

static boolean_t add_step()
{
	boolean_t res = FALSE;

	if (id_val > t2q_eng_id)
		goto EXIT;

	used_topo_id[topo_id_val] = id_val;

	if (id_val < lowest_eng_id)
		lowest_eng_id = id_val;

	if (id_val > highest_eng_id)
		highest_eng_id = id_val;

	if (fallback_to_val) {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Adding fallback engine:\n");
		print_step();
		fallbacks++;
		if (!osm_rch_set_fallback_engine(p_osm, fallback_to_val, id_val,
						use_val, engine_val,
						topo_id_val, p_subn_val,
						path_bit_val, config_val))
			goto EXIT;
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Adding engine:\n");
		print_step();
		engines++;
		if (!osm_rch_add_engine(p_osm, id_val, use_val, engine_val,
				       topo_id_val, p_subn_val, path_bit_val,
				       config_val))
			goto EXIT;
	}
	res = TRUE;
EXIT:
	return res;
}

static boolean_t legal_engine(char *engine)
{
	if (!strcmp(engine, "minhop"))
		return TRUE;
	if (!strcmp(engine, "ftree"))
		return TRUE;
	if (!strcmp(engine, "updn"))
		return TRUE;
	if (!strcmp(engine, "dor"))
		return TRUE;
	if (!strcmp(engine, "pqft"))
		return TRUE;
	if (!strcmp(engine, "sssp"))
		return TRUE;
	if (!strcmp(engine, "torus-2QoS")) {
		if (t2q_eng_id != MAX_ENGINE_ID) {
			inform_illegal_config("901F", "engine", engine,
					      "T2Q engine already defined in"
					      " rch");
			return FALSE;
		}
		t2q_eng_id = id_val;
		return TRUE;
	}
	if (!strcmp(engine, "dfp"))
		return TRUE;
	if (!strcmp(engine, "ar_dor"))
		return TRUE;
	if (!strcmp(engine, "ar_updn"))
		return TRUE;
	if (!strcmp(engine, "ar_ftree"))
		return TRUE;
	if (!strcmp(engine, "ar_torus"))
		return TRUE;

	inform_illegal_config("9007", "engine", engine, "Not legal engine");
	return FALSE;
}

static boolean_t legal_engines_list(char *list)
{
	char *token;
	char *context = NULL;
	char *list_copy = strdup(list);
	boolean_t res = TRUE;

	token = strtok_r(list_copy, ",", &context);
	while (token) {
		if (!legal_engine(trim(token))) {
			res = FALSE;
			break;
		}
		token = strtok_r(NULL, ",", &context);
	}
	free(list_copy);
	return res;
}

void __rch_parser_error(const char *str)
{
        OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "%.200s\n",str);
}

int __rch_parser_wrap()
{
        return 1;
}

int __rch_parser_parse();
int __rch_parser_lex();

int parse_rch_policy(struct osm_opensm *osm_ptr)
{
	int res;
	p_osm = osm_ptr;
	if (!p_osm->subn.opt.rch_policy_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 900F: RCH policy "
			"file not configured - Quit routing chains parser\n");
		return 1;
	}
	__rch_parser_in = fopen(p_osm->subn.opt.rch_policy_file, "r");
	if (!__rch_parser_in) {
		__rch_parser_error("ERR 9010: RCH Policy file specified but "
				    "not found\n");
		return 1;
	}
	t2q_eng_id = MAX_ENGINE_ID;
	engines = 0;
	lowest_eng_id = MAX_ENGINE_ID;
	highest_eng_id = 0;
	memset(used_topo_id, 0, MAX_ENGINES_NUMBER * sizeof(size_t));
	fallbacks = 0;

        res =  __rch_parser_parse();
	if (res)
		goto EXIT;

	if (in_step) {
		__rch_parser_error("ERR 9011: Policy file ended within routing"
				   " step\n");
		res = 1;
		goto EXIT;
	}

	if (!first_engine_valid()) {
		res = 1;
		goto EXIT;
	}

	if (t2q_eng_id < highest_eng_id) {
		__rch_parser_error("ERR 901C: T2Q engine "
				   "defined in RCH, T2Q "
				   "must be last engine");
		res = 1;
		goto EXIT;
	}
	if (!res)
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "RCH Parser detected %d"
			" engines and %d fallback engines\n", engines,
			fallbacks);
EXIT:
	if (__rch_parser_in)
	{
	        fclose(__rch_parser_in);
		__rch_parser_in = NULL;
	}

	clear_step();
	return res;
}

%}

%token COMMENT
%token STEP_START
%token STEP_END
%token TOPOLOGY
%token USE
%token WORD
%token CONFIG
%token ENGINE
%token PATH_BIT
%token FALL_BACK
%token ID

%%

rch_policy: steps;

steps: /* empty */
        |  steps step
        ;

comment: COMMENT
	{
	}
	;

step:
	start
	|id
	|use
	|word
	|topology
	|config
	|path_bit
	|fall_back
	|engine
	|end
	|comment
        ;

word: WORD
	{
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9000: Unexpected "
			"input: |%.200s|, Aborting rch parser!\n",
			__rch_parser_yylval);
		return 1;
	}
	;

use: USE WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9014: "
				"Parsed use qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		if (use_val) {
			inform_illegal_config("9001", "use", __rch_parser_yylval,
					     "use already defined");
			return 1;
		}
		use_val = strdup(__rch_parser_yylval);
	}
	;

id: ID WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9015: "
				"Parsed id qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		if (id_val) {
			inform_illegal_config("9002", "id", __rch_parser_yylval,
					     "id already defined");
			return 1;
		}
		int id = atoi(__rch_parser_yylval);
		if ((id <= 0) || (id > MAX_ENGINE_ID)) {
			inform_illegal_config("9003", "id", __rch_parser_yylval,
					     "not legal value");
			return 1;
		}
		id_val = (size_t)id;
	}
	;

topology: TOPOLOGY WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9016: "
				"Parsed topology qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}

		if (topo_id_val) {
			inform_illegal_config("901D", "topology",
					      __rch_parser_yylval,
					      "id already defined");
			return 1;
		}

		topo_id_val = (uint64_t)atoi(__rch_parser_yylval);
		if (topo_id_val < 0) {
			inform_illegal_config("900C", "topology",
					      __rch_parser_yylval,
					     "Negative id value");
			return 1;
		}
		if (topo_id_val == 0)
			p_subn_val = &p_osm->subn;
		else {
			osm_topology_t *p_topo = osm_topo_mgr_get_topo(p_osm,
								       topo_id_val);
			if (p_topo)
				p_subn_val = p_topo->p_subn;
			else {
				inform_illegal_config("900D", "topology",
						      __rch_parser_yylval,
						     "id not found by topology mgr");
				return 1;
			}
		}
	}
	;

config: CONFIG WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9017: "
				"Parsed config qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		if (config_val) {
			inform_illegal_config("9021", "config file",
					      __rch_parser_yylval,
					     "already defined");
			return 1;
		}
		FILE *f;
		f = fopen(__rch_parser_yylval, "r");
		if (!f) {
			inform_illegal_config("9004", "config",
					      __rch_parser_yylval,
					      "Can't open file");
			return 1;
		}
		fclose(f);
		config_val = strdup(__rch_parser_yylval);
	}
	;

path_bit: PATH_BIT WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9018: "
				"Parsed path-bit qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		if (!strlen(__rch_parser_yylval)) {
			inform_illegal_config("9020", "path_bit",
					      __rch_parser_yylval,
					     "Illegal Value");
			return 1;
		}

		char *endptr;
		uint8_t path_bit = strtoull(__rch_parser_yylval, &endptr, 0);

		if (*endptr!='\0') {
			inform_illegal_config("9020", "path_bit",
					      __rch_parser_yylval,
					     "Illegal Value");
			return 1;
		}

		unsigned lids_per_port = 1 << p_osm->subn.opt.lmc;

		if ((path_bit != OSM_SUBNET_DEFAULT_PATH_BITS) &&
		    ((path_bit < 0) || (path_bit >= lids_per_port))) {
			inform_illegal_config("9005", "path_bit",
					      __rch_parser_yylval,
					     "Out of legal range");
			return 1;
		}
		path_bit_val = path_bit;
	}
	;

fall_back: FALL_BACK WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9019: "
				"Parsed fallback-to qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		int fallback = atoi(__rch_parser_yylval);
		if (fallback <= 0) {
			inform_illegal_config("9006", "fallback-to",
					      __rch_parser_yylval,
					     "Not positive value");
			return 1;
		}
		fallback_to_val = (size_t)fallback;
	}
	;

engine: ENGINE WORD
	{
		if (!in_step) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 901A: "
				"Parsed engine qualifier outside a step "
				"(%.200s)\n", __rch_parser_yylval);
			return 1;
		}
		if (engine_val) {
			inform_illegal_config("9022", "engine",
					      __rch_parser_yylval,
					     "already defined");
			return 1;
		}
		engine_val = strdup(__rch_parser_yylval);
	}
	;

start:
        STEP_START
        {
		if (in_step) {
			__rch_parser_error("ERR 9008: Parsed step start "
					   "within a step");
			return 1;
		} else {
			clear_step();
			in_step = 1;
		}
        }
        ;

end:
        STEP_END
        {
		if (in_step) {
			in_step = 0;
			if (!engine_val) {
				__rch_parser_error("ERR 9009: No Engine "
						   "defined to step!");
				return 1;
			}
			if (!id_val) {
				__rch_parser_error("ERR 900A: No valid ID "
						   "defined to step!");
				return 1;
			}
			if (!legal_engines_list(engine_val)) {
				return 1;
			}
			if (!topo_id_legal()) {
				return 1;
			}
			if (!config_val)
				config_val = strdup(NONE_STR);
			if (!use_val)
				use_val = strdup(NONE_STR);
			if (!p_subn_val)
				p_subn_val = &p_osm->subn;
			if (!add_step())
				return 1;
		} else {
			__rch_parser_error("ERR 900B: Parsed step end while "
					   "not within a step");
			return 1;
		}
        }
        ;

