/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * 	This module responsible for processing received Routing Notification
 * 	MADs.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_RN_MAD_RCV_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>


void process_rn_sub_group_direction(osm_sm_t *sm, osm_switch_t *p_sw,
				    ib_rn_sub_group_direction_table_t *block,
				    uint16_t block_id)
{
	if (osm_ar_rn_sub_group_direction_set_block(&p_sw->ar_data, block_id,
						    block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7E03: Failed to store RNSubGroupDirectionTable"
			" response for switch 0x%" PRIx64 " (block id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			block_id);
	}
}

void process_rn_gen_string(osm_sm_t *sm, osm_switch_t *p_sw,
			   ib_rn_gen_string_table_t *block,
			   uint16_t block_id)
{
	if (osm_ar_rn_gen_string_set_block(&p_sw->ar_data, block_id, block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7E04: Failed to store RNGenStringTable"
			" response for switch 0x%" PRIx64 " (block id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			block_id);
	}
}

void process_rn_rcv_string(osm_sm_t *sm, osm_switch_t *p_sw,
			   ib_rn_rcv_string_table_t *block,
			   uint16_t block_id)
{
	if (osm_ar_rn_rcv_string_set_block(&p_sw->ar_data, block_id,
					   block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7E05: Failed to store RNRcvStringTable"
			" response for switch 0x%" PRIx64 " (block id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			block_id);
	}
}

void process_rn_xmit_port_mask(osm_sm_t *sm, osm_switch_t *p_sw,
			       ib_rn_xmit_port_mask_t *block,
			       uint16_t block_id)
{
	if (osm_ar_rn_xmit_port_mask_set_block(&p_sw->ar_data, block_id,
					       block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7E06: "
			"Failed to store RNXmitPortMask response for"
			" switch 0x%" PRIx64 " (block id %u)\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			block_id);
	}
}

void process_rn_gen_by_sub_group_priority(osm_sm_t *sm, osm_switch_t *p_sw,
					  ib_rn_gen_by_sub_group_priority_t *block)
{
	if (osm_ar_rn_subgroup_priority_flags_set_block(&p_sw->ar_data,
							block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7E07: Failed to store RNGenBySubGroupPriority"
			" response for switch 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
	}
}

void osm_rn_mad_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	osm_rn_mad_context_t *p_context;
	osm_node_t *p_node;
	ib_net64_t node_guid;
	uint16_t block_id;
	uint8_t *payload;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_rn_mad_context_ptr(p_madw);
	payload = ib_smp_get_payload_ptr(p_smp);

	node_guid = p_context->node_guid;
	block_id = p_context->block_id;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || !p_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7E01: Received Routing Notification MAD 0x%x(%s)"
			" for nonexistent switch node 0x%" PRIx64 "\n",
			cl_ntoh16(p_smp->attr_id),
			ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		p_node->sw->ar_configured = OSM_SW_AR_CONF_DISABLED;
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	/* Mark that the switch received the successful response */
	p_node->sw->ar_data.pending_mads--;

	switch (p_smp->attr_id) {
		case IB_MAD_ATTR_RN_SUB_GROUP_DIRECTION_TABLE:
			process_rn_sub_group_direction(sm, p_node->sw,
			    (ib_rn_sub_group_direction_table_t *)payload,
			    block_id);
			break;
		case IB_MAD_ATTR_RN_GEN_STRING_TABLE:
			process_rn_gen_string(sm, p_node->sw,
			    (ib_rn_gen_string_table_t *)payload,
			    block_id);
			break;
		case IB_MAD_ATTR_RN_RCV_STRING_TABLE:
			process_rn_rcv_string(sm, p_node->sw,
			    (ib_rn_rcv_string_table_t *)payload,
			    block_id);
			break;
		case IB_MAD_ATTR_RN_XMIT_PORT_MASK:
			process_rn_xmit_port_mask(sm, p_node->sw,
			    (ib_rn_xmit_port_mask_t *)payload,
			    block_id);
			break;
		case IB_MAD_ATTR_RN_GEN_BY_SUB_GROUP_PRIORITY:
			process_rn_gen_by_sub_group_priority(sm, p_node->sw,
			    (ib_rn_gen_by_sub_group_priority_t *)payload);
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7E02: Received unsupported MAD 0x%x(%s)"
				" for switch node 0x%" PRIx64 "\n",
				cl_ntoh16(p_smp->attr_id),
				ib_get_sm_attr_str(p_smp->attr_id),
				cl_ntoh64(node_guid));
	}

	CL_PLOCK_RELEASE(sm->p_lock);

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return;
}
