/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005,2015 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_router_t.
 * This object represents an Infiniband router.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_C
#include <opensm/osm_router.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

void osm_router_tbl_construct(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	cl_ptr_vector_construct(&p_router_tbl->blocks);
}

void osm_router_tbl_delete(IN osm_router_tbl_t * p_router_tbl)
{
	ib_rtr_adj_table_block_t *p_block;
	size_t num_blocks;
	int i;

	OSM_ASSERT(p_router_tbl);
	num_blocks = cl_ptr_vector_get_size(&p_router_tbl->blocks);
	if (num_blocks) {
		for (i = num_blocks - 1; i >= 0; i--) {
			if ((p_block =
				cl_ptr_vector_remove(&p_router_tbl->blocks, i)))
				free(p_block);
		}
	}
}

void osm_router_tbl_destroy(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	osm_router_tbl_delete(p_router_tbl);
	cl_ptr_vector_destroy(&p_router_tbl->blocks);
}

ib_api_status_t osm_router_tbl_init(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	if (cl_ptr_vector_init(&p_router_tbl->blocks, 0, 1) == CL_SUCCESS)
		return IB_SUCCESS;
	return IB_UNKNOWN_ERROR;
}

ib_api_status_t osm_adj_router_tbl_set(IN osm_router_tbl_t * p_router_tbl,
				       IN uint16_t block,
				       IN ib_rtr_adj_table_block_t * p_tbl)
{
	ib_rtr_adj_table_block_t *p_router_block;

	OSM_ASSERT(p_router_tbl);
	OSM_ASSERT(p_tbl);

	/* make sure the block is allocated */
	if (cl_ptr_vector_get_size(&p_router_tbl->blocks) > block)
		p_router_block = (ib_rtr_adj_table_block_t *)
				  cl_ptr_vector_get(&p_router_tbl->blocks,
						    block);
	else
		p_router_block = NULL;

	if (!p_router_block) {
		p_router_block = (ib_rtr_adj_table_block_t *)
			calloc(1, sizeof(ib_rtr_adj_table_block_t));
		if (!p_router_block)
			return IB_ERROR;
		cl_ptr_vector_set(&p_router_tbl->blocks, block, p_router_block);
	}

	/* sets the block values */
	memcpy(p_router_block, p_tbl, sizeof(ib_rtr_adj_table_block_t));


	return IB_SUCCESS;
}

ib_api_status_t osm_next_hop_router_tbl_set(IN osm_router_tbl_t *p_router_tbl,
					    IN uint16_t block_num,
					    IN ib_rtr_next_hop_table_block_t *p_block)
{
	ib_rtr_next_hop_table_block_t *p_router_block;

	OSM_ASSERT(p_router_tbl);
	OSM_ASSERT(p_block);

	/* make sure the block is allocated */
	if (cl_ptr_vector_get_size(&p_router_tbl->blocks) > block_num)
		p_router_block = (ib_rtr_next_hop_table_block_t *)
				  cl_ptr_vector_get(&p_router_tbl->blocks,
						    block_num);
	else
		p_router_block = NULL;

	if (!p_router_block) {
		p_router_block = (ib_rtr_next_hop_table_block_t *)
			calloc(1, sizeof(ib_rtr_next_hop_table_block_t));
		if (!p_router_block)
			return IB_ERROR;
		cl_ptr_vector_set(&p_router_tbl->blocks, block_num, p_router_block);
	}

	/* sets the block values */
	memcpy(p_router_block, p_block, sizeof(ib_rtr_next_hop_table_block_t));


	return IB_SUCCESS;
}

/********************** END OF ROUTER TABLE **********************************/
void osm_router_delete(IN OUT osm_router_t ** pp_rtr)
{
	cl_fmap_item_t *p_item, *p_next_item;
	osm_subnet_prefix_pkey_t *p_key;

	osm_router_tbl_destroy(&(*pp_rtr)->adj_router_tbl);
	osm_router_tbl_destroy(&(*pp_rtr)->next_hop_router_tbl);
	p_next_item = cl_fmap_head(&(*pp_rtr)->subnet_prefix_tbl);
	while (p_next_item != cl_fmap_end(&(*pp_rtr)->subnet_prefix_tbl)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&(*pp_rtr)->subnet_prefix_tbl, p_item);
		p_key = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		free(p_key);
		free(p_item);
	}
	free(*pp_rtr);
	*pp_rtr = NULL;
}

osm_router_t *osm_router_new(IN osm_port_t * p_port)
{
	osm_router_t *p_rtr;

	OSM_ASSERT(p_port);

	p_rtr = malloc(sizeof(*p_rtr));
	if (p_rtr) {
		memset(p_rtr, 0, sizeof(*p_rtr));
		p_rtr->p_port = p_port;
	}

	osm_router_tbl_construct(&p_rtr->adj_router_tbl);
	osm_router_tbl_init(&p_rtr->adj_router_tbl);
	osm_router_tbl_construct(&p_rtr->next_hop_router_tbl);
	osm_router_tbl_init(&p_rtr->next_hop_router_tbl);
	cl_fmap_init(&p_rtr->subnet_prefix_tbl, ib_compare_prefix_pkey);

	return p_rtr;
}

ib_api_status_t osm_router_find_pkey(
	IN osm_subn_t *p_subn,
	IN const osm_alias_guid_t *p_router_alias_guid,
	IN const ib_gid_t *p_remote_gid,
	IN const osm_pkey_tbl_t *pkey_tbl_src,
	IN const osm_pkey_tbl_t *pkey_tbl_dest,
	OUT ib_net16_t *pkey)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_map_item_t *p_item;
	cl_fmap_item_t *p_fitem;
	osm_router_t *p_rtr;
	osm_router_tbl_data_t *p_subn_pkey_data;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* find router by p_router_alias_guid */
	p_item = cl_qmap_get(&p_subn->rtr_guid_tbl,
			     p_router_alias_guid->base_port_guid);
	if (p_item == cl_qmap_end(&p_subn->rtr_guid_tbl))
	{
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 8000: No router found with guid 0x%016" PRIx64
			"\n", cl_ntoh64(p_router_alias_guid->base_port_guid));
		status = IB_NOT_FOUND;
		goto Exit;
	}
	p_rtr = (osm_router_t *)p_item;

	/*
	 * go over subnets of this router and take the first matching subnet
	 * p_remote_gid.subnet pair pkey.
	 */
	for (p_fitem = cl_fmap_head(&p_rtr->subnet_prefix_tbl);
	     p_fitem != cl_fmap_end(&p_rtr->subnet_prefix_tbl);
             p_fitem = cl_fmap_next(p_fitem)) {
		p_subn_pkey_data = (osm_router_tbl_data_t *)p_fitem;

		if (p_subn_pkey_data->subnet_prefix ==
			ib_gid_get_subnet_prefix(p_remote_gid)) {
			*pkey = p_subn_pkey_data->pkey;
			/*
			 * take the first pkey that both src and
			 * destination share
			 */
			if (osm_pkey_tbl_share_this_pkey(pkey_tbl_src,
							 pkey_tbl_dest,
							 *pkey,
							 p_subn->opt.
							     allow_both_pkeys))
				break;
		}
	}
	/* didn't find matching subnet in our data base */
	if (p_fitem == cl_fmap_end(&p_rtr->subnet_prefix_tbl))
		status = IB_NOT_FOUND;
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}
