/*
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_router_mgr_t.
 * This object implements the GUID manager object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_MGR_C
#include <opensm/osm_log.h>
#include <opensm/osm_router.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

/* Growth size for prefix to routers table */
#define GROWTH_SIZE 8

/* to insert dummy object into cl_map */
static int g_dummy_object = 1;

extern int wait_for_pending_transactions(osm_stats_t * stats);

static void osm_router_info_get(IN cl_map_item_t * p_object, IN void *context)
{
        ib_api_status_t status = IB_SUCCESS;
        osm_madw_context_t mad_context;
	osm_router_t *const p_router = (osm_router_t *) p_object;
	osm_sm_t *sm = context;
	osm_physp_t *p_physp = p_router->p_port->p_physp;

        OSM_LOG_ENTER(sm->p_log);

        memset(&mad_context, 0, sizeof(mad_context));
        mad_context.ri_context.port_guid = osm_physp_get_port_guid(p_physp);
        mad_context.ri_context.set_method = FALSE;
        mad_context.ri_context.light_sweep = FALSE;

        /* Request a RouterInfo attribute */
        status = osm_req_get(sm, osm_physp_get_dr_path_ptr(p_physp),
                             IB_MAD_ATTR_ROUTER_INFO, 0, TRUE, 0,
                             0, CL_DISP_MSGID_NONE, &mad_context);
        if (status != IB_SUCCESS)
                /* continue despite error */
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7701: "
                        "Failure initiating RouterInfo request (%s) "
			"in heavy sweep\n",
                        ib_get_err_str(status));
        OSM_LOG_EXIT(sm->p_log);
}

static ib_api_status_t osm_router_subnet_add(IN osm_router_t * p_rtr,
			   IN OUT osm_sm_t * sm,
			   IN osm_subnet_prefix_pkey_t *p_prefix_pkey)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_fmap_t *p_prefix_rtr_tbl;
	osm_router_prefix_rtrs_map_t *p_prefix_2_routers;
	osm_subnet_prefix_pkey_t *p_prefix_pkey_loc;
	ib_net64_t rtr_guid;

	OSM_LOG_ENTER(sm->p_log);
	p_prefix_rtr_tbl = &sm->p_subn->prefix_rtr_table;
	rtr_guid = osm_port_get_guid(p_rtr->p_port);

	p_prefix_2_routers = (osm_router_prefix_rtrs_map_t *)cl_fmap_get(
							     p_prefix_rtr_tbl,
							     p_prefix_pkey);
	if (p_prefix_2_routers !=
	    (osm_router_prefix_rtrs_map_t *)cl_fmap_end(p_prefix_rtr_tbl)) {
		/*
		 * the map holds routers guids as key,
		 * the value is not important
		 * */
		if (NULL == cl_map_insert(&p_prefix_2_routers->rtrs,
					  rtr_guid,
					  &g_dummy_object)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7702: Failed to add router 0x%016" PRIx64
				" to subnets prefixes table for"
				" subnet prefix 0x%016" PRIx64
				" and pkey 0x%04" PRIx16 ".\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Updating db for router 0x%016" PRIx64
			", subnet prefix 0x%016" PRIx64
			" with pkey 0x%04" PRIx16 " were found\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_prefix_pkey->subnet_prefix),
			cl_ntoh16(p_prefix_pkey->pkey));
	} else {
		/* need to create one */
		p_prefix_2_routers = (osm_router_prefix_rtrs_map_t *)
					calloc(1,
					sizeof(osm_router_prefix_rtrs_map_t));
		if (!p_prefix_2_routers) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7703: "
                                "Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 " could not allocate "
				"new entry in subnet prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			status = IB_INSUFFICIENT_MEMORY;
			goto Exit;
		}
		cl_map_construct(&p_prefix_2_routers->rtrs);
		if (CL_SUCCESS != cl_map_init(&p_prefix_2_routers->rtrs,
			GROWTH_SIZE)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7704: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 " failed to init new"
				" routers map\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			free(p_prefix_2_routers);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		if (NULL == cl_map_insert(&p_prefix_2_routers->rtrs, rtr_guid,
					  &g_dummy_object)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7705: Failed to add router 0x%016" PRIx64
				" to subnets prefixes table for"
				" subnet prefix 0x%016" PRIx64
				" and pkey 0x%04" PRIx16 ".\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			free(p_prefix_2_routers);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}

		p_prefix_pkey_loc = (osm_subnet_prefix_pkey_t *)
				      calloc(1,
				             sizeof(osm_subnet_prefix_pkey_t));
		if (!p_prefix_pkey_loc) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7706: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 ", could not allocate "
				"new key in subnets prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			free(p_prefix_2_routers);
			status = IB_INSUFFICIENT_MEMORY;
			goto Exit;
		}
		memcpy(p_prefix_pkey_loc, p_prefix_pkey,
			sizeof(osm_subnet_prefix_pkey_t));
		if (NULL == cl_fmap_insert(p_prefix_rtr_tbl, p_prefix_pkey_loc,
					   &p_prefix_2_routers->map_item)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 7707: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 ", failed to add "
				"new entry in subnets prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			free(p_prefix_2_routers);
			free(p_prefix_pkey_loc);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Updating db for router 0x%016" PRIx64
			", subnet prefix 0x%016" PRIx64
			" with pkey 0x%04" PRIx16 " weren't found\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_prefix_pkey->subnet_prefix),
			cl_ntoh16(p_prefix_pkey->pkey));
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

static boolean_t is_pkey_valid(osm_sm_t *sm, osm_router_t *p_router,
			       ib_net16_t pkey, uint32_t block_num,
			       uint32_t record_num)
{
	const osm_pkey_tbl_t *p_pkey_tbl;
	boolean_t is_valid = TRUE;
	ib_net64_t rtr_guid = osm_port_get_guid(p_router->p_port);
	ib_net16_t *found_pkey;


	p_pkey_tbl = osm_physp_get_pkey_tbl(
			osm_port_get_physp(p_router->p_port));

	/* check that we have full membership */
	if (!(pkey & IB_PKEY_TYPE_MASK)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 770D: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" pkey 0x%04" PRIx16 " with partial"
			" membership.\n",
			cl_ntoh64(rtr_guid),
			block_num, record_num, cl_ntoh16(pkey));
		is_valid = FALSE;
		goto Exit;
	}
	/* check that the same pkey is configured on port */
	found_pkey = osm_pkey_tbl_get_pkey(sm->p_log, pkey, p_pkey_tbl,
					   sm->p_subn->opt.allow_both_pkeys);
	if (!found_pkey) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 770E: router 0x%016" PRIx64
			", router table block %u, entry %u has invalid"
			" pkey 0x%04" PRIx16 " that is not configured"
			" on the port.\n",
			cl_ntoh64(rtr_guid),
			block_num, record_num, cl_ntoh16(pkey));
		is_valid = FALSE;
		goto Exit;
	}

	/* let's compare the pkeys' membership */
	if (*found_pkey != pkey) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 770F: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" pkey (0x%04" PRIx16 ") that has different"
			" membership than pkey configured on"
			" the port (0x%04" PRIx16 ").\n",
			cl_ntoh64(rtr_guid), block_num, record_num,
			cl_ntoh16(pkey), cl_ntoh16(*found_pkey));
		is_valid = FALSE;
	}
Exit:
	return is_valid;
}

static boolean_t osm_router_table_record_validate(osm_sm_t *sm,
				uint32_t block_num, uint32_t record_num,
				rtr_adj_site_local_subn_record_t *p_rt_record,
				osm_router_t *p_router)
{
	uint16_t sm_lid;
	boolean_t is_valid;

	OSM_LOG_ENTER(sm->p_log);
	OSM_ASSERT(sm);
	OSM_ASSERT(p_rt_record);
	OSM_ASSERT(p_router);

	is_valid = is_pkey_valid(sm, p_router, p_rt_record->pkey, block_num,
				 record_num);

	if (!is_valid)
		goto Exit;

	/* SM LID verification */
	sm_lid = cl_ntoh16(p_rt_record->master_sm_lid);
	if (sm_lid == 0 || sm_lid > IB_LID_MCAST_START_HO) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7710: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" sm lid 0x%04" PRIx16 ".\n",
			cl_ntoh64(osm_port_get_guid(p_router->p_port)),
			block_num, record_num, sm_lid);
		is_valid = FALSE;
		goto Exit;
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
	return is_valid;
}

static osm_subnet_prefix_pkey_t *create_entry_key(osm_sm_t *sm, ib_net16_t pkey,
						  ib_net64_t subnet_prefix,
						  ib_net64_t router_guid)
{
	osm_subnet_prefix_pkey_t *p_entry_key;

	p_entry_key = (osm_subnet_prefix_pkey_t *)calloc(1,
				sizeof(osm_subnet_prefix_pkey_t));
	if (!p_entry_key) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 7709: Failed to create key for router 0x%"
			PRIx64 " for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16"\n",
			cl_ntoh64(router_guid), cl_ntoh64(subnet_prefix), cl_ntoh16(pkey));
		goto Exit;
	}
	p_entry_key->subnet_prefix = subnet_prefix;
	p_entry_key->pkey = pkey;
Exit:
	return p_entry_key;
}

static osm_router_tbl_data_t *create_router_data(osm_sm_t * sm,
						 osm_subnet_prefix_pkey_t *p_entry_key,
						 boolean_t is_adjacent,
						 ib_net16_t master_sm_lid,
						 ib_net64_t rtr_guid)
{
	osm_router_tbl_data_t *p_rtr_data;

	p_rtr_data = (osm_router_tbl_data_t *)calloc(1,
			sizeof(osm_router_tbl_data_t));
	if (!p_rtr_data) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 770A: Failed to allocate"
			" data for router 0x%" PRIx64 "\n", cl_ntoh64(rtr_guid));
		goto Exit;
	}
	p_rtr_data->subnet_prefix = p_entry_key->subnet_prefix;
	p_rtr_data->pkey = p_entry_key->pkey;
	p_rtr_data->is_adjacent = is_adjacent;
	p_rtr_data->master_sm_lid = master_sm_lid;

Exit:
	return p_rtr_data;
}

static void add_router_to_db(osm_sm_t *sm, osm_router_t *p_router,
			     ib_net16_t pkey, ib_net64_t subnet_prefix,
			     uint32_t block_num, uint32_t record_num,
			     boolean_t adjacent, ib_net16_t master_sm_lid)
{
	osm_subnet_prefix_pkey_t *p_entry_key;
	osm_router_tbl_data_t *p_rtr_data;
	ib_net64_t rtr_guid;
	cl_fmap_t *p_router_prefixes;
	cl_fmap_item_t *p_check_item;

	OSM_LOG_ENTER(sm->p_log);

	p_router_prefixes = &p_router->subnet_prefix_tbl;
	rtr_guid = osm_port_get_guid(p_router->p_port);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Router 0x%016" PRIx64
		": adding subnet 0x%016" PRIx64 " pkey 0x%04" PRIx16
		"\n", cl_ntoh64(rtr_guid), cl_ntoh64(subnet_prefix), cl_ntoh16(pkey));

	p_entry_key = create_entry_key(sm, pkey, subnet_prefix, rtr_guid);

	if (!p_entry_key)
		goto Exit;

	if (adjacent)
		p_rtr_data = create_router_data(sm, p_entry_key, TRUE,
						master_sm_lid, rtr_guid);
	else
		p_rtr_data = create_router_data(sm, p_entry_key, FALSE, 0,
						rtr_guid);

	if (!p_rtr_data) {
		free(p_entry_key);
		goto Exit;
	}

	p_check_item = cl_fmap_insert(p_router_prefixes,
				      p_entry_key,
				      &p_rtr_data->map_item);
	/*
	 * TODO: use <subnet, pkey> in p_rtr_data, as a key for insert,
	 * so won't need to allocate/free separate key.
	 * */
	if (p_check_item != &p_rtr_data->map_item) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 770B: router 0x%016" PRIx64
			", duplicate entry in block %u, "
			"entry %u for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16 ".\n",
			cl_ntoh64(rtr_guid), block_num, record_num,
			cl_ntoh64(p_entry_key->subnet_prefix),
			cl_ntoh16(p_entry_key->pkey));
		free(p_entry_key);
		free(p_rtr_data);
		goto Exit;
	}

	/* update subnets to routers db */
	if (osm_router_subnet_add(p_router, sm, p_entry_key) != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 770C: router 0x%016" PRIx64
			", failed to update subnets prefixes table "
			"for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16 ".\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_entry_key->subnet_prefix),
			cl_ntoh16(p_entry_key->pkey));
		/* update failed, revert changes and free memory */
		cl_fmap_remove(p_router_prefixes, p_entry_key);
		free(p_entry_key);
		free(p_rtr_data);
	}
Exit:
	OSM_LOG_EXIT(sm->p_log);
}


static void osm_router_mgr_update_db(IN OUT osm_router_t * p_rtr, IN osm_sm_t * sm)
{
	uint32_t i, num_entries, idx = 0, block_num = 0;
	ib_rtr_adj_table_block_t *p_adj_block = NULL;
	ib_rtr_next_hop_table_block_t *p_nh_block = NULL;
	ib_net64_t subnet_prefix, rtr_guid;

	OSM_LOG_ENTER(sm->p_log);
	OSM_ASSERT(p_rtr);

	rtr_guid = osm_port_get_guid(p_rtr->p_port);
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Updating router 0x%016" PRIx64	" in db.\n",
		cl_ntoh64(rtr_guid));

	/* Update adjacent routers */
	num_entries = p_rtr->router_info.adjacent_site_local_subnets_tbl_top;

	/* go over router table blocks and build db */
	for (i = 0; i < num_entries; i++) {
		/* get the next block, initialized already to first */
		if ((i % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK) == 0) {
			block_num = i / IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;
			if (block_num >=
			    osm_router_tbl_get_num_blocks(&p_rtr->adj_router_tbl)) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 7708: Router 0x%016" PRIx64
					" adjacent block number %u in "
					"subnets prefixes table is not found, "
					"unconsistent to table top value %u",
					cl_ntoh64(rtr_guid), block_num, num_entries);
				break;
			}

			p_adj_block = osm_router_tbl_adj_block_get(&p_rtr->adj_router_tbl,
								   block_num);
		}
		/* skip NULL blocks */
		if (!p_adj_block) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Router 0x%016" PRIx64 " encountered NULL "
				"adjacent router table block, block number %u.\n",
				cl_ntoh64(rtr_guid), block_num);
			/* promote i to first element on the next block */
			i += IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;
			continue;
		}

		idx = i % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;

		/* skip adding invalid record to db */
		if (!osm_router_table_record_validate(sm, block_num, idx,
						 &p_adj_block->record[idx], p_rtr))
			continue;

		/* use site local prefix */
		subnet_prefix = IB_SITE_LOCAL_SUBNET_PREFIX_HO;
		/* complete received 16 bits to site local subnet prefix */
		subnet_prefix |= cl_ntoh16(p_adj_block->record[idx].subnet_prefix);

		add_router_to_db(sm, p_rtr, p_adj_block->record[idx].pkey,
				 cl_hton64(subnet_prefix), block_num, idx, TRUE,
				 p_adj_block->record[idx].master_sm_lid);

	}

	/* Update next hop routers */
	num_entries = cl_ntoh32(p_rtr->router_info.next_hop_table_top);

	/* go over router table blocks and build db */
	for (i = 0; i < num_entries; i++) {
		/* get the next block, initialized already to first */
		if ((i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK) == 0) {
			block_num = i / IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
			if (block_num >=
			    osm_router_tbl_get_num_blocks(&p_rtr->next_hop_router_tbl)) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 7711: Router 0x%016" PRIx64
					" next hop block number %u in "
					"subnets prefixes table is not found, "
					"unconsistent to table top value %u",
					cl_ntoh64(rtr_guid), block_num, num_entries);
				break;
			}
			p_nh_block = osm_router_tbl_nh_block_get(&p_rtr->next_hop_router_tbl,
								 block_num);
		}
		/* skip NULL blocks */
		if (!p_nh_block) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Router 0x%016" PRIx64 " encountered NULL "
				"next hop router table block, block number %u.\n",
				cl_ntoh64(rtr_guid), block_num);
			/* promote i to first element on the next block */
			i += IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
			continue;
		}

		idx = i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;

		/* skip adding invalid record to db */
		if (!is_pkey_valid(sm, p_rtr, p_nh_block->record[idx].pkey,
				   block_num, idx))
			continue;

		add_router_to_db(sm, p_rtr, p_nh_block->record[idx].pkey,
				 p_nh_block->record[idx].subnet_prefix, block_num,
				 idx, FALSE, 0);

	}
	OSM_LOG_EXIT(sm->p_log);
}

static void clear_subnet_prefix_tbl(cl_fmap_t *p_prefix_map)
{
	osm_subnet_prefix_pkey_t *p_key;
	cl_fmap_item_t *p_item;

	p_item = cl_fmap_head(p_prefix_map);

	while (p_item != cl_fmap_end(p_prefix_map))
	{
		p_key = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		cl_fmap_remove_item(p_prefix_map, p_item);
		free(p_item);
		free(p_key);
		p_item = cl_fmap_head(p_prefix_map);
	}

}
void osm_router_mgr_destroy_db(IN osm_sm_t * sm)
{
	cl_map_item_t *p_map_item;
	osm_router_t *p_router;

	osm_subn_destroy_prefix_router_db(sm->p_subn);

	/*
	 * remove the data base content
         * without removing routing table blocks
         */
	for (p_map_item = cl_qmap_head(&sm->p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&sm->p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;
		clear_subnet_prefix_tbl(&p_router->subnet_prefix_tbl);
	}
}

void osm_router_mgr_process(IN osm_sm_t * sm)
{
	cl_map_item_t *p_map_item;
	osm_router_t  *p_router;

	OSM_LOG_ENTER(sm->p_log);

	/* nothing to do */
	if (cl_is_qmap_empty(&sm->p_subn->rtr_guid_tbl)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Router guid list is empty\n");
		return;
	}

	cl_qmap_apply_func(&sm->p_subn->rtr_guid_tbl, osm_router_info_get, sm);

	/* barrier */
	if (wait_for_pending_transactions(&sm->p_subn->p_osm->stats)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Failed on waiting for transactions.\n");
	}

	/* clean previous data base */
	osm_router_mgr_destroy_db(sm);

	/* go over routers and update the FIB*/
	for (p_map_item = cl_qmap_head(&sm->p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&sm->p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;
		osm_router_mgr_update_db(p_router, sm);
	}

	OSM_LOG_EXIT(sm->p_log);
}
