/*
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * This file content handles receiving of next hop router table block
 * This object represents the RouterTable Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_NH_TABLE_RCV_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_router.h>

/*
 * WE ONLY RECEIVE GET only responses
 */
void osm_next_hop_rtr_table_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_net64_t port_guid;
	osm_ri_context_t *p_context;
	osm_router_t *p_router;
	uint32_t block_num, top;
	ib_rtr_next_hop_table_block_t *p_rt;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);
	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_ri_context_ptr(p_madw);
	port_guid = p_context->port_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	/* the router should exist in router table from NI rcv flow */
	p_router = (osm_router_t *)cl_qmap_get(&sm->p_subn->rtr_guid_tbl,
					       port_guid);

	if (PF(p_router ==
		(osm_router_t *)cl_qmap_end(&sm->p_subn->rtr_guid_tbl))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6C00: "
			"Failed to find router object for port 0x%016"
			PRIx64 " in router next hop table\n",
                        cl_ntoh64(port_guid));
		goto Exit;
	}

	block_num = cl_ntoh32(p_smp->attr_mod);
	top = cl_ntoh32(p_router->router_info.next_hop_table_top);
	if (PF(block_num > ((top + IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK - 1) /
			     IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3C01: "
			"Got invalid block_num (%u) for next hop router top "
			"(%u), port 0x%" PRIx64 "\n",
			block_num, top, cl_ntoh64(port_guid));
		goto Exit;
	}

	p_rt = ib_smp_get_payload_ptr(p_smp);

	osm_dump_next_hop_router_block(sm->p_log, port_guid, block_num, p_rt,
				       FILE_ID, OSM_LOG_DEBUG);

	osm_next_hop_router_tbl_set(&p_router->next_hop_router_tbl, block_num,
				    p_rt);
Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
