%{
/*
 * Copyright (c) 2016, Bay Microsystems, Inc All Rights Reserved
 * Copyright (c) 2016 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
/*
 * Abstract:
 *    Lexer of OSM Route parser.
 *
 * Environment:
 *    Linux User Mode
 *
 * Author:
 *    Manickavel Subramani, Bay Microsystems Inc
 */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_PARSER_L_L
#include <opensm/osm_router_policy.h>
#include "osm_router_parser_y.h"

#define HANDLE_IF_IN_DESCRIPTION   if (inr_description) { __route_parser_yylval = strdup(yytext); return TK_TEXT; }

#define SAVE_POS save_pos()
static void save_pos();

extern int column_num;
extern int line_num;

extern FILE * __route_parser_in;
extern char *__route_parser_yylval;


boolean_t inr_description = FALSE;
boolean_t inr_list_of_hex_num_ranges = FALSE;
boolean_t inr_list_of_numbers = FALSE;
boolean_t inr_list_of_strings = FALSE;
boolean_t inr_list_of_num_pairs = FALSE;
boolean_t inr_asterisk_or_list_of_numbers = FALSE;
boolean_t inr_list_of_num_ranges = FALSE;
boolean_t inr_single_string = FALSE;
boolean_t inr_single_number = FALSE;
boolean_t inr_asterisk_or_list_of_hex_num_ranges = FALSE;

static void reset_new_line_flags();
#define RESET_NEW_LINE_FLAGS reset_new_line_flags()

#define START_NAME		{inr_single_string = TRUE;}   /* single string w/o whitespaces (port group name) */
#define START_SUBNET_PREFIX	{inr_asterisk_or_list_of_hex_num_ranges = TRUE;} /* comma-separated list of hex num ranges */
#define START_ROUTER_GUID	{inr_single_number = TRUE;}  /* single number */
#define START_ROUTE_WEIGHT	{inr_single_number = TRUE;}   /* single number */

#define START_LOAD_SHARE_SUBNET_PREFIX  {inr_single_number = TRUE;} /* single hex nummber  */
#define START_LOAD_SHARE_GUID           {inr_list_of_hex_num_ranges = TRUE;}   /* List */

%}

%option nounput noinput

ROUTE_RULES_START	router\-rules
ROUTE_RULES_END		end\-router\-rules
ROUTE_RULE_START	router\-rule
ROUTE_RULE_END		end\-router\-rule
NAME			name
SUBNET_PREFIX		subnet\-prefix
ROUTER_GUID		router\-guid
ROUTE_WEIGHT		weight

LOAD_SHARE_RULES_START       load\-share\-rules
LOAD_SHARE_RULES_END         end\-load\-share\-rules
LOAD_SHARE_RULE_START        load\-share\-rule
LOAD_SHARE_RULE_END          end\-load\-share\-rule

WHITE			[ \t]+
NEW_LINE		\n
COMMENT			\#.*\n
WHITE_DOTDOT_WHITE	[ \t]*:[ \t]*
WHITE_COMMA_WHITE	[ \t]*,[ \t]*
QUOTED_TEXT		\"[^\"]*\"
NUMBER			[0-9]+
HEXNUMBER		0[xX][0-9a-fA-F]+		
UNDERSCORE		_
%%


{COMMENT}			{ SAVE_POS; RESET_NEW_LINE_FLAGS; } /* swallow comment */
{WHITE}{NEW_LINE}		{ SAVE_POS; RESET_NEW_LINE_FLAGS; } /* trailing blanks with new line */
{WHITE}				{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; }
{NEW_LINE}			{ SAVE_POS; RESET_NEW_LINE_FLAGS; }
{HEXNUMBER}{WHITE}{NUMBER}	{ SAVE_POS; if(inr_list_of_hex_num_ranges) return TK_WHITE; HANDLE_IF_IN_DESCRIPTION; }
{UNDERSCORE}			{	SAVE_POS;
					if(inr_list_of_hex_num_ranges || inr_asterisk_or_list_of_hex_num_ranges)
						yyterminate();
				}	


{ROUTE_RULES_START}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_ROUTE_RULES_START; }
{ROUTE_RULES_END}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_ROUTE_RULES_END; }
{ROUTE_RULE_START}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_ROUTE_RULE_START; }
{ROUTE_RULE_END}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_ROUTE_RULE_END; }

{NAME}{WHITE_DOTDOT_WHITE}		{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; START_NAME;	return TK_NAME;}
{SUBNET_PREFIX}{WHITE_DOTDOT_WHITE}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; START_SUBNET_PREFIX;	return TK_SUBNET_PREFIX;}
{ROUTER_GUID}{WHITE_DOTDOT_WHITE}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; START_ROUTER_GUID; START_LOAD_SHARE_GUID;	return TK_ROUTER_GUID;}
{ROUTE_WEIGHT}{WHITE_DOTDOT_WHITE}	{ SAVE_POS; HANDLE_IF_IN_DESCRIPTION; START_ROUTE_WEIGHT;	return TK_ROUTE_WEIGHT;}

{LOAD_SHARE_RULES_START}        { SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_LOAD_SHARE_RULES_START; }
{LOAD_SHARE_RULES_END}          { SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_LOAD_SHARE_RULES_END; }
{LOAD_SHARE_RULE_START}         { SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_LOAD_SHARE_RULE_START; }
{LOAD_SHARE_RULE_END}           { SAVE_POS; HANDLE_IF_IN_DESCRIPTION; return TK_LOAD_SHARE_RULE_END; }

0[xX][0-9a-fA-F]+	{
				SAVE_POS;
				__route_parser_yylval = strdup(yytext);
				if (inr_description || inr_list_of_strings || inr_single_string)
					return TK_TEXT;
				if (inr_asterisk_or_list_of_hex_num_ranges || inr_list_of_hex_num_ranges){
					if (yyleng >18)
						yyterminate();
				}
						
				return TK_NUMBER;
			}

[0-9]+		{
			SAVE_POS;
			__route_parser_yylval = strdup(yytext);
			if (inr_description || inr_list_of_strings || inr_single_string)
				return TK_TEXT;
			return TK_NUMBER;
		}


-		{
			SAVE_POS;
			if (inr_description || inr_list_of_strings || inr_single_string)
			{
				__route_parser_yylval = strdup(yytext);
				return TK_TEXT;
			}
			return TK_DASH;
		}

:		{
			SAVE_POS;
			if (inr_description || inr_list_of_strings || inr_single_string)
			{
				__route_parser_yylval = strdup(yytext);
				return TK_TEXT;
			}
			return TK_DOTDOT;
		}

,		{
			SAVE_POS;
			if (inr_description)
			{
				__route_parser_yylval = strdup(yytext);
				return TK_TEXT;
			}
			return TK_COMMA;
		}

\*		{
			SAVE_POS;
			if (inr_description || inr_list_of_strings || inr_single_string)
			{
				__route_parser_yylval = strdup(yytext);
				return TK_TEXT;
			}
			return TK_ASTERISK;
		}

{QUOTED_TEXT}	{
			SAVE_POS;
			__route_parser_yylval = strdup(&yytext[1]);
			__route_parser_yylval[strlen(__route_parser_yylval)-1] = '\0';
			return TK_TEXT;
		}

.		{ SAVE_POS; __route_parser_yylval = strdup(yytext); return TK_TEXT;}

<<EOF>>		{
			YY_NEW_FILE;
			yyterminate();
		}
%%


/*********************************************
 *********************************************/

static void save_pos()
{
	int i;
	for (i = 0; i < yyleng; i++)
	{
		if (yytext[i] == '\n') {
			line_num ++;
			column_num = 1;
		}
		else
			column_num ++;
	}
}

/*********************************************
 *********************************************/

static void reset_new_line_flags()
{
	inr_description = FALSE;
	inr_list_of_hex_num_ranges = FALSE;
	inr_list_of_numbers = FALSE;
	inr_list_of_strings = FALSE;
	inr_list_of_num_pairs = FALSE;
	inr_asterisk_or_list_of_numbers = FALSE;
	inr_list_of_num_ranges = FALSE;
	inr_single_string = FALSE;
	inr_single_number = FALSE;
	inr_asterisk_or_list_of_hex_num_ranges = FALSE;
}
