%{
/*
 * Copyright (c) 2016, Bay Microsystems, Inc All Rights Reserved
 * Copyright (c) 2016 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
/*
 * Abstract:
 *    Grammar of OSM route parser.
 *
 * Environment:
 *    Linux User Mode
 *
 * Author:
 *    Manickavel Subramani, Bay Microsystems
 */

#include <stdio.h>
#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_PARSER_Y_Y
#include <opensm/osm_opensm.h>
#include <opensm/osm_router_policy.h>

#define OSM_ROUTE_POLICY_MAX_LINE_LEN         1024*10

typedef struct tmp_parser_struct_t_ {
	char       str[OSM_ROUTE_POLICY_MAX_LINE_LEN];
	uint64_t   num_pair[2];
	cl_list_t  str_list;
	cl_list_t  num_list;
	cl_list_t  num_pair_list;
} tmp_parser_struct_t;

static void __parser_tmp_struct_init();
static void __parser_tmp_struct_reset();
static void __parser_tmp_struct_destroy();

static char * __parser_strip_white(char * str);
static int __parser_str2uint64(uint64_t * p_val, char * str);
static int __validate_uint64bit_string(char *str);

static int __parser_route_rules_end();
static int __parser_route_rule_end();
static void __parser_add_prefix_range_to_prefix_list(
	uint64_t ** range_arr,
	unsigned range_len);

static int __rangelist2rangearr(
	cl_list_t    * p_list,
	uint64_t  ** * p_arr,
	unsigned     * p_arr_len);
static int __validate_route_rule();

static int __parser_load_share_rule_end();
static int __loadguidtomap(cl_list_t * p_list);
static int __validate_load_share_rule();

static void yyerror(const char *format, ...);
extern FILE * __route_parser_in;
char *__route_parser_yylval = NULL;
int __route_parser_parse();
int __route_parser_lex();

extern int errno;

#define RESET_BUFFER  __parser_tmp_struct_reset()
#define INVALID_GUID (0xFFFFFFFFFFFFFFFFULL)

tmp_parser_struct_t tmp_parser_struct;

int column_num;
int line_num;

int sn_count;
int guid_count;
int weight_count;
int default_count;


osm_router_policy_t	* p_router_policy;
osm_route_rule_t	* p_current_route_rule = NULL;
osm_load_share_rule_t	* p_load_share_rule = NULL;
osm_log_t		* p_route_parser_osm_log;

/***************************************************/

%}

%token TK_NUMBER
%token TK_DASH
%token TK_DOTDOT
%token TK_COMMA
%token TK_ASTERISK
%token TK_TEXT
%token TK_WHITE


%token TK_ROUTE_RULES_START
%token TK_ROUTE_RULES_END
%token TK_ROUTE_RULE_START
%token TK_ROUTE_RULE_END

%token TK_NAME
%token TK_SUBNET_PREFIX
%token TK_ROUTER_GUID
%token TK_ROUTE_WEIGHT

%token TK_LOAD_SHARE_RULES_START
%token TK_LOAD_SHARE_RULES_END
%token TK_LOAD_SHARE_RULE_START
%token TK_LOAD_SHARE_RULE_END

%token TK_LOAD_SHARE_SUBNET_PREFIX
%token TK_LOAD_SHARE_GUID

%%

head:	route_policy_entries
	;

route_policy_entries: 	/* empty */
			| route_policy_entries route_policy_entry
			;

route_policy_entry:	route_rules_section
			| load_share_rules_section
			;

	/*
	 * Parsing route rules:
	 * -------------------
	 *  route-rules
	 *       route-rule
	 *          name: dynamic/static
	 *          guid: 0x1000000000000001,0x1000000000000002
	 *          ...
	 *      end-route-rule
	 *  end-route-rules
	 */

route_rules_section:	route_rules_start route_rules route_rules_end
			;

route_rules_start:	TK_ROUTE_RULES_START {
				default_count = 0;
			}
			;	

route_rules_end:	TK_ROUTE_RULES_END {	
			if( __parser_route_rules_end() )
					return 1;
			}
			;	
			
route_rules:	route_rule
		| route_rules route_rule
		;

route_rule:	route_rule_start route_rule_entries route_rule_end
		;

route_rule_start:TK_ROUTE_RULE_START {
			if (p_current_route_rule) {
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
					"ERR 9F00: Current rule already exist "
					"on rule start\n");
					return 1;
			}
			p_current_route_rule =
			(osm_route_rule_t *) calloc(1, sizeof(osm_route_rule_t));
			sn_count = 0;
			guid_count = 0;
			weight_count = 0;
		}
		;

route_rule_end:	TK_ROUTE_RULE_END {
			if ( __parser_route_rule_end() )
				return 1;
		}
		;

route_rule_entries:
		| route_rule_entries route_rule_entry
		;

route_rule_entry:	route_rule_name
			| route_rule_subnet_prefix
			| route_rule_router_guid
			| route_rule_weight
			;

	/*
	 *  route_rule_entry values:
	 *      route_rule_name
	 *      route_rule_subnet_prefix
	 *      route_rule_router_guid
	 *      route_rule_weight
	 */

route_rule_name:	route_rule_name_start single_string {
			/* 'name' of 'route-rule' - one instance */
			cl_list_iterator_t    list_iterator;
			char	* tmp_str;

			if (p_current_route_rule && p_current_route_rule->name) {
				yyerror("route-rule has multiple 'name' tags");
				list_iterator =
					cl_list_head(&tmp_parser_struct.str_list);
				if ( list_iterator !=
					cl_list_end(&tmp_parser_struct.str_list) ) {
					tmp_str = (char*)cl_list_obj(list_iterator);
					if (tmp_str)
						free(tmp_str);
				}
				cl_list_remove_all(&tmp_parser_struct.str_list);
				return 1;
			}

			list_iterator =
				cl_list_head(&tmp_parser_struct.str_list);
			if ( list_iterator !=
				cl_list_end(&tmp_parser_struct.str_list) ) {
				tmp_str = (char*)cl_list_obj(list_iterator);
				if (tmp_str)
					p_current_route_rule->name = tmp_str;
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
					"policy Name:(%s)\n",
					p_current_route_rule->name);
			}
			cl_list_remove_all(&tmp_parser_struct.str_list);
		}
		;

route_rule_name_start:	TK_NAME {
				RESET_BUFFER;
			}
			;

route_rule_subnet_prefix:route_rule_subnet_prefix_start asterisk_or_list_of_ranges {
			/* 'route-guid' in 'route-rule' - any num of instances */
			/* list of guid ranges */
			sn_count++;
			if (cl_list_count(&tmp_parser_struct.num_pair_list))
			{
				uint64_t ** range_arr;
				unsigned range_len;

				if(!__rangelist2rangearr(
					&tmp_parser_struct.num_pair_list,
					&range_arr, &range_len )) {
					return -1;
				}

				__parser_add_prefix_range_to_prefix_list(
							range_arr, range_len);
			}
		}
		;

route_rule_subnet_prefix_start:	TK_SUBNET_PREFIX {
					RESET_BUFFER;
				}
				;

asterisk_or_list_of_ranges:	route_rule_asterisk
				| list_of_ranges
				;

route_rule_asterisk:	TK_ASTERISK {
				uint64_t * num_pair =
					(uint64_t*)malloc(sizeof(uint64_t)*2);
				num_pair[0] = 0x0;
				num_pair[1] = 0x0;
				cl_list_insert_tail(
					&tmp_parser_struct.num_pair_list,
					num_pair);
			}
			;
route_rule_router_guid:	route_rule_router_guid_start single_number {
				/* 'route-guid' in 'route-rule' -
				 * any num of instances
				 */
				cl_list_iterator_t    list_iterator;
				uint64_t	* p_tmp_num;

				guid_count++;

				list_iterator = cl_list_head(&tmp_parser_struct.num_list);
				p_tmp_num = (uint64_t*)cl_list_obj(list_iterator);
				if (p_tmp_num && p_current_route_rule) {
					p_current_route_rule->router_guid =
						cl_hton64((uint64_t)(*p_tmp_num));
					OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
						"GUID  0x%016" PRIx64 "\n",
						cl_ntoh64(p_current_route_rule->router_guid));
					free(p_tmp_num);
				}

				cl_list_remove_all(&tmp_parser_struct.num_list);
			}
			;

route_rule_router_guid_start:	TK_ROUTER_GUID {
					RESET_BUFFER;
				}
				;

route_rule_weight:	route_rule_weight_start single_number {
			/* 'weight' in 'route-rule' -
			 * one instance of one number
			 */
			cl_list_iterator_t    list_iterator;
			uint64_t	* p_tmp_num;

			weight_count++;
			list_iterator = cl_list_head(&tmp_parser_struct.num_list);
			p_tmp_num = (uint64_t*)cl_list_obj(list_iterator);
			if (p_tmp_num && p_current_route_rule) {
				p_current_route_rule->weight =
					(uint32_t)(*p_tmp_num);
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
					"Weight:%d\n",
					p_current_route_rule->weight);
				free(p_tmp_num);
			}
			cl_list_remove_all(&tmp_parser_struct.num_list);
			}
			;

route_rule_weight_start:	TK_ROUTE_WEIGHT {
					RESET_BUFFER;
				}
				;
	/*
	 * Parsing Load Share rules:
	 * -------------------
	 * load-share-rules
	 *	load-share-rule
	 *		subnet-prefix: 0xfec0000000000022
	 *		router-guid: 0x1000000000000001 - 60
	 *		router-guid: 0x1000000000000002 - 40
	 *	end-load-share-rule
	 * end-load-share-rules
	 */

load_share_rules_section:	TK_LOAD_SHARE_RULES_START load_share_rules TK_LOAD_SHARE_RULES_END
				;

load_share_rules:	load_share_rule
			| load_share_rules load_share_rule
			;

load_share_rule:	load_share_rule_start load_share_rule_entries load_share_rule_end
			;

load_share_rule_start:	TK_LOAD_SHARE_RULE_START {
				if (p_load_share_rule) {
					OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
						"ERR 9F01: Current load share rule already exist "
						"on rule start\n");
					return 1;
				}

				p_load_share_rule =
						(osm_load_share_rule_t *) calloc(1,
							sizeof(osm_load_share_rule_t));

				cl_map_init(&p_load_share_rule->load_share_map, 16);
				sn_count = 0;
				guid_count = 0;
			}
			;

load_share_rule_end:	TK_LOAD_SHARE_RULE_END {
				if ( __parser_load_share_rule_end() )
					return 1;
			}
			;

load_share_rule_entries:
			| load_share_rule_entries load_share_rule_entry
			;

load_share_rule_entry:	load_share_rule_subnet_prefix
			| load_share_rule_router_guid
			;
	/*
	 * load_share_rule_entry values:
	 *	load_share_subnet_prefix
	 *	load_share_rule_guid
	 */

load_share_rule_subnet_prefix:	load_share_rule_subnet_prefix_start single_number {
				/* 'subnet-prefix' in 'load-share-rule' - any num of instances */
				sn_count++;
				cl_list_iterator_t list_iterator;
				uint64_t * p_tmp_num;

				list_iterator = cl_list_head(&tmp_parser_struct.num_list);
				p_tmp_num = (uint64_t*)cl_list_obj(list_iterator);
				if (p_tmp_num) {
					p_load_share_rule->subnet_prefix = cl_hton64((uint64_t)(*p_tmp_num));

					OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
						"PREFIX  0x%016" PRIx64 "\n",
						cl_ntoh64(p_load_share_rule->subnet_prefix));
					free(p_tmp_num);
				}

				cl_list_remove_all(&tmp_parser_struct.num_list);
			}
			;

load_share_rule_subnet_prefix_start:	TK_SUBNET_PREFIX {
						RESET_BUFFER;
					}
					;

load_share_rule_router_guid:	load_share_rule_router_guid_start load_share_list {
				/* 'route-guid' in 'load-share-rule' - any num of instances */
				guid_count++;
				if (cl_list_count(&tmp_parser_struct.num_pair_list))
					if(__loadguidtomap(&tmp_parser_struct.num_pair_list) < 0)
						return -1;
				}
				;

load_share_rule_router_guid_start:	TK_ROUTER_GUID {
						RESET_BUFFER;
					}
					;

load_share_list:	load_share_list_with_dash
			;

load_share_list_with_dash:	single_number_load_share {
					uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
					num_pair[0] = tmp_parser_struct.num_pair[0];
					num_pair[1] = tmp_parser_struct.num_pair[1];
					cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
				}
				| number_load_share_1 TK_DASH number_load_share_2 {
					uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
					num_pair[0] = tmp_parser_struct.num_pair[0];
					num_pair[1] = tmp_parser_struct.num_pair[1];
					cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
				}
				|TK_WHITE {
					yyerror("GUID-Percent syntex error");
				}
				;

single_number_load_share:	TK_NUMBER {
					if(__parser_str2uint64(&tmp_parser_struct.num_pair[0],
						__route_parser_yylval))
						return -1;

					tmp_parser_struct.num_pair[1] = 50;

					OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
						"GUID  0x%016" PRIx64 "\n",
						cl_ntoh64(tmp_parser_struct.num_pair[0]));
					free(__route_parser_yylval);
				}
				;

number_load_share_1:	TK_NUMBER {
				if(__parser_str2uint64(&tmp_parser_struct.num_pair[0],
					__route_parser_yylval))
					return -1;

				OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
					"GUID 0x%016" PRIx64 "\n",
					cl_ntoh64(tmp_parser_struct.num_pair[0]));
				free(__route_parser_yylval);
			}
			;

number_load_share_2:	TK_NUMBER {
				if(__parser_str2uint64(&tmp_parser_struct.num_pair[1],
					__route_parser_yylval))
					return -1;

				OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
					"Percentage :%d\n",
					(uint32_t )tmp_parser_struct.num_pair[1]);
				free(__route_parser_yylval);
			}
			;

	/*
	 * Common part
	 */


single_string:	single_string_elems {
			cl_list_insert_tail(&tmp_parser_struct.str_list,
				strdup(__parser_strip_white(
						tmp_parser_struct.str)));
			tmp_parser_struct.str[0] = '\0';
		}
		;

single_string_elems:	single_string_element
			| single_string_elems single_string_element
			;

single_string_element:	TK_TEXT {
				int len = strlen(__route_parser_yylval);
				if(len > 1024)
					yyerror("String length is grater then 1024");	
				strcat(tmp_parser_struct.str,
					__route_parser_yylval);
				free(__route_parser_yylval);
			}
			;

single_number:	number
		;

number:	TK_NUMBER {
		uint64_t * p_num = (uint64_t*)malloc(sizeof(uint64_t));
		if(__parser_str2uint64(p_num,__route_parser_yylval)) {
			free(p_num);
			return -1;
		}

		free(__route_parser_yylval);
		cl_list_insert_tail(&tmp_parser_struct.num_list, p_num);
	}
	;

list_of_ranges:	num_list_with_dash
		;

num_list_with_dash:	single_number_from_range {
				uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
				num_pair[0] = tmp_parser_struct.num_pair[0];
				num_pair[1] = tmp_parser_struct.num_pair[1];
				cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
			}
			| number_from_range_1 TK_DASH number_from_range_2 {
				uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
				if (tmp_parser_struct.num_pair[0] < tmp_parser_struct.num_pair[1]) {
					num_pair[0] = tmp_parser_struct.num_pair[0];
					num_pair[1] = tmp_parser_struct.num_pair[1];
				}
				else {
					OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
						"ERR 9F18:"
						" Invalid Subnet Prefix Range"
						" Min: 0x%016" PRIx64 ""
						" Max: 0x%016" PRIx64 "\n",
						tmp_parser_struct.num_pair[0],
						tmp_parser_struct.num_pair[1]);
					free(num_pair);
					return -1;
				}
				cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
			}
			| num_list_with_dash TK_COMMA number_from_range_1 TK_DASH number_from_range_2 {
				uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
				if (tmp_parser_struct.num_pair[0] < tmp_parser_struct.num_pair[1]) {
					num_pair[0] = tmp_parser_struct.num_pair[0];
					num_pair[1] = tmp_parser_struct.num_pair[1];
				}
				else {
					OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
						"ERR 9F19:"
						" Invalid Subnet Prefix Range"
						" Min: 0x%016" PRIx64 ""
						" Max: 0x%016" PRIx64 "\n",
						tmp_parser_struct.num_pair[0],
						tmp_parser_struct.num_pair[1]);
					free(num_pair);
					return -1;
				}
				cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
			}
			| num_list_with_dash TK_COMMA single_number_from_range {
				uint64_t * num_pair = (uint64_t*)malloc(sizeof(uint64_t)*2);
				num_pair[0] = tmp_parser_struct.num_pair[0];
				num_pair[1] = tmp_parser_struct.num_pair[1];
				cl_list_insert_tail(&tmp_parser_struct.num_pair_list, num_pair);
			}
			;

single_number_from_range:	TK_NUMBER {
					if(__parser_str2uint64(&tmp_parser_struct.num_pair[0],
						__route_parser_yylval)) 
						return -1;

					if(__parser_str2uint64(&tmp_parser_struct.num_pair[1],
						__route_parser_yylval))
						return -1;

					free(__route_parser_yylval);
				}
				;

number_from_range_1:	TK_NUMBER {
				if(__parser_str2uint64(&tmp_parser_struct.num_pair[0],
					__route_parser_yylval))
						return -1;

				free(__route_parser_yylval);
			}
			;

number_from_range_2:	TK_NUMBER {
				if(__parser_str2uint64(&tmp_parser_struct.num_pair[1],
					__route_parser_yylval))
					return -1;
				free(__route_parser_yylval);
			}
			;

%%

/***************************************************
 ***************************************************/

void osm_route_parse_policy_file(IN osm_subn_t * p_subn)
{
	int res = 0;
	p_route_parser_osm_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_route_parser_osm_log);

	__route_parser_in = fopen (p_subn->opt.router_policy_file, "r");
	if (!__route_parser_in) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F02: "
			"fopen(%s) failed: %s",
			p_subn->opt.router_policy_file, strerror(errno));
		goto Exit;
	}

	column_num = 1;
	line_num = 1;
	p_router_policy = &p_subn->p_osm->sm.router_policy_mgr;
	__parser_tmp_struct_init();

	res = __route_parser_parse();

	__parser_tmp_struct_destroy();

	if (res != 0) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F03: "
			"Failed parsing Route policy file (%s)\n",
			p_subn->opt.router_policy_file);
		p_subn->opt.router_policy_file = NULL;
		goto Exit;
	}

Exit:
	if (__route_parser_in)
		fclose(__route_parser_in);

	OSM_LOG_EXIT(p_route_parser_osm_log);
}

/***************************************************
 ***************************************************/

static void yyerror(const char *format, ...)
{
	char s[256];
	va_list pvar;

	OSM_LOG_ENTER(p_route_parser_osm_log);

	va_start(pvar, format);
	vsnprintf(s, sizeof(s), format, pvar);
	va_end(pvar);

	OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F04: "
		"Syntax error (line %d:%d): %s\n",
		line_num, column_num, s);
	OSM_LOG_EXIT(p_route_parser_osm_log);
}

/***************************************************
 ***************************************************/

static char * __parser_strip_white(char * str)
{
	char *p;

	while (isspace(*str))
		str++;
	if (!*str)
		return str;
	p = str + strlen(str) - 1;
	while (isspace(*p))
		*p-- = '\0';

	return str;
}

int __route_parser_wrap()
{
	return 1;
}

/***************************************************
 ***************************************************/

static int __parser_str2uint64(uint64_t * p_val, char * str)
{
	if(__validate_uint64bit_string(str)){
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F1A "
			"Syntax error (line %d:%d): String: %s \n",
			line_num, column_num, str);
		free(str);		
		return -1;
	}
	
	*p_val = strtoull(str, NULL, 0);
	return 0;
}

/***************************************************
 ***************************************************/
static int __parser_route_rule_end()
{
	if(__validate_route_rule()) {
		if(p_current_route_rule){
			if (p_current_route_rule->name)
				free(p_current_route_rule->name);

			if(sn_count){
				uint64_t *prefix;
				cl_list_iterator_t list_iterator;
				for (list_iterator = 
					cl_list_head(&p_current_route_rule->prefix_list);
					list_iterator != 
					cl_list_end(&p_current_route_rule->prefix_list);
					list_iterator = cl_list_next(list_iterator)){
					prefix = (uint64_t *) cl_list_obj(list_iterator);
					if(prefix)
						free(prefix);
				}
				cl_list_remove_all(&p_current_route_rule->prefix_list);
				cl_list_destroy(&p_current_route_rule->prefix_list);
			}
			free(p_current_route_rule);
			p_current_route_rule = NULL;
		}
		return -1;
	}

	cl_list_insert_tail(&p_router_policy->route_rules, p_current_route_rule);
	p_current_route_rule = NULL;
	return 0;
}

static int __parser_route_rules_end()
{
	cl_list_iterator_t list_iterator;
	cl_list_iterator_t t_list_iterator;
	osm_route_rule_t *p_route_rule = NULL;
	osm_load_share_rule_t *p_load_share_rule = NULL;

	if(default_count > 1) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F1C "
			"Error More than One default route rule in a policy\n");

		/* Destroy route rules */
		list_iterator = cl_list_head(&p_router_policy->route_rules);
		while (list_iterator != cl_list_end(&p_router_policy->route_rules)) {
			p_route_rule = (osm_route_rule_t *) cl_list_obj(list_iterator);
			if (p_route_rule){
				osm_router_policy_route_rule_destroy(p_route_rule);
				t_list_iterator = cl_list_next(list_iterator);	
				cl_list_remove_item(
						&p_router_policy->route_rules,
						list_iterator);
				list_iterator = t_list_iterator;
			}
			else
				list_iterator = cl_list_next(list_iterator);	
		}

		/* Destroy load share rules */
		list_iterator = cl_list_head(&p_router_policy->load_share_rules);
		while (list_iterator != 
				cl_list_end(&p_router_policy->load_share_rules)) {
			p_load_share_rule = (osm_load_share_rule_t *)
						cl_list_obj(list_iterator);
			if (p_load_share_rule){
				osm_router_policy_load_share_rule_destroy(
								p_load_share_rule);
				t_list_iterator = cl_list_next(list_iterator);	
				cl_list_remove_item(
						&p_router_policy->load_share_rules,
						list_iterator);
				list_iterator = t_list_iterator;
			}
			else 
				list_iterator = cl_list_next(list_iterator);	
		}
		return -1;
	}
	return 0;
}

/***************************************************
 ***************************************************/
static int __parser_load_share_rule_end()
{
	cl_map_iterator_t map_it;

	if(__validate_load_share_rule()) {
		if(p_load_share_rule) {
			for (map_it = cl_map_head(&p_load_share_rule->load_share_map);
				map_it != cl_map_end(&p_load_share_rule->load_share_map);
				map_it = cl_map_next(map_it)) {
				free((void*)cl_map_obj(map_it));
			}
		cl_map_remove_all(&p_load_share_rule->load_share_map);
			cl_map_destroy(&p_load_share_rule->load_share_map);
		}
		return -1;
	}

	for (map_it = cl_map_head(&p_load_share_rule->load_share_map);
		map_it != cl_map_end(&p_load_share_rule->load_share_map);
		map_it = cl_map_next(map_it)) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
			"GUID 0x%016" PRIx64" - %d\n",
			cl_ntoh64(cl_map_key(map_it)),
			*((uint32_t*)cl_map_obj(map_it)));
	}

	cl_list_insert_tail(&p_router_policy->load_share_rules, p_load_share_rule);
	p_load_share_rule = NULL;
	return 0;
}

/***************************************************
 ***************************************************/

static void __parser_tmp_struct_init()
{
	tmp_parser_struct.str[0] = '\0';
	cl_list_construct(&tmp_parser_struct.str_list);
	cl_list_init(&tmp_parser_struct.str_list, 10);
	cl_list_construct(&tmp_parser_struct.num_list);
	cl_list_init(&tmp_parser_struct.num_list, 10);
	cl_list_construct(&tmp_parser_struct.num_pair_list);
	cl_list_init(&tmp_parser_struct.num_pair_list, 10);
}

/***************************************************
 ***************************************************/

/*
 * Do NOT free objects from the temp struct.
 * Either they are inserted into the parse tree data
 * structure, or they are already freed when copying
 * their values to the parse tree data structure.
 */
static void __parser_tmp_struct_reset()
{
	tmp_parser_struct.str[0] = '\0';
	cl_list_remove_all(&tmp_parser_struct.str_list);
	cl_list_remove_all(&tmp_parser_struct.num_list);
	cl_list_remove_all(&tmp_parser_struct.num_pair_list);
}

/***************************************************
 ***************************************************/

static void __parser_tmp_struct_destroy()
{
	__parser_tmp_struct_reset();
	cl_list_destroy(&tmp_parser_struct.str_list);
	cl_list_destroy(&tmp_parser_struct.num_list);
	cl_list_destroy(&tmp_parser_struct.num_pair_list);
}

/***************************************************
 ***************************************************/

static int __cmp_num_range(const void * p1, const void * p2)
{
	uint64_t * pair1 = *((uint64_t **)p1);
	uint64_t * pair2 = *((uint64_t **)p2);

	if (pair1[0] < pair2[0])
		return -1;
	if (pair1[0] > pair2[0])
		return 1;

	if (pair1[1] < pair2[1])
		return -1;
	if (pair1[1] > pair2[1])
		return 1;

	return 0;
}

/***************************************************
 ***************************************************/

static void __sort_reduce_rangearr(
	uint64_t  **   arr,
	unsigned       arr_len,
	uint64_t  ** * p_res_arr,
	unsigned     * p_res_arr_len )
{
	unsigned i = 0;
	unsigned j = 0;
	unsigned last_valid_ind = 0;
	unsigned valid_cnt = 0;
	uint64_t ** res_arr;
	boolean_t * is_valid_arr;

	*p_res_arr = NULL;
	*p_res_arr_len = 0;

	qsort(arr, arr_len, sizeof(uint64_t*), __cmp_num_range);

	is_valid_arr = (boolean_t *)malloc(arr_len * sizeof(boolean_t));
	is_valid_arr[last_valid_ind] = TRUE;
	valid_cnt++;
	for (i = 1; i < arr_len; i++)
	{
		if (arr[i][0] <= arr[last_valid_ind][1])
		{
			if (arr[i][1] > arr[last_valid_ind][1])
				arr[last_valid_ind][1] = arr[i][1];
			free(arr[i]);
			arr[i] = NULL;
			is_valid_arr[i] = FALSE;
		}
		else if ((arr[i][0] - 1) == arr[last_valid_ind][1])
		{
			arr[last_valid_ind][1] = arr[i][1];
			free(arr[i]);
			arr[i] = NULL;
			is_valid_arr[i] = FALSE;
		}
		else
		{
			is_valid_arr[i] = TRUE;
			last_valid_ind = i;
			valid_cnt++;
		}
	}

	res_arr = (uint64_t **)malloc(valid_cnt * sizeof(uint64_t *));
	for (i = 0; i < arr_len; i++)
	{
		if (is_valid_arr[i]){
			res_arr[j++] = arr[i];
		}
	}
	free(is_valid_arr);
	free(arr);

	*p_res_arr = res_arr;
	*p_res_arr_len = valid_cnt;
}

/***************************************************
 ***************************************************/
static int __validate_subnet_prefix_range(
			uint64_t  **   arr,
			unsigned       arr_len)
{
	int i;
	for (i = 0; i < arr_len; i++)
	{
		if ((arr[i][1] - arr[i][0]) >= 4096) {
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, 
				"ERR 9F17: "
				"Subnet Prefix range is more than"
				" allowed (4096) Configured range :0x%016" PRIx64""
				" Low prefix: 0x%016" PRIx64""
				" High Prefix:0x%016" PRIx64"\n",
				(arr[i][1] - arr[i][0]), arr[i][0], arr[i][1]);
				return 0;
		}
	}
	return 1;
}

/***************************************************
 ***************************************************/

static int __rangelist2rangearr(
	cl_list_t    * p_list,
	uint64_t  ** * p_arr,
	unsigned     * p_arr_len)
{
	cl_list_iterator_t list_iterator;
	unsigned len = cl_list_count(p_list);
	unsigned i = 0;
	uint64_t ** tmp_arr;
	uint64_t ** res_arr = NULL;
	unsigned res_arr_len = 0;

	tmp_arr = (uint64_t **)malloc(len * sizeof(uint64_t *));

	list_iterator = cl_list_head(p_list);
	while( list_iterator != cl_list_end(p_list) )
	{
		tmp_arr[i++] = (uint64_t *)cl_list_obj(list_iterator);
		list_iterator = cl_list_next(list_iterator);
	}
	cl_list_remove_all(p_list);

	if(!__validate_subnet_prefix_range(tmp_arr,len)) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F14: "
			"Invalid subnet-prefix range \n");
		free(tmp_arr);	
		return 0;
	}
	__sort_reduce_rangearr( tmp_arr, len, &res_arr,
				&res_arr_len );
	*p_arr = res_arr;
	*p_arr_len = res_arr_len;
	return 1;
}

/***************************************************
 ***************************************************/

static void __parser_add_prefix_range_to_prefix_list(
	uint64_t  ** range_arr,
	unsigned     range_len)
{
	unsigned i;
	uint64_t *prefix;
	uint64_t prefix_ho;
	cl_qlist_t *ql = &(&p_current_route_rule->prefix_list)->list;

	if (!range_arr || !range_len)
		return;

	if(!p_current_route_rule){
		for (i = 0; i < range_len; i++)
			free(range_arr[i]);
		free(range_arr);
		return;
	}

	if(ql->state != CL_INITIALIZED){
		cl_list_construct(&p_current_route_rule->prefix_list);
		cl_list_init(&p_current_route_rule->prefix_list, 32);
	}

	for (i = 0; i < range_len; i++) {
		for (prefix_ho = range_arr[i][0];
			prefix_ho <= range_arr[i][1]; prefix_ho++) {
			prefix = (uint64_t*)malloc(sizeof(uint64_t));
			*prefix = cl_hton64(prefix_ho);
			cl_list_insert_tail(&p_current_route_rule->prefix_list, prefix);
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
				"PREFIX:  0x%016" PRIx64 " \n",
				cl_ntoh64(*prefix));
		}
		free(range_arr[i]);
	}
	free(range_arr);
}

/***************************************************
 ***************************************************/
static int __validate_route_rule()
{
	cl_list_iterator_t list_iterator;
	uint64_t prefix;

	if(!p_current_route_rule)
		return -1;

	/* validate route rule name */
	if(p_current_route_rule->name){
		if(strcmp(p_current_route_rule->name, "static")
			&& strcmp(p_current_route_rule->name, "dynamic")
			&& strcmp(p_current_route_rule->name, "default")) {
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
				"ERR 9F05: "
				"Invalid name specified"
				" Rule End line No: %d\n",
				line_num-3);
			return -1;
		}
		if(!strcmp(p_current_route_rule->name, "default"))
			default_count++;
	}
	else {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F06: "
			"No name specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}

	/* Validate subnet_prefix */
	if(sn_count && cl_list_count(&p_current_route_rule->prefix_list) > 0) {
		for (list_iterator = cl_list_head(&p_current_route_rule->prefix_list);
			list_iterator != cl_list_end(&p_current_route_rule->prefix_list);
			list_iterator = cl_list_next(list_iterator)){
			prefix = *((uint64_t *) cl_list_obj(list_iterator));

			if(!strcmp(p_current_route_rule->name, "static")
				&& !((prefix & IB_LOCAL_SUBNET_MASK) ==
				IB_SITE_LOCAL_SUBNET_PREFIX)){
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
					"ERR 9F07: "
					"Invalid static Subnet Prefix"
					": 0x%016" PRIx64
					" Rule End line No: %d\n",
					cl_ntoh64(prefix), line_num);
				return -1;
			}
			else if(!strcmp(p_current_route_rule->name, "default")
				&& prefix != 0x0){
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
					"ERR 9F08: "
					"Invalid Default Subnet Prefix"
					": 0x%016" PRIx64
					" Rule End line No: %d\n",
					cl_ntoh64(prefix), line_num);
				return -1;
			}
		}
	}

	else {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,"ERR 9F09: "
			"No subnet prefix specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}

	if(sn_count > 1) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0A: "
			"Multiple subnet prefix entry specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}

	/* Validate router Guid */
	if(guid_count == 0) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0B: "
			"No Router GUID specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}
	else if(guid_count > 1) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0C: "
			"Multiple Router GUID specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}
	else if(!p_current_route_rule->router_guid) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0D: "
			"Invalid Router GUID"
			": 0x%016" PRIx64
			" Rule End line No: %d\n",
			p_current_route_rule->router_guid, 
			line_num);
		return -1;
	}

	/*Validate weight 0 - 10*/
	if(weight_count == 0) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0E: "
			"No weight specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}
	else if (weight_count > 1) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F0F: "
			"Multiple weight entry specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;

	}
	else if(p_current_route_rule->weight > 10) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F10: "
			"Invalid weight :%d "
			" Rule End line No: %d\n",
			p_current_route_rule->weight,
			line_num);
		return -1;
	}

	return 0;
}

/***************************************************
 ***************************************************/
static int __loadguidtomap(cl_list_t * p_list)
{
	int ret = 0; 
	void *ptr= NULL;
	uint32_t *percent = NULL;
	uint64_t *range_arr = NULL;
	cl_list_iterator_t list_iterator;

	list_iterator = cl_list_head(p_list);


	while( list_iterator != cl_list_end(p_list) ) {
		range_arr = (uint64_t *)cl_list_obj(list_iterator);
		/* Check for duplicate key */
		ptr = cl_map_get(&p_load_share_rule->load_share_map, cl_hton64(range_arr[0]));

		if(!ptr){
			percent = (uint32_t*)malloc(sizeof(uint32_t));
			if(!percent) {
				OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
					"ERR 9F1C:"
					" Memory allocation failure\n");
				ret = -1;
				free(range_arr);
				break;
			}

			*percent = range_arr[1];
			cl_map_insert(&p_load_share_rule->load_share_map,
					cl_hton64(range_arr[0]), percent);
			list_iterator = cl_list_next(list_iterator);
			free(range_arr);
			ret = 1;
		}
		else {	
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
				"ERR 9F1E:"
				" Dublicate Router GUID "
				": 0x%016" PRIx64"\n", range_arr[0]);
			free(range_arr);
			ret = -1;
			break;
		}
	}
	
	cl_list_remove_all(p_list);
	return ret;
}

/***************************************************
 ***************************************************/
static int __validate_load_share_rule()
{
	int percent = 0;
	int ret = -1;
	uint64_t guid;
	cl_map_iterator_t map_it;

	if(sn_count == 0) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F14: "
			"No subnet prefix entry specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}
	else if(sn_count > 1) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F15: "
			"Multiple subnet prefix entry specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}

	if(!((p_load_share_rule->subnet_prefix & IB_LOCAL_SUBNET_MASK) ==
		IB_SITE_LOCAL_SUBNET_PREFIX)){
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
			"ERR 9F11: "
			"Invalid Subnet Prefix"
			": 0x%016" PRIx64
			" Rule End line No: %d\n",
			cl_ntoh64(p_load_share_rule->subnet_prefix),
			line_num);
		return -1;
	}

	/* Validate router Guid */
	if(guid_count == 0) {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F16: "
			"No Router GUID specified"
			" Rule End line No: %d\n",
			line_num);
		return -1;
	}

	for (map_it = cl_map_head(&p_load_share_rule->load_share_map);
		map_it != cl_map_end(&p_load_share_rule->load_share_map);
		map_it = cl_map_next(map_it)) {
		guid = cl_map_key(map_it);
		percent = percent + *((uint32_t*)cl_map_obj(map_it));

		if((guid == INVALID_GUID) || guid == 0x0) {
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
				"ERR 9F12: "
				"Invalid GUID"
				": 0x%016" PRIx64
				" Rule End line No: %d\n",
				cl_ntoh64(guid), line_num);
			return -1;
		}

		if(percent > 100) {
			OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR,
				"ERR 9F13: "
				"Invalid total percentage:%d"
				" Rule End line No: %d\n",
				percent, line_num);	
			return -1;
		}
		ret = 0;
	}

	return ret;
}

/***************************************************
 ***************************************************/
static int __validate_uint64bit_string(char *str)
{
	int len, i;
	char *ptr = str;
	len= strlen(ptr);

	if(len == 18 && !strncmp(ptr, "0x", 2)){
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
			"String :%s String len:%d\n",
			ptr, len);
		ptr +=2;
		for(i = 2; i < len; i++) {
			if(!isxdigit(*ptr))
				return -1; 
			ptr += 1;
		}
	}
	else if(len <= 3){
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_DEBUG,
			"String :%s String len:%d\n",
			ptr, len);
		for(i = 0; i < len; i++) {
			if(!isxdigit(*ptr))
				return -1;
			ptr += 1;
		}
	}
	else {
		OSM_LOG(p_route_parser_osm_log, OSM_LOG_ERROR, "ERR 9F1B "
			"Invalid String :%s String len:%d\n",
			ptr, len);
		return -1; 
	}
	return 0;
}
