/*
 * Copyright (c) 2016, Bay Microsystems, Inc All Rights Reserved
 * Copyright (c) 2016 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
/*
 * Abstract:
 *    OSM Route Policy functions.
 *
 * Author:
 *    Manickavel Subramani, Bay Microsystems Inc
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <stdbool.h>
#include <pthread.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <complib/cl_qlist.h>
#include <complib/cl_map.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_POLICY_C
#include <opensm/osm_log.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_sa.h>
#include <opensm/osm_router.h>
#include <opensm/osm_prefix_route.h>
#include <opensm/osm_router_policy.h>

typedef struct router_data {
	uint64_t guid;
	uint16_t type;
	uint16_t weight;
	int percent;
	uint16_t count;
	boolean_t dup;
	boolean_t active;
	boolean_t ls_enabled;
} router_data_t;

typedef struct router_factor{
	uint16_t type;
	uint64_t guid;
	int factor;
	uint16_t selected;
	boolean_t is_active;
} router_factor_t;

/* Route data array temporarily holds
 * route data pointers during route
 * process
 */
#define MAX_TEMP_ROUTE_DATA 32
router_data_t *r_data[MAX_TEMP_ROUTE_DATA];
boolean_t is_inited = false; 

#define ROUTE_TYPE_DEFAULT 0
#define ROUTE_TYPE_STATIC 1
#define ROUTE_TYPE_DYNAMIC 2

typedef struct osm_router_item {
	cl_list_item_t list_item;
	osm_router_t *p_rtr;
} osm_router_item_t;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

/***************************************************
 ***************************************************/

void osm_router_policy_route_rule_destroy(osm_route_rule_t * p)
{
	cl_list_iterator_t list_iterator;
	uint64_t *prefix;

	if (!p)
		return;
	if (p->name)
		free(p->name);

	 for (list_iterator = cl_list_head(&p->prefix_list);
		list_iterator != cl_list_end(&p->prefix_list);
		list_iterator = cl_list_next(list_iterator)){
		prefix = (uint64_t *) cl_list_obj(list_iterator);
		if(prefix)
			free(prefix);
	}
		
	cl_list_remove_all(&p->prefix_list);
	cl_list_destroy(&p->prefix_list);

	free(p);
}
/***************************************************
 ***************************************************/

void osm_router_policy_load_share_rule_destroy(osm_load_share_rule_t * p)
{
	cl_map_iterator_t map_it;

	if (!p)
		return;

	for (map_it = cl_map_head(&p->load_share_map);
		map_it != cl_map_end(&p->load_share_map);
		map_it = cl_map_next(map_it))
		free((void*)cl_map_obj(map_it));

	cl_map_remove_all(&p->load_share_map);
	cl_map_destroy(&p->load_share_map);

	free(p);
}

/***************************************************
 ***************************************************/

void osm_router_policy_construct(IN osm_router_policy_t * p_rp)
{
	cl_list_construct(&p_rp->route_rules);
	cl_list_init(&p_rp->route_rules, 32);

	cl_list_construct(&p_rp->load_share_rules);
	cl_list_init(&p_rp->load_share_rules, 32);

	cl_map_construct(&p_rp->load_share_state);
	cl_map_init(&p_rp->load_share_state,4);
}

/***************************************************
 ***************************************************/
void osm_router_policy_destroy(IN osm_router_policy_t * p_rp)
{
	cl_list_iterator_t list_iterator;
	cl_map_iterator_t map_it;
	cl_list_t *t_list;
	osm_route_rule_t *p_route_rule = NULL;
	osm_load_share_rule_t *p_load_share_rule = NULL;
	router_factor_t * r_factor = NULL;

	if (!p_rp)
		return;

	/* Destroy route rules */
	list_iterator = cl_list_head(&p_rp->route_rules);
	while (list_iterator != cl_list_end(&p_rp->route_rules)) {
		p_route_rule = (osm_route_rule_t *) cl_list_obj(list_iterator);
		if (p_route_rule)
			osm_router_policy_route_rule_destroy(p_route_rule);
		list_iterator = cl_list_next(list_iterator);
	}
	cl_list_remove_all(&p_rp->route_rules);
	cl_list_destroy(&p_rp->route_rules);

	/* Destroy load share rules */
	list_iterator = cl_list_head(&p_rp->load_share_rules);
	while (list_iterator != cl_list_end(&p_rp->load_share_rules)) {
		p_load_share_rule = (osm_load_share_rule_t *)
					cl_list_obj(list_iterator);
		if (p_load_share_rule)
			osm_router_policy_load_share_rule_destroy(
							p_load_share_rule);
		list_iterator = cl_list_next(list_iterator);
	}
	cl_list_remove_all(&p_rp->load_share_rules);
	cl_list_destroy(&p_rp->load_share_rules);

	/* Destroy load share state */
	for (map_it = cl_map_head(&p_rp->load_share_state);
		map_it != cl_map_end(&p_rp->load_share_state);
		map_it = cl_map_next(map_it)) {
		t_list = cl_map_obj(map_it);
		list_iterator = cl_list_head(t_list);
		while (list_iterator != cl_list_end(t_list)){
			r_factor = (router_factor_t*)cl_list_obj(list_iterator);
			if(r_factor)
				free(r_factor);
			list_iterator = cl_list_next(list_iterator);
		}
		cl_list_remove_all(t_list);
		cl_list_destroy(t_list);
		free(t_list);
	}

	cl_map_remove_all(&p_rp->load_share_state);
	cl_map_destroy(&p_rp->load_share_state);
	pthread_mutex_destroy(&mutex);
}

/***************************************************
 ***************************************************/
void osm_router_policy_init(OUT osm_sm_t * p_sm)
{
	if (!p_sm->p_subn->opt.router_policy_file)
		return;

	is_inited = false;
	OSM_LOG_ENTER(&p_sm->p_subn->p_osm->log);

	/* read Route policy config file */
	osm_route_parse_policy_file(p_sm->p_subn);

	OSM_LOG_EXIT(&p_sm->p_subn->p_osm->log);
	is_inited = true;
}

/***************************************************
 ***************************************************/
static void osm_select_load_share_route(const osm_subn_t * p_subn,
					uint64_t prefix,
					uint16_t num_items,
					osm_router_t ** r_rtr)
{
	int i, ftr, pt;
	int f_temp = -100;
	boolean_t found = false;
	cl_map_item_t *p_item;
	cl_list_iterator_t list_iterator;
	cl_list_iterator_t t_list_iterator;
	cl_list_t *t_list = NULL;
	router_factor_t * r_factor = NULL;
	router_factor_t * t_factor = NULL;
	osm_router_policy_t * p_router_policy=NULL;
	osm_log_t *p_osm_log = &p_subn->p_osm->log;

	p_router_policy = &p_subn->p_osm->sm.router_policy_mgr;

	if(!p_router_policy)
		return;

	OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "Prefix: 0x%016" PRIx64
		" Number of selected Routes:%d\n",
		cl_ntoh64(prefix), num_items);

	/* Get the current state ls_factor list for Prefix */
	t_list = (cl_list_t *)cl_map_get(&p_router_policy->load_share_state, prefix);

	if(!t_list)
	{
		/*create new element map for the prefix */
		t_list = (cl_list_t *)malloc(sizeof(cl_list_t));
		if(!t_list){
			OSM_LOG(p_osm_log, OSM_LOG_ERROR,"ERR 9C00: "
			"Memory allocation failure\n");
			return;
		}
		cl_list_construct(t_list);
		cl_list_init(t_list, 32);
		cl_map_insert(&p_router_policy->load_share_state, prefix, t_list);
	}

	/* Purge load sharing state list */	
	list_iterator = cl_list_head(t_list);
	while (list_iterator != cl_list_end(t_list)){
		r_factor = (router_factor_t*)cl_list_obj(list_iterator);
		found = false;
		for(i = 0; i < num_items; i++)
			if(r_factor != NULL && r_factor->guid == r_data[i]->guid)
				found = true;				
		if(!found && r_factor){
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Removing load sharing state"
				"GUID: 0x%016" PRIx64" \n",
				cl_ntoh64(r_factor->guid));
			t_list_iterator = cl_list_next(list_iterator);
			cl_list_remove_item(t_list, list_iterator);
			free(r_factor);
			list_iterator = t_list_iterator;
		}
		else
			list_iterator = cl_list_next(list_iterator);
	}

	for(i = 0; i < num_items; i++) {
		if(r_data[i]->dup || r_data[i]->percent == 0)
			continue;
		pt = (100 - r_data[i]->percent);
		found = false;
		for (list_iterator = cl_list_head(t_list);
			list_iterator != cl_list_end(t_list);
			list_iterator = cl_list_next(list_iterator)) {
			r_factor = (router_factor_t*)cl_list_obj(list_iterator);

			if(r_factor != NULL && (r_factor->guid == r_data[i]->guid)){
				if(pt == 0)
					ftr = -1;
				else {
					if(r_factor->factor == 0
						&& !r_factor->selected)
						ftr = r_data[i]->percent;
					else if(r_factor->factor < 0){
						ftr = r_data[i]->percent
							- ((r_factor->selected)
							* (r_data[i]->percent))
							+ r_factor->factor;
					}
					else
						ftr = (r_factor->factor)
							- ((r_factor->selected)
							* (pt));
				}

				r_factor->is_active = true;
				r_factor->factor = ftr;
				found = true;
			}
		}

		if(!found) {
			r_factor = (router_factor_t *)
					malloc(sizeof(router_factor_t));
			memset(r_factor, 0x0, sizeof(router_factor_t));
			if (!pt)
				r_factor->factor = -1;
			else
				r_factor->factor = r_data[i]->percent;

			r_factor->is_active = true;
			r_factor->guid = r_data[i]->guid;
			r_factor->type = r_data[i]->type;
			cl_list_insert_tail(t_list, r_factor);
		}
	}

	for (list_iterator = cl_list_head(t_list);
		list_iterator != cl_list_end(t_list);
		list_iterator = cl_list_next(list_iterator)) {
		r_factor = (router_factor_t*)cl_list_obj(list_iterator);

		if(!r_factor->is_active)
			continue;

		if(r_factor->factor > f_temp){
			f_temp = r_factor->factor;
			t_factor = r_factor;
		}

		if(r_factor->factor ==  f_temp && r_factor->selected){
			f_temp = r_factor->factor;
			t_factor = r_factor;
		}

		r_factor->selected = 0;
		r_factor->is_active = false;
	}

	if(t_factor) {
		t_factor->selected=1;

		p_item = cl_qmap_get(&p_subn->rtr_guid_tbl, t_factor->guid);
		if (p_item == cl_qmap_end(&p_subn->rtr_guid_tbl)) {
			OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C01: No"
				" load share router found guid 0x%016" PRIx64
				"\n", cl_ntoh64(t_factor->guid));
		}
		else {
			*r_rtr = (osm_router_t *)p_item;
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Selected load share GUID: 0x%016" PRIx64 "\n",
				cl_ntoh64(t_factor->guid));
		}
	}
	else {
		OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
			" No load share route for prefix: 0x%016" PRIx64 "\n",
			cl_ntoh64(prefix));
	}
}

/***************************************************
 ***************************************************/
static void osm_resolve_route(const osm_subn_t * p_subn,
				uint64_t prefix,
				cl_list_t *rtr_list,
				osm_router_t ** r_rtr)
{
	int *percent = NULL;
	int num_routes = 0, idx = 0;
	int p_remain, t_weight, t_type, found;
	uint32_t i, j, num_entries = 0, block_num = 0;
	uint64_t t_prefix = 0;
	uint64_t t_guid = 0;
	uint64_t dyn_guid[MAX_TEMP_ROUTE_DATA];
	int num_dyn_routes = 0;
	boolean_t dyn_found = false;
	cl_list_iterator_t list_iterator;
	cl_list_iterator_t prefix_iterator;
	cl_map_item_t *p_item = NULL;
	osm_route_rule_t *p_route_rule = NULL;
	osm_router_t *p_rtr = NULL;
	ib_rtr_next_hop_table_block_t *p_nh_block = NULL;
	osm_load_share_rule_t * p_ls_rule = NULL;
	osm_router_policy_t * p_router_policy=NULL;
	osm_log_t *p_osm_log = &p_subn->p_osm->log;
	cl_map_iterator_t map_it;

	OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "Prefix 0x%016" PRIx64
		" Number of Routers:%d\n", cl_ntoh64(prefix),
		(int)cl_list_count(rtr_list));

	/* Get dynamic route for the prefix */
	for (list_iterator = cl_list_head(rtr_list);
		list_iterator != cl_list_end(rtr_list);
		list_iterator = cl_list_next(list_iterator)) {
		p_rtr = (osm_router_t *) cl_list_obj(list_iterator);
		num_entries = 0;
		if(p_rtr)
			num_entries =
				cl_ntoh32(p_rtr->router_info.next_hop_table_top);
		else
			continue;

		OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "Router GUID: 0x%016" PRIx64
			" Next Hop entries: %d\n",
			cl_ntoh64(p_rtr->p_port->guid), num_entries);

		for (i = 0; i < num_entries; i++) {
			if ((i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK) == 0) {
				block_num = i / IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
				if (block_num >=
					cl_ptr_vector_get_size
						(&p_rtr->next_hop_router_tbl.blocks)) {
					OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C02: "
						"next hop block number %u in "
						"subnets prefixes table is not found, "
						"unconsistent to table top value %u",
						block_num, num_entries);
					break;
				}
				p_nh_block = (ib_rtr_next_hop_table_block_t *)
						cl_ptr_vector_get(
							&p_rtr->next_hop_router_tbl.blocks,
							block_num);
			}

			/* skip NULL blocks */
			if (!p_nh_block) {
				OSM_LOG(p_osm_log, OSM_LOG_INFO,
					"Encountered NULL next hop router "
					"table block, block number %u.\n",
					block_num);

				/* promote i to first element on the next block */
				i += IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
				continue;
			}

			idx = i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;

			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Dynamic Route Index:%d prefix 0x%016"
				PRIx64 " Weight:%d \n", idx,
				cl_ntoh64(p_nh_block->record[idx].subnet_prefix),
				p_nh_block->record[idx].weight);

			if(p_nh_block->record[idx].subnet_prefix == prefix) {
				r_data[num_routes] = (router_data_t *)
							calloc(1, sizeof(router_data_t));

				if(r_data[num_routes] == NULL) {
					OSM_LOG(p_osm_log, OSM_LOG_ERROR,
						"ERR 9C05: "
						" Memory Allocation failure \n");
					goto Exit;
				}
				r_data[num_routes]->guid = p_rtr->p_port->guid;
				r_data[num_routes]->weight = p_nh_block->record[idx].weight;
				r_data[num_routes]->type = ROUTE_TYPE_DYNAMIC;
				r_data[num_routes]->active = true;

				dyn_guid[num_dyn_routes] = p_rtr->p_port->guid;

				OSM_LOG(p_osm_log, OSM_LOG_ERROR,
					"Selected Dynamic Router GUID: 0x%016"
					PRIx64 " Weight: %d\n",
					cl_ntoh64(p_rtr->p_port->guid),
					r_data[num_routes]->weight);
				num_dyn_routes++;
				num_routes++;
				if(num_routes == MAX_TEMP_ROUTE_DATA) {
					OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C06: "
						"Number of router reachability"
						" exceeded \n");
					goto Exit;
				}
			}
		}
	}

	/* Get static route for the prefix */
	if(!is_inited){
		if(num_routes)
			goto route_select;
		else
			goto Exit;
	}

	p_router_policy = &p_subn->p_osm->sm.router_policy_mgr;
	if (!(cl_list_count(&p_router_policy->route_rules)))
		goto load_share;

	for (list_iterator = cl_list_head(&p_router_policy->route_rules);
		list_iterator != cl_list_end(&p_router_policy->route_rules);
		list_iterator = cl_list_next(list_iterator)) {

		p_route_rule = (osm_route_rule_t *) cl_list_obj(list_iterator);
		if (p_route_rule && !strcmp(p_route_rule->name, "static")) {
			for(prefix_iterator = cl_list_head(&p_route_rule->prefix_list);
				prefix_iterator != cl_list_end(&p_route_rule->prefix_list);
				prefix_iterator = cl_list_next(prefix_iterator)) {
				t_prefix = *((uint64_t*)cl_list_obj(prefix_iterator));

				if(t_prefix == prefix){
					/* Check dyanmic route exist for the static one */
					for(i = 0; i < num_dyn_routes; i++) {
						if(dyn_guid[i] == p_route_rule->router_guid) {
							dyn_found = true;
							break;
						}
					}
					
					if(!dyn_found)
						continue;

					dyn_found = false;
						
					r_data[num_routes] = (router_data_t *)
								calloc(1, sizeof(router_data_t));
					
					if(r_data[num_routes] == NULL) {
						OSM_LOG(p_osm_log, OSM_LOG_ERROR,
							"ERR 9C07: "
							" Memory Allocation"
							" failure \n");
						goto Exit;
					}
					r_data[num_routes]->guid = p_route_rule->router_guid;
					r_data[num_routes]->weight = p_route_rule->weight;
					r_data[num_routes]->type = ROUTE_TYPE_STATIC;
					p_item = cl_qmap_get(
							&p_subn->rtr_guid_tbl,
							r_data[num_routes]->guid);
					if (p_item ==
						cl_qmap_end(&p_subn->rtr_guid_tbl)) {
						r_data[num_routes]->active = false;
						OSM_LOG(p_osm_log, OSM_LOG_INFO,
							"Static Route GUID: 0x%016"
							PRIx64 " Weight: %d"
							" Not found or Inactive\n",
							cl_ntoh64(r_data[num_routes]->guid),
							r_data[num_routes]->weight);
					}
					else
						r_data[num_routes]->active = true;

					OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
						"Static Route GUID: 0x%016"
						PRIx64 " Weight: %d\n",
						cl_ntoh64(r_data[num_routes]->guid),
						r_data[num_routes]->weight);
					num_routes++;
					if(num_routes == MAX_TEMP_ROUTE_DATA) {
						OSM_LOG(p_osm_log, OSM_LOG_ERROR,
							"ERR 9C08: "
							"Number of router"
							" reachability exceeded \n");
						goto Exit;
					}
					break;
				}
			}
		}
	}

	/* Get Default route for the prefix */
	for (list_iterator = cl_list_head(&p_router_policy->route_rules);
		list_iterator != cl_list_end(&p_router_policy->route_rules);
		list_iterator = cl_list_next(list_iterator)) {

		p_route_rule = (osm_route_rule_t *) cl_list_obj(list_iterator);
		if (p_route_rule && !strcmp(p_route_rule->name, "default")) {
			for(i = 0; i < num_dyn_routes; i++) {
				if(dyn_guid[i] == p_route_rule->router_guid) {
					dyn_found = true;
					break;
				}
			}

			if(!dyn_found)
				continue;

			dyn_found = false;

			r_data[num_routes] = (router_data_t *)
						calloc(1, sizeof(router_data_t));
			
			if(r_data[num_routes] == NULL) {
				OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C09: "
					" Memory Allocation failure \n");
				goto Exit;
			}
			r_data[num_routes]->guid = p_route_rule->router_guid;
			r_data[num_routes]->weight = p_route_rule->weight;
			r_data[num_routes]->type = ROUTE_TYPE_DEFAULT;
			p_item = cl_qmap_get(&p_subn->rtr_guid_tbl,
						r_data[num_routes]->guid);
			if (p_item == cl_qmap_end(&p_subn->rtr_guid_tbl)) {
				r_data[num_routes]->active = false;
				OSM_LOG(p_osm_log, OSM_LOG_INFO,
					"Default Route GUID: 0x%016"
					PRIx64 " Weight: %d Not found\n",
					cl_ntoh64(r_data[num_routes]->guid),
					r_data[num_routes]->weight);
			}
			else 
				r_data[num_routes]->active = true;

			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Default Route GUID: 0x%016"
				PRIx64 " Weight: %d\n",
				cl_ntoh64(r_data[num_routes]->guid),
				r_data[num_routes]->weight);

			num_routes++;
			if(num_routes == MAX_TEMP_ROUTE_DATA) {
				OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C0A: "
					"Number of router reachability"
					" exceeded \n");
				goto Exit;
			}
			break;
		}
	}

load_share:
	if(num_routes == 0)
		goto Exit;

	/* Get load-share policy pointer */
	if (!(cl_list_count(&p_router_policy->load_share_rules)))
		goto route_select;

	for (list_iterator = cl_list_head(&p_router_policy->load_share_rules);
		list_iterator != cl_list_end(&p_router_policy->load_share_rules);
		list_iterator = cl_list_next(list_iterator)) {
		p_ls_rule = (osm_load_share_rule_t *) cl_list_obj(list_iterator);
		if(p_ls_rule->subnet_prefix == prefix){
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "Load Share match"
				"prefix 0x%016" PRIx64 "\n",
				cl_ntoh64(prefix));
			break;
		}
		p_ls_rule = NULL;
	}

	if(!p_ls_rule)
		goto route_select;

	/* Check prefix reachability via routers in the r_data list 
	   if not mark them inactive
	 */
	found = false;
	for (map_it = cl_map_head(&p_ls_rule->load_share_map);
		map_it != cl_map_end(&p_ls_rule->load_share_map);
		map_it = cl_map_next(map_it)) {
		OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
			"GUID 0x%016" PRIx64" - %d\n",
			cl_ntoh64(cl_map_key(map_it)),
			*((uint32_t*)cl_map_obj(map_it)));

		t_guid = cl_map_key(map_it);
		for(i = 0; i< num_routes; i++) {
			if(t_guid == r_data[i]->guid
				&& r_data[i]->type == ROUTE_TYPE_DYNAMIC) {
				found = true;
				break;
			}
		}
		if(!found) {
			OSM_LOG(p_osm_log, OSM_LOG_INFO,
				"No Load sharing route for Prefix: 0x%016" PRIx64
				" via router GUID:0x%016" PRIx64"\n",
				cl_ntoh64(prefix), cl_ntoh64(t_guid));
		}
		else {
			for(i = 0; i< num_routes; i++) {
				found = false;
				if(r_data[i]->type == ROUTE_TYPE_DYNAMIC) {
					r_data[i]->ls_enabled = true;
					continue;
				}
				for(j = 0; j < num_routes; j++) {
					if(r_data[j]->type == ROUTE_TYPE_STATIC
						|| r_data[j]->type
							== ROUTE_TYPE_DEFAULT)
						continue;
					if(r_data[i]->guid == r_data[j]->guid){
						r_data[j]->ls_enabled = true;
						found = true;
					}
				}
				if(!found)
					OSM_LOG(p_osm_log, OSM_LOG_INFO,
						"No Load sharing route for"
						" Prefix: 0x%016" PRIx64
						" via router GUID:0x%016"
						 PRIx64"\n",
						cl_ntoh64(prefix),
						cl_ntoh64(r_data[i]->guid));
			}
		}	
		found = false;
	}

	idx = -1;
	p_remain = 100;
	t_weight = 0;
	t_type = ROUTE_TYPE_DYNAMIC;

	/* Load guid percent value to the r_data */
	for(i = 0; i < num_routes; i++) {
		for(j = 0; j< num_routes; j++) {
			if(r_data[i]->guid == r_data[j]->guid) {
				percent = cl_map_get(&p_ls_rule->load_share_map,
							r_data[j]->guid);
				if(percent){
					if(!r_data[j]->ls_enabled) {
						percent = NULL;
						continue;
					}
				 	if(!r_data[j]->count && r_data[j]->active
						&& r_data[j]->weight >= t_weight
						&& r_data[j]->type <= t_type){
						t_weight = r_data[j]->weight;
						t_type = r_data[j]->type;
						r_data[j]->percent = *percent;
						r_data[j]->dup = false;
						r_data[j]->count = 1;
						p_remain = p_remain
							- r_data[j]->percent;

						if(idx >=0) {
							p_remain = p_remain
								+ r_data[idx]->percent;
							r_data[idx]->percent = 0;
							r_data[idx]->dup = true;
						}
						idx = j;
					}
					else if (!r_data[j]->count){
						r_data[j]->dup = true;
						r_data[j]->count = 1;
					}
				}
			}
			percent = NULL;
		}
		if(idx == -1)
			continue;

		OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "GUID: 0x%016" PRIx64
			" Percent:%d remain:%d \n",
			cl_ntoh64(r_data[idx]->guid),
			r_data[idx]->percent, p_remain);

		idx = -1;
		t_weight = 0;
		t_type = ROUTE_TYPE_DYNAMIC;
	}

	idx = -1;
	if(p_remain < 0)
		goto Exit;
	else if(p_remain > 0 && p_remain < 100){
		/* Distribution remaining percent to the
		 * best route in the r_data list
		 */
		for (i = 0; i < num_routes; i++){
			if(r_data[i]->percent == 0
				&& !r_data[i]->dup
				&& r_data[i]->ls_enabled) {

				if(r_data[i]->active
					&& r_data[i]->weight >= t_weight
					&& r_data[i]->type <= t_type){
					t_weight = r_data[i]->weight;
					t_type = r_data[i]->type;
					idx = i;
				}
			}
		}
	}

	if(idx >= 0){
		r_data[idx]->percent = p_remain;
		OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
			"Load distributed to GUID: 0x%016"
			PRIx64 " Percent:%d\n",
			cl_ntoh64(r_data[idx]->guid),
			r_data[idx]->percent);
	}
	
	osm_select_load_share_route(p_subn, prefix, num_routes, r_rtr);

	if(*r_rtr)
		goto Exit;
route_select:
	idx = -1;
	t_weight = 0;
	t_type = ROUTE_TYPE_DYNAMIC;

	 for (i = 0; i < num_routes; i++){
		if(r_data[i]->active
			&&r_data[i]->weight >= t_weight
			&& r_data[i]->type <= t_type){
			t_weight = r_data[i]->weight;
			t_type = r_data[i]->type;
			idx = i;
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG, "Route Selected"
				" GUID: 0x%016" PRIx64 " Type: %d\n",
				cl_ntoh64(r_data[i]->guid), r_data[i]->type);
		}
	}

	if(idx >= 0) {
		p_item = cl_qmap_get(&p_subn->rtr_guid_tbl, r_data[idx]->guid);
		if (p_item == cl_qmap_end(&p_subn->rtr_guid_tbl)) {
			if(r_data[idx]->type == ROUTE_TYPE_DYNAMIC){
				OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C03: No"
					" router found with guid 0x%016"
					PRIx64 "\n",
					cl_ntoh64(r_data[idx]->guid));
			}
			else {
				OSM_LOG(p_osm_log, OSM_LOG_ERROR, "ERR 9C04: No"
					" router found with guid 0x%016" PRIx64
					" May be invalid GUID in the static route rule\n",
					cl_ntoh64(r_data[idx]->guid));
			}
		}
		else {
			*r_rtr = (osm_router_t *)p_item;
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Selected  GUID: 0x%016" PRIx64 ""
				 " Type:%d \n",
				cl_ntoh64(r_data[idx]->guid), r_data[idx]->type);
		}
	}
	else {
		OSM_LOG(p_osm_log, OSM_LOG_INFO,
			" No route for Prefix: 0x%016" PRIx64 "\n",
			cl_ntoh64(prefix));
	}
Exit:
	/* Free r_data elements */
	for(i = 0; i < num_routes; i++){
		if(r_data[i]) {
			free(r_data[i]);
			r_data[i] = NULL;
		}
	}
}
/***************************************************
 ***************************************************/

void pr_rcv_filter_routers(IN const osm_subn_t *p_subn,
				IN const cl_qlist_t * p_list,
				OUT osm_sa_item_t ** p_pr_item)
{
	uint64_t prefix;
	cl_list_item_t *qlist_item;
	osm_router_t *p_rtr = NULL;
	cl_list_t rtr_list;
	const ib_path_rec_t *p_pr = NULL;
	const osm_port_t *p_port = NULL;
	const osm_node_t *p_node = NULL;
	osm_log_t *p_osm_log = &p_subn->p_osm->log;
	boolean_t found = false;

	OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
		"Filter Router Path Record Count:%d\n",
		cl_qlist_count(p_list));

	pthread_mutex_lock(&mutex);

	cl_list_construct(&rtr_list);
	cl_list_init(&rtr_list, 2);

	for (qlist_item = cl_qlist_head(p_list);
		qlist_item != cl_qlist_end(p_list);
		qlist_item = cl_qlist_next(qlist_item)) {
		p_pr = &(((osm_sa_item_t *)qlist_item)->resp.path_rec);
		p_port = osm_get_port_by_lid(p_subn, p_pr->dlid);
		p_node = osm_port_get_node(p_port);
		if (osm_node_get_type(p_node) == IB_NODE_TYPE_ROUTER
			&& p_node->router != NULL) 
			cl_list_insert_tail(&rtr_list, p_node->router);
		else
			continue;
	}

	if(cl_list_count(&rtr_list) == 0)
		goto Exit;

	prefix = ib_gid_get_subnet_prefix(&p_pr->dgid);
	osm_resolve_route(p_subn, prefix,
				&rtr_list, &p_rtr);
	if(!p_rtr){
		OSM_LOG(p_osm_log, OSM_LOG_INFO,
			"No route for Prefix: 0x%016" PRIx64 "\n",
			cl_ntoh64(prefix));
		goto Exit;
	}

	for (qlist_item = cl_qlist_head(p_list);
		qlist_item != cl_qlist_end(p_list);
		qlist_item = cl_qlist_next(qlist_item)) {

		p_pr = &(((osm_sa_item_t *)qlist_item)->resp.path_rec);
		if(p_pr->dlid == osm_port_get_base_lid(p_rtr->p_port)) {
			OSM_LOG(p_osm_log, OSM_LOG_DEBUG,
				"Selected Prefix 0x%" PRIx64 ": DGUID "
				"0x%" PRIx64 " DLID %u\n",
				cl_ntoh64(p_pr->dgid.unicast.prefix),
				cl_ntoh64(p_pr->dgid.unicast.interface_id),
				cl_ntoh16(p_pr->dlid));

			*p_pr_item = (osm_sa_item_t *)qlist_item;
			found = true;
			break;
		}
	}
	
	if(!found){
		OSM_LOG(p_osm_log, OSM_LOG_INFO,
			"No route for Prefix: 0x%016" PRIx64
			" via router GUID:  0x%016" PRIx64 "\n",
			cl_ntoh64(prefix),
			cl_ntoh64(p_rtr->p_port->guid));
	}
	
Exit:
	cl_list_remove_all(&rtr_list);
	cl_list_destroy(&rtr_list);
	pthread_mutex_unlock(&mutex);
}
