/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Routing chain manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_RCH_MGR_C
#include <opensm/osm_switch.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing_chain.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_mcast_mgr.h>

#define MAX_DUMP_PREFIX_LEN 20
#define MAX_RCH_INFO_LINE_LEN 100
#define MAX_RCH_INFO_LEN ((MAX_ENGINES_NUMBER + 1) * MAX_RCH_INFO_LINE_LEN)
extern int parse_rch_policy(struct osm_opensm *p_osm);

const osm_physp_t *INVALID = (osm_physp_t *) - 1;
#ifdef RCH_DEBUG
static void print_lft(osm_log_t * p_log, uint8_t * p_lft, uint64_t size)
{
	uint64_t i;
	uint8_t *port;

	port = p_lft;
	for (i = 0; i <= size; i++) {
		if (*port != 255)
			OSM_LOG(p_log, OSM_LOG_DEBUG, "LID %" PRIu64 ": %"
				PRIu8 "\n", i, *port);
		port++;
	}
}

static void print_sw_lft(struct osm_opensm *p_osm, IN osm_switch_t * p_sw)
{
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Printing lft for SW (%p) 0x%"
		PRIx64 "\n", p_sw,
		cl_ntoh64(p_sw->p_node->node_info.node_guid));
	print_lft(p_osm->sm.p_log, p_sw->new_lft, p_sw->lft_size - 1);
}

static void osm_print_all_lfts(IN struct osm_opensm *p_osm, IN struct osm_subn *p_subn)
{
	cl_map_item_t *p_sw;
	cl_qmap_t *sw_table = &p_subn->sw_guid_tbl;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Printing lft for all SW's\n");
	p_sw = cl_qmap_head(sw_table);
	while (p_sw != cl_qmap_end(sw_table)) {
		print_sw_lft(p_osm, (osm_switch_t*)p_sw);
		p_sw = cl_qmap_next(p_sw);
	}
}

static void print_hop(struct osm_opensm *p_osm, IN osm_switch_t * p_sw,
		      uint16_t lids)
{
	uint64_t i;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Printing min hops (lids = %"
		PRIu16 ") for SW (%p) 0x%" PRIx64 "\n", lids, p_sw,
		cl_ntoh64(p_sw->p_node->node_info.node_guid));
	for (i = 0; i < lids; i++) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "LID %" PRIu64
			" hops: %" PRIu8 "\n", i,
			osm_switch_get_least_hops(p_sw, i));
	}
}

static void osm_print_all_sw_hops(IN struct osm_opensm *p_osm,
			   IN struct osm_subn *p_subn)
{
	osm_switch_t *p_sw;
	cl_map_item_t *p_item;
	cl_qmap_t *sw_table = &p_subn->sw_guid_tbl;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Printing min hops for "
		"all SW's\n");
	p_item = cl_qmap_head(sw_table);
	while (p_item != cl_qmap_end(sw_table)) {
		p_sw = (osm_switch_t*)p_item;
		print_hop(p_osm, p_sw, p_sw->max_lid_ho);
		p_item = cl_qmap_next(p_item);
	}
}
#endif

static void merge_ar_lfts(osm_log_t * p_log, osm_switch_t * p_from_sw, osm_switch_t * p_to_sw,
			  osm_subn_t * from_subn, osm_subn_t * to_subn)
{
	osm_node_t  *p_from_node, *p_to_node;

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Merging AR LFT Tables for switch 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_from_sw->p_node)));

	osm_ar_data_copy(to_subn, from_subn, p_to_sw, p_from_sw);

	/* copy extended node info */
	p_to_node = p_to_sw->p_node;
	p_from_node = p_from_sw->p_node;
	p_to_node->ext_node_info = p_from_node->ext_node_info;
	memcpy(&p_to_node->new_ext_node_info, &p_from_node->new_ext_node_info,
	       sizeof(p_to_node->new_ext_node_info));
	p_to_node->ext_node_info_updated = p_from_node->ext_node_info_updated;
}

static void merge_lfts(osm_log_t * p_log, osm_switch_t * p_from_sw,
		       osm_switch_t * p_to_sw)
{
	uint64_t i;
	uint8_t *from_port;
	uint8_t *to_port;
	boolean_t b;
	uint8_t *from_lft;
	uint8_t *to_lft;
	uint16_t max_lid_from;

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Merging LFT Tables\n");
	from_lft = p_from_sw->new_lft;
	to_lft = p_to_sw->new_lft;
	max_lid_from = p_from_sw->max_lid_ho;
	from_port = from_lft;
	to_port = to_lft;
	for (i = 0; i <= max_lid_from; i++) {
		b = (*from_port != *to_port);
		if ((*from_port != OSM_NO_PATH) && (b)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG, "changing lid %" PRIu64
				",new value: %" PRIu8 "\n", i, *from_port);
			*to_port = *from_port;
			/* just update the ar table static port */
			if (p_to_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED) {
				osm_ar_lft_update_port(&p_to_sw->ar_data, i,
						       ADV_ROUTING_ALL_PLFTS, *from_port);
				OSM_LOG(p_log, OSM_LOG_DEBUG, "AR table changing lid %" PRIu64
					",new value: %" PRIu8 "\n", i, *from_port);
			}
		}
		from_port++;
		to_port++;
	}
}

static void merge_subn_lfts(osm_log_t * p_log, osm_subn_t * from_subn,
			    osm_subn_t * to_subn)
{
	osm_switch_t *p_from_sw;
	osm_switch_t *p_to_sw;
	cl_map_item_t *p_from_item;
	cl_map_item_t *p_to_item;
	cl_qmap_t *from_sw_tbl;
	cl_qmap_t *to_sw_tbl;
	ib_net64_t node_guid;
	uint16_t lft_size;

	from_sw_tbl = &from_subn->sw_guid_tbl;
	to_sw_tbl = &to_subn->sw_guid_tbl;
	p_from_item = cl_qmap_head(from_sw_tbl);
	while (p_from_item != cl_qmap_end(from_sw_tbl)) {
		p_from_sw = (osm_switch_t *)p_from_item;
		node_guid = osm_node_get_node_guid(p_from_sw->p_node);
		p_to_item = cl_qmap_get(to_sw_tbl, node_guid);
		if (p_to_item == cl_qmap_end(to_sw_tbl))
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Merge of SW 0x%" PRIx64
				" failed - SW not found in target subnet\n",
				cl_ntoh64(node_guid));
		else {
			p_to_sw = (osm_switch_t *)p_to_item;
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Merging Switch (%p) 0x%"
				PRIx64 "\n", p_to_sw, cl_ntoh64(node_guid));
			if (!p_to_sw->new_lft) {
				lft_size =
				    (from_subn->max_lid / IB_SMP_DATA_SIZE + 1)
				    * IB_SMP_DATA_SIZE;
				p_to_sw->new_lft = malloc(lft_size);
				memset(p_to_sw->new_lft, OSM_NO_PATH, lft_size);
			}
			if (p_to_sw->max_lid_ho < p_from_sw->max_lid_ho)
				p_to_sw->max_lid_ho = p_from_sw->max_lid_ho;
			/* currently we support only one AR, either 'to' or 'from' */
			if (p_from_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED)
				merge_ar_lfts(p_log, p_from_sw, p_to_sw, from_subn, to_subn);
			else
				merge_lfts(p_log, p_from_sw, p_to_sw);
		}
		p_from_item = cl_qmap_next(p_from_item);
	}

}

static void merge_subn_ar_configuration(osm_log_t * p_log, osm_subn_t * from_subn,
					osm_subn_t * to_subn)
{
	if (from_subn->ar_conf_params.ar_mode > to_subn->ar_conf_params.ar_mode) {
		to_subn->ar_conf_params.ar_mode = from_subn->ar_conf_params.ar_mode;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Changed ar_mode in main subnet to %d\n",
			to_subn->ar_conf_params.ar_mode);
	}
	if (from_subn->ar_conf_params.shield_mode > to_subn->ar_conf_params.shield_mode) {
		to_subn->ar_conf_params.shield_mode = from_subn->ar_conf_params.shield_mode;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Changed shield_mode in main subnet to %d\n",
			to_subn->ar_conf_params.shield_mode);
	}
}

static void print_engine(osm_opensm_t * p_osm, osm_routing_chain_t * p_engine)
{
	int fallback_id;
	osm_routing_chain_t *p_fallback;

	if (p_engine->fallback_engine) {
		p_fallback = (osm_routing_chain_t *) p_engine->fallback_engine;
		fallback_id = (int)p_fallback->id;
	} else
		fallback_id = -1;
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
		"\nRouting Engine: %zd\nUse: %s\n"
		"Algorithm: %s\nTopo_id: %" PRIu64 "\nConfig File: %.200s\n"
		"Fallback: %d\n", p_engine->id, p_engine->use, p_engine->engine,
		p_engine->topo_id, p_engine->config_file, fallback_id);
}

static inline boolean_t is_id_free(struct osm_opensm *p_osm, size_t id)
{
	boolean_t is_free;

	if ((id > 0) && (id <= MAX_ENGINE_ID))
		is_free = !(p_osm->sm.routing_chain_mgr.used_ids[id] ||
			    p_osm->sm.routing_chain_mgr.fallback_ids[id]);
	else {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3914: Illegal"
			" engine id (%zd) - Valid values are 1-%d\n",
			id, MAX_ENGINE_ID);
		is_free = FALSE;
	}
	return is_free;
}

osm_routing_chain_t *osm_rch_get_engine(struct osm_opensm *p_osm, size_t id)
{
	osm_routing_chain_mgr_t *p_mgr = &p_osm->sm.routing_chain_mgr;
	osm_routing_chain_t *p_item;
	boolean_t found = FALSE;

	if (is_id_free(p_osm, id))
		goto EXIT;

	p_item = (osm_routing_chain_t *) cl_qlist_head(&p_mgr->engine_list);
	while ((cl_list_item_t *) p_item != cl_qlist_end(&p_mgr->engine_list)
	       && !found) {
		if (p_item->id == id)
			found = TRUE;
		else
			p_item =
			    (osm_routing_chain_t *)
			    cl_qlist_next((cl_list_item_t *) p_item);
	}
	if (found)
		return p_item;
	else
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3900: ID %zd "
			"marked used but engine not found!!!\n", id);
EXIT:
	return NULL;
}

static osm_routing_chain_t *create_engine(osm_opensm_t * p_osm, size_t id,
					  char *use, char *engine,
					  uint64_t topo_id,
					  struct osm_subn *p_subn,
					  uint8_t path_bit, char *config_file)
{
	osm_routing_chain_t *new_engine;

	new_engine = calloc(1, sizeof(osm_routing_chain_t));

	if (!new_engine) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3901: Failed "
			"allocating new routing engine: %zd\n", id);
		return NULL;
	}
	new_engine->id = id;
	new_engine->use = strdup(use);
	new_engine->engine = strdup(engine);
	new_engine->config_file = strdup(config_file);
	new_engine->path_bit = path_bit;
	new_engine->subn = p_subn;
	new_engine->topo_id = topo_id;
	new_engine->fallback_engine = NULL;
	new_engine->used_fallback = FALSE;
	p_subn->routing_chain_engine_id = id;
	return new_engine;
}

static void destroy_engine(osm_routing_chain_t * p_engine)
{
	free(p_engine->use);
	free(p_engine->engine);
	free(p_engine->config_file);
	if (p_engine->fallback_engine)
		destroy_engine((osm_routing_chain_t *) p_engine->fallback_engine);
	free(p_engine);
}

static void rch_box_clear_group(osm_mgrp_t * p_mgrp) {

	cl_map_item_t *p_item;

	OSM_ASSERT(p_mgrp);

	while (cl_qmap_count(&p_mgrp->mcm_port_tbl)) {
		p_item = cl_qmap_head(&p_mgrp->mcm_port_tbl);
		cl_qmap_remove_item(&p_mgrp->mcm_port_tbl, p_item);
		osm_mcm_port_delete((osm_mcm_port_t *)p_item);
	}
}

typedef struct osm_rch_mgrp_box {
	cl_map_item_t map_item;
	osm_mgrp_box_t * mbox;
	ib_net64_t prim_bs_guid;
	uint64_t topology_id;
	osm_mgrp_t members_group;
	osm_mgrp_t borders_group;
} osm_rch_box_t;

static void rch_box_remove_members(osm_rch_box_t * p_rch_box) {
	rch_box_clear_group(&p_rch_box->members_group);
	rch_box_clear_group(&p_rch_box->borders_group);
}

static void rch_box_delete(osm_sm_t * sm, osm_rch_box_t * p_rch_box)
{
	osm_topology_t * p_topology;

	p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
					   p_rch_box->topology_id);

	osm_mtree_destroy(p_rch_box->mbox->root);

	rch_box_remove_members(p_rch_box);
	if (p_topology)
		mcast_mgr_decrease_mtree_root_weight(p_topology->p_subn,
						     p_rch_box->mbox);

	free(p_rch_box->mbox);
	free(p_rch_box);
}

static void rch_rbox_map_destroy(osm_sm_t * sm, cl_qmap_t * p_rbox_map) {
	cl_map_item_t * p_item;

	while (cl_qmap_count(p_rbox_map)) {
		p_item = cl_qmap_head(p_rbox_map);
		cl_qmap_remove_item(p_rbox_map, p_item);
		rch_box_delete(sm, (osm_rch_box_t *)p_item);
	}
}

void osm_rch_rbox_map_destroy(IN osm_routing_chain_mgr_t *p_mgr, uint16_t mlid)
{
	cl_qmap_t *p_rbox_map;

	OSM_ASSERT(mlid >= IB_LID_MCAST_START_HO &&
		   mlid <= IB_LID_MCAST_END_HO);

	p_rbox_map = &p_mgr->p_rch_rboxes[mlid - IB_LID_MCAST_START_HO];
	rch_rbox_map_destroy(&p_mgr->p_osm->sm, p_rbox_map);
}

/* destroy all rbox objects for all mlids and topologies */
static void rch_destroy_rbox_maps(IN osm_routing_chain_mgr_t *p_mgr)
{
	unsigned i;

	for (i = 0; i < IB_LID_MCAST_END_HO - IB_LID_MCAST_START_HO + 1; i++)
		rch_rbox_map_destroy(&p_mgr->p_osm->sm,
				     &p_mgr->p_rch_rboxes[i]);
}

/* destroy rbox objects for a certain topology */
static void rch_destroy_topo_rbox_items(IN osm_routing_chain_mgr_t * p_mgr,
					IN uint64_t topo_id)
{
	cl_map_item_t *p_item;
	unsigned i;

	for (i = 0; i < IB_LID_MCAST_END_HO - IB_LID_MCAST_START_HO + 1; i++) {
		p_item = cl_qmap_get(&p_mgr->p_rch_rboxes[i], topo_id);
		if (p_item != cl_qmap_end(&p_mgr->p_rch_rboxes[i])) {
			cl_qmap_remove_item(&p_mgr->p_rch_rboxes[i], p_item);
			rch_box_delete(&p_mgr->p_osm->sm,
				       (osm_rch_box_t *)p_item);
		}
	}
}

static void osm_rch_mgr_clean_main_topo_engines(osm_routing_chain_mgr_t * p_mgr)
{
	osm_subn_t *p_subn = &p_mgr->p_osm->subn;
	struct osm_routing_engine *backup_list = p_subn->routing_engine_list;
	struct osm_routing_engine *backup_last = p_subn->last_routing_engine;
	struct osm_routing_engine *backup_used = p_subn->routing_engine_used;
	struct osm_routing_engine *backup_def = p_subn->default_routing_engine;

	/* Clean main topology routing engines information */
	p_subn->routing_engine_list = p_mgr->main_topo_routing_engine_list;
	p_subn->last_routing_engine = p_mgr->main_topo_last_routing_engine;
	p_subn->routing_engine_used = p_mgr->main_topo_routing_engine_used;
	p_subn->default_routing_engine = p_mgr->main_topo_default_routing_engine;
	p_mgr->main_topo_routing_engine_list = NULL;
	p_mgr->main_topo_last_routing_engine = NULL;
	p_mgr->main_topo_routing_engine_used = NULL;
	p_mgr->main_topo_default_routing_engine = NULL;
	osm_destroy_routing_engines(p_subn);
	osm_destroy_default_routing_engine(p_subn);

	/* Restore subnets routing engines inforamtion */
	p_subn->routing_engine_list = backup_list;
	p_subn->last_routing_engine = backup_last;
	p_subn->routing_engine_used = backup_used;
	p_subn->default_routing_engine = backup_def;
}

static void osm_rch_mgr_clean(IN osm_routing_chain_mgr_t * p_mgr)
{
	cl_list_item_t *p_item;
	osm_routing_chain_t *p_chain;
	unsigned i;

	p_mgr->is_itr = FALSE;

	while (cl_qlist_count(&p_mgr->engine_list) > 0) {
		p_item = cl_qlist_remove_head(&p_mgr->engine_list);
		p_chain = (osm_routing_chain_t *) p_item;

		/* clean rbox objects for removed topologies */
		if (!osm_topo_mgr_get_topo(p_mgr->p_osm, p_chain->topo_id))
			rch_destroy_topo_rbox_items(p_mgr, p_chain->topo_id);

		destroy_engine(p_chain);
	}

	for (i = 0; i <= MAX_ENGINES_NUMBER; i++) {
		p_mgr->used_ids[i] = FALSE;
		p_mgr->fallback_ids[i] = FALSE;
	}

	osm_rch_mgr_clean_main_topo_engines(p_mgr);

}

void osm_rch_mgr_destroy(IN osm_routing_chain_mgr_t * p_mgr)
{
	osm_rch_mgr_clean(p_mgr);
	rch_destroy_rbox_maps(p_mgr);
	free(p_mgr->p_rch_rboxes);
}

boolean_t osm_rch_add_engine(IN osm_opensm_t * p_osm, IN size_t id,
			     IN char *use, IN char *engine, IN uint64_t topo_id,
			     IN struct osm_subn *p_subn, IN uint8_t path_bit,
			     IN char *config_file)
{
	osm_routing_chain_t *new_engine;
	boolean_t res = FALSE;

	if (cl_qlist_count(&p_osm->sm.routing_chain_mgr.engine_list) >=
	    MAX_ENGINES_NUMBER) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3915: Could not"
			" add engine %zd (use: %s) - Engines list reached "
			" maximum length (%d)\n", id, use, MAX_ENGINES_NUMBER);
		goto EXIT;
	}

	if (!is_id_free(p_osm, id)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3902: Could not"
			" add engine %zd (use: %s) - ID Illegal or Used!\n",
			id, use);
		goto EXIT;
	}

	new_engine = create_engine(p_osm, id, use, engine, topo_id, p_subn,
				   path_bit, config_file);

	if (!new_engine)
		goto EXIT;

	cl_qlist_insert_tail(&p_osm->sm.routing_chain_mgr.engine_list,
			     (cl_list_item_t *) new_engine);
	p_osm->sm.routing_chain_mgr.used_ids[id] = TRUE;
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "New Routing Engine added to"
		" Routing Chain:\n");
	print_engine(p_osm, new_engine);
	res = TRUE;
EXIT:
	return res;
}

boolean_t osm_rch_set_fallback_engine(IN struct osm_opensm *p_osm,
				      IN size_t target_engine_id, IN size_t id,
				      IN char *use, IN char *engine,
				      IN uint64_t topo_id,
				      IN struct osm_subn *p_subn,
				      IN uint8_t path_bit, IN char *config_file)
{
	boolean_t res = FALSE;
	osm_routing_chain_t *fallback_engine;
	osm_routing_chain_t *target_engine;

	if (!is_id_free(p_osm, id)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3903: Could not"
			" add fallback engine %zd (%s) - ID Illegal or Used\n"
			, id, use);
		goto EXIT;
	}

	target_engine = osm_rch_get_engine(p_osm, target_engine_id);

	if (!target_engine) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3904: Target "
			"Engine (%zd) for fallback setup not found, fallback "
			"engine %zd ignored!!!\n", target_engine_id, id);
		goto EXIT;
	}

	if (target_engine->fallback_engine) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 390B: Target "
			"Engine (%zd) already set with a fallback engine (%zd),"
			" engine %zd ignored!!!\n", target_engine_id,
			target_engine->fallback_engine->id, id);
		goto EXIT;
	}

	fallback_engine = create_engine(p_osm, id, use, engine, topo_id, p_subn,
					path_bit, config_file);

	if (!fallback_engine)
		goto EXIT;

	target_engine->fallback_engine =
	    (struct osm_routing_chain *)fallback_engine;
	p_osm->sm.routing_chain_mgr.fallback_ids[id] = TRUE;
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Routing engine %zd Created and"
		" set as fallback for routing engine %zd\n",
		fallback_engine->id, target_engine->id);
	res = TRUE;
EXIT:
	return res;
}

static void set_engine_main_subn(osm_routing_chain_mgr_t * p_mgr,
				 osm_routing_chain_t * engine,
				 struct osm_routing_engine **backup_list,
				 struct osm_routing_engine **backup_last,
				 struct osm_routing_engine **backup_used,
				 struct osm_routing_engine **backup_def,
				 boolean_t *backup_fallback)
{
	engine->subn->path_bit = engine->path_bit;
	*backup_list = engine->subn->routing_engine_list;
	*backup_last = engine->subn->last_routing_engine;
	*backup_used = engine->subn->routing_engine_used;
	*backup_def = engine->subn->default_routing_engine;
	*backup_fallback = engine->subn->no_fallback_routing_engine;

	/*
	 * Verify we dont already keep information about the routing engine
	 * used for the main subnet.
	 */
	OSM_ASSERT(p_mgr->main_topo_routing_engine_list == NULL);
	OSM_ASSERT(p_mgr->main_topo_last_routing_engine == NULL);
	OSM_ASSERT(p_mgr->main_topo_routing_engine_used == NULL);
	OSM_ASSERT(p_mgr->main_topo_default_routing_engine == NULL);

	engine->subn->routing_engine_list = NULL;
	engine->subn->last_routing_engine = NULL;
	engine->subn->routing_engine_used = NULL;
	engine->subn->default_routing_engine = NULL;
	osm_setup_routing_engines(engine->subn, engine->engine);
}

static void unset_engine_main_subn(osm_routing_chain_mgr_t * p_mgr,
				   osm_routing_chain_t * engine,
				   struct osm_routing_engine *backup_list,
				   struct osm_routing_engine *backup_last,
				   struct osm_routing_engine *backup_used,
				   struct osm_routing_engine *backup_def,
				   boolean_t backup_fallback)
{
	osm_subn_t *p_subn = engine->subn;

	p_subn->path_bit = OSM_SUBNET_DEFAULT_PATH_BITS;

	/* save the routing engine that was used for the main topology */
	p_mgr->main_topo_routing_engine_list = engine->subn->routing_engine_list;
	p_mgr->main_topo_last_routing_engine = engine->subn->last_routing_engine;
	p_mgr->main_topo_routing_engine_used = engine->subn->routing_engine_used;
	p_mgr->main_topo_default_routing_engine = engine->subn->default_routing_engine;

	p_subn->routing_engine_list = backup_list;
	p_subn->last_routing_engine = backup_last;
	p_subn->routing_engine_used = backup_used;
	p_subn->default_routing_engine = backup_def;
	p_subn->no_fallback_routing_engine = backup_fallback;
}

static boolean_t set_engine_subn(osm_opensm_t *p_osm,
				 osm_routing_chain_t *p_engine)
{
	p_engine->subn->path_bit = p_engine->path_bit;
	/* calculate_missing_routes should applied only on main subnet */
	/* to avoid potential credit loops */
	p_engine->subn->opt.calc_missing_routes = FALSE;
	if (strcmp(NONE_STR, p_engine->config_file)) {
		if (osm_subn_parse_conf_file(p_engine->config_file,
					     &p_engine->subn->opt,
					     TRUE, FALSE)) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3917: "
				"Failed to parse engine subn opt file (%s)\n",
				p_engine->config_file);
			return FALSE;
		}
	}
	osm_setup_routing_engines(p_engine->subn, p_engine->engine);
	if (!p_engine->subn->routing_engine_list) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3930: "
			"Setup or routing engine failed\n");
		return FALSE;
	}
	osm_setup_advanced_routing_engine(p_engine->subn,
					  p_engine->subn->opt.adv_routing_engine_name);
	return TRUE;
}
#ifdef RCH_DEBUG
/* DEBUG METHDOD - FOR INTERNAL TESTING - DO NOT DELETE */
static void print_subn(struct osm_opensm *p_osm, osm_subn_t * p_subn,
		       char *name)
{
	osm_node_t *p_node;
	cl_map_item_t *p_item;
	uint8_t i, port;
	osm_physp_t *p_physp;
	ib_net64_t guid;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "\nPrinting subn of %s:\n"
		"ports: %d\nnodes: %d\nswtiches: %d\n", name,
		cl_qmap_count(&p_subn->port_guid_tbl),
		cl_qmap_count(&p_subn->node_guid_tbl),
		cl_qmap_count(&p_subn->sw_guid_tbl));

	p_item = cl_qmap_head(&p_subn->node_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->node_guid_tbl)) {
		p_node = (osm_node_t *)p_item;
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "\nNode: 0x%" PRIx64
			"(%p), SW addr: %p, Lid: %" PRIu16 "\n",
			cl_ntoh64(p_node->node_info.node_guid), p_node,
			p_node->sw, cl_ntoh16(osm_node_get_base_lid(p_node, 1)));
		for (i = 0; i < p_node->physp_tbl_size; i++) {
			p_physp = &p_node->physp_table[i];
			if (p_physp->p_remote_physp) {
				if (p_physp->p_remote_physp !=
				    (osm_physp_t *) INVALID) {
					guid =
					    p_physp->p_remote_physp->port_guid;
					port =
					    p_physp->p_remote_physp->port_num;
					OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
						"\nPort %d (%p) (node: %p): r-guid: 0x%"
						PRIx64 " (%p), r-port: %d\n", i,
						p_physp, p_physp->p_node,
						cl_ntoh64(guid),
						p_physp->p_remote_physp, port);
				} else
					OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
						"\nPort %d INVALID", i);

			} else
				OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
					"\nPort %d NILL (%p)", i, p_physp);
		}
		p_item = cl_qmap_next(p_item);
	}
}
#endif

/*
 * Get port and check if this port require ITR routing.
 *
 * Decision is taken base on topology list of the port:
 *
 * If topology list is empty the port is only on main subnet and doesn't
 * required ITR.
 *
 * IF topology list is longer than 1, this port must be border sw and therefore
 * doesn't require ITR.
 *
 * Only internal sw of ITR topology require ITR so Conditions for port to require ITR:
 *
 *	1) Topology list count is exactly 1.
 *	2) Routing engine of the topology implement ITR methods
 *	3) Port is not a border switch of the topology
 */
static boolean_t itr_required(IN osm_port_t *p_port,
			      OUT osm_topology_t **p_itr_topo,
			      OUT boolean_t *is_border)
{
	boolean_t result = FALSE;
	osm_topology_t *p_topo = NULL;
	cl_list_iterator_t iter;
	boolean_t is_border_sw = FALSE;
	ib_net64_t guid;


	if (!cl_list_count(&p_port->topology_list))
		goto Exit;

	iter = cl_list_head(&p_port->topology_list);
	guid = osm_node_get_node_guid(p_port->p_node);

	while (iter != cl_list_end(&p_port->topology_list)) {
		p_topo = osm_port_get_topo_from_list_iter(&iter);

		iter = cl_list_next(iter);

		if (!osm_topo_is_engine_itr(p_topo))
			continue;

		if (!osm_topo_is_border_sw(p_topo, guid))
			result = TRUE;
		else
			is_border_sw = TRUE;
		break;

	}
Exit:
	/* return the ITR topo  */
	if (p_itr_topo)
		(*p_itr_topo) = (result || is_border_sw) ? p_topo: NULL;

	if (is_border)
		*is_border = is_border_sw;

	return result;
}

/*
 * Get sw guid, border switch lid and destination lid.
 * Set the outgoing port to destination lid equal to outgoing port
 * to border switch. If no border switch (bs_lid = FFFF), set NO_PATH
 */
static void itr_set_sw_port_for_lid(uint64_t sw_guid, osm_switch_t *p_sw,
				    ib_net16_t bs_lid,
				    ib_net16_t dest_lid,
				    osm_subn_t *p_main_subn)
{
	uint8_t port_num = OSM_NO_PATH;

	if (bs_lid != 0xFFFF) {
		bs_lid = cl_ntoh16(bs_lid);

		OSM_LOG(&p_main_subn->p_osm->log,
			OSM_LOG_DEBUG, "SW 0x%" PRIx64
			" Routing to dlid %" PRIu16 " old port (%u)"
			" via border switch %" PRIu16 " new port (%u)\n",
			cl_ntoh64(sw_guid), dest_lid, p_sw->new_lft[dest_lid],
			bs_lid, p_sw->new_lft[bs_lid]);

		port_num = p_sw->new_lft[bs_lid];
	}
	p_sw->new_lft[dest_lid] = port_num;
}

static void itr_route_sw_to_dest(osm_port_t * p_sw_port, osm_port_t *p_dest_port,
				 osm_port_t * p_bs_port, osm_subn_t *p_main_subn,
				 uint16_t dest_lid_ho)
{
	osm_switch_t  *p_sw;
	osm_switch_t  *p_bs_sw;
	ib_net16_t bs_lid;
	ib_net64_t sw_guid;

	if (p_bs_port) {
		p_bs_sw = p_bs_port->p_node->sw;
		if (p_bs_sw->new_lft[dest_lid_ho] == OSM_NO_PATH)
			bs_lid = 0xFFFF;
		else
			bs_lid = p_bs_port->lid;
	} else
		bs_lid = 0xFFFF;

	sw_guid = osm_node_get_node_guid(p_sw_port->p_node);

	p_sw = p_sw_port->p_node->sw;

	itr_set_sw_port_for_lid(sw_guid, p_sw, bs_lid, dest_lid_ho, p_main_subn);
}

/* get border switch port for given port
 * If main subnet given, return port object belonging to main subnet
 * If not given, return port object belonging to topology subnet
 */
static osm_port_t * itr_get_bs_by_port(IN osm_port_t *p_port,
				       osm_subn_t *p_main_subn)
{
	struct osm_routing_engine *r;
	ib_net64_t guid, bs_guid;
	osm_port_t *p_res = NULL, *p_subn_port;
	osm_node_t *p_node;
	osm_topology_t *p_topo;
	osm_subn_t *p_subn;

	OSM_ASSERT(p_port);

	p_subn_port = osm_get_port_from_itr_topo(p_port, &p_topo);

	if (!p_subn_port)
		goto Exit;

	if (osm_node_get_type(p_subn_port->p_node) != IB_NODE_TYPE_SWITCH) {
		p_node = p_subn_port->p_physp->p_remote_physp->p_node;
		guid = osm_node_get_node_guid(p_node);
	} else {
		/* this is border switch of the topoliogy*/
		p_node = p_subn_port->p_node;

		guid = osm_node_get_node_guid(p_node);

	        if (osm_topo_is_border_sw(p_topo, guid)) {
			p_res = p_subn_port;
			goto Exit;
		}
	}
	r = p_topo->p_subn->routing_engine_used;

	if (r && r->ITR_get_bs) {
		if (!r->ITR_get_bs(r->context, guid, &bs_guid)) {
			if (p_main_subn)
				p_subn = p_main_subn;
			else
				p_subn = p_topo->p_subn;
			p_res = osm_get_port_by_guid(p_subn,
						     bs_guid);
		}
	}
Exit:
	return p_res;
}

/* Route from source  sw port to destination port via source switch topology */
static void itr_to_port(osm_port_t *p_sw_port, osm_topology_t *p_sw_topo,
			osm_port_t *p_dest_port, osm_subn_t *p_main_subn,
			uint16_t lid_ho)
{
	osm_topology_t *p_topo;
	osm_subn_t *p_subn;
	osm_port_t *port_to_check, *p_itr_port;
	osm_port_t  *p_port, *p_bs_port;
	/* source switch is belonging to ITR topology */
	if (p_sw_topo != NULL) {
		port_to_check = p_dest_port;
		p_itr_port = p_sw_port;
		p_topo = p_sw_topo;
	} else {
		if (itr_required(p_dest_port, &p_topo, NULL)) {
			port_to_check = p_sw_port;
			p_itr_port = p_dest_port;
		} else
			return;
	}

	p_subn = p_topo->routing_engine->subn;

	p_port = osm_get_port_by_guid(p_subn, port_to_check->guid);

	if (!p_port) {
		/* both ports are on different topologies - set ITR */
			p_bs_port = itr_get_bs_by_port(p_itr_port, p_main_subn);
			itr_route_sw_to_dest(p_sw_port, p_dest_port, p_bs_port,
					     p_main_subn, lid_ho);
	}
}

/* Get port of sw and subnet, set itr from sw to each port in the subn */
static void set_sw_itr(osm_port_t *sw_port, osm_subn_t *p_main_subn,
			    osm_log_t *p_log)
{
	cl_map_item_t *p_item = cl_qmap_head(&p_main_subn->port_guid_tbl);
	osm_topology_t *p_topo = NULL;
	osm_port_t *dest_port;
	boolean_t is_border;
	uint16_t min_lid_ho = 0;
	uint16_t max_lid_ho = 0;
	uint16_t lid_ho = 0;

	itr_required(sw_port, &p_topo, &is_border);

	/* no routing is needed for border switch */
	if (is_border)
		return;

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Routing switch 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(sw_port->p_node)));

	while (p_item != cl_qmap_end(&p_main_subn->port_guid_tbl)) {
		dest_port = (osm_port_t*)p_item;
		if (sw_port->guid != dest_port->guid) {
			osm_port_get_lid_range_ho(dest_port, &min_lid_ho,
						  &max_lid_ho);
			for (lid_ho = min_lid_ho ; lid_ho <= max_lid_ho ; lid_ho++)
				itr_to_port(sw_port, p_topo, dest_port,
					    p_main_subn, lid_ho);
		}
		p_item = cl_qmap_next(p_item);
	}
}

static int  process_itr(osm_subn_t *p_main_subn)
{
	cl_map_item_t *p_item = cl_qmap_head(&p_main_subn->port_guid_tbl);
	osm_port_t *p_port;
	osm_log_t *p_log = &p_main_subn->p_osm->log;

	while (p_item != cl_qmap_end(&p_main_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH)
			set_sw_itr(p_port, p_main_subn, p_log);
		p_item = cl_qmap_next(p_item);
	}

	return 0;
}

static inline boolean_t engine_use_main_subn(struct osm_opensm *p_osm,
					     osm_routing_chain_t * p_engine)
{
	return (&p_osm->subn == p_engine->subn);
}

static void remark_topo_in_sw(struct osm_opensm *p_osm,
			      osm_routing_chain_t * p_engine)
{
	cl_map_item_t *p_item;
	osm_topology_t *p_topo;
	osm_node_t *p_node;
	uint64_t guid;

	p_topo = osm_topo_mgr_get_topo(p_osm, p_engine->topo_id);
	p_item = cl_qmap_head(&p_topo->p_subn->sw_guid_tbl);
	while (p_item != cl_qmap_end(&p_topo->p_subn->sw_guid_tbl)) {
		guid = cl_qmap_key(p_item);
		p_node = osm_get_node_by_guid(p_topo->p_subn, guid);
		OSM_ASSERT(p_node);
		p_node->physp_table[0].topo_id = p_topo->id;
		p_item = cl_qmap_next(p_item);
	}
}

static void print_rch_info(IN osm_opensm_t *p_osm, IN char *comp_suffix,
			   IN osm_log_level_t log_level)
{
	char eng_msg[MAX_RCH_INFO_LINE_LEN];
	char msg[MAX_RCH_INFO_LEN];
	size_t i, sw, nodes;
	osm_routing_chain_t *p_engine;
	osm_routing_chain_mgr_t *p_mgr = &p_osm->sm.routing_chain_mgr;

	if (osm_log_is_active_v2(p_osm->sm.p_log, log_level, FILE_ID)) {
		strcpy(msg, "Routing Chain Info:\n");

		for (i = 0; i <= MAX_ENGINES_NUMBER; i++) {
			if (p_mgr->used_ids[i]) {
				p_engine = osm_rch_get_engine(p_osm, i);
				sw = cl_qmap_count(&p_engine->subn->sw_guid_tbl);
				nodes = cl_qmap_count(&p_engine->subn->node_guid_tbl);
				snprintf(eng_msg, MAX_RCH_INFO_LINE_LEN,
					 "Engine %zd: routing engine: %s, "
					 "Topo: %" PRId64 ", Sw/Nodes: %zd/%zd%s\n",
					 p_engine->id, p_engine->engine,
					 p_engine->topo_id, sw, nodes,
					 comp_suffix);
				strcat(msg, eng_msg);
				comp_suffix = "";
			}
		}
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "%s", msg);
	}
}

static int route_engine(struct osm_ucast_mgr *ucast_mgr,
			osm_routing_chain_t * p_engine)
{
	int failed = -1;
	struct osm_opensm *p_osm = ucast_mgr->p_subn->p_osm;
	struct osm_routing_engine *backup_list, *backup_last,
		*backup_used, *backup_def;
	boolean_t backup_fallback = FALSE;
	char engine_prefix[MAX_DUMP_PREFIX_LEN];
	osm_topology_t *p_topo;
	struct osm_routing_engine *r;
	osm_routing_chain_mgr_t *p_mgr = &ucast_mgr->sm->routing_chain_mgr;

	backup_list = backup_last = backup_used = backup_def = NULL;
	OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Routing engine %zd\n",
		p_engine->id);

	if (engine_use_main_subn(p_osm, p_engine)) {
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Engine %zd set with"
		" main topology\n", p_engine->id);
		set_engine_main_subn(p_mgr, p_engine, &backup_list,
				     &backup_last, &backup_used, &backup_def,
				     &backup_fallback);
	} else {
		ucast_mgr->p_subn = p_engine->subn;
		if (!set_engine_subn(p_osm, p_engine))
			goto EXIT;

		/*
		 * If the first routing engine algorithm on the
		 * engines list is an ITR engine it must have border switches
		 */
		if (p_engine->subn->routing_engine_list->ITR_init) {
			if (!osm_subn_bs_count(p_engine->subn)) {
				OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR, "ERR 391B: "
					"No border switches on subnet, Fail to add "
					"routing engine %zd (topo id: %" PRIu64 "\n",
					p_engine->id, p_engine->topo_id);
				goto EXIT;
			}
		}
	}

	/* This will assure that if routing will fail for this engine, the default
	 * routing will not be used, we would like the failback engine to be used
	 * instead */
	if (p_engine->fallback_engine)
		p_engine->subn->no_fallback_routing_engine = TRUE;

	failed = osm_ucast_mgr_process(ucast_mgr);
	r = ucast_mgr->p_subn->routing_engine_used;
	if (!failed && r && r->ITR_init) {
		if (r->ITR_init(r->context)) {
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR, "ERR 391A: "
				"Engine %zd - ITR init Failed\n", p_engine->id);
			failed = -1;
		} else {
			remark_topo_in_sw(p_osm, p_engine);
			p_mgr->is_itr = TRUE;
		}
	}
	if (failed)
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR, "ERR 390D: Routing "
			"Engine %zd Failed!\n", p_engine->id);

EXIT:
	if (engine_use_main_subn(p_osm, p_engine))
		unset_engine_main_subn(p_mgr, p_engine, backup_list,
				       backup_last, backup_used, backup_def,
				       backup_fallback);
	else {
		ucast_mgr->p_subn = &p_osm->subn;
		if (failed)
			goto END;
		/* Mark enigne in topology */
		p_topo = osm_topo_mgr_get_topo(p_osm, p_engine->topo_id);
		p_topo->routing_engine = p_engine;
		p_topo->p_subn->routing_chain_engine_id = p_engine->id;

		/* Dump engine data */
		memset(engine_prefix, 0, MAX_DUMP_PREFIX_LEN);
		snprintf(engine_prefix, MAX_DUMP_PREFIX_LEN, ".%zd.%s",
			 p_engine->id, p_engine->engine);
		osm_dump_subn_all(p_osm, p_engine->subn, engine_prefix, FALSE);
	}
END:
	return failed;
}

static int route_engine_with_fallback(struct osm_ucast_mgr *ucast_mgr,
				      osm_routing_chain_t * p_engine)
{
	int failed;

	p_engine->used_fallback = FALSE;
	failed = route_engine(ucast_mgr, p_engine);

	if (failed && p_engine->fallback_engine) {
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_INFO,
			"Using Fallback " "Engine %zd\n",
			p_engine->fallback_engine->id);
		p_engine->used_fallback = TRUE;
		if (p_engine->topo_id == p_engine->fallback_engine->topo_id)
			osm_destroy_routing_engines(p_engine->fallback_engine->subn);
		failed = route_engine(ucast_mgr, p_engine->fallback_engine);
	}

	if (failed) {
		if (p_engine->fallback_engine)
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR,
				"ERR 390E: Routing Engine %zd and "
				"fallback engine %zd Failed!\n", p_engine->id,
				p_engine->fallback_engine->id);
		else
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR,
				"ERR 390F: Routing Engine %zd failed!"
				" (No failback)\n", p_engine->id);
		return -1;
	}
	return 0;
}

static osm_routing_chain_t *get_main_engine(struct osm_opensm *p_osm,
					    osm_routing_chain_mgr_t *p_mgr)
{
	size_t i;
	osm_routing_chain_t *p_engine;

	for (i = 1; i <= MAX_ENGINES_NUMBER; i++) {
		if (!p_mgr->used_ids[i])
			continue;
		p_engine = osm_rch_get_engine(p_osm, i);
		if (engine_use_main_subn(p_osm, p_engine))
			return p_engine;
		else
			continue;
	}
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3913: "
		"Can't find main routing engine\n");
	return NULL;
}

/*
 * In the engines list we know that first one use main subnet with topo_id 0
 * (this is ensured by topo parser)
 * This method:
 *
 * 1) Run all engines which are not using main subnet
 * 2) If ITR engine used then build completer topology and set it to the first
 *	engine instead of the main subnet
 * 3) Run first engine
 * 4) Merge all engines lfts
 */
int osm_rch_ucast_process(IN void * context)
{
	struct osm_ucast_mgr *ucast_mgr = (struct osm_ucast_mgr *) context;
	osm_routing_chain_mgr_t *p_mgr = &ucast_mgr->sm->routing_chain_mgr;
	osm_routing_chain_t *p_engine;
	osm_routing_chain_t *main_engine;
	osm_routing_chain_t *used_engine;
	size_t i;
	struct osm_opensm *p_osm;
	osm_topology_t *main_topo;
	int res = -1;
	char *completer_suffix = "";

	p_osm = ucast_mgr->p_subn->p_osm;
	if (osm_topologies_count(p_osm) == 0) {
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR, "ERR 3911: No "
			"topologies Found - Aborting Routing Chains Process\n");
		goto EXIT;
	}
	if (!osm_rch_mgr_create(ucast_mgr->p_subn->p_osm))
		goto EXIT;

	OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Clearing removed hcas maps"
		" on all topologies\n");
	clear_all_removed_hcas_maps(p_osm);

	for (i = 1; i <= MAX_ENGINES_NUMBER; i++) {
		if (!p_mgr->used_ids[i])
			continue;
		p_engine = osm_rch_get_engine(p_osm, i);
		if (!p_engine) {
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR,
				"ERR 3912: Can't find routing engine %zd\n", i);
			goto EXIT;
		}
		if (engine_use_main_subn(p_osm, p_engine))
			continue;

		if (route_engine_with_fallback(ucast_mgr, p_engine))
			goto EXIT;
	}

	main_engine = get_main_engine(p_osm, p_mgr);

	OSM_ASSERT(main_engine);

	/*
	 * Now there are three options:
	 *
	 * 1) No ITR engine - In that case main engine should run with
	 *                    entire fabric
	 *
	 * 2) Only main and ITR engine in chain - Main engine should use
	 *	the completer subnet (since main engine already set to
	 *	completer and ran there is nothing to do
	 *
	 * 3) Main, ITR and more engines in chain - build and route
	 *	non-itr topology in main engine.
	 */

	if (p_mgr->is_itr) {
		/*
		 * for multi-cast routing we need to build and route the
		 * completer subnet
		 */
		if (osm_topo_mgr_create_completer_topo(p_osm))
			goto EXIT;
		main_topo = osm_topo_mgr_get_topo(p_osm, MAIN_TOPO_ID);
		if (!osm_subn_bs_count(main_topo->p_subn)) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 391C: "
				"No border switches in completer subnet\n");
			goto EXIT;
		}
		main_engine->subn = main_topo->p_subn;
		if (route_engine_with_fallback(ucast_mgr, main_engine))
			goto EXIT;

		if (osm_rch_engines_count(p_osm) > 2) {
			/* option 3*/
			if (osm_topo_mgr_create_non_itr_topo(p_osm))
				goto EXIT;
			main_topo = osm_topo_mgr_get_topo(p_osm, NON_ITR_TOPO_ID);
			if (!osm_subn_bs_count(main_topo->p_subn)) {
				OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR,
					"ERR 391D: No border switches in "
					"non-itr subnet\n");
				goto EXIT;
			}
			main_engine->subn = main_topo->p_subn;
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Main routing "
				"engine set with non-itr topology\n");
			completer_suffix = " (non-itr)";
		} else {
			/* option 2 */
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Main routing "
				"engine set with completer topology\n");
			completer_suffix = " (completer)";
			goto MERGE;
		}
		/* reset new lfts from previous routing in main subnet, before merge*/
		osm_ucast_mgr_reset_lfts_all_switches(&p_osm->subn);
	} else {
		/* option 1 */
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Main routing engine "
			"set with main subnet\n");
		completer_suffix = " (main)";
		main_engine->subn = &p_osm->subn;
	}

	/* Route main engine */
	if (route_engine_with_fallback(ucast_mgr, main_engine))
		goto EXIT;
MERGE:
	for (i = 1; i <= MAX_ENGINES_NUMBER; i++) {
		if (!p_mgr->used_ids[i])
			continue;
		p_engine = osm_rch_get_engine(p_osm, i);
		if (p_engine->used_fallback)
			used_engine = p_engine->fallback_engine;
		else
			used_engine = p_engine;
		if (!engine_use_main_subn(p_osm, used_engine)) {
			OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "Merging "
				"engine %zd (%s)\n", used_engine->id,
				used_engine->engine);
			merge_subn_lfts(ucast_mgr->p_log, used_engine->subn,
					&p_osm->subn);
			merge_subn_ar_configuration(ucast_mgr->p_log,
						    used_engine->subn,
						    &p_osm->subn);
		}
	}

	OSM_LOG(ucast_mgr->p_log, OSM_LOG_DEBUG, "All engines routed, "
		"perfoming ITR...\n");
	if (p_mgr->is_itr && process_itr(&p_osm->subn)) {
		OSM_LOG(ucast_mgr->p_log, OSM_LOG_ERROR, "ERR 3919: "
			" Errors during ITR, RCH routing failed!\n");
		p_mgr->is_itr = FALSE;
		goto EXIT;
	}

	print_rch_info(p_osm, completer_suffix, OSM_LOG_DEBUG);
	res = 0;

EXIT:
	if (res)
		rch_destroy_rbox_maps(p_mgr);

	return res;
}

boolean_t osm_rch_mgr_create(IN osm_opensm_t * p_osm)
{
	boolean_t res = TRUE;
	osm_routing_chain_mgr_t *p_mgr;

	p_mgr = &p_osm->sm.routing_chain_mgr;
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
		"creating routing chain manager\n");

	osm_rch_mgr_clean(p_mgr);

	if (parse_rch_policy(p_osm)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 3910: RCH "
			"policy file parsing failed!!!\n");
		res =  FALSE;
	} else
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "RCH "
			"policy file parsed successfully!!!\n");
	return res;
}

void osm_rch_mgr_construct(IN osm_routing_chain_mgr_t *p_mgr)
{
	cl_qlist_init(&p_mgr->engine_list);
}

ib_api_status_t osm_rch_mgr_init(IN osm_routing_chain_mgr_t * p_mgr,
				 IN osm_sm_t * sm)
{
        unsigned i, mlid_range;

	OSM_LOG_ENTER(sm->p_log);

	mlid_range = IB_LID_MCAST_END_HO - IB_LID_MCAST_START_HO + 1;
        p_mgr->p_rch_rboxes = malloc(sizeof(cl_qmap_t) * mlid_range);
        if (!p_mgr->p_rch_rboxes) {
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3905: Failed to "
			"allocate rbox objects array\n");
                return CL_INSUFFICIENT_MEMORY;
        }

        for(i = 0; i < mlid_range; i++)
                cl_qmap_init(&p_mgr->p_rch_rboxes[i]);

	p_mgr->p_osm = sm->p_subn->p_osm;

	OSM_LOG_EXIT(sm->p_log);
        return CL_SUCCESS;
}

static boolean_t osm_rch_is_chain_itr(struct osm_ucast_mgr *p_mgr)
{
	osm_routing_chain_mgr_t *p_chain_mgr;
	p_chain_mgr = &p_mgr->sm->routing_chain_mgr;
	return p_chain_mgr->is_itr;
}

static
boolean_t osm_rch_is_physp_topo_itr(IN struct osm_opensm *p_osm,
				    IN osm_physp_t *p_physp,
				    OUT struct osm_routing_engine **r)
{
	struct osm_routing_engine * p_itr_r;
	osm_topology_t *p_topo;

	OSM_ASSERT(p_physp);

	p_topo = osm_topo_mgr_get_topo(p_osm, p_physp->topo_id);

	if (!osm_topo_is_engine_itr(p_topo))
		return FALSE;

	p_itr_r = p_topo->p_subn->routing_engine_used;

	if (r)
		*r = p_itr_r;

	return TRUE;
}

static void osm_rch_physp_re(IN struct osm_opensm *p_osm,
			     IN osm_physp_t *p_physp,
			     OUT struct osm_routing_engine **r)
{
	struct osm_routing_engine *p_re;
	osm_topology_t *p_topo;
	OSM_ASSERT(p_physp);

	p_topo = osm_topo_mgr_get_topo(p_osm, p_physp->topo_id);
	if (!p_topo)
		return;

	p_re = p_topo->p_subn->routing_engine_used;

	if (r)
		*r = p_re;
}

static osm_physp_t * osm_rch_get_physp_from_topo(IN struct osm_opensm *p_osm,
						 IN osm_physp_t *p_physp)
{
	osm_topology_t *p_topo;
	osm_node_t *p_node, *p_topo_node;
	ib_net64_t node_guid;

	OSM_ASSERT(p_physp);

	p_topo = osm_topo_mgr_get_topo(p_osm, p_physp->topo_id);

	if (!p_topo)
		return NULL;

	p_node = osm_physp_get_node_ptr(p_physp);

	node_guid = osm_node_get_node_guid(p_node);

	p_topo_node = osm_get_node_by_guid(p_topo->p_subn, node_guid);

	OSM_ASSERT(p_topo_node);

	return osm_node_get_physp_ptr(p_topo_node, p_physp->port_num);
}


static
void osm_rch_update_vlarb(void *context, osm_physp_t *osm_phys_port,
			  uint8_t port_num, ib_vl_arb_table_t *block,
			  unsigned block_length, unsigned block_num)
{
	struct osm_ucast_mgr *p_mgr = (struct osm_ucast_mgr *) context;
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;
	struct osm_routing_engine * r = NULL;
	osm_physp_t *p_remote_physp;
	osm_node_t *node = osm_physp_get_node_ptr(osm_phys_port);
	unsigned i, next;

	if (!osm_rch_is_chain_itr(p_mgr))
		return;

	/* the physp port is belonging to ITR topology */
	if (osm_rch_is_physp_topo_itr(p_osm, osm_phys_port, &r)) {
		if (!r || !r->update_vlarb)
			return;
		/* need to get physp from the topology */
		osm_phys_port = osm_rch_get_physp_from_topo(p_osm,
							    osm_phys_port);

		OSM_ASSERT(osm_phys_port);

		r->update_vlarb(r->context, osm_phys_port, port_num,
				block, block_length, block_num);
		return;
	}
	/* need to configure ports external to the ITR topology
	 * the current algorithm will execute similarly to the Torus one
	 *
	 * If osm_phys_port is a switch port that connects to a CA, then
	 * we're using at most VL 0 (for QoS level 0) and VL 1 (for QoS
	 * level 1).  We've been passed the  VLarb values for a switch
	 * external port, so we need to fix them up to avoid unexpected
	 * results depending on how the switch handles VLarb values for
	 * unprogrammed VLs.
	 *
	 * So, leave VL 0 alone, remap VL 4 to VL 1, zero out the rest,
	 * and compress out the zero entries to the end.
	 */
	if (!node->sw || !port_num)
		return;

	p_remote_physp = osm_phys_port->p_remote_physp;

	if (p_remote_physp && (osm_node_get_type(node) == IB_NODE_TYPE_SWITCH) &&
	   (osm_node_get_type(p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH))
		return;

	/* remote port is HCA port */
	next = 0;
	for (i = 0; i < block_length; i++) {
		switch (block->vl_entry[i].vl) {
		case 4:
			block->vl_entry[i].vl = 1;
			/* fall through */
		case 0:
			block->vl_entry[next].vl = block->vl_entry[i].vl;
			block->vl_entry[next].weight = block->vl_entry[i].weight;
			next++;
			/*
			 * If we didn't update vl_entry[i] in place,
			 * fall through to zero it out.
			 */
			if (next > i)
				break;
		default:
			block->vl_entry[i].vl = 0;
			block->vl_entry[i].weight = 0;
			break;
		}
	}
}

/* the osm_phys_port is representing physp of the oport_num */
static uint8_t osm_rch_sl2vl_entry(osm_subn_t *p_subn,
				   osm_node_t *p_node,
				   uint8_t iport_num,
				   uint8_t oport_num,
				   uint8_t sl)

{
	uint8_t vl = 0;
	if (!p_node->sw) {
		if (p_subn->min_data_vls >= 2)
			vl = sl >> 3 & 0x1;
	} else {
		if (p_subn->min_sw_data_vls >= 2)
			 vl = sl >> 3 & 0x1;
	}
	return vl;
}

void osm_rch_update_sl2vl(void *context, osm_physp_t *osm_phys_port,
			  uint8_t iport_num, uint8_t oport_num,
			  ib_slvl_table_t *osm_oport_sl2vl)
{
	struct osm_ucast_mgr *p_mgr = (struct osm_ucast_mgr *) context;
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;
	struct osm_routing_engine * r = NULL;
	osm_physp_t *p_physp_in;
	osm_node_t *node = osm_physp_get_node_ptr(osm_phys_port);
	uint8_t sl, vl;

	if (!osm_rch_is_chain_itr(p_mgr)) {
		/* check DF+ */
		osm_rch_physp_re(p_osm, osm_phys_port, &r);
		/* DF+ has update_sl2vl too */
		if (!r || !r->update_sl2vl)
			return;
		/* p_physp belongs to topology with update_sl2vl defined */

		/* switch is in DF+ topology */
		if (osm_node_get_type(node) == IB_NODE_TYPE_SWITCH) {
			p_physp_in = osm_node_get_physp_ptr(node, iport_num);

			/* in port and out port both in the DF+ */
			if (p_physp_in && p_physp_in->topo_id == osm_phys_port->topo_id) {
				/* need to get physp from the topology */
				osm_phys_port = osm_rch_get_physp_from_topo(p_osm,
							    		    osm_phys_port);

				OSM_ASSERT(osm_phys_port);

				r->update_sl2vl(r->context, osm_phys_port, iport_num,
						oport_num, osm_oport_sl2vl);
			}
		}
		return;
	}

	/* outport is on ITR topology */
	if (osm_rch_is_physp_topo_itr(p_osm, osm_phys_port, &r)) {

		if (!r || !r->update_sl2vl)
			return;

		if (osm_node_get_type(node) == IB_NODE_TYPE_SWITCH) {
			p_physp_in = osm_node_get_physp_ptr(node, iport_num);

			if (!p_physp_in)
				return;

			iport_num =
				(p_physp_in->topo_id != NO_TOPO_ID &&
				p_physp_in->topo_id == osm_phys_port->topo_id) ?
				iport_num:	/* both ports are belonging
						   to the same topology */
				0;
		}
		/* need to get physp from the topology */
		osm_phys_port = osm_rch_get_physp_from_topo(p_osm,
							    osm_phys_port);

		OSM_ASSERT(osm_phys_port);

		r->update_sl2vl(r->context, osm_phys_port, iport_num,
				oport_num, osm_oport_sl2vl);
		return;
	}

	for (sl = 0; sl < 16; sl++) {
		vl = osm_rch_sl2vl_entry(p_mgr->p_subn, node, iport_num, oport_num, sl);
		ib_slvl_table_set(osm_oport_sl2vl, sl, vl);
	}
}

static uint8_t osm_rch_path_sl(void *context, uint8_t path_sl_hint,
                      const ib_net16_t slid, const ib_net16_t dlid)
{
	struct osm_ucast_mgr *p_mgr = (struct osm_ucast_mgr *) context;
	osm_subn_t *p_main_subn;
	ib_net16_t s_lid, d_lid;
	boolean_t is_src_border, is_dest_border;
	osm_port_t *p_src_port, *p_dest_port, *p_bs_port;
	uint8_t sl_result = path_sl_hint;
	osm_topology_t *p_topo, *p_dest_topo, *p_src_topo;
	struct osm_routing_engine * p_itr_r;

	if (!osm_rch_is_chain_itr(p_mgr))
		goto Exit;

	p_main_subn = p_mgr->p_subn;

	p_src_port = osm_get_port_by_lid(p_main_subn, slid);

	if (!p_src_port) {
		OSM_LOG(&p_main_subn->p_osm->log, OSM_LOG_ERROR, "ERR 392A: "
			"Failed to get port for slid %" PRIu16 "\n",
			cl_ntoh16(slid));
		goto Exit;
	}

	p_dest_port = osm_get_port_by_lid(p_main_subn, dlid);

	if (!p_dest_port) {
		OSM_LOG(&p_main_subn->p_osm->log, OSM_LOG_ERROR, "ERR 392B: "
			"Failed to get port for dlid %" PRIu16 "\n",
			cl_ntoh16(dlid));
		goto Exit;
	}

	itr_required(p_src_port, &p_src_topo, &is_src_border);

	itr_required(p_dest_port, &p_dest_topo, &is_dest_border);

	/* both ports belong to MAIN topology*/
	if (!p_src_topo && !p_dest_topo)
		goto Exit;

	/* both ports are ITR */
	if (p_src_topo && p_dest_topo) {
		if (p_src_topo == p_dest_topo) {
			p_topo = p_src_topo;
			s_lid = slid;
			d_lid = dlid;
			goto Calc_sl;
		} else {
			OSM_LOG(&p_main_subn->p_osm->log, OSM_LOG_ERROR,
				"Two QoS topologies with Routing chains"
				" are not supported\n");
			goto Exit;
		}
	}
	if (!p_src_topo) {
		/* source port is on MAIN topology */
		/* dest port is on ITR topology, */
		/* need to find corresponding border switch  */
		/* dest is border, no path_sl is required*/
		if (is_dest_border)
			goto Exit;
		if (!(p_bs_port = itr_get_bs_by_port(p_dest_port, NULL)))
			goto Exit;
		p_topo = p_dest_topo;
		s_lid = p_bs_port->lid;
		d_lid = dlid;
	} else {
		/* dest port is on MAIN topology */
		/* source port is on ITR topology */
		if (is_src_border)
			goto Exit;
		if (!(p_bs_port = itr_get_bs_by_port(p_src_port, NULL)))
			goto Exit;
		p_topo = p_src_topo;
		s_lid = slid;
		d_lid = p_bs_port->lid;
	}

Calc_sl:
	p_itr_r = p_topo->p_subn->routing_engine_used;

	if (p_itr_r && p_itr_r->path_sl) {
		sl_result = p_itr_r->path_sl(p_itr_r->context,
					     path_sl_hint,
					     s_lid,
					     d_lid);
	}
Exit:
	return sl_result;
}

static osm_mgrp_box_t * rch_mgrp_box_new(uint16_t mlid)
{
        osm_mgrp_box_t *mbox = malloc(sizeof(*mbox));
        if (!mbox)
                return NULL;

        memset(mbox, 0, sizeof(*mbox));
        mbox->mlid = mlid;
        cl_qlist_init(&mbox->mgrp_list);

        return mbox;
}

static osm_rch_box_t * rch_box_new(osm_topology_t * p_topology, uint16_t mlid)
{
	osm_rch_box_t *p_rch_box;
	osm_mgrp_box_t *p_mbox;

	p_rch_box = (osm_rch_box_t *)malloc(sizeof(*p_rch_box));
	if (!p_rch_box)
		return NULL;

	memset(p_rch_box, 0, sizeof(*p_rch_box));
	p_mbox = rch_mgrp_box_new(mlid);
	if (!p_mbox) {
		free(p_rch_box);
		return NULL;
	}

	p_rch_box->mbox = p_mbox;
	p_rch_box->prim_bs_guid = 0;
	p_rch_box->topology_id = p_topology->id;

	cl_qmap_init(&p_rch_box->members_group.mcm_port_tbl);
	p_rch_box->members_group.mlid = mlid;
	cl_qlist_insert_tail(&p_mbox->mgrp_list,
			     &p_rch_box->members_group.list_item);

	cl_qmap_init(&p_rch_box->borders_group.mcm_port_tbl);
	p_rch_box->borders_group.mlid = mlid;
	cl_qlist_insert_tail(&p_mbox->mgrp_list,
			     &p_rch_box->borders_group.list_item);
	return  p_rch_box;
}

static ib_net64_t rch_rbox_get_prim_bs_guid(osm_sm_t * sm,
					    osm_topology_t *p_topology)
{
	cl_map_item_t * p_item;
	ib_net64_t guid;

	OSM_ASSERT(osm_subn_bs_count(p_topology->p_subn));

	/*
	 * Note: if we want to reuse previous primary border switch, we should
	 * keep it per topology and not per mlid
	 */

	/* Prefer to use hub switches to pass mc traffic between subnets */
	guid = osm_subn_get_hub_switch_guid(sm->p_subn);
	if (guid && osm_topo_is_border_sw(p_topology, guid))
		return guid;

	/* select first border switch of this toplogy */
	p_item = cl_qmap_head(&p_topology->p_subn->border_sw_tbl);
	return cl_qmap_key(p_item);
}

static boolean_t rch_merge_mft_dfs(osm_sm_t * sm,
				   cl_qmap_t * p_rbox_map,
				   osm_switch_t * p_main_sw,
				   osm_switch_t * p_topo_sw,
				   uint8_t us_port,
				   uint16_t mlid)
{
	osm_node_t * p_topo_node;
	osm_node_t * p_main_node;
	uint8_t rem_us_port;
	uint8_t ds_port;
	boolean_t trim_branch = FALSE;
	boolean_t has_leaves = FALSE;
	boolean_t result;

	for (ds_port = 0; ds_port < p_main_sw->num_ports; ds_port++) {

		if (us_port && ds_port == us_port)
			continue;

		/* Check if the switch wasn't already visited */
		if (osm_mcast_tbl_is_port(&p_main_sw->mcast_tbl, mlid, ds_port))
			trim_branch = TRUE;

		if (osm_mcast_tbl_is_port(&p_topo_sw->mcast_tbl, mlid,
					  ds_port) == FALSE)
			continue;

		if (ds_port == 0) {
			osm_mcast_tbl_set(&p_main_sw->mcast_tbl, mlid, 0);
			has_leaves = TRUE;
			continue;
		}

		p_topo_node =
		    osm_node_get_remote_node(p_topo_sw->p_node, ds_port,
					     &rem_us_port);
		p_main_node =
		    osm_node_get_remote_node(p_main_sw->p_node, ds_port, NULL);

		OSM_ASSERT(p_topo_node && p_main_node && rem_us_port);

		if (osm_node_get_type(p_main_node) != IB_NODE_TYPE_SWITCH) {
			osm_mcast_tbl_set(&p_main_sw->mcast_tbl, mlid, ds_port);
			has_leaves = TRUE;
			continue;
		}

		result = rch_merge_mft_dfs(sm, p_rbox_map, p_main_node->sw,
					   p_topo_node->sw, rem_us_port, mlid);
		if (result) {
			/* Branch is part of the multicast tree */
			has_leaves = TRUE;
			osm_mcast_tbl_set(&p_main_sw->mcast_tbl, mlid, ds_port);
		}
	}

	/*
	 * Verify that we dont break paths toward sub-topologies primary
	 * border switches.
	 */
	if (trim_branch && p_rbox_map) {
		cl_map_item_t *p_item;
		ib_net64_t sw_guid = osm_node_get_node_guid(p_main_sw->p_node);
		for (p_item = cl_qmap_head(p_rbox_map);
		     p_item != cl_qmap_end(p_rbox_map);
		     p_item = cl_qmap_next(p_item)) {
			osm_rch_box_t * rbox = (osm_rch_box_t *)p_item;
			if (sw_guid == rbox->prim_bs_guid) {
				trim_branch = FALSE;
				break;
			}
		}
	}

	/*
	 * If this node doesn't leads to leaves or was already visited,
	 * we dont need to use it in the multicast tree and return FALSE
	 */
	if (trim_branch || !has_leaves)
		return FALSE;

	if (us_port)
		osm_mcast_tbl_set(&p_main_sw->mcast_tbl, mlid, us_port);

	return TRUE;
}

/*
 * This function merges mft tables of a sub-topology subnet into the mft tables
 * of the main subnet using DFS from the primary border switch of the
 * sub-topology.
 *
 * The algorithm avoids adding mc routes to switches that were already merged
 * into the main subnet, unless p_rbox_map is defined, and the switch is a
 * primary border switch of one of the rbox objects in the map.
 */
static void rch_merge_mft_rbox(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
			       osm_rch_box_t * rbox,
			       uint16_t mlid)
{

	osm_switch_t *p_topo_sw;
	osm_switch_t *p_main_sw;
	osm_topology_t *p_topology;
	ib_net64_t dfs_root_guid;

	p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
					   rbox->topology_id);
	OSM_ASSERT(p_topology);

	if (!rbox->mbox->root) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Skipping MFT for topology id %" PRIu64"  mlid 0x%x\n",
			p_topology->id, mlid);
		return;
	}

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Merging MFT for topology id %" PRIu64"  mlid 0x%x\n",
		p_topology->id, mlid);

	dfs_root_guid = rbox->prim_bs_guid ? rbox->prim_bs_guid :
			osm_node_get_node_guid(rbox->mbox->root->p_sw->p_node);
	p_topo_sw =
	    (osm_switch_t *)cl_qmap_get(&p_topology->p_subn->sw_guid_tbl,
					dfs_root_guid);
	p_main_sw =
	    (osm_switch_t *)cl_qmap_get(&sm->p_subn->sw_guid_tbl,
					dfs_root_guid);

	OSM_ASSERT(p_main_sw && p_topo_sw);
	rch_merge_mft_dfs(sm, p_rbox_map, p_main_sw, p_topo_sw, 0, mlid);
}

static void rch_merge_mft(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
			  uint16_t mlid)
{

	osm_rch_box_t *rbox;
	osm_rch_box_t *rbox0 = NULL;
	cl_map_item_t *p_item;
	osm_switch_t *p_sw;

	OSM_ASSERT(cl_qmap_count(p_rbox_map) == 1 ||
		   cl_qmap_get(p_rbox_map, 0) != cl_qmap_end(p_rbox_map));

	/*
	 * We mark all primary border switches of every sub topology, as
	 * members on the MFT tables of the main subnet object, in order to
	 * avoid connecting them to the multicast tree of other sub topologies
	 */
	for (p_item = cl_qmap_head(p_rbox_map);
	     p_item != cl_qmap_end(p_rbox_map);
	     p_item = cl_qmap_next(p_item)) {
		rbox = (osm_rch_box_t *)p_item;
		if (!rbox->prim_bs_guid)
			continue;
		p_sw = (osm_switch_t *)cl_qmap_get(&sm->p_subn->sw_guid_tbl,
						   rbox->prim_bs_guid);
		osm_mcast_tbl_set(&p_sw->mcast_tbl, mlid, 0);
	}

	for (p_item = cl_qmap_head(p_rbox_map);
	     p_item != cl_qmap_end(p_rbox_map);
	     p_item = cl_qmap_next(p_item)) {

		/*
		 * We keep rbox0 for last in order to give precedence to sub
		 * topologies
		 */
		if (!cl_qmap_key(p_item)) {
			rbox0 = (osm_rch_box_t *)p_item;
			continue;
		}
		rbox = (osm_rch_box_t *)p_item;
		rch_merge_mft_rbox(sm, NULL, rbox, mlid);
	}

	if (rbox0)
		rch_merge_mft_rbox(sm, p_rbox_map, rbox0, mlid);

	return;
}

static void rch_remove_port_from_mft(osm_switch_t * p_sw, uint8_t in_port,
				     uint16_t mlid)
{
	osm_node_t *p_remote_node;
	osm_mcast_tbl_t *p_tbl;
	int port;
	uint8_t ds_port;

	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
	osm_mcast_tbl_clear(p_tbl, mlid, in_port);
	port = osm_mcast_tbl_get_sole_active_port(p_tbl, mlid);

	/*
	 * If current switch is member or it connects between two sub-trees
	 * we stop trimming the tree.
	 */
	if (port <= 0 || port > IB_NODE_NUM_PORTS_MAX)
		return;

	osm_mcast_tbl_clear(p_tbl, mlid, (uint8_t)port);
	p_remote_node = osm_node_get_remote_node(p_sw->p_node, (uint8_t)port,
						 &ds_port);
	OSM_ASSERT(p_remote_node && ds_port);
	if (osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH)
		rch_remove_port_from_mft(p_remote_node->sw, ds_port, mlid);
}

static void rch_remove_borders_from_mft(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
					uint16_t mlid)
{
	cl_map_item_t *p_item;
	cl_map_item_t *p_port_item;
	cl_map_item_t *p_mcm_item;
	osm_rch_box_t *p_rch_box;
	osm_mcm_port_t *p_mcm_port;
	osm_port_t *p_port;
	osm_subn_t *p_subn = sm->p_subn;

	OSM_LOG_ENTER(sm->p_log);

	for (p_item = cl_qmap_head(p_rbox_map);
	     p_item != cl_qmap_end(p_rbox_map);
	     p_item = cl_qmap_next(p_item)) {

		p_rch_box = (osm_rch_box_t *)p_item;

		osm_mgrp_t * p_mgrp = &p_rch_box->borders_group;
		for (p_mcm_item = cl_qmap_head(&p_mgrp->mcm_port_tbl);
		     p_mcm_item != cl_qmap_end(&p_mgrp->mcm_port_tbl);
		     p_mcm_item = cl_qmap_next(p_mcm_item)) {
			p_mcm_port = (osm_mcm_port_t *)p_mcm_item;
			p_port_item = cl_qmap_get(&p_subn->port_guid_tbl,
						  p_mcm_port->port->guid);
			OSM_ASSERT(p_port_item !=
				   cl_qmap_end(&p_subn->port_guid_tbl));
			p_port = (osm_port_t *)p_port_item;
			rch_remove_port_from_mft(p_port->p_node->sw, 0, mlid);
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}

static int rch_add_port_to_rbox_mgrp(osm_sm_t * sm, osm_port_t * p_port,
				     osm_mgrp_t * p_mgrp)
{
	osm_mcm_port_t *p_mcm_port;
	p_mcm_port = osm_mcm_port_new(p_port, p_mgrp);
	if (!p_mcm_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 3925: Faild to create mcm_port_t object\n");
		return 1;
	}
	p_mcm_port->num_receivers = 1;
	p_mgrp->full_members++;
	cl_qmap_insert(&p_mgrp->mcm_port_tbl, p_port->guid,
		       &p_mcm_port->map_item);
	return 0;
}

static int rch_add_borders_to_mgroups(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
				       cl_qmap_t * p_mcm_ports_map)
{
	osm_rch_box_t *rbox0 = NULL;
	osm_subn_t *p_subn0;
	osm_subn_t *p_subn;
	cl_map_item_t *p_box_item;
	osm_rch_box_t *rbox;
	osm_topology_t *p_topology;
	cl_map_item_t *p_item;
	osm_mgrp_t *p_mgrp;
	osm_port_t * p_port;
	osm_switch_t * p_sw;
	ib_net64_t guid;

	int status = 0;

	/*
	 * If all the members are from a single topology we dont need to add
	 * border switches for the multicast group.
	 */
	if (cl_qmap_count(p_rbox_map) < 2)
		return status;

	/*
	 * We add the borders of each sub-toplogy as members to the group in
	 * order to ensure that when we merge the multicast forwarding tables
	 * we would give precedence to the multicast trees of sub-topologies
	 * over the multicast tree of the completer topology.
	 * By doing so, we ensure that the only cross topologies turns would
	 * be through the primary border switch, and reduce chances for credit
	 * loops.
	 *
	 * We also add one border switch of each sub-topology, as a member
	 * to the completer topology in order to ensure connectivity.
	*/

	for (p_box_item = cl_qmap_head(p_rbox_map);
	     p_box_item != cl_qmap_end(p_rbox_map);
	     p_box_item = cl_qmap_next(p_box_item)) {

		if (!cl_qmap_key(p_box_item)) {
			rbox0 = (osm_rch_box_t *)p_box_item;
			continue;
		}
		rbox = (osm_rch_box_t *)p_box_item;
		p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
						   rbox->topology_id);
		rbox->prim_bs_guid = rch_rbox_get_prim_bs_guid(sm, p_topology);
		p_subn = p_topology->p_subn;

		for (p_item = cl_qmap_head(&p_subn->border_sw_tbl);
		     p_item != cl_qmap_end(&p_subn->border_sw_tbl);
		     p_item = cl_qmap_next(p_item)) {

			p_sw =
			    (osm_switch_t *)cl_qmap_get(&p_subn->sw_guid_tbl,
							cl_qmap_key(p_item));
			p_port = osm_switch_get_port0(p_sw, p_subn);
			guid = osm_port_get_guid(p_port);

			/* If border switch is already a member, do nothing */
			p_mgrp = &rbox->members_group;
			if (cl_qmap_get(&p_mgrp->mcm_port_tbl, guid) !=
			    cl_qmap_end(&p_mgrp->mcm_port_tbl))
				continue;

			OSM_ASSERT(p_port);
			if (rch_add_port_to_rbox_mgrp(sm, p_port,
						      &rbox->borders_group)) {
				status = 1;
				goto Exit;
			}
		}
	}

	if (rbox0) {
		p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
						   rbox0->topology_id);
		p_subn0 = p_topology->p_subn;

		for (p_box_item = cl_qmap_head(p_rbox_map);
		     p_box_item != cl_qmap_end(p_rbox_map);
		     p_box_item = cl_qmap_next(p_box_item)) {

			if (!cl_qmap_key(p_box_item))
				continue;
			rbox = (osm_rch_box_t *)p_box_item;
			p_sw =
			    (osm_switch_t *)cl_qmap_get(&p_subn0->sw_guid_tbl,
							rbox->prim_bs_guid);
			p_port = osm_switch_get_port0(p_sw, p_subn0);
			guid = osm_port_get_guid(p_port);

			if (cl_qmap_get(p_mcm_ports_map, guid) !=
			    cl_qmap_end(p_mcm_ports_map))
				p_mgrp = &rbox0->members_group;
			else
				p_mgrp = &rbox0->borders_group;

			if (rch_add_port_to_rbox_mgrp(sm, p_port, p_mgrp)) {
				status = 1;
				goto Exit;
			}

			/*
			 * define the guid of the switch from which we start
			 * merging MFT tables.
			 */
			rbox0->prim_bs_guid = rbox->prim_bs_guid;
		}
	}

Exit:
	return status;
}

static int rch_add_port_to_rbox_map(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
				    osm_rch_box_t * rbox0, osm_port_t * p_port,
				    uint16_t mlid)
{
	cl_list_iterator_t iter;
	cl_map_item_t * p_item;
	osm_port_t *p_topology_port;
	osm_topology_t *p_topology;
	osm_subn_t *p_subn;
	osm_rch_box_t *rbox;
	osm_mgrp_t *p_mgrp;
	int status = 0;

	for (iter = cl_list_head(&p_port->topology_list);
	     iter != cl_list_end(&p_port->topology_list);
	     iter = cl_list_next(iter)) {

		p_topology = osm_port_get_topo_from_list_iter(&iter);
		if (p_topology->id == 0) {
			/*
			 * We avoid adding ports to the completer topology in
			 * order to avoid building its multicast tree.
			 */
			if (cl_list_count(&p_port->topology_list) == 1)
				rbox = rbox0;
			else
				continue;
		} else {
			p_item = cl_qmap_get(p_rbox_map, p_topology->id);
			if (p_item == cl_qmap_end(p_rbox_map)) {
				/* Create new rch_box object */
				rbox = rch_box_new(p_topology, mlid);
				if (!rbox) {
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 3924: Faild to create "
						"RCH box object for topology "
						"id %" PRIu64 "for mlid "
						"0x%04x\n",
						p_topology->id, mlid);
					status = 1;
					goto Exit;
				}
				cl_qmap_insert(p_rbox_map, p_topology->id,
					       &rbox->map_item);
			} else {
				rbox = (osm_rch_box_t *)p_item;
			}
		}

		p_subn = p_topology->p_subn;
		p_topology_port =
		    (osm_port_t*)cl_qmap_get(&p_subn->port_guid_tbl,
					     p_port->guid);
		p_mgrp = &rbox->members_group;
		if (rch_add_port_to_rbox_mgrp(sm, p_topology_port, p_mgrp)) {
			status = 1;
			goto Exit;
		}

	}

Exit:
	return status;
}

static int rch_create_rbox_map(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
				   cl_qlist_t * p_mcm_ports_list,
				   uint16_t mlid)
{
	osm_rch_box_t *rbox;
	osm_rch_box_t *rbox0;
	osm_topology_t * p_topology;
	osm_topology_t * p_topology0;
	cl_map_item_t * p_item;
	cl_map_item_t * p_next_item;
	cl_list_item_t *p_list_item;
	osm_port_t *p_port;
	int status = 0;

	OSM_LOG_ENTER(sm->p_log);

	/*
	 * Remove all members from the rch_boxes that were already allocated
	 * before this call, and clear rbox's primary border switch.
	 */
	p_item = cl_qmap_head(p_rbox_map);
	while (p_item != cl_qmap_end(p_rbox_map)) {
		p_next_item = cl_qmap_next(p_item);
		p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
						   cl_qmap_key(p_item));
		if (p_topology) {
			rbox = (osm_rch_box_t *)p_item;
			rbox->prim_bs_guid = 0;
			rch_box_remove_members(rbox);
			rbox->topology_id = p_topology->id;
		}
		else {
			cl_qmap_remove_item(p_rbox_map, p_item);
			rch_box_delete(sm, (osm_rch_box_t *)p_item);
		}
		p_item = p_next_item;
	}

	/* Create or restore rbox0 object */
	p_item = cl_qmap_get(p_rbox_map, MAIN_TOPO_ID);
	if (p_item != cl_qmap_end(p_rbox_map)) {
		rbox0 = (osm_rch_box_t *)p_item;
		cl_qmap_remove_item(p_rbox_map, p_item);
	} else {
		p_topology0 = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
						    MAIN_TOPO_ID);
		OSM_ASSERT(p_topology0);
		rbox0 = rch_box_new(p_topology0, mlid);
		if (!rbox0) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3923: "
				"Faild to create RCH box object object for "
				"the completer topology\n");
				status = 1;
				goto Exit;
		}
	}

	/*
	 * Add each mc member port to rbox object of the port's sub topology.
	 */
	for (p_list_item = cl_qlist_head(p_mcm_ports_list);
	     p_list_item != cl_qlist_end(p_mcm_ports_list);
	     p_list_item = cl_qlist_next(p_list_item)) {
		p_port = ((osm_mcast_work_obj_t *)p_list_item)->p_port;
		if (rch_add_port_to_rbox_map(sm, p_rbox_map, rbox0, p_port,
					     mlid)) {
			/* clear all rch_box objects and exit */
			rch_rbox_map_destroy(sm, p_rbox_map);
			rch_box_delete(sm, rbox0);
			status = 1;
			goto Exit;
		}
	}

	/* Decide whether to keep each rbox or not */
	p_next_item = cl_qmap_head(p_rbox_map);
	while (p_next_item != cl_qmap_end(p_rbox_map)) {
		p_item = p_next_item;
		p_next_item = cl_qmap_next(p_next_item);
		rbox = (osm_rch_box_t *)p_item;
		if (!cl_qmap_count(&rbox->members_group.mcm_port_tbl)) {
			cl_qmap_remove_item(p_rbox_map, p_item);
			rch_box_delete(sm, rbox);
		}
	}

	/* Decide whether to keep rbox0 or not */
	if (cl_qmap_count(p_rbox_map) > 1 ||
	    cl_qmap_count(&rbox0->members_group.mcm_port_tbl)) {
		cl_qmap_insert(p_rbox_map, 0, &rbox0->map_item);
	} else {
		rch_box_delete(sm, rbox0);
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
	return status;
}

static osm_mtree_node_t * rch_create_mtree_dfs(osm_sm_t * sm,
					       osm_switch_t * p_sw,
					       osm_mtree_node_t * p_up,
					       uint8_t up_port,
					       uint16_t mlid)
{
	osm_mtree_node_t *p_mtn;
	osm_mcast_tbl_t *p_tbl;
	osm_node_t *p_rem_node;
	osm_mtree_node_t *p_child;
	uint8_t port, rem_port, active_ports = 0;

	if (p_sw->multicast_priv) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3931: "
			"Found loop in multicast tree of MLID 0x%X on switch "
			"0x%" PRIx64 "\n",
			mlid,
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		return NULL;
	}

	p_mtn = osm_mtree_node_new(p_sw);
        if (p_mtn == NULL) {
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 392F: "
                        "Insufficient memory to build multicast tree for MLID"
			"0x%X\n",
			mlid);
                return NULL;
        }

	p_mtn->up_port = up_port;
	p_mtn->p_up = p_up;
	p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
	p_sw->multicast_priv = p_mtn;

        for (port = 0; port < p_sw->num_ports; port++) {
		if (up_port && port == up_port) {
			if (osm_mcast_tbl_is_port(p_tbl, mlid, port))
				active_ports++;
			continue;
		}

                if (!osm_mcast_tbl_is_port(p_tbl, mlid, port))
			continue;

		if (port == 0) {
			p_mtn->child_array[port] = OSM_MTREE_LEAF;
			continue;
		}

		p_rem_node = osm_node_get_remote_node(p_sw->p_node, port,
						      &rem_port);
		OSM_ASSERT(p_rem_node);
		if (osm_node_get_type(p_rem_node) != IB_NODE_TYPE_SWITCH) {
			p_mtn->child_array[port] = OSM_MTREE_LEAF;
			active_ports++;
		} else {
			p_child = rch_create_mtree_dfs(sm, p_rem_node->sw,
						       p_mtn, rem_port, mlid);
			if (p_child) {
				p_mtn->child_array[port] = p_child;
				active_ports++;
			}
		}
	}

	if (!active_ports) {
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3931: "
                        "Reached dead-end switch 0x%" PRIx64 " while creating "
			"multicast tree for MLID 0x%X\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			mlid);
		osm_mtree_destroy(p_mtn);
		p_mtn = NULL;
		p_sw->multicast_priv = NULL;
	}

	return p_mtn;
}

static int rch_create_mtree(osm_sm_t * sm, cl_qmap_t * p_rbox_map,
			    cl_qlist_t * p_mcm_ports_list, osm_mgrp_box_t *mbox)
{
	cl_map_item_t *p_map_item;
	cl_list_item_t *p_list_item;
	osm_switch_t *p_sw = NULL;
	osm_mcast_work_obj_t *p_wobj;
	osm_port_t *p_port;
	osm_mtree_node_t *p_mtn;
	osm_physp_t *p_physp;
	uint64_t guid;
	uint8_t port_num;

	/* Clear sw->multicast_priv */
	for (p_map_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_map_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item))
		((osm_switch_t *)p_map_item)->multicast_priv = NULL;

	/* Choose root switch for the tree */
	for (p_map_item = cl_qmap_head(p_rbox_map);
	     p_map_item != cl_qmap_end(p_rbox_map);
	     p_map_item = cl_qmap_next(p_map_item)) {
		osm_rch_box_t *p_rbox = (osm_rch_box_t *)p_map_item;
		if (p_rbox->mbox->root && p_rbox->mbox->root->p_sw) {
			guid =
			    osm_node_get_node_guid(p_rbox->mbox->root->p_sw->p_node);
			p_sw = (osm_switch_t *)cl_qmap_get(&sm->p_subn->sw_guid_tbl,
							   guid);
			if (p_sw &&
			    ((cl_map_item_t *)p_sw !=
			     cl_qmap_end(&sm->p_subn->sw_guid_tbl))) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Building multicast tree for MLID 0x%X "
					"from switch 0x%" PRIx64 "\n",
					mbox->mlid,
					cl_ntoh64(guid));
				break;
			}
		}
	}

	if (!p_sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 392C: Failed to find root switch for "
			"multicast tree of MLID 0x%X\n",
			mbox->mlid);
		return 1;
	}

	/* build tree from root */
	mbox->root = rch_create_mtree_dfs(sm, p_sw, NULL, 0, mbox->mlid);
	if (!mbox->root) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 392D: Failed to build multicast tree for "
			"MLID 0x%X\n",
			mbox->mlid);
		return 1;
	}

	/* Update tree information */
	for (p_list_item = cl_qlist_head(p_mcm_ports_list);
	     p_list_item != cl_qlist_end(p_mcm_ports_list);
	     p_list_item = cl_qlist_next(p_list_item)) {
		p_wobj = cl_item_obj(p_list_item, p_wobj, list_item);
                p_port = p_wobj->p_port;
                if (p_port->p_node->sw) {
                        p_sw = p_port->p_node->sw;
                        port_num = 0;
                } else {
                        p_physp = osm_physp_get_remote(p_port->p_physp);
			if (!p_physp) {
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Physical port 0x%" PRIx64 " could not "
					"be added to multicast tree of MLID "
					"0x%X\n",
					cl_ntoh64(osm_port_get_guid(p_port)),
					mbox->mlid);
				continue;
			}
                        OSM_ASSERT(p_physp->p_node);
                        p_sw = p_physp->p_node->sw;
                        port_num = osm_physp_get_port_num(p_physp);
                }

		p_mtn = (osm_mtree_node_t *)p_sw->multicast_priv;
		if (!p_mtn) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Physical port 0x%" PRIx64 " could not "
				"be added to multicast tree of MLID "
				"0x%X, switch 0x%" PRIx64 " is not in tree\n",
				cl_ntoh64(osm_port_get_guid(p_port)),
				mbox->mlid,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		if (p_wobj->is_receiver) {
			p_mtn->child_array[port_num] = OSM_MTREE_LEAF;
			/* Updated number of receivers */
			while (!p_mtn->num_receivers++ && p_mtn->p_up)
				p_mtn = p_mtn->p_up;
		} else {
			p_mtn->child_array[port_num] = OSM_MTREE_SENDER;
		}
	}

	return IB_SUCCESS;
}

static ib_api_status_t osm_rch_mcast_stree(void *context, osm_mgrp_box_t *mbox)
{
	osm_ucast_mgr_t *p_mgr = (struct osm_ucast_mgr *)context;
	osm_sm_t *sm = p_mgr->sm;
	osm_routing_chain_mgr_t *p_rch_mgr;
	ib_api_status_t status_f = IB_SUCCESS;
	cl_qmap_t *p_rbox_map;
	cl_qmap_t mcm_ports_map;
	cl_qlist_t mcm_ports_list;
	cl_map_item_t * p_box_item;
	osm_subn_t * p_main_subn;
	struct osm_routing_engine *re;
	uint16_t max_used_mlid;

	OSM_LOG_ENTER(sm->p_log);

	p_rch_mgr = &sm->routing_chain_mgr;
	p_rbox_map = &p_rch_mgr->p_rch_rboxes[mbox->mlid - IB_LID_MCAST_START_HO];
	p_main_subn = sm->ucast_mgr.p_subn;
	max_used_mlid = osm_subn_get_max_mlid(sm->ucast_mgr.p_subn);

	if (p_rch_mgr->is_itr) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Building multicast tree "
				"for ITR subnet\n");
		/* Cleanup data of non-ITR multicast routing */
		osm_mtree_destroy(mbox->root);
		mbox->root = NULL;
	} else {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Building multicast tree "
				"for non-ITR subnet\n");
		/*
		 * Cleanup data of ITR multicast routing and use default
		 * multicast routing
		 */
		rch_rbox_map_destroy(sm, p_rbox_map);
		
		re = p_rch_mgr->main_topo_routing_engine_used;
		if (re && re->mcast_build_stree)
			status_f = re->mcast_build_stree(re->context, mbox);
		else if (p_main_subn->opt.const_multicast)
			status_f =
			    osm_mcast_mgr_build_spanning_tree_const(p_main_subn,
								    mbox);
		else
			status_f =
			    osm_mcast_mgr_build_spanning_tree(p_main_subn,
							      mbox);

		goto Exit;
	}

	if (osm_mcast_make_port_list_and_map(&mcm_ports_list, &mcm_ports_map, mbox)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3920: "
			"Unable to initilize map for multicast members for "
			"mlid 0x%x\n",
			mbox->mlid);
		status_f = IB_ERROR;
		goto Exit;
        }

	if (cl_qlist_count(&mcm_ports_list) < 2) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"MLID 0x%X has %u members - nothing to do\n",
			mbox->mlid, cl_qlist_count(&mcm_ports_list));
		osm_purge_mtree(p_main_subn, mbox);
		goto Exit_Free;
	}

	/* Create mbox object for each sub topology */
	if (rch_create_rbox_map(sm, p_rbox_map, &mcm_ports_list,
				    mbox->mlid)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3922: "
			"Unable to create mbox objects for mlid 0x%x\n",
			mbox->mlid);
		status_f = IB_ERROR;
		goto Exit_Free;
	}

	if (!cl_qmap_count(p_rbox_map)) {
		status_f = IB_SUCCESS;
		goto Exit_Free;
	}

	/* Add border switches as members */
	if (rch_add_borders_to_mgroups(sm, p_rbox_map, &mcm_ports_map)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3926: "
			"Unable to add border switches to mbox obejects for "
			"mlid 0x%x\n",
			mbox->mlid);
		rch_rbox_map_destroy(sm, p_rbox_map);
		status_f = IB_ERROR;
		goto Exit_Free;
	}

	/*
	 * For each topology, call it's routing engine's multicast algorithm
	 * function with the vritual mbox created for it.
	 */
	for (p_box_item = cl_qmap_head(p_rbox_map);
	     p_box_item != cl_qmap_end(p_rbox_map);
	     p_box_item = cl_qmap_next(p_box_item)) {

		osm_rch_box_t *rbox;
		osm_topology_t *p_topology;
		osm_subn_t * p_topology_subn;
		ib_api_status_t status = IB_SUCCESS;
		rbox = (osm_rch_box_t *)p_box_item;

		p_topology = osm_topo_mgr_get_topo(sm->p_subn->p_osm,
						   rbox->topology_id);
		p_topology_subn = p_topology->p_subn;

		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Building multicast spanning tree for topology "
			"id %" PRIu64 "\n",
			p_topology->id);

		status =
		    osm_mcast_mgr_alloc_mfts(p_topology_subn, max_used_mlid);
		if (status != IB_SUCCESS) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3927: "
				"Unable to allocate MFT table (%s) for "
				"mlid 0x%x for topology id %" PRIu64 "\n",
				ib_get_err_str(status), mbox->mlid,
				p_topology->id);
			goto Exit_Free;
		}

		osm_mcast_mgr_clear_mft(p_topology_subn, mbox->mlid);
		re = p_topology_subn->routing_engine_used;

		mcast_mgr_decrease_mtree_root_weight(p_topology_subn,
						     rbox->mbox);

		if (re && re->mcast_build_stree)
			status = re->mcast_build_stree(re->context, rbox->mbox);
		else if (p_topology_subn->opt.const_multicast)
			status =
			    osm_mcast_mgr_build_spanning_tree_const(p_topology_subn,
								    rbox->mbox);
		else
			status =
			    osm_mcast_mgr_build_spanning_tree(p_topology_subn,
							      rbox->mbox);

		if (status != IB_SUCCESS) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3928: "
				"Unable to create spanning tree (%s) for mlid "
				"0x%x for topology id %" PRIu64 "\n",
				ib_get_err_str(status), mbox->mlid,
				p_topology->id);
			rch_rbox_map_destroy(sm, p_rbox_map);
			status_f = IB_UNKNOWN_ERROR;
			goto Exit_Free;
		} else
			mcast_mgr_increase_mtree_root_weight(p_topology_subn,
							     rbox->mbox);
	}

	/* Merge MFT tables created on each sub-topology */
	rch_merge_mft(sm, p_rbox_map, mbox->mlid);

	/* Remove the border switches that were added to the multicast groups */
	rch_remove_borders_from_mft(sm, p_rbox_map, mbox->mlid);

	/* Build multicast tree structure */
	if (rch_create_mtree(sm, p_rbox_map, &mcm_ports_list, mbox)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 392E: Failed to build multicast tree for "
			"MLID 0x%X\n",
			mbox->mlid);
		goto Exit_Free;
	}

Exit_Free:
	/* Destroy ports list */
	osm_mcast_drop_port_list(&mcm_ports_list);

Exit:
	if (status_f != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 3929: "
			"Unable to create spanning tree (%s) for mlid "
			"0x%x\n", ib_get_err_str(status_f), mbox->mlid);
	}

        OSM_LOG_EXIT(sm->p_log);
        return status_f;
}

int osm_ucast_chain_setup(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	r->context = &osm->sm.ucast_mgr;
	r->build_lid_matrices = ucast_dummy_build_lid_matrices;
	r->ucast_build_fwd_tables = osm_rch_ucast_process;
	r->path_sl = osm_rch_path_sl;
	r->update_sl2vl = osm_rch_update_sl2vl;
	r->update_vlarb = osm_rch_update_vlarb;
	r->mcast_build_stree = osm_rch_mcast_stree;
	return 0;
}

uint32_t osm_rch_engines_count(IN struct osm_opensm *p_osm)
{
	osm_routing_chain_mgr_t *p_mgr = &p_osm->sm.routing_chain_mgr;
	return cl_qlist_count(&p_mgr->engine_list);
}
