/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2019 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	This module is responsible for handling rejected SA requests
 *	that needs to be reported.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SA_SECURITY_REPORT_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_sa.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_inform.h>

#define MAD_SOURCE_ADDR_STR_LEN 48

static char * get_mad_source_addr_str(osm_mad_addr_t *p_mad_addr, char *buf, size_t n)
{
	if (p_mad_addr->addr_type.gsi.global_route)
		snprintf(buf, n, "LID %u GUID 0x%" PRIx64,
			 cl_ntoh16(p_mad_addr->dest_lid),
			 cl_ntoh64(p_mad_addr->addr_type.gsi.grh_info.dest_gid.unicast.interface_id));
	else
		snprintf(buf, n, "LID %u", cl_ntoh16(p_mad_addr->dest_lid));

	return buf;
}

static void report_sa_notice(osm_mad_addr_t *p_mad_addr, ib_sa_mad_t *p_sa_mad, osm_sa_t *sa)
{
	struct _vendor_data_details ntc_data;
	osm_alias_guid_t *p_alias_guid;

	ntc_data.v_data_details.ntc_1300.slid = p_mad_addr->dest_lid;
	ntc_data.v_data_details.ntc_1300.attr_id = p_sa_mad->attr_id;
	ntc_data.v_data_details.ntc_1300.method = p_sa_mad->method;
	ntc_data.v_data_details.ntc_1300.sa_key = p_sa_mad->sm_key;

	CL_PLOCK_ACQUIRE(sa->p_lock);
	p_alias_guid = osm_get_alias_guid_by_mad_addr(sa->p_log, sa->p_subn,
						      p_mad_addr);
	if (p_alias_guid) {
		ntc_data.v_data_details.ntc_1300.sgid.unicast.prefix = sa->p_subn->opt.subnet_prefix;
		ntc_data.v_data_details.ntc_1300.sgid.unicast.interface_id = p_alias_guid->alias_guid;
		ntc_data.v_data_details.ntc_1300.physp_sguid = p_alias_guid->base_port_guid;
	} else {
		ntc_data.v_data_details.ntc_1300.sgid.unicast.prefix = 0;
		ntc_data.v_data_details.ntc_1300.sgid.unicast.interface_id = 0;
		ntc_data.v_data_details.ntc_1300.physp_sguid = 0;
	}
	ntc_data.trap_num = SA_VS_TRAP_NUM_KEY_VIOLATION;
	osm_report_sa_security_notice(sa->p_subn, &ntc_data);
	CL_PLOCK_RELEASE(sa->p_lock);
}

void osm_sa_security_report_sa_key(IN void *context, IN void *data)
{
	char mad_src_str[MAD_SOURCE_ADDR_STR_LEN];
	osm_sa_t *sa = context;
	osm_madw_t *p_madw = data;
	osm_mad_addr_t *p_mad_addr;
	ib_sa_mad_t *p_sa_mad;

	OSM_ASSERT(sa);

	OSM_LOG_ENTER(sa->p_log);

	p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	if (!sa->p_subn->opt.sa_enhanced_trust_model ||
	    sa_etm_is_log_print(sa->sm, p_mad_addr->dest_lid)) {
		p_sa_mad = osm_madw_get_sa_mad_ptr(p_madw);

		report_sa_notice(p_mad_addr, p_sa_mad, sa);
		OSM_LOG(sa->p_log, OSM_LOG_ERROR, "ERR 8B04: "
			"Non-Zero MAD SM_Key: 0x%" PRIx64 " != SM_Key: 0x%"
			PRIx64 "; SA MAD ignored for method 0x%X attribute 0x%X (%s) "
			"From %s\n", cl_ntoh64(p_sa_mad->sm_key),
			cl_ntoh64(sa->p_subn->opt.sa_key), p_sa_mad->method,
			cl_ntoh16(p_sa_mad->attr_id), ib_get_sa_attr_str(p_sa_mad->attr_id),
			get_mad_source_addr_str(p_mad_addr, mad_src_str, sizeof(mad_src_str)));
	}

	OSM_LOG_EXIT(sa->p_log)
}

void osm_sa_security_report_untrusted(IN void *context, IN void *data)
{
	char mad_src_str[MAD_SOURCE_ADDR_STR_LEN];
	osm_sa_t *sa = context;
	osm_madw_t *p_madw = data;
	osm_mad_addr_t *p_mad_addr;
	ib_sa_mad_t *p_sa_mad;

	OSM_ASSERT(sa);

	OSM_LOG_ENTER(sa->p_log);

	p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	if (!sa->p_subn->opt.sa_enhanced_trust_model ||
	    sa_etm_is_log_print(sa->sm, p_mad_addr->dest_lid)) {
		p_sa_mad = osm_madw_get_sa_mad_ptr(p_madw);

		report_sa_notice(p_mad_addr, p_sa_mad, sa);
		OSM_LOG(sa->p_log, OSM_LOG_ERROR,
			"ERR 8B12: Dropping untrusted request %s from %s"
			" according to SA enhanced trust model\n",
			ib_get_sa_attr_str(p_sa_mad->attr_id),
			get_mad_source_addr_str(p_mad_addr, mad_src_str, sizeof(mad_src_str)));
	}

	OSM_LOG_EXIT(sa->p_log)
}
