/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2011 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2008 Xsigo Systems Inc.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_sm_t.
 * This object represents the SM Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <complib/cl_thread.h>
#include <complib/cl_crc.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SM_C
#include <opensm/osm_sm.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_perfmgr.h>
#include <opensm/osm_opensm.h>

#define  OSM_SM_INITIAL_TID_VALUE 0x1233

extern void osm_lft_rcv_process(IN void *context, IN void *data);
extern void osm_mft_rcv_process(IN void *context, IN void *data);
extern void osm_nd_rcv_process(IN void *context, IN void *data);
extern void osm_ni_rcv_process(IN void *context, IN void *data);
extern void osm_pkey_rcv_process(IN void *context, IN void *data);
extern void osm_pi_rcv_process(IN void *context, IN void *data);
extern void osm_gi_rcv_process(IN void *context, IN void *data);
extern void osm_slvl_rcv_process(IN void *context, IN void *p_data);
extern void osm_sminfo_rcv_process(IN void *context, IN void *data);
extern void osm_si_rcv_process(IN void *context, IN void *data);
extern void osm_trap_rcv_process(IN void *context, IN void *data);
extern void osm_vla_rcv_process(IN void *context, IN void *data);
extern void osm_mlnx_epi_rcv_process(IN void *context, IN void *data);
extern void osm_vpi_rcv_process(IN void *context, IN void *data);
extern void osm_vi_rcv_process(IN void *context, IN void *data);
extern void osm_vps_rcv_process(IN void *context, IN void *data);
extern void osm_general_info_rcv_process(IN void *context, IN void *data);
extern void osm_vni_rcv_process(IN void *context, IN void *data);
extern void osm_vpkey_rcv_process(IN void *context, IN void *data);
extern void osm_vnd_rcv_process(IN void *context, IN void *data);
extern void osm_ri_rcv_process(IN void *context, IN void *data);
extern void osm_adj_rtr_table_rcv_process(IN void *context, IN void *data);
extern void osm_next_hop_rtr_table_rcv_process(IN void *context, IN void *data);
extern void osm_qos_config_sl_rcv_process(IN void *context, IN void *data);

extern void osm_ar_mad_rcv_process(IN void *context, IN void *data);
extern void osm_rn_mad_rcv_process(IN void *context, IN void *data);
extern void osm_plft_mad_rcv_process(IN void *context, IN void *data);
extern void osm_mlnx_eni_rcv_process(IN void *context, IN void *data);

extern void osm_state_mgr_process(IN osm_sm_t * sm, IN osm_signal_t signal);
extern void osm_sm_state_mgr_polling_callback(IN void *context);
extern void osm_sm_dump_stat_callback(IN void *context);
extern int parse_verbose_bypass_policy(struct osm_opensm *osm_ptr);
extern void osm_spst_rcv_process(IN void *context, IN void *data);

static void sm_process(osm_sm_t * sm, osm_signal_t signal)
{
#ifdef ENABLE_OSM_PERF_MGR
	if (signal == OSM_SIGNAL_PERFMGR_SWEEP)
		osm_perfmgr_process(&sm->p_subn->p_osm->perfmgr);
	else
#endif
		osm_state_mgr_process(sm, signal);
}

#define OPT_OFFSET(opt) offsetof(osm_subn_opt_t, opt)

static osm_crc_element_t crc_tbl[] = {
	{ OSM_CRC_ROOT_GUID_FILE, OPT_OFFSET(root_guid_file), 0, FALSE, TRUE },
	{ OSM_CRC_CN_GUID_FILE, OPT_OFFSET(cn_guid_file), 0, FALSE, TRUE },
	{ OSM_CRC_IO_GUID_FILE, OPT_OFFSET(io_guid_file), 0, FALSE, TRUE },
	{ OSM_CRC_IDS_GUID_FILE, OPT_OFFSET(ids_guid_file), 0, FALSE, TRUE },
	{ OSM_CRC_GUID_ROUTING_ORDER_FILE, OPT_OFFSET(guid_routing_order_file), 0, FALSE, TRUE },
	{ OSM_CRC_PORT_PROF_IGNORE_FILE, OPT_OFFSET(port_prof_ignore_file), 0, FALSE, TRUE },
	{ OSM_CRC_HOP_WEIGHTS_FILE, OPT_OFFSET(hop_weights_file), 0, FALSE, TRUE },
	{ OSM_CRC_PORT_SEARCH_ORDERING_FILE, OPT_OFFSET(port_search_ordering_file), 0, FALSE, TRUE },
	{ OSM_CRC_PGRP_POLICY_FILE, OPT_OFFSET(pgrp_policy_file), 0, FALSE, TRUE },
	{ OSM_CRC_RCH_POLICY_FILE, OPT_OFFSET(rch_policy_file), 0, FALSE, TRUE },
	{ OSM_CRC_TOPO_POLICY_FILE, OPT_OFFSET(topo_policy_file), 0, FALSE, TRUE },
	{ OSM_CRC_PARTITION_FILE, OPT_OFFSET(partition_config_file), 0, FALSE, FALSE },
	{ OSM_CRC_CONFIG_FILE, OPT_OFFSET(config_file), 0, FALSE, FALSE },
	{ OSM_CRC_PREFIX_ROUTES_FILE, OPT_OFFSET(prefix_routes_file), 0, FALSE, FALSE },
	{ OSM_CRC_PER_MODULE_LOG_FILE, OPT_OFFSET(per_module_logging_file), 0, FALSE, FALSE },
	{ OSM_CRC_HEALTH_POLICY_FILE, OPT_OFFSET(hm_ports_health_policy_file), 0, FALSE, TRUE },
	{ OSM_CRC_HELDBACK_FILE, OPT_OFFSET(held_back_sw_file), 0, FALSE, FALSE },
	{ OSM_CRC_QOS_POLICY_FILE, OPT_OFFSET(qos_policy_file), 0, FALSE, FALSE },
	{ OSM_CRC_ENHANCED_QOS_POLICY_FILE, OPT_OFFSET(enhanced_qos_policy_file), 0, FALSE, FALSE},
	{ OSM_CRC_MC_ROOTS_FILE, OPT_OFFSET(mc_roots_file), 0, FALSE, TRUE },
	{ OSM_CRC_ROUTE_POLICY_FILE, OPT_OFFSET(router_policy_file), 0, FALSE, FALSE },
	{ OSM_CRC_VERBOSE_BYPASS_POLICY_FILE, OPT_OFFSET(verbose_bypass_policy_file), 0, FALSE, FALSE },
	{ OSM_CRC_NODE_NAME_MAP_FILE, OPT_OFFSET(node_name_map_name), 0, FALSE, TRUE}
	/* node_name_map_name invalidate cache set to true due to the port group feature, which allows regex to
	 * create port groups which effect the routing */
};

static void sm_sweeper(IN void *p_ptr)
{
	ib_api_status_t status;
	osm_sm_t * p_sm = p_ptr;
	unsigned signals, i;

	OSM_LOG_ENTER(p_sm->p_log);

	while (p_sm->thread_state == OSM_THREAD_STATE_RUN) {
		/*
		 * Wait on the event with a timeout.
		 * Sweeps may be initiated "off schedule" by simply
		 * signaling the event.
		 */
		status = cl_event_wait_on(&p_sm->signal_event,
					  EVENT_NO_TIMEOUT, TRUE);

		if (status == CL_SUCCESS)
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
				"Off schedule sweep signalled\n");
		else {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2E01: "
				"Event wait failed (%s)\n",
				CL_STATUS_MSG(status));
			continue;
		}

		if (osm_exit_flag)
			break;

		cl_spinlock_acquire(&p_sm->signal_lock);
		signals = p_sm->signal_mask;
		p_sm->signal_mask = 0;
		cl_spinlock_release(&p_sm->signal_lock);

		for (i = 0; signals; signals >>= 1, i++)
			if (signals & 1)
				sm_process(p_sm, i);
	}

	OSM_LOG_EXIT(p_sm->p_log);
}

static void sm_sweep(void *arg)
{
	osm_sm_t *sm = arg;

	/*  do the sweep only if we are in MASTER state */
	if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER ||
	    sm->p_subn->sm_state == IB_SMINFO_STATE_DISCOVERING)
		osm_sm_signal(sm, OSM_SIGNAL_SWEEP);
}

static void sweep_fail_process(IN void *context, IN void *p_data)
{
	osm_sm_t *sm = context;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "light sweep failed\n");
	sm->p_subn->force_heavy_sweep = TRUE;
}

void osm_sm_construct(IN osm_sm_t * p_sm)
{
	memset(p_sm, 0, sizeof(*p_sm));
	p_sm->thread_state = OSM_THREAD_STATE_NONE;
	p_sm->sm_trans_id = OSM_SM_INITIAL_TID_VALUE;
	cl_spinlock_construct(&p_sm->signal_lock);
	cl_spinlock_construct(&p_sm->state_lock);
	cl_timer_construct(&p_sm->polling_timer);
	cl_event_construct(&p_sm->signal_event);
	cl_event_construct(&p_sm->subnet_up_event);
	cl_event_wheel_construct(&p_sm->trap_aging_tracker);
	cl_event_wheel_construct(&p_sm->reboot_tracker);
	cl_event_wheel_construct(&p_sm->sa_etm_drop_tracker);
	cl_thread_construct(&p_sm->sweeper);
	osm_sm_mad_ctrl_construct(&p_sm->mad_ctrl);
	osm_lid_mgr_construct(&p_sm->lid_mgr);
	osm_ucast_mgr_construct(&p_sm->ucast_mgr);
	osm_port_grp_mgr_construct(&p_sm->port_group_mgr);
	osm_cc_mgr_construct(&p_sm->cc_mgr);
	osm_rch_mgr_construct(&p_sm->routing_chain_mgr);
	osm_router_policy_construct(&p_sm->router_policy_mgr);
	osm_verbose_bypass_construct(&p_sm->verbose_bypass);
}

void osm_sm_shutdown(IN osm_sm_t * p_sm)
{
	boolean_t signal_event = FALSE;

	OSM_LOG_ENTER(p_sm->p_log);

	/*
	 * Signal our threads that we're leaving.
	 */
	if (p_sm->thread_state != OSM_THREAD_STATE_NONE)
		signal_event = TRUE;

	p_sm->thread_state = OSM_THREAD_STATE_EXIT;

	/*
	 * Don't trigger unless event has been initialized.
	 * Destroy the thread before we tear down the other objects.
	 */
	if (signal_event)
		cl_event_signal(&p_sm->signal_event);

	cl_timer_stop(&p_sm->polling_timer);
	cl_timer_stop(&p_sm->sweep_timer);
	cl_thread_destroy(&p_sm->sweeper);

	/*
	 * Always destroy controllers before the corresponding
	 * receiver to guarantee that all callbacks from the
	 * dispatcher are complete.
	 */
	osm_sm_mad_ctrl_destroy(&p_sm->mad_ctrl);
	cl_disp_unregister(p_sm->ni_disp_h);
	cl_disp_unregister(p_sm->pi_disp_h);
	cl_disp_unregister(p_sm->vpi_disp_h);
	cl_disp_unregister(p_sm->vi_disp_h);
	cl_disp_unregister(p_sm->gi_disp_h);
	cl_disp_unregister(p_sm->si_disp_h);
	cl_disp_unregister(p_sm->nd_disp_h);
	cl_disp_unregister(p_sm->lft_disp_h);
	cl_disp_unregister(p_sm->mft_disp_h);
	cl_disp_unregister(p_sm->sm_info_disp_h);
	cl_disp_unregister(p_sm->sm_info_sminfo_get_disp_h );
	cl_disp_unregister(p_sm->trap_disp_h);
	cl_disp_unregister(p_sm->slvl_disp_h);
	cl_disp_unregister(p_sm->vla_disp_h);
	cl_disp_unregister(p_sm->pkey_disp_h);
	cl_disp_unregister(p_sm->mlnx_epi_disp_h);
	cl_disp_unregister(p_sm->sweep_fail_disp_h);
	cl_disp_unregister(p_sm->vps_disp_h);
	cl_disp_unregister(p_sm->general_info_disp_h);
	cl_disp_unregister(p_sm->vni_disp_h);
	cl_disp_unregister(p_sm->vpkey_disp_h);
	cl_disp_unregister(p_sm->vnd_disp_h);
	cl_disp_unregister(p_sm->qos_config_sl_h);
	cl_disp_unregister(p_sm->ar_mad_disp_h);
	cl_disp_unregister(p_sm->rn_mad_disp_h);

	OSM_LOG_EXIT(p_sm->p_log);
}

void osm_sm_destroy(IN osm_sm_t * p_sm)
{
	OSM_LOG_ENTER(p_sm->p_log);
	osm_enhanced_qos_mgr_destroy(&p_sm->enhanced_qos_mgr);
	osm_rch_mgr_destroy(&p_sm->routing_chain_mgr);
	osm_lid_mgr_destroy(&p_sm->lid_mgr);
	osm_virt_mgr_destroy(&p_sm->virt_mgr, p_sm);
	osm_ucast_mgr_destroy(&p_sm->ucast_mgr);
	osm_port_grp_mgr_destroy(&p_sm->port_group_mgr);
	osm_topo_mgr_destroy(&p_sm->topology_mgr);
	osm_activity_mgr_destroy(&p_sm->activity_mgr);
	osm_router_policy_destroy(&p_sm->router_policy_mgr);
	cl_event_wheel_destroy(&p_sm->reboot_tracker);
	cl_event_wheel_destroy(&p_sm->trap_aging_tracker);
	cl_event_wheel_destroy(&p_sm->sa_etm_drop_tracker);
	cl_timer_destroy(&p_sm->sweep_timer);
	cl_timer_destroy(&p_sm->polling_timer);
	cl_timer_destroy(&p_sm->dump_statistics_timer);
	cl_event_destroy(&p_sm->signal_event);
	cl_event_destroy(&p_sm->subnet_up_event);
	cl_spinlock_destroy(&p_sm->signal_lock);
	cl_spinlock_destroy(&p_sm->state_lock);
	free(p_sm->mlids_req);
	osm_log_v2(p_sm->p_log, OSM_LOG_SYS, FILE_ID, "Exiting SM\n");	/* Format Waived */
	osm_verbose_bypass_destroy(p_sm, &p_sm->verbose_bypass);
	osm_cap_mgr_destroy(&p_sm->cap_mgr);
	OSM_LOG_EXIT(p_sm->p_log);
}

ib_api_status_t osm_sm_init(IN osm_sm_t * p_sm, IN osm_subn_t * p_subn,
			    IN osm_db_t * p_db, IN osm_vendor_t * p_vendor,
			    IN osm_mad_pool_t * p_mad_pool,
			    IN osm_vl15_t * p_vl15, IN osm_log_t * p_log,
			    IN osm_stats_t * p_stats,
			    IN cl_dispatcher_t * p_disp,
			    IN cl_dispatcher_t * p_sminfo_get_disp, IN cl_plock_t * p_lock)
{
	ib_api_status_t status;

	OSM_LOG_ENTER(p_log);

	p_sm->p_subn = p_subn;
	p_sm->p_db = p_db;
	p_sm->p_vendor = p_vendor;
	p_sm->p_mad_pool = p_mad_pool;
	p_sm->p_vl15 = p_vl15;
	p_sm->p_log = p_log;
	p_sm->p_disp = p_disp;
	p_sm->p_sminfo_get_disp = p_sminfo_get_disp;
	p_sm->p_lock = p_lock;

	status = cl_spinlock_init(&p_sm->signal_lock);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_spinlock_init(&p_sm->state_lock);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_event_init(&p_sm->signal_event, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_event_init(&p_sm->subnet_up_event, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_timer_init(&p_sm->sweep_timer, sm_sweep, p_sm);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_timer_init(&p_sm->polling_timer,
			       osm_sm_state_mgr_polling_callback, p_sm);
	if (status != CL_SUCCESS)
		goto Exit;

	status = cl_timer_init(&p_sm->dump_statistics_timer,
			       osm_sm_dump_stat_callback, p_sm);
	if (status != CL_SUCCESS)
		goto Exit;


	p_sm->mlids_req_max = 0;
	p_sm->mlids_req = malloc((IB_LID_MCAST_END_HO - IB_LID_MCAST_START_HO +
				  1) * sizeof(p_sm->mlids_req[0]));
	if (!p_sm->mlids_req)
		goto Exit;
	memset(p_sm->mlids_req, 0,
	       (IB_LID_MCAST_END_HO - IB_LID_MCAST_START_HO +
		1) * sizeof(p_sm->mlids_req[0]));

	status = osm_sm_mad_ctrl_init(&p_sm->mad_ctrl, p_sm->p_subn,
				      p_sm->p_mad_pool, p_sm->p_vl15,
				      p_sm->p_vendor,
				      p_log, p_stats, p_lock, p_disp, p_sminfo_get_disp);
	if (status != IB_SUCCESS)
		goto Exit;

	status = cl_event_wheel_init(&p_sm->trap_aging_tracker);
	if (status != IB_SUCCESS)
		goto Exit;

	status = cl_event_wheel_wnd_init(&p_sm->reboot_tracker);
	if (status != IB_SUCCESS)
		goto Exit;

	status = cl_event_wheel_init(&p_sm->sa_etm_drop_tracker);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_lid_mgr_init(&p_sm->lid_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_ucast_mgr_init(&p_sm->ucast_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_rch_mgr_init(&p_sm->routing_chain_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_virt_mgr_init(&p_sm->virt_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_cap_mgr_init(&p_sm->cap_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_activity_mgr_init(&p_sm->activity_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = osm_enhanced_qos_mgr_init(&p_sm->enhanced_qos_mgr, p_sm);
	if (status != IB_SUCCESS)
		goto Exit;

	status = IB_INSUFFICIENT_RESOURCES;
	p_sm->sweep_fail_disp_h = cl_disp_register(p_disp,
						   OSM_MSG_LIGHT_SWEEP_FAIL,
						   sweep_fail_process, p_sm);
	if (p_sm->sweep_fail_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->ni_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_NODE_INFO,
					   osm_ni_rcv_process, p_sm);
	if (p_sm->ni_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->pi_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_PORT_INFO,
					   osm_pi_rcv_process, p_sm);
	if (p_sm->pi_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vpi_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VPORT_INFO,
					   osm_vpi_rcv_process, p_sm);
	if (p_sm->vpi_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vi_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VIRTUALIZATION_INFO,
					   osm_vi_rcv_process, p_sm);
	if (p_sm->vi_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vni_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VNODE_INFO,
					   osm_vni_rcv_process, p_sm);
	if (p_sm->vni_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vnd_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VNODE_DESC,
					   osm_vnd_rcv_process, p_sm);
	if (p_sm->vnd_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vpkey_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VPKEY,
					   osm_vpkey_rcv_process, p_sm);
	if (p_sm->vpkey_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->gi_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_GUID_INFO,
					   osm_gi_rcv_process, p_sm);
	if (p_sm->gi_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->si_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_SWITCH_INFO,
					   osm_si_rcv_process, p_sm);
	if (p_sm->si_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->nd_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_NODE_DESC,
					   osm_nd_rcv_process, p_sm);
	if (p_sm->nd_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->lft_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_LFT,
					    osm_lft_rcv_process, p_sm);
	if (p_sm->lft_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->mft_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_MFT,
					    osm_mft_rcv_process, p_sm);
	if (p_sm->mft_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->sm_info_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_SM_INFO,
						osm_sminfo_rcv_process, p_sm);
	if (p_sm->sm_info_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->sm_info_sminfo_get_disp_h = cl_disp_register(p_sminfo_get_disp, OSM_MSG_MAD_SM_INFO,
						   osm_sminfo_rcv_process, p_sm);
	if (p_sm->sm_info_sminfo_get_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->trap_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_NOTICE,
					     osm_trap_rcv_process, p_sm);
	if (p_sm->trap_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->slvl_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_SLVL,
					     osm_slvl_rcv_process, p_sm);
	if (p_sm->slvl_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vla_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VL_ARB,
					    osm_vla_rcv_process, p_sm);
	if (p_sm->vla_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->pkey_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_PKEY,
					     osm_pkey_rcv_process, p_sm);
	if (p_sm->pkey_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->mlnx_epi_disp_h = cl_disp_register(p_disp,
						 OSM_MSG_MAD_MLNX_EXT_PORT_INFO,
						 osm_mlnx_epi_rcv_process, p_sm);
	if (p_sm->mlnx_epi_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->vps_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_VPORT_STATE,
					   osm_vps_rcv_process, p_sm);
	if (p_sm->vps_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->general_info_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_GENERAL_INFO,
					   osm_general_info_rcv_process, p_sm);
	if (p_sm->general_info_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->ri_disp_h = cl_disp_register(p_disp, OSM_MSG_MAD_ROUTER_INFO,
                                           osm_ri_rcv_process, p_sm);
	if (p_sm->ri_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->adj_rtr_disp_h = cl_disp_register(p_disp,
					   OSM_MSG_MAD_RTR_ADJ_SITE_LOCAL_TBL,
					   osm_adj_rtr_table_rcv_process, p_sm);
	if (p_sm->adj_rtr_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->next_hop_rtr_disp_h = cl_disp_register(p_disp,
					   OSM_MSG_MAD_RTR_NEXT_HOP_TBL,
					   osm_next_hop_rtr_table_rcv_process, p_sm);
	if (p_sm->next_hop_rtr_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;
	p_sm->qos_config_sl_h = cl_disp_register(p_disp,
					OSM_MSG_MAD_QOS_CONFIG_SL,
					osm_qos_config_sl_rcv_process, p_sm);
	if (p_sm->qos_config_sl_h == CL_DISP_INVALID_HANDLE)
		goto Exit;
	p_sm->sps_disp_h = cl_disp_register(p_disp,
					OSM_MSG_MAD_SWITCH_PORT_STATE_TBL,
					osm_spst_rcv_process, p_sm);
	if (p_sm->sps_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->ar_mad_disp_h =
	    cl_disp_register(p_disp, OSM_MSG_MAD_ADAPTIVE_ROUTING_MAD,
			     osm_ar_mad_rcv_process, p_sm);
	if (p_sm->ar_mad_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->rn_mad_disp_h =
	    cl_disp_register(p_disp, OSM_MSG_MAD_ROUTING_NOTIFICATION_MAD,
			     osm_rn_mad_rcv_process, p_sm);
	if (p_sm->rn_mad_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->plft_mad_disp_h =
	    cl_disp_register(p_disp, OSM_MSG_MAD_PLFT_MAD,
			     osm_plft_mad_rcv_process, p_sm);
	if (p_sm->plft_mad_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	p_sm->eni_disp_h =
	    cl_disp_register(p_disp, OSM_MSG_MAD_EXTENDED_NODE_INFO,
			     osm_mlnx_eni_rcv_process, p_sm);
	if (p_sm->eni_disp_h == CL_DISP_INVALID_HANDLE)
		goto Exit;

	cl_spinlock_acquire(&p_sm->state_lock);
	p_subn->sm_state = p_subn->opt.sm_inactive ?
	    IB_SMINFO_STATE_NOTACTIVE : IB_SMINFO_STATE_DISCOVERING;
	cl_spinlock_release(&p_sm->state_lock);
	osm_log_sm_state(p_sm);

	/* Update SM active priority */
	osm_set_active_sm_priority(p_sm);

	/*
	 * Now that the component objects are initialized, start
	 * the sweeper thread if the user wants sweeping.
	 */
	p_sm->thread_state = OSM_THREAD_STATE_RUN;
	status = cl_thread_init(&p_sm->sweeper, sm_sweeper, p_sm,
				"opensm sweeper");
	if (status != IB_SUCCESS)
		goto Exit;

	p_sm->files_crc = crc_tbl;
	p_sm->crc_changed = 0;

	if(p_sm->p_subn->opt.osm_stats_interval)
		cl_timer_start(&p_sm->dump_statistics_timer,
			       p_sm->p_subn->opt.osm_stats_interval);

	parse_verbose_bypass_policy(p_sm->p_subn->p_osm);

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}

void osm_sm_signal(osm_sm_t * p_sm, osm_signal_t signal)
{
	cl_spinlock_acquire(&p_sm->signal_lock);
	p_sm->signal_mask |= 1 << signal;
	cl_event_signal(&p_sm->signal_event);
	cl_spinlock_release(&p_sm->signal_lock);
}

void osm_sm_sweep(IN osm_sm_t * p_sm)
{
	OSM_LOG_ENTER(p_sm->p_log);
	osm_sm_signal(p_sm, OSM_SIGNAL_SWEEP);
	OSM_LOG_EXIT(p_sm->p_log);
}

ib_api_status_t osm_sm_bind(IN osm_sm_t * p_sm, IN ib_net64_t port_guid)
{
	ib_api_status_t status;

	OSM_LOG_ENTER(p_sm->p_log);

	status = osm_sm_mad_ctrl_bind(&p_sm->mad_ctrl, port_guid);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2E10: "
			"SM MAD Controller bind failed (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_sm->p_log);
	return status;
}

void osm_sm_reroute_mlid(osm_sm_t * sm, ib_net16_t mlid)
{
	mlid = cl_ntoh16(mlid) - IB_LID_MCAST_START_HO;
	sm->mlids_req[mlid] = 1;
	if (sm->mlids_req_max < mlid)
		sm->mlids_req_max = mlid;
	osm_sm_signal(sm, OSM_SIGNAL_IDLE_TIME_PROCESS_REQUEST);
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "rerouting requested for MLID 0x%x\n",
		mlid + IB_LID_MCAST_START_HO);
}

void osm_sm_reroute_all_mlids(osm_sm_t * sm)
{
	sm->p_subn->reroute_all_mlids = TRUE;
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "rerouting requested for all MLIDs\n");
	osm_sm_signal(sm, OSM_SIGNAL_IDLE_TIME_PROCESS_REQUEST);
}

static void _osm_set_active_sm_priority(osm_sm_t * sm, uint8_t * new_priority)
{
	osm_subn_opt_t *p_opt = &sm->p_subn->opt;
	uint8_t old_pri = p_opt->sm_priority;
	cl_qmap_t *p_sm_tbl;
	osm_remote_sm_t *p_rsm;


	if (new_priority)
		p_opt->sm_priority = *new_priority;

	if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER) {
		/* SM priority reduced by user, need to get back to sm_priority
		   during heavy sweep. Probably new SM will be selected */
		if (p_opt->sm_priority < old_pri) {
			sm->p_subn->active_sm_priority = p_opt->sm_priority;
			return;
		}
		/* active SM priority already set in previous heavy sweep */
		if (p_opt->master_sm_priority &&
		    sm->p_subn->active_sm_priority == p_opt->master_sm_priority)
			return;

		p_sm_tbl = &sm->p_subn->sm_guid_tbl;
		p_rsm = (osm_remote_sm_t *) cl_qmap_head(p_sm_tbl);
		while (p_rsm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl)) {
			if (ib_sminfo_get_state(&p_rsm->smi) !=
			    IB_SMINFO_STATE_NOTACTIVE)
				break;
			p_rsm = (osm_remote_sm_t *) cl_qmap_next(&p_rsm->map_item);
		}
		/* coming_out_of_standby or active SM found */
		if (sm->p_subn->coming_out_of_standby ||
		    p_rsm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl))
			sm->p_subn->active_sm_priority =
			    (p_opt->master_sm_priority > p_opt->sm_priority) ?
			    p_opt->master_sm_priority : p_opt->sm_priority;
		else /* only one SM in the fabric */
			sm->p_subn->active_sm_priority =
			    p_opt->sm_priority;
	} else {
		uint8_t old_act_pri = sm->p_subn->active_sm_priority;
		sm->p_subn->active_sm_priority = p_opt->sm_priority;
		if (old_act_pri < sm->p_subn->active_sm_priority &&
		    sm->p_subn->sm_state == IB_SMINFO_STATE_STANDBY)
			osm_send_trap144(sm, TRAP_144_MASK_SM_PRIORITY_CHANGE);
	}
}

void osm_set_sm_priority(osm_sm_t * sm, uint8_t priority)
{
	_osm_set_active_sm_priority(sm, &priority);
}

void osm_set_active_sm_priority(osm_sm_t * sm)
{
	_osm_set_active_sm_priority(sm, NULL);
}


/* calculate the crc for a given buffer */
uint32_t osm_calc_file_crc32(IN const char * file_name)
{
	uint32_t crc = 0;
	char *p;
	int  count = 0;
	char line[256];
	FILE *f;

	if (!(f = fopen(file_name, "r")))
		return -1L;
	while (fgets(line, sizeof(line), f)) {
		p = line;
		while (isspace(*p))
			p++;
		if (*p == '\0' || *p == '\n' || *p == '#')
			continue;
		count = strlen(p);
		/* do the calculation */
		crc = cl_calc_crc32((void *)p, count, crc);
	}
	fclose(f);
	return crc;
}

/*
 * Check if crc chagned for specific file
 * Return 0 if not changed and 1 if changed
 */
static uint16_t check_crc_change(osm_sm_t * p_sm, osm_crc_id_enum file_id,
				 boolean_t update)
{
	uint32_t local_crc;
	uint16_t changed = 0;
	osm_crc_element_t *p_crc_element;
	char *file_name;

	if (file_id >= OSM_CRC_MAX) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2E12: "
			"Got request to check crc for illegal file id %u\n",
			file_id);
		goto EXIT;
	}

	p_crc_element = &p_sm->files_crc[file_id];
	file_name = *(char**)((void*)&p_sm->p_subn->opt + p_crc_element->opt_offset);
	/* Skip file if it does not exist */
	if (!file_name || access(file_name, F_OK) == -1)
		goto EXIT;

	if (p_crc_element) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"Checking crc for file id %u (%s)\n",
			p_crc_element->id, file_name);
		local_crc = osm_calc_file_crc32(file_name);
		if (local_crc == (uint32_t)-1) {
			OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE,
				"cannot calculate crc for %s\n",file_name);
			local_crc = 0;
		}

		if (local_crc == p_crc_element->crc_value) {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "%s crc  "
				"unchanged (%u)\n",
				file_name, local_crc);
			if (update)
				p_crc_element->crc_changed = FALSE;
		} else {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "%s crc "
				"changed (old val: %u, new val %u)\n",
				file_name, p_crc_element->crc_value,
				local_crc);
			if (update) {
				p_crc_element->crc_value = local_crc;
				p_crc_element->crc_changed = TRUE;
			}
			changed = 1;
		}
	}
EXIT:
	return changed;
}

uint16_t osm_calc_files_crc(IN osm_sm_t *p_sm, IN osm_crc_id_enum file_id,
			    IN boolean_t update)
{
	p_sm->crc_changed = 0;
	osm_crc_id_enum id;

	if (file_id < OSM_CRC_MAX) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Calculating crc for file "
			"id %u\n", file_id);
		p_sm->crc_changed += check_crc_change(p_sm, file_id, update);
		goto EXIT;
	}

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Calculating crc for all files\n");
	for (id = 0 ; id < OSM_CRC_MAX ; id++)
		p_sm->crc_changed += check_crc_change(p_sm, id, update);

EXIT:
	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "%u files crc changed during last "
		"check\n", p_sm->crc_changed);
	return p_sm->crc_changed;
}

boolean_t osm_is_crc_changed(IN osm_sm_t *p_sm, IN osm_crc_id_enum file_id)
{
	boolean_t res = FALSE;

	if (file_id < OSM_CRC_MAX)
		res =  p_sm->files_crc[file_id].crc_changed;
	else
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 2E11: "
			"Request to check crc change for illegal file id "
			"(%u)\n", file_id);

	return res;
}

boolean_t osm_is_crc_invalidates_cache(IN osm_sm_t *p_sm)
{
	osm_crc_id_enum id= 0;
	boolean_t invalidate = FALSE;

	while ((id < OSM_CRC_MAX) && !invalidate) {
		invalidate =  (p_sm->files_crc[id].crc_changed &&
			       p_sm->files_crc[id].invalidates_cache);
		id++;
	}

	return invalidate;
}
