/*
 * Copyright (c) 2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_sps_rcv_t.
 * This object represents the SwitchPortStateTable Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <iba/ib_types.h>
#include <opensm/osm_opensm.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SW_PORT_STATE_RCV_C

extern
void osm_update_discovery(osm_sm_t *sm, osm_node_t *p_node,
			  osm_port_t *p_port, uint8_t port_num,
			  boolean_t check_force);
extern
void pi_rcv_process_switch_external_port(IN osm_sm_t * sm,
					 IN osm_node_t * p_node,
					 IN osm_port_t * p_port,
					 IN uint8_t ext_port,
					 IN osm_physp_t * p_physp,
					 IN osm_madw_context_t *p_context);

extern
void pi_rcv_process_switch_ext_port(IN osm_sm_t * sm,
                                    IN osm_node_t * p_node,
                                    IN osm_port_t * p_port,
                                    IN osm_physp_t * p_physp,
                                    IN ib_port_info_t * p_pi);


static uint8_t osm_spst_get_block_number(IN const ib_mad_t *p_mad)
{
	return (uint8_t) (cl_ntoh32(p_mad->attr_mod) &
			   OSM_SPST_PORT_STATE_MASK_ATTR_MOD_HO);
}

static inline uint8_t
osm_spst_get_index_state(IN uint8_t index,
			 IN const ib_switch_port_state_table_t *p_spst)
{
	uint8_t shift = (index % 2) ? IB_SPST_STATE_ODD_STATE_SHIFT : IB_SPST_STATE_EVEN_STATE_SHIFT;
	uint8_t mask = (index % 2) ? IB_SPST_STATE_ODD_STATE_MASK : IB_SPST_STATE_EVEN_STATE_MASK;
	return (p_spst->port_states[index / 2] & mask) >> shift;
}

void osm_spst_rcv_process(IN void *context, IN void *data) {
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	osm_sps_context_t *p_context;
	ib_net64_t port_guid, node_guid;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_physp_t *p_physp, *p_ext_port_physp;
	uint32_t block;
	ib_switch_port_state_table_t *p_spst;
	uint8_t i, ext_port, port_state, old_state, num_ports;
	osm_madw_context_t pi_context;
	osm_dr_path_t *p_dr_path;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_sps_context_ptr(p_madw);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_SWITCH_PORT_STATE_TBL);

	port_guid = p_context->port_guid;
	node_guid = p_context->node_guid;
	block = osm_spst_get_block_number(p_madw->p_mad);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received SwPortStateTbl for port "
		"with GUID 0x%" PRIx64 ", for parent node GUID "
		"0x%" PRIx64", block number %u\n",
		cl_ntoh64(port_guid), cl_ntoh64(node_guid), block);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (PF(!p_port)) {
		CL_PLOCK_RELEASE(sm->p_lock);
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0C01: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid),
			cl_ntoh64(node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0C02: "
			"SwitchPortStateTable MAD with status 0x%x received, "
			"for port GUID 0x%" PRIx64 "\n",
			cl_ntoh16(ib_smp_get_status(p_smp)),
			cl_ntoh64(port_guid));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0C03: "
			"Received SwitchPortStateTable MAD for node "
			"(node GUID 0x%" PRIx64 ", port GUID 0x%" PRIx64 "),"
			" that is not switch, TID 0x%" PRIx64 "\n",
			cl_ntoh64(node_guid), cl_ntoh64(port_guid),
			cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}
	num_ports = osm_node_get_num_physp(p_node);

	p_spst = ib_smp_get_payload_ptr(p_smp);
	/* can receive this MAD only on port 0 of the switch */
	p_physp = osm_node_get_physp_ptr(p_node, 0);
	OSM_ASSERT(p_physp);

	memset(&pi_context, 0, sizeof(osm_madw_context_t));
	pi_context.pi_context.node_guid = node_guid;
	pi_context.pi_context.port_guid = port_guid;
	pi_context.pi_context.set_method = FALSE;
	pi_context.pi_context.light_sweep = FALSE;
	pi_context.pi_context.active_transition = FALSE;
	pi_context.pi_context.client_rereg = FALSE;

	ext_port = block ? block * IB_SPST_BLOCK_SIZE : 1;
	/*
	 * Go over external ports of switch for the current block.
	 * i - index in block.
	 * If it's block 0, entry 0 is for port 0, which is not external.
	 * num_ports is number of external ports + 1 port 0
	 */
	for (i = block ? 0 : 1;
	     i < IB_SPST_BLOCK_SIZE && ext_port < num_ports; i++, ext_port ++) {
		p_ext_port_physp = osm_node_get_physp_ptr(p_node, ext_port);
		port_state = osm_spst_get_index_state(i, p_spst);

		OSM_ASSERT(p_ext_port_physp);
		if (!p_ext_port_physp)
			break;

		old_state = ib_port_info_get_port_state(&p_ext_port_physp->port_info);

		if (old_state != port_state) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Switch 0x%" PRIx64 " %s port %u"
				" changed state from %s to %s \n",
				cl_ntoh64(port_guid), p_node->print_desc, ext_port,
				ib_get_port_state_str(old_state),
				ib_get_port_state_str(port_state));

			/* If port state has changed, send PortInfo */
			pi_rcv_process_switch_external_port(sm, p_node, p_port,
							    ext_port, p_physp,
							    &pi_context);
		} else {
			ib_port_info_t dummy_port_info;
			memcpy(&dummy_port_info, &p_ext_port_physp->port_info,
			       sizeof(dummy_port_info));
			p_dr_path = osm_physp_get_dr_path_ptr(p_ext_port_physp);
			osm_dr_path_init(p_dr_path, p_physp->dr_path.hop_count,
					 p_physp->dr_path.path);

			pi_rcv_process_switch_ext_port(sm, p_node, p_port, p_ext_port_physp,
						       &dummy_port_info);
		}
	}
Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
