%{
/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_TOPO_PARSER_Y_Y
#include <opensm/osm_port_group.h>
#include <opensm/osm_opensm.h>

#define YYSTYPE char*
extern FILE * __topo_parser_in;
static int in_topo;
static char *sw_grp_val = NULL;
static osm_port_group_t *p_sw_grp = NULL;
static char *hca_grp_val = NULL;
static osm_port_group_t *p_hca_grp = NULL;
static uint64_t id_val;
static struct osm_opensm *p_osm;
char *__topo_parser_yylval = NULL;

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void clear_topo()
{
	clear_val(&__topo_parser_yylval);
	clear_val(&sw_grp_val);
	clear_val(&hca_grp_val);
	p_sw_grp = NULL;
	p_hca_grp = NULL;
	id_val = 0;
}

void __topo_parser_error(const char *str)
{
        OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "%.200s\n", str);
}

static void print_topo()
{
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Topology: Id=|%" PRIu64 "|, "
		"Sw-Grp=|%.200s|, Hca-Grp=|%.200s|\n", id_val, sw_grp_val,
		hca_grp_val);
}

int __topo_parser_wrap()
{
        return 1;
}

int __topo_parser_parse();
int __topo_parser_lex();

int parse_topo_policy(struct osm_opensm *osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__topo_parser_in = NULL;
	if (!p_osm->subn.opt.topo_policy_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "No policy file to parse -"
			" Quit topology parser\n");
		res = 1;
		goto EXIT;
	}
	__topo_parser_in = fopen(p_osm->subn.opt.topo_policy_file, "r");
	if (!__topo_parser_in) {
		__topo_parser_error("ERR 9900: Topology policy file specified "
				    "but not found\n");
		res = 1;
		goto EXIT;
	}
        res =  __topo_parser_parse();
	if (!res && in_topo) {
		__topo_parser_error("ERR 9901: Topology policy file ended "
				    "within topology paragraph\n");
		res = 1;
		goto EXIT;
	}

	if (res) {
		__topo_parser_error("ERR 9902: Topology policy file parsing "
				    "failed\n");
		osm_topo_mgr_destroy(&p_osm->sm.topology_mgr);
	} else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Topology "
			"policy file parsed successfully\n");
EXIT:
	if (__topo_parser_in)
	{
	        fclose(__topo_parser_in);
		__topo_parser_in = NULL;
	}
	clear_topo();
	return res;
}

%}

%token COMMENT
%token TOPO_START
%token TOPO_END
%token SW_GRP
%token HCA_GRP
%token ID
%token WORD

%%

topo_policy: topologies;

topologies: /* empty */
        |  topologies topo
        ;

comment: COMMENT
	{
	}
	;

topo:
	start
	|word
	|sw_grp
	|hca_grp
	|id
	|end
	|comment
        ;

word: WORD
	{
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9903: Unexpected "
			"input |%.200s|, Aborting parsing!\n",
			__topo_parser_yylval);
		return 1;
	}
	;

id: ID WORD
	{
		if (!in_topo) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9910: "
				"Parsed id qualifier outside a topology "
				"(%.200s)\n", __topo_parser_yylval);
			return 1;
		}
		if (id_val) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9904: "
				"Id already defined for this topology "
				"- Aborting parsing!\n");
			return 1;
		}
		int id = atoi(__topo_parser_yylval);
		if (id <= 0) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9905: "
				"topology id must be positive "
				"- Aborting parsing!\n");
			return 1;
		}
		osm_topology_t *p_topo = osm_topo_mgr_get_topo(p_osm,
							       (uint64_t)id);
		if (p_topo) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9906: "
				"topology id already known by topology manager"
				" - Aborting parsing!\n");
			return 1;
		}
		id_val = (uint64_t)id;
	}
	;

sw_grp: SW_GRP WORD
	{
		if (!in_topo) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9911: "
				"Parsed sw-grp qualifier outside a topology "
				"(%.200s)\n", __topo_parser_yylval);
			return 1;
		}
		if (p_sw_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9907: Sw-Grp "
				"already defined for this topology, Aborting "
				" parsing!\n");
			return 1;
		}
		p_sw_grp = osm_get_port_grp(p_osm, __topo_parser_yylval);
		if (p_sw_grp)
			sw_grp_val = strdup(__topo_parser_yylval);
		else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9908: Sw-Grp"
				" %.200s not found, Aborting topology "
				"parsing!\n", __topo_parser_yylval);
			return 1;
		}
	}
	;

hca_grp: HCA_GRP WORD
	{
		if (!in_topo) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9912: "
				"Parsed hca-grp qualifier outside a topology "
				"(%.200s)\n", __topo_parser_yylval);
			return 1;
		}
		if (p_hca_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9909: Hca-Grp "
				"already defined for this topology, Aborting "
				" parsing!\n");
			return 1;
		}
		p_hca_grp = osm_get_port_grp(p_osm, __topo_parser_yylval);
		if (p_hca_grp)
			hca_grp_val = strdup(__topo_parser_yylval);
		else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 990A: Hca-Grp"
				" %.200s not found, Aborting topology "
				"parsing!\n", __topo_parser_yylval);
			return 1;
		}
	}
	;

start:
        TOPO_START
        {
		if (in_topo) {
			__topo_parser_error("ERR 990B: Parsed topology "
					    "within a topology\n");
			return 1;
		} else {
			clear_topo();
			in_topo = 1;
		}
        }
        ;

end:
        TOPO_END
        {
		if (in_topo) {
			in_topo = 0;
			if (!id_val) {
				__topo_parser_error("ERR 990C: No Id defined"
						    " for topology - Aborting "
						    "parsing!\n");
				print_topo();
				return 1;
			}
			if (!p_sw_grp) {
				__topo_parser_error("ERR 990D: No Sw-Grp defined"
						    " for topology - Aborting "
						    "parsing!\n");
				print_topo();
				return 1;
			}
			if (!p_hca_grp) {
				__topo_parser_error("ERR 990E: No Hca-Grp defined"
						    " for topology - Aborting "
						    "parsing!\n");
				print_topo();
				return 1;
			}

			if (osm_topo_mgr_add_topo(p_osm, id_val, p_sw_grp, p_hca_grp))
				return 1;
		} else {
			__topo_parser_error("ERR 990F: Parsed topology end while"
					    " not within a topology\n");
			return 1;
		}
        }
        ;

