/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Topologies manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_TOPO_MGR_C
#include <opensm/osm_switch.h>
#include <opensm/osm_log.h>
#include <opensm/osm_topology_mgr.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

extern int parse_topo_policy(struct osm_opensm *p_osm);

typedef struct bfs_item {
	cl_list_item_t list_item;
	osm_port_group_member_t *p_member;
} bfs_item_t;

boolean_t osm_topo_physp_on_main(struct osm_physp *p_physp)
{
	return (p_physp->p_remote_physp &&
		((p_physp->topo_id == NO_TOPO_ID) ||
			(p_physp->topo_id == MAIN_TOPO_ID)));
}

uint64_t osm_topologies_count(IN struct osm_opensm *p_osm)
{
	uint64_t i, cnt = 0;

	for (i = 0 ; i < TOPO_ARRAY_SIZE ; i++)
		if (p_osm->sm.topology_mgr.topologies[i])
			cnt++;

	return cnt;
}

/* Mark physp as not belong to a topology */
static void mark_physp_out_of_topo(osm_physp_t *p_physp)
{
	p_physp->p_remote_physp = p_physp;
}

/* If physp marked as not belong to a topology - remove the mark */
static void unmark_physp_out_of_topo(osm_physp_t *p_physp)
{
        if (p_physp->p_remote_physp == p_physp)
		p_physp->p_remote_physp = NULL;
}

static boolean_t is_physp_in_topo(osm_physp_t *p_physp)
{
        return (p_physp->p_remote_physp != p_physp);
}

void osm_topo_switch_delete(cl_map_item_t *pp_sw)
{
	osm_switch_delete((osm_switch_t **)&pp_sw);
}

void osm_topo_node_delete(cl_map_item_t *pp_sw)
{
	osm_node_t *p_node;

	p_node = (osm_node_t *)pp_sw;

	free(p_node->print_desc);

	p_node->print_desc = NULL;

	free(p_node);
}

/* Get a pointer to a node on the 'main' subn and a port number
 * Set a pointer to the remote node on the specific topology
 * and return the number of the remote port on this node */
static osm_physp_t *topo_remote_physp(struct osm_opensm *p_osm,
				      const osm_node_t * p_main_node,
				      const uint8_t port_num,
				      osm_topology_t * p_topo)
{
	osm_physp_t *p_main_physp_remote;
	osm_physp_t *p_main_physp;
	osm_physp_t *p_topo_remote_physp = NULL;

	osm_port_t *p_topo_port_remote;
	uint8_t remote_port_num;
	ib_net64_t remote_port_guid;

	p_main_physp = (osm_physp_t *) & (p_main_node->physp_table[port_num]);
	p_main_physp_remote = osm_physp_get_remote(p_main_physp);
	if (!p_main_physp_remote)
		goto Exit;

	remote_port_guid = osm_physp_get_port_guid(p_main_physp_remote);
	remote_port_num = p_main_physp_remote->port_num;

	/* Verify that remote port belongs to topology */
	p_topo_port_remote = osm_get_port_by_guid(p_topo->p_subn, remote_port_guid);

	if (!p_topo_port_remote) {
		ib_net64_t guid;
		guid = osm_node_get_node_guid(p_main_node);
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Topology %" PRIu64
			": Guid 0x%" PRIx64 ", Port %d - Remote "
			"physp (Guid 0x%" PRIx64 ", Port %d) "
			"not part of this topology - Clearing Physp\n",
			p_topo->id, cl_ntoh64(guid), port_num,
			cl_ntoh64(remote_port_guid), remote_port_num);
		goto Exit;
	}

	p_topo_remote_physp =
		&(p_topo_port_remote->p_node->physp_table[remote_port_num]);

	/* Verify that specific physp on remote node belongs to topology */
	if (!is_physp_in_topo(p_topo_remote_physp))
		p_topo_remote_physp = NULL;

Exit:
	return p_topo_remote_physp;
}

/*
 * Return Value:
 *
 * True - physp connected successfully to remote physp on topology
 * False - physp not connected to remote physp on topology
 */
static boolean_t handle_phy_port(struct osm_opensm *p_osm,
				 osm_topology_t * p_topo,
				 osm_node_t * p_node, uint8_t port_num,
				 osm_node_t * p_main_node)
{
	osm_physp_t *p_phy;
	osm_physp_t *p_remote_phy;
	boolean_t result = TRUE;

	p_phy = &p_node->physp_table[port_num];

	/*
	 * If the remote is not NULL this mean that this physp was alreaedy
	 * set as the remote of other physp so we don't need to handle it.
	 */
	if (p_phy->p_remote_physp != NULL)
		goto EXIT;

	/* Find the remote node and port number ON THIS TOPOLOGY */
	p_remote_phy = topo_remote_physp(p_osm, p_main_node, port_num,
					   p_topo);

	if (!p_remote_phy) {
		/* The remote phy port is not part of this topology
		 *  so we need to clear it
		 */
		osm_physp_construct(p_phy);
		result = FALSE;
		goto EXIT;
	}
	/* Make sure the remote phy_port is not already set */
	if (p_remote_phy->p_remote_physp != NULL) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6609: Trying to "
			"set remote port which is already set, remote port "
			"guid: 0x%" PRIx64 " , remote remote val: %p, remote"
			" addr: %p\n", cl_ntoh64(p_remote_phy->port_guid),
			p_remote_phy->p_remote_physp,
			&p_remote_phy->p_remote_physp);
		result = FALSE;
		goto EXIT;
	}

	/* Set remote phy_ports to each other */
	p_phy->p_remote_physp = p_remote_phy;
	p_remote_phy->p_remote_physp = p_phy;

	/* Mark on main subnet that these ports added to this topology */
	if (p_topo->id != NON_ITR_TOPO_ID) {
		p_main_node->physp_table[port_num].topo_id = p_topo->id;
		p_main_node->physp_table[port_num].p_remote_physp->topo_id
		    = p_topo->id;
	}
EXIT:
	return result;
}

static void remove_port_from_topo(osm_topology_t * p_topo, osm_port_t *p_port)
{
	struct osm_opensm *p_osm = p_topo->p_subn->p_osm;
	osm_subn_t *main_subn = &p_osm->subn;
	osm_port_t *p_main_port;
	cl_list_iterator_t iter;
	osm_topology_t *p_iter_topo;

	cl_qmap_remove(&p_topo->p_subn->port_guid_tbl, p_port->guid);

	p_main_port = osm_get_port_by_guid(main_subn, p_port->guid);
	if (p_main_port) {
		iter = cl_list_head(&p_main_port->topology_list);

		while (iter != cl_list_end(&p_main_port->topology_list)) {
			p_iter_topo = osm_port_get_topo_from_list_iter(&iter);
			if (!p_iter_topo || (p_topo->id == p_iter_topo->id)) {
				cl_list_remove_item(&p_main_port->topology_list,
						    iter);
				break;
			}
			iter = cl_list_next(iter);
		}
	} else {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6615: Can't find "
			"port on main subnet: 0x%" PRIx64 "\n",
			cl_ntoh64(p_port->guid));
			return;
	}

	free(p_port);
}

/*
 * 1) Delete logical port of hca physp which is not connected
 * 1) Delete logical port of sw which all its physps are not connected
 * 2) Delete every node which all its physps are not connected
 * 3) Delete switch for switch node which all its physps are not connected
 */
static void  topo_remove_disconnected_nodes(osm_topology_t * p_topo)
{
	osm_subn_t *p_subn = p_topo->p_subn;
	cl_map_item_t *p_item;
	cl_map_item_t *sw_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	ib_net64_t guid;

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		p_item = cl_qmap_next(p_item);
		p_node = p_port->p_node;
		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			if (!p_node->discovery_count)
				remove_port_from_topo(p_topo, p_port);
		} else {
			if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
				if (p_port->p_physp->p_remote_physp &&
				    is_physp_in_topo(p_port->p_physp))
					p_node->discovery_count = 1;
				else {
					remove_port_from_topo(p_topo, p_port);
				}
			}
		}
	}

	p_item = cl_qmap_head(&p_subn->node_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->node_guid_tbl)) {
		p_node = (osm_node_t*)p_item;
		p_item = cl_qmap_next(p_item);
		if (!p_node->discovery_count){
			guid = osm_node_get_node_guid(p_node);
			if (p_node->sw) {
				sw_item = cl_qmap_remove(&p_subn->sw_guid_tbl,
							 guid);
				osm_topo_switch_delete(sw_item);
			}
			cl_qmap_remove(&p_subn->node_guid_tbl, guid);
			osm_topo_node_delete((cl_map_item_t *)p_node);
		}
	}
}

/*
 * Iterate all phy ports for all nodes in the topology nodes table, for each phy
 * port check if the remote phy port also belongs to a node that is on this
 * topology, if so we set the pointers of their remote_phy_port to point on
 * each other and if not we reset this phy port because it is not part of this
 * topology.
 * NOTE: When we copied the nodes we set remote_phy on all phy ports to NULL
 *       so now if we run into a remote_phy which is not NULL we can assume that
 *       it was filled by us and contain a reference to a phy port from this
 *       topology, only if it is NULL we need to go to the 'main' subn to
 *       check what is the remote node and if it is part of this topology
 */
static boolean_t link_phy_ports(struct osm_opensm *p_osm,
				osm_topology_t * p_topo)
{
	osm_switch_t *p_topo_sw;
	osm_node_t *p_topo_node;
	osm_node_t *p_main_node;
	osm_physp_t *physp_on_main;
	osm_physp_t *topo_physp;
	osm_physp_t *main_physp_remote;
	cl_map_item_t *p_item;
	osm_subn_t *p_subn = p_topo->p_subn;
	uint8_t i;
	ib_net64_t guid;
	cl_map_item_t *p_dummy_item;
	int ports_in_topo, remote_sw_in_main, remote_sw_in_topo;
	boolean_t remote_sw;
	osm_node_t *p_rem_node;

	p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->sw_guid_tbl)) {
		p_topo_sw = (osm_switch_t *)p_item;
		p_topo_node = p_topo_sw->p_node;
		guid = osm_node_get_node_guid(p_topo_node);
		ports_in_topo = 0;
		remote_sw_in_main = 0;
		remote_sw_in_topo = 0;
		p_main_node = osm_get_node_by_guid(&p_osm->subn,
						   p_topo_node->node_info.
						   node_guid);
		if (!p_main_node) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR,
				"ERR 6608: guid not found on main subnet 0x%"
				PRIx64 "\n", cl_ntoh64(guid));
			return FALSE;
		}

		for (i = 1; i < p_topo_node->physp_tbl_size; i++) {
			physp_on_main = &p_main_node->physp_table[i];
			topo_physp = &p_topo_node->physp_table[i];
			main_physp_remote = physp_on_main->p_remote_physp;
			if (main_physp_remote) {
				p_rem_node = main_physp_remote->p_node;
				remote_sw =
					(osm_node_get_type(p_rem_node) ==
						IB_NODE_TYPE_SWITCH);
				if (remote_sw)
					remote_sw_in_main++;
				if (is_physp_in_topo(topo_physp)) {
					if (handle_phy_port(p_osm, p_topo,
							     p_topo_node, i,
							     p_main_node)) {
						ports_in_topo++;
						if (remote_sw)
							remote_sw_in_topo++;
					}
				}
			}
		}

		if (ports_in_topo) {
			/* Added at least one physp of the node to the topo */
			p_topo_node->discovery_count = 1;
			osm_physp_t *p_phy_sw0 = &p_main_node->physp_table[0];
			if ((p_topo->id != NON_ITR_TOPO_ID) &&
			    (p_phy_sw0->topo_id == NO_TOPO_ID))
				p_phy_sw0->topo_id = p_topo->id;
		}

		if (remote_sw_in_topo &&
			(remote_sw_in_topo < remote_sw_in_main)) {
			/* This is border Switch */
			p_dummy_item = calloc(1, sizeof(cl_map_item_t));
			if (p_dummy_item) {
				OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
					"Added border switch 0x%" PRIx64 "\n",
					cl_ntoh64(guid));
				cl_qmap_insert(&p_subn->border_sw_tbl, guid,
					       p_dummy_item);
			} else {
				OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR,
					"ERR 660F: Fail to allocate p_item for"
					" border switch 0x%" PRIx64 "\n",
					cl_ntoh64(guid));
				return FALSE;
			}
		}

		p_item = cl_qmap_next(p_item);
	}

	topo_remove_disconnected_nodes(p_topo);

	/*
	 * Now that we finished to link the physps of the topology we can
	 * unmark the physps that are marked as out of topology.
	 */
	p_item = cl_qmap_head(&p_subn->node_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->node_guid_tbl)) {
		p_topo_node = (osm_node_t *)p_item;
		for (i = 1; i < p_topo_node->physp_tbl_size; i++)
			unmark_physp_out_of_topo(&p_topo_node->physp_table[i]);
		p_item = cl_qmap_next(p_item);
	}

	return TRUE;
}

static osm_port_t *copy_port(struct osm_opensm *p_osm, osm_port_t * p_port)
{
	osm_port_t *p_port_copy;

	p_port_copy = malloc(sizeof(osm_port_t));
	if (!p_port_copy) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6604: Failed to "
			"allocate memory for osm_port 0x%" PRIx64,
			cl_ntoh64(p_port->guid));
		return NULL;
	}
	memcpy(p_port_copy, p_port, sizeof(osm_port_t));
	p_port_copy->priv = NULL;
	return p_port_copy;
}

static osm_node_t *copy_node(struct osm_opensm *p_osm,
			     osm_port_group_member_t * p_member)
{
	osm_node_t *p_node_copy;
	osm_node_t *p_orig_node;
	size_t total_node_size;
	int i;
	osm_port_t *p_port;
	osm_physp_t *p_physp;

	p_port = p_member->port;
	p_orig_node = p_port->p_node;
	total_node_size = sizeof(osm_node_t) +
	    sizeof(osm_physp_t) * p_orig_node->node_info.num_ports;
	p_node_copy = malloc(total_node_size);
	if (!p_node_copy) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6605: Failed to "
			"allocate memory for osm_node 0x%" PRIx64,
			cl_ntoh64(p_port->guid));
		return NULL;
	}
	/* Copy the node with all it's phy ports */
	memcpy(p_node_copy, p_orig_node, total_node_size);
	p_node_copy->discovery_count = 0;
	if (p_orig_node->print_desc)
		p_node_copy->print_desc = strdup(p_orig_node->print_desc);

	/* For each phy port we need to update the pointers to the node and
	 *  to the remote phy port so they will point on the local topology
	 *  structures and not on the 'main' ones.
	 *  For now we can update the node pointer, when all nodes will be
	 *  copied we'll be able to iterate them and set all remote phy
	 *  port pointers (we can do that only when we know for sure that
	 *  all nodes were copied)
	 */
	for (i = 1; i < p_node_copy->physp_tbl_size; i++) {
		p_physp = &p_node_copy->physp_table[i];
		p_physp->p_node = p_node_copy;
		/* Mark out physp that are not members of the group */
		if (p_member->physp_members[i])
			p_physp->p_remote_physp = NULL;
		else {
			if (p_orig_node->physp_table[i].p_remote_physp)
				mark_physp_out_of_topo(p_physp);
		}
	}
	return p_node_copy;
}

static osm_switch_t *create_sw(struct osm_opensm *p_osm, osm_port_t * p_port)
{
	osm_switch_t *p_sw_copy;
	uint8_t num_ports;

	p_sw_copy = malloc(sizeof(osm_switch_t));
	if (!p_sw_copy) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6606: Failed to "
			"allocate memory for osm_sw 0x%" PRIx64,
			cl_ntoh64(p_port->guid));
		return NULL;
	}

	memset(p_sw_copy, 0, sizeof(osm_switch_t));

	num_ports = p_port->p_node->sw->num_ports;
	p_sw_copy->p_prof = malloc(sizeof(osm_port_profile_t) * num_ports);
	if (!p_sw_copy->p_prof) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6607: Failed to "
			"allocate memory for p_prof, osm_sw 0x%" PRIx64,
			cl_ntoh64(p_port->guid));
		free(p_sw_copy);
		return NULL;
	}
	p_sw_copy->num_ports = num_ports;
	memset(p_sw_copy->p_prof, 0, sizeof(osm_port_profile_t) * num_ports);

	p_sw_copy->switch_info = p_port->p_node->sw->switch_info;
	osm_mcast_tbl_init(&p_sw_copy->mcast_tbl, num_ports,
			   cl_ntoh16(p_sw_copy->switch_info.mcast_cap));
	osm_ar_data_init(&p_sw_copy->ar_data);
	p_sw_copy->ar_data.status = p_port->p_node->sw->ar_data.status;
	p_sw_copy->ar_data.pending_mads = p_port->p_node->sw->ar_data.pending_mads;
	p_sw_copy->ar_data.ar_info = p_port->p_node->sw->ar_data.ar_info;
	p_sw_copy->ar_data.ar_info_updated = p_port->p_node->sw->ar_data.ar_info_updated;
	p_sw_copy->ar_data.plft_info = p_port->p_node->sw->ar_data.plft_info;
	p_sw_copy->ar_data.plft_info_updated = p_port->p_node->sw->ar_data.plft_info_updated;
	return p_sw_copy;
}

/*
 * Get topology and return pointer to the pointer that the topology manager
 * holds for the topology with this id.
 */
static osm_topology_t **get_topo_pp(struct osm_opensm *p_osm,
				    osm_topology_t *p_topo)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;

	OSM_ASSERT(p_topo->id < TOPO_ARRAY_SIZE);
	return &p_mgr->topologies[p_topo->id];
}

/* Add a port that belongs to this topology to this topology structure.
 * This method adds all port related data to the subn tables:
 *	1) Add osm_port to the port_guid_tbl
 *	2) Add osm_node to the node_guid_tbl
 *	3) If SW - Add osm_sw to the sw_guid_tbl
 *	4) If SW Port 0 - Add phy_port and link to port/node (no remote node)
 *	5) If there is remote node which already exist in this topology
 *	   then create phy_port for this node and for the remote one and link
 *	   them to their nodes phy_port table (and to each other).
 *	   NOTE: If remote node does not exist yet we do nothing, the second node
 *	         which is found is responsible to create both phy_ports
 */
static void add_port_to_tables(struct osm_opensm *p_osm,
			       osm_topology_t * p_topo,
			       osm_port_group_member_t * p_member)
{
	osm_port_t *p_port;
	osm_port_t *p_topo_port = NULL;
	osm_node_t *p_topo_node = NULL;
	osm_switch_t *p_topo_sw = NULL;
	osm_subn_t *p_subn = p_topo->p_subn;
	uint16_t min_lid, max_lid;
	boolean_t new_node = FALSE;
	ib_net64_t node_guid;

	p_port = p_member->port;
	p_topo_port = copy_port(p_osm, p_port);
	if (!p_topo_port)
		goto FREE_PORT;

	/* If this is CA it might be that we already created a copy for it's
	 *  node for a different port so we need to check if the node already
	 *  exist in this topology nodes table
	 */
	node_guid = osm_node_get_node_guid(p_port->p_node);

	if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA)
		p_topo_node = osm_get_node_by_guid(p_subn, node_guid);

	/* If not found we can create the node now */
	if (!p_topo_node) {
		p_topo_node = copy_node(p_osm, p_member);
		new_node= TRUE;
	}

	if (!p_topo_node)
		goto FREE_PORT;

	p_topo_port->p_node = p_topo_node;
	p_topo_port->p_physp = osm_node_get_physp_ptr(p_topo_node,
						      p_port->p_physp->
						      port_num);

	osm_port_get_lid_range_ho(p_topo_port, &min_lid, &max_lid);
	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "max lid: %" PRIu16
		", topology max: %" PRIu16 "\n", max_lid,
		p_subn->max_lid);
	if (max_lid > p_subn->max_lid) {
		p_subn->max_lid = max_lid;
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG,
			"Set topology max lid to be %" PRIu16 "\n", max_lid);
	}

	if (osm_node_get_type(p_topo_node) == IB_NODE_TYPE_SWITCH) {
		p_topo_sw = create_sw(p_osm, p_port);
		if (!p_topo_sw)
			goto FREE_NODE;
		p_topo_node->sw = p_topo_sw;
		p_topo_sw->p_node = p_topo_node;
	}

	/* Add topology to port topology list (port on main subnet) */
	if (p_topo->id != NON_ITR_TOPO_ID) {
		if (cl_list_insert_tail(&p_port->topology_list,
					get_topo_pp(p_osm, p_topo))
		    != CL_SUCCESS) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6610: "
				"Failed to add topo %" PRIu64 "to port 0x%"
				PRIx64 " topology list\n", p_topo->id,
				cl_ntoh64(p_port->guid));
			goto FREE_SW;
		}
	}

	if (osm_node_get_type(p_topo_node) == IB_NODE_TYPE_SWITCH) {
		cl_qmap_insert(&p_subn->sw_guid_tbl,
				node_guid,
				(cl_map_item_t *) p_topo_sw);
	}

	/* Now we sure that all elements created and we can keep them
	 * No need to check if insert result, we call this method during
	 * iteration over node table so there is no option
	 * to have duplicate insert
	*/

	cl_qmap_insert(&p_subn->port_guid_tbl, p_port->guid,
		      (cl_map_item_t *) p_topo_port);

	if (new_node) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Adding 0x%" PRIx64
			" to topology %" PRIu64 "\n",
			cl_ntoh64(node_guid),
			p_topo->id);
		cl_qmap_insert(&p_subn->node_guid_tbl,
			       node_guid,
			       (cl_map_item_t *) p_topo_node);
	}
	return;
FREE_SW:
	osm_topo_switch_delete((cl_map_item_t *)p_topo_sw);
FREE_NODE:
	osm_topo_node_delete((cl_map_item_t *)p_topo_node);
FREE_PORT:
	free(p_topo_port);
}

static boolean_t fill_topo_tables(struct osm_opensm *p_osm,
				  osm_topology_t *p_topo)
{
	osm_port_group_t *group;
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	osm_port_group_member_t *p_member;
	osm_node_t *p_node;

	p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while (p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		p_port = (osm_port_t *)p_item;
		p_node = p_port->p_node;
		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
			group = p_topo->sw_grp;
		else
			group = p_topo->hca_grp;

		p_member = osm_pgrp_get_member(group, p_port->guid);
		if (p_member)
			add_port_to_tables(p_osm, p_topo, p_member);
		p_item = cl_qmap_next(p_item);
	}
	/* Now that all nodes which belong to this topology were copied
	 *  to the nodes table of the topology we can set all remote phy
	 *  port pointers for all phy ports
	 */
	return link_phy_ports(p_osm, p_topo);
}

static void osm_topo_subn_construct(IN osm_subn_t * p_subn,
				    IN struct osm_opensm *p_osm)
{
	memset(p_subn, 0, sizeof(*p_subn));
	cl_qmap_init(&p_subn->sw_guid_tbl);
	cl_qmap_init(&p_subn->node_guid_tbl);
	cl_qmap_init(&p_subn->port_guid_tbl);
	cl_qmap_init(&p_subn->border_sw_tbl);
	cl_qmap_init(&p_subn->held_back_sw_tbl);
	p_subn->p_osm = p_osm;
	p_subn->ignore_existing_lfts = TRUE;
	p_subn->need_update = TRUE;
	p_subn->min_sw_data_vls = p_osm->subn.min_sw_data_vls;
	p_subn->min_data_vls = p_osm->subn.min_data_vls;
	p_subn->path_bit = OSM_SUBNET_DEFAULT_PATH_BITS;
	cl_fmap_init(&p_subn->mc_roots_tbl, ib_compare_gids);
}

static boolean_t create_topo_subn(struct osm_opensm *p_osm,
				  osm_topology_t *p_topo)
{
	osm_subn_t *subn;

	/* Create new subn structure */
	subn = malloc(sizeof(osm_subn_t));
	if (!subn) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6602: Failed to "
			"allocate new subn for topology %" PRIu64 "\n",
			p_topo->id);
		goto EXIT;
	}
	osm_topo_subn_construct(subn, p_osm);

	/* Apply specific configuration file is specified */
	osm_subn_set_default_opt(&subn->opt);
	if (osm_subn_parse_conf_file(p_osm->subn.opt.config_file, &subn->opt,
				     FALSE, FALSE)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6603: Failed to "
			"parse main subn opt file (%s) for topology %" PRIu64
			"\n", p_osm->subn.opt.config_file, p_topo->id);
		goto FREE_AND_EXIT;
	}

	if (cl_list_init(&subn->leaf_switches_list, 10) != CL_SUCCESS) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6615: Failed to "
			"allocate leaf switches list for topology %" PRIu64
			"\n", p_topo->id);
		goto FREE_AND_EXIT;
	}

	osm_subn_parse_mc_roots_file(subn);

	p_topo->p_subn = subn;
	return TRUE;

FREE_AND_EXIT:
	free(subn);
EXIT:
	return FALSE;
}

static boolean_t build_subn(IN struct osm_opensm *p_osm,
			    IN osm_topology_t *p_topo)
{
	return (create_topo_subn(p_osm, p_topo) &&
		fill_topo_tables(p_osm, p_topo));
}

osm_topology_t *osm_topo_mgr_get_topo(IN struct osm_opensm *p_osm, IN uint64_t id)
{
	if (!p_osm->sm.topology_mgr.initialized)
		return NULL;

	if (id < TOPO_ARRAY_SIZE)
		return p_osm->sm.topology_mgr.topologies[id];
	else
		return NULL;
}

static void empty_tbl(cl_qmap_t *p_map, void (*del_func)(cl_map_item_t*))
{
	cl_map_item_t *p_item;
	cl_map_item_t *p_next_item;

	p_next_item = cl_qmap_head(p_map);
	while (p_next_item != cl_qmap_end(p_map)) {
		p_item = p_next_item;
		p_next_item = cl_qmap_next(p_next_item);
		if (del_func)
			del_func(p_item);
		else
			free(p_item);
	}
	cl_qmap_remove_all(p_map);
}

static void clear_removed_hcas_map(osm_topology_t * p_topo)
{
	if (p_topo->removed_hcas) {
		empty_tbl(p_topo->removed_hcas, NULL);
		free(p_topo->removed_hcas);
		p_topo->removed_hcas = NULL;
	}
}

void clear_all_removed_hcas_maps(IN struct osm_opensm * p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;
	uint64_t i;

	for (i = 0 ; i < TOPO_ARRAY_SIZE ; i++)
		if (p_mgr->topologies[i])
			clear_removed_hcas_map(p_mgr->topologies[i]);

}

boolean_t osm_topo_is_border_sw(IN osm_topology_t * p_topo,
				IN ib_net64_t sw_guid) {
	cl_map_item_t *p_item;
	osm_subn_t *p_subn = p_topo->p_subn;

	p_item = cl_qmap_get(&p_subn->border_sw_tbl, sw_guid);

        return ((p_item != cl_qmap_end(&p_subn->border_sw_tbl)) ?
		TRUE :
		FALSE);

}

boolean_t osm_topo_is_engine_itr(osm_topology_t *p_topo)
{
	struct osm_routing_engine *r;
	if (!p_topo)
		return FALSE;
	r = p_topo->p_subn->routing_engine_used;
	return (!r || !r->ITR_init) ? FALSE : TRUE;
}

/* return port object from ITR topology the port belongs too */
osm_port_t * osm_get_port_from_itr_topo(IN osm_port_t *p_port,
					OUT osm_topology_t **p_topo)
{
	osm_topology_t *p_iter_topo = NULL;
	cl_list_iterator_t iter;
	osm_port_t *p_res_port = NULL;

	if (!cl_list_count(&p_port->topology_list))
		goto Exit;

	iter = cl_list_head(&p_port->topology_list);

	while (iter != cl_list_end(&p_port->topology_list)) {
		p_iter_topo = osm_port_get_topo_from_list_iter(&iter);
		if (osm_topo_is_engine_itr(p_iter_topo)) {
			p_res_port = osm_get_port_by_guid(
					p_iter_topo->p_subn,
					p_port->guid);
			break;
		}
		iter = cl_list_next(iter);
	}

Exit:
	if (p_topo)
		*p_topo = p_iter_topo;
	return p_res_port;
}

static void clear_topo_data_from_nodes(IN osm_opensm_t * p_osm)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	int i;
	osm_node_t *p_node;

	p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	while(p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl)) {
		p_port = (osm_port_t*) p_item;
		p_node = p_port->p_node;
		cl_list_remove_all(&p_port->topology_list);
		for (i = 0; i < p_node->physp_tbl_size; i++)
			p_node->physp_table[i].topo_id = NO_TOPO_ID;

		p_item = cl_qmap_next(p_item);
	}
}

static void destroy_topo(osm_topology_t * p_topo)
{
	osm_subn_t *p_subn = p_topo->p_subn;

	if (p_subn) {
		empty_tbl(&p_subn->sw_guid_tbl, osm_topo_switch_delete);
		empty_tbl(&p_subn->node_guid_tbl, osm_topo_node_delete);
		empty_tbl(&p_subn->port_guid_tbl, NULL);
		empty_tbl(&p_subn->border_sw_tbl, NULL);
		subn_opt_destroy(&p_subn->opt);
		free(p_subn->opt.file_opts);
		osm_destroy_routing_engines(p_subn);
		osm_destroy_default_routing_engine(p_subn);
		osm_destroy_advanced_routing_engine(p_subn);
		cl_list_destroy(&p_subn->leaf_switches_list);
		osm_subn_clear_mc_roots_tbl(p_subn);
		free(p_subn);
	}
	clear_removed_hcas_map(p_topo);
	free(p_topo);
}

static void destroy_topo_array(IN osm_topo_mgr_t *p_mgr,
			       osm_topo_arr_type_t arr_type,
			       boolean_t clear_main_subn)
{
	uint64_t i;

	if (arr_type == PRIMARY_ARRAY) {
		/* IF we destroy main topo array we should clear the main
		 * subnet unless we specifically asked not to
		 */
		if (clear_main_subn)
			clear_topo_data_from_nodes(p_mgr->p_osm);
		for (i = 0 ; i < TOPO_ARRAY_SIZE ; i++) {
			if (p_mgr->topologies[i]) {
				destroy_topo(p_mgr->topologies[i]);
				p_mgr->topologies[i] = NULL;
			}
		}
	} else {
		for (i = 0 ; i < TOPO_ARRAY_SIZE ; i++) {
			if (p_mgr->prev_topologies[i]) {
				destroy_topo(p_mgr->prev_topologies[i]);
				p_mgr->prev_topologies[i] = NULL;
			}
		}
	}
}

void osm_topo_mgr_destroy(IN osm_topo_mgr_t *p_mgr)
{
	if (!p_mgr->initialized)
		return;
	destroy_topo_array(p_mgr, PRIMARY_ARRAY, TRUE);
	destroy_topo_array(p_mgr, PREV_ARRAY, FALSE);
	p_mgr->initialized = FALSE;
}

static void remark_main_topo_on_main(IN osm_opensm_t * p_osm)
{
	cl_map_item_t *p_item, *p_port_item;
	osm_port_t *p_port, *p_main_port;
	osm_node_t *p_node, *p_rem_node, *p_main_node;
	osm_subn_t *p_subn, *p_main_subn;
	osm_topology_t **main_topo_pp;
	osm_topology_t *p_topo = osm_topo_mgr_get_topo(p_osm, MAIN_TOPO_ID);
	uint8_t i;
	osm_physp_t *p_main_phy, *p_rem_phy;

	if (!p_topo)
		return;

	p_subn = p_topo->p_subn;
	p_main_subn = &p_osm->subn;
	main_topo_pp = get_topo_pp(p_osm, p_topo);

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while(p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*) p_item;
		p_item = cl_qmap_next(p_item);
		p_port_item = cl_qmap_get(&p_main_subn->port_guid_tbl, p_port->guid);
		if (p_port_item == cl_qmap_end(&p_main_subn->port_guid_tbl))
			continue;
		p_main_port = (osm_port_t*)p_port_item;
		if (cl_list_insert_tail(&p_main_port->topology_list, main_topo_pp)
		    != CL_SUCCESS) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6619: "
				"Failed to remark main topo %" PRIu64 " in port"
				" 0x%" PRIx64 " topology list\n", p_topo->id,
				cl_ntoh64(p_port->guid));
			return;
		}

		p_node = p_port->p_node;
		if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH)
			continue;

		if (p_port->p_physp->topo_id == NO_TOPO_ID)
			p_port->p_physp->topo_id = MAIN_TOPO_ID;
		for (i = 1 ; i < p_node->physp_tbl_size ; i++) {
			p_rem_node = osm_node_get_remote_node(p_node, i, NULL);
			if (p_rem_node) {
				p_main_node = p_main_port->p_node;
				p_main_phy = osm_node_get_physp_ptr(p_main_node,
								    i);
				p_rem_phy = p_main_phy->p_remote_physp;
				if (p_rem_phy) {
					p_main_phy->topo_id = MAIN_TOPO_ID;
					if (osm_node_get_type(p_rem_node) !=
					    IB_NODE_TYPE_SWITCH)
						p_rem_phy->topo_id = MAIN_TOPO_ID;
				}
			}
		}
	}
}

static boolean_t topo_mgr_validate(IN osm_opensm_t * p_osm)
{
	boolean_t res = TRUE;
	osm_subn_t *p_subn = &p_osm->subn;
	cl_map_item_t *p_item;
	osm_port_t *p_port;

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		if (cl_list_count(&p_port->topology_list) >
		    p_subn->opt.max_topologies_per_sw) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6616: "
				"port 0x%" PRIx64 " is used in more than max"
				" allowed topologies (%u)\n",
				cl_ntoh64(p_port->guid),
				p_subn->opt.max_topologies_per_sw);
			res = FALSE;
			break;
		}
		p_item = cl_qmap_next(p_item);
	}
	return res;
}

static void move_topologies_to_prev(osm_topo_mgr_t *p_mgr)
{
	destroy_topo_array(p_mgr, PREV_ARRAY, FALSE);
	memcpy(&p_mgr->prev_topologies, &p_mgr->topologies, TOPO_ARRAY_LEN);
	clear_topo_data_from_nodes(p_mgr->p_osm);
	memset(&p_mgr->topologies, 0, TOPO_ARRAY_LEN);
}

void osm_topo_mgr_revert_prev(IN struct osm_opensm * p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;

	if (!p_mgr->initialized)
		return;

	destroy_topo_array(p_mgr, PRIMARY_ARRAY, FALSE);
	memcpy(&p_mgr->topologies, &p_mgr->prev_topologies, TOPO_ARRAY_LEN);
	memset(&p_mgr->prev_topologies, 0, TOPO_ARRAY_LEN);

	remark_main_topo_on_main(p_osm);
}

/*
 * Add a physp of hca into a topology even though it is not defined in the
 * topology.
 * We do not expect that the remote physp on the SW will be part of the
 * topology BUT the remote SW must be part of this topology.
 */
static void force_hca_into_topo(osm_opensm_t * p_osm, osm_topology_t *p_topo,
				osm_port_t *p_port, uint8_t port_num,
				osm_node_t *p_remote_sw) {
	osm_port_group_member_t * p_member = NULL;
	osm_subn_t *p_subn = p_topo->p_subn;
	osm_switch_t *p_topo_sw;
	osm_port_t *p_topo_port;
	ib_net64_t sw_guid;
	cl_map_item_t *p_item;
	osm_physp_t *p_sw_main_physp, *p_sw_topo_physp, *p_hca_topo_physp;
	uint8_t sw_port_num;

	/* Verify that hca is not already in topology */
	p_item = cl_qmap_get(&p_subn->port_guid_tbl, p_port->guid);
	if (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6617: Fail to "
			"force hca port 0x%" PRIx64 " into topology %" PRId64
			", hca already in topology\n",
			cl_ntoh64(p_port->guid), p_topo->id);
		return;
	}

	/* Verify that remote sw is part of topology */
	sw_guid = osm_node_get_node_guid(p_remote_sw);
	p_item = cl_qmap_get(&p_subn->sw_guid_tbl, sw_guid);
	if (p_item == cl_qmap_end(&p_subn->sw_guid_tbl)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6618: Fail to "
			"force hca port 0x%" PRIx64 " into topology %" PRId64
			", Remote sw 0x%" PRIx64 " is not part of the topology\n",
			cl_ntoh64(p_port->guid), p_topo->id, cl_ntoh64(sw_guid));
		return;
	}
	p_topo_sw = (osm_switch_t*)p_item;

	/* Add hca port to topology subnet tables */
	p_member = osm_create_port_grp_member(p_osm, p_port, FALSE);
	if (!p_member)
		return;

	if (!osm_set_physp_reliable(p_osm, p_member, port_num, TRUE)) {
		osm_destroy_port_grp_member(p_member);
		return;
	}
	add_port_to_tables(p_osm, p_topo, p_member);
	osm_destroy_port_grp_member(p_member);

	/* Link hca and sw */
	p_item = cl_qmap_get(&p_subn->port_guid_tbl, p_port->guid);
	OSM_ASSERT(p_item != cl_qmap_end(&p_subn->port_guid_tbl));
	p_topo_port = (osm_port_t*)p_item;
	p_hca_topo_physp = p_topo_port->p_physp;

	p_sw_main_physp = osm_physp_get_remote(p_port->p_physp);
	sw_port_num = p_sw_main_physp->port_num;
	p_sw_topo_physp = &p_topo_sw->p_node->physp_table[sw_port_num];
	memcpy(p_sw_topo_physp, p_sw_main_physp, sizeof(osm_physp_t));
	p_sw_topo_physp->p_node = p_topo_sw->p_node;

	p_hca_topo_physp->p_remote_physp = p_sw_topo_physp;
	p_sw_topo_physp->p_remote_physp = p_hca_topo_physp;

	p_sw_main_physp->topo_id = p_topo->id;
	p_port->p_physp->topo_id = p_topo->id;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Forced Hca port 0x%" PRIx64
			" into topology %" PRId64 ", Connected to Sw 0x%" PRIx64
			" at port %u\n", cl_ntoh64(p_port->guid), p_topo->id,
			cl_ntoh64(sw_guid), sw_port_num);
}

/*
 * Locate hca's which are not joined to any topology and join them to the
 * first topology that their sw is part of (If the sw is not part of any
 * topology they will both be part of completer topology anyway).
 */
static void auto_join_lost_hcas(IN osm_opensm_t * p_osm)
{
	osm_subn_t *p_subn = &p_osm->subn;
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	uint8_t port_num;
	osm_node_t *p_remote_node;
	uint64_t sw_topo_id;
	osm_topology_t *p_topo;

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		p_item = cl_qmap_next(p_item);
		if ((osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA) &&
		    !cl_list_count(&p_port->topology_list)){
			port_num = p_port->p_physp->port_num;
			p_remote_node = osm_node_get_remote_node(p_port->p_node,
								 port_num, NULL);
			if (!p_remote_node)
				continue;

			sw_topo_id = osm_node_get_any_topo_id(p_remote_node);

			if (sw_topo_id == NO_TOPO_ID)
				continue;


			p_topo = p_osm->sm.topology_mgr.topologies[sw_topo_id];
			OSM_ASSERT(p_topo);
			force_hca_into_topo(p_osm, p_topo, p_port, port_num,
					    p_remote_node);
		}
	}
}

/*
 * If the hub switch of topology 0 was removed from the fabric, invalidates
 * ucast cache data.
 */
static void topo_mgr_check_topo0_hub_switch(osm_opensm_t * p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;
	osm_topology_t *p_prev_topology;
	osm_subn_t *p_main_subn = &p_osm->subn;
	osm_switch_t *p_prev_sw;
	cl_map_item_t *p_item;

	p_prev_topology = p_mgr->prev_topologies[MAIN_TOPO_ID];
	if (!p_prev_topology || !p_prev_topology->p_subn)
		return;

	p_prev_sw = osm_subn_get_hub_switch(p_prev_topology->p_subn);
	if (!p_prev_sw)
		return;

	p_item = cl_qmap_get(&p_main_subn->sw_guid_tbl,
			     osm_node_get_node_guid(p_prev_sw->p_node));
	if (p_item == cl_qmap_end(&p_main_subn->sw_guid_tbl)) {
		osm_subn_clear_hub_switch(p_prev_topology->p_subn);
		if (p_osm->subn.opt.use_ucast_cache) {
			OSM_LOG(p_osm->sm.p_log,
				OSM_LOG_INFO,
				"Cache invalidate required for topology %d "
				"due to hub switch disconnection\n",
				MAIN_TOPO_ID);
			p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
		}
	}

	return;
}

static inline boolean_t same_count(cl_qmap_t *p_map1, cl_qmap_t *p_map2)
{
	return (cl_qmap_count(p_map1) == cl_qmap_count(p_map2));
}

/*
 * Set hca as removed in a topology, if removed map not exist in the topology
 * create it, if hca already in the map do nothing (return TRUE)
 * Return True/False if succeed (False is only for allocation failure)
 */
static boolean_t set_removed_hca(IN struct osm_opensm *p_osm,
				 IN osm_topology_t *p_topo, IN ib_net64_t guid)
{
	cl_map_item_t *p_item;

	if (!p_topo->removed_hcas){
		p_topo->removed_hcas = malloc(sizeof(cl_qmap_t));
		if (!p_topo->removed_hcas){
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 660D: "
				"Fail to allocate removed hcas map for "
				" topology %" PRIu64 "\n", p_topo->id);
			return FALSE;
		}
		cl_qmap_init(p_topo->removed_hcas);
	}
	p_item = cl_qmap_get(p_topo->removed_hcas, guid);
	if (p_item == cl_qmap_end(p_topo->removed_hcas)) {
		p_item = malloc(sizeof(cl_map_item_t));
		if (!p_item){
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 660E: "
				"Fail to allocate removed hcas item for guid "
				"0x%" PRIx64 " in map of topology %" PRIu64 "\n",
				cl_ntoh64(guid), p_topo->id);
			return FALSE;
		}
		cl_qmap_insert(p_topo->removed_hcas, guid, p_item);
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Added guid 0x%"
			PRIx64 " to removed hcas map on topology %" PRIu64
			"\n", cl_ntoh64(guid), p_topo->id);
	}
	return TRUE;
}

/*
 * Unset hca from removed map, if removed map not exist in the topology
 * return false, otherwise return TRUE (even if hca to remove not found)
 */
static void unset_removed_hca(IN struct osm_opensm *p_osm,
				   IN osm_topology_t *p_topo, IN ib_net64_t guid)
{
	cl_map_item_t *p_item;

	OSM_ASSERT(p_topo->removed_hcas);

	p_item = cl_qmap_remove(p_topo->removed_hcas, guid);
	if (p_item != cl_qmap_end(p_topo->removed_hcas)) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Removed guid 0x%"
			PRIx64 " from removed hcas map on topology %" PRIu64
			"\n", cl_ntoh64(guid), p_topo->id);
		free(p_item);
	}
}

/* Check if a certain guid is set as removed hca in a topology */
static boolean_t is_removed_hca(IN osm_topology_t *p_topo, IN ib_net64_t guid)
{
	cl_map_item_t *p_item;

	if (p_topo->removed_hcas) {
		p_item = cl_qmap_get(p_topo->removed_hcas, guid);
		return (p_item != cl_qmap_end(p_topo->removed_hcas));
	} else
		return FALSE;
}

/* Handle added hca and return decision whether to invalidate cache */
static boolean_t handle_added_hca(IN struct osm_opensm *p_osm,
				  IN osm_topology_t *p_topo,
				  IN ib_net64_t guid)
{
	boolean_t invalidate = FALSE;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Handling added HCA port: 0x%"
		PRIx64 "\n", cl_ntoh64(guid));
	if (is_removed_hca(p_topo, guid))
		unset_removed_hca(p_osm, p_topo, guid);
	else {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "0x%" PRIx64
			" not found in removed map, new member"
			" - Invalidating cache\n", cl_ntoh64(guid));
		invalidate = TRUE;
	}

	return invalidate;
}

/* Handle removed hca and return decision whether to invalidate cache */
static boolean_t handle_removed_hca(IN struct osm_opensm *p_osm,
				    IN osm_topology_t *p_topo,
				    IN ib_net64_t guid)
{
	osm_subn_t *main_subn = &p_osm->subn;
	boolean_t invalidate = FALSE;
	osm_port_t *p_port;

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Handling removed HCA port: 0x%"
		PRIx64 "\n", cl_ntoh64(guid));
	p_port = osm_get_port_by_guid(main_subn, guid);
	if (p_port) {
		if (p_port->is_new)
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "0x%" PRIx64
				" is new - Ignore\n", cl_ntoh64(guid));
		else {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "0x%" PRIx64
				" is not new - Invalidate\n", cl_ntoh64(guid));
			invalidate = TRUE;
		}
	} else
		if (!set_removed_hca(p_osm, p_topo, guid))
			invalidate = TRUE;

	return invalidate;
}

static boolean_t compare_sw(IN struct osm_opensm *p_osm,
			    osm_subn_t *p_old_subn, osm_topology_t *p_topo,
			    osm_switch_t *p_old, osm_switch_t *p_new)
{
	boolean_t invalidate = FALSE;
	int i;
	osm_physp_t *p_physp_r_old;
	osm_physp_t *p_physp_r_new;
	ib_net64_t old_r_guid;
	ib_net64_t new_r_guid;
	uint8_t old_port;
	uint8_t new_port;
	osm_node_t *p_old_node = p_old->p_node;
	osm_node_t *p_new_node = p_new->p_node;

	for(i=1; (i < p_old_node->physp_tbl_size) && !invalidate; i++) {
		p_physp_r_old = p_old_node->physp_table[i].p_remote_physp;
		p_physp_r_new = p_new_node->physp_table[i].p_remote_physp;

		if (!p_physp_r_old && !p_physp_r_new)
			continue;

		if (p_physp_r_old && p_physp_r_new) {
			old_r_guid = p_physp_r_old->port_guid;
			new_r_guid = p_physp_r_new->port_guid;
			old_port = p_physp_r_old->port_num;
			new_port = p_physp_r_new->port_num;
			if ((old_r_guid != new_r_guid) || (old_port != new_port))
				invalidate = TRUE;
			continue;
		}

		if (p_physp_r_old) {
			/* No new -> port removed */
			if (p_physp_r_old->p_node->sw)
				invalidate = TRUE;
			else {
				old_r_guid = p_physp_r_old->port_guid;
				invalidate = handle_removed_hca(p_osm, p_topo,
								old_r_guid);
			}
		} else {
			/* No old -> port added */
			if (p_physp_r_new->p_node->sw)
				invalidate = TRUE;
			else {
				new_r_guid = p_physp_r_new->port_guid;
				invalidate = handle_added_hca(p_osm, p_topo,
							      new_r_guid);
			}
		}
	}
	return invalidate;
}

/* compare and copy data between subnets and invalidate cache if needed */
static void handle_subnet_changes(IN struct osm_opensm *p_osm,
				  IN struct osm_subn *p_old_subn,
				  IN osm_topology_t *p_topo)
{
	boolean_t invalidate = FALSE;
	struct osm_subn *p_new_subn = p_topo->p_subn;
	cl_map_item_t *p_old_item;
	cl_map_item_t *p_new_item;
	osm_switch_t *p_old_switch;
	osm_switch_t *p_new_switch;
	ib_net64_t old_switch_guid;
	ib_net64_t new_switch_guid;
	ib_net64_t hub_switch_guid;

	if (!same_count(&p_old_subn->sw_guid_tbl, &p_new_subn->sw_guid_tbl))
		invalidate = TRUE;

	p_old_item = cl_qmap_head(&p_old_subn->sw_guid_tbl);
	p_new_item = cl_qmap_head(&p_new_subn->sw_guid_tbl);
	while (p_old_item != cl_qmap_end(&p_old_subn->sw_guid_tbl) &&
	       p_new_item != cl_qmap_end(&p_new_subn->sw_guid_tbl)) {
		p_old_switch = (osm_switch_t*)p_old_item;
		p_new_switch = (osm_switch_t*)p_new_item;
		old_switch_guid = osm_node_get_node_guid(p_old_switch->p_node);
		new_switch_guid = osm_node_get_node_guid(p_new_switch->p_node);

		/*
		 * If p_old_switch and p_new_switch are pointers to different
		 * switches, increase the pointer to the switch with the
		 * lower guid.
		 */
		if (old_switch_guid != new_switch_guid) {
			if (old_switch_guid < new_switch_guid)
				p_old_item = cl_qmap_next(p_old_item);
			else
				p_new_item = cl_qmap_next(p_new_item);
			invalidate = TRUE;
			continue;
		}

		invalidate |= compare_sw(p_osm, p_old_subn, p_topo,
					 p_old_switch,
					 p_new_switch);
		p_new_switch->mlid_root_num = p_old_switch->mlid_root_num;
		p_old_item = cl_qmap_next(p_old_item);
		p_new_item = cl_qmap_next(p_new_item);
	}

	if (invalidate && p_osm->subn.opt.use_ucast_cache) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_INFO,
			"Cache invalidate required for "
			"topology %" PRIu64 "\n", p_topo->id);
		p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	}

	/* Copy hub switch guid to the new subnet */
	hub_switch_guid = osm_subn_get_hub_switch_guid(p_old_subn);
	if (hub_switch_guid &&
	    (cl_qmap_get(&p_new_subn->sw_guid_tbl, hub_switch_guid) !=
	     cl_qmap_end(&p_new_subn->sw_guid_tbl)))
		p_new_subn->hub_switch_guid = hub_switch_guid;
}

static bfs_item_t *bfs_list_add_member(struct osm_opensm *p_osm,
				       cl_qlist_t *bfs_list,
				       osm_port_group_member_t *p_member)
{
	bfs_item_t *p_bfsi = calloc(1, sizeof(bfs_item_t));

	if (!p_bfsi) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9916: Failed to "
			"allocate bfs item\n");
		return NULL;
	}
	p_bfsi->p_member = p_member;
	cl_qlist_insert_tail(bfs_list, (cl_list_item_t*)p_bfsi);
	return p_bfsi;
}

/*
	Gets a bfs item (p_member) and a port number.
	Check if the node connected on this port shall be added to the bfs list.

	The conditions to add the connected node are:
	1) This port is in the switches port group.
	2) There is a remote node (SWITCH) on this port.
	3) The remote node (and remote port) are also in the switches port group.
	4) The remote node was not visited yet (not in the visited map).

	If all above conditions stands - add the remote node to the bfs list.

	Return Value:
	TRUE - Added to list or no need to add.
	FALSE - Need to add but insert failed.
*/
static boolean_t bfs_handle_port(struct osm_opensm *p_osm,
				 osm_port_group_t *p_sw_grp,
				 cl_qmap_t *visited_map, cl_qlist_t *bfs_list,
				 bfs_item_t *p_bfsi, uint8_t port_num) {
	uint8_t remote_port;
	osm_node_t *p_remote_node;
	ib_net64_t remote_port_guid;
	ib_net64_t remote_node_guid;
	osm_node_t *p_node;
	osm_port_group_member_t *p_remote_member;
	cl_map_item_t *p_item;
	cl_qmap_t *p_sw_map = &p_sw_grp->ports;

	p_node = p_bfsi->p_member->port->p_node;
	p_remote_node = osm_node_get_remote_node(p_node, port_num, &remote_port);

	/* 1) This port is in the switches port group. */
	if (!p_bfsi->p_member->physp_members[port_num])
		return TRUE;

	/* 2) There is a remote node (SWITCH) on this port. */
	if (!p_remote_node ||
	    (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH))
		return TRUE;

	/* 3) The remote node (and remote port) are also in the sw port grp. */
	remote_port_guid = p_remote_node->node_info.port_guid;
	p_item = cl_qmap_get(p_sw_map, remote_port_guid);
	/* remote node in the grp. */
	if (p_item == cl_qmap_end(p_sw_map))
		return TRUE;
	p_remote_member = (osm_port_group_member_t*)p_item;
	/* remote port in the grp. */
	if (!p_remote_member->physp_members[remote_port])
		return TRUE;

	/* 4) The remote node was not visited yet (not in the visited map). */
	remote_node_guid = p_remote_node->node_info.node_guid;
	p_item = cl_qmap_get(visited_map, remote_node_guid);
	if (p_item != cl_qmap_end(visited_map))
		return TRUE;

	if (bfs_list_add_member(p_osm, bfs_list, p_remote_member))
		return TRUE;

	return FALSE;
}

/* Check if the HCA is connected to at least one switch from the topology. */
static boolean_t check_hca(struct osm_opensm *p_osm, cl_qmap_t *p_sw_map,
			   osm_port_group_member_t *p_member)
{
	int i = 1;
	boolean_t connected = FALSE;
	osm_node_t *p_node;
	uint8_t remote_port;
	osm_node_t *p_remote_node;
	cl_map_item_t *p_item;
	ib_net64_t remote_guid;
	osm_port_group_member_t *p_remote_member;

	p_node = p_member->port->p_node;
	while ((i < p_node->physp_tbl_size) && !connected) {
		if (p_member->physp_members[i]) {
			p_remote_node = osm_node_get_remote_node(p_node, i,
								 &remote_port);
			if (p_remote_node) {
				remote_guid = p_remote_node->node_info.port_guid;
				p_item = cl_qmap_get(p_sw_map, remote_guid);
				if (p_item == cl_qmap_end(p_sw_map)) {
					i++;
					continue;
				}
				p_remote_member = (osm_port_group_member_t*)p_item;
				if (p_remote_member->physp_members[remote_port])
					connected = TRUE;
			}
		}
		i++;
	}
	if (!connected)
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Topology "
			"broken!, HCA port %" PRIx64 " is not connected to any"
			" switch\n", cl_ntoh64(p_member->port->guid));
	return connected;
}


static boolean_t check_hcas(struct osm_opensm *p_osm, cl_qmap_t *p_sw_map,
			    osm_port_group_t *p_hca_grp)
{
	cl_map_item_t *p_item;
	osm_node_t *p_node;
	osm_port_group_member_t *p_member;

	p_item = cl_qmap_head(&p_hca_grp->ports);
	while (p_item != cl_qmap_end(&p_hca_grp->ports)) {
		p_member = (osm_port_group_member_t*)p_item;
		p_node = p_member->port->p_node;
		if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA) {
			if (!check_hca(p_osm, p_sw_map, p_member))
				return FALSE;
		} else
			OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Port %" PRIx64
				" found in hca_grp but is not a HCA.\n",
				cl_ntoh64(p_member->port->guid));
		p_item = cl_qmap_next(p_item);
	}
	return TRUE;
}

static void free_resources(cl_qmap_t *visited_map, cl_qlist_t *bfs_list)
{
	cl_map_item_t *p_map_item;
	cl_list_item_t *p_list_item;

	p_map_item = cl_qmap_head(visited_map);
	while (p_map_item != cl_qmap_end(visited_map)) {
		cl_qmap_remove_item(visited_map, p_map_item);
		free(p_map_item);
		p_map_item = cl_qmap_head(visited_map);
	}

	while (cl_qlist_count(bfs_list)) {
		p_list_item = cl_qlist_remove_head(bfs_list);
		free(p_list_item);
	}
}

static boolean_t verify_topo_integrity(struct osm_opensm *p_osm,
				       osm_port_group_t *p_sw_grp,
				       osm_port_group_t *p_hca_grp)
{
	cl_qmap_t visited_map;
	cl_map_item_t *p_item;
	cl_qlist_t bfs_list;
	bfs_item_t *p_bfsi;
	osm_node_t *p_node;
	ib_net64_t guid;
	osm_port_group_member_t *p_member;
	int i;
	uint32_t sw_count;
	boolean_t res;

	cl_qmap_init(&visited_map);
	cl_qlist_init(&bfs_list);
	sw_count = 0;

	p_item = cl_qmap_head(&p_sw_grp->ports);
	while (p_item != cl_qmap_end(&p_sw_grp->ports)) {
		p_member = (osm_port_group_member_t*)p_item;
		p_node = p_member->port->p_node;
		if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
			/* add the first sw found to bfs list. */
			if (!sw_count &&
			    !bfs_list_add_member(p_osm, &bfs_list, p_member))
				return FALSE;
			sw_count++;
		} else
			OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Port %" PRIx64
				" found in sw_grp but is not a sw.\n",
				cl_ntoh64(p_member->port->guid));
		p_item = cl_qmap_next(p_item);
	}
	if (!sw_count) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "No switch found"
			" in sw group\n");
		return FALSE;
	}

	while (cl_qlist_count(&bfs_list)) {
		p_bfsi = (bfs_item_t*)cl_qlist_remove_head(&bfs_list);
		p_node = p_bfsi->p_member->port->p_node;
		guid = osm_node_get_node_guid(p_node);

		/* If we already visited this node we can move on */
		p_item = cl_qmap_get(&visited_map, guid);
		if (p_item != cl_qmap_end(&visited_map))
			goto END_LOOP;
		p_item = malloc(sizeof(cl_map_item_t));
		if (!p_item) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9917: "
				"Failed to allocate bfs item for integrity "
				"check\n");
			free_resources(&visited_map, &bfs_list);
			return FALSE;
		}
		cl_qmap_insert(&visited_map, guid, p_item);
		for (i = 1 ; i < p_node->physp_tbl_size ; i++) {
			if (!bfs_handle_port(p_osm, p_sw_grp, &visited_map,
					     &bfs_list, p_bfsi, i)) {
				free_resources(&visited_map, &bfs_list);
				return FALSE;
			}
		}
END_LOOP:
		free(p_bfsi);
	}
	res = (sw_count == cl_qmap_count(&visited_map));
	if (res)
		res = check_hcas(p_osm, &p_sw_grp->ports, p_hca_grp);

	free_resources(&visited_map, &bfs_list);
	return res;
}

ib_api_status_t osm_topo_mgr_add_topo(IN struct osm_opensm *p_osm, IN uint64_t id,
				      IN osm_port_group_t *sw_grp,
				      IN osm_port_group_t *hca_grp)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;
	ib_api_status_t res = IB_SUCCESS;
	osm_topology_t *p_topo = NULL;

	if (p_mgr->topologies[id]) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6601: "
			"Fail to add topology - id exist\n");
		res = IB_ERROR;
		goto EXIT;
	}

	p_topo = malloc(sizeof(osm_topology_t));
	if (!p_topo) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 660C: "
			"Fail to allocate new topology\n");
		res = IB_ERROR;
		goto EXIT;
	}

	p_topo->id = id;
	p_topo->sw_grp = sw_grp;
	p_topo->hca_grp = hca_grp;
	p_topo->routing_engine = NULL;
	p_topo->p_subn = NULL;
	p_topo->removed_hcas = NULL;

	if (!build_subn(p_osm, p_topo)) {
		destroy_topo(p_topo);
		res = IB_ERROR;
		goto EXIT;
	}

	p_mgr->topologies[id] = p_topo;
	if (verify_topo_integrity(p_osm, sw_grp, hca_grp))
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "Topology %" PRIu64
			" (Sw/Nodes: %" PRIu32 "/%" PRIu32 ") - Integrity "
			"verified\n", id,
			cl_qmap_count(&p_topo->p_subn->sw_guid_tbl),
			cl_qmap_count(&p_topo->p_subn->node_guid_tbl));
	else
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 661A: "
			"Topology %" PRIu64 " (Sw/Nodes: %" PRIu32 "/%" PRIu32
			") - Integrity verification failed\n", id,
			cl_qmap_count(&p_topo->p_subn->sw_guid_tbl),
			cl_qmap_count(&p_topo->p_subn->node_guid_tbl));

EXIT:
	return res;
}

static void handle_topology_changes(IN osm_topo_mgr_t * p_mgr,
				    IN osm_topology_t *p_topo)
{
	osm_opensm_t *p_osm = p_mgr->p_osm;
	osm_topology_t *prev_topo = p_mgr->prev_topologies[p_topo->id];

	if (!prev_topo) {
		if (p_osm->subn.opt.use_ucast_cache) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_INFO,
				"Topology %" PRIu64
				" not found in prev topologies,"
				" invalidating cache\n",
				p_topo->id);
			p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
		}
		return;
	}

	if (prev_topo->removed_hcas) {
		p_topo->removed_hcas = prev_topo->removed_hcas;
		prev_topo->removed_hcas = NULL;
	}

	handle_subnet_changes(p_osm, prev_topo->p_subn, p_topo);
}


int osm_topo_mgr_create_completer_topo(IN struct osm_opensm *p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;
	osm_port_group_t *sw_grp = NULL, *hca_grp = NULL, *p_grp;
	osm_subn_t *p_subn = &p_osm->subn;
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_topology_t *p_topo;
	cl_list_iterator_t iter;
	boolean_t in_topo, is_sw;
	boolean_t bs = FALSE;
	ib_net64_t guid;
	uint8_t i;
	int res = 0;

	/* Create sw and hca port groups for completer topology */
	sw_grp = osm_create_port_grp(p_osm, "completer topo sw", NULL, NULL,
				     FALSE);
	hca_grp = osm_create_port_grp(p_osm, "completer topo hca", NULL, NULL,
				      FALSE);

	if (!sw_grp || !hca_grp) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6612: "
			"Failed to create completer port groups\n");
		res = 1;
		goto EXIT;
	}

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		p_node = p_port->p_node;
		bs = FALSE;
		is_sw = (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH);
		if (is_sw)
			p_grp = sw_grp;
		else
			p_grp = hca_grp;

		in_topo = (cl_list_count(&p_port->topology_list) > 0);

		if (in_topo && is_sw) {
			/*
			 * If itr engine exist it was last so must be found
			 * at the tail
			 */
			iter =  cl_list_head(&p_port->topology_list);
			p_topo = osm_port_get_topo_from_list_iter(&iter);
			guid = osm_node_get_node_guid(p_node);
			bs = osm_topo_is_border_sw(p_topo, guid);
		}
		if (!in_topo || bs)
			osm_set_port_in_grp(p_subn, p_grp, p_port, TRUE);

		if (bs) {
			/* Remove physps of bs that belong to topology */
			for (i = 1; i < p_node->physp_tbl_size; i++) {
				if (p_node->physp_table[i].topo_id != NO_TOPO_ID) {
					osm_set_physp_in_grp_by_num(p_subn,
								    p_grp,
								    p_port, i,
								    FALSE);
				}
			}
		}
		p_item = cl_qmap_next(p_item);
	}


	/* Create topoology from the port gorups */
	if (osm_topo_mgr_add_topo(p_osm, MAIN_TOPO_ID, sw_grp, hca_grp) !=
    IB_SUCCESS) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6611: "
			"Failed to create completer topology\n");
		res = 1;
		goto EXIT;
	}

	handle_topology_changes(p_mgr, p_mgr->topologies[MAIN_TOPO_ID]);

EXIT:
	osm_destroy_port_grp(sw_grp);
	osm_destroy_port_grp(hca_grp);
	return res;
}

int osm_topo_mgr_create_non_itr_topo(IN struct osm_opensm *p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;
	osm_port_group_t *sw_grp = NULL, *hca_grp = NULL, *p_grp;
	osm_subn_t *p_subn = &p_osm->subn;
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_topology_t *p_topo;
	uint64_t topo_id;
	cl_list_iterator_t iter;
	boolean_t in_topo, is_sw, in_itr;
	boolean_t bs = FALSE;
	ib_net64_t guid;
	uint8_t i;
	int res = 0;
	struct osm_routing_engine *re;

	/* Create sw and hca port groups fornon-itr topology */
	sw_grp = osm_create_port_grp(p_osm, "non-itr topo sw", NULL, NULL,
				     FALSE);
	hca_grp = osm_create_port_grp(p_osm, "non-itr topo hca", NULL, NULL,
				      FALSE);
	if (!sw_grp || !hca_grp) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6613: "
			"Failed to create non-itr port groups\n");
		res = 1;
		goto EXIT;
	}

	p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t*)p_item;
		p_node = p_port->p_node;
		in_itr = FALSE;
		bs = FALSE;
		is_sw = (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH);
		if (is_sw)
			p_grp = sw_grp;
		else
			p_grp = hca_grp;

		in_topo = (cl_list_count(&p_port->topology_list) > 0);

		if (in_topo) {
			/*
			 * If itr engine exist it was last so must be found
			 * at the tail
			 */
			iter =  cl_list_head(&p_port->topology_list);
			p_topo = osm_port_get_topo_from_list_iter(&iter);
			re = p_topo->p_subn->routing_engine_used;
			if (re && re->ITR_init)
				in_itr = TRUE;
			if (is_sw) {
				guid = osm_node_get_node_guid(p_node);
				bs = osm_topo_is_border_sw(p_topo, guid);
			}
		}

		if (!in_itr || bs)
			osm_set_port_in_grp(p_subn, p_grp, p_port, TRUE);

		if (in_itr && bs) {
			/* Remove physps of bs that belong to topology */
			for (i = 1; i < p_node->physp_tbl_size; i++) {
				topo_id = p_node->physp_table[i].topo_id;
				if ((topo_id != NO_TOPO_ID) &&
				    (topo_id != MAIN_TOPO_ID))
					osm_set_physp_in_grp_by_num(p_subn,
								    p_grp,
								    p_port, i,
								    FALSE);
			}
		}
		p_item = cl_qmap_next(p_item);
	}


	/* Create topoology from the port gorups */
	if (osm_topo_mgr_add_topo(p_osm, NON_ITR_TOPO_ID, sw_grp, hca_grp) !=
    IB_SUCCESS) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 6614: "
			"Failed to create non-itr topology\n");
		res = 1;
		goto EXIT;
	}

	handle_topology_changes(p_mgr, p_mgr->topologies[MAIN_TOPO_ID]);

EXIT:
	osm_destroy_port_grp(sw_grp);
	osm_destroy_port_grp(hca_grp);
	return res;
}

static void handle_topologies_changes(IN osm_topo_mgr_t *p_mgr)
{
	uint8_t i;
	osm_opensm_t *p_osm = p_mgr->p_osm;

	for (i = 1 ; i < TOPO_ARRAY_SIZE ; i++) {
		if (p_mgr->topologies[i])
			handle_topology_changes(p_mgr, p_mgr->topologies[i]);
		else if (p_mgr->prev_topologies[i] &&
			 p_osm->subn.opt.use_ucast_cache) {
			OSM_LOG(p_osm->sm.p_log, OSM_LOG_INFO, "Topology %"
				PRIu8 " found in prev topologies but not in "
				"current topologies - Invalidating cache\n", i);
			p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
		}
	}
}

ib_api_status_t osm_topo_mgr_create(IN osm_opensm_t * p_osm)
{
	osm_topo_mgr_t *p_mgr = &p_osm->sm.topology_mgr;

	p_mgr->p_osm = p_osm;
	p_mgr->initialized = TRUE;
	p_mgr->creation_result = IB_ERROR;

	if (!p_osm->subn.opt.topo_policy_file) {
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "No topology policy "
			"file found, topology manager not created\n");
		osm_topo_mgr_destroy(p_mgr);
		p_mgr->creation_result = IB_SUCCESS;
		goto EXIT;
	}

	OSM_LOG(p_osm->sm.p_log, OSM_LOG_DEBUG, "creating topologies manager\n");
	if (osm_ports_grp_count(p_osm) == 0) {
		osm_topo_mgr_destroy(p_mgr);
		OSM_LOG(p_osm->sm.p_log, OSM_LOG_ERROR, "ERR 660A: No Port "
			"Groups Found - Aborting topology parsing\n");
		goto EXIT;
	}

	move_topologies_to_prev(p_mgr);

	if (parse_topo_policy(p_osm)) {
		osm_topo_mgr_destroy(p_mgr);
		goto EXIT;
	}

	if (p_osm->subn.opt.auto_join_hca_wo_topo)
		auto_join_lost_hcas(p_osm);

	handle_topologies_changes(p_mgr);

	if (!topo_mgr_validate(p_osm)) {
		osm_topo_mgr_destroy(p_mgr);
		goto EXIT;
	}

	topo_mgr_check_topo0_hub_switch(p_osm);

	p_mgr->creation_result = IB_SUCCESS;
EXIT:
	return p_mgr->creation_result;
}
