/*
 * Copyright (c) 2016-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of DOR routing algorithm
 *      Calculation functions
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_DOR_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>


#define UCAST_HC_STATE_UNKNOWN	0
#define UCAST_HC_STATE_QUEUED	1
#define UCAST_HC_STATE_FAILED	2
#define UCAST_HC_STATE_COMPLETE	3

#define UCAST_HC_MAX_DIM	15

#define UCAST_HC_COORDS_FILE	"opensm-dor-coordinates"

struct hc_vertex;

typedef struct hc_vertex {
	unsigned int index;
	osm_switch_t *p_sw;
	uint8_t degree;
	struct hc_vertex *adjacents[UCAST_HC_MAX_DIM];
	unsigned int coord;
	unsigned int adj_dim_mask;
	uint8_t coord_assign_state;
} hc_vertex_t;
/*
 * FIELDS
 * 	index
 * 		Index of this vertex.
 *
 *	p_sw
 *		Pointer to osm_switch_t object represented by this vertex.
 *
 * 	degree
 * 		Degree of this vertex.
 *
 *	adjacents
 *		Array of adjacent vertices.
 *
 * 	coord
 * 		Coordinates assigned to this vertex.
 *
 * 	adj_dim_mask
 * 		Bit-mask of identified adjacent vertices by dimension.
 *
 * 	coord_assign_state
 * 		State of vertex during coordinates assigning phase.
 */

typedef struct hc {
	osm_subn_t *p_subn;
	hc_vertex_t *vertices;
	hc_vertex_t **coord_vector;
	unsigned int num_vertices;
	uint8_t dimensions;
} hc_t;
/*
 * FIELDS
 * 	p_subn
 * 		Pointer to osm_subn_t object.
 * 	vertices
 * 		Array of vertices.
 *
 * 	coord_vector
 * 		Map from coordinate to vertex.
 *
 * 	num_vertices
 * 		Number of vertices.
 *
 * 	dimensions
 * 		Number of dimensions.
 */


static inline
hc_vertex_t *ucast_hc_get_vertex(hc_t *hc, unsigned int index)
{
	OSM_ASSERT(hc->vertices[index].index == index);
	return &hc->vertices[index];
}


static inline
uint64_t ucast_hc_get_vertex_guid(hc_vertex_t *vertex)
{
	osm_node_t *p_node = vertex->p_sw->p_node;
	return cl_ntoh64(osm_node_get_node_guid(p_node));
}


static
int ucast_hc_count_bits(unsigned int n)
{
	int count = 0;

	while (n) {
		n &= (n - 1);
		count++;
	}

	return count;
}


static inline
unsigned int ucast_hc_get_edge_dim(hc_vertex_t *vertex1, hc_vertex_t *vertex2)
{
	return vertex1->coord ^ vertex2->coord;
}


static inline
void ucast_hc_set_vertex_coord(hc_t *hc, hc_vertex_t *vertex, unsigned int coord)
{
	hc->coord_vector[coord] = vertex;
	vertex->coord = coord;
	vertex->p_sw->coord = coord;
}


static inline
hc_vertex_t *ucast_hc_get_vertex_by_coord(hc_t *hc, unsigned int coord)
{
	return hc->coord_vector[coord];
}


static
void ucast_hc_notify_coord_adjacents(hc_t *hc, hc_vertex_t *vertex)
{
	unsigned int coord, adj_coord, dimension_bit;
	hc_vertex_t *adj;

	OSM_ASSERT(vertex->coord != OSM_SW_NO_COORD);
	coord = vertex->coord;

	dimension_bit = 1 << (hc->dimensions - 1);
	while (dimension_bit) {
		adj_coord = coord ^ dimension_bit;
		adj = ucast_hc_get_vertex_by_coord(hc, adj_coord);
		if (adj) {
			vertex->adj_dim_mask |= dimension_bit;
			adj->adj_dim_mask |= dimension_bit;
		}
		dimension_bit = dimension_bit >> 1;
	}
}


static
hc_vertex_t *ucast_hc_find_seed(hc_t *hc)
{
	hc_vertex_t *vertex;
	hc_vertex_t *seed = NULL;
	osm_port_t *p_port;
	unsigned int num_vertices = hc->num_vertices;
	unsigned int max_degree = 0;
	unsigned int max_adj_degree = 0;
	unsigned int index, adj_index;
	uint8_t *sum_adj_deg_array;

	sum_adj_deg_array = (uint8_t *)calloc(num_vertices, sizeof(uint8_t));
	if (!sum_adj_deg_array) {
		OSM_LOG(&hc->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A41: cannot allocate memory for seed"
			" selection\n");
		return NULL;
	}

	seed = ucast_hc_get_vertex(hc, 0);

	/*
	 * Seek for vertex with maximal number of 2 hops neighbours.
	 */
	for (index = 0; index < num_vertices; index++) {
		vertex = ucast_hc_get_vertex(hc, index);
		for (adj_index = 0; adj_index < vertex->degree; adj_index++)
			sum_adj_deg_array[index] +=
			    vertex->adjacents[adj_index]->degree;
	}

	for (index = 0; index < num_vertices; index++) {
		vertex = ucast_hc_get_vertex(hc, index);
		p_port = osm_switch_get_port0(vertex->p_sw, hc->p_subn);
		if (p_port->disc_through == OSM_DISC_NORMAL &&
		    (vertex->degree > max_degree ||
		     (vertex->degree == max_degree &&
		      sum_adj_deg_array[index] > max_adj_degree))) {
			seed = vertex;
			max_degree = vertex->degree;
			max_adj_degree = sum_adj_deg_array[index];
		}
	}

	free(sum_adj_deg_array);

	return seed;
}


static
int ucast_hc_init(hc_t *hc, osm_subn_t *p_subn)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	osm_switch_t *p_rem_sw;
	osm_node_t *p_node;
	osm_node_t *p_rem_node;
	osm_port_t *p_port;
	hc_vertex_t *vertices = NULL;
	hc_vertex_t *vertex, *adj;
	uint8_t *adj_matrix = NULL;
	hc_vertex_t **coord_vector = NULL;
	unsigned int num_vertices;
	unsigned int num_isolated_vertices = 0;
	unsigned int num_edges = 0;
	unsigned int index, adj_index;
	uint8_t degree, max_degree = 0;
	uint8_t port, port_num;
	int result = 0;

	/* Create vertices array. */
	num_vertices = cl_qmap_count(&p_subn->sw_guid_tbl);
	if (num_vertices == 0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A11: cannot route topology without switches\n");
		result = 1;
		goto exit;
	}


	vertices = (hc_vertex_t *)calloc(num_vertices, sizeof(hc_vertex_t));
	if (!vertices) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A12: cannot allocate memory for vertices\n");
		result = 1;
		goto exit;
	}

	index = 0;
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		vertex = &vertices[index];
		vertex->index = index;
		vertex->p_sw = p_sw;
		p_sw->priv = vertex;
		vertex->coord = OSM_SW_NO_COORD;
		index++;
		/* Count number of isolated nodes */
		p_port = osm_switch_get_port0(vertex->p_sw, p_subn);
		if (p_port->disc_through != OSM_DISC_NORMAL)
			num_isolated_vertices++;
	}

	/* Create adjacency matrix and adjacency vector. */
	adj_matrix = (uint8_t *)calloc(num_vertices, sizeof(uint8_t));
	if (!adj_matrix) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A13: cannot allocate memory for adjacency matrix\n");
		result = 1;
		goto exit;
	}

	for (index = 0; index < num_vertices; index++) {
		memset(adj_matrix, 0, num_vertices);
		vertex = &vertices[index];
		p_sw = vertex->p_sw;
		degree = 0;
		for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
			port = osm_switch_get_dimn_port(p_sw, port_num);
			p_node = p_sw->p_node;
			p_rem_node = osm_node_get_remote_node(p_node, port,
							      NULL);
			if (!p_rem_node || !p_rem_node->sw ||
			    p_rem_node == p_node)
				continue;

			p_rem_sw = p_rem_node->sw;
			adj_index = ((hc_vertex_t *)p_rem_sw->priv)->index;

			if (adj_matrix[adj_index])
				continue;

			adj_matrix[adj_index] = 1;
			adj = &vertices[adj_index];
			vertex->adjacents[degree] = adj;

			degree++;

			if (degree > UCAST_HC_MAX_DIM) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7A14: cannot route topology with"
					" degree (%d) > %d\n",
					degree,
					UCAST_HC_MAX_DIM);
				result = 1;
				goto exit;
			}
		}

		vertex->degree = degree;
		num_edges += degree;

		if (max_degree < degree)
			max_degree = degree;

		max_degree = max_degree > degree ? max_degree : degree;
	}

	if (num_edges == 0 && num_vertices > 1) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A15: cannot route topology without switch to"
			" switch links\n");
		result = 1;
		goto exit;
	}

	if ((1 << max_degree) < (num_vertices - num_isolated_vertices)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A16: maximum degree (%d) does not match number"
			" of valid switches (%d/%d)\n",
			max_degree, num_vertices - num_isolated_vertices,
			num_vertices);
		result = 1;
		goto exit;
	}

	/* Create coordinates to vertex map. */
	coord_vector = (hc_vertex_t**)calloc(1 << max_degree, sizeof(hc_vertex_t*));
	if (!coord_vector) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A17: cannot allocate memory for"
			" coordinates vector\n");
		result = 1;
		goto exit;
	}

exit:
	if (result) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A18: Hyper_Cube analysis failed\n");

		free(vertices);
		free(coord_vector);
		vertices = NULL;
		coord_vector = NULL;
	}

	free(adj_matrix);

	hc->p_subn = p_subn;
	hc->vertices = vertices;
	hc->num_vertices = num_vertices;
	hc->coord_vector = coord_vector;
	hc->dimensions = max_degree;

	return result;
}


static
void ucast_hc_destroy(hc_t *hc)
{
	free(hc->vertices);
	free(hc->coord_vector);
}


static
unsigned int ucast_hc_get_last_missing_adj_bit(hc_t *hc, hc_vertex_t *vertex)
{
	unsigned int mask = (1 << hc->dimensions) - 1;
	unsigned int dimension_bit = mask & ~vertex->adj_dim_mask;

	if (!(dimension_bit & (dimension_bit - 1))) {
		/* Only one bit is on. */
		return dimension_bit;
	}

	return 0;
}


static
int ucast_hc_validate_coordinates(hc_t *hc)
{
	osm_log_t *p_log = &hc->p_subn->p_osm->log;
	hc_vertex_t *vertex;
	osm_port_t *p_port;
	unsigned int num_vertices = hc->num_vertices;
	unsigned int index;
	int result = 0;

	for (index = 0; index < num_vertices; index++) {
		vertex = ucast_hc_get_vertex(hc, index);
		OSM_ASSERT(vertex);
		if (vertex->coord == OSM_SW_NO_COORD) {
			p_port = osm_switch_get_port0(vertex->p_sw, hc->p_subn);
			if (p_port->disc_through == OSM_DISC_NORMAL) {
				OSM_LOG(p_log, OSM_LOG_VERBOSE,
					"Warning: failed to assign coordinates"
					" to switch 0x%" PRIx64 "\n",
					ucast_hc_get_vertex_guid(vertex));
				result = 1;
			} else {
				/* Avoid failing validation on isolated nodes */
				OSM_LOG(p_log, OSM_LOG_VERBOSE,
					"Warning: failed to assign coordinates"
					" to isolated switch 0x%" PRIx64 "\n",
					ucast_hc_get_vertex_guid(vertex));
			}
			continue;
		}
	}

	return result;
}


static
int ucast_hc_validate_port_numbers(hc_t *hc)
{
	osm_log_t *p_log = &hc->p_subn->p_osm->log;
	hc_vertex_t *vertex, *adj;
	unsigned int index;
	unsigned int edge_dim, prev_edge_dim;
	unsigned int num_vertices = hc->num_vertices;
	osm_switch_t *p_sw;
	osm_node_t *p_node1, *p_node2;
	uint8_t port1_order, port2_order, port1, port2;
	int result = 0;

	/*
	 * Performs the following validations:
	 * 1. Port number on both ends of switch-to-switch link are equal.
	 * 2. Link of lower dimension has lower port number.
	 */
	for (index = 0; index < num_vertices; index++) {
		vertex = ucast_hc_get_vertex(hc, index);
		OSM_ASSERT(vertex);

		p_sw = vertex->p_sw;
		p_node1 = p_sw->p_node;
		prev_edge_dim = 0;

		for (port1_order = 0; port1_order < p_sw->num_ports;
		     port1_order++) {
			port1 = osm_switch_get_dimn_port(p_sw, port1_order);
			p_node2 = osm_node_get_remote_node(p_node1, port1,
							   &port2);
			if (!p_node2 || !p_node2->sw)
				continue;

			port2_order = osm_switch_get_rev_dimn_port(p_node2->sw,
								   port2);
			adj = (hc_vertex_t *)p_node2->sw->priv;

			/* Validate remote port number.*/
			if (port1_order != port2_order) {
				OSM_LOG(p_log, OSM_LOG_VERBOSE,
					"Warning: switch 0x%" PRIx64 " port %d"
					" (search order %d) is connected to"
					" remote port (0x%" PRIx64 " %d) with"
					" different search order (%d)\n",
					ucast_hc_get_vertex_guid(vertex),
					port1, port1_order,
					ucast_hc_get_vertex_guid(adj),
					port2, port2_order);
				result = 1;
			}

			/* Validate dimensions order. */
			if (vertex->coord != OSM_SW_NO_COORD &&
			    adj->coord != OSM_SW_NO_COORD) {
				edge_dim = ucast_hc_get_edge_dim(vertex, adj);
				if (edge_dim < prev_edge_dim) {
					OSM_LOG(p_log, OSM_LOG_VERBOSE,
						"Warning: dimension of switch"
						" 0x%" PRIx64 " port %d does"
						" not match ports order"
						" (dimension: 0x%04x)\n",
						ucast_hc_get_vertex_guid(vertex),
						port1, edge_dim);
					result = 1;
				}
				prev_edge_dim = edge_dim;
			}
		} /* port */
	} /* vertex */

	return result;
}


/*
 * Calculate coordinates for a given vertex according to the coordinates
 * of its adjacent vertices.
 * Return values:
 * 0 - Failed to calculate coordinates
 * 1 - Calculated coordinates successfully
 * 2 - Error
 */
static
int ucast_hc_process_vertex(hc_t *hc, hc_vertex_t *vertex,
			    uint8_t enable_heuristics)
{
	osm_log_t *p_log = &hc->p_subn->p_osm->log;
	hc_vertex_t *adj1 = NULL;
	hc_vertex_t *adj2 = NULL;
	hc_vertex_t *adj3 = NULL;
	unsigned int adj1_coord = OSM_SW_NO_COORD;
	unsigned int adj2_coord = OSM_SW_NO_COORD;
	unsigned int adj3_coord = OSM_SW_NO_COORD;
	unsigned int dim_diff, coord = OSM_SW_NO_COORD;
	uint8_t adj1_index, adj2_index, adj3_index, degree;

	OSM_ASSERT(vertex->coord == OSM_SW_NO_COORD);

	/* Find 3 adjacent vertices with known coordinates. */
	degree = vertex->degree;
	for (adj1_index = 0; adj1_index < degree; adj1_index++) {
		adj1 = vertex->adjacents[adj1_index];
		adj1_coord = adj1->coord;
		if (adj1_coord != OSM_SW_NO_COORD)
			break;
	}
	for (adj2_index = adj1_index + 1; adj2_index < degree; adj2_index++) {
		adj2 = vertex->adjacents[adj2_index];
		adj2_coord = adj2->coord;
		if (adj2_coord != OSM_SW_NO_COORD)
			break;
	}
	for (adj3_index = adj2_index + 1; adj3_index < degree; adj3_index++) {
		adj3 = vertex->adjacents[adj3_index];
		adj3_coord = adj3->coord;
		if (adj3_coord != OSM_SW_NO_COORD)
			break;
	}

	if (adj3_coord != OSM_SW_NO_COORD) {
		/*
		 * Use the coordinates of 3 adjacent vertices to calculate the
		 * coordinates of this vertex.
		 */
		dim_diff = (adj1_coord ^ adj2_coord) &
			   (adj1_coord ^ adj3_coord);
		if (ucast_hc_count_bits(dim_diff) != 1) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7A31: failed to calculate coordinates for"
				" 0x%" PRIx64 " based on:"
				" 0x%" PRIx64 " (%04x),"
				" 0x%" PRIx64 " (%04x),"
				" 0x%" PRIx64 " (%04x)\n",
				ucast_hc_get_vertex_guid(vertex),
				ucast_hc_get_vertex_guid(adj1),
				adj1_coord,
				ucast_hc_get_vertex_guid(adj2),
				adj2_coord,
				ucast_hc_get_vertex_guid(adj3),
				adj3_coord);
			return 2;
		}

		coord = adj1_coord ^ dim_diff;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Assigning coordinate %04x to 0x%" PRIx64 " based on:"
			" 0x%" PRIx64 " (%04x),"
			" 0x%" PRIx64 " (%04x),"
			" 0x%" PRIx64 " (%04x)\n",
			coord,
			ucast_hc_get_vertex_guid(vertex),
			ucast_hc_get_vertex_guid(adj1),
			adj1_coord,
			ucast_hc_get_vertex_guid(adj2),
			adj2_coord,
			ucast_hc_get_vertex_guid(adj3),
			adj3_coord);

		goto assign_coordinates;
	}

	if (adj2_coord != OSM_SW_NO_COORD) {
		OSM_ASSERT(adj1_coord);
		/*
		 * Use the coordinates of two adjacent vertices to calculate
		 * the coordinates of this vertex.
		 * When we have coordinates of two opposite corners of a face,
		 * we can calculate the coordinates of the other two corners.
		 * If we already assigned one of the coordinates to a another
		 * vertex, the other set of coordinates is of this vertex.
		 */
		dim_diff = adj1_coord ^ adj2_coord;
		if (ucast_hc_count_bits(dim_diff) != 2) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7A32: failed to calculate coordinates for"
				" 0x%" PRIx64 " based on:"
				" 0x%" PRIx64 " (%04x),"
				" 0x%" PRIx64 " (%04x)\n",
				ucast_hc_get_vertex_guid(vertex),
				ucast_hc_get_vertex_guid(adj1),
				adj1_coord,
				ucast_hc_get_vertex_guid(adj2),
				adj2_coord);
			return 2;
		}

		/* Choose one dimension bit. */
		dim_diff &= (dim_diff - 1);

		if (ucast_hc_get_vertex_by_coord(hc, adj1_coord ^ dim_diff)) {
			coord = adj2_coord ^ dim_diff;

			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Assigning coordinate %04x to 0x%" PRIx64 " based on:"
				" 0x%" PRIx64 " (%04x),"
				" 0x%" PRIx64 " (%04x)\n",
				coord,
				ucast_hc_get_vertex_guid(vertex),
				ucast_hc_get_vertex_guid(adj1),
				adj1_coord,
				ucast_hc_get_vertex_guid(adj2),
				adj2_coord);

			goto assign_coordinates;
		}

		if (ucast_hc_get_vertex_by_coord(hc, adj2_coord ^ dim_diff)) {
			coord = adj1_coord ^ dim_diff;

			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Assigning coordinate %04x to 0x%" PRIx64 " based on:"
				" 0x%" PRIx64 " (%04x),"
				" 0x%" PRIx64 " (%04x)\n",
				coord,
				ucast_hc_get_vertex_guid(vertex),
				ucast_hc_get_vertex_guid(adj1),
				adj1_coord,
				ucast_hc_get_vertex_guid(adj2),
				adj2_coord);

			goto assign_coordinates;
		}
	}

	if (adj2_coord != OSM_SW_NO_COORD &&
	    (dim_diff = ucast_hc_get_last_missing_adj_bit(hc, adj2))) {
		coord = adj2_coord ^ dim_diff;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Assigning coordinate %04x to 0x%" PRIx64 " based on:"
			" 0x%" PRIx64 " (%04x)\n",
			coord,
			ucast_hc_get_vertex_guid(vertex),
			ucast_hc_get_vertex_guid(adj2),
			adj2_coord);

		goto assign_coordinates;
	}

	if (adj1_coord != OSM_SW_NO_COORD &&
	    (dim_diff = ucast_hc_get_last_missing_adj_bit(hc, adj1))) {
		coord = adj1_coord ^ dim_diff;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Assigning coordinate %04x to 0x%" PRIx64
			" based on: 0x%" PRIx64 " (%04x)\n",
			coord,
			ucast_hc_get_vertex_guid(vertex),
			ucast_hc_get_vertex_guid(adj1),
			adj1_coord);

		goto assign_coordinates;
	}

	if (enable_heuristics && adj2_coord != OSM_SW_NO_COORD) {
		/*
		 * There are two known adjacent vertices, and two possible
		 * coordinates to assign.
		 */
		dim_diff = adj1_coord ^ adj2_coord;
		dim_diff &= (dim_diff - 1);
		coord = adj2_coord ^ dim_diff;

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Assigning coordinate %04x to 0x%" PRIx64
			" using heuristics based on:"
			" 0x%" PRIx64 " (%04x),"
			" 0x%" PRIx64 " (%04x)\n",
			coord,
			ucast_hc_get_vertex_guid(vertex),
			ucast_hc_get_vertex_guid(adj1),
			adj1_coord,
			ucast_hc_get_vertex_guid(adj2),
			adj2_coord);

		goto assign_coordinates;
	}

	if (enable_heuristics && adj1_coord != OSM_SW_NO_COORD) {
		uint8_t edge_dim = 0;
		uint8_t offset;
		/* Single known adjacent vertex. */

		for (adj2_index = 0;
		     adj2_index < adj1->degree;
		     adj2_index++) {
			adj2 = adj1->adjacents[adj2_index];
			if (adj2 == adj1) {
				break;
			}

			if (adj2->coord != OSM_SW_NO_COORD) {
				edge_dim =
				    ffs(ucast_hc_get_edge_dim(adj1, adj2)) - 1;
			} else {
				edge_dim++;
			}
		}

		for (offset = 0; offset < hc->dimensions; offset++) {
			coord = adj1->coord ^
				(1 << ((edge_dim + offset) % hc->dimensions));
			if (!ucast_hc_get_vertex_by_coord(hc, coord)) {
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Assigning coordinate %04x to 0x%" PRIx64
					" using heuristics"
					" based on: 0x%" PRIx64 " (%04x)\n",
					coord,
					ucast_hc_get_vertex_guid(vertex),
					ucast_hc_get_vertex_guid(adj1),
					adj1_coord);

				goto assign_coordinates;
			}
		}
		coord = OSM_SW_NO_COORD;
	}

assign_coordinates:

	if (coord != OSM_SW_NO_COORD) {
		/*
		 * Validate that calculated coordinates are not assigned to
		 * another vertex.
		 */
		if (ucast_hc_get_vertex_by_coord(hc, coord)) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7A33: coordinates (%04x) calculated for"
				" 0x%" PRIx64 " were already assigned to"
				" 0x%" PRIx64 "\n",
				coord,
				ucast_hc_get_vertex_guid(vertex),
				ucast_hc_get_vertex_guid(
				    ucast_hc_get_vertex_by_coord(hc, coord)));
			return 2;
		}

		/* Assign coordinates */
		ucast_hc_set_vertex_coord(hc, vertex, coord);
		return 1;
	}

	/* Not enough information. */
	return 0;
}


static
void ucast_hc_dump_coordinates_pfn(FILE *file, void *cxt)
{
	hc_t *hc = (hc_t *)cxt;
	hc_vertex_t *vertex;
	unsigned int index;

	fprintf(file, "# hyper-cube number of dimensions: %d\n\n",
		hc->dimensions);

	for (index = 0; index < hc->num_vertices; index++) {
		vertex = ucast_hc_get_vertex(hc, index);
		fprintf(file, "0x%016" PRIx64 " 0x%04x\n",
			ucast_hc_get_vertex_guid(vertex),
			vertex->coord);
	}
}


static
int ucast_hc_dump_coordinates(hc_t *hc)
{
	char path[256];

	if (hc->p_subn->routing_chain_engine_id)
		snprintf(path, sizeof(path), "%s/%s.%d",
			 hc->p_subn->opt.dump_files_dir, UCAST_HC_COORDS_FILE,
			 hc->p_subn->routing_chain_engine_id);
	else
		snprintf(path, sizeof(path), "%s/%s",
			 hc->p_subn->opt.dump_files_dir, UCAST_HC_COORDS_FILE);

	osm_dump_func_to_file(hc->p_subn->p_osm, path,
			      ucast_hc_dump_coordinates_pfn,
			      hc);
	return 0;
}


static
int ucast_hc_set_arbitrary_coordinates(hc_t *hc, hc_vertex_t *vertex,
				       unsigned int coord)
{
	hc_vertex_t *adj;
	unsigned int dim, adj_index;
	/*
	 * Set coordinates of the seed vertex and it's adjacent vertices.
	 */
	ucast_hc_set_vertex_coord(hc, vertex, coord);
	dim = 1;
	for (adj_index = 0; adj_index < vertex->degree; adj_index++) {
		adj = vertex->adjacents[adj_index];
		ucast_hc_set_vertex_coord(hc, adj, coord ^ dim);
		dim = dim << 1;
	}

	return 0;
}

int osm_ucast_hc_process(osm_subn_t * p_subn)
{
	hc_t hc;
	hc_vertex_t *seed_vertex;
	hc_vertex_t *vertex, *adj;
	unsigned int adj_index;
	hc_vertex_t **queue = NULL;
	unsigned int queue_in = 0;
	unsigned int queue_out = 0;
	unsigned int queue_iter = 0;
	int iteration = 1;
	int result = 0;
	uint8_t enable_heuristics = 0;
	int discovered = 0;
	int res;


	/*
	 * This routine assigns coordinates to each switch of a Hyper-Cube
	 * topology. Coordinates calculation is performed in the followings
	 * steps:
	 * 1. Create graph of the switches of the topology (every vertex
	 *    represent switch, no parallel links, no representation to other
	 *    node types).
	 * 2. Select seed and assign coordinates to it, and to its adjacent
	 *    vertices.
	 * 3. Assign coordinates to the rest of the vertices, using BFS from
	 *    seed vertex.
	 */

	if (ucast_hc_init(&hc, p_subn)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A21: cannot initialize routing algorithm\n");
		result = 1;
		goto exit;
	}

	/* Select seed and assign coordinates to it and its adjacent vertices */
	seed_vertex = ucast_hc_find_seed(&hc);
	if (!seed_vertex) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A22: cannot find hypercube seed\n");
		result = 1;
		goto exit;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Selected switch 0x%016" PRIx64 " as seed\n",
		ucast_hc_get_vertex_guid(seed_vertex));

	ucast_hc_set_arbitrary_coordinates(&hc, seed_vertex, 0);

	/*
	 * Start assigning coordinates to the reset of the vertices using BFS
	 * from the selected seed.
	 *
	 * Each vertex can be renter the queue 3 times at most.
	 * Every time we successfully calculate the coordinates of a vertex, we
	 * add adjacent vertices without coordinates.
	 * If a vertex has 3 adjacent vertices, we calculate its coordinates.
	 */
	queue = (hc_vertex_t **)calloc(hc.num_vertices * 3,
				       sizeof(hc_vertex_t*));
	if (!queue) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A23: cannot allocate memory for routing"
			" engine\n");
		result = 1;
		goto exit;
	}

	queue[queue_in++] = seed_vertex;
	while (iteration < hc.num_vertices) {
		discovered = 0;
		res = 0;

		/* Coordinates assigning BFS */
		while (queue_out < queue_in) {
			vertex = queue[queue_out++];

			res = vertex->coord != OSM_SW_NO_COORD ? 1 :
			      ucast_hc_process_vertex(&hc, vertex,
						      enable_heuristics);
			switch (res) {
			case 0:
				/* Failed to calculate coordinates. */
				vertex->coord_assign_state = UCAST_HC_STATE_FAILED;
				break;
			case 1:
				/* Coordinate calculated successfully. */
				discovered++;
				vertex->coord_assign_state = UCAST_HC_STATE_COMPLETE;
				ucast_hc_notify_coord_adjacents(&hc, vertex);

				/* Add adjacent vertices to queue. */
				for (adj_index = 0; adj_index < vertex->degree;
				     adj_index++) {
					adj = vertex->adjacents[adj_index];
					if (adj->coord_assign_state == UCAST_HC_STATE_COMPLETE ||
					    adj->coord_assign_state == UCAST_HC_STATE_QUEUED)
						continue;
					adj->coord_assign_state = UCAST_HC_STATE_QUEUED;
					queue[queue_in++] = adj;
				}
				break;
			case 2:
				/* Topology error */
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7A24: found topology error on"
					" switch 0x%" PRIx64 "\n",
					ucast_hc_get_vertex_guid(vertex));
				result = 1;
				goto exit;
			}
		} /* Coordinates assigning BFS end */

		if (discovered == 0) {
			if (enable_heuristics) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7A25: Failed to assign new"
					" coordinates after %d iterations\n",
					iteration);
				result = 2;
				goto exit;
			} else {
				/* Enable heuristic coordinates calculation */
				enable_heuristics = 1;
			}
		}

		/* Retry failed calculation vertices to head of the queue */
		queue_in = 0;
		for (queue_iter = 0; queue_iter < queue_out; queue_iter++) {
			vertex = queue[queue_iter];
			if (vertex->coord_assign_state == UCAST_HC_STATE_FAILED) {
				vertex->coord_assign_state = UCAST_HC_STATE_QUEUED;
				queue[queue_in++] = vertex;
			}
		}

		if (queue_in == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Coordinates assignment completed in %d"
				" iterations\n",
				iteration);
			break;
		}

		queue_out = 0;
		iteration++;
	}

	/* Dump coordinates to file. */
	ucast_hc_dump_coordinates(&hc);

	/* Validate coordinates. */
	if (ucast_hc_validate_coordinates(&hc)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A26: coordinates assignment validation failed\n");
		result = 1;
		goto exit;
	}

	/* Check port numbers violations. */
	if (ucast_hc_validate_port_numbers(&hc)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A27: port numbers validation failed\n");
	}

exit:

	if (result == 0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Hyper-Cube analysis completed successfully\n");
	} else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7A28: Hyper-Cube analysis failed\n");
	}

	free(queue);

	/* Destroy data structure. */
	ucast_hc_destroy(&hc);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return result;
}


static
int ucast_dor_build_lid_matrices(void *context)
{

	int result;
	osm_ucast_mgr_t *p_ucast_mgr = (osm_ucast_mgr_t *)context;

	result = osm_ucast_mgr_build_lid_matrices(p_ucast_mgr);
	if (result) {
		return result;
	}

	if (p_ucast_mgr->p_subn->opt.dor_hyper_cube_mode) {
		osm_ucast_hc_process(p_ucast_mgr->p_subn);
	}

	return result;
}


int osm_ucast_dor_build_lfts(void *context)
{
	osm_ucast_mgr_t *mgr = context;
	int ret;

	mgr->is_dor = 1;
	ret = osm_ucast_mgr_build_lft_tables(mgr);
	mgr->is_dor = 0;

	return ret;
}


int osm_ucast_dor_setup(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	r->context = &osm->sm.ucast_mgr;
	r->build_lid_matrices = ucast_dor_build_lid_matrices;
	r->ucast_build_fwd_tables = osm_ucast_dor_build_lfts;
	return 0;
}

int osm_ucast_ar_dor_setup(struct osm_routing_engine *r,
			   osm_opensm_t *osm)
{
	int rc;

	rc = osm_ucast_dor_setup(r, osm);

	if (rc)
		return rc;

	/* set suitable auto AR parameters */
	r->ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->shield_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_AR_LAG;

	return 0;
}
