/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_ucast_mgr_t.
 * This file implements the Unicast Manager object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_MGR_C
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_routing_chain.h>
#include <opensm/osm_port_group.h>
#include <opensm/osm_advanced_routing.h>
extern int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
	                                      void *cxt, char *pi, void *p_osm, boolean_t goto_node);

void osm_ucast_mgr_construct(IN osm_ucast_mgr_t * p_mgr)
{
	memset(p_mgr, 0, sizeof(*p_mgr));
}

void osm_ucast_mgr_destroy(IN osm_ucast_mgr_t * p_mgr)
{
	OSM_ASSERT(p_mgr);

	OSM_LOG_ENTER(p_mgr->p_log);

	if (p_mgr->cache_valid)
		osm_ucast_cache_invalidate(p_mgr);

	OSM_LOG_EXIT(p_mgr->p_log);
}

ib_api_status_t osm_ucast_mgr_init(IN osm_ucast_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	osm_ucast_mgr_construct(p_mgr);

	p_mgr->sm = sm;
	p_mgr->p_log = sm->p_log;
	p_mgr->p_subn = sm->p_subn;
	p_mgr->p_lock = sm->p_lock;
	p_mgr->rebalancing_required = TRUE;

	if (sm->p_subn->opt.use_ucast_cache) {
		cl_qmap_init(&p_mgr->cache_sw_tbl);
		cl_qmap_init(&p_mgr->cache_vlid_tbl);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

int osm_ucast_mgr_pc_thread_pool_init(IN osm_ucast_mgr_t * p_mgr,
				      IN cl_pc_thread_pool_t * pc_thread_pool,
				      IN void (*p_fn) (void *, uint32_t),
				      IN const char * name,
				      OUT uint32_t * number_of_threads)
{
	cl_status_t status;
	int res = 0;
	int i;

	/* Initializing thread pool for multithreaded algorithm */
	status = cl_pc_thread_pool_init(
			pc_thread_pool,
			*number_of_threads,
			p_fn,
			p_mgr->p_subn->opt.max_threads_per_core,
			p_mgr,
			name);

	if (status != CL_SUCCESS) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A09: "
				"Thread pool failed to init."
				"Running single threaded routing\n");
				*number_of_threads = 1;
	} else {
		*number_of_threads = pc_thread_pool->p_thread_pool->running_count;

		p_mgr->sw_lists = (cl_qlist_t *)
				    calloc(1, sizeof(cl_qlist_t) * (*number_of_threads));

		if (p_mgr->sw_lists == NULL) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A12: "
				 "Not enough memory for sw_lists\n");
			res = -1;
			goto Exit;
		}
		for (i=0; i < *number_of_threads; i++) {
			cl_qlist_init(&p_mgr->sw_lists[i]);
		}

		osm_add_sw_to_list(&p_mgr->p_subn->sw_guid_tbl,
				    p_mgr->sw_lists,
				    *number_of_threads);
	}
Exit:
	return res;
}
/**********************************************************************
 Add each switch's own and neighbor LIDs to its LID matrix
**********************************************************************/
static void ucast_mgr_process_hop_0_1(IN cl_map_item_t * p_map_item,
				      IN void *context)
{
	osm_switch_t * p_sw = (osm_switch_t *) p_map_item;
	osm_node_t *p_remote_node;
	uint16_t lid, remote_lid;
	uint8_t i;

	lid = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
	osm_switch_set_hops(p_sw, lid, 0, 0);

	for (i = 1; i < p_sw->num_ports; i++) {
		osm_physp_t *p = osm_node_get_physp_ptr(p_sw->p_node, i);
		p_remote_node = (p && osm_link_is_healthy(p) && p->p_remote_physp) ?
		    p->p_remote_physp->p_node : NULL;

		if (p_remote_node && p_remote_node->sw &&
		    p_remote_node != p_sw->p_node) {
			remote_lid = osm_node_get_base_lid(p_remote_node, 0);
			remote_lid = cl_ntoh16(remote_lid);
			osm_switch_set_hops(p_sw, remote_lid, i, p->hop_wf);
		}
	}
}

static void ucast_mgr_process_neighbor(IN osm_ucast_mgr_t * p_mgr,
				       IN osm_switch_t * p_this_sw,
				       IN osm_switch_t * p_remote_sw,
				       IN uint8_t port_num,
				       IN uint8_t remote_port_num)
{
	osm_switch_t *p_sw;
	cl_map_item_t *item;
	uint16_t lid_ho;
	uint16_t hops;
	osm_physp_t *p;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Node 0x%" PRIx64 ", remote node 0x%" PRIx64
		", port %u, remote port %u\n",
		cl_ntoh64(osm_node_get_node_guid(p_this_sw->p_node)),
		cl_ntoh64(osm_node_get_node_guid(p_remote_sw->p_node)),
		port_num, remote_port_num);

	p = osm_node_get_physp_ptr(p_this_sw->p_node, port_num);

	for (item = cl_qmap_head(&p_mgr->p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_mgr->p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		lid_ho = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
		hops = osm_switch_get_least_hops(p_remote_sw, lid_ho);
		if (hops == OSM_NO_PATH)
			continue;
		hops += p->hop_wf;
		if (hops <
		    osm_switch_get_hop_count(p_this_sw, lid_ho, port_num)) {
			if (osm_switch_set_hops
			    (p_this_sw, lid_ho, port_num, (uint8_t)hops) != 0)
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A03: "
					"cannot set hops for lid %u at switch 0x%"
					PRIx64 "\n", lid_ho,
					cl_ntoh64(osm_node_get_node_guid
						  (p_this_sw->p_node)));
			p_mgr->some_hop_count_set = TRUE;
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

/* can be called only if p_physp was previously checked for NULL */
struct osm_remote_node *find_and_add_remote_sys(osm_physp_t * p_physp,
						boolean_t dor, struct
						osm_remote_guids_count *r)
{
	unsigned i;
	osm_node_t *node = p_physp->p_remote_physp->p_node;
	uint8_t rem_port = p_physp->p_remote_physp->port_num;

	for (i = 0; i < r->count; i++)
		if (r->guids[i].node == node)
			if (!dor || (r->guids[i].port == rem_port))
				return &r->guids[i];

	r->guids[i].node = node;
	r->guids[i].forwarded_to = 0;
	r->guids[i].port = rem_port;
	r->count++;
	return &r->guids[i];
}

static void ucast_mgr_process_port(IN osm_ucast_mgr_t * p_mgr,
				   IN osm_switch_t * p_sw,
				   IN osm_port_t * p_port,
				   IN unsigned lid_offset,
				   IN uint32_t id,
				   OUT uint8_t * outport,
				   OUT uint32_t * weight)
{
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;
	uint16_t lid_ho;
	uint8_t port;
	boolean_t is_ignored_by_port_prof;
	ib_net64_t node_guid;
	unsigned start_from = 1;
	osm_physp_t *p_physp = NULL;

	OSM_LOG_ENTER(p_mgr->p_log);

	*outport = OSM_NO_PATH;
	*weight = 0;

	osm_port_get_lid_range_ho(p_port, &min_lid_ho, &max_lid_ho);

	lid_ho = min_lid_ho + lid_offset;

	if (lid_ho > max_lid_ho)
		goto Exit;

	if (lid_offset && !p_mgr->is_dor)
		/* ignore potential overflow - it is handled in osm_switch.c */
		start_from =
		    osm_switch_get_port_by_lid(p_sw, lid_ho - 1, OSM_NEW_LFT) +
		    1;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing port 0x%" PRIx64
		" (\'%s\' port %u), LID %u [%u,%u]\n",
		cl_ntoh64(osm_port_get_guid(p_port)),
		p_port->p_node->print_desc, p_port->p_physp->port_num, lid_ho,
		min_lid_ho, max_lid_ho);

	/* TODO - This should be runtime error, not a OSM_ASSERT() */
	OSM_ASSERT(max_lid_ho <= IB_LID_UCAST_END_HO);

	node_guid = osm_node_get_node_guid(p_sw->p_node);

	/*
	   The lid matrix contains the number of hops to each
	   lid from each port.  From this information we determine
	   how best to distribute the LID range across the ports
	   that can reach those LIDs.
	 */
	port = osm_switch_recommend_path(p_sw, p_port, lid_ho, start_from,
					 p_mgr->p_subn->ignore_existing_lfts,
					 p_mgr->p_subn->opt.lmc,
					 p_mgr->is_dor,
					 p_mgr->p_subn->opt.port_shifting,
					 !lid_offset && p_port->use_scatter ?
					 &p_mgr->scatter_bufs[id] : NULL,
					 OSM_LFT,
					 p_mgr->p_subn->opt.port_profile_switch_nodes);
	*outport = port;

	if (port == OSM_NO_PATH) {
		/* do not try to overwrite the ppro of non existing port ... */
		is_ignored_by_port_prof = TRUE;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"No path to get to LID %u from switch 0x%" PRIx64 "\n",
			lid_ho, cl_ntoh64(node_guid));
	} else {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);

		if (!p_physp)
			goto Exit;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Routing LID %u to port %u for switch 0x%" PRIx64 "\n",
			lid_ho, port, cl_ntoh64(node_guid));

		/*
		   we would like to optionally ignore this port in equalization
		   as in the case of the Mellanox Anafa Internal PCI TCA port
		 */
		is_ignored_by_port_prof = p_physp->is_prof_ignored;

		/*
		   We also would ignore this route if the target lid is of
		   a switch and the port_profile_switch_node is not TRUE
		 */
		if (!p_mgr->p_subn->opt.port_profile_switch_nodes)
			is_ignored_by_port_prof |=
			    (osm_node_get_type(p_port->p_node) ==
			     IB_NODE_TYPE_SWITCH);
	}

	/*
	   We have selected the port for this LID.
	   Write it to the forwarding tables.
	 */

	p_sw->new_lft[lid_ho] = port;
	if (!is_ignored_by_port_prof) {
		struct osm_remote_node *rem_node_used;
		osm_switch_count_path(p_sw, port);
		*weight = 1;
		if (port > 0 && p_port->priv &&
		    (rem_node_used = find_and_add_remote_sys(p_physp,
							     p_mgr->is_dor,
							     p_port->priv)))
			rem_node_used->forwarded_to++;
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static int copy_port_list(osm_ucast_mgr_t * mgr, cl_qlist_t *port_list,
			  uint32_t number_of_threads)
{
	osm_port_t *port, *copy_port;
	cl_list_item_t *item;
	int i;
	int ret = 0;

        for (item = cl_qlist_head(port_list); item != cl_qlist_end(port_list);
                item = cl_qlist_next(item)) {
                port = cl_item_obj(item, port, list_item);
		for (i = 1; i < number_of_threads; i++) {
			copy_port = (osm_port_t *) malloc(sizeof(osm_port_t));
			if (copy_port == NULL) {
				ret = -1;
				goto Exit;
			}
			memcpy(copy_port, port, sizeof(osm_port_t));
			copy_port->priv = NULL;
			cl_qlist_insert_tail(&mgr->port_order_lists[i], &copy_port->list_item);
		}
        }
Exit:
	return ret;
}

int alloc_ports_priv(osm_ucast_mgr_t * mgr, cl_qlist_t *port_list)
{
	cl_list_item_t *item;
	struct osm_remote_guids_count *r;
	unsigned lmc;
	int status = 0;

	for (item = cl_qlist_head(port_list); item != cl_qlist_end(port_list);
			item = cl_qlist_next(item)) {
		osm_port_t *port = cl_item_obj(item, port, list_item);
		lmc = ib_port_info_get_lmc(&port->p_physp->port_info);
		r = malloc(sizeof(*r) + sizeof(r->guids[0]) * (1 << lmc));
		if (r == NULL) {
			/* need to reset to NULL non-allocated priv pointer */
			/* so don't return here */
			status = -1;
		}
		port->priv=r;
	}
	return status;
}

void init_ports_priv(osm_ucast_mgr_t * mgr, cl_qlist_t *port_list)
{
	cl_list_item_t *item;
	struct osm_remote_guids_count *r;
	unsigned lmc;

	for (item = cl_qlist_head(port_list); item != cl_qlist_end(port_list);
			item = cl_qlist_next(item)) {
		osm_port_t *port = cl_item_obj(item, port, list_item);
		lmc = ib_port_info_get_lmc(&port->p_physp->port_info);
		r = (struct osm_remote_guids_count *) port->priv;
		memset(r, 0, sizeof(*r) + sizeof(r->guids[0]) * (1 << lmc));
	}
}

void free_ports_priv(osm_ucast_mgr_t * mgr, uint32_t i)
{
	cl_qlist_t *port_list = &mgr->port_order_lists[i];
	cl_list_item_t *item;

	for (item = cl_qlist_head(port_list); item != cl_qlist_end(port_list);
	     item = cl_qlist_next(item)) {
		osm_port_t *port = cl_item_obj(item, port, list_item);
		free(port->priv);
		port->priv = NULL;
	}
}

static void handle_base_lid_offset(IN osm_ucast_mgr_t *p_mgr,
				   IN uint32_t id,
				   IN osm_switch_t * p_sw,
				   IN unsigned min_lid_offset,
				   IN unsigned max_lid_offset)
{
	int i;
	cl_qlist_t *list = &p_mgr->port_order_lists[id];
	cl_list_item_t *item;
	uint8_t outport = OSM_NO_PATH;
	uint32_t weight = 0;
	uint32_t port_arr[256];

	for (item = cl_qlist_head(list); item != cl_qlist_end(list);) {
		osm_port_t *port = cl_item_obj(item, port, list_item);
		/* calculate outport for base_lid separately */
		if (min_lid_offset == 0 || min_lid_offset == max_lid_offset) {
			ucast_mgr_process_port(p_mgr, p_sw, port,
					       min_lid_offset, id,
					       &outport, &weight);

			if ((outport == OSM_NO_PATH) ||
			    (min_lid_offset == max_lid_offset))
				goto go_next;
		}

		memset(&port_arr[0], 0, p_sw->num_ports * sizeof(uint32_t));

		/* calculate the outport for rest of lids */
                for (i = min_lid_offset + 1; i <= max_lid_offset; i++) {
			ucast_mgr_process_port(p_mgr, p_sw, port, i, id,
					       &outport, &weight);
			/* in case of switch ucast_mgr_process_port may return */
			/* NO_PATH if LMC is not set on ESP0 */
			if (port->p_node->sw && outport == OSM_NO_PATH)
				goto go_next;

			OSM_ASSERT(outport != OSM_NO_PATH);
			port_arr[outport] += weight;
		}
		/* reduce the weigth of outports for LMC>0 */
		/* this will ensure balanced routing of the base_lid */
		for (i = 1; i < p_sw->num_ports; i++)
			osm_port_prof_path_count_dec(&p_sw->p_prof[i],
						     port_arr[i]);
go_next:
		item = cl_qlist_next(item);
		if (item != cl_qlist_end(list)){
			osm_port_t *p_next_port = cl_item_obj(item, port,
							      list_item);
			if (port->use_scatter == 0 && p_next_port->use_scatter)
			for (i = 0; i < p_sw->num_ports; i++)
				osm_port_prof_path_count_clean(&p_sw->p_prof[i]);
		}
	}
}

static void ucast_mgr_process_tbl_base(IN cl_map_item_t * p_map_item,
				       IN void *context, IN uint32_t id)
{
	osm_ucast_mgr_t *p_mgr = context;
	osm_switch_t * p_sw = (osm_switch_t *) p_map_item;
	unsigned i, lids_per_port;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_sw && p_sw->p_node);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing switch 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));

	/* Initialize LIDs in buffer to invalid port number. */
	memset(p_sw->new_lft, OSM_NO_PATH, p_sw->max_lid_ho + 1);

	init_ports_priv(p_mgr, &p_mgr->port_order_lists[id]);

	if (p_mgr->p_subn->path_bit != OSM_SUBNET_DEFAULT_PATH_BITS) {
		p_mgr->p_subn->opt.lmc = 0;
		handle_base_lid_offset(p_mgr, id, p_sw,
				       p_mgr->p_subn->path_bit,
				       p_mgr->p_subn->path_bit);
	} else {
		lids_per_port = 1 << p_mgr->p_subn->opt.lmc;
		if (p_mgr->p_subn->opt.improved_lmc_path_distribution)
			handle_base_lid_offset(p_mgr, id, p_sw,
					       0, lids_per_port - 1);
		else {
			for (i = 0; i < lids_per_port; i++) {
				handle_base_lid_offset(p_mgr, id, p_sw, i, i);
			}
		}

	}
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void ucast_mgr_process_tbl_par(IN void *context, uint32_t my_id)
{
	osm_ucast_mgr_t *p_mgr = context;
	cl_list_item_t * item;
	osm_switch_t * p_sw;
	cl_qlist_t * p_list;

	p_list = &p_mgr->sw_lists[my_id];

	while (cl_qlist_count(p_list)) {

		item = cl_qlist_remove_head(p_list);

		p_sw = cl_item_obj(item, p_sw, list_item);

		ucast_mgr_process_tbl_base(&p_sw->map_item, context, my_id);

	}
}

static void ucast_mgr_process_tbl(IN cl_map_item_t * p_map_item,
					IN void *context)
{
	ucast_mgr_process_tbl_base(p_map_item, context, 0);
}

static void ucast_mgr_process_neighbors(IN cl_map_item_t * p_map_item,
					IN void *context)
{
	osm_switch_t * p_sw = (osm_switch_t *) p_map_item;
	osm_ucast_mgr_t * p_mgr = context;
	osm_node_t *p_node;
	osm_node_t *p_remote_node;
	uint32_t port_num;
	uint8_t remote_port_num;
	uint32_t num_ports;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);
	OSM_ASSERT(osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing switch with GUID 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)));

	num_ports = osm_node_get_num_physp(p_node);

	/*
	   Start with port 1 to skip the switch's management port.
	 */
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_remote_node = osm_node_get_remote_node(p_node,
							 (uint8_t) port_num,
							 &remote_port_num);
		if (p_remote_node && p_remote_node->sw
		    && (p_remote_node != p_node)) {
			/* make sure the link is healthy. If it is not - don't
			   propagate through it. */
			p_physp = osm_node_get_physp_ptr(p_node, port_num);

			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			ucast_mgr_process_neighbor(p_mgr, p_sw,
						   p_remote_node->sw,
						   (uint8_t) port_num,
						   remote_port_num);
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static boolean_t ucast_mgr_lmx_process_switch(osm_ucast_mgr_t *p_mgr,
					 osm_switch_t * p_sw,
					 uint16_t * lids_arr)
{
	boolean_t some_hop_count_set = FALSE;
	osm_node_t *p_node;
	osm_node_t *p_remote_node;
	osm_switch_t *p_remote_sw;
	osm_physp_t *p_physp;
	uint8_t port_num;
	uint8_t num_ports;
	uint8_t hops;
	uint16_t lid;
	unsigned i;

	p_node = p_sw->p_node;
	num_ports = p_sw->num_ports;

	for (port_num = 1; port_num < num_ports; port_num++) {
		p_remote_node = osm_node_get_remote_node(p_node, port_num,
							 NULL);
		if (!p_remote_node || !p_remote_node->sw ||
		    p_remote_node == p_node)
			continue;

		p_remote_sw = p_remote_node->sw;

		/* make sure the link is healthy. If it is not - don't
		   propagate through it. */
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp || !osm_link_is_healthy(p_physp))
			continue;

		i = 0;
		for (lid = lids_arr[0]; lid != 0; lid = lids_arr[i++]) {

			hops = osm_switch_get_least_hops(p_remote_sw, lid);
			if (hops == OSM_NO_PATH)
				continue;
			hops += p_physp->hop_wf;
			if (hops >= osm_switch_get_hop_count(p_sw, lid,
							     port_num))
				continue;

			if (osm_switch_set_hops(p_sw, lid, port_num, hops)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A22: "
					"cannot set hops for lid %u at "
					"switch 0x%" PRIx64 "\n", lid,
					cl_ntoh64(osm_node_get_node_guid(p_node)));
			} else
				some_hop_count_set = TRUE;
		}
	}

	return some_hop_count_set;
}

static void ucast_mgr_build_lmx_par(IN void *context, uint32_t my_id)
{
	osm_ucast_mgr_t *p_mgr = context;
	cl_map_item_t *p_sw_item;
	osm_switch_t *p_sw;
	cl_list_item_t *p_list_item;
	cl_qlist_t *p_list = &p_mgr->sw_lists[my_id];
	uint16_t *lids_arr;
	unsigned i = 0;
	boolean_t some_hop_count_set = TRUE;

	/*
	 * In order to improve performance, we convert the list of the switches
	 * in to an array of the lids of the switches.
	 */
	lids_arr = (uint16_t *)malloc((cl_qlist_count(p_list) + 1) *
				      sizeof(uint16_t));
	if (!lids_arr) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A21: "
			"Failed to allocate array for lid matrix "
			"calculation.\n");
		goto Exit;
	}

	for (p_list_item = cl_qlist_head(p_list);
	     p_list_item != cl_qlist_end(p_list);
	     p_list_item = cl_qlist_next(p_list_item)) {
		p_sw = cl_item_obj(p_list_item, p_sw, list_item);
		lids_arr[i++] = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
	}
	lids_arr[i] = 0;

	/*
	 * Update LMX table of every switch in the fabric according to it's
	 * neighbor switches.
	 */
	while (some_hop_count_set) {
		some_hop_count_set = FALSE;
		for (p_sw_item = cl_qmap_head(&p_mgr->p_subn->sw_guid_tbl);
			p_sw_item != cl_qmap_end(&p_mgr->p_subn->sw_guid_tbl);
			p_sw_item = cl_qmap_next(p_sw_item)) {
			p_sw = (osm_switch_t *) p_sw_item;
			some_hop_count_set |=
				ucast_mgr_lmx_process_switch(p_mgr,
							     p_sw,
							     lids_arr);
		}
	}

Exit:
	free(lids_arr);
}


static int set_hop_wf(void *ctx, uint64_t guid, char *p)
{
	osm_ucast_mgr_t *m = ctx;
	osm_node_t *node = osm_get_node_by_guid(m->p_subn, cl_hton64(guid));
	osm_physp_t *physp;
	unsigned port, hop_wf;
	char *e;

	if (!node || !node->sw) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"switch with guid 0x%016" PRIx64 " is not found\n",
			guid);
		return 0;
	}

	if (!p || !*p || !(port = strtoul(p, &e, 0)) || (p == e) ||
	    port >= node->sw->num_ports) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"bad port specified for guid 0x%016" PRIx64 "\n", guid);
		return 0;
	}

	p = e + 1;

	if (!*p || !(hop_wf = strtoul(p, &e, 0)) || p == e || hop_wf >= 0x100) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"bad hop weight factor specified for guid 0x%016" PRIx64
			"port %u\n", guid, port);
		return 0;
	}

	physp = osm_node_get_physp_ptr(node, port);
	if (!physp)
		return 0;

	physp->hop_wf = hop_wf;

	return 0;
}

static void set_default_hop_wf(cl_map_item_t * p_map_item, void *ctx)
{
	osm_switch_t *sw = (osm_switch_t *) p_map_item;
	int i;

	for (i = 1; i < sw->num_ports; i++) {
		osm_physp_t *p = osm_node_get_physp_ptr(sw->p_node, i);
		if (p)
			p->hop_wf = 1;
	}
}

static int set_search_ordering_ports(void *ctx, uint64_t guid, char *p)
{
	osm_subn_t *p_subn = ctx;
	osm_node_t *node = osm_get_node_by_guid(p_subn, cl_hton64(guid));
	osm_switch_t *sw;
	uint8_t *search_ordering_ports = NULL;
	uint8_t port;
	unsigned int *ports = NULL;
	const int bpw = sizeof(*ports)*8;
	int words;
	int i = 1; /* port 0 maps to port 0 */

	if (!node || !(sw = node->sw)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"switch with guid 0x%016" PRIx64 " is not found\n",
			guid);
		return 0;
	}

	if (sw->search_ordering_ports) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"switch with guid 0x%016" PRIx64 " already listed\n",
			guid);
		return 0;
	}

	search_ordering_ports = malloc(sizeof(*search_ordering_ports)*sw->num_ports);
	if (!search_ordering_ports) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 3A07: cannot allocate memory for search_ordering_ports\n");
		return -1;
	}
	memset(search_ordering_ports, 0, sizeof(*search_ordering_ports)*sw->num_ports);

	/* the ports array is for record keeping of which ports have
	 * been seen */
	words = (sw->num_ports + bpw - 1)/bpw;
	ports = malloc(words*sizeof(*ports));
	if (!ports) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 3A08: cannot allocate memory for ports\n");
		free(search_ordering_ports);
		return -1;
	}
	memset(ports, 0, words*sizeof(*ports));

	while ((*p != '\0') && (*p != '#')) {
		char *e;

		port = strtoul(p, &e, 0);
		if ((p == e) || (port == 0) || (port >= sw->num_ports) ||
		    !osm_node_get_physp_ptr(node, port)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
				"bad port %d specified for guid 0x%016" PRIx64 "\n",
				port, guid);
			free(search_ordering_ports);
			free(ports);
			return 0;
		}

		if (ports[port/bpw] & (1u << (port%bpw))) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
				"port %d already specified for guid 0x%016" PRIx64 "\n",
				port, guid);
			free(search_ordering_ports);
			free(ports);
			return 0;
		}

		ports[port/bpw] |= (1u << (port%bpw));
		search_ordering_ports[i++] = port;

		p = e;
		while (isspace(*p)) {
			p++;
		}
	}

	if (i > 1) {
		for (port = 1; port < sw->num_ports; port++) {
			/* fill out the rest of the search_ordering_ports array
			 * in sequence using the remaining unspecified
			 * ports.
			 */
			if (!(ports[port/bpw] & (1u << (port%bpw)))) {
				search_ordering_ports[i++] = port;
			}
		}
		sw->search_ordering_ports = search_ordering_ports;
	} else {
		free(search_ordering_ports);
	}

	free(ports);
	return 0;
}

int osm_ucast_mgr_build_lid_matrices(IN osm_ucast_mgr_t * p_mgr)
{
	uint32_t i;
	uint32_t iteration_max;
	cl_qmap_t *p_sw_guid_tbl;
	uint32_t number_of_threads = p_mgr->p_subn->opt.routing_threads_num;
	int res = 0;
	cl_pc_thread_pool_t pc_thread_pool;

	p_sw_guid_tbl = &p_mgr->p_subn->sw_guid_tbl;

	if (number_of_threads != 1) {

		res = osm_ucast_mgr_pc_thread_pool_init(p_mgr,
							&pc_thread_pool,
							&ucast_mgr_build_lmx_par,
							"LMX",
							&number_of_threads);
		if (res != 0)
			goto Exit;
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
		"Starting switches' Min Hop Table Assignment\n");
	/*
	   Set up the weighting factors for the routing.
	 */
	cl_qmap_apply_func(p_sw_guid_tbl, set_default_hop_wf, NULL);
	if (p_mgr->p_subn->opt.hop_weights_file) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Fetching hop weight factor file \'%s\'\n",
			p_mgr->p_subn->opt.hop_weights_file);
		if (parse_node_map(p_mgr->p_subn->opt.hop_weights_file,
				   set_hop_wf, p_mgr)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A05: "
				"cannot parse hop_weights_file \'%s\'\n",
				p_mgr->p_subn->opt.hop_weights_file);
		}
	}

	/*
	   Set the switch matrices for each switch's own port 0 LID(s)
	   then set the lid matrices for the each switch's leaf nodes.
	 */
	cl_qmap_apply_func(p_sw_guid_tbl, ucast_mgr_process_hop_0_1, p_mgr);

	/*
	   Get the switch matrices for each switch's neighbors.
	   This process requires a number of iterations equal to
	   the number of switches in the subnet minus 1.

	   In each iteration, a switch learns the lid/port/hop
	   information (as contained by a switch's lid matrix) from
	   its immediate neighbors.  After each iteration, a switch
	   (and it's neighbors) know more routing information than
	   it did on the previous iteration.
	   Thus, by repeatedly absorbing the routing information of
	   neighbor switches, every switch eventually learns how to
	   route all LIDs on the subnet.

	   Note that there may not be any switches in the subnet if
	   we are in simple p2p configuration.
	 */
	iteration_max = cl_qmap_count(p_sw_guid_tbl);

	/*
	   If there are switches in the subnet, iterate until the lid
	   matrix has been constructed.  Otherwise, just immediately
	   indicate we're done if no switches exist.
	 */
	if (!iteration_max)
		goto Exit;

	/*
	   we need to find out when the propagation of
	   hop counts has relaxed. So this global variable
	   is preset to 0 on each iteration and if
	   if non of the switches was set will exit the
	   while loop
	 */
	if (number_of_threads == 1) {
		iteration_max--;
		p_mgr->some_hop_count_set = TRUE;
		for (i = 0; (i < iteration_max) && p_mgr->some_hop_count_set;
		     i++) {
			p_mgr->some_hop_count_set = FALSE;
			cl_qmap_apply_func(p_sw_guid_tbl,
					   ucast_mgr_process_neighbors, p_mgr);
		}
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Min-hop propagated in %d steps\n", i);
	} else {
		cl_pc_thread_pool_start(&pc_thread_pool);
		cl_pc_thread_pool_wait(&pc_thread_pool, EVENT_NO_TIMEOUT);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Min-hop propagated using parallel computation\n");
	}

Exit:
	if (number_of_threads != 1) {
		for (i=0; i < number_of_threads; i++)
			cl_qlist_remove_all(&p_mgr->sw_lists[i]);

		cl_pc_thread_pool_destroy(&pc_thread_pool);
		free(p_mgr->sw_lists);
		p_mgr->sw_lists = NULL;
	}
	return res;
}

static void ucast_mgr_reset_switch_lfts(IN cl_map_item_t *p_map_item,
					IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	memset(p_sw->new_lft, OSM_NO_PATH, p_sw->lft_size);
}

void osm_ucast_mgr_reset_lfts_all_switches(osm_subn_t *p_subn)
{
	cl_qmap_apply_func(&p_subn->sw_guid_tbl,
			   ucast_mgr_reset_switch_lfts, NULL);
}

int ucast_mgr_setup_all_switches(osm_subn_t * p_subn)
{
	osm_switch_t *p_sw;
	uint16_t lids;

	lids = p_subn->max_lid;

	for (p_sw = (osm_switch_t *) cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
		if (osm_switch_prepare_path_rebuild(p_sw, lids, TRUE)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 3A0B: "
				"cannot setup switch 0x%016" PRIx64 "\n",
				cl_ntoh64(osm_node_get_node_guid
					  (p_sw->p_node)));
			return -1;
		}
		if (p_sw->search_ordering_ports) {
			free(p_sw->search_ordering_ports);
			p_sw->search_ordering_ports = NULL;
		}
	}

	if (p_subn->opt.port_search_ordering_file) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Fetching dimension ports file \'%s\'\n",
			p_subn->opt.port_search_ordering_file);
		if (parse_node_map_str(p_subn->opt.port_search_ordering_file,
				       osm_apply_func_on_port_group_nodes,
				       set_search_ordering_ports, p_subn, p_subn->p_osm, TRUE)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 3A0F: "
				"cannot parse port_search_ordering_file \'%s\'\n",
				p_subn->opt.port_search_ordering_file);
		}
	}

	return 0;
}

static boolean_t osm_ucast_port_is_zero_lid(IN osm_ucast_mgr_t *p_mgr,
				       IN osm_port_t *p_port)
{
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;

	osm_port_get_lid_range_ho(p_port, &min_lid_ho, &max_lid_ho);

	/* If the lids are zero - then there was some problem with
	 * the initialization. Don't handle this port. */
	if (min_lid_ho == 0 || max_lid_ho == 0) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A04: "
			"Port 0x%"PRIx64" (%s port %d) has LID 0. An "
			"initialization error occurred. Ignoring port\n",
			cl_ntoh64(osm_port_get_guid(p_port)),
			p_port->p_node->print_desc,
			p_port->p_physp->port_num);
		return TRUE;
	}

	return FALSE;
}

static int add_guid_to_order_list(void *ctx, uint64_t guid, char *p)
{
	osm_ucast_mgr_t *m = ctx;
	osm_port_t *port = osm_get_port_by_guid(m->p_subn, cl_hton64(guid));

	if (!port) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"port guid not found: 0x%016" PRIx64 "\n", guid);
		return 0;
	}

	if (port->flag) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"port guid specified multiple times 0x%016" PRIx64 "\n",
			guid);
		return 0;
	}

	if (osm_ucast_port_is_zero_lid(m, port))
		return 0;

	cl_qlist_insert_tail(&m->port_order_lists[0], &port->list_item);
	port->flag = 1;
	port->use_scatter =  (m->p_subn->opt.guid_routing_order_no_scatter == TRUE) ? 0 : m->p_subn->opt.scatter_ports;

	return 0;
}

void osm_add_sw_to_list(cl_qmap_t * sw_guid_tbl, cl_qlist_t * sw_lists, uint32_t num_threads)
{
	cl_qmap_t *tbl;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	unsigned k = 0;

	tbl = sw_guid_tbl;
	for (item = cl_qmap_head(tbl); item != cl_qmap_end(tbl);
                     item = cl_qmap_next(item)) {
			p_sw = (osm_switch_t *) item;
			cl_qlist_insert_tail(&sw_lists[k], &p_sw->list_item);
			k = (k + 1) % num_threads;
	}
}

static void add_port_to_order_list(cl_map_item_t * p_map_item, void *ctx)
{
	osm_port_t *port = (osm_port_t *) p_map_item;
	osm_ucast_mgr_t *m = ctx;

	if (!port->flag) {
		if (osm_ucast_port_is_zero_lid(m, port))
			return;

		if (port->p_node->sw && !m->p_subn->opt.use_scatter_for_switch_lid)
			port->use_scatter = 0;
		else
			port->use_scatter = m->p_subn->opt.scatter_ports;

		cl_qlist_insert_tail(&m->port_order_lists[0], &port->list_item);
	} else
		port->flag = 0;
}

static int mark_ignored_port(void *ctx, uint64_t guid, char *p)
{
	osm_ucast_mgr_t *m = ctx;
	osm_node_t *node = osm_get_node_by_guid(m->p_subn, cl_hton64(guid));
	osm_physp_t *physp;
	unsigned port;

	if (!node || !node->sw) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"switch with guid 0x%016" PRIx64 " is not found\n",
			guid);
		return 0;
	}

	if (!p || !*p || !(port = strtoul(p, NULL, 0)) ||
	    port >= node->sw->num_ports) {
		OSM_LOG(m->p_log, OSM_LOG_DEBUG,
			"bad port specified for guid 0x%016" PRIx64 "\n", guid);
		return 0;
	}

	physp = osm_node_get_physp_ptr(node, port);
	if (!physp)
		return 0;

	physp->is_prof_ignored = 1;

	return 0;
}

static void clear_prof_ignore_flag(cl_map_item_t * p_map_item, void *ctx)
{
	osm_switch_t *sw = (osm_switch_t *) p_map_item;
	int i;

	for (i = 1; i < sw->num_ports; i++) {
		osm_physp_t *p = osm_node_get_physp_ptr(sw->p_node, i);
		if (p)
			p->is_prof_ignored = 0;
	}
}

static void add_sw_endports_to_order_list(osm_switch_t * sw,
					  osm_ucast_mgr_t * m)
{
	osm_port_t *port;
	osm_physp_t *p;
	int i;

	for (i = 1; i < sw->num_ports; i++) {
		p = osm_node_get_physp_ptr(sw->p_node, i);
		if (p && p->p_remote_physp && !p->p_remote_physp->p_node->sw) {
			port = osm_get_port_by_guid(m->p_subn,
						    p->p_remote_physp->
						    port_guid);
			if (!port || port->flag)
				continue;

			if (osm_ucast_port_is_zero_lid(m, port))
				continue;

			cl_qlist_insert_tail(&m->port_order_lists[0],
					     &port->list_item);
			port->flag = 1;
			port->use_scatter = m->p_subn->opt.scatter_ports;
		}
	}
}

static void sw_count_endport_links(osm_switch_t * sw)
{
	osm_physp_t *p;
	int i;

	sw->endport_links = 0;
	for (i = 1; i < sw->num_ports; i++) {
		p = osm_node_get_physp_ptr(sw->p_node, i);
		if (p && p->p_remote_physp && !p->p_remote_physp->p_node->sw)
			sw->endport_links++;
	}
}

static int compar_sw_load(const void *s1, const void *s2)
{
#define get_sw_endport_links(s) (*(osm_switch_t **)s)->endport_links
	return get_sw_endport_links(s2) - get_sw_endport_links(s1);
}

static void sort_ports_by_switch_load(osm_ucast_mgr_t * m)
{
	int i, num = cl_qmap_count(&m->p_subn->sw_guid_tbl);
	void **s = malloc(num * sizeof(*s));
	if (!s) {
		OSM_LOG(m->p_log, OSM_LOG_ERROR, "ERR 3A0C: "
			"No memory, skip by switch load sorting.\n");
		return;
	}
	s[0] = cl_qmap_head(&m->p_subn->sw_guid_tbl);
	for (i = 1; i < num; i++)
		s[i] = cl_qmap_next(s[i - 1]);

	for (i = 0; i < num; i++)
		sw_count_endport_links(s[i]);

	qsort(s, num, sizeof(*s), compar_sw_load);

	for (i = 0; i < num; i++)
		add_sw_endports_to_order_list(s[i], m);
	free(s);
}

int alloc_random_bufs(osm_ucast_mgr_t * p_mgr, int num_of_bufs)
{
	if (num_of_bufs <= 0)
		return 0;

	p_mgr->scatter_bufs = (struct random_data *)
			      malloc (sizeof(struct random_data) * num_of_bufs);

	if (p_mgr->scatter_bufs == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 3A15: Not enough memory for scatter_bufs\n");
		goto Exit;
	}

	memset(p_mgr->scatter_bufs, 0 , sizeof(struct random_data) * num_of_bufs);

	p_mgr->scatter_statebufs = (char *)
				   malloc (sizeof(char) * SCATTER_BUF_SIZE * num_of_bufs);

	if (p_mgr->scatter_statebufs == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 3A16: Not enough memory for scatter_statebufs\n");
		free(p_mgr->scatter_bufs);
		p_mgr->scatter_bufs = NULL;
		goto Exit;
	}

	memset(p_mgr->scatter_statebufs, 0 , sizeof(char) * SCATTER_BUF_SIZE * num_of_bufs);

	return 0;
Exit:
	return -1;
}

int osm_ucast_mgr_init_port_order_lists(IN osm_ucast_mgr_t *p_mgr,
					IN uint32_t number_of_threads)
{
	cl_qlist_t * port_lists;
	uint32_t i;

	OSM_ASSERT(!p_mgr->port_order_lists);

	p_mgr->port_order_lists = (cl_qlist_t *)malloc(sizeof(*port_lists) *
						       number_of_threads);
	if (p_mgr->port_order_lists == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A14: "
			"Not enough memory for port_order_lists\n");
		return 1;
	}

	for (i = 0; i < number_of_threads; i++) {
		cl_qlist_init(&p_mgr->port_order_lists[i]);
	}

	if (p_mgr->p_subn->opt.guid_routing_order_file) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Fetching guid routing order file \'%s\'\n",
			p_mgr->p_subn->opt.guid_routing_order_file);

		if (parse_node_map_str(p_mgr->p_subn->opt.guid_routing_order_file,
				       osm_apply_func_on_port_group_nodes,
				       add_guid_to_order_list, p_mgr, p_mgr->p_subn->p_osm, FALSE))
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A0D: "
				"cannot parse guid routing order file \'%s\'\n",
				p_mgr->p_subn->opt.guid_routing_order_file);
	}
	sort_ports_by_switch_load(p_mgr);

	cl_qmap_apply_func(&p_mgr->p_subn->port_guid_tbl,
			   add_port_to_order_list, p_mgr);

	if (number_of_threads > 1 &&
	    copy_port_list(p_mgr, &p_mgr->port_order_lists[0],
			   number_of_threads) != 0 ) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A13: "
			"Not enough memory for copy_port_list\n");
		return 1;
	}

	return 0;
}

void osm_ucast_mgr_release_port_order_lists(IN osm_ucast_mgr_t *p_mgr,
					    IN uint32_t number_of_threads)
{
	cl_qlist_t *port_list;
	cl_list_item_t *item;
	osm_port_t *port;
	uint32_t i;

	for (i = 0; i < number_of_threads; i++) {
		port_list = &p_mgr->port_order_lists[i];
		while (cl_qlist_count(port_list) != 0) {
			item = cl_qlist_remove_head(port_list);
			port = cl_item_obj(item, port, list_item);
			/*
			 * Should free all copy port objects. list[0] has
			 * original port objects
			 */
			if (i) {
				free(port->priv);
				port->priv = NULL;
				free(port);
			}
		}
	}

	free(p_mgr->port_order_lists);
	p_mgr->port_order_lists = NULL;
}

static int ucast_mgr_build_lfts(osm_ucast_mgr_t * p_mgr)
{
	int i, ret = -1;
	uint32_t number_of_threads = p_mgr->p_subn->opt.routing_threads_num;
	cl_pc_thread_pool_t pc_thread_pool;
	/* ucast_mgr_process_tbl overwrite lmc value so we need to recover it*/
	uint8_t subn_lmc = p_mgr->p_subn->opt.lmc;

	/* reset new lfts from previous routing */
	osm_ucast_mgr_reset_lfts_all_switches(p_mgr->p_subn);

	if (number_of_threads != 1) {
		ret = osm_ucast_mgr_pc_thread_pool_init(p_mgr,
							&pc_thread_pool,
							&ucast_mgr_process_tbl_par,
							"routing",
							&number_of_threads);
		if (ret != 0)
			goto Exit;
	}
	if (p_mgr->p_subn->opt.scatter_ports) {
		if (alloc_random_bufs(p_mgr, number_of_threads) < 0) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A1A: "
					"Out for memory for alloc_random_bufs\n");
				goto Exit;
		}
	}

	if (osm_ucast_mgr_init_port_order_lists(p_mgr, number_of_threads)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A1B: "
			"Failed to initialize port order lists\n");
		goto Exit;
	}

	if (p_mgr->p_subn->opt.port_prof_ignore_file) {
		cl_qmap_apply_func(&p_mgr->p_subn->sw_guid_tbl,
				   clear_prof_ignore_flag, NULL);
		if (parse_node_map(p_mgr->p_subn->opt.port_prof_ignore_file,
				   mark_ignored_port, p_mgr)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A0E: "
				"cannot parse port prof ignore file \'%s\'\n",
				p_mgr->p_subn->opt.port_prof_ignore_file);
		}
	}

	/* Single threaded algorithm */
	if (number_of_threads == 1) {
		if (p_mgr->p_subn->opt.scatter_ports) {
			ret = initstate_r(p_mgr->p_subn->opt.scatter_ports,
				           &p_mgr->scatter_statebufs[0],
				           SCATTER_BUF_SIZE,
				           &p_mgr->scatter_bufs[0]);

			if (ret < 0) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					 "ERR 3A1E: Failed init random function\n");
				goto Exit2;
			}
		}
		if (alloc_ports_priv(p_mgr,
				     &p_mgr->port_order_lists[0]) < 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A1C: "
				"Out for memory for alloc_port_priv\n");
			goto Exit2;
		}

		cl_qmap_apply_func(&p_mgr->p_subn->sw_guid_tbl,
				   ucast_mgr_process_tbl,
				   p_mgr);
		ret = 0;

		goto Exit2;

	}

	/* Multithreaded part */
	for (i=0; i < number_of_threads; i++) {
		if (p_mgr->p_subn->opt.scatter_ports) {
			int r_status;
			r_status = initstate_r(p_mgr->p_subn->opt.scatter_ports+i,
				    &p_mgr->scatter_statebufs[SCATTER_BUF_SIZE*i],
				    SCATTER_BUF_SIZE, &p_mgr->scatter_bufs[i]);
			if (r_status < 0) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A1F: "
					"Failed to init random function\n");
				goto Exit;
			}
		}
	}

	for (i=0; i < number_of_threads; i++) {
		if (alloc_ports_priv(p_mgr,
				     &p_mgr->port_order_lists[i]) < 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A1D: "
				"Out of memory for alloc_port_priv\n");
			goto Exit;
		}
	}

	cl_pc_thread_pool_start(&pc_thread_pool);
	/* wait for completion */
	cl_pc_thread_pool_wait(&pc_thread_pool, EVENT_NO_TIMEOUT);
	/* We are done sucessfully */
	ret = 0;

Exit:
	if (number_of_threads > 1)
		cl_pc_thread_pool_destroy(&pc_thread_pool);

	free(p_mgr->sw_lists);

	p_mgr->sw_lists = NULL;
Exit2:

	for (i = 0; i < number_of_threads; i++) {
		free_ports_priv(p_mgr,i);
	}

	osm_ucast_mgr_release_port_order_lists(p_mgr, number_of_threads);

	if (p_mgr->p_subn->opt.scatter_ports) {

		free(p_mgr->scatter_bufs);

		p_mgr->scatter_bufs = NULL;

		free(p_mgr->scatter_statebufs);

		p_mgr->scatter_statebufs = NULL;
	}
	p_mgr->p_subn->opt.lmc = subn_lmc;
	return ret;
}

static boolean_t ucast_mgr_validate_hub_switch(osm_switch_t * p_sw,
					       osm_subn_t * p_subn)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	uint16_t lid, min_lid, max_lid;

	for (p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (osm_port_is_isolated(p_port))
			continue;

		osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);

		if (p_subn->path_bit != OSM_SUBNET_DEFAULT_PATH_BITS) {
			min_lid += p_subn->path_bit;
			if (min_lid > max_lid)
				continue;
			max_lid = min_lid;
		}

		for (lid = min_lid; lid <= max_lid; lid++) {
			if (p_sw->new_lft[lid] == OSM_NO_PATH)
				return FALSE;
		}
	}

	return TRUE;
}

static void ucast_mgr_create_hub_switch_dfs_tree(osm_switch_t * p_sw,
						 cl_map_t * p_sw_map,
						 uint16_t hub_lid)
{
	uint8_t port_num;
	uint8_t num_ports = osm_node_get_num_physp(p_sw->p_node);
	osm_node_t * p_remote_node;
	osm_switch_t * p_remote_sw;
	uint8_t ds_port;

	cl_map_insert(p_sw_map, p_sw->p_node->node_info.node_guid, p_sw);
	if (p_sw->rank == 0 || p_sw->rank == OSM_SW_NO_RANK)
		return;

	for (port_num = 1; port_num < num_ports; port_num++) {
		p_remote_node =
		    osm_node_get_remote_node(p_sw->p_node, port_num, &ds_port);

		if (!p_remote_node ||
		    osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
			continue;

		/*
		 * Check that the remote swith has lower rank and that we
		 * havn't visited it yet.
		 */
		p_remote_sw = p_remote_node->sw;
		if (p_remote_sw->rank == OSM_SW_NO_RANK ||
		    p_remote_sw->rank >= p_sw->rank ||
		    cl_map_get(p_sw_map,
			       p_remote_node->node_info.node_guid))
			continue;

		p_remote_sw->new_lft[hub_lid] = ds_port;
		ucast_mgr_create_hub_switch_dfs_tree(p_remote_sw, p_sw_map,
						     hub_lid);
	}
}

static void ucast_mgr_create_hub_switch_tree(osm_switch_t * p_sw, uint16_t lid)
{

	cl_map_t visited_sw_map;

	cl_map_init(&visited_sw_map, 10);
	ucast_mgr_create_hub_switch_dfs_tree(p_sw, &visited_sw_map, lid);
	cl_map_remove_all(&visited_sw_map);
	cl_map_destroy(&visited_sw_map);
}

int osm_ucast_mgr_calculate_missing_routes(osm_ucast_mgr_t * p_mgr)
{
	cl_map_item_t *p_item;
	cl_map_item_t *p_port_item;
	osm_subn_t *p_subn;
	osm_switch_t *p_sw = NULL;
	osm_switch_t *p_hub_sw = NULL;
	osm_port_t *p_port;
	int ret = 0;
	uint16_t hub_lid, lid, min_lid, max_lid;
	uint8_t port_num;
	cl_list_iterator_t leaf_switch_iter;
	cl_list_t *p_leaf_switches_list;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_mgr);

	p_subn = p_mgr->p_subn;
	p_leaf_switches_list = &p_subn->leaf_switches_list;

	if (!cl_list_count(p_leaf_switches_list)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"There is no optional hub switch for routing "
			"missing paths\n");
		p_subn->hub_switch_guid = 0;
		ret = -1;
		goto Exit;
	}

	/*
	 * Up/Down based routing does not allow LID routing communication
	 * between switches that are located inside spine "switch systems".
	 * The reason is that there is no way to allow a LID route between
	 * them that doesn't break the Up/Down rule.
	 *
	 * In order to connect those spine switches, forbidden Down/Up
	 * turns that are needed are concentrated into a single subtree
	 * with a root and thus no credit-loops are added to the fabric.
	 *
	 * Subtree root (hub switch) is chosen among one of the fabric
	 * leaf switches that must be set/cleared by:
	 * - osm_subn_add_leaf()
	 * - osm_subn_clear_leafs()
	 */

	/* Try to use previous hub switch if possible */
	p_sw = osm_subn_get_hub_switch(p_subn);
	if (p_sw && ucast_mgr_validate_hub_switch(p_sw, p_subn)){
		p_hub_sw = p_sw;
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Using previous hub switch 0x%016" PRIx64 " as "
			"hub switch.\n",
			cl_ntoh64(osm_node_get_node_guid(p_hub_sw->p_node)));
	}

	/* Try to find optimal hub switch on the fabric */
	if (!p_hub_sw) {
		for (leaf_switch_iter = cl_list_head(p_leaf_switches_list);
		     leaf_switch_iter != cl_list_end(p_leaf_switches_list);
		     leaf_switch_iter = cl_list_next(leaf_switch_iter)) {

			p_sw = (osm_switch_t *)cl_list_obj(leaf_switch_iter);

			/* Isolated switch can't be used as a hub */
			if (osm_switch_is_isolated(p_sw, p_subn))
				continue;

			if (!ucast_mgr_validate_hub_switch(p_sw, p_subn))
				continue;

			OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
				"Using switch 0x%016" PRIx64 " as hub switch.\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			p_hub_sw = p_sw;
			break;
		}
	}

	if (!p_hub_sw) {
		p_hub_sw = p_sw;
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Failed to find optimal hub switch, "
			"using switch 0x%016" PRIx64 ".\n",
			cl_ntoh64(osm_node_get_node_guid(p_hub_sw->p_node)));
	}

	p_subn->hub_switch_guid = osm_node_get_node_guid(p_hub_sw->p_node);

	hub_lid = cl_ntoh16(osm_physp_get_base_lid(
                            osm_node_get_physp_ptr(p_hub_sw->p_node, 0)));

	/*
	 * Create an updside-down subtree from the hub switch to the roots.
	 * Every allowd down/up turn will be in this subtree.
	 */
	ucast_mgr_create_hub_switch_tree(p_hub_sw, hub_lid);

	/*
	 * Now go over all the switches in the fabric
	 * and route the missing LIDs to the selected
	 * hub switch.
	 * In short, this hub switch now poses a target
	 * for all those missing LIDs.
	 */
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		/* Avoid routing isolated/ignored switches */
		if (osm_switch_is_isolated(p_sw, p_subn))
			continue;

		/*
		 * If the a switch doesn't have path to the hub switch, then
		 * it can't use it to complete missing routes
		 */
		if (p_sw->new_lft[hub_lid] == OSM_NO_PATH) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Can not calculate missing paths for switch"
				"0x%" PRIx64 "\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node))
			);
			continue;
		}

		for (p_port_item = cl_qmap_head(&p_subn->port_guid_tbl);
		     p_port_item != cl_qmap_end(&p_subn->port_guid_tbl);
		     p_port_item = cl_qmap_next(p_port_item)) {
			p_port = (osm_port_t *)p_port_item;

			if (osm_port_is_isolated(p_port))
				continue;

			osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);

			if (p_subn->path_bit != OSM_SUBNET_DEFAULT_PATH_BITS) {
				min_lid += p_mgr->p_subn->path_bit;
				if (min_lid > max_lid)
					continue;
				max_lid = min_lid;
			}

			for (lid = min_lid; lid <= max_lid; lid++) {

				/*
				 * Switch already has a route to destination
				 * lid
				 */
				if (p_sw->new_lft[lid] != OSM_NO_PATH)
					continue;

				/*
				 * Hub switch doesn't have path to destination
				 * lid
				 */
				if (p_hub_sw->new_lft[lid] == OSM_NO_PATH)
					continue;

				/*
				 * the missing LID will be routed through the
				 * same port that routes to the selected leaf
				 * switch
				 */

				port_num = p_sw->new_lft[hub_lid];

				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Switch 0x%" PRIx64 " : setting path "
					"to LID %u through port %u\n",
					cl_ntoh64(osm_node_get_node_guid(
						   p_sw->p_node)),
					lid, port_num);

				p_sw->new_lft[lid] = port_num;
			}
		}
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

int osm_ucast_mgr_build_lft_tables(IN osm_ucast_mgr_t * p_mgr)
{
	return ucast_mgr_build_lfts(p_mgr);
}

static void ucast_mgr_set_fwd_top(IN cl_map_item_t * p_map_item,
				  IN void *cxt)
{
	osm_ucast_mgr_t *p_mgr = cxt;
	osm_switch_t * p_sw = (osm_switch_t *) p_map_item;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	osm_madw_context_t context;
	ib_api_status_t status;
	ib_switch_info_t si;
	boolean_t set_swinfo_require = FALSE;
	ib_net16_t lin_top;
	uint8_t life_state;

	OSM_ASSERT(p_mgr);

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_sw);

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);

	if (p_sw->max_lid_ho == 0) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Skipping switch 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		return;
	}

	if (p_mgr->max_lid < p_sw->max_lid_ho)
		p_mgr->max_lid = p_sw->max_lid_ho;

	p_physp = osm_node_get_physp_ptr(p_node, 0);

	OSM_ASSERT(p_physp);

	p_path = osm_physp_get_dr_path_ptr(p_physp);

	/*
	   Set the top of the unicast forwarding table.
	 */
	si = p_sw->switch_info;
	lin_top = cl_hton16(p_sw->max_lid_ho);
	if (lin_top != si.lin_top) {
		set_swinfo_require = TRUE;
		si.lin_top = lin_top;
		context.si_context.lft_top_change = TRUE;
	} else
		context.si_context.lft_top_change = FALSE;

	life_state = si.life_state;
	ib_switch_info_set_life_time(&si, p_mgr->p_subn->opt.packet_life_time);

	if (life_state != si.life_state)
		set_swinfo_require = TRUE;

	if (set_swinfo_require) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Setting switch FT top to LID %u\n", p_sw->max_lid_ho);

		context.si_context.light_sweep = FALSE;
		context.si_context.node_guid = osm_node_get_node_guid(p_node);
		context.si_context.set_method = TRUE;

		status = osm_req_set(p_mgr->sm, p_path, (uint8_t *) & si,
				     sizeof(si), IB_MAD_ATTR_SWITCH_INFO,
				     0, FALSE,
				     ib_port_info_get_m_key(&p_physp->port_info),
				     0, CL_DISP_MSGID_NONE, &context);

		if (status != IB_SUCCESS)
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A06: "
				"Sending SwitchInfo attribute failed (%s)\n",
				ib_get_err_str(status));
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static osm_madw_t *set_lft_block(IN osm_switch_t *p_sw,
				 IN osm_ucast_mgr_t *p_mgr,
				 IN uint16_t block_id_ho)
{
	osm_madw_context_t context;
	osm_dr_path_t *p_path;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;

	/*
	   Send linear forwarding table blocks to the switch
	   as long as the switch indicates it has blocks needing
	   configuration.
	 */
	if (!p_sw->new_lft) {
		/* any routing should provide the new_lft */
		OSM_ASSERT(p_mgr->p_subn->opt.use_ucast_cache &&
			  p_mgr->cache_valid && !p_sw->need_update);
		return NULL;
	}

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp)
		return NULL;

	p_path = osm_physp_get_dr_path_ptr(p_physp);

	context.lft_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
	context.lft_context.set_method = TRUE;

	/* Do not send LFT, block is AR configured */
	/* HACK: Need to update old LFTs for routing algorithms
	 * that are based on existing LFTs
	 */
	if (p_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED ||
	    p_sw->ar_configured == OSM_SW_AR_CONF_PLUGIN) {
		memcpy(p_sw->lft + block_id_ho * IB_SMP_DATA_SIZE,
		       p_sw->new_lft + block_id_ho * IB_SMP_DATA_SIZE,
		       IB_SMP_DATA_SIZE);
		return NULL;
	}

	if (p_sw->ar_configured != OSM_SW_AR_CONF_DISABLED &&
	    !p_sw->need_update &&
	    !p_mgr->p_subn->need_update &&
	    !memcmp(p_sw->new_lft + block_id_ho * IB_SMP_DATA_SIZE,
		    p_sw->lft + block_id_ho * IB_SMP_DATA_SIZE,
		    IB_SMP_DATA_SIZE))
		return NULL;

	/*
	 * Zero the stored LFT block, so in case the MAD will end up
	 * with error, we will resend it in the next sweep.
	 */
	memset(p_sw->lft + block_id_ho * IB_SMP_DATA_SIZE, 0,
	       IB_SMP_DATA_SIZE);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Writing FT block %u to switch 0x%" PRIx64 "\n", block_id_ho,
		cl_ntoh64(context.lft_context.node_guid));

	p_madw = osm_prepare_req_set(p_mgr->sm, p_path,
			     p_sw->new_lft + block_id_ho * IB_SMP_DATA_SIZE,
			     IB_SMP_DATA_SIZE, IB_MAD_ATTR_LIN_FWD_TBL,
			     cl_hton32(block_id_ho), FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);

	if (p_madw == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 3A10: "
			"Out of memory\n");
		return NULL;
	}

	return p_madw;
}

static void ucast_mgr_pipeline_fwd_tbl(osm_ucast_mgr_t * p_mgr)
{
	cl_qmap_t *tbl;
	cl_map_item_t *item;
	osm_madw_t *p_madw;
	osm_switch_t *p_sw;
	cl_qlist_t mad_list;

	cl_qlist_init(&mad_list);
	unsigned i, max_block = p_mgr->max_lid / IB_SMP_DATA_SIZE + 1;

	tbl = &p_mgr->p_subn->sw_guid_tbl;
	for (i = 0; i < max_block; i++)
		for (item = cl_qmap_head(tbl); item != cl_qmap_end(tbl);
		     item = cl_qmap_next(item)){
			p_sw = (osm_switch_t *)item;
			if (p_sw->max_lid_ho == 0)
				continue;

			if (lookup_held_back_switch(p_mgr->p_subn,
						    p_sw->p_node))
				continue;

			p_madw = set_lft_block(p_sw, p_mgr, i);
			if (p_madw != NULL){
				cl_qlist_insert_tail(&mad_list, &p_madw->list_item);
			}

			if (i == max_block - 1) {
				if (p_sw->ar_configured == OSM_SW_AR_CONF_DISABLED)
					p_sw->ar_configured = OSM_SW_AR_CONF_NONE;

				/*
				 * Invalidate ar_info and plft_info cache in case
				 * ar_configured is none but AR is enabled on the switch,
				 * since the current ar_info on the switch is
				 * going to change as a result of LFT set
				 */
				if (p_sw->ar_configured == OSM_SW_AR_CONF_NONE &&
				    (p_sw->ar_data.ar_info.enable_features &
				     IB_AR_INFO_FEATURE_MASK_AR_ENABLED))
					osm_ar_invalidate_ar_cache_params(p_sw);
                        }
		}
	osm_send_req_mad_batch(p_mgr->sm, &mad_list);
}

void osm_ucast_mgr_set_fwd_tables(osm_ucast_mgr_t * p_mgr)
{
	osm_opensm_report_event(p_mgr->p_subn->p_osm,
				OSM_EVENT_ID_START_LFT_UPDATE, NULL);

	p_mgr->max_lid = 0;

	cl_qmap_apply_func(&p_mgr->p_subn->sw_guid_tbl, ucast_mgr_set_fwd_top,
			   p_mgr);

	if (!p_mgr->p_subn->is_ar_plugin_active &&
	    osm_ar_set_subn_ar_tables(p_mgr->sm))
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 3A26: failed to set advanced routing configuration"
			" on the subnet\n");

	ucast_mgr_pipeline_fwd_tbl(p_mgr);
}

static void ucast_mgr_clear_switch_settings(IN cl_map_item_t * p_map_item,
					    IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	p_sw->rank = OSM_SW_NO_RANK;
	p_sw->coord = OSM_SW_NO_COORD;
}

static void update_subn_ar_conf_params(IN osm_log_t * p_log, OUT osm_subn_t *p_subn,
				       IN struct osm_routing_engine *routing_engine)
{
	if (p_subn->opt.ar_mode == OSM_ADV_ROUTING_MODE_AUTO)
		p_subn->ar_conf_params.ar_mode = routing_engine->ar_mode_auto;

	if (p_subn->opt.shield_mode == OSM_ADV_ROUTING_MODE_AUTO)
		p_subn->ar_conf_params.shield_mode = routing_engine->shield_mode_auto;
}

static int ucast_mgr_route(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	int ret = -1;
	osm_ucast_mgr_t * p_mgr = &osm->sm.ucast_mgr;
	event_data_t event_data;

	update_subn_ar_conf_params(&osm->log, p_mgr->p_subn, r);
	memset(&event_data, 0, sizeof(event_data_t));
	memcpy(&event_data.name, r->name, strlen(r->name));
	event_data.event_id = OSM_ACT_MGR_ROUTE_SUCCESS;

	OSM_LOG(&osm->log, OSM_LOG_VERBOSE,
		"building routing with \'%s\' routing algorithm...\n", r->name);

	osm_subn_clear_leafs(p_mgr->p_subn);
	cl_qmap_apply_func(&p_mgr->p_subn->sw_guid_tbl,
			   ucast_mgr_clear_switch_settings, NULL);

	if (!r->build_lid_matrices ||
	    (ret = r->build_lid_matrices(r->context)) > 0)
		ret = osm_ucast_mgr_build_lid_matrices(&osm->sm.ucast_mgr);

	if (ret < 0) {
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
			"%s: cannot build lid matrices.\n", r->name);
		event_data.event_id = OSM_ACT_MGR_ROUTE_FAIL;
		goto Exit;
	}

	if (!r->ucast_build_fwd_tables ||
	    (ret = r->ucast_build_fwd_tables(r->context)) > 0)
		ret = ucast_mgr_build_lfts(&osm->sm.ucast_mgr);

	if (ret < 0) {
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
			"%s: cannot build fwd tables.\n", r->name);
		event_data.event_id = OSM_ACT_MGR_ROUTE_FAIL;
		goto Exit;
	}

	p_mgr->p_subn->routing_engine_used = r;

	ret = 0;
Exit:
	osm_act_mgr_post_event(&osm->sm.activity_mgr, &event_data);
	return ret;
}

static int ucast_mgr_adv_route(osm_ucast_mgr_t *p_mgr,
			       osm_adv_routing_engine_t *adv_routing_engine)
{
	int result = 0;

	OSM_LOG_ENTER(p_mgr->p_log);

	result = adv_routing_engine->route(adv_routing_engine->context);
	if (result) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 3A23: advanced routing engine \'%s\' failed on"
			" build forwarding tables stage\n",
			adv_routing_engine->name);
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return result;
}

static void ucast_mgr_choose_advanced_routing_engine(IN osm_log_t * p_log, OUT osm_subn_t *p_subn)
{
	osm_adv_routing_type_enum last_adv_routing_type = OSM_ADV_ROUTING_TYPE_NONE;
	osm_adv_routing_type_enum adv_routing_type;

	if (!strcmp(p_subn->opt.adv_routing_engine_name, "auto"))
		adv_routing_type = get_adv_routing_auto_type(p_subn->routing_engine_used);
	else
		adv_routing_type = get_adv_routing_type_from_name(p_subn->opt.adv_routing_engine_name);
	/* update the ar configured parameter */
	p_subn->ar_conf_params.ar_algorithm = adv_routing_type;

	if (p_subn->adv_routing_engine)
		last_adv_routing_type = get_adv_routing_type_from_name(p_subn->adv_routing_engine->name);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"adv_routing_engine type is %d, last adv_routing_engine type is %d\n",
		adv_routing_type, last_adv_routing_type);

	if (adv_routing_type != last_adv_routing_type) {
		if (p_subn->adv_routing_engine)
			osm_destroy_advanced_routing_engine(p_subn);

		if (adv_routing_type == OSM_ADV_ROUTING_TYPE_NONE)
			return;

		osm_create_advanced_routing_engine(p_subn,
						   get_adv_routing_name_from_type(adv_routing_type));
	}
}

int osm_ucast_mgr_process(IN osm_ucast_mgr_t * p_mgr)
{
	osm_opensm_t *p_osm;
	struct osm_routing_engine *p_routing_eng;
	cl_qmap_t *p_sw_guid_tbl;
	int failed = 0;
	osm_subn_t *p_subn = p_mgr->p_subn;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_sw_guid_tbl = &p_subn->sw_guid_tbl;
	p_osm = p_subn->p_osm;
	p_routing_eng = p_subn->routing_engine_list;
	/*
	   If there are no switches in the subnet, we are done.
	 */
	if (cl_qmap_count(p_sw_guid_tbl) == 0 ||
	    ucast_mgr_setup_all_switches(p_subn) < 0)
		goto Exit;

	failed = -1;
	p_subn->routing_engine_used = NULL;

	/* set right ar configuration parameters according to given configuration */
	p_subn->ar_conf_params.ar_mode = p_subn->opt.ar_mode;
	p_subn->ar_conf_params.shield_mode = p_subn->opt.shield_mode;
	p_subn->ar_conf_params.ar_algorithm =
		get_adv_routing_type_from_name(p_subn->opt.adv_routing_engine_name);

	while (p_routing_eng) {
		/* update ar configuration parameters according to the routing engine */
		failed = ucast_mgr_route(p_routing_eng, p_osm);
		if (!failed)
			break;
		p_routing_eng = p_routing_eng->next;
	}

	if (!p_subn->routing_engine_used &&
	    p_subn->no_fallback_routing_engine != TRUE) {
		/* If configured routing algorithm failed, use default MinHop */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Using Default Engine\n");
		failed = ucast_mgr_route(p_subn->default_routing_engine, p_osm);
	}

	/* if for some reason different routing engine is used */
	/* cleanup last unused routing engine */
	p_routing_eng = p_subn->last_routing_engine;
	if (p_routing_eng && p_routing_eng != p_subn->default_routing_engine) {
		if (p_routing_eng->context &&
		    p_routing_eng->destroy &&
		    p_subn->routing_engine_used &&
		    p_subn->routing_engine_used->type !=
		    osm_routing_engine_type(p_routing_eng->name))
			p_routing_eng->destroy(p_routing_eng->context);
		free(p_routing_eng);
	}

	p_subn->last_routing_engine = NULL;

	ucast_mgr_choose_advanced_routing_engine(p_mgr->p_log, p_subn);

	if (!failed && p_subn->adv_routing_engine &&
	    !p_mgr->p_subn->is_ar_plugin_active) {
		/* apply advanced routing engine */
		if (ucast_mgr_adv_route(p_mgr, p_subn->adv_routing_engine)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 3A25: failed to configure advanced"
				" routing engine \'%s\'\n",
				p_subn->adv_routing_engine->name);
		}
	}

	if (p_subn->routing_engine_used) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
			"%s tables configured on all switches\n",
			osm_routing_engine_type_str(p_subn->
					routing_engine_used->type));
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"advanced configuration rch id %d: ar_mode %d, shield_mode %d, "
		"adv_routing_engine type %d,"
		"ar_sl_mask 0x%x, ar_transport_mask 0x%x adv_routing_engine %s, "
		"is_ar_plugin_active %d\n",
		p_subn->routing_chain_engine_id,
		p_subn->ar_conf_params.ar_mode, p_subn->ar_conf_params.shield_mode,
		p_subn->ar_conf_params.ar_algorithm,
		p_subn->opt.ar_sl_mask,
		p_subn->opt.ar_transport_mask,
		p_subn->adv_routing_engine ? p_subn->adv_routing_engine->name : "none",
		p_mgr->p_subn->is_ar_plugin_active);

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return failed;
}

static int ucast_build_lid_matrices(void *context)
{
	return osm_ucast_mgr_build_lid_matrices(context);
}

static int ucast_build_lfts(void *context)
{
	return ucast_mgr_build_lfts(context);
}

int osm_ucast_minhop_setup(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	r->context = &osm->sm.ucast_mgr;
	r->build_lid_matrices = ucast_build_lid_matrices;
	r->ucast_build_fwd_tables = ucast_build_lfts;
	return 0;
}


int ucast_dummy_build_lid_matrices(void *context)
{
	return 0;
}

void osm_ucast_process_tbl(IN cl_map_item_t * const p_map_item,
			    IN void *context)
{
	ucast_mgr_process_tbl(p_map_item, context);
}

int osm_ucast_add_guid_to_order_list(void *ctx, uint64_t guid, char *p)
{
	return add_guid_to_order_list(ctx, guid, p);
}

void osm_ucast_clear_prof_ignore_flag(cl_map_item_t * const p_map_item, void *ctx)
{
	clear_prof_ignore_flag(p_map_item, ctx);
}

int osm_ucast_mark_ignored_port(void *ctx, uint64_t guid, char *p)
{
	return mark_ignored_port(ctx, guid, p);
}

void osm_ucast_add_port_to_order_list(cl_map_item_t * const p_map_item, void *ctx)
{
	add_port_to_order_list(p_map_item, ctx);
}
