/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of OpenSM PGFT/QFT Routing that is formula based
 *    but also supports cases where links and even switcehs are missing.
 *
 * Limitations:
 *    Some fault cases are not handled yet (when all paths through different
 *    switches (not connected directly) break routing.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <regex.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PQFT_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_switch.h>

/***************************************************************************
 *
 * Definitions
 *
 */

#define PQFT_ASSERT( log, exp, fmt , ... ) do {				\
    if (!PT(exp)) {							\
      osm_log(log, OSM_LOG_ERROR, "ASSERT @ %s %d %s :" fmt,		\
	      __FILE__, __LINE__, __FUNCTION__, ## __VA_ARGS__);	\
      printf("Assert at: %s %d. Aborting\n", __FILE__, __LINE__);	\
      exit(1);								\
    }									\
  } while (0)

#define PQFT_MIN_HEIGHT 2
#define PQFT_MAX_HEIGHT 4
#define PQFT_TUPLE_LEN	5
#define PQFT_TUPLE_BUF_LEN 256

typedef enum {
	PGFT_TYPE = 0,
	QFT_TYPE
} osm_pqft_type_t;

/* Forward references */
struct osm_pqft_;
struct osm_pqft_node_;
struct osm_pqft_port_;

/* the digits in the d are level, d1, d2 ... dh */
typedef union {
	uint64_t key;
	uint8_t d[PQFT_TUPLE_LEN];
} osm_pqft_tuple_t;

/*
 * Each connection between nodes is through port objects
 */
typedef struct osm_pqft_port_t_ {
	uint8_t port_num;	/* the matching phys_port num - 0 means unknown */
	uint8_t rem_neighbor_idx;	/* index of the neighbor on the remote pqft node */
	uint8_t rem_port_idx;	/* port index on the remote pqft node */
	boolean_t added;	/* set to TRUE if the port just added */
	boolean_t removed;	/* set to TRUE if the port just removed */
} osm_pqft_port_t;

/*
 * All ports and data about a specific neighbor
 */
typedef struct osm_pqft_neighbor_t_ {
	struct osm_pqft_node_t_ *rem_node;	/* remote pqft node */
} osm_pqft_neighbor_t;

/* track port masks */
typedef struct osm_pqft_port_mask_ {
	cl_map_item_t map_item;
	ib_net64_t mask[4];
	uint8_t max_mask;
} osm_pqft_port_mask_t;

/*
 * A ndoe on the FT
 */
typedef struct osm_pqft_node_t_ {
	cl_map_item_t tuple_map_item;	/* by tuple */
	cl_map_item_t pguid_map_item;	/* by tuple */
	osm_node_t *p_osm_node;	/* NULL if not matched */
	ib_net64_t port_guid;	/* PGUID of the matching node */
	uint16_t base_lid;	/* host order LID of this node */
	osm_pqft_tuple_t tuple;	/* note the last digit of the tuple is the level */
	osm_pqft_neighbor_t *down_neighbors;
	uint8_t num_down_neighbors;
	uint8_t next_down_neighbor_idx;
	osm_pqft_port_t *down_ports;
	uint8_t num_down_ports;
	osm_pqft_neighbor_t *up_neighbors;
	uint8_t num_up_neighbors;
	uint8_t next_up_neighbor_idx;
	osm_pqft_port_t *up_ports;
	uint8_t num_up_ports;
	uint16_t LID;		/* valid only during routing and only on leafs */
	osm_pqft_port_mask_t available_up_ports_mask;	/* to find missing ports */
	uint8_t qft_num_over_mapped;	/* Cnt for fix QFT level of P > 1 with
					   > 1 mapped osm nodes.  */
} osm_pqft_node_t;

/*
 * PQFT switch struct
 */
typedef struct osm_pqft_sw {
	cl_map_item_t map_item;
	osm_pqft_node_t *p_pqft_node;
	boolean_t is_matched;
} osm_pqft_sw_t;

/*
 * The complete subnet
 */
typedef struct osm_pqft_t_ {
	osm_pqft_type_t subn_type;
	uint8_t H;		/* height */
	uint8_t *M;		/* children by level - 1 */
	uint8_t *W;		/* parents by level - 1 */
	uint8_t *P;		/* parallelizm by level -1 - different meaning for PGFT/QFT */
	unsigned int *Q;	/* PI(Mi) for i <= l */
	unsigned int *R;	/* PI(WiPi) for i <= l */
	osm_opensm_t *p_osm;
	cl_qmap_t node_by_tuple;	/* by tuple */
	cl_qmap_t node_by_pguid;	/* by PGUID */
	cl_qmap_t missing_up_port_nodes_by_tuple;	/* item node's
							   available_up_ports_mask */
	unsigned int num_leafs;
	osm_pqft_node_t **leafs;
	boolean_t is_matched;
	boolean_t any_missing_links;
	boolean_t random_on_non_full_tree;
} osm_pqft_t;

static char *tuple2str(IN uint8_t h, IN uint8_t * t, IN char * buf)
{
	int i, j;
	/* level first */
	j = snprintf(buf, PQFT_TUPLE_BUF_LEN, "%d", t[0]);
	sprintf(buf,"%d", t[0]);
	for (i = h; i > 0; i--) {
		/* need a comma ? */
		if (i == h)
			j += snprintf(&buf[j], PQFT_TUPLE_BUF_LEN - j,
				      ":%d", t[i]);
		else
			j += snprintf(&buf[j], PQFT_TUPLE_BUF_LEN - j,
				      ",%d", t[i]);
	}
	return (buf);
}

static inline osm_subn_t *pqft_get_subnet(IN osm_pqft_t * p_pqft)
{
	return p_pqft->p_osm->sm.ucast_mgr.p_subn;
}

/***************************************************/

static uint8_t get_path_bit(IN osm_pqft_t * p_pqft)
{
	uint8_t subn_path_bit;

	subn_path_bit = pqft_get_subnet(p_pqft)->path_bit;
	if (subn_path_bit == OSM_SUBNET_DEFAULT_PATH_BITS)
		return 0;
	else
		return subn_path_bit;
}

/***************************************************/

static uint8_t get_sw_path_bit(IN osm_pqft_t * p_pqft)
{
	if (pqft_get_subnet(p_pqft)->opt.lmc_esp0)
		return get_path_bit(p_pqft);
	else
		return 0;
}

/***************************************************************************
 *
 * Buidling of the defined PGFT/PQFT
 *
 */

static char *get_regex_field(char *s, regmatch_t * p_matches, int i)
{
	s[p_matches[i].rm_eo] = '\0';
	return &s[p_matches[i].rm_so];
}

/* parse the textual definition and extract the parameters */
static int parse_pqft_option(osm_pqft_t * p_pqft,
			     osm_pqft_type_t * T, uint8_t * H,
			     uint8_t ** M, uint8_t ** W, uint8_t ** P)
{
	osm_opensm_t * p_osm = p_pqft->p_osm;
	osm_subn_t * p_subn;
	char *p_ch;
	char *p_type;
	regex_t pqft_exp;
	regmatch_t matches[6];
	char buf[128];
	unsigned int i;
	int ret = 1, res = 0;

	p_subn = pqft_get_subnet(p_pqft);

	res = regcomp(&pqft_exp,
		      "^(PGFT|QFT)[(]([0-9]+);([0-9,]+);([0-9,]+);([0-9,]+)[)]$",
		      REG_EXTENDED);

	if (res) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6408: "
			"Could not compile regexp for pqft_structure parsing");
		goto Exit;
	}

	if (NULL == p_subn->opt.pqft_structure) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6401: provided empty pqft structure string\n");
		goto Free_reg;
	}

	if (strlen(p_subn->opt.pqft_structure) > 127) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6402: provided pqft structure %s is too long!\n",
			p_osm->subn.opt.pqft_structure);
		goto Free_reg;
	}

	if (sizeof(buf) <= strlen(p_subn->opt.pqft_structure)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6418: Provided pqft structure |%s| is too long! "
			"(longer than 128 characters)\n", buf);
		goto Free_reg;
	}
	strncpy(buf, p_subn->opt.pqft_structure, sizeof(buf));
	if (regexec(&pqft_exp, buf, 6, matches, 0)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6403: provided pqft structure %s not a PGFT/QFT syntax!\n",
			buf);
		goto Free_reg;
	}

	p_type = get_regex_field(buf, matches, 1);
	if (!strcmp("PGFT", p_type)) {
		*T = PGFT_TYPE;
	} else if (!strcmp("QFT", buf)) {
		*T = QFT_TYPE;
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6404: provided pqft structure type: %s, is unknown type\n",
			p_type);
		goto Free_reg;
	}

	*H = atoi(get_regex_field(buf, matches, 2));
	if (*H < PQFT_MIN_HEIGHT || *H > PQFT_MAX_HEIGHT) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6405: provided pqft height: %d, is out of range [%d,%d]\n",
			*H, PQFT_MIN_HEIGHT, PQFT_MAX_HEIGHT);
		goto Free_reg;
	}

	*M = (uint8_t *) malloc(*H);
	if (! *M) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6406: out of memory while allocating M of size %d\n",
			(unsigned int)*H);
		goto Free_reg;
	}

	*W = (uint8_t *) malloc(*H);
	if (! *W) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 6409: out of memory while allocating W of size %d\n",
			(unsigned int)*H);
		goto Free_M;
	}

	*P = (uint8_t *) malloc(*H);
	if (! *P) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 640A: out of memory while allocating P of size %d\n",
			(unsigned int)*H);
		goto Free_W;
	}

	/* now it must have the height and 3 vectors */
	p_ch = strtok(get_regex_field(buf, matches, 3), ",");
	for (i = 0; i < *H; i++) {
		if (NULL == p_ch) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 6407: out of fields parsing M vector field %d\n",
				i);
			goto Free_P;
		}
		(*M)[i] = atoi(p_ch);
		p_ch = strtok(NULL, ",");
	}

	p_ch = strtok(get_regex_field(buf, matches, 4), ",");
	for (i = 0; i < *H; i++) {
		if (NULL == p_ch) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 6410: out of field parsing W vector field %d\n",
				i);
			goto Free_P;
		}
		(*W)[i] = atoi(p_ch);
		p_ch = strtok(NULL, ",");
	}

	p_ch = strtok(get_regex_field(buf, matches, 5), ",");
	for (i = 0; i < *H; i++) {
		if (NULL == p_ch) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 6411: out of field parsing P vector field %d\n",
				i);
			goto Free_P;
		}
		(*P)[i] = atoi(p_ch);
		p_ch = strtok(NULL, ",");
	}

	ret = 0;
	goto Free_reg;

Free_P:
	free(*P);
Free_W:
	free(*W);
Free_M:
	free(*M);
Free_reg:
	regfree(&pqft_exp);
Exit:
	return ret;
}

/* a debug feature = dump out all that is known about the tree */
static void dump_pqft(osm_pqft_t * p_pqft)
{
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	osm_pqft_node_t *p_pqft_node;
	char path[1024];
	FILE *p_pqft_dump_file;
	char *filename = "opensm-pqft.dump";
	uint8_t i;

	snprintf(path, sizeof(path), "%s/%s",
		 pqft_get_subnet(p_pqft)->opt.dump_files_dir, filename);
	p_pqft_dump_file = fopen(path, "w");
	if (!p_pqft_dump_file) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR, "ERR 6447: "
			"cannot open file \'%s\': %s\n", filename,
			strerror(errno));
		return;
	}

	/* for each node */
	for (p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
	     p_pqft_node !=
	     (osm_pqft_node_t *) cl_qmap_end(&p_pqft->node_by_tuple);
	     p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->tuple_map_item)) {

		fprintf(p_pqft_dump_file, "NODE: %s\n",
			tuple2str(p_pqft->H, p_pqft_node->tuple.d, tuple_buf));
		if (p_pqft_node->p_osm_node) {
			fprintf(p_pqft_dump_file,
				"\tMATCH: base_lid: 0x%04x desc: %s\n",
				p_pqft_node->base_lid,
				p_pqft_node->p_osm_node->print_desc);
		}
		fprintf(p_pqft_dump_file, "\tUP NEIGHBORS: %d\n",
			p_pqft_node->num_up_neighbors);
		for (i = 0; i < p_pqft_node->num_up_neighbors; i++) {
			if (p_pqft_node->up_neighbors[i].rem_node)
				fprintf(p_pqft_dump_file, "\t\t%d: %s %s\n", i,
					tuple2str(p_pqft->H,
						  p_pqft_node->up_neighbors[i].
						  rem_node->tuple.d, tuple_buf),
					(p_pqft_node->up_neighbors[i].rem_node->
					 p_osm_node ? p_pqft_node->
					 up_neighbors[i].rem_node->p_osm_node->
					 print_desc : ""));
			else
				fprintf(p_pqft_dump_file, "\t\t%d: UDEFINED?\n",
					i);
		}
		fprintf(p_pqft_dump_file, "\tUP PORTS: %d\n",
			p_pqft_node->num_up_ports);
		for (i = 0; i < p_pqft_node->num_up_ports; i++) {
			fprintf(p_pqft_dump_file,
				"\t\t%d: port_num=%d rem_neighbor_idx=%d rem_port_idx=%d\n",
				i, p_pqft_node->up_ports[i].port_num,
				p_pqft_node->up_ports[i].rem_neighbor_idx,
				p_pqft_node->up_ports[i].rem_port_idx);
		}

		fprintf(p_pqft_dump_file, "\tDOWN NEIGHBORS: %d\n",
			p_pqft_node->num_down_neighbors);
		for (i = 0; i < p_pqft_node->num_down_neighbors; i++) {
			if (p_pqft_node->down_neighbors[i].rem_node)
				fprintf(p_pqft_dump_file, "\t\t%d: %s %s\n", i,
					tuple2str(p_pqft->H,
						  p_pqft_node->
						  down_neighbors[i].rem_node->
						  tuple.d, tuple_buf),
					(p_pqft_node->down_neighbors[i].
					 rem_node->p_osm_node ? p_pqft_node->
					 down_neighbors[i].rem_node->
					 p_osm_node->print_desc : ""));
			else
				fprintf(p_pqft_dump_file, "\t\t%d: UDEFINED?\n",
					i);
		}
		fprintf(p_pqft_dump_file, "\tDOWN PORTS: %d\n",
			p_pqft_node->num_down_ports);
		for (i = 0; i < p_pqft_node->num_down_ports; i++) {
			fprintf(p_pqft_dump_file,
				"\t\t%d: port_num=%d rem_neighbor_idx=%d rem_port_idx=%d\n",
				i, p_pqft_node->down_ports[i].port_num,
				p_pqft_node->down_ports[i].rem_neighbor_idx,
				p_pqft_node->down_ports[i].rem_port_idx);
		}
		fprintf(p_pqft_dump_file, "\n");
	}

	/* done going through all the leaf switches */
	fclose(p_pqft_dump_file);
}

/*
 * Advance the number stored in D by 1 and wrap based on the Base
 * return TRUE if done
 */
static boolean_t get_next_tuple(uint8_t H, uint8_t * B, uint8_t * D)
{
	/* 000 001 010 011 020 021 100 101 110 111 120 121 */
	uint8_t p = 0;
	while (p <= H) {
		if (++D[p] >= B[p]) {
			D[p] = 0;
			p++;
		} else
			return FALSE;
	}
	return TRUE;
}

/* allocate and initialize a new node by its tuple */
static int init_pqft_node_by_tuple(osm_pqft_t * s, osm_pqft_tuple_t t)
{
	int status = 0;
	size_t asize;
	unsigned int num_leafs;
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	uint8_t l = t.d[0];
	osm_pqft_node_t *p_pqft_node =
	    (osm_pqft_node_t *) malloc(sizeof(osm_pqft_node_t));
	if (p_pqft_node == NULL) {
		OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
			"ERR 6412: out of memory while allocating new node\n");
		return 1;
	}
	memset(p_pqft_node, 0, sizeof(osm_pqft_node_t));
	p_pqft_node->tuple = t;

	/* create the arrays of neighbors and ports */
	if (l > 0) {

		p_pqft_node->num_down_ports = s->M[l - 1] * s->P[l - 1];
		if (s->subn_type == QFT_TYPE)
			p_pqft_node->num_down_neighbors =
			    p_pqft_node->num_down_ports;
		else
			p_pqft_node->num_down_neighbors = s->M[l - 1];

		asize =
		    p_pqft_node->num_down_neighbors *
		    sizeof(osm_pqft_neighbor_t);
		p_pqft_node->down_neighbors =
		    (osm_pqft_neighbor_t *) malloc(asize);
		if (p_pqft_node->down_neighbors == NULL) {
			OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
				"ERR 6413: out of memory while allocating down neighbors\n");
			goto FailDnNeigh;
		} else
			memset(p_pqft_node->down_neighbors, 0, asize);

		asize = p_pqft_node->num_down_ports * sizeof(osm_pqft_port_t);
		p_pqft_node->down_ports = (osm_pqft_port_t *) malloc(asize);
		if (p_pqft_node->down_ports == NULL) {
			OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
				"ERR 6414: out of memory while allocating down ports\n");
			goto FailDnPorts;
		} else
			memset(p_pqft_node->down_ports, 0, asize);
	}

	if (l < s->H) {
		p_pqft_node->num_up_ports = s->W[l] * s->P[l];
		if (s->subn_type == QFT_TYPE)
			p_pqft_node->num_up_neighbors = p_pqft_node->num_up_ports;
		else
			p_pqft_node->num_up_neighbors = s->W[l];

		asize =
		    p_pqft_node->num_up_neighbors * sizeof(osm_pqft_neighbor_t);
		p_pqft_node->up_neighbors =
		    (osm_pqft_neighbor_t *) malloc(asize);
		if (p_pqft_node->up_neighbors == NULL) {
			OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
				"ERR 6415: out of memory while allocating up neighbors\n");
			goto FailUpNeigh;
		} else
			memset(p_pqft_node->up_neighbors, 0, asize);

		asize = p_pqft_node->num_up_ports * sizeof(osm_pqft_port_t);
		p_pqft_node->up_ports = (osm_pqft_port_t *) malloc(asize);
		if (p_pqft_node->up_ports == NULL) {
			OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
				"ERR 6416: out of memory while allocating up ports\n");
			goto FailUpPorts;
		} else
			memset(p_pqft_node->up_ports, 0, asize);

		num_leafs = 1;
		for (l = 0; l < s->H; l++)
			num_leafs *= s->M[l];

		if (num_leafs > s->num_leafs) {
			s->leafs =
			    (osm_pqft_node_t **) realloc(s->leafs, num_leafs *
							 sizeof(osm_pqft_node_t *));
			if (s->leafs == NULL) {
				OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
					"ERR 6417: out of memory while allocating leaf array\n");
				goto FailLeafs;
			}
			memset(&s->leafs[s->num_leafs], 0,
			       (num_leafs - s->num_leafs) * sizeof(osm_pqft_node_t *));
			s->num_leafs = num_leafs;
		}
	}

	cl_qmap_insert(&s->node_by_tuple, t.key, &p_pqft_node->tuple_map_item);

	OSM_LOG(&s->p_osm->log, OSM_LOG_VERBOSE,
		"Init new PQFT node with tuple: %s\n",
		tuple2str(s->H, t.d, tuple_buf));
	goto Exit;

FailLeafs:
	free(p_pqft_node->up_ports);
FailUpPorts:
	free(p_pqft_node->up_neighbors);
FailUpNeigh:
	free(p_pqft_node->down_ports);
FailDnPorts:
	free(p_pqft_node->down_neighbors);
FailDnNeigh:
	free(p_pqft_node);
	status = 1;

Exit:
	return status;
}

/* given a tuple - find the node or return NULL */
static osm_pqft_node_t *find_node_by_tuple(osm_pqft_t * s, osm_pqft_tuple_t t)
{
	cl_map_item_t *item = cl_qmap_get(&s->node_by_tuple, t.key);
	if (item == cl_qmap_end(&s->node_by_tuple))
		return NULL;
	return ((osm_pqft_node_t *) item);
}

static osm_pqft_node_t *find_node_by_pguid(osm_pqft_t * s, ib_net64_t pguid)
{
	cl_map_item_t *item = cl_qmap_get(&s->node_by_pguid, pguid);
	if (item == cl_qmap_end(&s->node_by_pguid))
		return NULL;
	return (PARENT_STRUCT(item, osm_pqft_node_t, pguid_map_item));
}

/* connect the given node down ports by the FT type */
static int connect_node_downwards(osm_pqft_t * s, osm_pqft_node_t * p_pqft_node)
{
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	uint8_t l = p_pqft_node->tuple.d[0];
	uint8_t j, k, p, port_idx, rem_port_idx, rem_neigh_idx;
	osm_pqft_node_t *p_rem_node;
	osm_pqft_tuple_t rem_tuple;
	uint8_t Dl, Dlp1;

	/* no children to level 0 */
	if (l == 0)
		return 0;

	if (s->subn_type == PGFT_TYPE) {
		/*
		 * Connection rule for PGFT:
		 * l,A1,A2...Al..,Ah -- l-1,B1,B2...Bl...,Bh
		 *   For each j 0..Ml-1
		 *     Connect to: l-1,B1,B2...Bl..,Bh | Bl=j and the rest i Bi=Ai
		 *      Via ports Pl (separated up by Wl and down by Ml)
		 */
		Dl = p_pqft_node->tuple.d[l];
		rem_tuple = p_pqft_node->tuple;
		for (j = 0; j < s->M[l - 1]; j++) {
			rem_tuple.d[l] = j;
			rem_tuple.d[0] = l - 1;
			p_rem_node = find_node_by_tuple(s, rem_tuple);
			if (NULL == p_rem_node) {
				OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
					"ERR 6420: can't find down neighbor: %d of node: %s "
					"with tuple: %s\n",
					j, tuple2str(s->H, p_pqft_node->tuple.d,
						     tuple_buf), tuple2str(s->H,
									   rem_tuple.
									   d,
									   tuple_buf));
			} else {
				/* connect the neighbors */
				PQFT_ASSERT(&s->p_osm->log,
					    j < p_pqft_node->num_down_neighbors,
					    "j %d >= num_down_neighbors %d\n",
					    j, p_pqft_node->num_down_neighbors);

				p_pqft_node->down_neighbors[j].rem_node =
				    p_rem_node;

				PQFT_ASSERT(&s->p_osm->log,
					    (Dl < p_rem_node->num_up_neighbors),
					    "tuple=%s l=%d, Dl %d >= num_up_neighbors %d\n",
					    tuple2str(s->H,
						      p_pqft_node->tuple.d,
						      tuple_buf), l, Dl,
					    p_rem_node->num_up_neighbors);

				p_rem_node->up_neighbors[Dl].rem_node =
				    p_pqft_node;

				/* now obtain the actual port numbers */
				for (p = 0; p < s->P[l - 1]; p++) {
					port_idx = j + p * s->M[l - 1];
					rem_neigh_idx = p_pqft_node->tuple.d[l];
					rem_port_idx =
					    rem_neigh_idx + p * s->W[l - 1];

					PQFT_ASSERT(&s->p_osm->log,
						    (port_idx <
						     p_pqft_node->
						     num_down_ports),
						    "port_idx (%d) >= p_pqft_node->num_down_ports (%d)",
						    port_idx,
						    p_pqft_node->
						    num_down_ports);

					p_pqft_node->down_ports[port_idx].
					    rem_port_idx = rem_port_idx;
					p_pqft_node->down_ports[port_idx].
					    rem_neighbor_idx = rem_neigh_idx;

					PQFT_ASSERT(&s->p_osm->log,
						    (rem_port_idx <
						     p_rem_node->num_up_ports),
						    "rem_port_idx (%d) >= p_rem_node->num_up_ports (%d)",
						    rem_port_idx,
						    p_rem_node->num_up_ports);

					p_rem_node->up_ports[rem_port_idx].
					    rem_port_idx = port_idx;
					p_rem_node->up_ports[rem_port_idx].
					    rem_neighbor_idx = j;
				}
			}
		}
	} else if (s->subn_type == QFT_TYPE) {
		/*
		 * Connection rule for QFT:
		 * l,A1,A2...Al..,Ah -- l-1,B1,B2...Bl...,Bh
		 *   For each j 0..Ml -1
		 *    For each k 0..Pl -1
		 *    Connect to:
		 *     if l < H
		 *      l-1,B1,B2..Bl,Bl+1..,Bh | Bl=j, Bl+1=k+Pl*int(Al+1/Pl) and Ai=Bi for the rest
		 *     else l=H
		 *      l-1,B1,B2..Bh-1,Bh | Bh=j, Bh-1=k+Ph*int(Ah-1/Ph) and Ai=Bi for the rest
		 *
		 * NOTE that for QFT there is only one per per neighbor
		 */
		Dl = p_pqft_node->tuple.d[l];
		if (l == s->H) {
			Dlp1 = p_pqft_node->tuple.d[l - 1];
		} else {
			Dlp1 = p_pqft_node->tuple.d[l + 1];
		}
		rem_tuple = p_pqft_node->tuple;
		rem_tuple.d[0] = l - 1;
		for (j = 0; j < s->M[l - 1]; j++) {
			for (k = 0; k < s->P[l - 1]; k++) {
				rem_tuple.d[l] = j;
				if (l < s->H)
					rem_tuple.d[l + 1] =
					    (Dlp1 / s->P[l - 1]) * s->P[l - 1] +
					    k;
				else
					rem_tuple.d[l - 1] =
					    (Dlp1 / s->P[l - 1]) * s->P[l - 1] +
					    k;

				p_rem_node = find_node_by_tuple(s, rem_tuple);
				if (NULL == p_rem_node) {
					OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR,
						"ERR 6421: can't find down neighbor: %s of node: %s\n",
						tuple2str(s->H, rem_tuple.d,
							  tuple_buf1),
						tuple2str(s->H,
							  p_pqft_node->tuple.d,
							  tuple_buf));
				} else {
					port_idx =
					    rem_tuple.d[l] + k * s->M[l - 1];
					rem_port_idx =
					    Dl + (Dlp1 % s->P[l - 1]) * s->W[l -
									     1];

					/* connect the neighbors */
					PQFT_ASSERT(&s->p_osm->log,
						    (port_idx <
						     p_pqft_node->
						     num_down_neighbors),
						    "j %d, k %d,port_idx %d >= num_down_neighbors %d on %s\n",
						    j, k, port_idx,
						    p_pqft_node->
						    num_down_neighbors,
						    tuple2str(s->H,
							      p_pqft_node->
							      tuple.d,
							      tuple_buf));

					p_pqft_node->down_neighbors[port_idx].
					    rem_node = p_rem_node;

					PQFT_ASSERT(&s->p_osm->log,
						    (rem_port_idx <
						     p_rem_node->
						     num_up_neighbors),
						    "j %d, k %d, l %d, Dl %d, rem_port_idx %d >= "
						    "num_up_neighbors %d on rem node: %s,\n",
						    j, k, l, Dl, rem_port_idx,
						    p_rem_node->
						    num_up_neighbors,
						    tuple2str(s->H,
							      p_rem_node->tuple.
							      d, tuple_buf));

					p_rem_node->up_neighbors[rem_port_idx].
					    rem_node = p_pqft_node;

					/* now obtain the actual port numbers */
					PQFT_ASSERT(&s->p_osm->log,
						    (port_idx <
						     p_pqft_node->
						     num_down_ports),
						    "port_idx (%d) >= p_pqft_node->num_down_ports (%d)",
						    port_idx,
						    p_pqft_node->
						    num_down_ports);

					p_pqft_node->down_ports[port_idx].
					    rem_port_idx = rem_port_idx;
					p_pqft_node->down_ports[port_idx].
					    rem_neighbor_idx = rem_port_idx;

					PQFT_ASSERT(&s->p_osm->log,
						    (rem_port_idx <
						     p_rem_node->num_up_ports),
						    "rem_port_idx (%d) >= p_rem_node->num_up_ports (%d)",
						    rem_port_idx,
						    p_rem_node->num_up_ports);

					p_rem_node->up_ports[rem_port_idx].
					    rem_port_idx = port_idx;
					p_rem_node->up_ports[rem_port_idx].
					    rem_neighbor_idx = port_idx;
				}
			}
		}
	}
	return 0;
}

static int pqft_subn_init(osm_opensm_t * p_osm,
			  osm_pqft_t * s, osm_pqft_type_t T, uint8_t H,
			  uint8_t * M, uint8_t * W, uint8_t * P)
{
	unsigned int i, l;
	uint8_t D[PQFT_TUPLE_LEN], B[PQFT_TUPLE_LEN];
	boolean_t done;
	osm_pqft_tuple_t t;
	osm_pqft_node_t *p_pqft_node;

	t.key = 0;
	s->subn_type = T;
	s->H = H;
	s->M = M;
	s->W = W;
	s->P = P;

	s->Q = (unsigned int *)malloc(sizeof(unsigned int) * H);
	if (! s->Q) {
		OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR, "ERR 640B: "
			"Out of memory while allocating Q array\n");
		goto Err;
	}

	s->Q[0] = s->M[0];
	for (i = 1; i < H; i++) {
		s->Q[i] = s->Q[i - 1] * s->M[i];
	}

	s->R = (unsigned int *)malloc(sizeof(unsigned int) * H);
	if (! s->R) {
		OSM_LOG(&s->p_osm->log, OSM_LOG_ERROR, "ERR 640C: "
		"Out of memory while allocating R array\n");
		goto Free_Q;
	}

	if (s->subn_type == PGFT_TYPE) {
		s->R[0] = s->W[0];
		for (i = 1; i < H; i++) {
			s->R[i] = s->R[i - 1] * s->W[i];
		}
	} else {
		s->R[0] = s->W[0] * s->P[0];
		for (i = 1; i < H; i++) {
			s->R[i] = s->R[i - 1] * s->W[i] * s->P[i];
		}
	}

	cl_qmap_init(&s->node_by_tuple);
	cl_qmap_init(&s->node_by_pguid);
	cl_qmap_init(&s->missing_up_port_nodes_by_tuple);

	static char m_buf[256];
	static char w_buf[256];
	static char p_buf[256];
	OSM_LOG(&p_osm->log, OSM_LOG_INFO,
		"Allocating tree by M=%s W=%s P=%s\n", tuple2str(s->H - 1, s->M,
								 m_buf),
		tuple2str(s->H - 1, s->W, w_buf), tuple2str(s->H - 1, s->P,
							    p_buf));

	for (l = 0; l <= s->H; l++) {
		/* prepare the arrays of digits and bases */
		memset(D, 0, PQFT_MAX_HEIGHT);
		memset(B, 1, PQFT_MAX_HEIGHT);

		/*
		 * Number of nodes in level l is PI(i=1..l, Wi)*PI(i=l+1..h, Mi)
		 * So Bi = { i <= l : Wi ; i > l : Mi }
		 */
		for (i = 1; i <= s->H; i++) {
			if (i <= l)
				B[i - 1] = W[i - 1];
			else
				B[i - 1] = M[i - 1];
		}

		/* generate all nodes by their tuple */
		done = FALSE;
		while (!done) {
			t.key = 0UL;
			memcpy(&t.d[1], D, PQFT_MAX_HEIGHT);
			t.d[0] = l;

			if (init_pqft_node_by_tuple(s, t))
				goto Free_R;

			/* advance */
			done = get_next_tuple(s->H, B, D);
		}
	}			/* levels */

	/* connect the nodes by the tree rule */
	for (p_pqft_node = (osm_pqft_node_t *) cl_qmap_head(&s->node_by_tuple);
	     p_pqft_node != (osm_pqft_node_t *) cl_qmap_end(&s->node_by_tuple);
	     p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->tuple_map_item)) {
		if (connect_node_downwards(s, p_pqft_node))
			goto Free_R;
	}

	/* TODO: put into config file */
	s->random_on_non_full_tree = FALSE;

	return 0;

Free_R:
	free(s->R);
Free_Q:
	free(s->Q);
Err:
	return 1;
}

/***************************************************
 * Matching Discovered Subnet Nodes to FT
 ***************************************************/
static inline uint8_t ptr_to_rank(void *ptr)
{
	uint8_t res = 0;
	memcpy(&res, &ptr, 1);
	if (0 == res)
		return (255);
	else
		return (res - 1);
}

static inline void *rank_to_ptr(uint8_t rank)
{
	void *res = 0;
	rank++;
	memcpy(&res, &rank, 1);
	return (res);
}

static int rank_nodes_by_distance_from_cas(osm_pqft_t * p_pqft,
					   cl_qmap_t * p_node_by_nguid,
					   cl_map_t * p_rank_by_nguid,
					   uint8_t * p_max_rank,
					   osm_node_t ** pp_max_rank_node)
{
	cl_list_t bfs_q;
	int ret =0;
	osm_node_t *p_node, *p_rem_node;
	ib_net64_t nguid, rem_nguid;
	uint8_t num_phys_ports, port_num, rem_port_num, rank = 0, prev_rank;

	*pp_max_rank_node = NULL;
	*p_max_rank = 0;

	cl_list_init(&bfs_q, 100);

	/* place all CAs into the BFS Q */
	p_node = (osm_node_t *) cl_qmap_head(p_node_by_nguid);
	while (p_node != (osm_node_t *) cl_qmap_end(p_node_by_nguid)) {
		if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH) {
			nguid = osm_node_get_node_guid(p_node);
			if (cl_map_get(p_rank_by_nguid, nguid) == NULL) {
				cl_list_insert_tail(&bfs_q, p_node);
				cl_map_insert(p_rank_by_nguid, nguid,
					      rank_to_ptr(0));
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
					"Rank leaf node: 0x%016" PRIx64
					" with rank: %d\n", cl_ntoh64(nguid),
					0);
			}
		}
		p_node = (osm_node_t *) cl_qmap_next(&p_node->map_item);
	}

	while (!cl_is_list_empty(&bfs_q)) {
		p_node = (osm_node_t *) cl_list_remove_head(&bfs_q);
		nguid = osm_node_get_node_guid(p_node);
		rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, nguid));

		/* go over all connected nodes and assign rank if unknown */
		num_phys_ports = osm_node_get_num_physp(p_node);
		for (port_num = 1; port_num < num_phys_ports; port_num++) {
			p_rem_node =
			    osm_node_get_remote_node(p_node, port_num,
						     &rem_port_num);
			if (p_rem_node) {
				rem_nguid = osm_node_get_node_guid(p_rem_node);

				/* ignore nodes not included in the scope map */
				if (cl_qmap_get(p_node_by_nguid, rem_nguid) ==
				    cl_qmap_end(p_node_by_nguid))
					continue;

				if (cl_map_get(p_rank_by_nguid, rem_nguid) ==
				    NULL) {
					cl_list_insert_tail(&bfs_q, p_rem_node);
					cl_map_insert(p_rank_by_nguid,
						      rem_nguid,
						      rank_to_ptr(rank + 1));
					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_VERBOSE,
						"Rank node: 0x%016" PRIx64
						" with rank: %d\n",
						cl_ntoh64(rem_nguid), rank + 1);
					if (rank + 1 > *p_max_rank) {
						*p_max_rank = rank + 1;
						*pp_max_rank_node = p_rem_node;
					}
				} else {
					prev_rank =
					    ptr_to_rank(cl_map_get
							(p_rank_by_nguid,
							 rem_nguid));
					if ((prev_rank != rank - 1)
					    && (prev_rank != rank + 1)) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR,
							"ERR 6424: Node: GUID 0x%016"
							PRIx64
							" ranked: %d connected node: GUID 0x%016"
							PRIx64
							" ranked: %d which breaks FT requirement of layering. "
							"The topology is NOT a PGFT or QFT\n",
							cl_ntoh64(nguid), rank,
							cl_ntoh64(rem_nguid),
							prev_rank);
						ret = 1;
						goto Exit;
					}
				}
			}
		}
	}

	*p_max_rank = rank;

Exit:
	cl_list_remove_all(&bfs_q);
	cl_list_destroy(&bfs_q);
	return ret;
}

/* collect the spine and leaf nodes into the given lists */
static void collect_spine_nodes_and_leaf_ports(osm_pqft_t * p_pqft,
					       cl_qmap_t * p_node_by_nguid,
					       cl_map_t * p_rank_by_nguid,
					       uint8_t max_rank,
					       unsigned int *p_num_spines,
					       cl_list_t * p_spines,
					       unsigned int *p_num_leafs,
					       cl_list_t * p_leafs)
{
	osm_node_t *p_node;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	ib_net64_t nguid;
	uint8_t rank = 0;
	unsigned int leafs = 0, spines = 0;
	uint8_t num_phys_ports, port_num;

	p_node = (osm_node_t *) cl_qmap_head(p_node_by_nguid);
	while (p_node != (osm_node_t *) cl_qmap_end(p_node_by_nguid)) {
		nguid = osm_node_get_node_guid(p_node);
		rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, nguid));
		if (rank == 0) {
			/* see what ports it has that are conencted elsewhere */
			num_phys_ports = osm_node_get_num_physp(p_node);
			for (port_num = 1; port_num < num_phys_ports;
			     port_num++) {
				p_physp =
				    osm_node_get_physp_ptr(p_node, port_num);
				if ((NULL != p_physp)
				    && osm_physp_has_any_link(p_physp)) {
					p_port =
					    osm_get_port_by_guid(pqft_get_subnet(p_pqft),
								 osm_physp_get_port_guid
								 (p_physp));
					cl_list_insert_tail(p_leafs, p_port);
					leafs++;
				}
			}
		} else if (rank == max_rank) {
			cl_list_insert_tail(p_spines, p_node);
			spines++;
		}
		p_node = (osm_node_t *) cl_qmap_next(&p_node->map_item);
	}
	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"Collected %d leaf port and %d spine nodes\n", leafs, spines);
	if (p_num_leafs)
		*p_num_leafs = leafs;
	if (p_num_spines)
		*p_num_spines = spines;
}

static void set_matching_node(osm_pqft_t * p_pqft,
			      osm_pqft_node_t * p_pqft_node,
			      osm_node_t * p_osm_node,
			      ib_net64_t pguid, ib_net16_t base_lid)
{
	p_pqft_node->port_guid = pguid;
	cl_qmap_insert(&p_pqft->node_by_pguid, pguid,
		       &p_pqft_node->pguid_map_item);
	p_pqft_node->p_osm_node = p_osm_node;
	p_pqft_node->base_lid = cl_ntoh16(base_lid);
}

static int connect_nodes(osm_pqft_t * p_pqft,
		         osm_pqft_node_t * p_pqft_node1,
	                 uint8_t port_num1,
		         osm_pqft_node_t * p_pqft_node2, uint8_t port_num2)
{
	osm_pqft_node_t *p_pqft_top_node, *p_pqft_bot_node;
	uint8_t top_port_num, bot_port_num;
	uint8_t p, top_port_idx, bot_port_idx, top_neighbor_idx,
	    bot_neighbor_idx;
	uint8_t l, topDl, botDl, topDlp1 = 0, botDlp1 = 0;
	boolean_t top_port_free, bot_port_free;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	char tuple_buf[PQFT_TUPLE_BUF_LEN];

	/* find the upper node */
	if (p_pqft_node1->tuple.d[0] < p_pqft_node2->tuple.d[0]) {
		p_pqft_bot_node = p_pqft_node1;
		bot_port_num = port_num1;
		p_pqft_top_node = p_pqft_node2;
		top_port_num = port_num2;
	} else {
		p_pqft_bot_node = p_pqft_node2;
		bot_port_num = port_num2;
		p_pqft_top_node = p_pqft_node1;
		top_port_num = port_num1;
	}

	l = p_pqft_top_node->tuple.d[0];
	topDl = p_pqft_top_node->tuple.d[l];
	botDl = p_pqft_bot_node->tuple.d[l];

	if (p_pqft->subn_type == PGFT_TYPE) {
		/* for PGFT we may need to select the first available port by its p value */
		bot_neighbor_idx = topDl;
		top_neighbor_idx = botDl;

		for (p = 0; p < p_pqft->P[l - 1]; p++) {
			top_port_idx = top_neighbor_idx + p * p_pqft->M[l - 1];
			bot_port_idx = bot_neighbor_idx + p * p_pqft->W[l - 1];
			PQFT_ASSERT(&p_pqft->p_osm->log,
				    (top_port_idx <
				     p_pqft_top_node->num_down_ports),
				    "top_port_idx %d > num_down_ports %d on: %s\n",
				    top_port_idx,
				    p_pqft_top_node->num_down_ports,
				    p_pqft_top_node->p_osm_node->print_desc);

			PQFT_ASSERT(&p_pqft->p_osm->log,
				    (bot_port_idx <
				     p_pqft_bot_node->num_up_ports),
				    "bot_port_idx %d > num_up_ports %d on: %s\n",
				    bot_port_idx, p_pqft_bot_node->num_up_ports,
				    p_pqft_bot_node->p_osm_node->print_desc);

			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"Assigning node: 0x%016" PRIx64
				" down port idx: %d to port num: %d and "
				"node: 0x%016" PRIx64
				" up port idx: %d to port num: %d\n",
				cl_ntoh64(p_pqft_top_node->port_guid),
				top_port_idx, top_port_num,
				cl_ntoh64(p_pqft_bot_node->port_guid),
				bot_port_idx, bot_port_num);

			top_port_free =
			    (p_pqft_top_node->down_ports[top_port_idx].
			     port_num == 0);
			bot_port_free =
			    (p_pqft_bot_node->up_ports[bot_port_idx].port_num ==
			     0);

			if (top_port_free && bot_port_free) {
				p_pqft_top_node->down_ports[top_port_idx].
				    port_num = top_port_num;
				p_pqft_bot_node->up_ports[bot_port_idx].
				    port_num = bot_port_num;
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
					"Assigned node: 0x%016" PRIx64
					" down port idx: %d to port num: %d and "
					"node: 0x%016" PRIx64
					" up port idx: %d to port num: %d\n",
					cl_ntoh64(p_pqft_top_node->port_guid),
					top_port_idx, top_port_num,
					cl_ntoh64(p_pqft_bot_node->port_guid),
					bot_port_idx, bot_port_num);
				return 0;
			} else {
				PQFT_ASSERT(&p_pqft->p_osm->log,
					    !(top_port_free || bot_port_free),
					    "top_port_idx %d is already mapped to %d but"
					    " bot_port_idx %d is already mapped to %d\n",
					    top_port_idx,
					    p_pqft_top_node->
					    down_ports[top_port_idx].port_num,
					    bot_port_idx,
					    p_pqft_bot_node->
					    up_ports[bot_port_idx].port_num);
			}

			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"Skip previously allocated ports node: 0x%016"
				PRIx64 " port idx: %d and node: 0x%016" PRIx64
				" port idx: %d\n",
				cl_ntoh64(p_pqft_top_node->port_guid),
				top_port_idx,
				cl_ntoh64(p_pqft_bot_node->port_guid),
				bot_port_idx);
		}

		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6436: Could not find empty ports to connect %s down "
			"neighbor %d to %s up neighbor %d\n",
			tuple2str(p_pqft->H, p_pqft_top_node->tuple.d,
				  tuple_buf1), top_neighbor_idx,
			tuple2str(p_pqft->H, p_pqft_bot_node->tuple.d,
				  tuple_buf), bot_neighbor_idx);
		return 1;
	} else if (p_pqft->subn_type == QFT_TYPE) {
		/* for QFT there is only one connection between the switches */
		if (l < p_pqft->H) {
			topDlp1 = p_pqft_top_node->tuple.d[l + 1];
			botDlp1 = p_pqft_bot_node->tuple.d[l + 1];
		} else {
			topDlp1 = p_pqft_top_node->tuple.d[l - 1];
			botDlp1 = p_pqft_bot_node->tuple.d[l - 1];
		}

		top_port_idx =
		    botDl + p_pqft->M[l - 1] * (botDlp1 % p_pqft->P[l - 1]);
		bot_port_idx =
		    topDl + p_pqft->W[l - 1] * (topDlp1 % p_pqft->P[l - 1]);

		if (top_port_idx >= p_pqft_top_node->num_down_ports) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6429: "
				"top_port_idx %d >= p_pqft_top_node->num_down_ports %d\n",
				top_port_idx, p_pqft_top_node->num_down_ports);
			return 1;
		}

		if (bot_port_idx >= p_pqft_bot_node->num_up_ports) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6430: bot_port_idx %d >= p_pqft_bot_node->num_up_ports %d\n",
				bot_port_idx, p_pqft_bot_node->num_up_ports);
			return 1;
		}

		/* we double check the pointer are back to correct nodes */
		bot_neighbor_idx =
		    p_pqft_top_node->down_ports[top_port_idx].rem_neighbor_idx;
		top_neighbor_idx =
		    p_pqft_bot_node->up_ports[bot_port_idx].rem_neighbor_idx;

		if (bot_neighbor_idx >= p_pqft_bot_node->num_up_neighbors) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6431: "
				"bot_neighbor_idx %d >= p_pqft_bot_node->num_up_neighbors %d\n",
				bot_neighbor_idx,
				p_pqft_bot_node->num_up_neighbors);
			return 1;
		}

		if (p_pqft_bot_node->up_neighbors[bot_neighbor_idx].rem_node !=
		    p_pqft_top_node) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6432: "
				"bottom node neighbor does not point back to top node!\n");
			return 1;
		}

		if (top_neighbor_idx >= p_pqft_top_node->num_down_neighbors) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6433: "
				"top_neighbor_idx %d >= p_pqft_top_node->num_down_neighbors %d\n",
				top_neighbor_idx,
				p_pqft_top_node->num_down_neighbors);
			return 1;
		}

		if (p_pqft_top_node->down_neighbors[top_neighbor_idx].
		    rem_node != p_pqft_bot_node) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6434: "
				"top node neighbor does not point back to bottom node!\n");
			return 1;
		}

		p_pqft_top_node->down_ports[top_port_idx].port_num =
		    top_port_num;
		p_pqft_bot_node->up_ports[bot_port_idx].port_num = bot_port_num;
	} else
		return 1;

	return 0;
}

/*TODO: move to subnet/osm_node */
static ib_net64_t get_port_guid_by_port_num_hint(osm_node_t * p_node,
						 uint8_t port_num)
{
	osm_physp_t *p_physp;
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	} else {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
	}
	if (!p_physp)
		return 0;

	return osm_physp_get_port_guid(p_physp);
}

struct pqft_tuple_calc_item {
	cl_map_item_t map_item;
	osm_pqft_tuple_t tuple;
	boolean_t lsb_assigned;
	boolean_t msb_assigned;
};

static struct pqft_tuple_calc_item *find_tuple_calc_item_by_pguid(cl_qmap_t *
								  p_tuple_calc_by_pguid,
								  ib_net64_t
								  pguid)
{
	cl_map_item_t *p_item = cl_qmap_get(p_tuple_calc_by_pguid, pguid);
	if (p_item == cl_qmap_end(p_tuple_calc_by_pguid))
		return NULL;

	return (struct pqft_tuple_calc_item *)p_item;
}

/* First descend down allocating the MSB of the nodes
   When leafs are reached go back up through unassigned nodes */
static int dfs_calc_tuples_msb(osm_pqft_t * p_pqft,
			       osm_node_t * p_node,
			       uint8_t in_port_num,
			       cl_map_t * p_rank_by_nguid,
			       cl_qmap_t * p_tuple_calc_by_pguid,
			       unsigned int *p_num_leafs_set,
			       boolean_t leaf_reached)
{
	uint8_t l, port_num, rem_port_num, sub_group;
	uint8_t num_phys_ports = osm_node_get_num_physp(p_node);
	uint8_t rank, rem_rank;
	osm_node_t *p_rem_node;
	ib_net64_t nguid, rem_nguid, pguid, rem_pguid;
	struct pqft_tuple_calc_item *p_tuple_item, *p_rem_tuple_item;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	uint8_t child_idx = 0;
	uint8_t *qft_sub_groups_alloc_idx = NULL;
	boolean_t qft_allocate_in_groups = FALSE;
	uint8_t qft_limit_descendants = 0;
	int ret = 0;

	nguid = osm_node_get_node_guid(p_node);
	rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, nguid));

	pguid =
	    get_port_guid_by_port_num_hint(p_node, in_port_num);
	p_tuple_item =
	    find_tuple_calc_item_by_pguid(p_tuple_calc_by_pguid, pguid);

	if (!p_tuple_item) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6482: Could not find p_tuple_item by PGUID: "
			"0x%016" PRIx64	" for node: %s\n", cl_ntoh64(pguid),
			p_node->print_desc);
		ret = 1;
		goto Exit;
	}

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"Visiting node: %s in_port: %d on the way: %s\n",
		p_node->print_desc, in_port_num, leaf_reached ? "UP" : "DOWN");

	/* prepare the QFT special sub gruop allocated index counters */
	if ((!leaf_reached) && (p_pqft->subn_type == QFT_TYPE)) {
		if ((rank - 2 > 0) && (p_pqft->P[rank - 2] > 1)) {
			qft_sub_groups_alloc_idx =
			    (uint8_t *) malloc(p_pqft->M[rank - 1] *
					       sizeof(uint8_t));
			if (!qft_sub_groups_alloc_idx) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6483: Fail to allocate "
					"qft_sub_groups_alloc_idx %u bytes\n",
					(unsigned int)(p_pqft->M[rank - 1] * sizeof(uint8_t)));
				ret = 1;
				goto Exit;
			}

			memset(qft_sub_groups_alloc_idx, 0,
			       p_pqft->M[rank - 1] * sizeof(uint8_t));
			qft_allocate_in_groups = TRUE;
		} else if ((rank - 1 > 0) && (p_pqft->P[rank - 1] > 1))
			qft_limit_descendants = p_pqft->M[rank - 1];
	}

	/* double setting for children but need to handle parents */
	p_tuple_item->msb_assigned = TRUE;

	/* go over all connections of current node */
	for (port_num = 1; port_num < num_phys_ports; port_num++) {
		p_rem_node =
		    osm_node_get_remote_node(p_node, port_num, &rem_port_num);
		if (NULL == p_rem_node)
			continue;

		rem_pguid =
		    get_port_guid_by_port_num_hint(p_rem_node,
								 rem_port_num);
		p_rem_tuple_item =
		    find_tuple_calc_item_by_pguid(p_tuple_calc_by_pguid,
						  rem_pguid);

		if (!p_rem_tuple_item) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6484: Could not find rem_tuple_item by "
				"pguid: 0x%016"	PRIx64 " for node: %s\n",
				cl_ntoh64(rem_pguid), p_rem_node->print_desc);
			ret = 1;
			goto Free_sub_groups;
		}

		rem_nguid = osm_node_get_node_guid(p_rem_node);
		rem_rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, rem_nguid));

		/* going up only if already going up or change dir at level 1 */
		if (rem_rank == rank + 1) {
			if (p_rem_tuple_item->msb_assigned)
				continue;

			/* copy the current node digits to the remote tupple */
			for (l = rank + 2; l <= p_pqft->H; l++)
				p_rem_tuple_item->tuple.d[l] =
				    p_tuple_item->tuple.d[l];
			p_rem_tuple_item->tuple.d[0] = rem_rank;
			p_rem_tuple_item->msb_assigned = TRUE;

			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"set UP node: %s in_port: %d tuple: %d to %d to: %s parent:%s\n",
				p_rem_node->print_desc, rem_port_num, rank + 2,
				p_pqft->H, tuple2str(p_pqft->H,
						     p_rem_tuple_item->tuple.d,
						     tuple_buf1),
				p_node->print_desc);

			/* ascend only up to one level below the top but mark the top */
			if (rem_rank < p_pqft->H)
				dfs_calc_tuples_msb(p_pqft, p_rem_node,
						    rem_port_num,
						    p_rank_by_nguid,
						    p_tuple_calc_by_pguid,
						    p_num_leafs_set, TRUE);
		} else if ((rem_rank == rank - 1) && !leaf_reached) {
			if (p_rem_tuple_item->msb_assigned
			    && !qft_allocate_in_groups)
				continue;

			/* skip next children if in QFT special case and already
			   allocated M[l] children */
			if ((qft_limit_descendants != 0)
			    && (child_idx >= qft_limit_descendants))
				continue;

			/* copy the current node digits to the remote tupple */
			for (l = rank + 1; l <= p_pqft->H; l++)
				p_rem_tuple_item->tuple.d[l] =
				    p_tuple_item->tuple.d[l];

			if (qft_allocate_in_groups) {
				if (p_rem_tuple_item->msb_assigned) {
					/* need to re-assign the Dl ... */
					sub_group =
					    p_rem_tuple_item->tuple.d[rem_rank +
								      1] /
					    p_pqft->P[rank - 2];
					p_rem_tuple_item->tuple.d[rem_rank +
								  1] +=
					    ++qft_sub_groups_alloc_idx
					    [sub_group];

					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_VERBOSE,
						"Override: %s in_port: %d tuple digit: %d now: %s\n",
						p_rem_node->print_desc,
						rem_port_num, rank,
						tuple2str(p_pqft->H,
							  p_rem_tuple_item->
							  tuple.d, tuple_buf1));
				} else {
					p_rem_tuple_item->tuple.d[rem_rank +
								  1] =
					    child_idx;
					child_idx += p_pqft->P[rank - 2];
				}
			} else {
				p_rem_tuple_item->tuple.d[rem_rank + 1] =
				    child_idx;
				child_idx++;
			}

			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"set DN node: %s in_port: %d tuple: %d to %d to: %s parent: %s\n",
				p_rem_node->print_desc, rem_port_num, rank,
				p_pqft->H, tuple2str(p_pqft->H,
						     p_rem_tuple_item->tuple.d,
						     tuple_buf1),
				p_node->print_desc);

			p_rem_tuple_item->tuple.d[0] = rem_rank;
			p_rem_tuple_item->msb_assigned = TRUE;
			if (rank > 1) {
				dfs_calc_tuples_msb(p_pqft, p_rem_node,
						    rem_port_num,
						    p_rank_by_nguid,
						    p_tuple_calc_by_pguid,
						    p_num_leafs_set, FALSE);
			} else {
				if (qft_limit_descendants != 0)
					(*p_num_leafs_set) +=
					    qft_limit_descendants;
				else
					(*p_num_leafs_set)++;
			}
		}
	}

Free_sub_groups:
	if (qft_sub_groups_alloc_idx)
		free(qft_sub_groups_alloc_idx);
Exit:
	return ret;
}

/* First descend up allocating the LSB of the nodes
   When spines are reached go back down through unassigned nodes */
static int dfs_calc_tuples_lsb(osm_pqft_t * p_pqft,
			       osm_node_t * p_node,
			       uint8_t in_port_num,
			       cl_map_t * p_rank_by_nguid,
			       cl_qmap_t * p_tuple_calc_by_pguid,
			       unsigned int *p_num_spines_set,
			       boolean_t spine_reached)
{
	uint8_t l, port_num, rem_port_num;
	uint8_t num_phys_ports = osm_node_get_num_physp(p_node);
	uint8_t rank, rem_rank;
	osm_node_t *p_rem_node;
	ib_net64_t nguid, rem_nguid, pguid, rem_pguid;
	struct pqft_tuple_calc_item *p_tuple_item, *p_rem_tuple_item;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	uint8_t parent_idx = 0;

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"Visiting node: %s in_port: %d on the way: %s\n",
		p_node->print_desc, in_port_num, spine_reached ? "DOWN" : "UP");

	nguid = osm_node_get_node_guid(p_node);
	rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, nguid));

	pguid =
	    get_port_guid_by_port_num_hint(p_node, in_port_num);
	p_tuple_item =
	    find_tuple_calc_item_by_pguid(p_tuple_calc_by_pguid, pguid);

	PQFT_ASSERT(&p_pqft->p_osm->log, (p_tuple_item != NULL),
		    "Could not find p_tuple_item by PGUID: 0x%016" PRIx64
		    " for node: %s\n", cl_ntoh64(pguid), p_node->print_desc);

	p_tuple_item->lsb_assigned = TRUE;

	/* go over all connections of current node */
	for (port_num = 1; port_num < num_phys_ports; port_num++) {
		p_rem_node =
		    osm_node_get_remote_node(p_node, port_num, &rem_port_num);
		if (NULL == p_rem_node)
			continue;

		rem_pguid =
		    get_port_guid_by_port_num_hint(p_rem_node,
								 rem_port_num);
		p_rem_tuple_item =
		    find_tuple_calc_item_by_pguid(p_tuple_calc_by_pguid,
						  rem_pguid);

		PQFT_ASSERT(&p_pqft->p_osm->log, (p_rem_tuple_item != NULL),
			    "Could not find rem_tuple_item by PGUID: 0x%016"
			    PRIx64 " for node: %s\n", cl_ntoh64(rem_pguid),
			    p_rem_node->print_desc);

		if (p_rem_tuple_item->lsb_assigned)
			continue;

		rem_nguid = osm_node_get_node_guid(p_rem_node);
		rem_rank = ptr_to_rank(cl_map_get(p_rank_by_nguid, rem_nguid));

		if (rem_rank == rank - 1) {
			for (l = 1; l < rank; l++)
				p_rem_tuple_item->tuple.d[l] =
				    p_tuple_item->tuple.d[l];
			p_rem_tuple_item->tuple.d[0] = rem_rank;
			p_rem_tuple_item->lsb_assigned = TRUE;

			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"set DN node: %s in_port: %d tuple: %d to %d to: %s parent: %s\n",
				p_rem_node->print_desc, rem_port_num, 1,
				rank - 1, tuple2str(p_pqft->H,
						    p_rem_tuple_item->tuple.d,
						    tuple_buf1),
				p_node->print_desc);

			if (rank > 1)
				dfs_calc_tuples_lsb(p_pqft, p_rem_node,
						    rem_port_num,
						    p_rank_by_nguid,
						    p_tuple_calc_by_pguid,
						    p_num_spines_set, TRUE);
		} else if ((rem_rank == rank + 1) && !spine_reached) {
			for (l = 1; l <= rank; l++)
				p_rem_tuple_item->tuple.d[l] =
				    p_tuple_item->tuple.d[l];
			p_rem_tuple_item->tuple.d[rem_rank] = parent_idx++;
			p_rem_tuple_item->tuple.d[0] = rem_rank;
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"set UP node: %s in_port: %d tuple: %d to %d to: %s parent: %s\n",
				p_rem_node->print_desc, rem_port_num, 1,
				rank + 1, tuple2str(p_pqft->H,
						    p_rem_tuple_item->tuple.d,
						    tuple_buf1),
				p_node->print_desc);

			p_rem_tuple_item->lsb_assigned = TRUE;
			if (rem_rank == p_pqft->H) {
				(*p_num_spines_set)++;
			}
			dfs_calc_tuples_lsb(p_pqft, p_rem_node, rem_port_num,
					    p_rank_by_nguid,
					    p_tuple_calc_by_pguid,
					    p_num_spines_set, FALSE);
		}
	}
	return 0;
}

static int match_nodes_to_pqft(osm_pqft_t * p_pqft,
			       cl_qmap_t * p_nodes_by_nguid,
			       cl_qmap_t * p_ports_by_pguid)
{
	/* rank the nodes by min BFS from CAs */
	cl_map_t rank_by_nguid;
	osm_node_t *p_start_node, *p_node, *p_rem_node;
	osm_port_t *p_port;
	osm_pqft_node_t *p_pqft_node, *p_rem_pqft_node;
	ib_net64_t pguid, rem_pguid;
	ib_net16_t base_lid;
	uint8_t l, num_phys_ports, port_num, rem_port_num, max_rank = 0;
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	cl_list_t spine_nodes;
	cl_list_t leaf_ports;
	struct pqft_tuple_calc_item *tmp_tuples = NULL, *p_tuple_item;
	unsigned int tuple_idx, num_ports = cl_qmap_count(p_ports_by_pguid);
	unsigned int num_spines, num_spines_set = 0, num_leafs;
	unsigned int num_leafs_set = 0, num_trials;
	cl_qmap_t tuple_calc_by_pguid;
	cl_list_iterator_t list_iter;
	int status = 0;

	cl_map_construct(&rank_by_nguid);
	cl_map_init(&rank_by_nguid, 100);

	cl_qmap_init(&tuple_calc_by_pguid);

	cl_list_init(&spine_nodes, 100);
	cl_list_init(&leaf_ports, 100);

	p_pqft->is_matched = FALSE;

	if (rank_nodes_by_distance_from_cas(p_pqft,
					    p_nodes_by_nguid,
					    &rank_by_nguid,
					    &max_rank, &p_start_node)) {
		status = 1;
		goto Exit;
	}

	if (max_rank > p_pqft->H) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 649A: max rank is larger than"
			" structured parameter (hight), rank"
			" = %" PRIu8 " hight =%" PRIu8 "\n",
			max_rank, p_pqft->H);
		status = 1;
		goto Exit;
	}

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_INFO,
		"PQFT Ranked %d nodes, max rank=%d\n",
		(unsigned int)cl_map_count(&rank_by_nguid), max_rank);

	collect_spine_nodes_and_leaf_ports(p_pqft,
					   p_nodes_by_nguid,
					   &rank_by_nguid,
					   max_rank,
					   &num_spines,
					   &spine_nodes,
					   &num_leafs, &leaf_ports);

	/* initialize the tmp map of port guids to tuple calc entries */
	tmp_tuples = (struct pqft_tuple_calc_item *)
	    calloc(num_ports, sizeof(struct pqft_tuple_calc_item));
	if (!tmp_tuples) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
		"ERR 6485: Failed allocate tmp_tuples\n");
		status = 1;
		goto Exit;
	}
	tuple_idx = 0;
	for (p_port = (osm_port_t *) cl_qmap_head(p_ports_by_pguid);
	     p_port != (osm_port_t *) cl_qmap_end(p_ports_by_pguid);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		if (!osm_physp_get_port_num(p_port->p_physp) ||
		    osm_physp_get_remote(p_port->p_physp)) {
			pguid = osm_port_get_guid(p_port);
			cl_qmap_insert(&tuple_calc_by_pguid, pguid,
				       &(tmp_tuples[tuple_idx++].map_item));
		}
	}

	/* DFS setting MSBs */
	list_iter = cl_list_head(&spine_nodes);
	num_trials = 0;
	while (((num_leafs_set < num_leafs)) &&
	       (list_iter != cl_list_end(&spine_nodes)) && (num_trials < 20)) {
		num_leafs_set = 0;
		/* zero out the flag marking MSB is set */
		for (tuple_idx = 0; tuple_idx < num_ports; tuple_idx++)
			tmp_tuples[tuple_idx].msb_assigned = FALSE;
		p_node = (osm_node_t *) cl_list_obj(list_iter);
		if (dfs_calc_tuples_msb(p_pqft,
					    p_node,
					    0,
					    &rank_by_nguid,
					    &tuple_calc_by_pguid,
					    &num_leafs_set, FALSE)) {
			status = 1;
			goto Free_tmp_tuples;
		}
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
			"Starting MSB settings from: %s found %d leafs expected %d\n",
			p_node->print_desc, num_leafs_set, num_leafs);
		list_iter = cl_list_next(list_iter);
		num_trials++;
	}
	if (num_leafs_set < num_leafs) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6470: "
			"could not find a spine switch with access to all leafs\n");
		status = 1;
		goto Free_tmp_tuples;
	}

	/* DFS setting LSBs */
	list_iter = cl_list_head(&leaf_ports);
	num_trials = 0;
	while (((num_spines_set < num_spines)) &&
	       (list_iter != cl_list_end(&leaf_ports)) && (num_trials < 20)) {
		num_spines_set = 0;
		for (tuple_idx = 0; tuple_idx < num_ports; tuple_idx++)
			tmp_tuples[tuple_idx].lsb_assigned = FALSE;
		p_port = (osm_port_t *) cl_list_obj(list_iter);
		p_node =
		    osm_physp_get_node_ptr(osm_physp_get_remote
					   (p_port->p_physp));
		dfs_calc_tuples_lsb(p_pqft, p_node, 0, &rank_by_nguid,
				    &tuple_calc_by_pguid, &num_spines_set, 0);	/* starting up dir */
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
			"Starting LSB settings from: %s found %d spines expected %d\n",
			p_node->print_desc, num_spines_set, num_spines);
		list_iter = cl_list_next(list_iter);
		num_trials++;
	}

	if (num_spines_set < num_spines) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6471: "
			"could not find a leaf switch with access to all spines\n");
		status = 1;
		goto Free_tmp_tuples;
	}

	/* perform the actual assignment */
	for (p_port = (osm_port_t *) cl_qmap_head(p_ports_by_pguid);
	     p_port != (osm_port_t *) cl_qmap_end(p_ports_by_pguid);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		if (osm_physp_get_port_num(p_port->p_physp) &&
		    !osm_physp_get_remote(p_port->p_physp))
			continue;

		p_node = p_port->p_node;
		pguid = osm_port_get_guid(p_port);
		p_tuple_item =
		    find_tuple_calc_item_by_pguid(&tuple_calc_by_pguid, pguid);

		if (!p_tuple_item) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6486: Could not find PGUID: 0x%016" PRIx64
				" for node: %s\n", cl_ntoh64(pguid),
				p_node->print_desc);
			status = 1;
			goto Free_tmp_tuples;
		}

		if (p_tuple_item->msb_assigned && p_tuple_item->lsb_assigned) {
			p_pqft_node =
			    find_node_by_tuple(p_pqft, p_tuple_item->tuple);
			if (!p_pqft_node) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6472: could not get PQFT node by tuple: %s\n",
					tuple2str(p_pqft->H,
						  p_tuple_item->tuple.d,
						  tuple_buf));
				status = 1;
			} else {
				/* was that particular node already matched? */
				if (p_pqft_node->p_osm_node) {
					/* we may allow and fix these if QFT and node level has P[l] > 1 */
					l = p_pqft_node->tuple.d[0];
					if ((p_pqft->subn_type == QFT_TYPE)
					    && (p_pqft->P[l - 1] > 1)) {
						/* change the tuple and get a new one */
						p_tuple_item->tuple.d[l + 1] +=
						    ++(p_pqft_node->
						       qft_num_over_mapped);
						p_pqft_node =
						    find_node_by_tuple(p_pqft,
								       p_tuple_item->
								       tuple);
						if (!p_pqft_node) {
							OSM_LOG(&p_pqft->p_osm->
								log,
								OSM_LOG_ERROR,
								"ERR 6478: "
								"could not get QFT remapped node by tuple: %s\n",
								tuple2str
								(p_pqft->H,
								 p_tuple_item->
								 tuple.d,
								 tuple_buf));
							status = 1;
						} else {
							OSM_LOG(&p_pqft->p_osm->
								log,
								OSM_LOG_VERBOSE,
								"Moved mapped node: %s PGUID: 0x%016"
								PRIx64
								" to PQFT tuple: %s\n",
								p_node->
								print_desc,
								cl_ntoh64
								(pguid),
								tuple2str
								(p_pqft->H,
								 p_tuple_item->
								 tuple.d,
								 tuple_buf));
						}
					}
				}

				if (p_pqft_node && !p_pqft_node->p_osm_node) {
					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_VERBOSE,
						"Matched node: %s PGUID: 0x%016"
						PRIx64 " to PQFT tuple: %s\n",
						p_node->print_desc,
						cl_ntoh64(pguid),
						tuple2str(p_pqft->H,
							  p_tuple_item->tuple.d,
							  tuple_buf));

					/* get the pqft node by the allocated tuple */
					base_lid =
					    osm_port_get_base_lid(p_port);
					set_matching_node(p_pqft, p_pqft_node,
							  p_node, pguid,
							  base_lid);
				} else {
					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_ERROR,
						"ERR 6477: "
						"node: %s is mapped to a previously PQFT node by tuple: %s\n",
						p_node->print_desc,
						tuple2str(p_pqft->H,
							  p_tuple_item->tuple.d,
							  tuple_buf));
					status = 1;
				}
			}
		} else {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"No tuple was assigned to node: %s PGUID: 0x%016"
				PRIx64 " MSB=%s LSB=%s\n", p_node->print_desc,
				cl_ntoh64(pguid),
				p_tuple_item->msb_assigned ? "YES" : "NO",
				p_tuple_item->lsb_assigned ? "YES" : "NO");
			status = 1;
		}
	}

	if (status)
		goto Free_tmp_tuples;

	/* last stage in matching is to populate the port numbers of the PQFT */
	for (p_port = (osm_port_t *) cl_qmap_head(p_ports_by_pguid);
	     p_port != (osm_port_t *) cl_qmap_end(p_ports_by_pguid);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		if (osm_physp_get_port_num(p_port->p_physp) &&
		    !osm_physp_get_remote(p_port->p_physp))
			continue;
		p_node = p_port->p_node;

		pguid = osm_port_get_guid(p_port);
		p_pqft_node = find_node_by_pguid(p_pqft, pguid);

		if (!p_pqft_node) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6487: Should never get here if there are"
				" unmatched nodes\n");
			status = 1;
			goto Free_tmp_tuples;
		}

		if (!p_pqft_node->p_osm_node) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6488: PQFT node has no osm node for port "
				"node is: %s PGUID: 0x%016" PRIx64 " \n",
				p_node->print_desc, cl_ntoh64(pguid));
			status = 1;
			goto Free_tmp_tuples;
		}

		if (p_port->p_node != p_pqft_node->p_osm_node) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6489: PQFT pointed node is: %s while port "
				"node is: %s\n",
				p_pqft_node->p_osm_node->print_desc,
				p_node->print_desc);
			status = 1;
			goto Free_tmp_tuples;
		}

		/* skip hosts as they have the lowest rank */
		if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH)
			continue;

		num_phys_ports = osm_node_get_num_physp(p_node);
		for (port_num = 1; port_num < num_phys_ports; port_num++) {

			p_rem_node =
			    osm_node_get_remote_node(p_node, port_num,
						     &rem_port_num);
			if (!p_rem_node)
				continue;

			rem_pguid =
			    get_port_guid_by_port_num_hint
			    (p_rem_node, rem_port_num);
			p_rem_pqft_node = find_node_by_pguid(p_pqft, rem_pguid);

			if (!p_rem_pqft_node) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 648A: Could not find the remote "
					"node by PGUID: 0x%016"	PRIx64 "\n",
					rem_pguid);
				status = 1;
				goto Free_tmp_tuples;
			}

			if (p_rem_pqft_node->p_osm_node != p_rem_node) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 648B: rem node is: %s while port "
					"node is: %s\n",
					p_rem_pqft_node->p_osm_node->print_desc,
					p_rem_node->print_desc);
				status = 1;
				goto Free_tmp_tuples;
			}

			/* finally connect them - but only from up to down to avoid double */
			if (p_pqft_node->tuple.d[0] >
			    p_rem_pqft_node->tuple.d[0]) {
				if (connect_nodes
				    (p_pqft, p_pqft_node, port_num,
				     p_rem_pqft_node, rem_port_num)) {
					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_ERROR,
						"ERR 6474: could not connect node: 0x%016"
						PRIx64 " to node: 0x%016" PRIx64
						" \n", cl_ntoh64(pguid),
						cl_ntoh64(rem_pguid));
					dump_pqft(p_pqft);
					status = 1;
				}
			}
		}
	}

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_INFO,
		"PQFT Matched the discovered subnet\n");
	p_pqft->is_matched = TRUE;

Free_tmp_tuples:
	free(tmp_tuples);
Exit:
	cl_list_remove_all(&spine_nodes);
	cl_list_destroy(&spine_nodes);

	cl_list_remove_all(&leaf_ports);
	cl_list_destroy(&leaf_ports);

	cl_map_remove_all(&rank_by_nguid);
	cl_map_destroy(&rank_by_nguid);

	cl_qmap_remove_all(&tuple_calc_by_pguid);

	return status;
}

static void dump_hca_ordering(osm_pqft_t * p_pqft)
{
	unsigned int leaf_idx = 0;
	osm_pqft_node_t *p_pqft_node;
	char path[1024];
	FILE *p_hca_ordering_file;
	char *filename = "opensm-pqft-ca-order.dump";

	snprintf(path, sizeof(path), "%s/%s",
		 pqft_get_subnet(p_pqft)->opt.dump_files_dir, filename);
	p_hca_ordering_file = fopen(path, "w");
	if (!p_hca_ordering_file) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR, "ERR 6468: "
			"cannot open file \'%s\': %s\n", filename,
			strerror(errno));
		return;
	}

	/* for each leaf */
	for (leaf_idx = 0; leaf_idx < p_pqft->num_leafs; leaf_idx++) {
		p_pqft_node = p_pqft->leafs[leaf_idx];
		if (p_pqft_node) {
			fprintf(p_hca_ordering_file, "0x%04x\t%s\n",
				p_pqft_node->base_lid,
				p_pqft_node->p_osm_node->print_desc);
		} else {
			fprintf(p_hca_ordering_file, "0xFFFF\tDUMMY\n");
		}
	}
	/* done going through all the leaf switches */
	fclose(p_hca_ordering_file);
}

#if 0
/* note the below function is for checking only */
static int double_check_pgft_matching(osm_pqft_t * p_pqft)
{
	osm_pqft_node_t *p_pqft_node, *p_rem_pqft_node;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	uint8_t port_idx, neigh_idx, port_num, l;
	uint8_t rem_port_num, rem_neigh_idx, p, rem_port_idx;
	uint8_t rev_port_num, rev_port_idx;
	osm_pqft_tuple_t t;
	osm_node_t *p_rem_osm_node, *p_rev_osm_node;
	unsigned int anyErr = 0;

	/* go over all the pgft nodes
	   if the node was matched - go over all the up and down ports
	   if the port was matched check the other side node matches both as neighbor
	   and port number just go over all nodes and assign ports by algorithm */
	for (p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
	     p_pqft_node !=
	     (osm_pqft_node_t *) cl_qmap_end(&p_pqft->node_by_tuple);
	     p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->tuple_map_item)) {

		if (p_pqft_node->p_osm_node == NULL) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"skipping un-matched node: %s\n",
				tuple2str(p_pqft->H, p_pqft_node->tuple.d,
					  tuple_buf1));
			continue;
		}

		l = p_pqft_node->tuple.d[0];
		for (port_idx = 0; port_idx < p_pqft_node->num_down_ports;
		     port_idx++) {
			port_num = p_pqft_node->down_ports[port_idx].port_num;
			if (port_num == 0) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
					"skipping un-matched node: %s port_idx: %d\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx);
				continue;
			}

			/* PGFT specific */
			neigh_idx = port_idx % p_pqft->M[l - 1];
			p = port_idx / p_pqft->M[l - 1];
			rem_neigh_idx = p_pqft_node->tuple.d[l];

			/* now get the neighbor by tuple and then by physp num */
			t.key = p_pqft_node->tuple.key;
			t.d[0] = l - 1;
			t.d[l] = neigh_idx;

			p_rem_pqft_node = find_node_by_tuple(p_pqft, t);
			if (p_rem_pqft_node == NULL) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6451: Could not obtain remote node by tuple for "
					"node: %s port_idx: %d neighbor: %d\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					neigh_idx);
				anyErr++;
				continue;
			}

			p_rem_osm_node =
			    osm_node_get_remote_node(p_pqft_node->p_osm_node,
						     port_num, &rem_port_num);
			if (p_rem_osm_node == NULL) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6453: Could not obtain remote osm node for "
					"node: %s port_idx: %d port_num: %d\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					port_num);
				anyErr++;
				continue;
			}

			/* remote nodes are equal */
			if (p_rem_osm_node != p_rem_pqft_node->p_osm_node) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6454: Remote PQFT node for node: %s port_idx: %d "
					"does not remote match osm node. "
					"on remote node %s. Expected: %s got: %s\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					tuple2str(p_pqft->H,
						  p_rem_pqft_node->tuple.d,
						  tuple_buf),
					(p_rem_pqft_node->
					 p_osm_node ? p_rem_pqft_node->
					 p_osm_node->print_desc : "NULL"),
					(p_rem_osm_node ? p_rem_osm_node->
					 print_desc : "NULL"));
				anyErr++;
				continue;
			}

			if (p_pqft_node->down_ports[port_idx].
			    rem_neighbor_idx != rem_neigh_idx) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6455: Missmatch neighbor index for node: %s down "
					"port_idx: %d. Expected %d got %d (to remote node %s)\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					rem_neigh_idx,
					p_pqft_node->down_ports[port_idx].
					rem_neighbor_idx, tuple2str(p_pqft->H,
								    p_rem_pqft_node->
								    tuple.d,
								    tuple_buf));
				anyErr++;
				continue;
			}

			rem_port_idx =
			    p_pqft_node->down_ports[port_idx].rem_port_idx;
			if (rem_port_idx !=
			    rem_neigh_idx + p * p_pqft->W[l - 1]) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6456: Missmatch neighbor port index for node: %s "
					"port_idx: %d.\n", tuple2str(p_pqft->H,
								     p_pqft_node->
								     tuple.d,
								     tuple_buf1),
					port_idx);
				anyErr++;
				continue;
			}

			if (rem_port_num !=
			    p_rem_pqft_node->up_ports[rem_port_idx].port_num) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6457: Missmatch neighbor port num for node: %s "
					"port_idx: %d.\n", tuple2str(p_pqft->H,
								     p_pqft_node->
								     tuple.d,
								     tuple_buf1),
					port_idx);
				anyErr++;
				continue;
			}

			/* reverse port must match too */
			p_rev_osm_node =
			    osm_node_get_remote_node(p_rem_osm_node,
						     rem_port_num,
						     &rev_port_num);
			if (p_rev_osm_node == NULL) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6458: Could not obtain reverse osm node for node: %s "
					"port_idx: %d port_num: %d\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					port_num);
				anyErr++;
				continue;
			}

			if (p_rev_osm_node != p_pqft_node->p_osm_node) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6459: Reverse PQFT node for node: %s port_idx: %d "
					"does not remote match osm node.\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx);
				anyErr++;
				continue;
			}

			if (p_rem_pqft_node->up_ports[rem_port_idx].
			    rem_neighbor_idx != neigh_idx) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 645A: Missmatch reverse neighbor index for node: %s "
					"port_idx: %d. Expected %d got %d (to remote node %s)\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					neigh_idx,
					p_rem_pqft_node->up_ports[rem_port_idx].
					rem_neighbor_idx, tuple2str(p_pqft->H,
								    p_rem_pqft_node->
								    tuple.d,
								    tuple_buf));
				anyErr++;
				continue;
			}

			rev_port_idx =
			    p_rem_pqft_node->up_ports[rem_port_idx].
			    rem_port_idx;
			if (rev_port_idx != port_idx) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 645B: Missmatch reverse neighbor port index for node: %s "
					"port_idx: %d. Expected %d got %d (to remote node %s)\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx,
					port_idx, rev_port_idx,
					tuple2str(p_pqft->H,
						  p_rem_pqft_node->tuple.d,
						  tuple_buf));
				anyErr++;
				continue;
			}

			if (rev_port_num != port_num) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 645C: Missmatch reverse neighbor port num for node: %s "
					"port_idx: %d.\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf1), port_idx);
				anyErr++;
				continue;
			}

		}
	}
	return anyErr;
}
#endif

static int collect_leafs(osm_pqft_t * p_pqft)
{
	osm_pqft_node_t *p_pqft_node;
	unsigned int leaf_idx = 0;
	unsigned int max_leaf_idx = 0;
	uint8_t l;

	/* forget everything */
	memset(p_pqft->leafs, 0, p_pqft->num_leafs * sizeof(osm_pqft_node_t *));

	for (p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
	     p_pqft_node !=
	     (osm_pqft_node_t *) cl_qmap_end(&p_pqft->node_by_tuple);
	     p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->tuple_map_item)) {
		if (p_pqft_node->tuple.d[0] != 0)
			continue;

		if (p_pqft_node->port_guid == 0)
			continue;

		for (l = 0; l < p_pqft->H; l++) {
			if (l == 0)
				leaf_idx = p_pqft_node->tuple.d[l + 1];
			else
				leaf_idx +=
				    p_pqft_node->tuple.d[l + 1] * p_pqft->Q[l -
									    1];
		}

		if (leaf_idx >= p_pqft->num_leafs) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 648C: leaf_idx %d >= p_pqft->num_leafs"
				"%d\n",	leaf_idx, p_pqft->num_leafs);
			return 1;
		}

		if (p_pqft->leafs[leaf_idx] != NULL) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 648D: leaf_idx %d reassigned!!!\n",
				leaf_idx);
			return 1;

		}

		p_pqft->leafs[leaf_idx] = p_pqft_node;
		if (leaf_idx > max_leaf_idx)
			max_leaf_idx = leaf_idx;

		/* double check LID assignment ... */
		if (p_pqft->leafs[leaf_idx]->base_lid == 0)
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6445: Leaf index %d with zero base_lid!\n",
				leaf_idx);
		else
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
				"Leafs %d has base_lid %d\n",
				leaf_idx, p_pqft->leafs[leaf_idx]->base_lid);
	}
	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"Collected %d PQT leafs\n", max_leaf_idx + 1);
	dump_hca_ordering(p_pqft);
	return 0;
}

static void port_mask_init(osm_pqft_port_mask_t * pm, uint8_t num_ports)
{
	memset(pm, 0, sizeof(osm_pqft_port_mask_t));
	pm->max_mask = num_ports / 64;
}

static void port_mask_set(osm_pqft_port_mask_t * pm, uint8_t pn)
{
	uint8_t i = pn / 64;
	uint8_t m = pn % 64;
	pm->mask[i] |= (ib_net64_t) 1 << m;
}

/*
static void
port_mask_clr(osm_pqft_port_mask_t *pm, uint8_t pn)
{
  uint8_t i = pn/64;
  uint8_t m = pn % 64;
  pm->mask[i] &= ~((ib_net64_t)1<<m);
}
*/

static boolean_t port_mask_is_set(osm_pqft_port_mask_t * pm, uint8_t pn)
{
	uint8_t i = pn / 64;
	uint8_t m = pn % 64;
	return ((pm->mask[i] & ((ib_net64_t) 1 << m)) ? TRUE : FALSE);
}

static inline void
port_mask_copy(osm_pqft_port_mask_t * i, osm_pqft_port_mask_t * o)
{
	memcpy(o->mask, i->mask, sizeof(ib_net64_t) * (1 + i->max_mask));
	o->max_mask = i->max_mask;
}

/* return FALSE if no ports in results */
static boolean_t port_mask_and(osm_pqft_port_mask_t * i1,
			       osm_pqft_port_mask_t * i2,
			       osm_pqft_port_mask_t * o)
{
	boolean_t any = FALSE;
	uint8_t i;
	if (i1->max_mask < i2->max_mask)
		o->max_mask = i1->max_mask;
	else
		o->max_mask = i2->max_mask;

	for (i = 0; i <= o->max_mask; i++) {
		o->mask[i] = i1->mask[i] & i2->mask[i];
		if (o->mask[i])
			any = TRUE;
	}
	return any;
}

/*
static void
port_mask_or(osm_pqft_port_mask_t *i1, osm_pqft_port_mask_t *i2,
	     osm_pqft_port_mask_t *o)
{
  uint8_t i;
  for (i = 0; i <= i1->max_mask && i < i2->max_mask; i++)
    o->mask[i] = i1->mask[i] | i2->mask[i];
}
*/

/* scan through the topology find missing ports and update the DB */
int find_missing_up_ports(osm_pqft_t * p_pqft)
{
	osm_pqft_node_t *p_pqft_node;
	uint8_t l, i, lvl;
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	boolean_t any_missing;
	boolean_t all_missing;

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_INFO, "Find missing up ports.\n");
	cl_qmap_remove_all(&p_pqft->missing_up_port_nodes_by_tuple);

	/* if no up going port on a node we need to "fake" the ports going up to
	   it as down we do it by clearing remote side of its down ports.
	   So handle one layer at a time */
	for (lvl = p_pqft->H; lvl > 0; lvl--) {
		for (p_pqft_node =
		     (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
		     p_pqft_node !=
		     (osm_pqft_node_t *) cl_qmap_end(&p_pqft->node_by_tuple);
		     p_pqft_node =
		     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->
						      tuple_map_item)) {

			if (NULL == p_pqft_node->p_osm_node)
				continue;

			l = p_pqft_node->tuple.d[0];
			if (l != lvl)
				continue;

			if (p_pqft_node->num_up_ports == 0)
				continue;

			/* are we missing any ports? store them */
			any_missing = FALSE;
			all_missing = TRUE;
			port_mask_init(&p_pqft_node->available_up_ports_mask,
				       p_pqft_node->num_up_ports);
			for (i = 0; i < p_pqft_node->num_up_ports; i++) {
				if (p_pqft_node->up_ports[i].port_num) {
					port_mask_set(&p_pqft_node->
						      available_up_ports_mask,
						      i);
					all_missing = FALSE;
				} else {
					OSM_LOG(&p_pqft->p_osm->log,
						OSM_LOG_VERBOSE,
						"Missing up port "
						"at index %d on switch %s\n",
						i,
						tuple2str(p_pqft->H,
							  p_pqft_node->tuple.
							  d, tuple_buf));
					any_missing = TRUE;
				}
			}

			if (any_missing) {
				cl_qmap_insert(&p_pqft->
					       missing_up_port_nodes_by_tuple,
					       p_pqft_node->tuple.key,
					       &p_pqft_node->
					       available_up_ports_mask.
					       map_item);
			}

			if (all_missing) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
					"ALL up ports switch %s are missing!\n",
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf));

				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6442: Un implemented handling of case where"
					"all up ports are missing\n");

				/*
				 * TODO: declare all ports feeding into this switch as down.
				 * Go over all down ports of the node,
				 * get remote side port and zero the port num
				 */
				return 1;
			}
		}
	}

	/* TODO: maybe can improve performance by having any missing flag per node */
	if (cl_is_qmap_empty(&p_pqft->missing_up_port_nodes_by_tuple))
		p_pqft->any_missing_links = FALSE;
	else {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_INFO,
			"Some nodes are missing up ports - "
			"routing will be blocking for full shifts\n");
		p_pqft->any_missing_links = TRUE;
	}
	return 0;
}

static boolean_t is_pgft_descendant(osm_pqft_t * p_pqft,
				    osm_pqft_node_t * p_pqft_parent,
				    osm_pqft_node_t * p_pqft_dest)
{
	uint8_t l;
	for (l = p_pqft_parent->tuple.d[0] + 1; l <= p_pqft->H; l++) {
		if (p_pqft_parent->tuple.d[l] != p_pqft_dest->tuple.d[l])
			return FALSE;
	}
	return TRUE;
}

static boolean_t is_qft_descendant(osm_pqft_t * p_pqft,
				   osm_pqft_node_t * p_pqft_parent,
				   osm_pqft_node_t * p_pqft_dest)
{
	uint8_t l, i, j;
	uint8_t la = p_pqft_parent->tuple.d[0];

	for (l = la + 1; l <= p_pqft->H; l++) {
		if (l == la + 1) {
			i = p_pqft_parent->tuple.d[l] / p_pqft->P[l - 2];
			j = p_pqft_dest->tuple.d[l] / p_pqft->P[l - 2];
		} else {
			i = p_pqft_parent->tuple.d[l];
			j = p_pqft_dest->tuple.d[l];
		}
		if (i != j)
			return FALSE;
	}
	return TRUE;
}

/* HACK: for now we do it the most inefficient way.
   But maybe it is OK as there should not be many such cases */
static uint8_t get_random_available_down_port(osm_pqft_node_t * p_pqft_node,
					      uint8_t base_out_port_idx,
					      uint8_t num_parallel_ports)
{
	static uint8_t possible_ports[256];
	uint8_t idx = 0;
	uint8_t p, port_idx = base_out_port_idx;
	for (p = 0; p < num_parallel_ports; p++) {
		if (p_pqft_node->down_ports[port_idx].port_num) {
			possible_ports[idx++] =
			    p_pqft_node->down_ports[port_idx].port_num;
		}
		port_idx += p_pqft_node->num_down_neighbors;
	}
	if (idx != 0)
		return (possible_ports[rand() % idx]);

	return 0;
}

static uint8_t get_random_available_up_port_idx(osm_pqft_port_mask_t * port_mask)
{
	uint8_t possible_ports[256];
	uint8_t idx = 0;
	uint8_t i, p;
	ib_net64_t m;
	for (i = 0; i <= port_mask->max_mask; i++) {
		m = 1;
		if (port_mask->mask[i]) {
			for (p = 0; p < 64; p++) {
				if (port_mask->mask[i] & m)
					possible_ports[idx++] = i * 64 + p;
				m = m << 1;
			}
		}
	}
	if (idx != 0)
		return (possible_ports[rand() % idx]);

	return 0;
}

static int get_rand_up_port_step_by_start_step_n_cnt(osm_pqft_port_mask_t * port_mask,
						     uint8_t start,
						     uint8_t step,
						     uint8_t cnt, uint8_t * res_step)
{
	uint8_t possible_port_steps[256];
	uint8_t idx = 0;
	uint8_t i, p, o = 0, c = 0;
	ib_net64_t m;
	uint8_t start_i = start / 64;
	uint8_t stop_i = (start + cnt * step) / 64;

	if (stop_i > port_mask->max_mask)
		stop_i = port_mask->max_mask;
	if (stop_i > 255)
		stop_i = 255;
	o = start - start_i * 64;

	for (i = start_i; i <= stop_i; i++) {
		m = 1 << o;

		/* Confirm we are not getting out of bounds */
		if (i > port_mask->max_mask)
			return 1;

		/* any bit for this word? */
		if (port_mask->mask[i]) {
			for (p = o; (p < 64) && (c < cnt); p += step) {
				if (port_mask->mask[i] & m)
					possible_port_steps[idx++] = c;
				m = m << step;
				c++;
			}
			o = p - 64;
		}
	}

	if (idx != 0) {
		*res_step = possible_port_steps[rand() % idx];
		return 0;
	}

	return 1;
}

static osm_pqft_port_mask_t *find_up_port_mask_by_tuple(osm_pqft_t * p_pqft,
							osm_pqft_tuple_t t)
{
	cl_map_item_t *item =
	    cl_qmap_get(&p_pqft->missing_up_port_nodes_by_tuple, t.key);
	if (item == cl_qmap_end(&p_pqft->missing_up_port_nodes_by_tuple))
		return NULL;
	return ((osm_pqft_port_mask_t *) item);
}

/* get port mask for the switch above dest on same tree as source */
static osm_pqft_port_mask_t *get_pgft_dst_parent_mask(osm_pqft_t * p_pqft,
						      osm_pqft_node_t *
						      p_pqft_src,
						      unsigned int leaf_idx)
{
	osm_pqft_tuple_t dest_parent_tuple;
	osm_pqft_node_t *p_pqft_dst = p_pqft->leafs[leaf_idx];
	uint8_t i, l = p_pqft_src->tuple.d[0];
	osm_pqft_port_mask_t *res;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	char tuple_buf3[PQFT_TUPLE_BUF_LEN];
	if (p_pqft_dst == NULL)
		return NULL;

	/* for PGFT it is clear requirement the requested node upper digits are
	   equals the dst and lower digits equal the src */
	/* level is same as src */
	dest_parent_tuple.key = 0;
	dest_parent_tuple.d[0] = l;
	for (i = 1; i <= l; i++)
		dest_parent_tuple.d[i] = p_pqft_src->tuple.d[i];
	for (i = l + 1; i <= p_pqft->H; i++)
		dest_parent_tuple.d[i] = p_pqft_dst->tuple.d[i];

	res = find_up_port_mask_by_tuple(p_pqft, dest_parent_tuple);
	if (res) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
			"Dst parent %s for switch %s and dst %s mask 0x%016"
			PRIx64 "\n", tuple2str(p_pqft->H, dest_parent_tuple.d,
					       tuple_buf1), tuple2str(p_pqft->H,
								      p_pqft_src->
								      tuple.d,
								      tuple_buf),
			tuple2str(p_pqft->H, p_pqft_dst->tuple.d, tuple_buf3),
			res->mask[0]);
	}
	return res;
}

/* get port mask for the switch at level l above dest (exactlt 2 levels below
   the source)
   the switch above the dest that is 2 levels below source that is one level
   above P>1 is shown to be Dh...Dl+1, Sl..S1 */
static osm_pqft_port_mask_t *get_qft_dst_parent_mask(osm_pqft_t * p_pqft,
						     osm_pqft_node_t *
						     p_pqft_src,
						     unsigned int leaf_idx)
{
	osm_pqft_tuple_t dest_parent_tuple;
	osm_pqft_node_t *p_pqft_dst = p_pqft->leafs[leaf_idx];
	uint8_t i, l = p_pqft_src->tuple.d[0], lq;
	osm_pqft_port_mask_t *res;
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	char tuple_buf3[PQFT_TUPLE_BUF_LEN];
	if (p_pqft_dst == NULL)
		return NULL;

	/* src must be high enough and also one level above a parallel conn */
	if ((l < 2) || p_pqft->P[l - 2] == 1)
		return NULL;

	lq = l - 2;

	/* level is 2 below src */
	dest_parent_tuple.key = 0;
	dest_parent_tuple.d[0] = lq;
	for (i = 1; i <= lq; i++)
		dest_parent_tuple.d[i] = p_pqft_src->tuple.d[i];
	for (i = lq + 1; i <= p_pqft->H; i++)
		dest_parent_tuple.d[i] = p_pqft_dst->tuple.d[i];

	res = find_up_port_mask_by_tuple(p_pqft, dest_parent_tuple);
	if (res) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
			"QFT dst parent %s for switch %s and dst %s mask 0x%016"
			PRIx64 "\n",
			tuple2str(p_pqft->H, dest_parent_tuple.d, tuple_buf1),
			tuple2str(p_pqft->H, p_pqft_src->tuple.d, tuple_buf),
			tuple2str(p_pqft->H, p_pqft_dst->tuple.d, tuple_buf3),
			res->mask[0]);
	}
	return res;
}

/* given source and destination find a possible port offset */
static int get_qft_random_down_port_to_dst(osm_pqft_t * p_pqft,
					   osm_pqft_node_t * p_pqft_src,
					   unsigned int leaf_idx,
					   uint8_t * p_dn_port_step)
{
	osm_pqft_port_mask_t *p_dest_parent_mask;
	uint8_t l = p_pqft_src->tuple.d[0];
	uint8_t p = *p_dn_port_step;
	p_dest_parent_mask =
	    get_qft_dst_parent_mask(p_pqft, p_pqft_src, leaf_idx);
	if (p_dest_parent_mask == NULL)
		return 0;

	/* scan through the relevant ports -
	   the start index equal digit l-1 then skip W[l-1] up to P[l-1] */
	if (get_rand_up_port_step_by_start_step_n_cnt(p_dest_parent_mask,
						      p_pqft_src->tuple.d[l -
									  1],
						      p_pqft->W[l - 2],
						      p_pqft->P[l - 2], &p))
		return 1;

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"get_rand_up_port_step_by_start_step_n_cnt returned %d "
		"for start: %d step: %d cnt: %d\n",
		p, p_pqft_src->tuple.d[l - 1], p_pqft->W[l - 2],
		p_pqft->P[l - 2]);
	*p_dn_port_step = p;
	return 0;
}

/* return true if the src and dst nodes are common to same parent */
static boolean_t is_under_same_parent(osm_pqft_t * p_pqft,
				      osm_pqft_node_t * p_pqft_node,
				      osm_pqft_node_t * p_pqft_dest_node)
{
	char tuple_buf1[PQFT_TUPLE_BUF_LEN];
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	uint8_t i;
	uint8_t l = p_pqft_node->tuple.d[0];
	boolean_t res = TRUE;
	uint8_t p;
	if (p_pqft->subn_type == QFT_TYPE)
		for (i = l + 2; i <= p_pqft->H; i++) {
			p = p_pqft->P[i - 2];
			if (p_pqft_node->tuple.d[i] / p !=
			    p_pqft_dest_node->tuple.d[i] / p)
				res = FALSE;
	} else
		for (i = l + 2; i <= p_pqft->H; i++)
			if (p_pqft_node->tuple.d[i] !=
			    p_pqft_dest_node->tuple.d[i])
				res = FALSE;

	OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
		"Under same parent for: %s (%s) %s (%s) is %s\n",
		p_pqft_node->p_osm_node->print_desc,
		tuple2str(p_pqft->H, p_pqft_node->tuple.d, tuple_buf1),
		p_pqft_dest_node->p_osm_node->print_desc,
		tuple2str(p_pqft->H, p_pqft_dest_node->tuple.d, tuple_buf),
		(res == TRUE ? "TRUE" : "FALSE"));
	return (res);
}

static int full_tree_routing(osm_pqft_t * p_pqft)
{
	osm_pqft_node_t *p_pqft_node, *p_pqft_dest;
	unsigned int leaf_idx;
	uint8_t path_bit = 0, out_port = 0, out_port_idx = 0, l, j, p, R, base_out_port_idx;
	osm_switch_t *p_sw;
	char tuple_buf[PQFT_TUPLE_BUF_LEN];
	boolean_t was_up = FALSE;
	osm_pqft_port_mask_t avail_up_ports;
	osm_pqft_port_mask_t *p_this_up_ports_mask = NULL;
	osm_pqft_port_mask_t *p_dest_up_ports_mask = NULL;

	/* just go over all nodes and assign ports by algorithm */
	for (p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
	     p_pqft_node !=
	     (osm_pqft_node_t *) cl_qmap_end(&p_pqft->node_by_tuple);
	     p_pqft_node =
	     (osm_pqft_node_t *) cl_qmap_next(&p_pqft_node->tuple_map_item)) {

		if (NULL == p_pqft_node->p_osm_node)
			continue;

		l = p_pqft_node->tuple.d[0];
		if (l == 0)
			continue;

		p_sw = p_pqft_node->p_osm_node->sw;
		if (!p_sw) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 648E: No switch for node 0x%016" PRIx64
				"\n", cl_ntoh64(osm_node_get_node_guid
						(p_pqft_node->p_osm_node)));
			return 1;
		}

		memset(p_sw->new_lft, OSM_NO_PATH, p_sw->lft_size);

		/* get the current node available up ports mask */
		if (p_pqft->any_missing_links) {
			/* note if full no mask found */
			p_this_up_ports_mask =
			    find_up_port_mask_by_tuple(p_pqft,
						       p_pqft_node->tuple);
		}

		/* assign routes to all leafs */
		for (leaf_idx = 0; leaf_idx < p_pqft->num_leafs; leaf_idx++) {
			out_port = OSM_NO_PATH;
			p_pqft_dest = p_pqft->leafs[leaf_idx];

			if (p_pqft_dest == NULL)
				continue;

			if (p_pqft->subn_type == PGFT_TYPE) {
				if (is_pgft_descendant
				    (p_pqft, p_pqft_node, p_pqft_dest)) {
					/* going down */
					was_up = FALSE;
					base_out_port_idx = out_port_idx =
					    p_pqft_dest->tuple.d[l];

					if ((l > 1) && (p_pqft->P[l - 1] != 1)) {
						out_port_idx +=
						    p_pqft->M[l -
							      1] * ((leaf_idx /
								     p_pqft->
								     R[l -
								       1]) %
								    p_pqft->
								    P[l - 1]);
					}

					if (out_port_idx >=
					    p_pqft_node->num_down_ports) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR, "ERR 648F: "
							"Overflow on down ports"
							" on node 0x%016" PRIx64
							" idx %d >= num_down_ports"
							" %d. l=%d, d[l]=%d M[l]=%d,"
							" leaf=%d, R[l]=%d, P[l]=%d\n",
							cl_ntoh64(osm_node_get_node_guid(p_pqft_node->p_osm_node)),
							out_port_idx,
							p_pqft_node->num_down_ports,
							l, p_pqft_dest->tuple.d[l],
							p_pqft->M[l - 1], leaf_idx,
							p_pqft->R[l - 1],
							p_pqft->P[l - 1]);
						return 1;
					}

					out_port =
					    p_pqft_node->
					    down_ports[out_port_idx].port_num;
					/* we may recover down port only if P>1 and other ports exist */
					if ((out_port == 0)
					    && (p_pqft->P[l - 1] > 1)) {
						out_port =
						    get_random_available_down_port
						    (p_pqft_node,
						     base_out_port_idx,
						     p_pqft->P[l - 1]);
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_VERBOSE,
							"Selected Random Down Parallel port on %s to dst "
							"lid: 0x%04x idx: %d out_port %d\n",
							tuple2str(p_pqft->H,
								  p_pqft_node->
								  tuple.d,
								  tuple_buf),
							p_pqft_dest->base_lid,
							leaf_idx, out_port);
					}

					/* if could not find a downport this may acceptible if there are
					   some missing links */
					if ((out_port == 0)
					    && p_pqft->any_missing_links) {
						out_port = 255;
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_VERBOSE,
							"PGFT routing to lid: 0x%04x idx: %d deadend on switch %s. "
							"out_port_idx %d going DOWN path up should never go"
							" through here\n",
							p_pqft_dest->base_lid,
							leaf_idx,
							tuple2str(p_pqft->H,
								  p_pqft_node->
								  tuple.d,
								  tuple_buf),
							out_port_idx);
					}

				} else {
					/* going up */
					was_up = TRUE;
					PQFT_ASSERT(&p_pqft->p_osm->log,
						    (l < p_pqft->H),
						    "Routing up of top node 0x%016"
						    PRIx64 "\n",
						    cl_ntoh64
						    (osm_node_get_node_guid
						     (p_pqft_node->
						      p_osm_node)));

					out_port_idx =
					    (leaf_idx / p_pqft->R[l - 1]) %
					    (p_pqft->W[l] * p_pqft->P[l]);
					if (p_pqft->any_missing_links) {
						/* in case of some errors we need to make sure up port masks
						   are fine... */
						p_dest_up_ports_mask =
						    get_pgft_dst_parent_mask
						    (p_pqft, p_pqft_node,
						     leaf_idx);

						if ((p_dest_up_ports_mask !=
						     NULL)
						    || (p_this_up_ports_mask !=
							NULL)) {
							if (p_dest_up_ports_mask
							    == NULL) {
								port_mask_copy
								    (p_this_up_ports_mask,
								     &avail_up_ports);
							} else
							    if
							    (p_this_up_ports_mask
							     == NULL) {
								port_mask_copy
								    (p_dest_up_ports_mask,
								     &avail_up_ports);
							} else {
								/* both are not NULL */
								port_mask_and
								    (p_this_up_ports_mask,
								     p_dest_up_ports_mask,
								     &avail_up_ports);
							}

							/* do random if on any missing link or this particularl
							   path is absent ... */
							if (p_pqft->
							    random_on_non_full_tree
							    || (FALSE ==
								port_mask_is_set
								(&avail_up_ports,
								 out_port_idx)))
							{
								out_port_idx =
								    get_random_available_up_port_idx
								    (&avail_up_ports);
								OSM_LOG
								    (&p_pqft->
								     p_osm->log,
								     OSM_LOG_VERBOSE,
								     "Selected Random up port on %s to dst lid: 0x%04x "
								     "idx: %d out_port_idx %d\n",
								     tuple2str
								     (p_pqft->H,
								      p_pqft_node->
								      tuple.d,
								      tuple_buf),
								     p_pqft_dest->
								     base_lid,
								     leaf_idx,
								     out_port_idx);
							}
						}
					}

					if (out_port_idx >=
					    p_pqft_node->num_up_ports) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR, "ERR 6490: "
							"Overflow on up ports on"
							" node 0x%016" PRIx64
							" idx %d >= num_up_ports"
							" %d. leaf_idx %d R[l-1]"
							" %d W[l+1] %d P[l+1] %d\n",
							cl_ntoh64(osm_node_get_node_guid(p_pqft_node->p_osm_node)),
							out_port_idx,
							p_pqft_node->num_up_ports,
							leaf_idx, p_pqft->R[l - 1],
							p_pqft->W[l], p_pqft->P[l]);
						return 1;
					}

					out_port =
					    p_pqft_node->up_ports[out_port_idx].
					    port_num;
				}	/* up port */

			} else if (p_pqft->subn_type == QFT_TYPE) {
				if (is_qft_descendant
				    (p_pqft, p_pqft_node, p_pqft_dest)) {
					/* going down */
					was_up = FALSE;

					if ((l > 1) && (p_pqft->P[l - 2] > 1)) {
						/* >1 parallel paths exists */
						base_out_port_idx =
						    (p_pqft_dest->tuple.d[l] /
						     p_pqft->P[l -
							       2]) *
						    p_pqft->P[l - 2];
						if (l > 2)
							R = p_pqft->R[l - 3];
						else
							R = 1;

						/* the calculated offset within the "bundle" */
						j = ((leaf_idx / R) %
						     (p_pqft->W[l - 2] *
						      p_pqft->P[l -
								2])) /
						    p_pqft->W[l - 2];

						/* for faulty network check the passby switch does not
						   miss some ports ... */
						if (p_pqft->any_missing_links) {
							p = j;
							if (get_qft_random_down_port_to_dst(p_pqft, p_pqft_node, leaf_idx, &p)) {
								OSM_LOG
								    (&p_pqft->
								     p_osm->log,
								     OSM_LOG_ERROR,
								     "ERR 6480: Could not find QFT down port valid from:"
								     " %s to %s \n",
								     p_pqft_node->
								     p_osm_node->
								     print_desc,
								     p_pqft_dest->
								     p_osm_node->
								     print_desc);
								return 1;
							}
							OSM_LOG(&p_pqft->p_osm->
								log,
								OSM_LOG_VERBOSE,
								"QFT routing from: %s to: %s using alternate port: "
								"%d (instead of %d) to avoid missing 2 level down.\n",
								p_pqft_node->
								p_osm_node->
								print_desc,
								p_pqft_dest->
								p_osm_node->
								print_desc,
								base_out_port_idx
								+ p,
								base_out_port_idx
								+ j);
							j = p;
						}

						out_port_idx =
						    base_out_port_idx + j;

					} else {
						/* single path - use connection rule */
						out_port_idx =
						    p_pqft_dest->tuple.d[l];
						if (p_pqft->P[l - 1] > 1) {
							if (l < p_pqft->H) {
								out_port_idx +=
								    p_pqft->
								    M[l -
								      1] *
								    (p_pqft_dest->
								     tuple.d[l +
									     1]
								     %
								     p_pqft->
								     P[l - 1]);
							} else {
								out_port_idx +=
								    p_pqft->
								    M[l -
								      1] *
								    (p_pqft_dest->
								     tuple.d[l -
									     1]
								     %
								     p_pqft->
								     P[l - 1]);
							}
						}
						j = 0;
					}

					if (out_port_idx >=
					    p_pqft_node->num_down_ports) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR, "ERR 6491: "
							"Overflow on down ports"
							" on node 0x%016" PRIx64
							" idx %d >= num_down_ports"
							" %d. Calculations used: "
							"l=%d, leaf=%d, d[l]=%d "
							"d[l+1]=%d M[l]=%d, "
							"P[l-1]=%d, W[l-1]=%d,"
							" j=%d\n",
							cl_ntoh64(osm_node_get_node_guid(p_pqft_node->p_osm_node)),
							out_port_idx,
							p_pqft_node->num_down_ports,
							l, leaf_idx,
							p_pqft_dest->tuple.d[l],
							p_pqft_dest->tuple.d[l + 1],
							p_pqft->M[l - 1],
							p_pqft->P[l - 2],
							p_pqft->W[l - 2], j);
						return 1;
					}
					PQFT_ASSERT(&p_pqft->p_osm->log,
						    (out_port_idx <
						     p_pqft_node->
						     num_down_ports),
						    "Overflow on down ports on node 0x%016"
						    PRIx64
						    " idx %d >= num_down_ports %d. Calculations used: "
						    "l=%d, leaf=%d, d[l]=%d d[l+1]=%d M[l]=%d, "
						    "P[l-1]=%d, W[l-1]=%d, j=%d\n",
						    cl_ntoh64
						    (osm_node_get_node_guid
						     (p_pqft_node->p_osm_node))
						    , out_port_idx,
						    p_pqft_node->num_down_ports,
						    l, leaf_idx,
						    p_pqft_dest->tuple.d[l],
						    p_pqft_dest->tuple.d[l + 1],
						    p_pqft->M[l - 1],
						    p_pqft->P[l - 2],
						    p_pqft->W[l - 2], j);

					out_port =
					    p_pqft_node->
					    down_ports[out_port_idx].port_num;

					if ((out_port == 0)
					    && p_pqft->any_missing_links) {
						out_port = 255;
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_VERBOSE,
							"QFT routing to lid: 0x%04x idx: %d deadend on switch %s. "
							"out_port_idx %d going DOWN path up should never go"
							" through here\n",
							p_pqft_dest->base_lid,
							leaf_idx,
							tuple2str(p_pqft->H,
								  p_pqft_node->
								  tuple.d,
								  tuple_buf),
							out_port_idx);
					}
				} else {
					/* going up */
					was_up = TRUE;

					if (l >= p_pqft->H) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR, "ERR 6492: "
							"Routing up of top node"
							" 0x%016" PRIx64 "\n",
							cl_ntoh64(osm_node_get_node_guid(p_pqft_node->p_osm_node)));
						return 1;
					}

					/* criteria for being a merge */
					if ((l > 1) && (p_pqft->P[l - 2] > 1)) {
						if (l > 2)
							R = p_pqft->R[l - 3];
						else
							R = 1;
						j = ((leaf_idx / R) %
						     (p_pqft->W[l - 2] *
						      p_pqft->P[l -
								2])) /
						    p_pqft->W[l - 2];
						out_port_idx =
						    ((leaf_idx /
						      p_pqft->R[l -
								1]) *
						     p_pqft->P[l - 2] +
						     j) % (p_pqft->W[l] *
							   p_pqft->P[l]);
					} else {
						out_port_idx =
						    (leaf_idx /
						     p_pqft->R[l -
							       1]) %
						    (p_pqft->W[l] *
						     p_pqft->P[l]);
					}

					if (p_pqft->any_missing_links) {
						/* if just below a parallel and not on same subtree -
						   only take local ports else do like PGFT */
						if ((p_pqft->P[l] == 1) ||
						    is_under_same_parent(p_pqft,
									 p_pqft_node,
									 p_pqft_dest))
							/* in case of some errors we need to make sure up port masks
							   are fine... */
							p_dest_up_ports_mask =
							    get_pgft_dst_parent_mask
							    (p_pqft,
							     p_pqft_node,
							     leaf_idx);
						else
							p_dest_up_ports_mask =
							    NULL;

						if ((p_dest_up_ports_mask !=
						     NULL)
						    || (p_this_up_ports_mask !=
							NULL)) {
							if (p_dest_up_ports_mask
							    == NULL) {
								port_mask_copy
								    (p_this_up_ports_mask,
								     &avail_up_ports);
							} else
							    if
							    (p_this_up_ports_mask
							     == NULL) {
								port_mask_copy
								    (p_dest_up_ports_mask,
								     &avail_up_ports);
							} else {
								/* both are not NULL */
								port_mask_and
								    (p_this_up_ports_mask,
								     p_dest_up_ports_mask,
								     &avail_up_ports);
							}

							/* do random if on any missing link or this particularl path
							   is absent ... */
							if (p_pqft->
							    random_on_non_full_tree
							    || (FALSE ==
								port_mask_is_set
								(&avail_up_ports,
								 out_port_idx)))
							{
								out_port_idx =
								    get_random_available_up_port_idx
								    (&avail_up_ports);
								OSM_LOG
								    (&p_pqft->
								     p_osm->log,
								     OSM_LOG_VERBOSE,
								     "QFT Selected Random up port on %s to dst lid: "
								     "0x%04x idx: %d out_port_idx %d\n",
								     tuple2str
								     (p_pqft->H,
								      p_pqft_node->
								      tuple.d,
								      tuple_buf),
								     p_pqft_dest->
								     base_lid,
								     leaf_idx,
								     out_port_idx);
							}
						}
					}

					if (out_port_idx >=
					    p_pqft_node->num_up_ports) {
						OSM_LOG(&p_pqft->p_osm->log,
							OSM_LOG_ERROR, "ERR 6493: "
							"Overflow on up ports on"
							" node 0x%016" PRIx64
							" idx %d >= num_up_ports"
							" %d. leaf_idx %d R[l-1]"
							" %d W[l+1] %d P[l+1] %d\n",
							cl_ntoh64(osm_node_get_node_guid(p_pqft_node->p_osm_node)),
							out_port_idx,
							p_pqft_node->num_up_ports,
							leaf_idx, p_pqft->R[l - 1],
							p_pqft->W[l], p_pqft->P[l]);
						return 1;
					}

					out_port =
					    p_pqft_node->up_ports[out_port_idx].
					    port_num;
				}
			}

			if ((out_port == 0) && was_up) {
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
					"ERR 6481: Full routing to lid: 0x%04x idx: %d deadend on "
					"switch %s. out_port_idx %d going %s\n",
					p_pqft_dest->base_lid, leaf_idx,
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf), out_port_idx,
					(was_up ? "UP" : "DOWN"));
				return 1;
			} else {
				/* TODO: support LMC? */
				path_bit = (p_pqft_dest->p_osm_node->sw) ?
				    get_sw_path_bit(p_pqft) : get_path_bit(p_pqft);
				OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_VERBOSE,
					"Routing to lid: 0x%04x idx: %d switch %s. "
					"out_port_idx %d out_port %d going %s\n",
					p_pqft_dest->base_lid + path_bit, leaf_idx,
					tuple2str(p_pqft->H,
						  p_pqft_node->tuple.d,
						  tuple_buf), out_port_idx,
					out_port, (was_up ? "UP" : "DOWN"));
				p_sw->new_lft[p_pqft_dest->base_lid + path_bit] = out_port;
			}
		}
	}

	return 0;
}

/***************************************************
 * MAIN ENTRY POINTS - osm_routing_engine functions
 ***************************************************/

osm_pqft_t *construct_pqft(osm_opensm_t * p_osm)
{
	osm_pqft_t *p_pqft = (osm_pqft_t *) malloc(sizeof(osm_pqft_t));
	if (NULL == p_pqft)
		return NULL;

	memset(p_pqft, 0, sizeof(osm_pqft_t));

	p_pqft->p_osm = p_osm;

	/* find and parse the pqft definition from the opts */
	osm_pqft_type_t T = PGFT_TYPE;
	uint8_t H = 0;
	uint8_t *M = NULL;
	uint8_t *W = NULL;
	uint8_t *P = NULL;
	if (parse_pqft_option(p_pqft, &T, &H, &M, &W, &P)) {
		free(p_pqft);
		return NULL;
	}

	if (pqft_subn_init(p_osm, p_pqft, T, H, M, W, P)) {
		free(p_pqft);
		return NULL;
	}

	OSM_LOG(&p_osm->log, OSM_LOG_INFO, "PGFT/QFT Constructed\n");

	return p_pqft;
}

static int subn_match(IN void *context)
{
	osm_pqft_t *p_pqft = (osm_pqft_t *) context;
	osm_pqft_node_t *p_pqft_node;
	cl_map_item_t *p_map_item;
	osm_switch_t *p_sw;
	unsigned int i, h;

	/* Clear p_qft_node data */
	p_pqft_node = (osm_pqft_node_t *) cl_qmap_head(&p_pqft->node_by_tuple);
	while (&p_pqft_node->tuple_map_item !=
	       cl_qmap_end(&p_pqft->node_by_tuple)) {
		p_pqft_node->p_osm_node = NULL;
		p_pqft_node->port_guid = 0;
		p_pqft_node->base_lid = 0;
		p_pqft_node->next_down_neighbor_idx = 0;
		p_pqft_node->next_up_neighbor_idx = 0;
		/* the port_num == 0 means no matching done yet of this port */
		for (i = 0; i < p_pqft_node->num_up_ports; i++)
			p_pqft_node->up_ports[i].port_num = 0;
		for (i = 0; i < p_pqft_node->num_down_ports; i++)
			p_pqft_node->down_ports[i].port_num = 0;
		memset(&p_pqft_node->available_up_ports_mask, 0, sizeof(osm_pqft_port_mask_t));
		p_pqft_node->qft_num_over_mapped = 0;
		p_pqft_node = (osm_pqft_node_t *)cl_qmap_next(&p_pqft_node->tuple_map_item);
	}

	cl_qmap_remove_all(&p_pqft->node_by_pguid);
	cl_qmap_remove_all(&p_pqft->missing_up_port_nodes_by_tuple);
	p_pqft->is_matched = FALSE;
	p_pqft->any_missing_links = FALSE;
	p_pqft->random_on_non_full_tree = FALSE;
	cl_qmap_t *via_nodes_by_nguid = &pqft_get_subnet(p_pqft)->node_guid_tbl;
	if (match_nodes_to_pqft(p_pqft, via_nodes_by_nguid,
				&pqft_get_subnet(p_pqft)->port_guid_tbl))
		return -1;

#if 0
	/* TODO: remove or make verbose mode */
	dump_pqft(p_pqft);

	/* TODO: remove double check */
	if ((p_pqft->subn_type == PGFT_TYPE) &&
	    (double_check_pgft_matching(p_pqft) > 0))
		return -1;
#endif
	/* Build the full lid matrices needed for multicast routing */
	if (p_pqft->p_osm->subn.opt.calc_missing_routes)
		osm_ucast_mgr_build_lid_matrices(&p_pqft->p_osm->sm.ucast_mgr);

	/* Assign ranks to switches */
	for (p_map_item = cl_qmap_head(&p_pqft->node_by_tuple);
	     p_map_item != cl_qmap_end(&p_pqft->node_by_tuple);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_pqft_node = (osm_pqft_node_t *) p_map_item;
		if (!p_pqft_node->p_osm_node ||
		    !(p_sw = p_pqft_node->p_osm_node->sw))
			continue;

		h = p_pqft_node->tuple.d[0];
		if (h > 0 && h <= p_pqft->H)
			p_sw->rank = p_pqft->H - h;
	}

	return 0;
}

static int find_sw_2_sw_path(osm_pqft_t *p_pqft, cl_qmap_t *all_sw_tbl,
			     cl_qmap_t *up_sw_tbl, osm_pqft_node_t *from_node,
			     osm_pqft_node_t *to_node, uint8_t up_port)
{
	osm_pqft_sw_t *p_pqft_sw;
	cl_map_item_t *p_map_item, *tmp_map_item;
	uint8_t down_port, rem_port;
	uint16_t lid;
	osm_switch_t *p_sw;
	osm_physp_t *rem_physp;
	int ret = -1;

	p_pqft_sw = (osm_pqft_sw_t *)cl_qmap_get(all_sw_tbl,
						 from_node->port_guid);

	OSM_ASSERT(from_node->p_osm_node->sw);

	/* Update path to itself */
	lid = from_node->base_lid + get_sw_path_bit(p_pqft);
	from_node->p_osm_node->sw->new_lft[lid] = 0;

	/* Wire all up ports through up_port */
	for (p_map_item = cl_qmap_head(up_sw_tbl);
	     p_map_item != cl_qmap_end(up_sw_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_pqft_sw = (osm_pqft_sw_t *)p_map_item;
		p_sw = from_node->p_osm_node->sw;
		OSM_ASSERT(p_sw);
		lid = p_pqft_sw->p_pqft_node->base_lid +
		    get_sw_path_bit(p_pqft);
		p_sw->new_lft[lid] = up_port;
	}

	/* If reached the hub switch */
	if (from_node == to_node->up_neighbors[0].rem_node) {
		ret = 0;
		return ret;
	}

	/* Find the port that going towards the hub switch */
	lid = to_node->base_lid + get_path_bit(p_pqft);
	down_port = from_node->p_osm_node->sw->new_lft[lid];

	if (down_port == OSM_NO_PATH) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6966: Failed to build sw_2_sw routing - couldn't "
			"find down_port from 0x%016" PRIx64 "\n",
			cl_ntoh64(from_node->port_guid));
		return ret;
	}
	/* Wire all the unrouted switches through the down_port */
	for (p_map_item = cl_qmap_head(all_sw_tbl);
	     p_map_item != cl_qmap_end(all_sw_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_pqft_sw = (osm_pqft_sw_t *)p_map_item;
		p_sw = from_node->p_osm_node->sw;
		OSM_ASSERT(p_sw);
		lid = p_pqft_sw->p_pqft_node->base_lid +
		      get_sw_path_bit(p_pqft);
		if (p_sw->new_lft[lid] != OSM_NO_PATH)
			continue;
		p_sw->new_lft[lid] = down_port;
	}

	rem_physp = from_node->p_osm_node->physp_table[down_port].p_remote_physp;
	if (!rem_physp) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR, "ERR 6496: "
			"Failed to find down remote port from 0x%" PRIx64
			" port %d\n",
			cl_ntoh64(osm_node_get_node_guid(from_node->p_osm_node)),
			down_port);
		return ret;
	}

	if (!p_pqft_sw->is_matched) {
		p_pqft_sw = calloc(1, sizeof(osm_pqft_sw_t));
		if (!p_pqft_sw) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6494: Out of memory\n");
			return ret;
		}
		p_pqft_sw->p_pqft_node = from_node;
		tmp_map_item = cl_qmap_insert(up_sw_tbl, from_node->port_guid,
					      &p_pqft_sw->map_item);
		if (tmp_map_item != &p_pqft_sw->map_item) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6495: insert to all_sw_tbl failed\n");
			free(p_pqft_sw);
			return ret;
		}
		p_pqft_sw->is_matched = TRUE;
	}

	rem_port = rem_physp->port_num;
	from_node = find_node_by_pguid(p_pqft, rem_physp->port_guid);
	ret = find_sw_2_sw_path(p_pqft, all_sw_tbl, up_sw_tbl,
				from_node, to_node, rem_port);
	return ret;
}

static int sw_2_sw_routing(osm_pqft_t *p_pqft)
{
	osm_pqft_node_t *p_pqft_node;
	osm_pqft_sw_t *p_pqft_sw;
	cl_map_item_t *p_map_item, *tmp_map_item;
	cl_qmap_t all_sw_tbl, up_sw_tbl;
	osm_switch_t *p_hub_sw = NULL;
	osm_node_t *p_node;
	unsigned int i, ret = 0;
	uint16_t lid;

	cl_qmap_init(&all_sw_tbl);
	cl_qmap_init(&up_sw_tbl);

	/* Populate all_sw_tbl */
	for (p_map_item = cl_qmap_head(&p_pqft->node_by_tuple);
	     p_map_item != cl_qmap_end(&p_pqft->node_by_tuple);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_pqft_node = (osm_pqft_node_t *) p_map_item;
		if (!p_pqft_node->p_osm_node ||
		    !p_pqft_node->p_osm_node->sw)
			continue;
		p_pqft_sw = calloc(1, sizeof(osm_pqft_sw_t));
		if (!p_pqft_sw) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6496: Out of memory\n");
			ret = -1;
			goto Exit;
		}
		p_pqft_sw->p_pqft_node = p_pqft_node;
		tmp_map_item = cl_qmap_insert(&all_sw_tbl,
					      p_pqft_node->port_guid,
					      &p_pqft_sw->map_item);
		if (tmp_map_item != &p_pqft_sw->map_item) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6497: insert to all_sw_tbl failed\n");
			ret = -1;
			goto Exit;
		}
	}

	/* Locate hub switch */
	for (i = 0; i < p_pqft->num_leafs; i++) {
		if (!p_pqft->leafs[i])
			continue;
		if (!p_pqft->leafs[i]->num_up_ports)
			continue;
		p_pqft_node = p_pqft->leafs[i]->up_neighbors[0].rem_node;
		OSM_ASSERT(p_pqft_node);
		p_node = p_pqft_node->p_osm_node;
		if (!p_node || !p_node->sw || !p_pqft->leafs[i]->base_lid)
			continue;
		p_hub_sw = p_node->sw;
		lid = p_pqft->leafs[i]->base_lid + get_path_bit(p_pqft);
		for (p_map_item = cl_qmap_head(&p_pqft->node_by_tuple);
		     p_map_item != cl_qmap_end(&p_pqft->node_by_tuple);
		     p_map_item = cl_qmap_next(p_map_item)) {
			p_pqft_node = (osm_pqft_node_t *) p_map_item;
			if (!p_pqft_node->p_osm_node ||
			    !p_pqft_node->p_osm_node->sw)
				continue;
			if (p_pqft_node->p_osm_node->sw->new_lft[lid] == OSM_NO_PATH) {
				p_hub_sw = NULL;
				break;
			}
		}
		if (p_hub_sw)
			break;
	}

	if (!p_hub_sw) {
		OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
			"ERR 6498: hub sw wasn't found\n");
		ret = -1;
		goto Exit;
	}

	for (p_map_item = cl_qmap_head(&all_sw_tbl);
	     p_map_item != cl_qmap_end(&all_sw_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_pqft_sw = (osm_pqft_sw_t *)p_map_item;
		if (p_pqft_sw->is_matched)
			continue;
		/* Build route from the hub switch and p_pqft_sw */
		ret = find_sw_2_sw_path(p_pqft, &all_sw_tbl, &up_sw_tbl,
					p_pqft_sw->p_pqft_node,
					p_pqft->leafs[i], 0);
		/* Clear up_sw_tbl */
		while (cl_qmap_count(&up_sw_tbl)) {
			tmp_map_item = cl_qmap_head(&up_sw_tbl);
			cl_qmap_remove_item(&up_sw_tbl, tmp_map_item);
			free(tmp_map_item);
		}
		if (ret) {
			OSM_LOG(&p_pqft->p_osm->log, OSM_LOG_ERROR,
				"ERR 6499: Failed to find route between 0x%"
				PRIx64 " and 0x%" PRIx64 "\n",
				cl_ntoh64(p_pqft_sw->p_pqft_node->port_guid),
				cl_ntoh64(p_pqft->leafs[i]->port_guid));
			ret = -1;
			goto Exit;
		}
	}
Exit:
	while (cl_qmap_count(&all_sw_tbl)) {
		tmp_map_item = cl_qmap_head(&all_sw_tbl);
		cl_qmap_remove_item(&all_sw_tbl, tmp_map_item);
		free(tmp_map_item);
	}
	return ret;
}

static int do_routing(IN void *context)
{
	osm_pqft_t *p_pqft = (osm_pqft_t *) context;
	osm_subn_t * p_subn = pqft_get_subnet(p_pqft);

	if (p_pqft->is_matched == FALSE)
		return -1;

	/* reset new lfts from previous routing */
	osm_ucast_mgr_reset_lfts_all_switches(p_subn);

	/* collect all leafs */
	if (collect_leafs(p_pqft))
		return -1;

	/* analyze missing ports */
	if (find_missing_up_ports(p_pqft))
		return -1;

	/* perform "optimal" routing */
	if (full_tree_routing(p_pqft))
		return -1;

	/* connect switches */

	if (p_pqft->p_osm->subn.opt.calc_missing_routes &&
	    sw_2_sw_routing(p_pqft))
		return -1;

	return 0;
}

static void delete_pqft_node(osm_pqft_node_t *p_pqft_node)
{
	if (!p_pqft_node)
		return;

	free(p_pqft_node->down_neighbors);
	free(p_pqft_node->down_ports);
	free(p_pqft_node->up_neighbors);
	free(p_pqft_node->up_ports);
	free(p_pqft_node);
}

static void delete(IN void *context)
{
	osm_pqft_t *p_pqft = (osm_pqft_t *) context;
	cl_map_item_t *p_map_item;

	if (NULL == p_pqft)
		return;

	free(p_pqft->Q);
	free(p_pqft->R);
	free(p_pqft->M);
	free(p_pqft->W);
	free(p_pqft->P);
	while (cl_qmap_count(&p_pqft->node_by_tuple)) {
		p_map_item = cl_qmap_head(&p_pqft->node_by_tuple);
		cl_qmap_remove_item(&p_pqft->node_by_tuple, p_map_item);
		delete_pqft_node((osm_pqft_node_t *)p_map_item);
	}
	free(p_pqft->leafs);
	free(p_pqft);

}

int osm_ucast_pqft_setup(struct osm_routing_engine *r, osm_opensm_t * p_osm)
{
	osm_pqft_t *p_pqft = construct_pqft(p_osm);
	if (NULL == p_pqft)
		return -1;

	r->context = (void *)p_pqft;
	r->build_lid_matrices = subn_match;
	r->ucast_build_fwd_tables = do_routing;
	r->destroy = delete;

	return 0;
}
