/*
 * Copyright (c) 2004-2008 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2013 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Up Down Algorithm using ranking & Min Hop
 *      Calculation functions
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_UPDN_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>

extern int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
                                              void *cxt, char *pi, void *p_osm, boolean_t goto_node);

/* //////////////////////////// */
/*  Local types                 */
/* //////////////////////////// */

/* direction */
typedef enum updn_switch_dir {
	UP = 0,
	DOWN
} updn_switch_dir_t;

typedef struct updn_node {
	cl_list_item_t list;
	osm_switch_t *sw;
	uint64_t id;
	updn_switch_dir_t dir;
	unsigned visited;
	uint8_t num_hca;
	boolean_t is_leaf;
	uint32_t *dlid_routes;
	uint32_t *dlid_sw_routes;
	uint32_t total_routes;
} updn_node_t;

/* updn structure */
typedef struct updn {
	unsigned num_roots;
	osm_opensm_t *p_osm;
	osm_subn_t *p_subn;
	unsigned num_nodes;
	updn_node_t **ranked_nodes;
} updn_t;

static inline char *direction_str(updn_switch_dir_t direction)
{
	return direction == UP ? "UP" : "DOWN";
}

static inline osm_subn_t *updn_get_subnet(IN updn_t * p_updn)
{
	return p_updn->p_subn;
}

/* This function returns direction based on rank and guid info of current &
   remote ports */
static updn_switch_dir_t updn_get_dir(unsigned cur_rank, unsigned rem_rank,
				      uint64_t cur_id, uint64_t rem_id,
				      uint8_t hops)
{
	/* HACK: comes to solve root nodes connection, in a classic subnet root nodes do not connect
	   directly, but in case they are we assign to root node an UP direction to allow UPDN to discover
	   the subnet correctly (and not from the point of view of the last root node).
	 */
	if (!cur_rank && !rem_rank)
		return UP;

	if (cur_rank < rem_rank) {
		/* allow one hop between "siblings" on a way down */
		if (hops > (rem_rank - cur_rank + 1))
			return UP;
		return DOWN;
	} else if (cur_rank > rem_rank) {
		return UP;
	} else {
		/* Equal rank, decide by id number, bigger == UP direction */
		if (cur_id > rem_id || hops > 1)
			return UP;
		else
			return DOWN;
	}
}

/**********************************************************************
 * This function does the bfs of min hop table calculation by guid index
 * as a starting point.
 **********************************************************************/
static int updn_bfs_by_node(IN osm_log_t * p_log, IN osm_subn_t * p_subn,
			    IN osm_switch_t * p_sw, uint32_t id)
{
	uint8_t pn, pn_rem;
	cl_qlist_t list;
	uint16_t lid;
	updn_switch_dir_t next_dir, current_dir;
	struct updn_node *u;
	OSM_LOG_ENTER(p_log);

	lid = osm_node_get_base_lid(p_sw->p_node, 0);
	lid = cl_ntoh16(lid);
	osm_switch_set_hops(p_sw, lid, 0, 0);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Starting from switch - port GUID 0x%" PRIx64 " lid %u\n",
		cl_ntoh64(p_sw->p_node->node_info.port_guid), lid);
	u =(struct updn_node *) (p_sw->priv);
	u = &u[id];
	u->dir = UP;

	/* Update list with the new element */
	cl_qlist_init(&list);
	cl_qlist_insert_tail(&list, &u->list);

	/* BFS the list till no next element */
	while (!cl_is_qlist_empty(&list)) {
		u = (struct updn_node *)cl_qlist_remove_head(&list);
		u->visited = 0;	/* cleanup */
		current_dir = u->dir;
		/* Go over all ports of the switch and find unvisited remote nodes */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			struct updn_node *rem_u;
			uint8_t current_min_hop, remote_min_hop,
			    set_hop_return_value;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn,
						     &pn_rem);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			/* Fetch remote guid only after validation of remote node */
			p_remote_sw = p_remote_node->sw;
			rem_u = (struct updn_node *) p_remote_sw->priv;
			rem_u = &rem_u[id];
			/* Decide which direction to mark it (UP/DOWN) */
			next_dir = updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->id, rem_u->id, 1);

			/* Check if this is a legal step : the only illegal step is going
			   from DOWN to UP */
			if ((current_dir == DOWN) && (next_dir == UP)) {
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Avoiding move from 0x%016" PRIx64
					" to 0x%016" PRIx64 "\n",
					cl_ntoh64(osm_node_get_node_guid(u->sw->p_node)),
					cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
				/* Illegal step */
				continue;
			}
			/* Set MinHop value for the current lid */
			current_min_hop = osm_switch_get_least_hops(u->sw, lid);
			/* Check hop count if better insert into list && update
			   the remote node Min Hop Table */
			remote_min_hop =
			    osm_switch_get_hop_count(p_remote_sw, lid, pn_rem);
			if (current_min_hop + 1 < remote_min_hop) {
				set_hop_return_value =
				    osm_switch_set_hops(p_remote_sw, lid,
							pn_rem,
							current_min_hop + 1);
				if (set_hop_return_value) {
					OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA01: "
						"Invalid value returned from set min hop is: %d\n",
						set_hop_return_value);
				}
				/* Check if remote port has already been visited */
				if (!rem_u->visited) {
					/* Insert updn_switch item into the list */
					rem_u->dir = next_dir;
					rem_u->visited = 1;
					cl_qlist_insert_tail(&list,
							     &rem_u->list);
				}
			}
		}
	}

	OSM_LOG_EXIT(p_log);
	return 0;
}

static void updn_bfs_by_node_par(IN void *context, uint32_t my_id)
{
	osm_ucast_mgr_t *p_mgr = context;
	cl_list_item_t * item;
	osm_switch_t * p_sw;
	cl_qlist_t * p_list;

	p_list = &p_mgr->sw_lists[my_id];

	while (cl_qlist_count(p_list)) {

		item = cl_qlist_remove_head(p_list);

		p_sw = cl_item_obj(item, p_sw, list_item);

		updn_bfs_by_node(p_mgr->p_log, p_mgr->p_subn, p_sw, my_id);
	}
}

/* NOTE : PLS check if we need to decide that the first */
/*        rank is a SWITCH for BFS purpose */
static int updn_subn_rank(IN updn_t * p_updn)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp, *p_remote_physp;
	cl_qlist_t list;
	cl_map_item_t *item;
	struct updn_node *u, *remote_u;
	uint8_t num_ports, port_num;
	osm_log_t *p_log = &p_updn->p_osm->log;
	unsigned max_rank = 0;
	osm_subn_t *p_subn;
	boolean_t is_leaf;

	OSM_LOG_ENTER(p_log);
	cl_qlist_init(&list);

	p_subn = updn_get_subnet(p_updn);

	/* add all roots to the list */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		u = p_sw->priv;
		if (!p_sw->rank)
			cl_qlist_insert_tail(&list, &u->list);
	}

	p_updn->num_nodes = 0;
	p_updn->ranked_nodes =
	    (updn_node_t**)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
				  sizeof(updn_node_t *));
	if (!p_updn->ranked_nodes) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA0D: Failed to allocate memory for ranked nodes"
			" array\n");
		return 1;
	}

	/* BFS the list till it's empty */
	while (!cl_is_qlist_empty(&list)) {
		u = (struct updn_node *)cl_qlist_remove_head(&list);
		p_updn->ranked_nodes[p_updn->num_nodes++] = u;
		is_leaf = TRUE;
		/* Go over all remote nodes and rank them (if not already visited) */
		p_sw = u->sw;
		num_ports = p_sw->num_ports;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Handling switch GUID 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		for (port_num = 1; port_num < num_ports; port_num++) {
			ib_net64_t port_guid;

			/* Current port fetched in order to get remote side */
			p_physp =
			    osm_node_get_physp_ptr(p_sw->p_node, port_num);

			if (!p_physp)
				continue;

			p_remote_physp = p_physp->p_remote_physp;

			/*
			   make sure that all the following occur on p_remote_physp:
			   1. The port isn't NULL
			   2. It is a switch
			 */
			if (!p_remote_physp)
				continue;

			if (p_remote_physp->p_node->sw) {
				remote_u = p_remote_physp->p_node->sw->priv;
				port_guid = p_remote_physp->port_guid;

				if (remote_u->sw->rank > u->sw->rank)
					is_leaf = FALSE;

				if (remote_u->sw->rank > u->sw->rank + 1) {
					/* Rank assigned to switch only once */
					OSM_ASSERT(remote_u->sw->rank ==
						   OSM_SW_NO_RANK);
					remote_u->sw->rank = u->sw->rank + 1;
					max_rank = remote_u->sw->rank;
					cl_qlist_insert_tail(&list,
							     &remote_u->list);
					OSM_LOG(p_log, OSM_LOG_DEBUG,
						"Rank of port GUID 0x%" PRIx64
						" = %u\n", cl_ntoh64(port_guid),
						remote_u->sw->rank);
				}
				continue;
			}

			if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_CA)
				continue;

			if (ib_mlnx_epi_get_is_special_port_bit(&p_remote_physp->ext_port_info))
				continue;

			u->num_hca++;
		}

		if (is_leaf) {
			osm_subn_add_leaf(p_subn, p_sw);
			u->is_leaf = TRUE;
		}
	}

	/* Print Summary of ranking */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Subnet ranking completed. Max Node Rank = %d\n", max_rank);
	OSM_LOG_EXIT(p_log);
	return 0;
}

/* hack: preserve min hops entries to any other root switches */
static void updn_clear_non_root_hops(updn_t * updn, osm_switch_t * sw)
{
	osm_port_t *port;
	unsigned i;

	for (i = 0; i < sw->num_hops; i++)
		if (sw->hops[i]) {
			port = osm_get_port_by_lid_ho(updn_get_subnet(updn), i);
			if (!port || !port->p_node->sw ||
			    port->p_node->sw->rank)
				memset(sw->hops[i], 0xff, sw->num_ports);
		}
}

static int updn_set_min_hop_table(IN updn_t * p_updn)
{
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	osm_switch_t *p_sw;
	cl_map_item_t *item;
	uint32_t number_of_threads = p_subn->opt.routing_threads_num;
	int i, res = 0;
	cl_pc_thread_pool_t pc_thread_pool;
	osm_ucast_mgr_t * p_mgr = &p_updn->p_osm->sm.ucast_mgr;

	OSM_LOG_ENTER(p_log);

	/* Go over all the switches in the subnet - for each init their Min Hop
	   Table */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches [\n");

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		/* Clear Min Hop Table */
		if (p_subn->opt.connect_roots)
			updn_clear_non_root_hops(p_updn, p_sw);
		else
			osm_switch_clear_hops(p_sw);
	}

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches ]\n");

	/* Now do the BFS for each port  in the subnet */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet [\n");

	if (number_of_threads != 1) {

		res = osm_ucast_mgr_pc_thread_pool_init(p_mgr,
							&pc_thread_pool,
							&updn_bfs_by_node_par,
							"updn",
							&number_of_threads);
		if (res != 0)
			goto Exit;
	}

	if (number_of_threads == 1) {
		for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     item = cl_qmap_next(item)) {
				p_sw = (osm_switch_t *)item;
				updn_bfs_by_node(p_log, p_subn, p_sw, 0);
		}
		res = 0;
		goto Exit;
	}
	/* Multithreaded part of the algorithm */

	/* create sw->priv array per thread */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		struct updn_node *u, *current;
		p_sw = (osm_switch_t *)item;
		current = (struct updn_node *) p_sw->priv;
		u = (struct updn_node *) malloc (
					sizeof(struct updn_node) * number_of_threads);
		if (u == NULL) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA11: "
				"Not enough memory for priv objects\n");
			res = -1;
			goto Exit;
		}
		for (i=0; i < number_of_threads; i++)
			memcpy(&u[i], current, sizeof(struct updn_node));
		/* change the pointer */
		p_sw->priv = u;
		free(current);
	}
	cl_pc_thread_pool_start(&pc_thread_pool);

	cl_pc_thread_pool_wait(&pc_thread_pool, EVENT_NO_TIMEOUT);
Exit:
	if (number_of_threads != 1) {
		cl_pc_thread_pool_destroy(&pc_thread_pool);
		free(p_mgr->sw_lists);
		p_mgr->sw_lists = NULL;
	}
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet ]\n");
	/* Cleanup */
	OSM_LOG_EXIT(p_log);
	return res;
}

static int updn_build_lid_matrices(IN updn_t * p_updn)
{
	int status;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Ranking all port guids in the list\n");
	if (!p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0A: "
			"No guids were provided or number of guids is 0\n");
		status = -1;
		goto _exit;
	}

	/* Check if it's not a switched subnet */
	if (cl_is_qmap_empty(&p_subn->sw_guid_tbl)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0B: "
			"This is not a switched subnet, cannot perform UPDN algorithm\n");
		status = -1;
		goto _exit;
	}

	/* Rank the subnet switches */
	if (updn_subn_rank(p_updn)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0E: "
			"Failed to assign ranks\n");
		status = -1;
		goto _exit;
	}

	/* After multiple ranking need to set Min Hop Table by UpDn algorithm  */
	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Setting all switches' Min Hop Table\n");
	status = updn_set_min_hop_table(p_updn);

_exit:
	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return status;
}

static struct updn_node *create_updn_node(osm_switch_t * sw)
{
	struct updn_node *u;

	u = malloc(sizeof(*u));
	if (!u)
		return NULL;
	memset(u, 0, sizeof(*u));
	u->sw = sw;
	u->id = cl_ntoh64(osm_node_get_node_guid(sw->p_node));
	u->is_leaf = FALSE;
	u->num_hca = 0;
	u->dlid_routes = NULL;
	u->dlid_sw_routes = NULL;
	u->total_routes = 0;
	sw->rank = OSM_SW_NO_RANK;
	return u;
}

static void delete_updn_node(struct updn_node *u)
{
	u->sw->priv = NULL;
	free(u->dlid_routes);
	free(u->dlid_sw_routes);
	free(u);
}

static void updn_delete_nodes(updn_t *p_updn)
{
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	cl_map_item_t *item;
	osm_switch_t *p_sw;

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		delete_updn_node(p_sw->priv);
	}

	free(p_updn->ranked_nodes);
	p_updn->ranked_nodes = NULL;
}

/* Find Root nodes automatically by Min Hop Table info */
static void updn_find_root_nodes_by_min_hop(OUT updn_t * p_updn)
{
	osm_opensm_t *p_osm = p_updn->p_osm;
	osm_switch_t *p_sw;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	cl_map_item_t *item;
	double thd1, thd2;
	unsigned i, cas_num = 0;
	unsigned *cas_per_sw;
	uint16_t lid_ho;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Current number of ports in the subnet is %d\n",
		cl_qmap_count(&p_subn->port_guid_tbl));

	lid_ho = p_subn->max_lid + 1;
	cas_per_sw = malloc(lid_ho * sizeof(*cas_per_sw));
	if (!cas_per_sw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR AA14: "
			"cannot alloc mem for CAs per switch counter array\n");
		goto _exit;
	}
	memset(cas_per_sw, 0, lid_ho * sizeof(*cas_per_sw));

	/* Find the Maximum number of CAs (and routers) for histogram normalization */
	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Finding the number of CAs and storing them in cl_map\n");
	for (item = cl_qmap_head(&p_subn->port_guid_tbl);
	     item != cl_qmap_end(&p_subn->port_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_port = (osm_port_t *)item;
		if (!p_port->p_node->sw) {
			p_physp = p_port->p_physp->p_remote_physp;
			if (!p_physp || !p_physp->p_node->sw ||
			    ib_mlnx_epi_get_is_special_port_bit(&p_physp->ext_port_info))
				continue;
			lid_ho = osm_node_get_base_lid(p_physp->p_node, 0);
			lid_ho = cl_ntoh16(lid_ho);
			cas_per_sw[lid_ho]++;
			cas_num++;
		}
	}

	thd1 = cas_num * 0.9;
	thd2 = cas_num * 0.05;
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Found %u CAs and RTRs, %u SWs in the subnet. "
		"Thresholds are thd1 = %f && thd2 = %f\n",
		cas_num, cl_qmap_count(&p_subn->sw_guid_tbl), thd1, thd2);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Passing through all switches to collect Min Hop info\n");
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		unsigned hop_hist[IB_SUBNET_PATH_HOPS_MAX];
		uint16_t max_lid_ho;
		uint8_t hop_val;
		uint16_t numHopBarsOverThd1 = 0;
		uint16_t numHopBarsOverThd2 = 0;

		p_sw = (osm_switch_t *) item;

		memset(hop_hist, 0, sizeof(hop_hist));

		max_lid_ho = p_sw->max_lid_ho;
		for (lid_ho = 1; lid_ho <= max_lid_ho; lid_ho++)
			if (cas_per_sw[lid_ho]) {
				hop_val =
				    osm_switch_get_least_hops(p_sw, lid_ho);
				if (hop_val >= IB_SUBNET_PATH_HOPS_MAX)
					continue;

				hop_hist[hop_val] += cas_per_sw[lid_ho];
			}

		/* Now recognize the spines by requiring one bar to be
		   above 90% of the number of CAs and RTRs */
		for (i = 0; i < IB_SUBNET_PATH_HOPS_MAX; i++) {
			if (hop_hist[i] > thd1)
				numHopBarsOverThd1++;
			if (hop_hist[i] > thd2)
				numHopBarsOverThd2++;
		}

		/* If thd conditions are valid - rank the root node */
		if (numHopBarsOverThd1 == 1 && numHopBarsOverThd2 == 1) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Ranking GUID 0x%" PRIx64 " as root node\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			p_sw->rank = 0;
			p_updn->num_roots++;
		}
	}

	free(cas_per_sw);
_exit:
	OSM_LOG_EXIT(&p_osm->log);
	return;
}

static void dump_roots(cl_map_item_t *item, FILE *file, void *cxt)
{
	osm_switch_t *sw = (osm_switch_t *)item;
	if (!sw->rank)
		fprintf(file, "0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(sw->p_node)));
}

static int update_id(void *cxt, uint64_t guid, char *p)
{
	updn_t *p_updn = cxt;
	osm_opensm_t *osm = p_updn->p_osm;
	osm_switch_t *sw;
	uint64_t id;
	char *e;

	sw = osm_get_switch_by_guid(updn_get_subnet(p_updn), cl_hton64(guid));
	if (!sw) {
		OSM_LOG(&osm->log, OSM_LOG_VERBOSE,
			"switch with guid 0x%" PRIx64 " is not found\n", guid);
		return 0;
	}

	id = strtoull(p, &e, 0);
	if (*e && !isspace(*e)) {
		OSM_LOG(&osm->log, OSM_LOG_ERROR,
			"ERR AA05: cannot parse node id \'%s\'", p);
		return -1;
	}

	OSM_LOG(&osm->log, OSM_LOG_DEBUG,
		"update node 0x%" PRIx64 " id to 0x%" PRIx64 "\n", guid, id);

	((struct updn_node *)sw->priv)->id = id;

	return 0;
}

static int rank_root_node(void *cxt, uint64_t guid, char *p)
{
	updn_t *updn = cxt;
	osm_switch_t *sw;

	sw = osm_get_switch_by_guid(updn_get_subnet(updn),
				    cl_hton64(guid));
	if (!sw) {
		OSM_LOG(&updn->p_osm->log, OSM_LOG_ERROR, "ERR AA04: "
			"Root switch with guid 0x%" PRIx64 " is not found\n", guid);
		return 0;
	}

	OSM_LOG(&updn->p_osm->log, OSM_LOG_DEBUG,
		"Ranking root port GUID 0x%" PRIx64 "\n", guid);

	sw->rank = 0;
	updn->num_roots++;

	return 0;
}

/* UPDN callback function */
static int updn_lid_matrices(void *ctx)
{
	updn_t *p_updn = ctx;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	int ret = 0;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	char root_file_path[1024];

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_sw->priv = create_updn_node(p_sw);
		if (!p_sw->priv) {
			OSM_LOG(&(p_updn->p_osm->log), OSM_LOG_ERROR, "ERR AA0C: "
				"cannot create updn node\n");
			OSM_LOG_EXIT(&p_updn->p_osm->log);
			return -1;
		}
	}

	/* First setup root nodes */
	p_updn->num_roots = 0;

	if (p_subn->opt.root_guid_file) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"UPDN - Fetching root nodes from file \'%s\'\n",
			p_subn->opt.root_guid_file);

		ret = parse_node_map_str(p_subn->opt.root_guid_file,
					 osm_apply_func_on_port_group_nodes, rank_root_node,
					 p_updn, p_updn->p_osm, TRUE);
		if (ret) {
			OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA02: "
				"cannot parse root guids file \'%s\'\n",
				p_subn->opt.root_guid_file);
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);
			updn_find_root_nodes_by_min_hop(p_updn);
		} else if (p_subn->opt.connect_roots &&
			   p_updn->num_roots > 1)
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);
	} else {
		osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);
		updn_find_root_nodes_by_min_hop(p_updn);
	}

	if (p_subn->opt.ids_guid_file) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"UPDN - update node ids from file \'%s\'\n",
			p_subn->opt.ids_guid_file);

		ret = parse_node_map(p_subn->opt.ids_guid_file,
				     update_id, p_updn);
		if (ret)
			OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA03: "
				"cannot parse node ids file \'%s\'\n",
				p_subn->opt.ids_guid_file);
	}

	/* Only if there are assigned root nodes do the algorithm, otherwise perform do nothing */
	if (p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"activating UPDN algorithm\n");
		ret = updn_build_lid_matrices(p_updn);
	} else {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_INFO,
			"disabling UPDN algorithm, no root nodes were found\n");
		ret = -1;
	}

	if (OSM_LOG_IS_ACTIVE_V2(&p_updn->p_osm->log, OSM_LOG_ROUTING)) {
		snprintf(root_file_path, sizeof(root_file_path), "%s/opensm-updn-roots.dump",
			 p_updn->p_osm->subn.opt.dump_files_dir);
		osm_dump_qmap_to_file(p_updn->p_osm, root_file_path,
				      &p_subn->sw_guid_tbl,
				      dump_roots, NULL);
	}

	if (ret)
		updn_delete_nodes(p_updn);

	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return ret;
}

/*
 * When converge_routes is TRUE, an adjacent switch with higher value of dlid_routes
 * should be selected. Otherwise an adjacent switch with lower value of dlid_routes
 * should be chosen
*/

inline static boolean_t updn_lid_tracking_compare_dlid_routes(
					 uint32_t curr_dlid_routes,
					 uint32_t adj_sw_dlid_routes,
					 boolean_t converge_routes)
{
	return (converge_routes) ?
		(curr_dlid_routes < adj_sw_dlid_routes):
		(curr_dlid_routes > adj_sw_dlid_routes);
}

static int updn_lid_tracking_process_switch(updn_t *p_updn, struct updn_node *u,
					    enum updn_switch_dir direction,
					    uint8_t path_bit)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	cl_qlist_t *port_list;
	cl_list_item_t *list_item;
	osm_switch_t *p_sw = u->sw;
	osm_port_t *p_dest_port;
	osm_physp_t *p_dest_physp, *p_dest_sw_physp = NULL, *p_exit_physp;
	osm_node_t *p_adj_node;
	osm_switch_t *p_dest_sw, *p_adj_sw;
	updn_node_t *exit_node, *adj_node;
	boolean_t is_lmc_routing;
	uint32_t weights[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint32_t dlid_routes;
	uint32_t dlid_sw_routes;
	uint32_t total_routes;
	uint16_t dest_sw_lid, dest_port_base_lid, dest_port_lid, num_routes;
	uint8_t least_hops, port, offset, p, num_ports = p_sw->num_ports;
	uint8_t exit_port;

	OSM_ASSERT(p_sw->rank != OSM_SW_NO_RANK);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing switch 0x%" PRIx64 " direction %s\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
		direction_str(direction));

	is_lmc_routing = p_subn->path_bit == OSM_SUBNET_DEFAULT_PATH_BITS &&
		       path_bit;

	port_list = &p_mgr->port_order_lists[0];
	for (list_item = cl_qlist_head(port_list);
	     list_item != cl_qlist_end(port_list);
	     list_item = cl_qlist_next(list_item)) {

		exit_port = OSM_NO_PATH;
		exit_node = NULL;
		dlid_sw_routes = 0;
		dlid_routes = 0;
		total_routes = 0;

		p_dest_port = cl_item_obj(list_item, p_dest_port, list_item);
		p_dest_physp = osm_port_get_physp(p_dest_port);

		if (path_bit >= (1 << osm_physp_get_lmc(p_dest_physp)))
			continue;

		dest_port_base_lid = cl_ntoh16(p_dest_port->lid);
		if (!dest_port_base_lid)
			continue;

		dest_port_lid = dest_port_base_lid + path_bit;

		/* Destination is this switch */
		if (p_dest_port->p_node == p_sw->p_node) {
			if (direction == UP) {
				p_sw->new_lft[dest_port_lid] = 0;
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Routing LID %u to port 0 for switch"
					" 0x%" PRIx64 "\n",
					dest_port_lid,
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			}
			continue;
		}

		/* Find switch connected to destination port */
		if (p_dest_port->p_node->sw) {
			p_dest_sw = p_dest_port->p_node->sw;
		} else {
			p_dest_sw_physp = osm_physp_get_remote(p_dest_physp);
			if (!p_dest_sw_physp)
				/* Isolated node ... */
				continue;
			p_dest_sw = osm_physp_get_node_ptr(p_dest_sw_physp)->sw;
			OSM_ASSERT(p_dest_sw);
		}

		/* Destination port connected directly to this switch */
		if (p_dest_sw == p_sw) {
			if (direction == UP) {
				OSM_ASSERT(p_dest_sw_physp);
				p_sw->new_lft[dest_port_lid] =
				    osm_physp_get_port_num(p_dest_sw_physp);
			}
			continue;
		}

		if (p_dest_sw->rank == OSM_SW_NO_RANK)
			/* Unranked switch ... */
			continue;

		dest_sw_lid = cl_ntoh16(osm_node_get_base_lid(p_dest_sw->p_node, 0));
		least_hops = osm_switch_get_least_hops(p_sw, dest_sw_lid);
		if (least_hops == OSM_NO_PATH) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"No path to get to LID %u from switch"
				" 0x%" PRIx64 "\n",
				dest_port_lid,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		/* Validate destination direction */
		if (direction !=
		    updn_get_dir(p_sw->rank, p_dest_sw->rank,
				 u->id,
				 ((updn_node_t *)p_dest_sw->priv)->id,
				 least_hops))
			continue;

		/*
		 * When routing for LMC, start port search after last exit port
		 * that was used for this destination
		 */
		offset = (is_lmc_routing ? p_sw->new_lft[dest_port_lid - 1] : 0) +
			 1;
		OSM_ASSERT(offset != OSM_NO_PATH);
		for (p = offset; p < num_ports + offset; p++) {
			port = osm_switch_get_dimn_port(p_sw, p % num_ports);
			if (!port)
				continue;

			/* Check if port leads to shortest route */
			if (osm_switch_get_hop_count(p_sw, dest_sw_lid, port) !=
						     least_hops)
				continue;

			/* Check if port is connected and healthy */
			p_exit_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_exit_physp || !osm_physp_is_healthy(p_exit_physp) ||
			    !p_exit_physp->p_remote_physp)
				continue;

			/* Get remote switch */
			p_adj_node = osm_node_get_remote_node(p_sw->p_node,
							      port, NULL);
			OSM_ASSERT(p_adj_node && p_adj_node->sw);
			p_adj_sw = p_adj_node->sw;
			adj_node = (updn_node_t *)p_adj_sw->priv;

			/* Validate port direction.
			 * Don't skip the port in case the link
			 * is between "siblings".
			 */
			if (p_sw->rank != p_adj_sw->rank &&
			    direction !=
			    updn_get_dir(p_sw->rank, p_adj_sw->rank,
					 u->id, adj_node->id, 1))
				continue;

			/*
			 * Exit port selection criterias by precedence:
			 * 1. Exit port that leads to switch with maximal/minimal
			 * number of routes to destination port.
			 *  (depends on opt.updn_lid_tracking_converge_routes)
			 * 2. Exit port that leads to a switch with least routes
			 *    that goes through the switch connected to
			 *    destination port.
			 * 3. Exit port with with least routes.
			 * 4. Exit port that leads to switch that used for
			 *    lesser number of routes.
			 * Criterias 2,3,4 are ignored when routing for path
			 * bit != 0.
			 */
			if (exit_port == OSM_NO_PATH ||
			    updn_lid_tracking_compare_dlid_routes(
				dlid_routes,
				adj_node->dlid_routes[dest_port_lid],
				p_subn->opt.updn_lid_tracking_converge_routes) ||
			    (!is_lmc_routing &&
			     dlid_routes == adj_node->dlid_routes[dest_port_lid] &&
			     (dlid_sw_routes > adj_node->dlid_sw_routes[dest_sw_lid] ||
			      (dlid_sw_routes == adj_node->dlid_sw_routes[dest_sw_lid] &&
			       (weights[exit_port] > weights[port] ||
			        (weights[exit_port] == weights[port] &&
				 total_routes > adj_node->total_routes)))))) {
				exit_port = port;
				exit_node = adj_node;
				total_routes = adj_node->total_routes;
				dlid_sw_routes = adj_node->dlid_sw_routes[dest_sw_lid];
				dlid_routes = adj_node->dlid_routes[dest_port_lid];
			}
		}

		if (exit_port == OSM_NO_PATH) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR AA20: Failed to find %s path to get to LID %u"
				" from switch 0x%" PRIx64 "\n",
				direction == UP ? "up" : "down",
				dest_port_lid,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Routing LID %u to port %u for switch 0x%" PRIx64 "\n",
			dest_port_lid, exit_port,
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));

		p_sw->new_lft[dest_port_lid] = exit_port;
		/* Update weights */
		num_routes = u->dlid_routes[dest_port_lid] +
			     (u->num_hca ? u->num_hca : !!u->is_leaf);

		if (num_routes) {
			exit_node->dlid_routes[dest_port_lid] += num_routes;
			if (p_dest_port->p_node->sw) {
				continue;
			}
			exit_node->dlid_sw_routes[dest_sw_lid] += num_routes;
			exit_node->total_routes += num_routes;
			weights[exit_port] += 1;
		}
	}

	return 0;
}


static int updn_node_rank_cmp_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;

	return (updn_get_dir(p_left->sw->rank, p_right->sw->rank,
			     p_left->id, p_right->id, 1) == DOWN ? -1 : 1);
}

static int updn_lid_tracking_build_lft_tables(updn_t *p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	struct updn_node *u;
	int i, result = 0;
	uint16_t num_lids;
	uint8_t path_bit, min_path_bit, max_path_bit;

	OSM_LOG_ENTER(p_log);

	/* reset new lfts from previous routing */
	osm_ucast_mgr_reset_lfts_all_switches(p_subn);

	if (osm_ucast_mgr_init_port_order_lists(p_mgr, 1)) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR AA08: Failed to initialize port order list\n");
		return 1;
	}

	/* Initialize DLID arrays */
	for (i = 0; i < p_updn->num_nodes; i++) {
		u = p_updn->ranked_nodes[i];
		num_lids = u->sw->max_lid_ho + 1;

		u->dlid_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));
		u->dlid_sw_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));

		if (!u->dlid_routes || !u->dlid_sw_routes) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR AA09: Failed to allocates destination LIDs"
				" array\n");
			result = -1;
			goto exit;
		}

		u->total_routes = 0;
	}

	qsort(p_updn->ranked_nodes, p_updn->num_nodes, sizeof(updn_node_t *),
	      updn_node_rank_cmp_func);

	if (p_subn->path_bit != OSM_SUBNET_DEFAULT_PATH_BITS) {
		min_path_bit = p_subn->path_bit;
		max_path_bit = p_subn->path_bit;
	} else {
		min_path_bit = 0;
		max_path_bit = (1 << p_subn->opt.lmc) - 1;
	}

	for (path_bit = min_path_bit; path_bit <= max_path_bit; path_bit++) {

		/* From leafs to spines */
		for (i = p_updn->num_nodes - 1; i >= 0; i--) {
			u = p_updn->ranked_nodes[i];
			updn_lid_tracking_process_switch(p_updn, u, UP,
							 path_bit);
		}

		/* Clear switch weights */
		for (i = 0; i < p_updn->num_nodes; i++) {
			u = p_updn->ranked_nodes[i];
			u->total_routes = 0;
		}

		/* From spines to leafs */
		for (i = 0; i < p_updn->num_nodes; i++) {
			u = p_updn->ranked_nodes[i];
			updn_lid_tracking_process_switch(p_updn, u, DOWN,
							 path_bit);
		}
	}

exit:
	osm_ucast_mgr_release_port_order_lists(p_mgr, 1);

	OSM_LOG_EXIT(p_log);

	return result;
}

static void updn_calculate_missing_routes(updn_t * p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_log_t *p_log = &p_updn->p_osm->log;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(p_log);

	if (p_updn->num_roots > 1 &&
	    p_subn->opt.calc_missing_routes &&
	    osm_ucast_mgr_calculate_missing_routes(p_mgr)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA06: leaf switches were not set properly, "
			"no missing LID routes will be calculated\n");
	}

	osm_subn_clear_leafs(p_subn);

	OSM_LOG_EXIT(p_log);
}

/* UPDN callback function */
static int updn_fwd_tables(void * ctx)
{
	updn_t *p_updn = ctx;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	int ret = 0;

	OSM_LOG_ENTER(p_log);

	if (p_subn->opt.updn_lid_tracking_mode) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Up/Down routing with LID tracking\n");
		ret = updn_lid_tracking_build_lft_tables(p_updn);
	} else {
		ret = osm_ucast_mgr_build_lft_tables(&p_updn->p_osm->sm.ucast_mgr);
	}

	if (ret < 0)
		goto Exit;

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Connecting switches that are unreachable within "
		"Up/Down rules\n");
	updn_calculate_missing_routes(p_updn);

Exit:
	updn_delete_nodes(p_updn);

	OSM_LOG_EXIT(p_log);
	return ret;
}

static void updn_delete(void *context)
{
	free(context);
}

static ib_api_status_t updn_mcast_stree(void *context,
					IN OUT struct osm_mgrp_box *mgb)
{
	updn_t *p_updn = (updn_t *)context;
	return osm_mcast_mgr_build_clf_stree(updn_get_subnet(p_updn), mgb);
}

int osm_ucast_updn_setup(struct osm_routing_engine *r, osm_opensm_t *osm)
{
	updn_t *updn;

	updn = malloc(sizeof(updn_t));
	if (!updn)
		return -1;
	memset(updn, 0, sizeof(updn_t));

	updn->p_osm = osm;
	updn->p_subn = osm->sm.ucast_mgr.p_subn;

	r->context = updn;
	r->destroy = updn_delete;
	r->build_lid_matrices = updn_lid_matrices;
	r->ucast_build_fwd_tables = updn_fwd_tables;
	r->mcast_build_stree = updn_mcast_stree;

	return 0;
}

int osm_ucast_ar_updn_setup(struct osm_routing_engine *r,
			    osm_opensm_t *osm)
{
	int rc;

	rc = osm_ucast_updn_setup(r, osm);

	if (rc)
		return rc;

	/* set suitable auto AR parameters */
	r->ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->shield_mode_auto = OSM_ADV_ROUTING_MODE_RN;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_AR_TREE;

	return 0;
}
