/*
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UNHEALTHY_PORTS_C
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_node.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <complib/cl_timer.h>
#include <complib/cl_qlist.h>
#include <complib/cl_event_wheel.h>
#include <complib/cl_math.h>


const char *osm_hm_cond_str[] = {"REBOOT", "FLAPPING",
				 "UNRESPONSIVE", "NOISY",
				 "SET_ERR", "ILLEGAL",
				 "MANUAL"};

struct opensm_dump_context {
	osm_sm_t *p_osm;
	FILE *file;
};

static int
opensm_dump_to_file(osm_sm_t * p_osm, const char *file_name,
		    void (*dump_func) (osm_sm_t * p_osm, FILE * file))
{
	char orig[1024], tmp[1024];
	FILE *file;
	int fd;

	snprintf(orig, sizeof(orig), "%s/%s",
		 p_osm->p_subn->opt.dump_files_dir, file_name);

	snprintf(tmp, sizeof(tmp), "%s/%sXXXXXX",
		 p_osm->p_subn->opt.dump_files_dir, file_name);

	mode_t mask = umask(S_IRUSR | S_IWUSR);
	fd = mkstemp(tmp);
	umask(mask);
	if (fd < 0) {
		OSM_LOG(p_osm->p_log, OSM_LOG_ERROR, "ERR C201: "
			"cannot open tmp file \"%s\" : %s\n", tmp,
			strerror(errno));
		return -1;
	}

	file = fdopen(fd, "w");
	if (!file) {
		OSM_LOG(p_osm->p_log, OSM_LOG_ERROR, "ERR C202: "
			"cannot open tmp file \"%s\" : %s\n", tmp,
			strerror(errno));
		return -1;
	}

	dump_func(p_osm, file);
	fclose(file);

	if(rename(tmp, orig)){
		OSM_LOG(p_osm->p_log, OSM_LOG_ERROR, "ERR C206:"
			"cannot rename tmp file \"%s\" to file \"%s\" : %s\n",
			tmp, orig, strerror(errno));
		return -1;
	}

	if (chmod(orig, S_IRUSR | S_IWUSR)){
		OSM_LOG(p_osm->p_log, OSM_LOG_ERROR, "ERR C207:"
	                "cannot change mode of orig file \"%s\" : %s\n", orig,
	                strerror(errno));
		return -1;
	}

	return 0;
}

static inline uint64_t ptr_to_key(void const *p)
{
	uint64_t k = 0;

	memcpy(&k, &p, sizeof(void *));
	return k;
}

static inline void *key_to_ptr(uint64_t k)
{
	void *p = 0;

	memcpy(&p, &k, sizeof(void *));
	return p;
}

void osm_unhealthy_port_remove(IN osm_sm_t * sm,
			       IN OUT osm_unhealthy_port_t **pp_unhealthy_port)
{
	/* Clear event_wheel object for the unhealthy port */
	cl_event_wheel_unreg(&sm->reboot_tracker,
			     ptr_to_key(*pp_unhealthy_port));

	osm_unhealthy_port_delete(pp_unhealthy_port);
}

osm_unhealthy_port_t *osm_unhealthy_port_new(IN ib_net64_t guid,
					     IN uint8_t port_num,
					     IN char *node_desc,
					     IN uint8_t node_type,
					     IN ib_net64_t peer_guid,
					     IN uint8_t peer_port_num,
					     IN char *peer_node_desc)

{

	osm_unhealthy_port_t *p_unhealthy_port =
	    calloc(1, sizeof(osm_unhealthy_port_t));
	if (!p_unhealthy_port)
		return NULL;

	p_unhealthy_port->answered_last_ls = FALSE;
	p_unhealthy_port->p_obj = p_unhealthy_port;
	p_unhealthy_port->guid = guid;
	p_unhealthy_port->port_num = port_num;
	p_unhealthy_port->peer_guid = peer_guid;
	p_unhealthy_port->peer_port_num = peer_port_num;
	if (node_desc)
		p_unhealthy_port->node_desc = strdup(node_desc);
	p_unhealthy_port->node_type = node_type;
	if (peer_node_desc)
		p_unhealthy_port->peer_node_desc = strdup(peer_node_desc);

	return p_unhealthy_port;
}

void osm_unhealthy_port_delete(IN OUT osm_unhealthy_port_t **pp_unhealthy_port)
{
	free((*pp_unhealthy_port)->node_desc);
	free((*pp_unhealthy_port)->peer_node_desc);
	free(*pp_unhealthy_port);
	*pp_unhealthy_port = NULL;
}

static osm_hm_action_type_enum fill_cond_buf(osm_sm_t * p_sm,
					     osm_unhealthy_port_t *p_unhealthy_port,
					     osm_hm_cond_type_enum cond)
{
	osm_subn_opt_t *p_opt = &p_sm->p_subn->opt;

	switch(cond) {
	case OSM_HM_COND_SET_ERR:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_seterr_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_seterr_action);
		else
			return MAX(p_opt->hm_sw_seterr_action,
				   p_opt->hm_ca_seterr_action);
		break;
	case OSM_HM_COND_REBOOT:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_reboot_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_reboot_action);
		else
			return MAX(p_opt->hm_sw_reboot_action,
				   p_opt->hm_ca_reboot_action);
		break;
	case OSM_HM_COND_FLAPPING:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_flapping_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_flapping_action);
		else
			return MAX(p_opt->hm_sw_flapping_action,
				   p_opt->hm_ca_flapping_action);
		break;
	case OSM_HM_COND_UNRESPONSIVE:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_unresponsive_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_unresponsive_action);
		else
			return MAX(p_opt->hm_sw_unresponsive_action,
				   p_opt->hm_ca_unresponsive_action);
		break;
	case OSM_HM_COND_NOISY:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_noisy_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_noisy_action);
		else
			return MAX(p_opt->hm_sw_noisy_action,
				   p_opt->hm_ca_noisy_action);
		break;
	case OSM_HM_COND_ILLEGAL:
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_illegal_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_illegal_action);
		else
			return MAX(p_opt->hm_sw_illegal_action,
				   p_opt->hm_ca_illegal_action);
		break;
	case OSM_HM_COND_MANUAL:
		if (p_unhealthy_port->manual_action != OSM_HM_ACTION_IGNORE)
			return p_unhealthy_port->manual_action;
		if (p_unhealthy_port->node_type == IB_NODE_TYPE_SWITCH)
			return (p_opt->hm_sw_manual_action);
		else if (p_unhealthy_port->node_type == IB_NODE_TYPE_CA ||
			 p_unhealthy_port->node_type == IB_NODE_TYPE_ROUTER)
			return (p_opt->hm_ca_manual_action);
		else
			return MAX(p_opt->hm_sw_manual_action,
				   p_opt->hm_ca_manual_action);
		break;
	default:
		break;
	}

	return OSM_HM_ACTION_IGNORE;
}

static void sm_hm_set_action(IN osm_sm_t * sm,
			     IN osm_unhealthy_port_t * p_unhealthy_port,
			     IN osm_hm_cond_type_enum cond)
{
	osm_hm_action_type_enum action;

	action = fill_cond_buf(sm, p_unhealthy_port, cond);
	switch (action) {
	case OSM_HM_ACTION_ISOLATE:
		if (!p_unhealthy_port->isolate) {
			p_unhealthy_port->isolate = TRUE;
		}
		break;
	case OSM_HM_ACTION_NO_DISCOVER:
		if (!p_unhealthy_port->no_discover) {
			p_unhealthy_port->no_discover = TRUE;
		}
		break;
	case OSM_HM_ACTION_IGNORE:
		p_unhealthy_port->ignore = TRUE;
		break;
	case OSM_HM_ACTION_DISABLE:
		p_unhealthy_port->disable = TRUE;
		break;
	case OSM_HM_ACTION_REPORT:
		p_unhealthy_port->report = TRUE;
		break;
	default:
		break;
	}
}

void osm_hm_update_actions(IN osm_sm_t * sm,
			   IN osm_unhealthy_port_t * p_unhealthy_port)
{
	osm_hm_cond_type_enum cond;

	p_unhealthy_port->isolate = FALSE;
	p_unhealthy_port->no_discover = FALSE;
	p_unhealthy_port->ignore = FALSE;
	p_unhealthy_port->disable = FALSE;
	p_unhealthy_port->report = FALSE;
	for (cond = OSM_HM_COND_REBOOT; cond < OSM_HM_COND_MAX; cond++) {
		if (osm_hm_is_cond_set(p_unhealthy_port, cond))
			sm_hm_set_action(sm, p_unhealthy_port, cond);
	}
}

void osm_hm_set_cond(IN struct osm_sm *sm,
		     IN osm_unhealthy_port_t * p_unhealthy_port,
		     IN osm_hm_cond_type_enum cond)
{
	uint32_t num_reboots;
	boolean_t err = FALSE;

	switch (cond) {
	case OSM_HM_COND_SET_ERR:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->set_err_hist |= 0x1;
			if (__builtin_popcount(p_unhealthy_port->set_err_hist) >=
			    sm->p_subn->opt.hm_num_set_err_sweeps) {
				p_unhealthy_port->cond_mask |=
				    1 << OSM_HM_COND_SET_ERR;
				err = TRUE;
			}
		}
		break;
	case OSM_HM_COND_REBOOT:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			cl_event_wheel_wnd_reg(&sm->reboot_tracker,
					       ptr_to_key(p_unhealthy_port),
					       cl_get_time_stamp() +
					       sm->p_subn->opt.hm_reboots_period_secs * 1000000,
					       NULL, NULL);
			num_reboots = cl_event_wheel_num_regs(&sm->reboot_tracker,
							      ptr_to_key(p_unhealthy_port));
			if (num_reboots >= sm->p_subn->opt.hm_num_reboots) {
				p_unhealthy_port->cond_mask |=
				    1 << OSM_HM_COND_REBOOT;
				err = TRUE;
			}
		}
		break;
	case OSM_HM_COND_FLAPPING:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->flapping_hist |= 0x1;
			if (__builtin_popcount(p_unhealthy_port->flapping_hist) >=
			     sm->p_subn->opt.hm_num_flapping_sweeps) {
				p_unhealthy_port->cond_mask |=
				    1 << OSM_HM_COND_FLAPPING;
				err = TRUE;
			}
		}
		break;
	case OSM_HM_COND_UNRESPONSIVE:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->no_resp_hist |= 0x1;
			if (__builtin_popcount(p_unhealthy_port->no_resp_hist) >=
			    sm->p_subn->opt.hm_num_no_resp_sweeps) {
				p_unhealthy_port->cond_mask |=
				    1 << OSM_HM_COND_UNRESPONSIVE;
				err = TRUE;
			}
		}
		break;
	case OSM_HM_COND_NOISY:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->cond_mask |= 1 << OSM_HM_COND_NOISY;
			err = TRUE;
		}
		break;
	case OSM_HM_COND_ILLEGAL:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->illegal_count++;
			if (p_unhealthy_port->illegal_count >=
			    sm->p_subn->opt.hm_num_illegal) {
				p_unhealthy_port->cond_mask |= 1 << OSM_HM_COND_ILLEGAL;
				err = TRUE;
			}
		}
		break;
	case OSM_HM_COND_MANUAL:
		if (!osm_hm_is_cond_set(p_unhealthy_port, cond)) {
			p_unhealthy_port->cond_mask |= 1 << OSM_HM_COND_MANUAL;
			err = TRUE;
		}
		break;
	default:
		break;
	}
	if (err) {
		if (p_unhealthy_port->timestamp.tv_sec == 0 &&
		    p_unhealthy_port->timestamp.tv_usec == 0)
			gettimeofday(&p_unhealthy_port->timestamp, NULL);
		sm->hm_dirty = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Bad port found: 0x%016"
			PRIx64 "[%d] \"%s\" peer of 0x%016" PRIx64 "[%d] "
			"\"%s\" condition: \"%s\"\n",
			cl_ntoh64(p_unhealthy_port->guid),
			p_unhealthy_port->port_num,
			p_unhealthy_port->node_desc,
			cl_ntoh64(p_unhealthy_port->peer_guid),
			p_unhealthy_port->peer_port_num,
			p_unhealthy_port->peer_node_desc,
			osm_hm_cond_str[cond]);

		sm_hm_set_action(sm, p_unhealthy_port, cond);
	} else
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Found port in bad state: "
			"0x%016" PRIx64 "[%d] \"%s\" peer of 0x%016" PRIx64
			"[%d] \"%s\" condition: %s set_err_hist: %d "
			"flapping_hist: %d reboot_num: %d no_resp_hist: %d\n",
			cl_ntoh64(p_unhealthy_port->guid),
			p_unhealthy_port->port_num,
			p_unhealthy_port->node_desc,
			cl_ntoh64(p_unhealthy_port->peer_guid),
			p_unhealthy_port->peer_port_num,
			p_unhealthy_port->peer_node_desc,
			osm_hm_cond_str[cond], p_unhealthy_port->set_err_hist,
			p_unhealthy_port->flapping_hist,
			cl_event_wheel_num_regs(&sm->reboot_tracker,
						ptr_to_key(p_unhealthy_port)),
			p_unhealthy_port->no_resp_hist);
}

static void sm_dump_one_unhealthy_port(cl_fmap_item_t * p_item, void *cxt)
{
	struct opensm_dump_context *dump_cxt = cxt;
	FILE *file = dump_cxt->file;
	osm_unhealthy_port_t *p_unhealthy_port = (osm_unhealthy_port_t *) p_item;
	osm_hm_cond_type_enum cond;
	char cond_buf[256] = { 0 };
	char *p;
	boolean_t is_unhealthy = FALSE;

	p = cond_buf;
	for (cond = OSM_HM_COND_REBOOT; cond < OSM_HM_COND_MAX; cond++) {
		if (((1 << cond) & p_unhealthy_port->cond_mask) &&
		    ((cond == OSM_HM_COND_MANUAL) ||
		     (fill_cond_buf(dump_cxt->p_osm, p_unhealthy_port, cond) >
		      OSM_HM_ACTION_IGNORE))) {
			if (is_unhealthy) {
				*p = ',';
				p++;
			}
			p = strncpy(p, osm_hm_cond_str[cond],
				    strlen(osm_hm_cond_str[cond]));
			p += strlen(osm_hm_cond_str[cond]);
			is_unhealthy = TRUE;
		}
	}

	if (is_unhealthy == FALSE)
		return;

	fprintf(file, "0x%016" PRIx64 ", %d, \"%s\", 0x%016" PRIx64 ", %d, "
		"\"%s\", {%s}, %s", cl_ntoh64(p_unhealthy_port->guid),
		p_unhealthy_port->port_num,
		p_unhealthy_port->node_desc,
		cl_ntoh64(p_unhealthy_port->peer_guid),
		p_unhealthy_port->peer_port_num,
		p_unhealthy_port->peer_node_desc,
		cond_buf, ctime(&p_unhealthy_port->timestamp.tv_sec));
}

static void sm_hm_dump_all(osm_sm_t * p_osm, FILE * file)
{
	struct opensm_dump_context dump_context;

	dump_context.p_osm = p_osm;
	dump_context.file = file;
	OSM_LOG(p_osm->p_log, OSM_LOG_DEBUG, "Dump unhealthy ports\n");
	fprintf(file, "# NodeGUID, PortNum, NodeDesc, PeerNodeGUID, PeerPortNum, "
		" PeerNodeDesc, {BadCond1, BadCond2, ...}, TimeStamp\n");
	CL_PLOCK_ACQUIRE(p_osm->p_lock);
	cl_fmap_apply_func(&p_osm->p_subn->unhealthy_ports_tbl,
			   sm_dump_one_unhealthy_port, &dump_context);
	CL_PLOCK_RELEASE(p_osm->p_lock);
}

static int osm_hm_file_dump(IN osm_sm_t * p_osm)
{
	int res = 1;
	res = opensm_dump_to_file(p_osm, "opensm-unhealthy-ports.dump",
				  sm_hm_dump_all);
	return res;
}

static osm_physp_t *find_peer(IN osm_sm_t * sm, IN uint8_t hop_count,
			      IN uint8_t *path)
{
	osm_node_t *p_node;
	osm_port_t *p_sm_port;
	osm_physp_t *p_physp = NULL;
	uint8_t hop;

	OSM_LOG_ENTER(sm->p_log);

	p_sm_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);

	if (!p_sm_port)
		goto Exit;

	/* hop_count == 0: destination port guid is SM */
	if (hop_count == 0) {
		p_physp = p_sm_port->p_physp;
		goto Exit;
	}

	p_node = osm_port_get_node(p_sm_port);
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, path[1]);
	else
		p_physp = p_sm_port->p_physp;

	/* hop_count == 1: outgoing physp is SM physp */
	for (hop = 2; p_physp && hop <= hop_count; hop++) {
		p_physp = p_physp->p_remote_physp;
		if (!p_physp)
			break;
		p_node = p_physp->p_node;
		p_physp = osm_node_get_physp_ptr(p_node, path[hop]);
	}

	/* At this point, p_physp points at the outgoing physp on the
	   last hop, or NULL if we don't know it.
	*/

Exit:
	OSM_LOG_EXIT(sm->p_log);

	return p_physp;
}

void osm_hm_set_by_dr_path(IN osm_sm_t * sm, IN uint8_t hop_count,
			   IN uint8_t *path, IN osm_hm_cond_type_enum cond)
{
	osm_unhealthy_port_t *p_unhealthy_port;
	cl_fmap_item_t *p_item;
	osm_physp_t *p_peer_physp;

	if (!sm->p_subn->opt.hm_unhealthy_ports_checks)
		return;

	if (hop_count == 0)
		return;

	p_peer_physp = find_peer(sm, hop_count, path);
	if (!p_peer_physp)
		return;

	p_unhealthy_port = osm_unhealthy_port_new(0, 0, NULL, 0,
						  osm_node_get_node_guid(p_peer_physp->p_node),
						  osm_physp_get_port_num(p_peer_physp),
						  p_peer_physp->p_node->print_desc);
	if (!p_unhealthy_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C203: Couldn't "
			"create new unhealthy port UNKNOWN[0]<-->0x%016"
			PRIx64 "[%d]\n",
			osm_node_get_node_guid(p_peer_physp->p_node),
			osm_physp_get_port_num(p_peer_physp));
		return;
	}

	p_item = cl_fmap_insert(&sm->p_subn->unhealthy_ports_tbl,
				&p_unhealthy_port->p_obj,
				&p_unhealthy_port->fmap_item);

	if (p_item != &p_unhealthy_port->fmap_item)
		osm_unhealthy_port_remove(sm, &p_unhealthy_port);
	osm_hm_set_cond(sm, (osm_unhealthy_port_t *)p_item, cond);
}

void osm_hm_set_by_physp(IN osm_sm_t * sm, IN osm_physp_t * p_physp,
			 IN osm_hm_cond_type_enum cond)
{
	osm_unhealthy_port_t *p_unhealthy_port;
	osm_node_t *p_node;
	osm_physp_t *p_remote_physp;
	cl_fmap_item_t *p_item;
	char *peer_node_desc = NULL;

	if (!sm->p_subn->opt.hm_unhealthy_ports_checks)
		return;

	p_node = osm_physp_get_node_ptr(p_physp);

	/* In case we're treating sp0, go over its' external ports */
	if (p_node->sw && osm_physp_get_port_num(p_physp) == 0) {
		osm_hm_set_by_node(sm, p_node, cond);
		return;
	}

	p_remote_physp = p_physp->p_remote_physp;
	if (!p_remote_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Remote port of "
			" 0x%016" PRIx64 "[%d] couldn't be found\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			osm_physp_get_port_num(p_physp));
		return;
	}

	peer_node_desc = p_remote_physp->p_node->print_desc;

	p_unhealthy_port = osm_unhealthy_port_new(osm_node_get_node_guid(p_node),
						  osm_physp_get_port_num(p_physp),
						  p_node->print_desc,
						  osm_node_get_type(p_node),
						  osm_node_get_node_guid(p_remote_physp->p_node),
						  osm_physp_get_port_num(p_remote_physp),
						  peer_node_desc);
	if (!p_unhealthy_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C204: Couldn't "
			"create new unhealthy port 0x%016" PRIx64
			"[%d]<-->0x%016" PRIx64 "[%d]\n",
			osm_node_get_node_guid(p_node),
			osm_physp_get_port_num(p_physp),
			osm_node_get_node_guid(p_remote_physp->p_node),
			osm_physp_get_port_num(p_remote_physp));
		return;
	}

	p_item = cl_fmap_insert(&sm->p_subn->unhealthy_ports_tbl,
				&p_unhealthy_port->p_obj,
				&p_unhealthy_port->fmap_item);

	if (p_item != &p_unhealthy_port->fmap_item)
		osm_unhealthy_port_remove(sm, &p_unhealthy_port);
	osm_hm_set_cond(sm, (osm_unhealthy_port_t *)p_item, cond);
}

void osm_hm_set_by_node(IN osm_sm_t * sm, IN osm_node_t * p_node,
			IN osm_hm_cond_type_enum cond)
{
	uint8_t i, port_num = osm_node_get_num_physp(p_node);
	osm_physp_t *p_physp;

	if (!sm->p_subn->opt.hm_unhealthy_ports_checks)
		return;

	for (i = 1; i < port_num; i++) {
		p_physp = osm_node_get_physp_ptr(p_node, i);
		if (p_node->physp_discovered[i] &&
		    ib_port_info_get_port_state(&p_physp->port_info) !=
		    IB_LINK_DOWN)
			osm_hm_set_by_physp(sm, p_physp, cond);
	}
	return;
}

void osm_hm_recover_all(IN osm_subn_t * p_subn)
{
	cl_fmap_item_t *p_item;
	osm_unhealthy_port_t *p_unhealthy_port;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Recovering ALL unhealthy ports\n");
	while ((p_item = cl_fmap_head(&p_subn->unhealthy_ports_tbl)) !=
	       cl_fmap_end(&p_subn->unhealthy_ports_tbl)) {
		cl_fmap_remove_item(&p_subn->unhealthy_ports_tbl, p_item);
		p_unhealthy_port = (osm_unhealthy_port_t *) p_item;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "Recovered NodeGUID 0x%016"
			PRIx64 " port %d\n", cl_ntoh64(p_unhealthy_port->peer_guid),
			p_unhealthy_port->peer_port_num);
		p_subn->p_osm->sm.hm_dirty = TRUE;
		osm_unhealthy_port_remove(&p_subn->p_osm->sm, &p_unhealthy_port);
	}
}

osm_unhealthy_port_t *osm_hm_get_unhealthy_port_by_guid(IN struct osm_sm *sm,
							IN ib_net64_t guid,
							IN const uint8_t port_num)
{
	osm_unhealthy_port_t *p_unhealthy_port, search_port;

	search_port.p_obj = &search_port;
	search_port.peer_guid = guid;
	search_port.peer_port_num = port_num;

	p_unhealthy_port = (osm_unhealthy_port_t *)
	    cl_fmap_get(&sm->p_subn->unhealthy_ports_tbl,
			&search_port.p_obj);

	if (&p_unhealthy_port->fmap_item ==
	    cl_fmap_end(&sm->p_subn->unhealthy_ports_tbl))
		return NULL;

	return p_unhealthy_port;
}

osm_unhealthy_port_t *osm_hm_get_unhealthy_port(IN struct osm_sm *sm,
						IN struct osm_node *p_node,
						IN const uint8_t port_num)
{
	ib_net64_t guid;

	OSM_ASSERT(p_node);

	guid = osm_node_get_node_guid(p_node);
	return osm_hm_get_unhealthy_port_by_guid(sm, guid, port_num);
}

static boolean_t is_sm_port(IN osm_sm_t *sm, IN osm_node_t * p_node,
			    IN const uint8_t port_num)
{
	osm_port_t *smport;
	osm_node_t *p_remote_node;
	osm_physp_t *p_remote_physp;
	boolean_t is_sm_port = FALSE;

	smport = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
	OSM_ASSERT(smport);
	if (smport->p_node == p_node &&
	    smport->p_physp->port_num == port_num)
		is_sm_port = TRUE;
	else if ((p_remote_physp = smport->p_physp->p_remote_physp)) {
		p_remote_node = p_remote_physp->p_node;
		if (p_remote_node == p_node &&
		    p_remote_physp->port_num == port_num)
			is_sm_port = TRUE;
	}
	return is_sm_port;
}

boolean_t lookup_isolated_port(IN osm_sm_t *sm, IN osm_node_t * p_node,
			       IN const uint8_t port_num)
{
	osm_unhealthy_port_t *p_unhealthy_port;

	p_unhealthy_port = osm_hm_get_unhealthy_port(sm, p_node, port_num);

	if (!p_unhealthy_port)
		return FALSE;

	if (p_unhealthy_port->isolate && is_sm_port(sm, p_node, port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C205: SM port/peer "
			"0x%" PRIx64 "[%u] configured for isolation, "
			"Ignoring this configuration (Not allowed)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), port_num);
		return FALSE;
	}

	return (p_unhealthy_port->isolate);
}

boolean_t lookup_no_discover_port(IN osm_sm_t *sm, IN osm_node_t *p_node,
				  IN const uint8_t port_num)
{
	osm_unhealthy_port_t *p_unhealthy_port;
	ib_net64_t guid = osm_node_get_node_guid(p_node);

	p_unhealthy_port = osm_hm_get_unhealthy_port_by_guid(sm, guid,
							     port_num);

	if (!p_unhealthy_port)
		return FALSE;

	if (p_unhealthy_port->no_discover && is_sm_port(sm, p_node, port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR C208: SM port/peer "
			"0x%" PRIx64 "[%u] configured for no_discover, "
			"Ignoring this configuration (Not allowed)\n",
			cl_ntoh64(guid), port_num);
		return FALSE;
	}

	return (p_unhealthy_port->no_discover);
}

boolean_t osm_hm_set_port_answer_ls(IN struct osm_sm *sm,
				    IN struct osm_node * p_node,
				    IN const uint8_t port_num,
				    IN boolean_t answered)
{
	osm_unhealthy_port_t *p_unhealthy_port;

	p_unhealthy_port = osm_hm_get_unhealthy_port(sm, p_node, port_num);

	if (!p_unhealthy_port)
		return FALSE;

	p_unhealthy_port->answered_last_ls = answered;
	return TRUE;
}

void osm_hm_mgr_process(IN struct osm_sm * sm)
{
	if (sm->p_subn->opt.hm_unhealthy_ports_checks == FALSE)
		return;

	if (!sm->hm_dirty) {
		/*
		 * Dump the file anyway in the first sweep,
		 * in order to clear previously found unhealthy ports
		 */
		if (sm->p_subn->first_time_master_sweep)
			osm_hm_file_dump(sm);
		return;
	}

	osm_hm_file_dump(sm);

	sm->hm_dirty = FALSE;
}

boolean_t osm_unhealthy_port_is_healthy(IN struct osm_sm *sm,
					IN osm_unhealthy_port_t *p_unhealthy_port)
{
	if (p_unhealthy_port->cond_mask ||
	    p_unhealthy_port->set_err_hist ||
	    p_unhealthy_port->flapping_hist ||
	    p_unhealthy_port->no_resp_hist ||
	    p_unhealthy_port->illegal_count ||
	    cl_event_wheel_num_regs(&sm->reboot_tracker,
				    ptr_to_key(p_unhealthy_port)))
		return FALSE;

	return TRUE;
}
