/*
 * Copyright (c) 2016 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Topologies manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VERBOSE_BYPASS_C
#include <opensm/osm_sm.h>

void osm_verbose_bypass_construct(osm_log_verbose_bypass_t *p_verbose_bypass)
{
	cl_map_construct(&p_verbose_bypass->attr_ids);
	cl_map_init(&p_verbose_bypass->attr_ids, MAP_GROWTH_SIZE);

	cl_map_construct(&p_verbose_bypass->guids);
	cl_map_init(&p_verbose_bypass->guids, MAP_GROWTH_SIZE);

	cl_map_construct(&p_verbose_bypass->lids);
	cl_map_init(&p_verbose_bypass->lids, MAP_GROWTH_SIZE);
}

void osm_verbose_bypass_clear(struct osm_sm *p_sm, osm_log_verbose_bypass_t *p_verbose_bypass)
{
	cl_map_remove_all(&p_verbose_bypass->attr_ids);
	cl_map_remove_all(&p_verbose_bypass->guids);
	cl_map_remove_all(&p_verbose_bypass->lids);
}

void osm_verbose_bypass_destroy(struct osm_sm *p_sm, osm_log_verbose_bypass_t *p_verbose_bypass)
{
	osm_verbose_bypass_clear(p_sm, p_verbose_bypass);
	cl_map_destroy(&p_verbose_bypass->attr_ids);
	cl_map_destroy(&p_verbose_bypass->guids);
	cl_map_destroy(&p_verbose_bypass->lids);
}

void osm_verbose_bypass_add_attr_id(struct osm_sm *p_sm, uint64_t key)
{
	osm_log_verbose_bypass_t *p_verbose_bypass = &p_sm->verbose_bypass;

	cl_map_insert(&p_verbose_bypass->attr_ids, key, p_verbose_bypass);

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Adding attribute id 0x%"
		PRIx64 " to log verbose bypass policy, new table size: %zu\n",
		key, cl_map_count(&p_verbose_bypass->attr_ids));
}

void osm_verbose_bypass_add_guid(struct osm_sm *p_sm, ib_net64_t key)
{
	osm_log_verbose_bypass_t *p_verbose_bypass = &p_sm->verbose_bypass;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Adding guid 0x%" PRIx64
		" to log verbose bypass policy\n", cl_ntoh64(key));

	cl_map_insert(&p_verbose_bypass->guids, key, p_verbose_bypass);
}

void osm_verbose_bypass_add_lid(struct osm_sm *p_sm, ib_net64_t key)
{
	osm_log_verbose_bypass_t *p_verbose_bypass = &p_sm->verbose_bypass;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Adding lid %" PRIu64
		" to log verbose bypass policy\n", key);

	cl_map_insert(&p_verbose_bypass->lids, key, p_verbose_bypass);
}

boolean_t is_verbose_bypass_allowed(struct osm_sm *p_sm,
				    osm_verbose_bypass_data_t *p_data)
{
	boolean_t res = FALSE;
	osm_log_verbose_bypass_t *p_verbose_bypass = &p_sm->verbose_bypass;
	uint64_t key;
	uint16_t lid, min_lid, max_lid;
	uint8_t i = 0;

	if (p_data->attr_id) {
		key = (uint64_t)cl_ntoh16(p_data->attr_id);
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Searching attribute id 0x%"
			PRIx64 " in log verbose bypass policy\n", key);
		if (cl_map_get(&p_verbose_bypass->attr_ids, key)) {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Attribute id 0x%"
				PRIx64 " authorised for bypass\n", key);
			res = TRUE;
			goto EXIT;
		}
	}

	while (i < VERBOSE_DATA_ELEMENTS && p_data->guids[i]) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Searching guid 0x%" PRIx64
			" in log verbose bypass policy\n",
			cl_ntoh64(p_data->guids[i]));
		if (cl_map_get(&p_verbose_bypass->guids, p_data->guids[i])) {
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Guid 0x%" PRIx64
				" authorised for bypass\n",
				cl_ntoh64(p_data->guids[i]));
			res = TRUE;
			goto EXIT;
		}
		i++;
	}

	i = 0;

	while (i < VERBOSE_DATA_ELEMENTS && p_data->base_lids[i]) {
		min_lid = cl_ntoh16(p_data->base_lids[i]);

		/* For MLIDs we ignore lmc */
		if (min_lid < IB_LID_MCAST_START_HO)
			max_lid = min_lid + (1 << p_sm->p_subn->opt.lmc) - 1;
		else
			max_lid = min_lid;

		for (lid = min_lid ; lid <= max_lid ; lid++) {
			key = (uint64_t)lid;
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Searching lid %"
				PRIu64 " in log verbose bypass policy\n", key);
			if (cl_map_get(&p_verbose_bypass->lids, key)) {
				OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Lid %"
					PRIu64 " authorised for bypass\n", key);
				res = TRUE;
				goto EXIT;
			}
		}
		i++;
	}

EXIT:
	return res;
}

boolean_t osm_verbose_bypass_handle_pi(struct osm_sm *p_sm,
				       ib_net16_t base_lid,
				       ib_net64_t port_guid,
				       ib_net64_t node_guid)
{
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;

	if (p_sm->p_subn->opt.verbose_bypass_policy_file) {
		memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
		verbose_bypass_data.attr_id = IB_MAD_ATTR_PORT_INFO;
		verbose_bypass_data.guids[0] = node_guid;
		verbose_bypass_data.guids[1] = port_guid;
		verbose_bypass_data.base_lids[0] = base_lid;
		res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
		if (res)
			osm_log_register_admin_pid(p_sm->p_log);
	}
	return res;
}

static boolean_t osm_verbose_bypass_handle_node(struct osm_sm *p_sm,
						ib_node_info_t *p_ni,
						ib_net16_t attr_id)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;
	uint8_t i;

	if (p_sm->p_subn->opt.verbose_bypass_policy_file) {
		p_node = osm_get_node_by_guid(p_sm->p_subn, p_ni->node_guid);
		memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
		verbose_bypass_data.attr_id = attr_id;
		verbose_bypass_data.guids[0] = p_ni->node_guid;
		if (p_node) {
			if (p_ni->node_type == IB_NODE_TYPE_SWITCH) {
				p_physp = osm_node_get_physp_ptr(p_node, 0);
				if (p_physp)
					verbose_bypass_data.base_lids[0] =
						p_physp->port_info.base_lid;
			} else {
				for (i = 1 ; i <= p_ni->num_ports ; i++) {
					p_physp = osm_node_get_physp_ptr(p_node, i);
					if (p_physp)
						verbose_bypass_data.base_lids[i - 1] =
							p_physp->port_info.base_lid;
				}
			}
		}
		res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
		if (res)
			osm_log_register_admin_pid(p_sm->p_log);
	}
	return res;
}

boolean_t osm_verbose_bypass_handle_ni(struct osm_sm *p_sm, ib_node_info_t *p_ni)
{
	return osm_verbose_bypass_handle_node(p_sm, p_ni, IB_MAD_ATTR_NODE_INFO);
}

boolean_t osm_verbose_bypass_handle_nd(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_NODE_DESC);
}

boolean_t osm_verbose_bypass_handle_ar_info(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO);
}

boolean_t osm_verbose_bypass_handle_ar_lft(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_ADAPTIVE_ROUTING_LFT);
}

boolean_t osm_verbose_bypass_handle_ar_group_table(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE);
}

boolean_t osm_verbose_bypass_handle_ar_group_table_copy(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE_COPY);
}

boolean_t osm_verbose_bypass_handle_si(struct osm_sm *p_sm, ib_net64_t node_guid)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;

	if (p_sm->p_subn->opt.verbose_bypass_policy_file) {
		memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
		verbose_bypass_data.attr_id = IB_MAD_ATTR_SWITCH_INFO;
		verbose_bypass_data.guids[0] = node_guid;
		p_node = osm_get_node_by_guid(p_sm->p_subn, node_guid);
		if (p_node && p_node->sw) {
			p_physp = osm_node_get_physp_ptr(p_node, 0);
			verbose_bypass_data.base_lids[0] = p_physp->port_info.base_lid;
		}
		res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
		if (res)
			osm_log_register_admin_pid(p_sm->p_log);
	}
	return res;
}

static boolean_t osm_verbose_bypass_handle_port(struct osm_sm *p_sm,
						ib_net16_t attr_id,
						osm_port_t *p_port,
						ib_net64_t extra_guid)
{
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;
	ib_net64_t port_guid = 0 , node_guid = 0;

	if (p_port && p_sm->p_subn->opt.verbose_bypass_policy_file) {
		port_guid = osm_port_get_guid(p_port);
		node_guid = osm_node_get_node_guid(p_port->p_node);
		memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
		verbose_bypass_data.attr_id = attr_id;
		verbose_bypass_data.guids[0] = node_guid;
		verbose_bypass_data.guids[1] = port_guid;
		if (extra_guid)
			verbose_bypass_data.guids[2] = extra_guid;
		verbose_bypass_data.base_lids[0] = osm_port_get_base_lid(p_port);
		res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
		if (res)
			osm_log_register_admin_pid(p_sm->p_log);
	}
	return res;
}

boolean_t osm_verbose_bypass_handle_pkt(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_P_KEY_TABLE,
					      p_port, 0);
}

boolean_t osm_verbose_bypass_handle_mepi(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm,
					      IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO,
					      p_port, 0);
}

boolean_t osm_verbose_bypass_handle_gi(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_GUID_INFO,
					      p_port, 0);
}

boolean_t osm_verbose_bypass_handle_geni(struct osm_sm *p_sm, osm_node_t *p_node)
{
	return osm_verbose_bypass_handle_node(p_sm, &p_node->node_info,
					      IB_MAD_ATTR_GENERAL_INFO);
}

boolean_t osm_verbose_bypass_handle_vi(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm,
					      IB_MAD_ATTR_VIRTUALIZATION_INFO,
					      p_port, 0);
}

boolean_t osm_verbose_bypass_handle_vps(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_VPORT_STATE,
					      p_port, 0);
}

boolean_t osm_verbose_bypass_handle_vnd(struct osm_sm *p_sm, osm_port_t *p_port,
					ib_net64_t vnode_guid)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_VNODE_DESC,
					      p_port, vnode_guid);
}

boolean_t osm_verbose_bypass_handle_vni(struct osm_sm *p_sm, osm_port_t *p_port,
					ib_net64_t vnode_guid)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_VNODE_INFO,
					      p_port, vnode_guid);
}

boolean_t osm_verbose_bypass_handle_vpi(struct osm_sm *p_sm, osm_port_t *p_port,
					ib_net64_t vport_guid)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_VPORT_INFO,
					      p_port, vport_guid);
}

boolean_t osm_verbose_bypass_handle_pr(struct osm_sm *p_sm, ib_path_rec_t *p_pr,
				       ib_net64_t comp_mask,
				       const osm_alias_guid_t *p_req_alias_guid)
{
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;
	uint8_t i = 0;
	ib_net16_t lid;


	if (!p_sm->p_subn->opt.verbose_bypass_policy_file)
		goto EXIT;

	memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
	verbose_bypass_data.attr_id = IB_MAD_ATTR_PATH_RECORD;

	if (p_req_alias_guid)
		verbose_bypass_data.guids[i++] = p_req_alias_guid->alias_guid;

	if (comp_mask & IB_PR_COMPMASK_SGID)
		verbose_bypass_data.guids[i++] = p_pr->sgid.unicast.interface_id;

	if (comp_mask & IB_PR_COMPMASK_DGID)
		verbose_bypass_data.guids[i++] = p_pr->dgid.unicast.interface_id;

	i = 0;

	if (p_req_alias_guid) {
		lid = osm_port_get_base_lid(p_req_alias_guid->p_base_port);
		verbose_bypass_data.base_lids[i++] = lid;
	}

	if (comp_mask & IB_PR_COMPMASK_SLID)
		verbose_bypass_data.base_lids[i++] = p_pr->slid;

	if (comp_mask & IB_PR_COMPMASK_DLID)
		verbose_bypass_data.base_lids[i++] = p_pr->dlid;

	res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
	if (res)
		osm_log_register_admin_pid(p_sm->p_log);

EXIT:
	return res;
}

boolean_t osm_verbose_bypass_handle_mcmpr(struct osm_sm *p_sm,
					  ib_member_rec_t *p_mcm_pr)
{
	osm_verbose_bypass_data_t verbose_bypass_data;
	boolean_t res = FALSE;
	uint8_t i = 0;
	ib_net16_t lid;
	osm_port_t *p_port;
	ib_net64_t port_guid;


	if (!p_sm->p_subn->opt.verbose_bypass_policy_file)
		goto EXIT;

	memset(&verbose_bypass_data, 0, sizeof(verbose_bypass_data));
	verbose_bypass_data.attr_id = IB_MAD_ATTR_MCMEMBER_RECORD;

	port_guid = p_mcm_pr->port_gid.unicast.interface_id;
	verbose_bypass_data.guids[0] = port_guid;

	p_port = osm_get_port_by_guid(p_sm->p_subn, port_guid);
	if (p_port) {
		lid = osm_port_get_base_lid(p_port);
		verbose_bypass_data.base_lids[i++] = lid;
	}

	verbose_bypass_data.base_lids[i++] = p_mcm_pr->mlid;

	res = is_verbose_bypass_allowed(p_sm, &verbose_bypass_data);
	if (res)
		osm_log_register_admin_pid(p_sm->p_log);

EXIT:
	return res;
}

boolean_t
osm_verbose_bypass_handle_qcsl(struct osm_sm *p_sm, osm_port_t *p_port)
{
	return osm_verbose_bypass_handle_port(p_sm, IB_MAD_ATTR_QOS_CONFIG_SL,
					      p_port, 0);
}
