/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of OpenSM Virtualization manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_qlist.h>
#include <complib/cl_pool.h>
#include <complib/cl_debug.h>
#include <complib/cl_vector.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VIRT_MGR_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_node.h>
#include <opensm/osm_port.h>
#include <opensm/osm_vport.h>
#include <opensm/osm_virt_mgr.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>

extern int wait_for_pending_transactions(osm_stats_t * stats);
extern void osm_send_vi_set(osm_sm_t * sm, osm_port_t * p_port,
			    boolean_t virt_enable, uint8_t vps_change_bit,
			    uint8_t client_rereg);

extern void pkey_mgr_process_pkey(IN osm_log_t * p_log,
				  IN osm_sm_t * sm,
				  IN const ib_net16_t pkey,
				  IN osm_pkey_tbl_t *p_pkey_tbl,
				  IN char *tbl_owner);

extern void pkey_mgr_process_pending_list(osm_log_t * p_log,
					  osm_sm_t * sm,
					  osm_pkey_tbl_t *p_pkey_tbl,
					  uint16_t max_num_of_blocks,
					  uint16_t max_num_of_pkeys,
					  char *owner_desc,
					  ib_net64_t owner_guid);

extern int pkey_mgr_update_peer_port(osm_log_t * p_log, osm_sm_t * sm,
                              const osm_port_t * const p_port,
                              osm_partition_enforce_type_enum enforce_type,
                              cl_map_t * vports_pkeys);

extern void add_pkeys_to_map(IN osm_pkey_tbl_t * p_pkey_tbl,
			     cl_map_t * pkey_map);

ib_api_status_t osm_virt_mgr_init(IN osm_virt_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	OSM_LOG_ENTER(sm->p_log);

	cl_qlist_init(&p_mgr->queue);
	cl_qmap_init(&p_mgr->partitions_data.limited_prtns_for_all);
	cl_qmap_init(&p_mgr->partitions_data.full_prtns_for_all);
	cl_qmap_init(&p_mgr->partitions_data.prtns_per_guid);
	cl_qmap_init(&p_mgr->vnodes);

	OSM_LOG_EXIT(sm->p_log);
	return IB_SUCCESS;
}

void virt_mgr_dequeue_port_fn(cl_list_item_t * const p_list_item,
			      void * context)
{
	osm_virt_data_t *p_virt_data = cl_item_obj(p_list_item, p_virt_data,
						   virt_mgr_queue_item);

	p_virt_data->is_virt_mgr_queue_pending = FALSE;
}

void osm_virt_mgr_empty_queue(IN osm_virt_mgr_t * p_mgr)
{
	cl_qlist_apply_func(&p_mgr->queue, virt_mgr_dequeue_port_fn, NULL);
	cl_qlist_remove_all(&p_mgr->queue);
	return;
}

void osm_virt_mgr_destroy(IN osm_virt_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	OSM_LOG_ENTER(sm->p_log);

	osm_virt_mgr_empty_queue(p_mgr);
	osm_virt_mgr_clear_prtn_data(sm);

	OSM_LOG_EXIT(sm->p_log);
	return;
}

static boolean_t virt_mgr_enqueue_port(osm_sm_t * sm, osm_port_t * p_port)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	osm_virt_data_t *p_virt_data = &p_port->virt_data;

	if (!p_virt_data->is_virt_mgr_queue_pending) {
		cl_qlist_insert_tail(&p_mgr->queue,
				     &p_virt_data->virt_mgr_queue_item);
		p_virt_data->is_virt_mgr_queue_pending = TRUE;
		return TRUE;
	}

	return FALSE;
}

/*
 * Start Virtual ports discover of a given osm_port_t object by sending
 * VirtualizationInfo Get to the port.
 */
static void virt_mgr_get_virtualization_info(size_t index, void * element,
					     void * context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;
	osm_physp_t *p_physp;
	osm_madw_context_t madw_context;
	ib_api_status_t status;

	if (!p_port)
		return;

	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	madw_context.vi_context.set_method = FALSE;
	madw_context.vi_context.logical_port_guid = osm_port_get_guid(p_port);
	madw_context.vi_context.state_change_bit_check =
	    p_port->virt_data.state_change_bit_check;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Sending VirtualizationInfo Get request to port "
		"0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(p_physp),
			     IB_MAD_ATTR_VIRTUALIZATION_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &madw_context);

	if (status != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 6B02: Failure initiating VirtualizationInfo "
			"Get request (%s) for port 0x%" PRIx64 ", returning "
			"port to the queue\n",
			ib_get_err_str(status),
			cl_ntoh64(osm_port_get_guid(p_port)));
		virt_mgr_enqueue_port(sm, p_port);
	} else
		osm_port_increase_virt_transactions(sm, p_port);
}

/* Changes VCAPortState of a virtual port to ACTIVE */
static void virt_mgr_activate_vport(osm_sm_t * sm, osm_port_t * p_port,
				    osm_vport_t *p_vport)
{
	osm_physp_t *p_physp;
	osm_madw_context_t context;
	ib_api_status_t status;
	ib_vport_info_t vpi;
	ib_net32_t attr_mod;
	uint16_t vport_index = p_vport->index;

	OSM_ASSERT(p_port);
	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	vpi = p_vport->vport_info;
	ib_vport_info_set_vport_state(&vpi, IB_LINK_ACTIVE);
	ib_vport_info_set_client_rereg(&vpi, 1);

	if (ib_vport_info_get_lid_required(&vpi)) {
		vpi.vport_lid = osm_lid_mgr_get_vport_lid(&sm->lid_mgr, p_port,
							  p_vport);
		osm_ucast_cache_remove_vport(&sm->ucast_mgr,
					     p_port,
					     vport_index,
					     vpi.vport_lid);
	} else
		vpi.vport_lid = 0;

	attr_mod = cl_hton32(((uint32_t)vport_index) << 16);

	context.vpi_context.set_method = TRUE;
	context.vpi_context.logical_port_guid = osm_port_get_guid(p_port);
	context.vpi_context.vnode_guid = p_vport->vnode_guid;
	context.vpi_context.vnode_port = p_vport->vnode_port;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Sending VirtualPortInfo Set request to virtual port %u "
		"of port 0x%" PRIx64 " (Activating)\n",
		vport_index,
		cl_ntoh64(osm_port_get_guid(p_port)));

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(p_physp),
			     (uint8_t *)&vpi, sizeof(vpi),
			     IB_MAD_ATTR_VPORT_INFO, attr_mod, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 6B04: Failure initiating VirtualPortInfo Set "
			"request (%s) for virtual %u of port 0x%" PRIx64 "\n",
			ib_get_err_str(status),
			vport_index,
			cl_ntoh64(osm_port_get_guid(p_port)));
	else
		osm_port_increase_virt_transactions(sm, p_port);

	return;
}

/*
 *  Verify that the reported lid by the ACTIVE vport is updated in the sm
 *  db, if not try to update it in the db and if the lid is not available
 *  update new lid to the vport (send VPortInfo set with new lid)
 */
static void virt_mgr_verify_active_vport_lid(osm_sm_t * sm, osm_port_t * p_port,
					     osm_vport_t *p_vport)
{
	osm_physp_t *p_physp;
	osm_madw_context_t context;
	ib_api_status_t status;
	ib_vport_info_t vpi;
	ib_net32_t attr_mod;
	ib_net16_t lid;
	uint16_t vport_index = p_vport->index;

	OSM_ASSERT(p_port);
	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	vpi = p_vport->vport_info;
	lid = osm_lid_mgr_get_vport_lid(&sm->lid_mgr, p_port, p_vport);
	if (p_vport->report_new_vport)
		osm_ucast_cache_remove_vport(&sm->ucast_mgr,
					     p_port,
					     vport_index,
					     lid);

	if (lid == vpi.vport_lid)
		return;

	/* Lid of ACTIVE vport changed, send VPortInfo Set to update lid */

	vpi.vport_lid = lid;
	ib_vport_info_set_vport_state(&vpi, IB_LINK_NO_CHANGE);
	ib_vport_info_set_client_rereg(&vpi, 1);
	attr_mod = cl_hton32(((uint32_t)vport_index) << 16);

	context.vpi_context.set_method = TRUE;
	context.vpi_context.logical_port_guid = osm_port_get_guid(p_port);
	context.vpi_context.vnode_guid = p_vport->vnode_guid;
	context.vpi_context.vnode_port = p_vport->vnode_port;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Sending VirtualPortInfo Set request to virtual port %u "
		"of port 0x%" PRIx64 " (Replacing Lid)\n",
		vport_index,
		cl_ntoh64(osm_port_get_guid(p_port)));

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(p_physp),
			     (uint8_t *)&vpi, sizeof(vpi),
			     IB_MAD_ATTR_VPORT_INFO, attr_mod, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 6B08: Failure initiating VirtualPortInfo Set "
			"request (%s) for virtual %u of port 0x%" PRIx64 "\n",
			ib_get_err_str(status),
			vport_index,
			cl_ntoh64(osm_port_get_guid(p_port)));
	else
		osm_port_increase_virt_transactions(sm, p_port);

	return;
}

/* Setup virtual ports */
static void virt_mgr_setup_vports(size_t index, void * element,
				  void * context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;
	osm_vport_t *p_vport;
	ib_vport_info_t *p_vpi;
	uint16_t vport_index_top;
	uint16_t vport_index;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_port)
		return;

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Configuring virtual ports of port 0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));

	if (!p_port->virt_data.vports) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Did not find virtual ports on port 0x%" PRIx64 ", "
			"skipping port\n",
			cl_ntoh64(osm_port_get_guid(p_port)));
		goto Exit;
	}

	vport_index_top =
	    cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);

	/*
	 * Our first priority is to hold the same lid for ACTIVE vports so we
	 * need first to iterate all vports and verify that all ACTIVE vports
	 * lids are updated in sm db (or try to hold them if they are not) and
	 * after that we can iterate the vports again and activate INITed
	 * vports
	 */

	for (vport_index = 0; vport_index <= vport_index_top; vport_index++) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport)
			continue;

		p_vpi = &p_vport->vport_info;
		if (ib_vport_info_get_lid_required(p_vpi) &&
		    ib_vport_info_get_vport_state(p_vpi) == IB_LINK_ACTIVE) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Verifying lid for ACTIVE virtual port %u of "
				"port 0x%" PRIx64 "\n",
				vport_index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			virt_mgr_verify_active_vport_lid(sm, p_port, p_vport);
		}
	}

	for (vport_index = 0; vport_index <= vport_index_top; vport_index++) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Skipping virtual port %u of physical port "
				"0x%" PRIx64 " due to missing data\n",
				vport_index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			continue;
		}

		p_vpi = &p_vport->vport_info;
		if (ib_vport_info_get_vport_state(p_vpi) == IB_LINK_INIT) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Activating virtual port %u of port "
				"0x%" PRIx64 "\n",
				vport_index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			virt_mgr_activate_vport(sm, p_port, p_vport);
		}
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

/* Setup pkeys on ports */
static void virt_mgr_setup_vports_pkeys(size_t index, void * element,
					void * context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;
	osm_physp_t *p_physp;
	osm_vport_t *p_vport;
	uint16_t vport_index_top;
	uint16_t vport_index;
	cl_map_t pkey_map;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_port)
		goto Exit;;

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Setting Pkeys data for virtual ports of port 0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));

	if (!p_port->virt_data.vports) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Did not find virtual ports on port 0x%" PRIx64 ", "
			"skipping port\n",
			cl_ntoh64(osm_port_get_guid(p_port)));
		goto Exit;
	}

	if (!(vport_index_top = osm_port_get_vport_top_index(p_port)))
		goto Exit;

	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	/* add pkeys of vPort 0*/
	cl_map_construct(&pkey_map);
	cl_map_init(&pkey_map, 16);

	add_pkeys_to_map(&p_physp->pkeys, &pkey_map);

	/* Start from 1, vport 0 does not have pkey table of its own */
	for (vport_index = 1; vport_index <= vport_index_top; vport_index++) {
		p_vport = osm_port_get_vport(p_port, vport_index);
		if (!p_vport)
			continue;

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Setting pkeys for virtual "
			"port %u of port 0x%" PRIx64 "\n", vport_index,
			cl_ntoh64(osm_port_get_guid(p_port)));

		osm_virt_mgr_apply_partitions(sm, p_vport, &pkey_map);
	}
	pkey_mgr_update_peer_port(sm->p_log, sm, p_port,
		sm->p_subn->opt.part_enforce_enum, &pkey_map);
	cl_map_remove_all(&pkey_map);
	cl_map_destroy(&pkey_map);

Exit:
	OSM_LOG_EXIT(sm->p_log);
}
/* Remove vports from multicast group if vport was removed from some pkeys*/
static void update_mcast_membership(size_t index, void *element,
				    void *context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;
	cl_list_item_t * p_mcm;;
	osm_mcm_port_t *mcm_port;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_port)
		goto Exit;

	if (!p_port->virt_data.vports) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Did not find virtual ports on port 0x%" PRIx64 ", "
			"skipping port\n",
			cl_ntoh64(osm_port_get_guid(p_port)));
		goto Exit;
	}

	if (!osm_port_get_vport_top_index(p_port))
		goto Exit;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Checking Mcast membership for virtual ports of port 0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));


	p_mcm = cl_qlist_head(&p_port->mcm_list);
	while (p_mcm != cl_qlist_end(&p_port->mcm_list)) {
			mcm_port = cl_item_obj(p_mcm,
				       mcm_port, list_item);
			p_mcm = cl_qlist_next(p_mcm);
			osm_port_update_vports_mcast_membership(sm,
								p_port,
								mcm_port->mgrp);
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
}
/* Send GID in notices for new virtual ports */
static void virt_mgr_report_new_vports(osm_sm_t * sm,
				       cl_ptr_vector_t * ports_to_process)
{
	osm_port_t *p_port;
	osm_vport_t *p_vport;
	ib_gid_t vport_gid;
        size_t i;
	uint16_t vport_index_top;
	uint16_t vport_index;

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

        for (i = 0; i < cl_ptr_vector_get_size(ports_to_process); i++) {
		p_port = (osm_port_t *)cl_ptr_vector_get(ports_to_process, i);
		if (!p_port || !p_port->virt_data.vports)
			continue;

		vport_index_top =
		    cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);

		/* We skip vport 0 of the port */
		for (vport_index = 1; vport_index <= vport_index_top;
		     vport_index++) {
			p_vport = osm_port_get_vport(p_port, vport_index);
			if (!p_vport || p_vport->report_new_vport == FALSE)
				continue;

			if (sm->p_subn->opt.aguid_inout_notice) {
				vport_gid.unicast.prefix =
				    sm->p_subn->opt.subnet_prefix;
				vport_gid.unicast.interface_id =
				    osm_vport_get_guid(p_vport);

				osm_report_notice_64_67(sm->p_subn, &vport_gid,
							64);
			}

			p_vport->report_new_vport = FALSE;
		}
	}

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}

/*
 * Return ports with virtualization errors to the tail of the queue
 * Ports that received timeouts will continue to be processed.
 */
static void virt_mgr_handle_timeouts(osm_sm_t * sm,
				     cl_ptr_vector_t * ports_to_process)
{
	osm_port_t *p_port;
        size_t i;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

        for (i = 0; i < cl_ptr_vector_get_size(ports_to_process); i++) {
		p_port = (osm_port_t *)cl_ptr_vector_get(ports_to_process, i);
		if (p_port && p_port->virt_data.transactions) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Returning port 0x%" PRIx64 " to "
				"Virtualization Manager's queue due to "
				"timeouts\n",
				cl_ntoh64(osm_port_get_guid(p_port)));
			cl_ptr_vector_set(ports_to_process, i, NULL);
			virt_mgr_enqueue_port(sm, p_port);
		}
	}

	CL_PLOCK_RELEASE(sm->p_lock);
}

static void virt_mgr_set_vi_client_rereg(size_t index, void * element,
					 void * context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;

	if (p_port && p_port->virt_data.first_virt_discovery) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"First Virt Discovery - Setting client rereg for "
			"port 0x%" PRIx64 "\n",
			cl_ntoh64(osm_port_get_guid(p_port)));
		osm_send_vi_set(sm, p_port, TRUE, 1, 1);
	}
}

static void virt_mgr_verify_mcast_membership(osm_sm_t *sm,
					    cl_ptr_vector_t *ports_to_process)
{
	OSM_LOG_ENTER(sm->p_log);
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	cl_ptr_vector_apply_func(ports_to_process, update_mcast_membership, sm);
	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
}

inline
static int virt_mgr_lock_apply_and_wait(osm_sm_t * sm,
					cl_ptr_vector_t * ports_to_process,
					cl_pfn_ptr_vec_apply_t pfn_callback)
{
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	cl_ptr_vector_apply_func(ports_to_process, pfn_callback, sm);
	CL_PLOCK_RELEASE(sm->p_lock);

	return wait_for_pending_transactions(&sm->p_subn->p_osm->stats);
}

void virt_mgr_handle_empty_vnode(IN struct osm_sm * sm, OUT osm_vnode_t *p_vnode,
				 IN boolean_t migrate_ports,
				 IN boolean_t del_if_empty)
{
	cl_list_item_t *p_list_item;
	cl_qlist_t *p_list;
	ib_net64_t vnode_guid;
	osm_port_t *p_port;
	osm_guid_list_item_t *p_guid_item;
	vnode_guid = p_vnode->vnode_info.vnode_guid;
	cl_qmap_t *p_map = &sm->virt_mgr.vnodes;

	/* If vnode not empty nothing to do */
	if (cl_qmap_count(&p_vnode->vports))
		return;

	p_list = &p_vnode->migration_ports;
	if (!cl_is_qlist_empty(p_list)) {
		p_list_item = cl_qlist_remove_head(p_list);
		while (p_list_item != cl_qlist_end(p_list)) {
			if (osm_exit_flag || !migrate_ports)
				goto NEXT;

			p_guid_item = (osm_guid_list_item_t*)p_list_item;
			p_port = osm_get_port_by_guid(sm->p_subn,
						      p_guid_item->port_guid);
				if (p_port) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Adding Port 0x%" PRIx64 " from"
					" vnode 0x%" PRIx64
					" migration ports list to virt"
					" mgr queue\n",
					cl_ntoh64(p_guid_item->port_guid),
					cl_ntoh64(vnode_guid));
				osm_virt_mgr_enqueue_port(sm, p_port);
			} else
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Port 0x%" PRIx64 " from vnode"
					" 0x%" PRIx64 " migration ports"
					" list not found - Ignoring it\n",
					cl_ntoh64(p_guid_item->port_guid),
					cl_ntoh64(vnode_guid));
NEXT:
			free(p_list_item);
			p_list_item = cl_qlist_remove_head(p_list);
		}
	}

	if (del_if_empty) {
		cl_qmap_remove_item(p_map, (cl_map_item_t*)p_vnode);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Deleting VNode 0x%"
			PRIx64 " (No VPorts Registered for it)\n",
			cl_ntoh64(vnode_guid));
		free(p_vnode);
	}
}

/* MUST BE CALLED UNDER LOCK! */
void virt_mgr_handle_empty_vnodes(struct osm_sm * sm, boolean_t migrate_ports,
				  boolean_t del_empty_vnodes)
{
	cl_map_item_t *p_item, *p_next_item;
	cl_qmap_t *p_map = &sm->virt_mgr.vnodes;
	osm_vnode_t *p_vnode;

	OSM_LOG_ENTER(sm->p_log);

	p_item = cl_qmap_head(p_map);
	while (p_item != cl_qmap_end(p_map)) {
		p_vnode = (osm_vnode_t*)p_item;
		p_next_item = cl_qmap_next(p_item);
		virt_mgr_handle_empty_vnode(sm, p_vnode, migrate_ports,
					    del_empty_vnodes);
		p_item = p_next_item;
	}
	OSM_LOG_EXIT(sm->p_log);
}

static void virt_mgr_process_vports_enhanced_qos(size_t index, void * element,
						 void * context)
{
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)element;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_port)
		goto Exit;

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Setting enhanced qos data for virtual ports of "
		"port 0x%" PRIx64 "\n", cl_ntoh64(osm_port_get_guid(p_port)));

	osm_enhanced_qos_mgr_process_vports(sm, p_port, NULL);
Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static boolean_t virt_mgr_process_ports(osm_sm_t * sm,
					cl_ptr_vector_t * ports_to_process)
{
	boolean_t shutdown = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Starting processing %d ports\n",
		cl_ptr_vector_get_size(ports_to_process));

	/* Discover virtual ports */
	if (virt_mgr_lock_apply_and_wait(sm, ports_to_process,
				     virt_mgr_get_virtualization_info)) {
		shutdown = TRUE;
		goto EXIT;
	}

	if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED) {
		/* On first virt discovery we set Client Rereg in port level */
		if (virt_mgr_lock_apply_and_wait(sm, ports_to_process,
					     virt_mgr_set_vi_client_rereg)) {
			shutdown = TRUE;
			goto EXIT;
		}

		/* Setup pkeys of virtual ports */
		if (virt_mgr_lock_apply_and_wait(sm, ports_to_process,
					     virt_mgr_setup_vports_pkeys)) {
			shutdown = TRUE;
			goto EXIT;
		}

		virt_mgr_verify_mcast_membership(sm, ports_to_process);

		/* Setup qos config sl of virtual ports */
		if (virt_mgr_lock_apply_and_wait(sm, ports_to_process,
				    virt_mgr_process_vports_enhanced_qos)) {
			shutdown = TRUE;
			goto EXIT;
		}
		/* Activate virtual ports */
		if (virt_mgr_lock_apply_and_wait(sm, ports_to_process,
					     virt_mgr_setup_vports)) {
			shutdown = TRUE;
			goto EXIT;
		}
	}

	/* Send GID in notices for new virtual ports */
	virt_mgr_report_new_vports(sm, ports_to_process);

	/* Handle ports that had timeouts */
	virt_mgr_handle_timeouts(sm, ports_to_process);
EXIT:
	OSM_LOG_EXIT(sm->p_log);
	return shutdown;
}

void osm_virt_mgr_process(IN osm_sm_t * sm)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	osm_vlid_mgr_t *p_vlid_mgr = &sm->lid_mgr.vlid_mgr;
	cl_list_item_t *p_list_item;
	osm_virt_data_t *p_virt_data;
	osm_port_t *p_port;
	cl_ptr_vector_t ports_to_process;
	int virt_max_ports_in_process;
	uint32_t num_ports_to_process, num_ports_left, i;
	osm_subn_t *p_subn = sm->p_subn;
	boolean_t shutdown = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	/* Decide how many ports to process */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	num_ports_to_process = cl_qlist_count(&p_mgr->queue);
	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"VirtMgr processing cycle start: %u ports pending in queue\n",
		num_ports_to_process);

	if (!num_ports_to_process) {
		CL_PLOCK_RELEASE(sm->p_lock);
		goto ExitNoSignal;
	}

	virt_max_ports_in_process = sm->p_subn->opt.virt_max_ports_in_process;
	if (virt_max_ports_in_process && num_ports_to_process > virt_max_ports_in_process)
		num_ports_to_process = virt_max_ports_in_process;

	/* Allocate vector to hold all ports that need to be processed */
	if (cl_ptr_vector_init(&ports_to_process, num_ports_to_process,
			       num_ports_to_process)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 6B01: Failed to allocate memory for "
			"processing virtual ports\n");
		num_ports_to_process = 0;
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	/*
	 * Move ports from the queue to a vector with the ports to be
	 * processed on this iteration.
	 */
	for (i = 0; i < num_ports_to_process; i++) {
		OSM_ASSERT(cl_qlist_count(&p_mgr->queue));

		p_list_item = cl_qlist_remove_head(&p_mgr->queue);
		p_virt_data = cl_item_obj(p_list_item, p_virt_data,
					  virt_mgr_queue_item);
		p_virt_data->is_virt_mgr_queue_pending = FALSE;
		p_virt_data->transactions = 0;
		p_port = p_virt_data->p_port;

		if (osm_port_get_base_lid(p_port))
			cl_ptr_vector_set(&ports_to_process, i, p_port);
		else
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 6B09: Virt Mgr dropping port without lid "
				" - 0X%" PRIx64 "\n",
				cl_ntoh64(osm_port_get_guid(p_port)));
	}

	CL_PLOCK_RELEASE(sm->p_lock);

	/* Process the ports */
	shutdown = virt_mgr_process_ports(sm, &ports_to_process);

	cl_ptr_vector_destroy(&ports_to_process);

Exit:
	/* If there are more ports in the queue, send signal to the SM */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	num_ports_left = cl_qlist_count(&p_mgr->queue);

	if ((!num_ports_left || !sm->p_subn->opt.virt_max_ports_in_process)
	    && osm_vlid_mgr_is_update_since_apply(p_vlid_mgr)) {
		if (sm->ucast_mgr.vlid_cache_valid == FALSE)
			osm_subn_apply_vlids_to_lfts(sm->p_subn);
		CL_PLOCK_RELEASE(sm->p_lock);
		osm_ucast_mgr_set_fwd_tables(&sm->ucast_mgr);
		CL_PLOCK_ACQUIRE(sm->p_lock);
		if (sm->ucast_mgr.p_subn->opt.use_ucast_cache)
			sm->ucast_mgr.vlid_cache_valid = TRUE;
		osm_dump_lfts(p_subn->p_osm, p_subn, "");
		osm_db_store(sm->lid_mgr.p_g2l,
		sm->p_subn->opt.fsync_high_avail_files);
		CL_PLOCK_RELEASE(sm->p_lock);
		CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	}

	/*
	 * If shutdown occured during ports processing then processing ended
	 * without waiting for all mads handling to end and therefore vports
	 * might still try to register to vnodes, therefore if shutdown occured
	 * we don't want to delete empty vnodes.
	 */
	virt_mgr_handle_empty_vnodes(sm, TRUE, !shutdown);

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"VirtMgr processing cycle end: %u ports processed, "
		"%u ports pending in queue\n",
		num_ports_to_process, num_ports_left);

	if (num_ports_left)
		osm_sm_signal(sm, OSM_SIGNAL_VIRTMGR_PROCESS_REQUEST);
	else
		osm_dump_virtualization(sm->p_subn->p_osm);

ExitNoSignal:
	OSM_LOG_EXIT(sm->p_log);
	return;
}

/* Write lock should be acquired when calling this function */
void osm_virt_mgr_enqueue_port(IN osm_sm_t * sm, IN osm_port_t * p_port)
{

	if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_IGNORED)
		return;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Adding port 0x%016" PRIx64
		" to virtMgr queue\n", cl_ntoh64(p_port->guid));

	p_port->virt_data.state_change_bit_check = FALSE;
	if (virt_mgr_enqueue_port(sm, p_port))
		osm_sm_signal(sm, OSM_SIGNAL_VIRTMGR_PROCESS_REQUEST);
}

/* Write lock should be acquired when calling this function */
void osm_virt_mgr_dequeue_port(IN osm_sm_t * sm, IN osm_port_t * p_port)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	osm_virt_data_t *p_virt_data = &p_port->virt_data;

	if (p_virt_data->is_virt_mgr_queue_pending) {
		cl_qlist_remove_item(&p_mgr->queue,
				     &p_virt_data->virt_mgr_queue_item);
		p_virt_data->is_virt_mgr_queue_pending = FALSE;
	}

	return;
}

void osm_virt_mgr_vport_state_change_check(IN osm_sm_t * sm)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	uint16_t ports_enqueued = 0;

	OSM_LOG_ENTER(sm->p_log);
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_item = cl_qmap_head(&sm->p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&sm->p_subn->port_guid_tbl)) {
		p_port = (osm_port_t *)p_item;
		p_item = cl_qmap_next(p_item);
		if (osm_port_is_virt_supported(p_port) &&
		    virt_mgr_enqueue_port(sm, p_port)) {
			p_port->virt_data.state_change_bit_check = TRUE;
			ports_enqueued++;
		}
	}

	if (ports_enqueued) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Enqueued %u ports for virtualization light sweep\n",
			ports_enqueued);
		osm_sm_signal(sm, OSM_SIGNAL_VIRTMGR_PROCESS_REQUEST);
	}

	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
}


static void virt_mgr_clear_prtns(cl_qmap_t *p_map)
{
	cl_map_item_t *p_item;
	osm_virt_mgr_partition_t *p_prtn_item;

	while (cl_qmap_count(p_map) != 0) {
		p_item = cl_qmap_head(p_map);
		p_prtn_item = (osm_virt_mgr_partition_t*)p_item;
		cl_qmap_remove_item(p_map, p_item);
		free(p_prtn_item);
        }
}

void osm_virt_mgr_clear_prtn_data(IN struct osm_sm *sm)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	cl_map_item_t *p_map_item;
	cl_qmap_t *p_per_guid_tbl;
	osm_virt_mgr_partitions_t *p_prtns_data;
	uint32_t count;
	ib_net64_t guid;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Clearing partitions data from virt "
		"mgr\n");

	count = cl_qmap_count(&p_mgr->partitions_data.limited_prtns_for_all);
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Clearing %u limited partitions "
		"defined for ALL guids\n", count);
	virt_mgr_clear_prtns(&p_mgr->partitions_data.limited_prtns_for_all);

	count = cl_qmap_count(&p_mgr->partitions_data.full_prtns_for_all);
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Clearing %u full partitions "
		"defined for ALL guids\n", count);
	virt_mgr_clear_prtns(&p_mgr->partitions_data.full_prtns_for_all);

	p_per_guid_tbl = &p_mgr->partitions_data.prtns_per_guid;
	p_map_item = cl_qmap_head(p_per_guid_tbl);
	while (p_map_item != cl_qmap_end(p_per_guid_tbl)) {
		p_prtns_data = (osm_virt_mgr_partitions_t*)p_map_item;
		guid = (ib_net64_t)cl_qmap_key(p_map_item);

		count = cl_qmap_count(&p_prtns_data->limited_prtns);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Clearing %u limited "
			"partitions defined for guid 0x%" PRIx64 "\n", count,
			cl_ntoh64(guid));
		virt_mgr_clear_prtns(&p_prtns_data->limited_prtns);

		count = cl_qmap_count(&p_prtns_data->full_prtns);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Clearing %u full "
			"partitions defined for guid 0x%" PRIx64 "\n", count,
			cl_ntoh64(guid));
		virt_mgr_clear_prtns(&p_prtns_data->full_prtns);

		p_map_item = cl_qmap_next(p_map_item);
	}
}

/*
 * This method iterate through all guids which were specifically specified for
 * partitions and remove a required partition with required membbership from
 * any guid which includes them
 */
static void remove_prtn_from_all_guids(IN struct osm_sm *sm, IN ib_net16_t pkey,
				       IN boolean_t full)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	cl_qmap_t *p_guid_map, *p_prtn_map;
	cl_map_item_t *p_item, *p_tmp_item;
	osm_virt_mgr_partitions_t *p_partitions;

	p_guid_map = &p_mgr->partitions_data.prtns_per_guid;
	p_item = cl_qmap_head(p_guid_map);

	while (p_item != cl_qmap_end(p_guid_map)) {
		p_partitions = (osm_virt_mgr_partitions_t*)p_item;
		p_item = cl_qmap_next(p_item);

		/* Set the map according to membership */
		if (full)
			p_prtn_map = &p_partitions->full_prtns;
		else
			p_prtn_map = &p_partitions->limited_prtns;

		/* Remove the prtn if found */
		p_tmp_item = cl_qmap_remove(p_prtn_map, pkey);
		if (p_tmp_item != cl_qmap_end(p_prtn_map))
			free(p_tmp_item);
	}
}

boolean_t osm_virt_mgr_save_partition(IN struct osm_sm *sm, IN ib_net64_t guid,
				      IN ib_net16_t pkey, IN boolean_t full,
				      IN boolean_t indx0)
{
	osm_virt_mgr_partition_t *p_prtn;
	cl_qmap_t *p_limited_map, *p_full_map;
	cl_map_item_t *p_map_item, *tmp_map_item;
	osm_virt_mgr_partitions_t *p_partitions;
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	cl_qmap_t *p_per_guid_map;
	boolean_t res = TRUE;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Saving partition in virt mgr "
		"partition data: guid 0x%" PRIx64 ", pkey 0x%04x, full %u, "
		"indx0 %u\n", cl_ntoh64(guid), cl_ntoh16(pkey), full, indx0);

	/* Find where should this partition shall be saved */
	if (guid) {
		p_per_guid_map = &p_mgr->partitions_data.prtns_per_guid;
		p_map_item = cl_qmap_get(p_per_guid_map, guid);
		if (p_map_item == cl_qmap_end(p_per_guid_map)) {
			/* First pkey for this guid, create item for it */
			p_partitions = calloc(1, sizeof(osm_virt_mgr_partitions_t));
			if (!p_partitions) {
				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6B05: "
					"Failed to allocate partitions lists "
					"for guid 0x%" PRIx64 " in virt mgr "
					" partitions data\n", cl_ntoh64(guid));
				res = FALSE;
				goto EXIT;
			}
			cl_qmap_insert(p_per_guid_map, guid,
				       (cl_map_item_t*)p_partitions);
			cl_qmap_init(&p_partitions->limited_prtns);
			cl_qmap_init(&p_partitions->full_prtns);
		} else
			p_partitions = (osm_virt_mgr_partitions_t*)p_map_item;

		p_full_map = &p_partitions->full_prtns;
		p_limited_map = &p_partitions->limited_prtns;
	} else {
		/*
		 * Since last definition should take effect if a pkey was
		 * defined to all with a certain membership we should remove
		 * from all guids the opposite membership for this pkey if
		 * they have it
		 */
		remove_prtn_from_all_guids(sm, pkey, !full);
		p_full_map = &p_mgr->partitions_data.full_prtns_for_all;
		p_limited_map = &p_mgr->partitions_data.limited_prtns_for_all;
	}


	/* Now we have the list this partition shall be saved in, save partition */
	p_prtn = calloc(1, sizeof(osm_virt_mgr_partition_t));
	if (!p_prtn) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6B06: "
			"Failed to allocate partition item for pkey 0x%04x "
			"guid (0x%" PRIx64 ") in virt mgr partitions data\n",
			cl_ntoh16(pkey), cl_ntoh64(guid));
		res = FALSE;
		goto EXIT;
	}
	p_prtn->pkey = pkey;
	p_prtn->full = full;
	p_prtn->indx0 = indx0;


	/*
	 * In vports pkey tables there is no option to allow both memberships
	 * so we ensure this pkey appears only in one of the maps
	 */
	tmp_map_item = cl_qmap_remove(p_limited_map, pkey);
	if (tmp_map_item != cl_qmap_end(p_limited_map))
		free(tmp_map_item);

	tmp_map_item = cl_qmap_remove(p_full_map, pkey);
	if (tmp_map_item != cl_qmap_end(p_full_map))
		free(tmp_map_item);

	/* No need to check return value, we just removed this key */
	if (full)
		cl_qmap_insert(p_full_map, pkey, &p_prtn->map_item);
	else
		cl_qmap_insert(p_limited_map, pkey, &p_prtn->map_item);

EXIT:
	return res;
}

static void add_partition_to_pending_list(struct osm_sm *sm,
					  osm_virt_mgr_partition_t *p_prtn,
					  osm_pkey_tbl_t *pkeys_tbl,
					  char *tbl_owner)
{
	ib_net16_t pkey;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Adding pkey 0x%04x "
		"(full %u, indx0 %u) to pkey table\n", cl_ntoh16(p_prtn->pkey),
		p_prtn->full, p_prtn->indx0);
	/* This method convert the pkey to full if needed */
	if (p_prtn->indx0)
		osm_pkey_tbl_set_indx0_pkey(sm->p_log, p_prtn->pkey,
					    p_prtn->full, pkeys_tbl);
	/*
	 * This method which adds the pkey to the pkey table pending list
	 * expects the pkey to already be converted to full if needed
	 */
	pkey = p_prtn->pkey;
	if (p_prtn->full)
		pkey |= cl_hton16(0x8000);
	pkey_mgr_process_pkey(sm->p_log, sm, pkey, pkeys_tbl, tbl_owner);
}

/*
 * Get two maps indexed by pkeys and a pkey
 * Return indication if the pkey exist in one of the maps
 */
static boolean_t is_pkey_in_maps(cl_qmap_t *p_map1, cl_qmap_t *p_map2,
				 ib_net16_t pkey)
{
	boolean_t found = FALSE;
	cl_map_item_t *p_item;

	if (!p_map1)
		goto CHECK_SECOND;

	p_item = cl_qmap_get(p_map1, pkey);
	if (p_item != cl_qmap_end(p_map1)) {
		found = TRUE;
		goto EXIT;
	}

CHECK_SECOND:

	if (!p_map2)
		goto EXIT;

	p_item = cl_qmap_get(p_map2, pkey);
	if (p_item != cl_qmap_end(p_map2))
		found = TRUE;
EXIT:
	return found;
}
/*
 * This method search the parititons data of the virt mgr for parititons
 * which are relevant for a given vport (partitions which were defined for
 * this specific vport guid or parititions which were defined for ALL)
 *
 * Relevant partitions are added to the vport pkey table pending list
 *
 * Adding principles:
 *	1) Limited memberships added first
 *	2) If parition defined both for ALL and for this specific guid we
 *		add with the membership defined for the guid
 */
static void update_pending_list(struct osm_sm *sm, osm_vport_t *p_vport,
			        char *tbl_owner)
{
	osm_virt_mgr_partition_t *p_prtn, p_limit_def, p_full_def;
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	cl_map_item_t *p_map_item;
	osm_virt_mgr_partitions_t *p_guid_prtns = NULL;
	ib_net64_t vport_guid = osm_vport_get_guid(p_vport);
	cl_qmap_t *p_guid_full_map = NULL, *p_guid_limited_map = NULL;
	cl_qmap_t *p_per_guid_map, *p_all_full_map, *p_all_limited_map;
	osm_pkey_tbl_t *pkeys_tbl = &p_vport->pkeys;

	memset(&p_limit_def, 0, sizeof(osm_virt_mgr_partition_t));
	memset(&p_full_def, 0, sizeof(osm_virt_mgr_partition_t));

	p_all_full_map = &p_mgr->partitions_data.full_prtns_for_all;
	p_all_limited_map = &p_mgr->partitions_data.limited_prtns_for_all;

	p_per_guid_map = &p_mgr->partitions_data.prtns_per_guid;
	p_map_item = cl_qmap_get(p_per_guid_map, vport_guid);
	if (p_map_item != cl_qmap_end(p_per_guid_map)) {
		p_guid_prtns = (osm_virt_mgr_partitions_t*)p_map_item;
		p_guid_full_map = &p_guid_prtns->full_prtns;
		p_guid_limited_map = &p_guid_prtns->limited_prtns;
	}

	/* Add prtns defined for this specific guid with limited membership */
	if (p_guid_limited_map) {
		p_map_item = cl_qmap_head(p_guid_limited_map);
		while (p_map_item != cl_qmap_end(p_guid_limited_map)) {
			p_prtn = (osm_virt_mgr_partition_t*)p_map_item;
			if (ib_pkey_get_base(p_prtn->pkey) == IB_DEFAULT_PARTIAL_PKEY)
				p_limit_def = *p_prtn;
			else
				add_partition_to_pending_list(sm, p_prtn, pkeys_tbl,
							      tbl_owner);
			p_map_item = cl_qmap_next(p_map_item);
		}
	}

	/*
	 * Add pkeys defined for ALL with limited membership and not defined for
	 * this specific guid
	 */
	p_map_item = cl_qmap_head(p_all_limited_map);
	while (p_map_item != cl_qmap_end(p_all_limited_map)) {
		p_prtn = (osm_virt_mgr_partition_t*)p_map_item;
		p_map_item = cl_qmap_next(p_map_item);
		if (is_pkey_in_maps(p_guid_limited_map, p_guid_full_map,
				    p_prtn->pkey))
			continue;
		if (ib_pkey_get_base(p_prtn->pkey) == IB_DEFAULT_PARTIAL_PKEY)
			p_limit_def = *p_prtn;
		else
			add_partition_to_pending_list(sm, p_prtn, pkeys_tbl,
						      tbl_owner);
	}

	/* Add prtns defined for this specific guid with full membership */
	if (p_guid_full_map) {
		p_map_item = cl_qmap_head(p_guid_full_map);
		while (p_map_item != cl_qmap_end(p_guid_full_map)) {
			p_prtn = (osm_virt_mgr_partition_t*)p_map_item;
			if (ib_pkey_get_base(p_prtn->pkey) == IB_DEFAULT_PARTIAL_PKEY)
				p_full_def = *p_prtn;
			else
				add_partition_to_pending_list(sm, p_prtn, pkeys_tbl,
							      tbl_owner);
			p_map_item = cl_qmap_next(p_map_item);
		}
	}

	/*
	 * Add pkeys defined for ALL with full membership and not defined for
	 * this specific guid
	 */
	p_map_item = cl_qmap_head(p_all_full_map);
	while (p_map_item != cl_qmap_end(p_all_full_map)) {
		p_prtn = (osm_virt_mgr_partition_t*)p_map_item;
		p_map_item = cl_qmap_next(p_map_item);
		if (is_pkey_in_maps(p_guid_limited_map, p_guid_full_map,
				    p_prtn->pkey))
			continue;
		if (ib_pkey_get_base(p_prtn->pkey) == IB_DEFAULT_PARTIAL_PKEY)
			p_full_def = *p_prtn;
		else
			add_partition_to_pending_list(sm, p_prtn, pkeys_tbl,
						      tbl_owner);
	}

	if (p_limit_def.pkey)
		add_partition_to_pending_list(sm, &p_limit_def, pkeys_tbl, tbl_owner);

	if (p_full_def.pkey)
		add_partition_to_pending_list(sm, &p_full_def, pkeys_tbl, tbl_owner);
}

void osm_virt_mgr_apply_partitions(IN struct osm_sm *sm, struct osm_vport *p_vport,
					  IN cl_map_t * pkey_map)
{
	ib_net64_t vport_guid;
	char tbl_owner[50];
	uint16_t max_num_of_blocks;
	uint16_t max_num_of_pkeys;

	if (!p_vport->index) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9306: "
			"Ignoring request to apply partitions for VPort 0x%"
			PRIx64	", vport index 0 of port 0x%" PRIx64 "\n",
			cl_ntoh64(osm_vport_get_guid(p_vport)),
			cl_ntoh64(osm_port_get_guid(p_vport->p_port)));
		return;
	}
	memset(tbl_owner, 0, sizeof(tbl_owner));
	sprintf(tbl_owner, "vport %u of port 0x%" PRIx64,
		p_vport->index,
		cl_ntoh64(osm_port_get_guid(p_vport->p_port)));

	vport_guid = osm_vport_get_guid(p_vport);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Applying partitions for vport 0x%"
		PRIx64 ", vport %u of port 0x%" PRIx64 "\n",
		cl_ntoh64(vport_guid), p_vport->index,
		cl_ntoh64(osm_port_get_guid(p_vport->p_port)));

	/* Fill pending list with relevant partitions */
	update_pending_list(sm, p_vport, tbl_owner);

	osm_vport_get_max_pkeys_and_blocks(sm, p_vport, &max_num_of_pkeys,
					   &max_num_of_blocks);

	/* Process pending list */
	pkey_mgr_process_pending_list(sm->p_log, sm, &p_vport->pkeys,
				      max_num_of_blocks,
				      max_num_of_pkeys, tbl_owner,
				      vport_guid);

	/* Send changed blocks to vport */
	osm_vport_update_pkey_tbl(sm, p_vport, tbl_owner);

	add_pkeys_to_map(&p_vport->pkeys, pkey_map);

}

static ib_api_status_t send_vnd_get(osm_sm_t *sm, osm_port_t *p_port,
				    ib_net64_t vnode_guid, uint16_t vport_index)
{
	ib_net32_t attr_mod;
	ib_api_status_t status;
	osm_madw_context_t context;
	osm_dr_path_t *p_dr_path;
	ib_net64_t m_key;

	memset(&context, 0, sizeof(context));
	context.vnd_context.vnode_guid = vnode_guid;
	context.vnd_context.port_guid = p_port->guid;
	p_dr_path = osm_physp_get_dr_path_ptr(p_port->p_physp);
	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	/* Set the vport index within the attr_mod in bits 16-31*/
	attr_mod = cl_hton32(((uint32_t)vport_index) << 16);

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_VNODE_DESC,
			     attr_mod, TRUE, m_key, 0, CL_DISP_MSGID_NONE,
			     &context);

	if (status == IB_SUCCESS)
		osm_port_increase_virt_transactions(sm, p_port);
	else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9301: "
			"Failed sending VND get request to vnode 0x%" PRIx64
			" while handling vport index %u of port 0x%" PRIx64 "\n",
			cl_ntoh64(vnode_guid), vport_index,
			cl_ntoh64(osm_port_get_guid(p_port)));

	return status;
}

/* This function need to be called under write p_lock */
boolean_t osm_virt_mgr_add_vnode(struct osm_sm *sm, IN ib_vnode_info_t *p_vni,
				 IN osm_port_t *p_port, IN uint16_t vport_index)
{
	osm_virt_mgr_t *p_mgr = &sm->virt_mgr;
	ib_net64_t vnode_guid = p_vni->vnode_guid;
	cl_map_item_t *p_item;
	osm_vnode_t *p_vnode;
	boolean_t res = TRUE;

	p_item = cl_qmap_get(&p_mgr->vnodes, vnode_guid);

	/* If this vnode already exist in vnodes map, there is nothing to do */
	if (p_item == cl_qmap_end(&p_mgr->vnodes)) {
		/* Create vnode, save in map, send vnode desc */
		p_vnode = calloc(1, sizeof(osm_vnode_t));
		if (!p_vnode) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6B07: "
				"Failed to allocate VNode object for VNode 0x%"
				PRIx64 ", required for vport %u of port 0x%"
				PRIx64 "\n", cl_ntoh64(vnode_guid), vport_index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			res = FALSE;
			goto EXIT;
		}
		p_vnode->vnode_info =  *p_vni;
		cl_qmap_init(&p_vnode->vports);
		p_vnode->node_guid = osm_node_get_node_guid(p_port->p_node);
		cl_qlist_init(&p_vnode->migration_ports);
		cl_qmap_insert(&p_mgr->vnodes, vnode_guid, &p_vnode->map_item);
		if (send_vnd_get(sm, p_port, vnode_guid, vport_index) !=
		    IB_SUCCESS)
			res = FALSE;
	}
EXIT:
	return res;
}

osm_vnode_t *osm_virt_mgr_get_vnode(IN struct osm_sm *sm,
				    IN ib_net64_t vnode_guid)
{
	cl_map_item_t *p_item;

	p_item = cl_qmap_get(&sm->virt_mgr.vnodes, vnode_guid);
	if (p_item == cl_qmap_end(&sm->virt_mgr.vnodes))
		return NULL;
	else
		return (osm_vnode_t*)p_item;
}

static void osm_vport_add_migration_port(osm_sm_t *sm, osm_vnode_t *p_vnode,
					 osm_port_t *p_port)
{
	osm_guid_list_item_t *p_guid_item;

	p_guid_item = calloc(1, sizeof(osm_guid_list_item_t));
	if (!p_guid_item) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9308: "
			"Can't allocate memory for guid item\n");
		return;
	}
	p_guid_item->port_guid = p_port->guid;
	cl_qlist_insert_head(&p_vnode->migration_ports, &p_guid_item->list_item);
}

boolean_t osm_virt_mgr_register_to_vnode(IN struct osm_sm *sm,
					 IN ib_net64_t vnode_guid,
					 IN struct osm_vport *p_vport)
{
	cl_map_item_t *p_item;
	ib_net64_t vport_guid = osm_vport_get_guid(p_vport);
	boolean_t res = TRUE;
	osm_vnode_t *p_vnode;
	osm_vport_t *p_found_vport;
	ib_net64_t vport_node_guid;

	p_vnode = osm_virt_mgr_get_vnode(sm, vnode_guid);
	OSM_ASSERT(p_vnode);

	vport_node_guid = osm_node_get_node_guid(p_vport->p_port->p_node);
	if (p_vnode->node_guid != vport_node_guid) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort 0x%" PRIx64
			" tried to register to VNode 0x%" PRIx64
			" which does not share it's node guid, Adding port "
			"0x%" PRIx64 " to the VNode migration port list\n",
			cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid),
			cl_ntoh64(p_vport->p_port->guid));
		osm_vport_add_migration_port(sm, p_vnode, p_vport->p_port);
		res = FALSE;
		goto EXIT;
	}

	p_item = cl_qmap_get(&p_vnode->vports, vport_guid);
	if (p_item == cl_qmap_end(&p_vnode->vports)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Registering VPort 0x%"
			PRIx64 " to VNode 0x%" PRIx64 "\n",
			cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid));
		cl_qmap_insert(&p_vnode->vports, vport_guid,
			       &p_vport->map_item);
	} else {
		p_found_vport = (osm_vport_t*)p_item;
		if (p_vport == p_found_vport)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9303: "
				"VPort 0x%" PRIx64 " already registered to "
				"VNode 0x%" PRIx64 "\n",
				cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid));
		else {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"A different VPort with guid 0x%" PRIx64
				" already registered to VNode 0x%" PRIx64
				" on port 0x%" PRIx64 ", Adding port 0x%" PRIx64
				" to the VNode migration port list\n",
				cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid),
				cl_ntoh64(p_found_vport->p_port->guid),
				cl_ntoh64(p_vport->p_port->guid));
			osm_vport_add_migration_port(sm, p_vnode,
						     p_vport->p_port);
			res = FALSE;
		}
	}
EXIT:
	return res;
}

void osm_virt_mgr_unregister_from_vnode(IN struct osm_sm *sm,
					IN ib_net64_t vnode_guid,
				        IN struct osm_vport *p_vport)
{
	cl_map_item_t *p_item;
	ib_net64_t vport_guid = osm_vport_get_guid(p_vport);
	osm_vnode_t *p_vnode;
	osm_vport_t *p_found_vport;

	p_vnode = osm_virt_mgr_get_vnode(sm, vnode_guid);
	OSM_ASSERT(p_vnode);

	p_item = cl_qmap_get(&p_vnode->vports, vport_guid);
	if (p_item == cl_qmap_end(&p_vnode->vports))
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9304: "
			"VPort 0x%" PRIx64 " is not registered to VNode 0x%"
			PRIx64 "\n",
			cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid));
	else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Unregistering VPort 0x%"
			PRIx64 " from VNode 0x%" PRIx64 "\n",
			cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid));
		p_found_vport = (osm_vport_t*)p_item;
		if (p_vport == p_found_vport) {
			p_item = cl_qmap_remove(&p_vnode->vports, vport_guid);
		} else {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9305: "
				"A different VPort with guid 0x%" PRIx64
				" is registered to VNode 0x%" PRIx64
				" - Ignoring unregister request\n",
				cl_ntoh64(vport_guid), cl_ntoh64(vnode_guid));
			return;
		}
	}
}

void osm_virt_mgr_reload_prtn_data(IN struct osm_sm *sm)
{
	cl_map_item_t *p_item;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_vport_t *p_vport;
	uint16_t index, top_index;
	cl_map_t pkey_map;

	cl_map_construct(&pkey_map);
	cl_map_init(&pkey_map, 16);

	OSM_LOG_ENTER(sm->p_log);

	p_item = cl_qmap_head(&sm->p_subn->port_guid_tbl);
	while (p_item != cl_qmap_end(&sm->p_subn->port_guid_tbl)) {
		p_port = (osm_port_t *)p_item;
		p_item = cl_qmap_next(p_item);
		if (osm_port_is_virt_supported(p_port)) {
			if  (!(top_index = osm_port_get_vport_top_index(p_port)))
				continue;
			/* add vport0 pkeys */
			p_physp = osm_port_get_physp(p_port);
			OSM_ASSERT(p_physp);
			add_pkeys_to_map(&p_physp->pkeys, &pkey_map);
			for (index = 1 ; index <= top_index ; index++) {
				p_vport = osm_port_get_vport(p_port, index);
				if (p_vport)
					osm_virt_mgr_apply_partitions(sm,
								      p_vport,
								      &pkey_map);
			}
		        pkey_mgr_update_peer_port(sm->p_log, sm, p_port,
				sm->p_subn->opt.part_enforce_enum, &pkey_map);
			cl_map_remove_all(&pkey_map);
		}
	}
	cl_map_destroy(&pkey_map);
	OSM_LOG_EXIT(sm->p_log);
}

void osm_virt_mgr_handle_vnd_change(IN struct osm_sm *sm,
				   IN struct osm_port *p_port,
				   IN uint16_t vport_index)
{
	osm_vport_t *p_vport = osm_port_get_vport(p_port, vport_index);

	if (p_vport) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Received trap 1144 for VPort index %u of port 0x%"
			PRIx64 " - Sending VNode Description get request\n",
			vport_index, cl_ntoh64(p_port->guid));
		send_vnd_get(sm, p_port, p_vport->vnode_guid, vport_index);
	} else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9307: "
			"Received trap 1144 for VPort which is not found "
			"(VPort index %u of port 0x%" PRIx64 " )\n",
			vport_index, cl_ntoh64(p_port->guid));
}
