/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vpi_rcv_t.
 * This object represents the VPortInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <iba/ib_types.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_virt_mgr.h>

#define FILE_ID OSM_FILE_VIRTUALIZATION_INFO_RCV_C

static boolean_t send_vps_to_all_blocks(IN osm_sm_t * sm, IN osm_port_t *p_port)
{
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t context;
	ib_api_status_t status;
	uint16_t vports_num;
	uint32_t blocks_num, block;
	boolean_t res = TRUE;
	ib_net64_t nguid;
	ib_net64_t m_key;

	OSM_LOG_ENTER(sm->p_log);

	/* First create dr path, context and m_key that can be used for all blocks */

	p_dr_path = osm_physp_get_dr_path_ptr(p_port->p_physp);

	memset(&context, 0, sizeof(context));
	context.vps_context.logical_port_guid = p_port->guid;

	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	vports_num = cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);
	blocks_num = (uint32_t)(vports_num/OSM_VPS_BLOCK_SIZE + 1);

	for (block = 0 ; block < blocks_num ; block++) {
		status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_VPORT_STATE,
				     cl_ntoh32(block), TRUE, m_key,
				     0, CL_DISP_MSGID_NONE, &context);

		if (status == IB_SUCCESS)
			osm_port_increase_virt_transactions(sm, p_port);
		else {
			nguid = osm_node_get_node_guid(p_port->p_node);
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9602: "
				"Port 0x%" PRIx64 " (Node 0x%" PRIx64
				") - Failed to send VPS get for block %u\n",
				cl_ntoh64(p_port->guid), cl_ntoh64(nguid),
				block);
			res = FALSE;
		}
	}

	OSM_LOG_EXIT(sm->p_log);
	return res;
}

static boolean_t validate_virualization_info(osm_log_t *p_log,
					     osm_port_t *p_port,
					     ib_virtualization_info_t *p_vi)
{
	boolean_t res = TRUE;
	ib_net16_t top, cap;

	if (p_port->virt_data.allocated_blocks &&
	    (p_vi->vport_cap != p_port->virt_data.virtualization_info.vport_cap)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 9603: "
			"VportCap (%u) is different than VportCap which is "
			"set in Vport (%u)\n", cl_ntoh16(p_vi->vport_cap),
			cl_ntoh16(p_port->virt_data.virtualization_info.vport_cap));
		res = FALSE;
	}

	top = cl_ntoh16(p_vi->vport_index_top);
	cap = cl_ntoh16(p_vi->vport_cap);

	if (top > cap) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 9604: "
			"VportTop (%u) is higher than VportCap (%u)\n",
			top, cap);
		res = FALSE;
	}

	if (p_vi->virt_revision != OSM_VIRT_REVISION) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 9605: "
			"Virt revision in VirtualizationInfo (%u) is different "
			"than OpenSm virt revision (%u)\n",
			p_vi->virt_revision, OSM_VIRT_REVISION);
		res = FALSE;
	}

	return res;
}

static void discover_port_vports(osm_sm_t *sm, osm_port_t *p_port,
				 ib_virtualization_info_t *p_vi)
{
	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Starting vports discovery for port 0x%" PRIx64 "\n",
		cl_ntoh64(osm_port_get_guid(p_port)));
	if (!osm_port_set_virt_info(sm, p_port, p_vi) ||
	    !send_vps_to_all_blocks(sm, p_port))
		osm_virt_mgr_enqueue_port(sm, p_port);
}

void osm_send_vi_set(osm_sm_t * sm, osm_port_t * p_port, boolean_t virt_enable,
		     uint8_t vps_change_bit, uint8_t client_rereg)
{
	osm_physp_t *p_physp;
	osm_madw_context_t madw_context;
	ib_api_status_t status;
	ib_virtualization_info_t vi = { 0 };

	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	madw_context.vi_context.set_method = TRUE;
	madw_context.vi_context.logical_port_guid = osm_port_get_guid(p_port);
	if (client_rereg)
		madw_context.vi_context.client_rereg_set = TRUE;
	else
		madw_context.vi_context.client_rereg_set = FALSE;

	ib_virtualization_info_set_virtualization_enabled(&vi, !!virt_enable);
	ib_virtualization_info_set_vport_state_change(&vi, vps_change_bit);
	ib_virtualization_info_set_vclient_rereg(&vi, client_rereg);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"sending VirtualizationInfo Set request to port "
		"0x%" PRIx64 "(virt_enabled: %s, vps_change_bit: %u)\n",
		cl_ntoh64(osm_port_get_guid(p_port)),
		virt_enable ? "TRUE" : "FALSE", vps_change_bit);

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(p_physp),
			     (uint8_t *)&vi, sizeof(vi),
			     IB_MAD_ATTR_VIRTUALIZATION_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &madw_context);

	if (status)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 9606: Failure initiating VirtualizationInfo "
			"Set request (%s) for port 0x%" PRIx64 "\n",
			ib_get_err_str(status),
			cl_ntoh64(osm_port_get_guid(p_port)));
	else
		osm_port_increase_virt_transactions(sm, p_port);
}


static void vi_set_process(osm_sm_t *sm, osm_port_t *p_port,
			   ib_virtualization_info_t *p_vi,
			   osm_vi_context_t *p_context)
{
	/* If virtualization was disabled we are done */
	if (ib_virtualization_info_get_virtualization_enabled(p_vi)) {
		if (p_context->client_rereg_set)
			/*
			 * After setting client rereg first virt discovery
			 * is done
			 */
			p_port->virt_data.first_virt_discovery = FALSE;
		else
			/*
			 * If the vi set was not for cliend rereg we can
			 * start discovery
			 */
			discover_port_vports(sm, p_port, p_vi);
	}
}

static void post_virt_discover_event(osm_sm_t *p_sm, ib_net64_t port_guid,
				     uint16_t index_top)
{
	event_data_t event_data;

	memset(&event_data, 0, sizeof(event_data_t));
	event_data.event_id = OSM_ACT_MGR_VIRT_DISCOVER;
	event_data.port_guid = port_guid;
	event_data.top_index = index_top;
	osm_act_mgr_post_event(&p_sm->activity_mgr, &event_data);
}

static void vi_get_process(osm_sm_t *sm, osm_port_t *p_port,
			   ib_virtualization_info_t *p_vi,
			   osm_vi_context_t *p_context)
{
	boolean_t virt_enabled;
	boolean_t state_change; /* vport state change bit */

	virt_enabled = ib_virtualization_info_get_virtualization_enabled(p_vi);

	if (sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_DISABLED) {
		if (virt_enabled)
			osm_send_vi_set(sm, p_port, FALSE, 0, 0);
		return;
	}

	post_virt_discover_event(sm, osm_port_get_guid(p_port),
				 p_vi->vport_index_top);

	state_change = ib_virtualization_info_get_vport_state_change(p_vi);

	/*
	 * If virt is not already enabled or change bit is not already
	 * zero we must set virtualization info
	 *
	 * If virtualization set is not required we can start discovery
	 * of vports
	 */
	if (!virt_enabled || state_change)
		osm_send_vi_set(sm, p_port, TRUE, 1, 0);
	else if (p_context->state_change_bit_check)
		p_port->virt_data.state_change_bit_check = FALSE;
	else
		discover_port_vports(sm, p_port, p_vi);
}

void osm_vi_rcv_process(IN void *context, IN void *data){
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_virtualization_info_t *p_vi, *p_old_vi;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_vi_context_t *p_context;
	ib_net64_t port_guid;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_vi_context_ptr(p_madw);
	p_vi = ib_smp_get_payload_ptr(p_smp);

	port_guid = p_context->logical_port_guid;
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);

	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9601: "
			"Received VirtualizationInfo for non existent port "
			"0x%" PRIx64 "\n",
			cl_ntoh64(port_guid));
		goto EXIT;
	}

	verbose_bypass = osm_verbose_bypass_handle_vi(sm, p_port);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received Virtualization Info "
		"%s for port GUID 0x%" PRIx64 " (Parent Node "
		"Guid 0x%" PRIx64 "), IndexTop = %u, TID 0x%" PRIx64 "\n",
		(p_context->set_method ? "SetResp" : "GetResp"),
		cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		cl_ntoh16(p_vi->vport_index_top),
		cl_ntoh64(p_smp->trans_id));

	osm_port_decrease_virt_transactions(sm, p_port);

	/* Check mad status */
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9600: "
			"Received error status for %s of virualization "
			"info mad\n",
			(p_context->set_method ? "SetResp" : "GetResp"));
		goto EXIT;
	}

	if (!validate_virualization_info(sm->p_log, p_port, p_vi))
		goto EXIT;

	/*
	 * If virtualization was off for this vport and it was turned on
	 * we need to dump this on next virtualization dump (as long as
	 * virtualization is indeed enabled)
	 */
	p_old_vi = &p_port->virt_data.virtualization_info;
	if ((sm->p_subn->opt.virt_enabled == OSM_CAPABILITY_ENABLED) &&
	    ib_virtualization_info_get_virtualization_enabled(p_vi) &&
	    !ib_virtualization_info_get_virtualization_enabled(p_old_vi))
		osm_virt_mgr_set_dump_required(&sm->virt_mgr, TRUE);


	if (p_context->set_method)
		vi_set_process(sm, p_port, p_vi, p_context);
	else
		vi_get_process(sm, p_port, p_vi, p_context);

EXIT:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);
	return;
}


