/*
 * Copyright (c) 2016 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Virtual Lid Manager
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VLID_MGR_C
#include <opensm/osm_vlid_mgr.h>

/*
 * Our goal is to set vlid <--> lid mapping as fast as possible and be able
 * to iterate over all lids and the vlids that are mapped for each one as
 * quickly as possible.
 *
 * The vlids array is used to map vlid to lid on immediate action, when new
 * vlid is mapped we set in the vlid index the value of the lid it is mapped to
 * and if the vlid is removed we simply clear its index value so the mapping
 * and removal is done in O(1).
 *
 * This function is constructing iterators for lids and their vlids, after it
 * is called once the iterators are valid as long as there are no changes in
 * the data (no vlids added or removed).
 *
 * When construction is done we would like to have:
 *
 *	A. In the 'lids' array a list of all lids which has vlids that belong
 *	   to them.
 *
 *	B. In the 'next' array in the index of a lid we should see the value
 *	   of the first vlid that belong to it, in the 'next' array in the
 *	   index of this vlid we should see the value of the next vlid which
 *	   belongs to this lid and so on
 *
 *	By having these two updated we can iterate the lids and for each one
 *	iterate the vlids which belongs to it.
 *
 *	We achieve this using the below algorithm:
 *
 * 1. We go over the vlids array and for each vlid read the lid it belongs to
 *
 * 2. If the 'last' array is empty in the lid index it means that this is the
 *    first vlid that belong tho this lid, we set the vlid value in the 'next'
 *    and in the 'last' arrays.
 *    In addition we also add the lid value to the next empty index of the
 *    'lids' array.
 *
 * 3. If the 'last' array has value set in it then this is the value of the
 *    last vlid that belong to this lis so we set the new vlid value in the
 *    'next' array in the index of the current last vlid and we set the new
 *    vlid also in the 'last' array in the index of the lid so it will be the
 *    next one to be updated if another vlid will be found for this lid
 */
void osm_vlid_mgr_construct_iter(osm_vlid_mgr_t *p_mgr)
{
	uint16_t vlid, phy_lid;
	uint16_t new_min_lid = IB_LID_UCAST_END_HO;
	uint16_t new_max_lid = 0;

	memset(p_mgr->next, 0, sizeof(p_mgr->next));
	memset(p_mgr->last, 0, sizeof(p_mgr->last));
	memset(p_mgr->lids, 0, sizeof(p_mgr->lids));
	p_mgr->lids_num = 0;

	for (vlid = p_mgr->min_vlid ; vlid <= p_mgr->max_vlid ; vlid++) {
		if (!p_mgr->vlids[vlid])
			continue;

		if (vlid < new_min_lid)
			new_min_lid = vlid;

		if (vlid > new_max_lid)
			new_max_lid = vlid;

		phy_lid = p_mgr->vlids[vlid];

		/* Did we have already vlid belong to this physical port? */
		if (p_mgr->last[phy_lid])
			/* Yes - Set this vlid as the next vlid belong to this
			 *       physcial port lid
			 */
			p_mgr->next[p_mgr->last[phy_lid]] = vlid;
		else {
			/* No - Set this vlid as the first vlid belong to this
			 *       physcial port lid
			 */
			p_mgr->next[phy_lid] = vlid;
			p_mgr->lids[p_mgr->lids_num++] = phy_lid;
		}

		/* In any case set this vlid as the last vlid belong to this
		 * physical port lid
		 */
		p_mgr->last[phy_lid] = vlid;
	}

	p_mgr->min_vlid = new_min_lid;
	p_mgr->max_vlid = new_max_lid;
	p_mgr->iter_constructed = TRUE;
}

