/*
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VNODE_DESC_RCV_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_unhealthy_ports.h>

#define VNODE_DESC_VPORT_INDEX_SHIFT		16

uint16_t osm_vnd_get_vport_index(IN const ib_mad_t *p_mad){
	return (uint16_t) (cl_ntoh32(p_mad->attr_mod) >> VNODE_DESC_VPORT_INDEX_SHIFT);
}

/*
 * WE ONLY RECEIVE GET responses
 */
void osm_vnd_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_node_desc_t *p_vnd;
	osm_vnd_context_t *p_vnd_context;
	osm_port_t *p_port;
	ib_net64_t vnode_guid, port_guid;
	uint16_t vport_index;
	osm_vnode_t *p_vnode;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_vnd = ib_smp_get_payload_ptr(p_smp);
	p_vnd_context = osm_madw_get_vnd_context_ptr(p_madw);

	port_guid = p_vnd_context->port_guid;
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	OSM_ASSERT(p_port);
	osm_port_decrease_virt_transactions(sm, p_port);

	vnode_guid = p_vnd_context->vnode_guid;
	vport_index = osm_vnd_get_vport_index(p_madw->p_mad);

	verbose_bypass = osm_verbose_bypass_handle_vnd(sm, p_port, vnode_guid);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received VNodeDesc GetResp for "
		"vnode 0x%" PRIx64 ", connected to vport index %u of port "
		"0x%" PRIx64 "\n",
		cl_ntoh64(vnode_guid), vport_index, cl_ntoh64(port_guid));

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 2700: "
			"VNodeDesc received with error\n");
		goto EXIT;
	}

	p_vnode = osm_virt_mgr_get_vnode(sm, vnode_guid);

	if (p_vnode) {
		p_vnode->vnode_desc = *p_vnd;
		osm_virt_mgr_set_dump_required(&sm->virt_mgr, TRUE);
	} else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 2701: "
			"VNode not found in virt mgr data\n");

EXIT:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);
	return;
}
