/*
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vni_rcv_t.
 * This object represents the VNodeInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <iba/ib_types.h>
#include <opensm/osm_opensm.h>
#define FILE_ID OSM_FILE_VNODE_INFO_RCV_C
#define VNODE_INFO_VNODE_INDEX_SHIFT		16

uint16_t osm_vni_get_vport_index(IN const ib_mad_t *p_mad){
	return (uint16_t) (cl_ntoh32(p_mad->attr_mod) >> VNODE_INFO_VNODE_INDEX_SHIFT);
}

static void send_vpi_get(osm_sm_t *sm, osm_port_t *p_port,
			 uint16_t vport_index, ib_vnode_info_t *p_vni)
{
	ib_net32_t attr_mod;
	ib_api_status_t status;
	osm_madw_context_t context;
	osm_dr_path_t *p_dr_path;
	ib_net64_t m_key;
	ib_net64_t port_guid;

	port_guid = osm_port_get_guid(p_port);
	memset(&context, 0, sizeof(context));
	context.vpi_context.logical_port_guid = port_guid;
	context.vpi_context.vnode_guid = p_vni->vnode_guid;
	context.vpi_context.vnode_port = p_vni->local_port_num;
	context.vpi_context.set_method = FALSE;
	p_dr_path = osm_physp_get_dr_path_ptr(p_port->p_physp);
	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	/* Set the vport index within the attr_mod in bits 16-31*/
	attr_mod = cl_hton32(((uint32_t)vport_index) << 16);

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_VPORT_INFO,
			     attr_mod, TRUE, m_key, 0, CL_DISP_MSGID_NONE,
			     &context);

	if (status == IB_SUCCESS)
		osm_port_increase_virt_transactions(sm, p_port);
	else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1201: "
			"Failed sending VPI get request for vport index %u "
			"of port 0x%" PRIx64 "\n",
			vport_index, cl_ntoh64(port_guid));
}

void osm_vni_rcv_process(IN void *context, IN void *data){
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_vnode_info_t *p_vni;
	osm_vni_context_t *p_vni_context;
	osm_port_t *p_port;
	ib_net64_t port_guid;
	uint16_t vport_index;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_vni = ib_smp_get_payload_ptr(p_smp);
	p_vni_context = osm_madw_get_vni_context_ptr(p_madw);

	port_guid = p_vni_context->port_guid;
	vport_index = osm_vni_get_vport_index(p_madw->p_mad);

	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	OSM_ASSERT(p_port);
	osm_port_decrease_virt_transactions(sm, p_port);

	verbose_bypass = osm_verbose_bypass_handle_vni(sm, p_port,
						       p_vni->vnode_guid);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received VNodeInfo getResp, Port 0x%"
		PRIx64 ", Vport Index %u\n", cl_ntoh64(port_guid), vport_index);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1200: "
			"Received error on VNodeInfo getResp, Port 0x%" PRIx64
			", Vport Index %u\n",
			cl_ntoh64(port_guid), vport_index);
		goto EXIT;
	}

	if (!osm_virt_mgr_add_vnode(sm, p_vni, p_port, vport_index))
		goto EXIT;

	send_vpi_get(sm, p_port, vport_index, p_vni);

EXIT:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);
	return;
}
