/*
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VPKEY_RCV_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_vport.h>

#define VPKEY_INDEX_SHIFT		16

uint16_t osm_vpkey_get_vport_index(IN const ib_mad_t *p_mad){
	return (uint16_t) (cl_ntoh32(p_mad->attr_mod) >> VPKEY_INDEX_SHIFT);
}

/*
 * WE ONLY RECEIVE GET or SET responses
 */
void osm_vpkey_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_pkey_table_t *p_pkey_tbl;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_node_t *p_node;
	osm_pkey_context_t *p_context;
	ib_net64_t port_guid, vport_guid;
	uint16_t block_num, max_block;
	uint16_t vport_index;
	osm_vport_t *p_vport;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);

	p_context = osm_madw_get_pkey_context_ptr(p_madw);
	p_pkey_tbl = ib_smp_get_payload_ptr(p_smp);

	port_guid = p_context->port_guid;

	cl_plock_excl_acquire(sm->p_lock);
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	OSM_ASSERT(p_port);
	osm_port_decrease_virt_transactions(sm, p_port);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		if (p_context->set_method)
			osm_hm_set_by_node(sm, osm_port_get_node(p_port),
					   OSM_HM_COND_SET_ERR);

		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	block_num = (uint16_t) ((cl_ntoh32(p_smp->attr_mod)) & 0x0000FFFF);

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_CA) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7200: "
			"Received VPkey tbl response from non CA node 0x%"
			PRIx64 " - Ignoring\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)));
		goto Exit;
	}

	vport_index = osm_vpkey_get_vport_index(p_madw->p_mad);

	if (!vport_index) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7203: "
			"Received VPkey table get response for "
			"vport index %u of port 0x%" PRIx64
			", VPkey is not relevant for index 0 - Ignoring "
			"response\n", vport_index, cl_ntoh64(p_port->guid));
		goto Exit;
	}

	p_vport = osm_port_get_vport(p_port, vport_index);
	vport_guid = osm_vport_get_guid(p_vport);

	if (!p_vport) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7201: "
			"VPort 0x%" PRIx64 " (vport index %u of port 0x%"
			PRIx64 ") not found - Ignoring response\n",
			cl_ntoh64(vport_guid), vport_index,
			cl_ntoh64(p_port->guid));
		goto Exit;
	}

	max_block = osm_vport_get_max_pkey_block(sm, p_vport);
	if (PF(block_num > max_block)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7202: "
			"VPort 0x%" PRIx64 " Received VPkey tbl response for "
			"illegal block number %u (max block is %u) - Ignoring "
			"response\n",
			cl_ntoh64(vport_guid), block_num, max_block);
		goto Exit;
	}

	/*
	 * We do not care if this is a result of a set or get -
	 *  all we want is to update the subnet.
	 */
	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		"Got GetResp(VPKey) block:%u VPort GUID 0x%"
		PRIx64 ", VPort index %u of port 0x%" PRIx64 ", TID 0x%" PRIx64
		"\n", block_num, cl_ntoh64(vport_guid), p_vport->index,
		cl_ntoh64(p_vport->p_port->guid), cl_ntoh64(p_smp->trans_id));

	osm_dump_vpkey_block_v2(sm->p_log, vport_guid, block_num, vport_index,
			       p_pkey_tbl, FILE_ID, OSM_LOG_DEBUG);

	osm_vport_set_pkey_tbl(sm->p_log, sm->p_subn,
			       p_vport, p_pkey_tbl, block_num,
			       p_context->set_method);

Exit:
	cl_plock_release(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
