/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of virtual ports (vports) support
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VPORT_C
#include <opensm/osm_vport.h>
#include <opensm/osm_opensm.h>

/* Definitions */

/* Forward references */

uint16_t osm_vport_get_max_pkey_block(IN struct osm_sm *p_sm,
				      IN osm_vport_t *p_vport)
{
	osm_vnode_t *p_vnode;
	uint16_t max_block = 0;

	p_vnode = osm_virt_mgr_get_vnode(p_sm, p_vport->vnode_guid);
	if (!p_vnode) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 5700: "
				"VNode 0x%" PRIx64 " for VPort 0x%" PRIx64
				" not found\n", cl_ntoh64(p_vport->vnode_guid),
				cl_ntoh64(osm_vport_get_guid(p_vport)));
		goto EXIT;
	}

	max_block = (cl_ntoh16(p_vnode->vnode_info.partition_cap) +
			IB_NUM_PKEY_ELEMENTS_IN_BLOCK - 1) /
				IB_NUM_PKEY_ELEMENTS_IN_BLOCK;
EXIT:
	return max_block;
}

void osm_vport_get_max_pkeys_and_blocks(IN struct osm_sm *p_sm,
					IN osm_vport_t *p_vport,
					OUT uint16_t *p_max_pkeys,
					OUT uint16_t *p_max_blocks)
{
	osm_vnode_t *p_vnode;

	p_vnode = osm_virt_mgr_get_vnode(p_sm, p_vport->vnode_guid);
	if (!p_vnode) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 5702: "
				"VNode 0x%" PRIx64 " for VPort 0x%" PRIx64
				" not found\n", cl_ntoh64(p_vport->vnode_guid),
				cl_ntoh64(osm_vport_get_guid(p_vport)));
		*p_max_pkeys = 0;
		*p_max_blocks = 0;
		return;
	}
	*p_max_pkeys = cl_ntoh16(p_vnode->vnode_info.partition_cap);

	*p_max_blocks = (*p_max_pkeys + IB_NUM_PKEY_ELEMENTS_IN_BLOCK - 1) /
			IB_NUM_PKEY_ELEMENTS_IN_BLOCK;
}

void osm_vport_set_pkey_tbl(IN osm_log_t * p_log, IN struct osm_subn *p_subn,
			    IN osm_vport_t * p_vport,
			    IN ib_pkey_table_t * p_pkey_tbl,
			    IN uint16_t block_num,
			    IN boolean_t is_set)
{
	uint16_t max_blocks;

	max_blocks = osm_vport_get_max_pkey_block(&p_subn->p_osm->sm, p_vport);

	if (block_num >= max_blocks) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 5701: "
			"Got illegal update for block number:%u max:%u "
			"for VPort 0x%" PRIx64 " Vport index %u in port 0x%"
			PRIx64 "\n", block_num, max_blocks,
			cl_ntoh64(osm_vport_get_guid(p_vport)),
			p_vport->index, cl_ntoh64(p_vport->p_port->guid));
		return;
	}

	/* decrement block received counter */
	if(!is_set)
		p_vport->pkeys.rcv_blocks_cnt--;
	osm_pkey_tbl_set(&p_vport->pkeys, block_num, p_pkey_tbl,
			 p_subn->opt.allow_both_pkeys);
}

static void update_pkey_entry(IN osm_sm_t * sm,
			      IN osm_vport_t * p_vport,
			      IN ib_pkey_table_t * block,
			      IN uint16_t block_index)
{
	osm_madw_context_t context;
	uint32_t attr_mod_ho;
	ib_net64_t m_key;
	osm_dr_path_t path;
	ib_api_status_t status;
	osm_port_t *p_port = p_vport->p_port;

	context.pkey_context.port_guid = p_port->guid;
	context.pkey_context.set_method = TRUE;

	path = *osm_physp_get_dr_path_ptr(p_port->p_physp);
	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	attr_mod_ho = block_index | (p_vport->index << 16);

	status = osm_req_set(sm, &path,
			     (uint8_t *) block, sizeof(*block),
			     IB_MAD_ATTR_V_PKEY_TABLE,
			     cl_hton32(attr_mod_ho), FALSE, m_key,
			     0, CL_DISP_MSGID_NONE, &context);

	if (status == IB_SUCCESS) {
		osm_port_increase_virt_transactions(sm, p_port);
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Sent PKeyTable set request to VPort 0x%" PRIx64
			", VPort index %u of port 0x%" PRIx64 " (%s)\n",
			cl_ntoh64(osm_vport_get_guid(p_vport)), p_vport->index,
			cl_ntoh64(p_port->guid), ib_get_err_str(status));
	} else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 5703: "
			"Failure initiating PKeyTable set request to "
			"VPort 0x%" PRIx64 ", VPort index %u of port "
			"0x%" PRIx64 " (%s)\n",
			cl_ntoh64(osm_vport_get_guid(p_vport)), p_vport->index,
			cl_ntoh64(p_port->guid), ib_get_err_str(status));
}

void osm_vport_update_pkey_tbl(IN osm_sm_t * sm, IN osm_vport_t *p_vport,
			       IN char *owner_desc)
{
	ib_pkey_table_t *block, *new_block;
	uint16_t block_index;
	ib_pkey_table_t empty_block;
	uint16_t num_of_blocks;
	osm_pkey_tbl_t *p_pkey_tbl = &p_vport->pkeys;

	memset(&empty_block, 0, sizeof(ib_pkey_table_t));
	num_of_blocks = osm_pkey_tbl_get_num_blocks(p_pkey_tbl);

	for (block_index = 0; block_index < num_of_blocks; block_index++) {
		block = osm_pkey_tbl_block_get(p_pkey_tbl, block_index);
		new_block = osm_pkey_tbl_new_block_get(p_pkey_tbl, block_index);
		if (!new_block)
			new_block = &empty_block;
		if (block && !memcmp(new_block, block, sizeof(*block)))
			continue;

		update_pkey_entry(sm, p_vport, new_block, block_index);
	}

}

ib_net16_t osm_vport_calc_lid(IN osm_vport_t *p_vport)
{
	osm_vport_t *p_lid_vport;
	ib_vport_info_t *p_vpi;
	uint16_t lid_vport_index;

	p_vpi = &p_vport->vport_info;

	if (ib_vport_info_get_lid_required(p_vpi))
		return p_vpi->vport_lid;

	lid_vport_index = cl_ntoh16(p_vpi->lid_by_vport_index);
	if (lid_vport_index) {
		p_lid_vport = osm_port_get_vport(p_vport->p_port, lid_vport_index);
		if (p_lid_vport)
			return osm_vport_get_lid(p_lid_vport);
		return 0;
	}

	return  osm_port_get_base_lid(p_vport->p_port);
}

ib_net16_t osm_vport_get_lid_by_aguid(IN const struct osm_alias_guid *p_aguid)
{
	osm_vport_t *p_vport;

	p_vport = osm_port_get_vport(p_aguid->p_base_port, p_aguid->vport_index);
	if (p_vport)
		return osm_vport_calc_lid(p_vport);

	return 0;
}

struct osm_alias_guid *osm_vport_get_aguid(IN struct osm_subn *p_subn,
					   IN osm_vport_t *p_vport)
{
	cl_qmap_t  *alias_guid_tbl = &p_subn->alias_port_guid_tbl;
	ib_net64_t vport_guid = p_vport->vport_info.port_guid;
	osm_alias_guid_t *p_alias_guid;

	p_alias_guid =(osm_alias_guid_t*)cl_qmap_get(alias_guid_tbl, vport_guid);

	if (p_alias_guid == (osm_alias_guid_t*)cl_qmap_end(alias_guid_tbl)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 5704: "
			"Vport guid 0x%" PRIx64 " not found in "
			"alias_port_guid_tbl (index %" PRIu16 " of port 0x%"
			PRIx64 ")\n", cl_ntoh64(vport_guid), p_vport->index,
			cl_ntoh64(p_vport->p_port->guid));
		p_alias_guid = NULL;
	}

	return p_alias_guid;
}
