/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vpi_rcv_t.
 * This object represents the VPortInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <iba/ib_types.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_vport.h>
#define FILE_ID OSM_FILE_VPORT_INFO_RCV_C
#define VPORT_INFO_VPORT_INDEX_SHIFT		16

uint16_t osm_vpi_get_vport_index(IN const ib_mad_t *p_mad){
	return (uint16_t) (cl_ntoh32(p_mad->attr_mod) >> VPORT_INFO_VPORT_INDEX_SHIFT);
}

static void get_pkey_table(osm_sm_t *sm, osm_vport_t *p_vport)
{
	osm_madw_context_t context;
	ib_api_status_t status;
	osm_dr_path_t path;
	ib_net64_t m_key;
	uint16_t block_num, max_blocks;
	uint32_t attr_mod_ho;
	osm_port_t *p_port = p_vport->p_port;

	OSM_LOG_ENTER(sm->p_log);

	path = *osm_physp_get_dr_path_ptr(p_port->p_physp);
	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	context.pkey_context.port_guid = p_port->guid;
	context.pkey_context.set_method = FALSE;

	max_blocks = osm_vport_get_max_pkey_block(sm, p_vport);

	p_vport->pkeys.rcv_blocks_cnt = max_blocks;
	for (block_num = 0; block_num < max_blocks; block_num++) {
		attr_mod_ho = block_num |
			      (p_vport->index << VPORT_INFO_VPORT_INDEX_SHIFT);

		status = osm_req_get(sm, &path, IB_MAD_ATTR_V_PKEY_TABLE,
				     cl_hton32(attr_mod_ho), FALSE,
				     m_key, 0, CL_DISP_MSGID_NONE, &context);

		if (status == IB_SUCCESS)
			osm_port_increase_virt_transactions(sm, p_port);
		else {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1503: "
				"Failure initiating PKeyTable get request to "
				"VPort 0x%" PRIx64 ", VPort index %u of port "
				"0x%" PRIx64 " (%s)\n",
				osm_vport_get_guid(p_vport), p_vport->index,
				cl_ntoh64(p_port->guid), ib_get_err_str(status));
			goto Exit;
		}
	}

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Sent PKeyTable get request to "
		"%u blocks of VPort 0x%" PRIx64 ", VPort index %u of port "
		"0x%" PRIx64 " (%s)\n",
		max_blocks, cl_ntoh64(osm_vport_get_guid(p_vport)), p_vport->index,
		cl_ntoh64(p_port->guid), ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

/* Handle SetResponse */
static void vpi_rcv_process_set(IN osm_sm_t * p_sm, IN osm_port_t *p_port,
				IN uint16_t vport_index, IN osm_vport_t *p_vport,
				IN ib_smp_t *p_smp, IN ib_vport_info_t *p_vpi,
				uint8_t state, osm_vpi_context_t *p_vpi_context)
{
	OSM_ASSERT(p_vport);

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Received Vport Info SetResp "
		"for port 0x%" PRIx64 " (Node 0x%" PRIx64 "), Index %u"
		", State %u, LidRequired: %" PRIu8 ", LidVal: %" PRIu16
		", LidByIndex: %" PRIu16 "\n", cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		vport_index, state, ib_vport_info_get_lid_required(p_vpi),
		cl_ntoh16(p_vpi->vport_lid), cl_ntoh16(p_vpi->lid_by_vport_index));

	if (state == IB_LINK_DOWN) {
		OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE, "VPort %u of Port 0x%"
			PRIx64 " (Node 0x%" PRIx64 ") found in state DOWN - "
			"Deleting VPort\n", vport_index,
			cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)));
		osm_port_delete_vport(p_sm, p_port, vport_index);
	} else
		osm_port_update_vport(p_sm, p_port, vport_index, p_vpi,
				      p_vpi_context);

	return;
}


/* Handle GetResponse */
static void vpi_rcv_process_get(IN osm_sm_t * p_sm, IN osm_port_t *p_port,
			       IN uint16_t vport_index, IN osm_vport_t *p_vport,
			       IN ib_vport_info_t *p_vpi, uint8_t state,
			       osm_vpi_context_t *p_vpi_context)
{
	uint8_t vport_had_lid = 0;

	OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "Received Vport Info GetResp "
		"for port 0x%" PRIx64 " (Node 0x%" PRIx64 "), Index %u"
		", State %u, LidRequired: %" PRIu8 ", LidVal: %" PRIu16
		", LidByIndex: %" PRIu16 "\n", cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		vport_index, state, ib_vport_info_get_lid_required(p_vpi),
		cl_ntoh16(p_vpi->vport_lid), cl_ntoh16(p_vpi->lid_by_vport_index));

	if (state == IB_LINK_DOWN) {
		/*
		 * We do not send VPort Info to Vports in state DOWN so
		 * this shouldn't have happen, seems like VPort went
		 * down during the short time that passed since we
		 * received VPort State and sent VPort Info get to all
		 * Vports which were not DOWN
		 */
		OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE, "VPort %u of Port 0x%"
			PRIx64 " (Node 0x%" PRIx64 ") - responded to "
			"VPort Info get with state DOWN\n",
			vport_index, cl_ntoh64(p_port->guid),
			cl_ntoh64(osm_node_get_node_guid(p_port->p_node)));
		if (p_vport)
			osm_port_delete_vport(p_sm, p_port, p_vport->index);
	} else {
		if (p_vport) {
			vport_had_lid =
			    ib_vport_info_get_lid_required(&p_vport->vport_info);
			if (p_vpi->port_guid == p_vport->vport_info.port_guid)
				osm_port_update_vport(p_sm, p_port, vport_index,
						      p_vpi, p_vpi_context);
			else {
				OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE,
					"VPort %u of Port 0x%" PRIx64 " - guid "
					"changed from 0x%" PRIx64 " to 0x%"
					PRIx64 ", Replacing VPort\n",
					vport_index, cl_ntoh64(p_port->guid),
					cl_ntoh64(p_vport->vport_info.port_guid),
					cl_ntoh64(p_vpi->port_guid));
					osm_port_delete_vport(p_sm, p_port,
							      p_vport->index);
					/*  Rare case - GUID replaced
					    - Invalidate vLID cache
					*/
					p_vport = osm_port_create_vport(p_sm,
									p_port,
									vport_index,
									p_vpi,
									p_vpi_context);
			}
		} else
			p_vport = osm_port_create_vport(p_sm, p_port,
							vport_index, p_vpi,
							p_vpi_context);

		/* VPort 0 does not have a pkey tbl of its own */
		if (p_vport && vport_index)
			get_pkey_table(p_sm, p_vport);

		/*
		 * If the vport does not require a lid we need to check if it
		 * already assigned to a lid and if so release it
		 */
		if (p_vport && p_vport->index && vport_had_lid &&
		    !ib_vport_info_get_lid_required(p_vpi)) {
			/*
			 * Vport not down and not require lid so we take the
			 * lid from the vport
			 * no need to add to the vLID cache
			 */
			osm_lid_mgr_clear_vport_lid(&p_sm->lid_mgr, p_port,
						    p_vport, TRUE);
		}
	}
}

static void vpi_rcv_check_and_fix_lid(osm_log_t * log, ib_vport_info_t * p_vpi)
{
	if (ib_vport_info_get_lid_required(p_vpi) &&
	    cl_ntoh16(p_vpi->vport_lid) > IB_LID_UCAST_END_HO) {
		OSM_LOG(log, OSM_LOG_VERBOSE,
			"Got invalid virtual port LID %u from GUID "
			"0x%" PRIx64 ", Corrected to 0\n",
			cl_ntoh16(p_vpi->vport_lid),
			cl_ntoh64(p_vpi->port_guid));
		p_vpi->vport_lid = 0;
	}
}

void osm_vpi_rcv_process(IN void *context, IN void *data){
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_vport_info_t *p_vpi;
	osm_vpi_context_t *p_vpi_context;
	osm_port_t *p_port;
	ib_net64_t port_guid;
	osm_vport_t *p_vport;
	uint16_t vport_index;
	uint8_t state;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_vpi = ib_smp_get_payload_ptr(p_smp);
	p_vpi_context = osm_madw_get_vpi_context_ptr(p_madw);

	port_guid = p_vpi_context->logical_port_guid;
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	OSM_ASSERT(p_port);
	osm_port_decrease_virt_transactions(sm, p_port);

	vport_index = osm_vpi_get_vport_index(p_madw->p_mad);

	verbose_bypass = osm_verbose_bypass_handle_vpi(sm, p_port,
						       p_vpi->port_guid);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1500: "
			"Received error for %s of VPORT Info, Index %u\n",
			(p_vpi_context->set_method ? "SetResp" : "GetResp"),
			vport_index);
		goto EXIT;
	}

	state = ib_vport_info_get_vport_state(p_vpi);
	p_vport = osm_port_get_vport(p_port, vport_index);

	if (!state) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1501: "
			"Received %s for VportInfo with unexpected state 0, "
			"port 0x%" PRIx64 ", vport index %u\n",
			(p_vpi_context->set_method ? "SetResp" : "GetResp"),
			cl_ntoh64(p_port->guid), vport_index);
		goto EXIT;
	}

	if (!p_vpi->port_guid) {
		if (p_vport) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO, "Received VPortInfo "
				"with vport guid 0 for existing vport "
				"index %u of port 0x%" PRIx64 " - Deleting "
				"vport (current vport guid 0x%" PRIx64 ")\n",
				vport_index, cl_ntoh64(port_guid),
				cl_ntoh64(p_vport->vport_info.port_guid));
			osm_port_delete_vport(sm, p_port, vport_index);
		} else
			OSM_LOG(sm->p_log, OSM_LOG_INFO, "Received VPortInfo "
				"with vport guid 0 for non-existing vport "
				"index %u of port 0x%" PRIx64 " - Ignoring\n",
				vport_index, cl_ntoh64(port_guid));

		goto EXIT;
	}


	/*
	 * Check that virtual port lid is valid, and if not, set it to 0.
	 */
	vpi_rcv_check_and_fix_lid(sm->p_log, p_vpi);

	if (p_vpi_context->set_method)
		vpi_rcv_process_set(sm, p_port, vport_index, p_vport, p_smp,
				    p_vpi, state, p_vpi_context);
	else {
		if ((vport_index == 0) && (p_vpi->port_guid != p_port->guid)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1502: "
				"Port 0x%" PRIx64 " - received VportInfo "
				"GetResp for index0 with guid 0x%" PRIx64
				" which is different from port guid\n",
				cl_ntoh64(p_port->guid),
				cl_ntoh64(p_vpi->port_guid));
			goto EXIT;
		}
		vpi_rcv_process_get(sm, p_port, vport_index, p_vport, p_vpi,
				    state, p_vpi_context);
	}

EXIT:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);
	return;
}
