/*
 * Copyright (c) 2014 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vps_rcv_t.
 * This object represents the VPortState Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <iba/ib_types.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_vport.h>
#include <opensm/osm_virt_mgr.h>
#define FILE_ID OSM_FILE_VPORT_STATE_RCV_C
#define VPORT_STATE_BLOCK_NUMBER_MASK		0x000001FF

uint8_t osm_vps_get_index_state(IN uint8_t index, IN const ib_vport_state_t *p_vps)
{
	uint8_t shift = (index % 2) ? IB_VPORT_STATE_ODD_STATE_SHIFT : IB_VPORT_STATE_EVEN_STATE_SHIFT;
	uint8_t mask = (index % 2) ? IB_VPORT_STATE_ODD_STATE_MASK : IB_VPORT_STATE_EVEN_STATE_MASK;
	return (p_vps->vports_states[index / 2] & mask) >> shift;
}

uint16_t osm_vport_state_get_block_number(IN const ib_mad_t *p_mad)
{
	return (uint16_t) (cl_ntoh32(p_mad->attr_mod) & VPORT_STATE_BLOCK_NUMBER_MASK);
}

static void send_vni_get(osm_sm_t *sm, osm_port_t *p_port,
			 uint16_t vport_index,
			 osm_madw_context_t *p_context,
			 osm_dr_path_t *p_dr_path,
			 ib_net64_t m_key)
{
	ib_net32_t attr_mod;
	ib_api_status_t status;
	osm_physp_t *p_physp;

	p_physp = osm_port_get_physp(p_port);
	OSM_ASSERT(p_physp);

	/* Set vport index within the attr_mod in bits 16-31*/
	attr_mod = cl_hton32(((uint32_t)vport_index) << 16);

	status = osm_req_get(sm, p_dr_path, IB_MAD_ATTR_VNODE_INFO, attr_mod,
			     FALSE, m_key, 0, CL_DISP_MSGID_NONE, p_context);

	if (status == IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Sent VNI get request to "
			"port 0x%" PRIx64 ", vport %u\n",
			cl_ntoh64(osm_port_get_guid(p_port)), vport_index);
		osm_port_increase_virt_transactions(sm, p_port);
	} else
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 1503: "
			"Failed sending VNodeInfo get request to port 0x%"
			PRIx64 ", vport %u\n",
			cl_ntoh64(osm_port_get_guid(p_port)), vport_index);
}

static void handle_vport(osm_sm_t *sm, osm_port_t *p_port,
			 ib_vport_state_t *p_vps,
			 uint16_t index_within_block,
			 uint16_t index_within_port,
			 osm_madw_context_t *p_context,
			 osm_dr_path_t *p_dr_path,
			 ib_net64_t m_key)
{
	uint8_t old_state, new_state;
	osm_vport_t *p_vport;
	boolean_t standalone_mode = sm->p_subn->opt.stand_alone != OSM_STANDALONE_NA && 
				    sm->master_sm_guid == p_port->guid;

	p_vport = osm_port_get_vport(p_port, index_within_port);

	if (p_vport)
		old_state = ib_vport_info_get_vport_state(&p_vport->vport_info);
	else
		old_state = IB_LINK_DOWN;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Handling VPort %u of port 0x%"
		PRIx64 " (Addr = %p)\n", index_within_port,
		cl_ntoh64(p_port->guid), p_vport);

	new_state = osm_vps_get_index_state(index_within_block, p_vps);

	if ((old_state == IB_LINK_DOWN) && (new_state == IB_LINK_DOWN) && !standalone_mode)
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort Index %u: "
			"Old State = %u, New State = %u - Nothing to do\n",
			index_within_port, old_state, new_state);
	else if ((old_state == IB_LINK_ACTIVE) && (new_state == IB_LINK_ACTIVE)) {
		if (p_port->virt_data.first_virt_discovery) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort Index %u: "
				"Old State = %u, New State = %u - first "
				"virt discovery, Sending VPI get\n",
				index_within_port, old_state, new_state);

			send_vni_get(sm, p_port, index_within_port, p_context,
				     p_dr_path, m_key);
		} else
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort Index %u: "
				"Old State = %u, New State = %u - not first "
				"virt discovery, Nothing to do\n",
				index_within_port, old_state, new_state);
	} else {
		if ((new_state == IB_LINK_DOWN) && p_vport && !standalone_mode) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort Index %u: "
				"Old State = %u, New State = %u - Deleting "
				"Vport\n", index_within_port, old_state,
				new_state);
			osm_port_delete_vport(sm, p_port, index_within_port);
			goto EXIT;
		}

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "VPort Index %u: "
			"Old State = %u, New State = %u - Sending VPI get\n",
			index_within_port, old_state, new_state);

		send_vni_get(sm, p_port, index_within_port, p_context,
			     p_dr_path, m_key);
	}
EXIT:
	return;
}

void osm_vps_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_vport_state_t *p_vps;
	osm_port_t *p_port;
	osm_vps_context_t *p_vps_context;
	osm_madw_context_t mad_cntx;
	osm_dr_path_t *p_path;
	ib_net64_t port_guid;
	uint16_t i, vport_index, block_number;
	uint16_t top, top_block_index;
	ib_net64_t m_key;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_vps = ib_smp_get_payload_ptr(p_smp);
	p_vps_context = osm_madw_get_vps_context_ptr(p_madw);

	port_guid = p_vps_context->logical_port_guid;
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	OSM_ASSERT(p_port);
	osm_port_decrease_virt_transactions(sm, p_port);

	verbose_bypass = osm_verbose_bypass_handle_vps(sm, p_port);

	block_number = osm_vport_state_get_block_number(p_madw->p_mad);

	/* Check mad status */
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9301: Received "
			"error status for GetResp of vport state mad, "
			"Block %u\n", block_number);
		goto EXIT;
	}

	/* Prepare context, dr_path and m_key for vnode info requests */
	memset(&mad_cntx, 0, sizeof(mad_cntx));
	mad_cntx.vni_context.port_guid = port_guid;
	p_path = osm_physp_get_dr_path_ptr(p_port->p_physp);
	m_key = ib_port_info_get_m_key(&p_port->p_physp->port_info);

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received VPort State (block %u) "
		"for port GUID 0x%" PRIx64 " (Parent Node Guid 0x%" PRIx64
		"), TID 0x%" PRIx64 "\n", block_number,	cl_ntoh64(p_port->guid),
		cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
		cl_ntoh64(p_smp->trans_id));

	/* vport_index represent the global index of the vport in the port */
	vport_index = block_number * OSM_VPS_BLOCK_SIZE;

	/* Find the highest index in this specific block that is still within
	 * the top limits (handle case that top is in the midle of this block)
	 * because we don't want to handle index that is higher than top
	 *
	 * vport_index is currently the index in the port of the first
	 * index in this block (e.g: if the size of a block is 128 vports than
	 * the vport_index of the first index in the second block is 128)
	 */
	top = cl_ntoh16(p_port->virt_data.virtualization_info.vport_index_top);
	if ((vport_index + OSM_VPS_BLOCK_SIZE) > top)
		top_block_index = top - vport_index + 1;
	else
		top_block_index = OSM_VPS_BLOCK_SIZE;

	/* Handle every internal index in the current block */
	for (i = 0 ; i < top_block_index ; i++) {
		handle_vport(sm, p_port, p_vps, i, vport_index, &mad_cntx,
			     p_path, m_key);
		vport_index++;
	}
EXIT:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);
	return;
}

