/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be before all other includes */
#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ARMGR_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "opensm/osm_armgr.hpp"
#include "opensm/osm_ar_algorithm.hpp"
#include "opensm/osm_ar_kdor_algorithm.hpp"

#define GROUP_BLOCKS_NUMBER(calculated_groups_number , ar_info)  \
        (((calculated_groups_number*(ar_info.sub_grps_active + 1)) + \
          (AR_GROUP_TABLE_BLOCK_SIZE-1)) / AR_GROUP_TABLE_BLOCK_SIZE)

#define IS_GENERATE_ARN(master_db, sw_db_entry) \
        (master_db.m_arn_enable && \
	 (sw_db_entry.m_ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_ARN_SUPPORTED) && \
         ib_ar_info_get_grp_tbl_cpy_sup(&sw_db_entry.m_ar_info))

#define IS_GENERATE_FRN(master_db, sw_db_entry) \
        (master_db.m_frn_enable && \
	 (sw_db_entry.m_ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_FRN_SUPPORTED) && \
         ib_ar_info_get_grp_tbl_cpy_sup(&sw_db_entry.m_ar_info))


//Bit 0 = 2 sub groups (1,3,5,7)
//Bit 1 = 3 sub groups (2,3,6,7)
//Bit 2 = 4 sub groups (4,5,6,7)
//Bits 3 to 7 are reserved
                               //0,1,2,3,4,5,6,7
//u_int8_t DFSubGrpsRequired[7] = {1,2,3,1,3,1,2,1};

extern "C" {
	int wait_for_pending_transactions(osm_stats_t * stats);
}


using namespace OSM;


/****************************************************/
/**************** private methods *******************/
/****************************************************/
//============================================================================//
void OSMAdaptiveRoutingManager::WaitForPendingTransactions()
{

	osm_subn_t *p_subn = m_p_osm_subn;
	osm_sm_t *sm = &p_subn->p_osm->sm;

	CL_PLOCK_RELEASE(sm->p_lock);

	wait_for_pending_transactions(&p_subn->p_osm->stats);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
}


void OSMAdaptiveRoutingManager::SetRequiredARInfo(ARSWDataBaseEntry &db_entry)
{
	CLEAR_STRUCT(db_entry.m_required_ar_info);

	if (m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) {
		u_int16_t dfp_en_vl_mask = m_master_db.m_dfp_en_vl_mask;
		if (m_master_db.m_en_sl_mask == OSM_ALL_SL_VL_AR_ENABLED) {
			//all sl/vl are AR enabled
			dfp_en_vl_mask = OSM_ALL_SL_VL_AR_ENABLED;
		}

		if (db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) {
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
			db_entry.m_required_ar_info.enable_by_sl_mask =
				cl_ntoh16((u_int16_t)(dfp_en_vl_mask & AR_ENABLE_BY_VL_MASK));
		} else if (dfp_en_vl_mask != OSM_ALL_SL_VL_AR_ENABLED) {
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
			db_entry.m_required_ar_info.enable_by_sl_mask =
				cl_ntoh16(dfp_en_vl_mask);
		}
		//(when using DF, one can not disable AR on a specific switch )
		db_entry.m_required_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_AR_ENABLED;
		//currently DF supports only 2 sub groups
		ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 1);
	} else {
		if (m_master_db.m_en_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) {
			if (!(db_entry.m_ar_info.enabled_and_cap_fields &
						IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8505: Switch GUID 0x%" PRIx64 ", LID %u, " \
					   "do not support by_sl_cap.\n",
					   db_entry.m_general_sw_info.m_guid,
					   db_entry.m_general_sw_info.m_lid);
				db_entry.m_option_on = false;
			} else {
				db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
				db_entry.m_required_ar_info.enable_by_sl_mask =
					cl_ntoh16(m_master_db.m_en_sl_mask);
			}
		}
		ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 0);
	}

	if (db_entry.m_option_on)
		db_entry.m_required_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_AR_ENABLED;
	else
		db_entry.m_required_ar_info.enable_features &= ~IB_AR_INFO_FEATURE_MASK_AR_ENABLED;

	if (m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) {
		if (!(db_entry.m_ar_info.enabled_and_cap_fields &
					IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8506: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "do not support by_transport_disable.\n",
				   db_entry.m_general_sw_info.m_guid,
				   db_entry.m_general_sw_info.m_lid);
			db_entry.m_option_on = false;
		} else {
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
				IB_AR_INFO_CAP_MASK_BY_TRANS_CAP;
			/* the bits 4-7 of 'by transport disable' field are reserved */
			db_entry.m_required_ar_info.by_trns_disable =
				(~m_master_db.m_en_tr_mask) & 0xF;
		}
	}

	db_entry.m_required_ar_info.ageing_time_value =
		cl_hton32(db_entry.m_ageing_time_value);

	//RN
	if (IS_GENERATE_ARN(m_master_db, db_entry) ||
	    IS_GENERATE_FRN(m_master_db, db_entry))
		db_entry.m_required_ar_info.enable_features |=
			IB_AR_INFO_FEATURE_MASK_RN_XMIT_ENABLED;
	else if (m_master_db.m_arn_enable || m_master_db.m_frn_enable) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8507: Switch GUID 0x%" PRIx64 ", LID %u, " \
			   "do not support routing notification.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
	}

	if ((db_entry.m_ar_info.enable_features &
		IB_AR_INFO_FEATURE_MASK_FR_SUPPORTED) &&
	    m_master_db.m_fr_enable) {
		db_entry.m_required_ar_info.enable_features |=
			IB_AR_INFO_FEATURE_MASK_FR_ENABLED;
	}
}

//============================================================================//
int OSMAdaptiveRoutingManager::Init()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (m_is_init)
		return 0;

	/*
	 * init m_ar_algorithm
	*/

	if (m_master_db.m_ar_algorithm == AR_ALGORITHM_KDOR_HC) {

		m_ar_algorithm = new ArHcAlgorithm(m_p_osm_log, m_sw_db.m_sw_map, *this);

		if (m_ar_algorithm == NULL) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "Failed to create HC algorithm\n");
			throw 1;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "Set AR KDOR HC algorithm\n");
	}

	if (m_sw_db.Init()) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8510: ARSWDataBase init failed.\n");
		throw 1;
	}

	if (m_thread_pool.Init(m_p_osm_subn->opt.routing_threads_num))
		throw 1;

	m_port_groups_calculator.Init();

	m_is_init = true;
	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

//============================================================================//

/***************/
void OSMAdaptiveRoutingManager::UpdateSW(const ARGeneralSWInfo& general_sw_info)
{
	osm_switch_t *p_sw;

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	GuidToSWDataBaseEntryIter sw_node_it;
	// Case new switch was added
	if ((sw_node_it = this->m_sw_db.m_sw_map.find(general_sw_info.m_guid)) ==
	     this->m_sw_db.m_sw_map.end()) {
		ARSWDataBaseEntry new_sw_node(general_sw_info, &m_port_groups_calculator);
		GuidToSWDataBaseEntryInsertRes res = this->m_sw_db.m_sw_map.insert(
			pair<u_int64_t, ARSWDataBaseEntry>(general_sw_info.m_guid, new_sw_node));
		sw_node_it = res.first;
		sw_node_it->second.Init();
#ifdef _DEBUG_AR_UPDATE_DB
		printf("Switch GUID " U64H_FMT " was added to db\n",
		       sw_node_it->second.m_general_sw_info.m_guid);
#endif
	}

	// Save general info even if we know this switch already because
	// this general info might have been changed by osm
	sw_node_it->second.SetARGeneralSWInfo(general_sw_info);
	sw_node_it->second.in_temporary_error = false;

	p_sw = general_sw_info.m_p_osm_sw;
	/* Update status according to required AR get MADS, that were sent earlier */
	osm_ar_update_status(p_sw);
	if (p_sw->ar_data.status == AR_STATUS_SUCCESS) {
		sw_node_it->second.m_ar_info = p_sw->ar_data.ar_info;
		MarkSWSupportAR(sw_node_it->second);
		sw_node_it->second.MarkSWSupport(SUPPORT_DF);
	} else {
		MarkSWNotSupportAR(sw_node_it->second, NO_ERR);
		sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, NO_ERR);
	}
	// Mark that update was done
	sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::AddNewAndUpdateExistSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_osm_sw;
	// Go over only on switches in osm database to find new ones or changed ones
	for (p_osm_sw = (osm_switch_t *) cl_qmap_head(&this->m_p_osm_subn->sw_guid_tbl);
	     p_osm_sw != (osm_switch_t *) cl_qmap_end(&this->m_p_osm_subn->sw_guid_tbl);
	     p_osm_sw = (osm_switch_t *) cl_qmap_next(&p_osm_sw->map_item)) {

		u_int16_t device_id = cl_ntoh16(p_osm_sw->p_node->node_info.device_id);

		ARGeneralSWInfo general_sw_info(cl_ntoh64(osm_node_get_node_guid(p_osm_sw->p_node)),// GUID
				cl_ntoh16(osm_node_get_base_lid(p_osm_sw->p_node, 0)),              // Lid
				osm_node_get_num_physp(p_osm_sw->p_node) - 1,                       // Num ports
				p_osm_sw);                                                          // Switch ptr

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "AddNewAndUpdateExistSwitches GUID 0x%" PRIx64 " Lid: %u\n",
			   general_sw_info.m_guid, general_sw_info.m_lid);

		this->UpdateSW(general_sw_info);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::RemoveAbsentSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	map<u_int64_t, ARSWDataBaseEntry>::iterator aux_sw_node_it;
	while (sw_node_it !=  this->m_sw_db.m_sw_map.end()) {
		if (sw_node_it->second.m_app_data == ARMGR_SW_UPDATED) {
			++sw_node_it;
			continue;
		}

#ifdef _DEBUG_AR_UPDATE_DB
		printf("Switch GUID " U64H_FMT " was removed from db\n",
				sw_node_it->second.m_general_sw_info.m_guid);
#endif

		aux_sw_node_it = sw_node_it;
		++sw_node_it;

		m_sw_db.UnMapLidToGroup(aux_sw_node_it->second.m_general_sw_info.m_lid);
		this->m_sw_db.m_sw_map.erase(aux_sw_node_it);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}


//============================================================================//
/***************/

void OSMAdaptiveRoutingManager::MarkSWNotSupportAR(ARSWDataBaseEntry &sw_db_entry,
		support_errs_t type)
{
	sw_db_entry.m_support[SUPPORT_AR] = NOT_SUPPORTED;
	sw_db_entry.m_error[SUPPORT_AR] = type;
}

bool OSMAdaptiveRoutingManager::IsARNotSupported(ARSWDataBaseEntry &sw_db_entry) {
	return (sw_db_entry.m_support[SUPPORT_AR] == NOT_SUPPORTED);
}

void OSMAdaptiveRoutingManager::MarkSWSupportAR(ARSWDataBaseEntry &sw_db_entry) {
	sw_db_entry.m_support[SUPPORT_AR] = SUPPORTED;
}

bool OSMAdaptiveRoutingManager::IsARActive(ARSWDataBaseEntry &sw_db_entry) {
	return (sw_db_entry.m_support[SUPPORT_AR] == SUPPORTED && sw_db_entry.m_option_on);
}

bool OSMAdaptiveRoutingManager::IsDFActive(ARSWDataBaseEntry &sw_db_entry) {
	return (IsARActive(sw_db_entry) &&
			sw_db_entry.m_support[SUPPORT_DF] == SUPPORTED);
}

int OSMAdaptiveRoutingManager::ARInfoGetGroupCapProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	bool wait_for_responses = false;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		int status;

		/*
		 * Step 1.1 -
		 * Skip if we know already that this switch not support AR
		 */
		if (this->IsARNotSupported(sw_node_it->second))
			continue;
		/*
		 * Step 1.2 -
		 * Determine required switch configuration.
		 */
		SetRequiredARInfo(sw_node_it->second);

		/*
		 * Step 1.3 -
		 * Query switch for it's capabilities according to required
		 * configuration.
		 */

		/* Check if the switch switch supports dynamic capability calculation */
		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
		    IB_AR_INFO_CAP_MASK_DYN_CAP_CALC_SUP))
			continue;

		/*
		 * Check if capabilities according to required configuration
		 * are already known.
		 * If not, query the switch.
		 */
		if (!osm_ar_capabilities_search(p_ar_data, &sw_node_it->second.m_required_ar_info)) {
			status = osm_ar_query_ar_info(&m_p_osm->sm, p_sw,
						      &sw_node_it->second.m_required_ar_info);
			if (status) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8502: failed to query ARInfo on switch"
					   " 0x%" PRIx64 "\n",
					   cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
				continue;
			}
			wait_for_responses = true;
		}
	}

	/* If we sent queries, wait for responses */
	if (wait_for_responses)
		this->WaitForPendingTransactions();

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

int OSMAdaptiveRoutingManager::ValidateSwitchesCapabilites()
{
	// TODO: enhance capabilities search
	// for example, the search fail if the switch supports RN, but the
	// does not requires it.

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		const ib_ar_info_t *p_req_ar_info;
		int status;

		if (this->IsARNotSupported(sw_node_it->second))
			continue;
		/*
		 * Check if the switch supports the required capabilities,
		 * if not, mark the switch not to use AR.
		 */
		p_req_ar_info = osm_ar_capabilities_search(p_ar_data,
							   &sw_node_it->second.m_required_ar_info);
		if (!p_req_ar_info) {
			sw_node_it->second.MarkSWNotSupport(SUPPORT_AR, AR_GROUP_CAP_ZERO);
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8503: switch 0x%" PRIx64 " does not support"
				   " required capabilities\n",
				   cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		} else {
			osm_ar_set_required_ar_info(p_ar_data, p_req_ar_info);
			// HACK
			sw_node_it->second.m_required_ar_info = *p_req_ar_info;
			sw_node_it->second.m_ar_info = *p_req_ar_info;
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

int OSMAdaptiveRoutingManager::ARInfoSetProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	ib_ar_info_t smp_ar_info;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

		if (this->IsARNotSupported(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported, ar information skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			continue;
		}

		smp_ar_info = sw_node_it->second.m_required_ar_info;
		osm_ar_set_required_ar_info(p_ar_data, &smp_ar_info);

		/*
		 *  Set Group Top
		 */
		osm_ar_set_required_group_top(p_ar_data,
					      sw_node_it->second.m_group_top);
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

void OSMAdaptiveRoutingManager::TreeRoutingNotificationProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (sw_node_it = this->m_sw_db.m_sw_map.begin();
	     sw_node_it != this->m_sw_db.m_sw_map.end();
	     ++sw_node_it) {
		if (this->IsARNotSupported(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported, ar information skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			//we already suppose to have this switch in not supported list
			continue;
		}

		if (sw_node_it->second.in_temporary_error ||
		    !sw_node_it->second.m_config_rn)
			continue;

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		/*
		 * Step 1 -
		 *  RNGenStringTable
		 */

		uint16_t max_string = (uint16_t)
			((0x1 << ib_ar_info_get_string_width_cap(&sw_node_it->second.m_ar_info)) - 1);

		if (p_sw->rank > max_string) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8508: Switch GUID 0x%" PRIx64 ", LID %u, will not genarate RN"
				   " because sw rank:%u is greater than max_string:%u.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   p_sw->rank, max_string);
		} else {
			osm_ar_rn_direction_string_set(&p_sw->ar_data, 0, p_sw->rank);
		}

		/*
		 * Step 2 -
		 *  RNGenBySubGroupPriority
		 */

		uint8_t flags;
		flags = (IS_GENERATE_ARN(m_master_db, sw_node_it->second) ? 0x1 : 0) |
			(IS_GENERATE_FRN(m_master_db, sw_node_it->second) ? 0x2 : 0);
		osm_ar_subgroup_pri_rn_flags_set(&p_sw->ar_data, 0, flags);

	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::ARDumpSWSettings(ARSWDataBaseEntry &sw_db_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
//Currently implemented for one sub group per group
void OSMAdaptiveRoutingManager::ARUpdateSWGroupToCopy(ARSWDataBaseEntry &sw_db_entry,
						      GroupsList &assign_groups)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	set<uint16_t> copy_to_groups;
	uint16_t copy_from_group = 0;
	uint16_t tmp_copy_from_group = 0;
	uint16_t copy_from_block = 0;
	uint16_t modified_block = 0;
	uint16_t prev_modified_block = 0;
	pair<set<uint16_t>::iterator,bool> insert_res;
	set<uint16_t>::iterator modified_group_iter_1;
	set<uint16_t>::iterator modified_group_iter_2;

	uint16_t group_number = 0;
	uint16_t group_top = 0;
	osm_ar_subgroup_t calculated_element;
	osm_ar_subgroup_t *p_current_element;
	int element_cmp;
	uint16_t group_block_idx;


	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "ARUpdateSWGroupToCopy "   \
		   "Switch GUID 0x%" PRIx64 ", LID %u, \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);

	/*
	 * For each group_bitmask we try to find group to copy from
	 * and a list of copy to groups.
	 * we search for copy from group in the following priority:
	 * 1) Two consecutive unmodified groups on the same block
	 * 2) One unmodified group - should be optimized
	 * 3) Two consecutive modified groups on the same block
	 * 4) One modified group.
	 */

	GroupsListIter list_iter = assign_groups.begin();
	for ( ;list_iter != assign_groups.end(); ++list_iter) {

		copy_to_groups.clear();
		copy_from_group = 0;
		copy_from_block = 0;
		tmp_copy_from_group = 0;
		modified_block = 0;
		prev_modified_block = 0;

		uint16_t min_group_number = 0xffff;
		uint16_t max_group_number = 0;

		if ((*list_iter)->m_sw_lid_to_group_number.size() == 0)
			continue;
		LidToGroupNumberMapConstIter iter = (*list_iter)->m_sw_lid_to_group_number.begin();
		for (; iter != (*list_iter)->m_sw_lid_to_group_number.end(); ++iter) {
			max_group_number = max(max_group_number, iter->second);
			min_group_number = min(min_group_number, iter->second);
		}

		(*list_iter)->m_group_bitmask.get_data(calculated_element);
		ARUpdateSWGroupElement(sw_db_entry, calculated_element, min_group_number);

		iter = (*list_iter)->m_sw_lid_to_group_number.begin();
		for (; iter != (*list_iter)->m_sw_lid_to_group_number.end(); ++iter) {
			if (iter->second == min_group_number)
				continue;
			osm_ar_groups_copy_group(&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data,
					         min_group_number, iter->second, true);
		}

		group_top = max(group_top, max_group_number);
	}

	sw_db_entry.m_group_top = group_top;

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ARUpdateSWGroupElement(
		ARSWDataBaseEntry &sw_db_entry,
		osm_ar_subgroup_t &calculated_element,
		uint16_t group_number)
{
	osm_ar_subgroup_t *p_subgroup;

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "ARUpdateSWGroupElement Switch GUID 0x%" PRIx64 ", LID %u " \
		   "group_number:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   group_number);

	p_subgroup = osm_ar_subgroup_get(&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data,
					 group_number, 0);
	memcpy(p_subgroup, &calculated_element, sizeof(osm_ar_subgroup_t));

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ARUpdateSWGroupTableTop(
		ARSWDataBaseEntry &sw_db_entry,
		uint16_t calculated_groups_number)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (calculated_groups_number == 0) {
		sw_db_entry.m_group_top = 0;
		return;
	}

	sw_db_entry.m_group_top = calculated_groups_number - 1;
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//will consume strings
// string == sw_rank - 1
// sw_rank < string < max_consume_rank
//if fabric uses parallel links max_consume_rank == sw_rank + 1
// will consume only string == sw_rank - 1 (like pure fat tree)
void OSMAdaptiveRoutingManager::UpdateRNRcvString(ARSWDataBaseEntry &sw_db_entry,
						  uint8_t max_rank,
						  uint8_t sw_rank,
						  uint8_t max_consume_rank)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_rn_string_handling_clear(p_ar_data);

	uint8_t consume_decision =
		(IS_GENERATE_FRN(m_master_db, sw_db_entry) ?
		 RN_RCV_STRING_CONSUME_ALL : RN_RCV_STRING_CONSUME_ARN);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "UpdateRNRcvString Switch GUID 0x%" PRIx64 ", LID %u "
		   " max_rank:%u sw_rank:%u"
		   " max_consume_rank:%u consume_decision:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   max_rank, sw_rank, max_consume_rank, consume_decision);

	if (sw_rank != 0) {
		// consume sw_rank - 1;
		osm_ar_rn_string_handling_set(p_ar_data, sw_rank - 1, consume_decision, 0);

		// pass on sw_rank
		osm_ar_rn_string_handling_set(p_ar_data, sw_rank, RN_RCV_STRING_PASS_ON, sw_rank);
	}

	uint8_t rank = sw_rank + 1;

	for ( ;rank < max_consume_rank; ++rank)
		osm_ar_rn_string_handling_set(p_ar_data, rank, consume_decision, 0);

	for ( ;rank < max_rank; ++rank)
		osm_ar_rn_string_handling_set(p_ar_data, rank, RN_RCV_STRING_PASS_ON, rank);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::UpdateRNXmitPortMask(ARSWDataBaseEntry &sw_db_entry,
						     PortsBitset &ca_ports,
						     PortsBitset &sw_ports,
						     bool is_down_sw)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	if (is_down_sw) {
		u_int8_t calc_mask = 0;

		if(IS_GENERATE_ARN(m_master_db, sw_db_entry))
			calc_mask |= RN_XMIT_PORT_MASK_GENERATE_ARN;

		if(IS_GENERATE_FRN(m_master_db, sw_db_entry))
			calc_mask |= RN_XMIT_PORT_MASK_GENERATE_FRN;

		calc_mask |= RN_XMIT_PORT_MASK_PASS_ON_RN;

		for (u_int8_t port_num = 1;
		     port_num <= sw_db_entry.m_general_sw_info.m_num_ports;
		     ++port_num) {

			if (ca_ports.test(port_num))
				/* Do not send RN to CA ports */
				osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, 0);
			else if (sw_ports.test(port_num))
				osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, calc_mask);
		}

		if (sw_db_entry.m_to_set_rn_xmit_port_mask) {
			PortsBitset disconnected_ports = ca_ports | sw_ports;
			disconnected_ports.flip();

			for (u_int8_t port_num = 1;
			     port_num <= sw_db_entry.m_general_sw_info.m_num_ports;
			     ++port_num) {
				if (disconnected_ports.test(port_num))
					osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, 0);
			}
		}

	} else
		//clear all xmit_port_mask
		osm_ar_rn_port_xmit_mask_clear(p_ar_data);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::ARCalculatePortGroupsParallelLinks()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	// some variables for calculation
	PortsBitset ports_per_remote_node[IB_LID_UCAST_END_HO + 1];
	u_int8_t port_2_group[MAX_SUPPORTED_PORTS];

	unsigned calculated_max_pg_id;
	unsigned calculated_max_lid;

	PortsBitset group_bitset;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate Port Groups Parallel Links. (switches number:%u)\n",
		   (int)this->m_sw_db.m_sw_map.size());

	/*
	 * Go through all the switches and create their port groups.
	 * In this algorithm ports groups are created out of "parallel"
	 * links - links that are connecting the same pair of nodes.
	 * So here's what we're doing here:
	 *   - for each switch
	 *       - for each port of the switch (port_num)
	 *           - get the remote node
	 *           - if the remote node is CA or RTR
	 *               - no need to create a group, as there can be
	 *                 only one link to its port - skip this port
	 *           - if remote node is switch
	 *               - get remote switch's LID (remote_lid)
	 *               - set bit num. port_num in the bitmask that holds
	 *                 bit for every port that has this LID as a remote
	 *                 side of the link (ports_per_remote_node[remote_lid])
	 *       - at this point we have all the ports organized in groups
	 *         by the remote node (ports_per_remote_node[])
	 *       - set the groups table according to ports_per_remote_node[]
	 *       - create port_2_group map
	 *       - set the lft table according to port_2_group and ucast_lft_port used before
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		/*
		 * Step 1.1 -
		 *  Ensure that this switch supports AR, and that AR is enabled.
		 */
		if (!this->IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64", LID %u: " \
				   "AR/FR not supported or not enabled, calculate of port groups skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			//we already suppose to have this switch in not supported list
			goto next_switch;
		}
		/*
		 * Step 1.2 -
		 *  Set default values for calculations
		 */
		memset(ports_per_remote_node, 0, sizeof(ports_per_remote_node));
		memset(port_2_group, OSM_NO_PATH, sizeof(port_2_group));// default is no group for this port

		calculated_max_pg_id = m_master_db.m_fr_enable ? 1 : 0;
		calculated_max_lid = 0;


		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate LAG groups for Switch GUID 0x%" PRIx64 ", LID %u \n",
			   sw_node_it->second.m_general_sw_info.m_guid,
			   sw_node_it->second.m_general_sw_info.m_lid);

		/*
		 * Step 2 -
		 *  Iterate through all the switch ports and fill ports_per_remote_node
		 *  We create bitmask for each remote node that is connected to this switch.
		 *  Each bit that is set represent a port number that connected to the remote node that
		 *  this bit mask belongs to.
		 */
		for (u_int8_t port_num = 1;
		     port_num <= sw_node_it->second.m_general_sw_info.m_num_ports;
		     ++port_num) {
			osm_node_t *p_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
			// Ignore any loopback connection on switch
			if (p_node == p_remote_node) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Ignoring loopback on Switch GUID 0x%" PRIx64
					   ", LID %u, during calculate of port groups.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
				continue;
			}

			u_int64_t remote_node_guid = cl_ntoh64(osm_node_get_node_guid(p_remote_node));
			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);
			uint16_t remote_node_lid;
			if (remote_node_type == IB_NODE_TYPE_ROUTER || remote_node_type == IB_NODE_TYPE_CA) {
				remote_node_lid = cl_ntoh16(osm_physp_get_base_lid(p_remote_physp));
			} else if (remote_node_type == IB_NODE_TYPE_SWITCH) {
				remote_node_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0));
			} else {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8509: Switch GUID 0x%" PRIx64 " - Unknown node type: "  \
					   "%u, calculate of port groups skipped.\n",
					   remote_node_guid, remote_node_type);
				throw 1;
			}

			// Mark that port number port_num routes to lid remote_node_lid.
			// Note that bit number represent port number.
			ports_per_remote_node[remote_node_lid].set(port_num);

			if (this->m_p_osm_log->level & OSM_LOG_DEBUG)
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Update ports_per_remote_node on "  \
					   "Switch GUID 0x%" PRIx64 ", LID: %u : port_num %u, " \
					   "ports_per_remote_node[%u]=%s.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num,
					   remote_node_lid,
					   ports_per_remote_node[remote_node_lid].to_string().c_str());

		}   /* End of loop switch ports */

		/*
		 * Step 3 -
		 *  Create all port groups in the following way:
		 *  At this point we have all the ports marked their bits in the ports_per_remote_node array
		 *  Bits that are marked in the same array entry effectively form port group.
		 *
		 *  Number of port groups <= number of switch ports.
		 *
		 *  Check all the entries of the ports_per_remote_node array.
		 *  If an entry contain more than one port and the entry is not all 0, create port group.
		 */
		for (unsigned lid_num = 1; lid_num <= m_p_osm_subn->max_lid; ++lid_num) {
			if (!ports_per_remote_node[lid_num].any())
				continue;

			// count number of ports that are connected directly to the same LID (remote node) */
			u_int8_t ports_counter = 0;
			for (unsigned i = 1; i <= sw_node_it->second.m_general_sw_info.m_num_ports; ++i)
				if (ports_per_remote_node[lid_num].test(i))
					++ports_counter;

			// If there is only one port in such group, then the AR mode would be Static, and there is no
			// need to create port group for it - the GroupNumber in the LFT entry for Static LID
			// State is actually a port number. So, if we see a port that is mapped to group
			// id OSM_NO_PATH, it means that the port is alone - no port group was created for it.
			if (ports_counter == 1)
				continue;

			osm_ar_subgroup_t *p_subgroup = osm_ar_subgroup_get(p_ar_data, calculated_max_pg_id, 0);
			ports_per_remote_node[lid_num].get_data(*p_subgroup);

			++calculated_max_pg_id;
		}      /* End of loop all lids */
		if (calculated_max_pg_id)
			--calculated_max_pg_id;

		/*
		 * Step 4.1 -
		 *  Init port-2-group mapping.
		 *  Value OSM_NO_PATH (255) in port-2-group mapping means that the port is mapped to itself,
		 *  which is valid in case of Static LID State.
		 */
		for (unsigned pg_id = 0; pg_id <= calculated_max_pg_id; ++pg_id) {
			osm_ar_subgroup_t *p_subgroup = osm_ar_subgroup_get(p_ar_data, pg_id, 0);
			group_bitset.set(*p_subgroup);
			if (!group_bitset.any())
				// no ports in this group - skip it
				continue;

			for (unsigned i = 0;
			     i <= sw_node_it->second.m_general_sw_info.m_num_ports;
			     ++i)
				if (group_bitset.test(i))
					port_2_group[i] = pg_id;
		}      /* End of loop valid port groups */
		//TODO enter assert regarding port zero

		/*
		 * Step 4.2 -
		 *  Now let's configure AR LFTs.
		 */
		for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {
			// Find out which port was used to route to lid_num in the usual unicast routing
			u_int8_t ucast_lft_port =
				OsmSwitchGetPortByLidUseNewLft(
						sw_node_it->second.m_general_sw_info.m_p_osm_sw,
						dest_lid);
			if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
				continue;
			calculated_max_lid = dest_lid;


			// Check port_2_group mapping of this port.
			// If there is no port group for this port (the mapped port group ID is OSM_NO_PATH),
			// then it means that the port is alone. In that case, LID state would be Static,
			// which means that the GroupNumber in the LFT entry is actually a port number.
			//
			// Note: if (lid_num == my LID), then (ucast_lft_port == 0), and its port_2_group
			// mapping will be OSM_NO_PATH. This is loopback scenario.
			unsigned pg_id = port_2_group[ucast_lft_port];

			osm_advanced_routing_data_t *p_ar_data =
				&sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data;
			if (pg_id != OSM_NO_PATH) {
				osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port,
					       (this->m_master_db.m_ar_enable ?
					       this->m_master_db.m_ar_mode :
					       OSM_AR_IB_LID_STATE_STATIC),
					       pg_id);
			} else
				osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port,
						OSM_AR_IB_LID_STATE_STATIC, 0);
		}      /* End of loop destination lids */

		/*
		 * Step 5 -
		 *  Now we will decide if the calculated settings are different from old,
		 *  or if we need to ignore old settings and use the new ones
		 */
		this->ARUpdateSWGroupTableTop(sw_node_it->second,
					      calculated_max_pg_id + 1);

next_switch:
		++sw_node_it;
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}


u_int16_t OSMAdaptiveRoutingManager::AllocateSwArGroup(uint16_t sw_lid,
						       uint16_t group_cap)
{
	//After we configure the fabric once, all leaf switches should be assigned
	//with AR groups so we optimize by not locking m_sw_db.m_db_lock on read
	uint16_t ar_group_number = m_sw_db.m_sw_lid_to_group_number[sw_lid];

	if (ar_group_number != AR_GROUP_UNASSIGNED && ar_group_number < group_cap)
		return ar_group_number;

	ar_group_number = AR_GROUP_UNASSIGNED;
	pthread_mutex_lock(&m_sw_db.m_db_lock);

	if (m_sw_db.m_sw_lid_to_group_number[sw_lid] == AR_GROUP_UNASSIGNED ||
	    m_sw_db.m_sw_lid_to_group_number[sw_lid] >= group_cap) {
		if (!m_sw_db.m_free_ar_group_numbers.empty()) {
			set <uint16_t>::iterator iter =
				m_sw_db.m_free_ar_group_numbers.begin();

			if (*iter < group_cap) {
				ar_group_number = *iter;
				m_sw_db.m_free_ar_group_numbers.erase(iter);
			}
		} else if (m_sw_db.m_next_ar_group_number < group_cap){
			ar_group_number = m_sw_db.m_next_ar_group_number;
			m_sw_db.m_next_ar_group_number++;
		}

		if (ar_group_number != AR_GROUP_UNASSIGNED) {
			if (m_sw_db.m_sw_lid_to_group_number[sw_lid] ==	AR_GROUP_UNASSIGNED) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "AllocateSwArGroup Sw_Lid:%u Group:%u.\n",
					   sw_lid, ar_group_number);
			} else {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Replace SwArGroup Sw_Lid:%u Group:%u with:%u.\n",
					   sw_lid,
					   m_sw_db.m_sw_lid_to_group_number[sw_lid],
					   ar_group_number);
				m_sw_db.m_free_ar_group_numbers.insert(
						m_sw_db.m_sw_lid_to_group_number[sw_lid]);
			}
			m_sw_db.m_sw_lid_to_group_number[sw_lid] = ar_group_number;

		} else
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Failed to AllocateSwArGroup Sw_Lid:%u.\n",
				   sw_lid);
	} else
		ar_group_number = m_sw_db.m_sw_lid_to_group_number[sw_lid];

	pthread_mutex_unlock(&m_sw_db.m_db_lock);

	return ar_group_number;
}

int OSMAdaptiveRoutingManager::SetHcaToSwLidMapping(osm_physp_t *p_hca_physp,
						    osm_node_t *p_remote_sw_node,
						    u_int16_t *hca_to_sw_lid_mapping)
{
	u_int16_t node_lid = cl_ntoh16(osm_physp_get_base_lid(p_hca_physp));
	if (p_remote_sw_node->node_info.node_type != IB_NODE_TYPE_SWITCH ||
	    node_lid == 0 )
		return -1;

	uint8_t lmc = osm_physp_get_lmc(p_hca_physp);

	u_int16_t remote_node_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_sw_node, 0));

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Map (HCA/Router) LID %u lmc:%u to SW LID %u.\n",
		   node_lid, lmc, remote_node_lid);

	u_int16_t numLidsPerPort = (u_int16_t)((u_int16_t)1 << lmc);
	u_int16_t lid_limit = node_lid+numLidsPerPort;

	for (u_int16_t l = node_lid; l < lid_limit; l++)
		hca_to_sw_lid_mapping[l] = remote_node_lid;

	return 0;
}

/***************/
void OSMAdaptiveRoutingManager::UpdateCalculatedArGroupTable(osm_ar_subgroup_t *p_subgroup,
					 		     uint16_t group_number,
							     PortsBitset &group_bitmask,
							     unsigned &calculated_max_pg_id)
{
	group_bitmask.get_data(*p_subgroup);
	if (group_number > calculated_max_pg_id)
		calculated_max_pg_id = group_number;
}

void OSMAdaptiveRoutingManager::ARCalculatePortGroupsTree()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    uint8_t max_rank = 0;

    //Minimal hops array contains information for switches only
    //find for each hca the connected switch and fill in hca_to_sw_lid_mapping
    u_int16_t hca_to_sw_lid_mapping[IB_LID_UCAST_END_HO + 1] = {};
    uint8_t sw_lid_to_rank_mapping[IB_LID_UCAST_END_HO + 1] = {};
    memset(sw_lid_to_rank_mapping, 0xFF, sizeof(sw_lid_to_rank_mapping));

    osm_node_t *p_osm_node;
    // Go over all nodes in osm database
    for (p_osm_node =  (osm_node_t *) cl_qmap_head(&m_p_osm_subn->node_guid_tbl);
         p_osm_node != (osm_node_t *) cl_qmap_end(&m_p_osm_subn->node_guid_tbl);
         p_osm_node =  (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

        if (p_osm_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
            //set sw_lid_to_rank_mapping
            if (m_master_db.m_frn_enable || m_master_db.m_arn_enable) {

                u_int16_t lid = cl_ntoh16(
                    osm_node_get_base_lid(p_osm_node, 0));

                if (p_osm_node->sw->rank == OSM_SW_NO_RANK) {
                    m_master_db.m_frn_enable = false;
                    m_master_db.m_arn_enable = false;

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                            "ERR 8517: Switch GUID 0x%" PRIx64  ", LID %u: "
                            "has no rank - can not support ARN/FRN.\n",
                            cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                            lid);

                } else if (p_osm_node->sw->rank > MAX_SUPPORTED_RANK) {
                    m_master_db.m_frn_enable = false;
                    m_master_db.m_arn_enable = false;

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
                            "Switch GUID 0x%" PRIx64  ", LID %u: "
                            "has rank:%u which exceeds maximum.\n",
                            cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                            lid,
                            p_osm_node->sw->rank);

                }

                sw_lid_to_rank_mapping[lid] = p_osm_node->sw->rank;
                max_rank = max(max_rank,p_osm_node->sw->rank);
            }
            continue;
        }

        for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);

            if (!p_remote_node || !p_physp || !osm_link_is_healthy(p_physp))
                continue;
            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;

            SetHcaToSwLidMapping(p_physp, p_remote_node, hca_to_sw_lid_mapping);
        }
    }

    /*
     * In this algorithm ports groups are created out of minimal hops
     * links - each lid will recieve a group containing all the ports
     * with minimal hops to this lid.
     * So here's what we're doing here:
     *   - for each switch
     *       - for each port of the switch (port_num)
     *           - calculate group bitmask for this lid
     *           - asisign groups acording to device limitations
     */
    m_port_groups_calculator.CalculatePortGroupsTree(
        max_rank, hca_to_sw_lid_mapping, sw_lid_to_rank_mapping);


    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void
OSMAdaptiveRoutingManager::ARCalculateSwitchPortGroupsTree(
    ARSWDataBaseEntry &sw_db_entry,
    uint8_t max_rank,
    u_int16_t *hca_to_sw_lid_mapping,
    uint8_t *sw_lid_to_rank_mapping)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    // some variables for calculation
    osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
    osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
    osm_ar_subgroup_t *p_subgroup;
    unsigned calculated_max_pg_id;
    unsigned calculated_max_lid;
    int rc = 0;

    TreeAlgorithmData algorithm_data;
    uint16_t max_lid = 0;
    uint8_t **hops = NULL;
    GroupData group_data;

    //for ARN/FRN consume string calculation
    PortsBitset down_sw_ports;
    PortsBitset ca_ports;
    PortsBitset sw_ports;

    osm_ar_subgroup_t empty_element;
    osm_ar_subgroup_clear(&empty_element);

    uint8_t sw_rank = OSM_SW_NO_RANK;
    uint8_t max_consume_rank = 0;

    bool is_group_per_leaf_sw = false;

    /*
     * Step 1.1 -
     *  Ensure that this switch supports AR, and that AR is enabled.
     */
    if (!this->IsARActive(sw_db_entry)) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Skip Switch GUID 0x%" PRIx64 ", LID %u: " \
                   "AR not supported or not enabled.\n",
                sw_db_entry.m_general_sw_info.m_guid,
                sw_db_entry.m_general_sw_info.m_lid);
        //we already suppose to have this switch in not supported list
        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    /*
     * Step 1.2 -
     *  Set default values for calculations
     */
    calculated_max_pg_id = 0;
    calculated_max_lid = 0;
    algorithm_data.Clear();

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Calculate group bitmask for GUID 0x%" PRIx64 ", LID %u: \n",
            sw_db_entry.m_general_sw_info.m_guid,
            sw_db_entry.m_general_sw_info.m_lid);

    /*
     * Step 1.3 -
     *  Build down switch ports and ca ports bitset
     */
    sw_db_entry.m_config_rn =
        (IS_GENERATE_ARN(m_master_db, sw_db_entry) ||
         IS_GENERATE_FRN(m_master_db, sw_db_entry));
    is_group_per_leaf_sw = sw_db_entry.m_config_rn;

    if (sw_db_entry.m_config_rn) {

        down_sw_ports.reset();
        ca_ports.reset();
        sw_ports.reset();

        sw_rank = sw_db_entry.m_general_sw_info.m_p_osm_sw->rank;

        osm_node_t *p_osm_node =
            sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;

        for( u_int8_t port_num = 1 ;
             port_num <= p_osm_node->node_info.num_ports ; port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;

            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp || !p_remote_physp->p_node)
                continue;


            if(p_remote_physp->p_node->sw){
                sw_ports.set(port_num);
                if (p_remote_physp->p_node->sw->rank > sw_rank)
                    down_sw_ports.set(port_num);
            } else
                ca_ports.set(port_num);
        }

    }


    /*
     * Step 2.1 -
     *  Iterate through all lids and calculate group bitmask for each lid.
     */
    max_lid = sw_db_entry.m_general_sw_info.m_p_osm_sw->max_lid_ho;
    hops = sw_db_entry.m_general_sw_info.m_p_osm_sw->hops;
    //u_int8_t last_port_num;

    uint16_t sw_lid_num;
    for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {

        if (!hops[lid_num]) {
            //this is not sw lid. get the sw lid connected to this lid
            sw_lid_num = hca_to_sw_lid_mapping[lid_num];
            if (sw_lid_num == 0) {
                continue;
            }
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] = hops[sw_lid:%u]+1 \n",lid_num , sw_lid_num);
        } else
            sw_lid_num = lid_num;

        if (!hops[sw_lid_num] || hops[sw_lid_num][0] < 1 || hops[sw_lid_num][0] == OSM_NO_PATH) {
            if(!hops[sw_lid_num]) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
            } else {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u][0] == %u \n", sw_lid_num, hops[sw_lid_num][0]);
            }
            continue;
        }

        LidToGroupDataMapIter lid_to_group_iter = algorithm_data.m_lid_to_group_map.find(sw_lid_num);
        if (lid_to_group_iter != algorithm_data.m_lid_to_group_map.end()) {

            this->AddLidToARGroup(lid_num, sw_lid_num, lid_to_group_iter->second,
                            algorithm_data, false,
                            is_group_per_leaf_sw);
            continue;
        }

        group_data.m_group_bitmask.reset();
        group_data.m_group_size = 0;


        /*
         *  Calculate group bitmask.
         */
        for (u_int8_t port_num = 1; port_num <= sw_db_entry.m_general_sw_info.m_num_ports; ++port_num) {
            if(hops[sw_lid_num][0] == hops[sw_lid_num][port_num]) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[%u][%u]:%u \n", sw_lid_num, port_num , hops[sw_lid_num][port_num]);
                group_data.m_group_bitmask.set(port_num);
                group_data.m_group_size++;
            }
        }

        if (group_data.m_group_size > 1 || sw_db_entry.m_config_rn){

            GroupBitmaskToDataMapInsertRes res =
                algorithm_data.m_groups_map.insert(
                    make_pair(group_data.m_group_bitmask,group_data));

            this->AddLidToARGroup(lid_num, sw_lid_num, &res.first->second,
                            algorithm_data, true,
                            is_group_per_leaf_sw);
        }
    }

    /*
     * Step 2.2 -
     *  try to assign group number and group table number(for IS4) for all groups.
     */
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 2.2 AssignPortGroups \n");

    if (!algorithm_data.m_lid_to_group_map.empty()) {       
        rc = this->AssignPortGroups(sw_db_entry, algorithm_data);
        if (rc) 
            OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    /*
     * Step 2.3 -
     *  Calculate min string to consume.
     */

    max_consume_rank = sw_rank;
    if (sw_db_entry.m_config_rn &&
        !algorithm_data.m_assign_groups.empty()) {

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Step 2.3 Calculate min string to consume."
                   " sw_rank:%u\n",
                   sw_rank);

        GroupsListIter list_iter = algorithm_data.m_assign_groups.begin();
        uint8_t grp_min_rank = sw_rank;

        for ( ;
              list_iter != algorithm_data.m_assign_groups.end();
              ++list_iter) {

            this->PrintGroupData("Calculate consume_rank for group: ",
                           *(*list_iter));

            if ((*list_iter)->m_group_size > 1 &&
                ((*list_iter)->m_group_bitmask & down_sw_ports).any() ) {

                    grp_min_rank = max_rank;

                    for (list<uint16_t>::iterator lid_iter =
                            (*list_iter)->m_lids_list.begin() ;
                         lid_iter != (*list_iter)->m_lids_list.end();
                         ++lid_iter)

                        grp_min_rank =
                            min(grp_min_rank,
                                sw_lid_to_rank_mapping[*lid_iter]);

                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "grp_min_rank:%u \n", grp_min_rank);

                    //ignore parallel links (grp_min_rank == sw_rank+1)
                    if (grp_min_rank < sw_rank+2 ||
                        max_consume_rank == grp_min_rank)
                        continue;

                    if (max_consume_rank == sw_rank)
                        max_consume_rank = grp_min_rank;
                    else {
                        OSM_AR_LOG(
                            m_p_osm_log, OSM_LOG_ERROR,
                            "ERR 8518: Group minimal rank (%u) "
                            "is different from current rank (%u) "
                            "for GUID 0x%" PRIx64 ", LID %u: bitmask:%s\n",
                            grp_min_rank, max_consume_rank,
                            sw_db_entry.m_general_sw_info.m_guid,
                            sw_db_entry.m_general_sw_info.m_lid,
                            (*list_iter)->m_group_bitmask.
                                to_string().c_str());

                        //set maximal rank
                        if (max_consume_rank > grp_min_rank)
                            max_consume_rank = grp_min_rank;
                    }

                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "max_consume_rank:%u \n", max_consume_rank);


            }


        }
    }

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 3\n");

    /*
     * Step 3 -
     *  Update calculated_ar_group_table:
     *  At this point we have all assigned groups
     *  in algorithm_data.m_assign_groups list
     */
    if (!algorithm_data.m_assign_groups.empty()) {
    
        GroupsListIter list_iter = algorithm_data.m_assign_groups.begin();

        //In case we use group_per_leaf_sw we optimize ARGroup MAD sending by using copy
        //otherwize we use UpdateCalculatedArGroupTable
        if (!is_group_per_leaf_sw) {
                for ( ;
                      list_iter != algorithm_data.m_assign_groups.end();
                      ++list_iter) {
                      p_subgroup = osm_ar_subgroup_get(p_ar_data, (*list_iter)->m_group_number, 0);
                      UpdateCalculatedArGroupTable(p_subgroup,
                                                   (*list_iter)->m_group_number,
                                                   (*list_iter)->m_group_bitmask,
                                                   calculated_max_pg_id);
                }
        }
    }

    /*
     * Step 4 -
     *  Now let's configure AR LFTs.
     */
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 4 configure AR LFTs \n");

    for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {
        uint16_t group_number = AR_GROUP_UNASSIGNED;
        enum OSM_AR_LID_STATE ar_mode = OSM_AR_IB_LID_STATE_STATIC;
        // Find out which port was used to route to lid_num in the usual unicast routing
        u_int8_t ucast_lft_port =
                OsmSwitchGetPortByLidUseNewLft(sw_db_entry.m_general_sw_info.m_p_osm_sw,
                        dest_lid);
        if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
            continue;
        calculated_max_lid = dest_lid;

        LidToGroupDataMapConstIter iter = algorithm_data.m_lid_to_group_map.find(dest_lid);

        if (iter == algorithm_data.m_lid_to_group_map.end()) {
        
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - Set lid sate Static Group:%u\n",
                       sw_db_entry.m_general_sw_info.m_lid, 
                       dest_lid, 
                       ucast_lft_port);
            // We already mark static in lid state by default
        } else if (iter->second->m_group_table_number == ARMGR_MAX_GROUP_TABLES) {
            if (m_p_osm_log->level & OSM_LOG_DEBUG) 
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "LID %u dest_lid:%u - was not mapped to group because group bitmask:%s was not mapped.\n",
                           sw_db_entry.m_general_sw_info.m_lid, 
                           dest_lid, 
                           iter->second->m_group_bitmask.to_string().c_str());

        } else {
            group_number = iter->second->m_group_number;
            if (is_group_per_leaf_sw) {

                if (!hops[dest_lid])
                    //this is not sw lid. get the sw lid connected to this lid
                    sw_lid_num = hca_to_sw_lid_mapping[dest_lid];
                else
                    sw_lid_num = dest_lid;

                LidToGroupNumberMapConstIter sw_grp_iter =
                    iter->second->m_sw_lid_to_group_number.find(sw_lid_num);

                if (sw_grp_iter ==
                    iter->second->m_sw_lid_to_group_number.end()) {
                    group_number = iter->second->m_group_number;
                } else
                    group_number = sw_grp_iter->second;
            }

            ar_mode = this->m_master_db.m_ar_mode;
            if ((is_group_per_leaf_sw && group_number == AR_GROUP_UNASSIGNED) ||
                !this->m_master_db.m_ar_enable) //only FR/FRN enabled
                ar_mode = OSM_AR_IB_LID_STATE_STATIC;

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - Set lid state %u Group:%u Table:%u (p_group:%p)\n",
                       sw_db_entry.m_general_sw_info.m_lid, dest_lid, 
                       ar_mode,
                       group_number,
                       iter->second->m_group_table_number,
                       iter->second);
        } 
        osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port, ar_mode,
                       group_number);

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "LID %u dest_lid:%u - GroupNumber:%u TableNumber:%u LidState:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid, dest_lid,
                   group_number, 0, ar_mode);
    }      /* End of loop destination lids */

    /*
     * Step 5 -
     * Update Copy group and RN tables if needed
     */

    //In case we use group_per_leaf_sw we optimize ARGroup MAD sending by using copy
    if (is_group_per_leaf_sw) {

        p_subgroup = osm_ar_subgroup_get(p_ar_data, 0, 0);
        //verify group 0 is set to empty group
        if(!sw_db_entry.m_is_group_table_valid[0]||
           memcmp(p_subgroup,
                  &empty_element, sizeof(osm_ar_subgroup_t))){

            this->ARUpdateSWGroupElement(sw_db_entry, empty_element, 0);
        }

        pthread_mutex_lock(&m_sw_db.m_db_lock);
        this->ARUpdateSWGroupToCopy(sw_db_entry,
                algorithm_data.m_assign_groups);
        pthread_mutex_unlock(&m_sw_db.m_db_lock);

    } else
        this->ARUpdateSWGroupTableTop(sw_db_entry, calculated_max_pg_id + 1);


    if (sw_db_entry.m_config_rn){
        this->UpdateRNRcvString(sw_db_entry, max_rank, sw_rank,
                                max_consume_rank);
        this->UpdateRNXmitPortMask(sw_db_entry, ca_ports, sw_ports,
                                   down_sw_ports.any());
    }
}

void OSMAdaptiveRoutingManager::AddLidToARGroup(uint16_t lid_num, uint16_t sw_lid_num,
						GroupData* p_group_data,
						TreeAlgorithmData &algorithm_data,
						bool is_new_group,
						bool is_group_per_leaf_sw)
{
	if (is_new_group) {
		p_group_data->m_lids_list.push_back(sw_lid_num);
		algorithm_data.m_lid_to_group_map.insert(
				make_pair(sw_lid_num, p_group_data));

		PrintGroupData("Add Group Data: ",*p_group_data);
	}

	if (lid_num != sw_lid_num) {
		p_group_data->m_lids_list.push_back(lid_num);
		algorithm_data.m_lid_to_group_map.insert(
				make_pair(lid_num, p_group_data));

		if (is_group_per_leaf_sw)
			p_group_data->m_leaf_switches.insert(sw_lid_num);

		if (!is_new_group)
			PrintGroupData("Update Group Data: ",*p_group_data);
	}
}

void OSMAdaptiveRoutingManager::PrintGroupData(const char * str,GroupData &group_data)
{

	if (this->m_p_osm_log->level & OSM_LOG_DEBUG) {

		stringstream sstr;
		list<uint16_t>::const_iterator iter =  group_data.m_lids_list.begin();
		int lid_count = 1;
		//print first 100 lids
		for (; iter != group_data.m_lids_list.end() ; ++iter) {
			sstr << " " << *iter;
			lid_count++;
			if (lid_count > 100) {
				sstr << " ...";
				break;
			}
		}

		sstr << " bitset:" << group_data.m_group_bitmask;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				"%s Group:%p size:%d lids_number:%u lids:%s\n",
				str,
				&group_data,
				group_data.m_group_size,
				(uint16_t)group_data.m_lids_list.size(),
				sstr.str().c_str());
	}
}

int OSMAdaptiveRoutingManager::AssignPortGroups(ARSWDataBaseEntry &sw_db_entry,
						TreeAlgorithmData &algorithm_data)
{
	int rc = 0;
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	uint16_t group_cap = sw_db_entry.GetGroupCap();

	if (sw_db_entry.m_config_rn) {
		rc = AssignPerLeafSwitchGroups(sw_db_entry, algorithm_data);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	GroupsList group_list;

	rc = GetOrderedGroupList(algorithm_data, group_list);
	if (rc) OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "AssignPortGroups \n");


	//reserve group 0 as empry group
	uint16_t next_group_number = (m_master_db.m_fr_enable ? 1 : 0);

	GroupsListIter list_iter = group_list.begin();
	for ( ; list_iter != group_list.end() && next_group_number < group_cap;
	     ++list_iter) {
		(*list_iter)->m_group_number = next_group_number;
		(*list_iter)->m_group_table_number = 0;

		next_group_number++;

		algorithm_data.m_assign_groups.push_front(*list_iter);
	}

	if (next_group_number == group_cap) {
		for ( ;list_iter != group_list.end(); ++list_iter)
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "SX Switch GUID 0x%" PRIx64
				   ", Can not assign group %s \n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   (*list_iter)->m_group_bitmask.to_string().c_str());
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::AssignPerLeafSwitchGroups(ARSWDataBaseEntry &sw_db_entry,
							 TreeAlgorithmData &algorithm_data)
{
	int rc = 0;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "AssignPerLeafSwitchGroups \n");

	uint16_t group_number = 0;
	bool is_group_cap_exceeded = false;

	uint16_t group_cap = sw_db_entry.GetGroupCap();

	GroupBitmaskToDataMapIter iter = algorithm_data.m_groups_map.begin();
	for (; iter != algorithm_data.m_groups_map.end(); ++iter) {
		// will not perform AR for switches that are not leaf switches
		iter->second.m_group_number = AR_GROUP_UNASSIGNED;
		iter->second.m_group_table_number = 0;
		algorithm_data.m_assign_groups.push_front(&iter->second);

		// Go over all LeafSwitches
		set<uint16_t>::iterator leafs_sw_iter = iter->second.m_leaf_switches.begin();
		for (; leafs_sw_iter != iter->second.m_leaf_switches.end(); ++leafs_sw_iter) {

			group_number = AllocateSwArGroup(*leafs_sw_iter, group_cap);

			if (group_number == AR_GROUP_UNASSIGNED) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "SX Switch GUID 0x%" PRIx64 ", "
					   "Can not assign group %s for lid:%u\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   iter->second.m_group_bitmask.to_string().c_str(),
					   *leafs_sw_iter);
				is_group_cap_exceeded = true;
				continue;
			}

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Assign group %u for leaf switch lid:%u\n",
				   group_number, *leafs_sw_iter);

			iter->second.m_sw_lid_to_group_number.insert(
					make_pair(*leafs_sw_iter, group_number));
		}
	}

	if (is_group_cap_exceeded)
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8519: SX Switch GUID 0x%" PRIx64
			   ", Can not assign separate groups number"
			   " for all leaf switches.\n",
			   sw_db_entry.m_general_sw_info.m_guid);

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::GetContainedGroupList(GroupsList& group_list,
						     PortsBitset &group_bitmask,
						     GroupsList& contained_group_list,
						     PortsBitset &total_bitmask)
{
	int rc = 0;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "GetContainedGroupList \n");

	GroupsListIter list_iter = group_list.begin();

	PortsBitset not_group_bitmask = group_bitmask;
	not_group_bitmask.flip();

	for ( ;
	     list_iter != group_list.end() ;
	     ++list_iter) {
		//the group is contained in group_bitmask
		if (!(not_group_bitmask & (*list_iter)->m_group_bitmask).any()) {
			contained_group_list.push_back(*list_iter);
			total_bitmask |= (*list_iter)->m_group_bitmask;

			if (total_bitmask == group_bitmask)
				break; //no need to continue
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

bool compare_groups_by_members(GroupData* p_group1, GroupData* p_group2)
{
	//compare by the following ordered by priority:
	//1) group members number (descending orded)
	//2) group ports number   (descending orded)
	//3) first lid number     (ascending orded) - for stable sorting
	if (p_group1->m_lids_list.size() != p_group2->m_lids_list.size())
		return p_group1->m_lids_list.size() > p_group2->m_lids_list.size();
	if (p_group1->m_group_size != p_group2->m_group_size)
		return p_group1->m_group_size > p_group2->m_group_size;

	return *p_group1->m_lids_list.begin() < *p_group2->m_lids_list.begin();
}

int OSMAdaptiveRoutingManager::GetOrderedGroupList(TreeAlgorithmData &algorithm_data,
						   GroupsList& group_list)
{
	int rc = 0;
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	GroupBitmaskToDataMapIter iter = algorithm_data.m_groups_map.begin();
	for (; iter != algorithm_data.m_groups_map.end() ; ++iter)
		group_list.push_back(&iter->second);

	group_list.sort(compare_groups_by_members);

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

/*
 * Preprocess the subnet and determine whether the routing algorithm can run
 * on the subnet or not.
 * Return 0 - Indicates the routing algorithm can route the subnet.
 * O/W - Indicates the routing algorithm can not route the subnet.
 */
int OSMAdaptiveRoutingManager::Preprocess()
{
	//TODO: Rename ARInfoGetGroupCapProcess to GetSwtichesCapabilites
	//TODO: Write appropriate error message if validation fails.

	int result = 0;

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	/* Update inventory with new/removed switches */
	this->UpdateFabricSwitches();

	this->UpdateUserOptions();

	if ((m_master_db.m_ar_algorithm != AR_ALGORITHM_DF_PLUS) ||
	    !this->m_master_db.m_enable) {
		// Mark all switches as not updated in this cycle
		map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
			this->m_sw_db.m_sw_map.begin();
		while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
			sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
			++sw_node_it;
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, result);
}

/***************/
int OSMAdaptiveRoutingManager::ARCycle()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

	if (m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) {

		if (m_p_osm->subn.opt.qos != TRUE) {
    			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8513: DF_PLUS Algorithm not supported together with disabled qos.\n");
			this->m_master_db.m_enable = false;
		}

		if ((m_master_db.m_en_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) &&
		    (m_p_osm->subn.opt.max_op_vls <= 2)) {
			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				"ERR 8514: DF_PLUS Algorithm do not support disabling AR on sl "
				" when op_vls equals %u\n", m_p_osm->subn.opt.max_op_vls);
			this->m_master_db.m_enable = false;
		}

		if (!m_master_db.m_ar_enable) {
			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				"ERR 8515: DF_PLUS Algorithm does not support ar_mode %u\n",
				m_p_osm->subn.opt.ar_mode);
			this->m_master_db.m_enable = false;
		}

		if (!this->m_master_db.m_enable) {
			ClearAllDragonflyConfiguration();
			rc = -1;
		}
		else
    			rc = ARDragonFlyCycle();

		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	if (!this->m_master_db.m_enable) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "AR and Shield are disabled. enable %d\n",
			   this->m_master_db.m_enable);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	/* Query for required switch capabilities */
	this->ARInfoGetGroupCapProcess();

	/* Validate switch capabilities */
	rc = this->ValidateSwitchesCapabilites();
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8501: failed to validate the required ar capabilities "
			"for advanced routing configuration\n");
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = osm_ar_prepare_tables(this->m_p_osm_subn);
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8504: failed to prepare switches for advanced"
			" routing configuration\n");
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	switch( m_master_db.m_ar_algorithm )
	{
		case AR_ALGORITHM_TREE:
			this->ARCalculatePortGroupsTree();
			break;
		case AR_ALGORITHM_LAG:
			this->ARCalculatePortGroupsParallelLinks();
			break;
		default:
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8516: Unknown algorithm: %d.\n",
				   m_master_db.m_ar_algorithm);
			OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
	};

	// Set ARInfo after group top calculation
	this->ARInfoSetProcess();

	// Currently RoutingNotification is supported only on Tree algorithm
	if ((m_master_db.m_ar_algorithm == AR_ALGORITHM_TREE) &&
			(m_master_db.m_frn_enable || m_master_db.m_arn_enable))
		this->TreeRoutingNotificationProcess();

	//mark the switches as updated
	sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
	if (this->IsARActive(sw_node_it->second))
		// Mark that update was done for this switch in this cycle
		sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::Route()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;

	m_is_temporary_error = false;
	m_is_permanent_error = false;

	static unsigned cycle_num = 0;
	++cycle_num;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Started configuration cycle\n");
	try {
		this->Preprocess();

		if (m_ar_algorithm)
			rc = m_ar_algorithm->RunCycle();
		else
			rc = ARCycle();

	} catch (int error) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8500: Caught exception = %u, exiting AR Manager...\n",
			   error);
		rc = 1;
	}

	this->AROSMIntegrationProcess();
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Configuration cycle (number %u) failed\n", cycle_num);
		fprintf(stdout, "AR Manager - Configuration cycle (number %u) failed\n", cycle_num);
	} else {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Configuration cycle (number %u) completed successfully\n", cycle_num);
		fprintf(stdout, "AR Manager - Configuration cycle (number %u) completed successfully\n", cycle_num);
	}
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//============================================================================//
/***************/
void OSMAdaptiveRoutingManager::AROSMIntegrationProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	/*
	 * Step 1 -
	 *  Decide for each switch the status
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		osm_switch_t *p_osm_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		if ((sw_node_it->second.m_app_data == ARMGR_SW_UPDATED) &&
		    (this->IsARActive(sw_node_it->second)))
			/*
			 * Step 1.1 -
			 *  we configured Adaptive Routing for this switch, need to mark it
			 */
			p_osm_sw->ar_configured = OSM_SW_AR_CONF_CONFIGURED;
		else {
			// All AR configuration will be deleted from the switch
			// in case we had temp error (timeout) we will need to reconfigure.
			if (ib_ar_info_get_group_cap_ho(&sw_node_it->second.m_ar_info) != 0)
				sw_node_it->second.ClearARData();
			/*
			 * Step 1.2 -
			 *  we didn't configured ar for this switch, need to mark it
			 *  and also abort ar if it was enabled in previous cycle
			 */
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "AR NOT configured on Switch GUID 0x%" PRIx64
				   ", LID %u\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);

			if (p_osm_sw->ar_configured) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Clear old lft for Switch GUID 0x%" PRIx64
					   ", LID %u (because was configured in previous cycle)\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
				// mark to SM to perform LFT set
				p_osm_sw->ar_configured = OSM_SW_AR_CONF_DISABLED;
			}
		}

		++sw_node_it;
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/****************************************************/
/***************** public methods *******************/
/****************************************************/
//============================================================================//
OSMAdaptiveRoutingManager::OSMAdaptiveRoutingManager(osm_opensm_t *p_osm,
						     ar_algorithm_t ar_algorithm):
	m_p_osm(p_osm),
	m_p_osm_vendor(p_osm->p_vendor),
	m_p_osm_subn(p_osm->sm.ucast_mgr.p_subn),
	m_p_osm_log(&p_osm->log),
	m_thread_pool(&p_osm->log),
	m_port_groups_calculator(this, &m_thread_pool, &p_osm->log, &m_sw_db),
	m_p_error_window_arr(NULL),
	m_is_permanent_error(false),
	m_is_temporary_error(false),
	m_df_configured(false),
	m_ar_algorithm(NULL),
	m_sw_info_configured(false),
	m_max_df_group_number(0),
	m_group_discovered(false),
	m_options_file_crc(0),
	m_is_init(false)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	m_master_db.m_ar_algorithm = ar_algorithm;
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
OSMAdaptiveRoutingManager::~OSMAdaptiveRoutingManager()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Exiting...\n");
	if (m_p_error_window_arr != NULL)
		delete[] m_p_error_window_arr;
	delete m_ar_algorithm;
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
void OSMAdaptiveRoutingManager::UpdateFabricSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	// Mark all switches as not updated
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
		++sw_node_it;
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "UpdateFabricSwitches\n");

	// Add new switches from osm database and update details on exist ones
	this->AddNewAndUpdateExistSwitches();

	// Remove non exists switches (in osm database) from our database
	this->RemoveAbsentSwitches();

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
void
OSMAdaptiveRoutingManager::UpdateUserOptions()
{
	ar_configuration_params_t *p_ar_conf = &m_p_osm_subn->ar_conf_params;

	this->SetDefaultConfParams();

	switch (p_ar_conf->ar_mode) {
	case OSM_ADV_ROUTING_MODE_ENABLED:
		m_master_db.m_ar_enable = true;
		m_master_db.m_arn_enable = false;
		break;
	case OSM_ADV_ROUTING_MODE_RN:
		m_master_db.m_ar_enable = true;
		m_master_db.m_arn_enable = true;
		break;
	default:
		m_master_db.m_ar_enable = false;
		m_master_db.m_arn_enable = false;
	}

	switch (p_ar_conf->shield_mode) {
	case OSM_ADV_ROUTING_MODE_ENABLED:
		m_master_db.m_flfr_enable = true;
		m_master_db.m_flfr_remote_disable = true;
		break;
	case OSM_ADV_ROUTING_MODE_RN:
		m_master_db.m_flfr_enable = true;
		m_master_db.m_flfr_remote_disable = false;
		break;
	default:
		m_master_db.m_flfr_enable = false;
		m_master_db.m_flfr_remote_disable = true;
	}

	m_master_db.m_enable =     m_master_db.m_enable && (m_master_db.m_ar_enable || m_master_db.m_flfr_enable);
	m_master_db.m_ar_enable =  m_master_db.m_enable && m_master_db.m_ar_enable;
	m_master_db.m_arn_enable = m_master_db.m_enable && m_master_db.m_arn_enable;
	m_master_db.m_fr_enable =  m_master_db.m_enable && m_master_db.m_flfr_enable;
	m_master_db.m_frn_enable = m_master_db.m_fr_enable && !m_master_db.m_flfr_remote_disable;
	m_master_db.m_en_sl_mask = m_p_osm_subn->opt.ar_sl_mask;
	m_master_db.m_en_tr_mask = m_p_osm_subn->opt.ar_transport_mask;
	m_master_db.m_max_cas_on_spine = m_p_osm_subn->opt.dfp_max_cas_on_spine;

	//Currently only TREE supports FRN/ARN
	//Only TREE and LAG supports FR
	if (this->m_master_db.m_ar_algorithm != AR_ALGORITHM_TREE) {
		if (this->m_master_db.m_arn_enable || this->m_master_db.m_frn_enable) {
			this->m_master_db.m_arn_enable = false;
			this->m_master_db.m_frn_enable = false;

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
				   "Currently only tree ar algorithm supports " \
				   "ar routing notifications, "
				   "disabling routing notifications\n");
		}
		if (this->m_master_db.m_ar_algorithm != AR_ALGORITHM_LAG) {
			this->m_master_db.m_enable = this->m_master_db.m_ar_enable;
			if (this->m_master_db.m_fr_enable) {
				this->m_master_db.m_fr_enable = false;
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
					   "Currently only tree and lag ar algorithms " \
					   "support shield mode, disabling shield mode\n");
			}
		}
	}
	if (this->m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS)
		/* spine-to-spine non-DragonFly routes allowed */
		m_master_db.m_op_mode = 1;
}

int OSMAdaptiveRoutingManager::Construct()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;

	this->UpdateFabricSwitches();

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
		++sw_node_it;
	}

	// Initialize - if already been initialized than do nothing
	this->Init();
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void OSMAdaptiveRoutingManager::SetDefaultConfParams()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	/* Set Master Database */
	m_master_db.m_enable                 = AR_CFG_DEFAULT_ENABLE;
	m_master_db.m_ar_enable              = AR_CFG_DEFAULT_AR_ENABLE;
	m_master_db.m_arn_enable             = AR_CFG_DEFAULT_ARN_ENABLE;
	m_master_db.m_flfr_enable            = AR_CFG_DEFAULT_FLFR_ENABLE;
	m_master_db.m_flfr_remote_disable    = AR_CFG_DEFAULT_FLFR_REMOTE_DISABLE;
	m_master_db.m_ar_mode                = AR_CFG_DEFAULT_SW_AR_MODE;
	m_master_db.m_en_sl_mask             = AR_CFG_DEFAULT_EN_SL_MASK;
	m_master_db.m_dfp_en_vl_mask         = AR_CFG_DEFAULT_DFP_EN_VL_MASK;
	m_master_db.m_en_tr_mask             = AR_CFG_DEFAULT_TRANSPORT_MASK;
	m_master_db.m_op_mode                = 0x0;

	/* Set SW Database */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_ageing_time_value = AR_CFG_DEFAULT_SW_AGEING_TIME;
		sw_node_it->second.m_option_on = AR_CFG_DEFAULT_SW_ENABLE;
		++sw_node_it;
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

int OSMAdaptiveRoutingManager::GetSL2VL(osm_physp_t *osm_phys_port,
					uint8_t iport_num, uint8_t oport_num,
					uint8_t (*sl2vl)[IB_NUMBER_OF_SLS])
{
	if (m_ar_algorithm)
		return m_ar_algorithm->GetSL2VL(osm_phys_port, iport_num, oport_num, sl2vl);

	if (m_df_configured)
		return GetSL2VLDF(osm_phys_port, iport_num, oport_num, sl2vl);

	return -1;
}
