/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DFMGR_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "opensm/osm_armgr.hpp"

//VlCap 2 is : VL0, VL1
#define MIN_VL_CAP 2


#define SW_TYPE_TO_STR(sw_type) \
        ((sw_type == SW_TYPE_UNKNOWN) ? "UNKNOWN" : \
            ((sw_type == SW_TYPE_LEAF) ? "LEAF" : "SPINE"))


/****************************************************/
/**************** private methods *******************/
/****************************************************/

int OSMAdaptiveRoutingManager::ARDragonFlyCycle()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;
	AnalizeDFSetupData *setup_data = new AnalizeDFSetupData();
	int unsupported_num;

	for (int i = 0 ; i < 1 ; i++) {
		rc = InitDragonFlyPlus(*setup_data);
		if (rc)
			break;

		if(!SetDragonFlyPlusCapable()) {
			rc = -1;
			break;
		}

		rc = AnalizeDragonFlyPlusSetup(*setup_data);
		if (rc) {
			m_is_permanent_error = true;
			break;
		}

		unsupported_num = this->ARInfoGetGroupCapProcess();
		if (unsupported_num) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "%d switches do not support DFP Cap.\n" ,
				   unsupported_num);
		}

		/* Validate switch capabilities */
		rc = this->ValidateSwitchesCapabilites();
		if (rc) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C01: failed to validate the required ar capabilities "
				   "for advanced routing configuration\n");
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		}
		rc = osm_ar_prepare_tables(this->m_p_osm_subn);
		if (rc) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C02: failed to prepare switches for advanced"
				   " routing configuration\n");
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		}

		rc = ARCalculatePortGroupsDF(*setup_data);
		ARCalculatePortGroupsDFCleanup();
		if (rc) {
			m_is_permanent_error = true;
			break;
		}

		unsupported_num = this->ARInfoSetProcess();
		if (unsupported_num) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C03: %d switches do not support AR Set.\n" , unsupported_num);
			m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
		}

		rc = ARDefinePLFTs();
		if (rc)
			break;

		rc = ARMapPLFTsAndSetSL2VLAct();
		if (rc)
			break;

		UpdateSmDbSwInfo();
	}

	ARDragonFlyCycleEnd(rc);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
		   "ARDragonFlyCycle Ended. rc:%d\n", rc);
	delete setup_data;
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void OSMAdaptiveRoutingManager::ARDragonFlyCycleEnd(int rc)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (m_is_permanent_error)
		m_is_temporary_error = false;
	else if(rc && !m_is_temporary_error)
		m_is_permanent_error = true;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data == NULL)
			continue;

		//set priv to point to it's origin value
		sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv =
			sw_node_it->second.m_p_df_data->m_saved_priv;
	}

	if (m_is_permanent_error) {
		ClearAllDragonflyConfiguration();
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ClearAllDragonflyConfiguration()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_osm_switch;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Clear All Dragonfly Configuration\n");

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		// clear AR configuration from sw - SM should configure this sw
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;

		if (sw_node_it->second.m_p_df_data == NULL)
			continue;

		if (m_sw_info_configured) {
			//Clear rank and coordination from smdb switch info
			p_osm_switch = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
			p_osm_switch->rank = OSM_SW_NO_RANK;
			p_osm_switch->coord = OSM_SW_NO_COORD;
		}
	}

	m_sw_info_configured = false;

	sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data != NULL) {
			delete sw_node_it->second.m_p_df_data;
			sw_node_it->second.m_p_df_data = NULL;
			//Setting/Unsetting PLFT invalidate groups data
			sw_node_it->second.ClearARGroupsData();
		}
	}
	m_df_configured = false;

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

int OSMAdaptiveRoutingManager::InitDragonFlyPlus(AnalizeDFSetupData &setup_data)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "InitDragonFlyPlus.\n");
	osm_switch_t *p_sw;

	m_df_configured = true;
	m_group_discovered = false;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ;sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data == NULL) {
			//Allocate Dragon Flay Data for all switches
			sw_node_it->second.m_p_df_data = new DfSwData();
			if (sw_node_it->second.m_p_df_data == NULL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C04: Failed to allocate Dragon Fly Data.\n");
				m_df_configured = false;
				OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
			}
			//Setting/Unsetting PLFT invalidate groups data
			sw_node_it->second.ClearARGroupsData();
			//On moving to DF clear ARLFTs
			sw_node_it->second.ClearARLftData();
		}
		p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		//when using DF, one can not disable AR on a specific switch
		sw_node_it->second.m_option_on = true;

		//Init DF Data
		sw_node_it->second.m_p_df_data->
			m_df_sw_setup[NEW_DATA].Clear();
		//consider port 0 as down port
		sw_node_it->second.m_p_df_data->
			m_df_sw_setup[NEW_DATA].m_down_ports.set(0);

		//set priv to point to current ARSWDataBaseEntry
		sw_node_it->second.m_p_df_data->m_saved_priv =
			sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv;
		sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv = &(sw_node_it->second);

		//Init group_number
		sw_node_it->second.m_p_df_data->m_df_prev_group_number =
			sw_node_it->second.m_p_df_data->m_df_group_number;
		sw_node_it->second.m_p_df_data->m_df_group_number =
			INVALID_DFP_GROUP_NUMBER;
		//update sw_lid_to_df_data_mapping
		setup_data.sw_lid_to_df_data_mapping[sw_node_it->second.m_general_sw_info.m_lid] =
			sw_node_it->second.m_p_df_data;

		if(sw_node_it->second.m_osm_update_needed) {
			sw_node_it->second.m_p_df_data->plft_number = 0;
			sw_node_it->second.m_p_df_data->m_df_sw_setup[OLD_DATA].Clear();
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

bool OSMAdaptiveRoutingManager::SetRequiredPLFTInfo(ARSWDataBaseEntry *p_sw_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_advanced_routing_data_t *p_ar_data = &p_sw_entry->m_general_sw_info.m_p_osm_sw->ar_data;
	ib_plft_info_t *p_plft_info = &p_ar_data->plft_info;

	uint8_t ModeCap = p_plft_info->mode_cap, mode;
	if (ModeCap == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C05: Switch GUID 0x%" PRIx64 ", LID %u, "  \
			   "Do not have sufficient PrivateLFT cap: (num tables , size). \n",
			   p_sw_entry->m_general_sw_info.m_guid,
			   p_sw_entry->m_general_sw_info.m_lid);
		p_sw_entry->MarkSWNotSupport(SUPPORT_DF, DF_PLFT_CAP_ZERO);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, false);
	}

	u_int8_t NumOfLFTables = 0;
	ib_plft_bank_entry_t *p_mode;
	uint8_t active_mode = 0;

	for (int i = 0 ; i < ModeCap; i++) {
		p_mode = &p_plft_info->modes[i];

		if (p_mode->num_banks < MIN_PLFT_NUMBER ||
				p_mode->size < MIN_PLFT_TABLE_SIZE)
			continue;

		if (p_mode->num_banks > NumOfLFTables) {
			NumOfLFTables = p_mode->num_banks;
			mode = i;
			if (NumOfLFTables >= MAX_PLFT_NUMBER)
				break;
		}
	}

	if (NumOfLFTables == 0) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C06: Switch GUID 0x%" PRIx64 ", LID %u, "  \
			   "Do not have sufficient PrivateLFT cap: (num tables , size). \n",
			   p_sw_entry->m_general_sw_info.m_guid,
			   p_sw_entry->m_general_sw_info.m_lid);

		p_sw_entry->MarkSWNotSupport(SUPPORT_DF, DF_PLFT_CAP_LESS_THAN_MIN);

		OSM_AR_LOG_RETURN(this->m_p_osm_log, false);
	} else {
		active_mode = mode + 1;
		osm_ar_set_required_plft_active_mode(p_ar_data, active_mode);
		p_sw_entry->m_p_df_data->plft_number = MIN_PLFT_NUMBER;
		osm_ar_set_required_plft_number(p_ar_data,
						p_sw_entry->m_p_df_data->plft_number);

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "PLFT Info Get Switch GUID 0x%" PRIx64 ", LID %u:"
			   " desired mode: %d Active_Mode: %d.\n",
			   p_sw_entry->m_general_sw_info.m_guid,
			   p_sw_entry->m_general_sw_info.m_lid,
			   active_mode,
			   p_plft_info->active_mode);
	}
	OSM_AR_LOG_RETURN(this->m_p_osm_log, true);
}

bool OSMAdaptiveRoutingManager::SetDragonFlyPlusCapable()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int unsupported_num = 0;
	int disable_by_tr_err = 0;
	bool new_dev = false;

	if (m_p_osm_subn->opt.max_op_vls < MIN_VL_CAP ) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C07: Invalid maximal operational VLs %u configuration"\
			   "for Dragonfly setup.\n",
			   m_p_osm_subn->opt.max_op_vls);

		OSM_AR_LOG_RETURN(this->m_p_osm_log, false);
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

		if (sw_node_it->second.m_osm_update_needed) {
			sw_node_it->second.m_support[SUPPORT_DF] = SUPPORT_UNKNOWN;
		} else if(sw_node_it->second.m_support[SUPPORT_DF] == NOT_SUPPORTED) {
			unsupported_num++;
			continue;
		}

		if (!IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported or not enabled.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			unsupported_num++;
			continue;
		}

		//All switches should support PLFT at least 2 banks of at least 48K (PrivateLFTInfo)
		if (!this->SetRequiredPLFTInfo(&sw_node_it->second)) {
			unsupported_num++;
			continue;
		}

		if (sw_node_it->second.m_support[SUPPORT_DF] == SUPPORTED) {
			continue;
		}

		if (sw_node_it->second.in_temporary_error)
			continue;

		new_dev = true;

		//All switches should support vl2vl configuration (ExtendedNodeInfo SL2VLCap bit 0 is on)
		osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

		if (!(p_osm_node->ext_node_info.sl_2_vl_cap & 0x1)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C08: Switch GUID 0x%" PRIx64 ", LID %u, "  \
				   "does not support VL2VL\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_VL2VL_CAP);
			unsupported_num++;
			continue;
		}

		if ((m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) &&
		    !(sw_node_it->second.m_ar_info.enabled_and_cap_fields &
		     IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C09: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "does not support by_transport_disable.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			disable_by_tr_err++;
		}

		//Bit 0 = 2 sub groups
		//Bit 1 = 3 sub groups
		//Bit 2 = 4 sub groups
		//Bits 3 to 7 are reserved
		if (!(sw_node_it->second.m_ar_info.sub_grps_sup & 0x1)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C10: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "2 Subgroups not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
		}

		//All switches should support adaptive Routing checked by ARInfoProcess
		if (!(sw_node_it->second.m_ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C11: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "BySLCap not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_AR_BY_SL_CAP);
			unsupported_num++;
			continue;
		}

		if (!(sw_node_it->second.m_ar_info.enabled_and_cap_fields &
		    IB_AR_INFO_CAP_MASK_GLB_GROUPS)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C12: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Only global AR groups supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
		}

		//All switches should support at least 2 VLs in PortInfo VLCap
		osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
		if (!p_physp_0) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C13: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Failed to get phys port 0.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DB_ERR);
		}

		if (p_physp_0 &&
				!(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C14: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "SL to VL map not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_SL_MAP_CAP);
		}

		for (u_int8_t port_num = 1;
		     port_num <= p_osm_node->node_info.num_ports; port_num++) {
			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
			if (vl_cap < MIN_VL_CAP) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C15: Invalid VL CAP %u on Switch GUID 0x%" PRIx64 \
					   ", LID %u, port:%u\n",
					   vl_cap,
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num);

				sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_VL_CAP_LESS_THAN_MIN);
				break;

			}
		}

	}

	if (new_dev) {
		unsupported_num = 0;
		sw_node_it = this->m_sw_db.m_sw_map.begin();
		for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

			if (sw_node_it->second.in_temporary_error)
				continue;

			if(sw_node_it->second.m_support[SUPPORT_DF] == NOT_SUPPORTED)
				unsupported_num++;
			else if (sw_node_it->second.m_support[SUPPORT_DF] ==
					SUPPORT_UNKNOWN)
				sw_node_it->second.MarkSWSupport(SUPPORT_DF);
		}
	}

	if (unsupported_num || disable_by_tr_err) {
		if (unsupported_num) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "%d switches are not Dragonfly Capable.\n",
				   unsupported_num);
		}
		if (disable_by_tr_err) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
				   "%d switches do not support by_transport_disable.\n",
				   disable_by_tr_err);
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, true);
}

/***************/

int OSMAdaptiveRoutingManager::DiscoverLeaf(AnalizeDFSetupData &setup_data,
					    ARSWDataBaseEntry* p_db_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;
	/* For each connected port not in the DownPortsBitset
	   and not in UpPortsBitset go to connected SW

	   if remote is not SW - skip
	   if remote is LEAF - report an error
	   else set as SPINE

	   If remote DfSwInfo. GroupNumber == 0 //new Spine in group
	   Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
	   push remote DfSwInfo to GroupStack.
	   Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
	   Incorrect fabric architecture
	*/

	osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
		if (curr_sw_setup.m_up_ports.test(port_num) ||
				curr_sw_setup.m_down_ports.test(port_num))
			continue;

		osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp || !osm_link_is_healthy(p_physp))
			continue;
		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp || !p_remote_node)
			continue;

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip non sw node GUID 0x%" PRIx64 "\n",
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));

			continue;
		}

		if (p_remote_node->sw == NULL) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C16: Invalid SW DB on node GUID 0x%" PRIx64 "\n",
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
			rc = -1;
			break;
		}

		ARSWDataBaseEntry *p_remote_db_entry =
			(ARSWDataBaseEntry*)(p_remote_node->sw->priv);
		DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
			m_df_sw_setup[NEW_DATA];

		if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C17: Unexpected leaf GUID: 0x%" PRIx64 " LID: %u"
				   " connected to leaf GUID: 0x%" PRIx64 " LID: %u\n",
				   cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
				   cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)),
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
				   cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0)));
			rc = -1;
			break;
		}

		rc = SetSpine(setup_data, p_remote_node);
		if (rc) {
			break;
		}

		if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
				INVALID_DFP_GROUP_NUMBER) {

			SetGroupNumber(p_remote_db_entry,
					p_db_entry->m_p_df_data->m_df_group_number);

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Set SPINE GUID: 0x%" PRIx64 " LID: %u group: %u\n",
				   p_remote_db_entry->m_general_sw_info.m_guid,
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   p_remote_db_entry->m_p_df_data->m_df_group_number);

			setup_data.m_group_stack.push(p_remote_db_entry);

		} else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
				p_db_entry->m_p_df_data->m_df_group_number) {

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C18: Unexpected leaf GUID: 0x%" PRIx64 " LID: %u from DF group %d"
				   " connected to spine GUID: 0x%" PRIx64 " LID: %u from group %d\n",
				   cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
				   cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)),
				   p_db_entry->m_p_df_data->m_df_group_number,
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
				   cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0)),
				   p_remote_db_entry->m_p_df_data->m_df_group_number);

			rc = -1;
			break;
		}

	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::DiscoverSpine(AnalizeDFSetupData &setup_data,
					     ARSWDataBaseEntry* p_db_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;
	/* For each connected port not in the DownPortsBitset
	   and not in UpPortsBitset go to connected SW

	   Skip non sw node
	   If remote type == Spine Set Spine
	   If Remote DfSwInfo. GroupNumber == local DfSwInfo. GroupNumber
	   Incorrect fabric architecture

	   Else // line
	   If remote DfSwInfo. GroupNumber == 0 //new line in group
	   Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
	   push remote DfSwInfo to GroupStack.
	   Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
	   Incorrect fabric architecture
	   */

	osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
		if (curr_sw_setup.m_up_ports.test(port_num) ||
				curr_sw_setup.m_down_ports.test(port_num))
			continue;

		osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp || !osm_link_is_healthy(p_physp))
			continue;

		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp || !p_remote_node)
			continue;

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip non sw node GUID 0x%" PRIx64 "\n",
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));

			continue;
		}

		if (p_remote_node->sw == NULL) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C19: Invalid SW DB on node GUID 0x%" PRIx64 "\n",
				   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
			rc = -1;
			break;
		}

		ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
		DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
			m_df_sw_setup[NEW_DATA];

		if (remote_sw_setup.m_sw_type == SW_TYPE_SPINE) {

			rc = SetSpine(setup_data, p_remote_node);
			if (rc)
				break;

			if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
					p_db_entry->m_p_df_data->m_df_group_number) {

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C20: Unexpected spine from DF group %d"
					   " connected to spine sw from the same group.\n",
					   p_db_entry->m_p_df_data->m_df_group_number);
				rc = -1;
				break;
			}

		} else if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {

			if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
					INVALID_DFP_GROUP_NUMBER) {


				SetGroupNumber(p_remote_db_entry,
						p_db_entry->m_p_df_data->m_df_group_number);

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
					   p_remote_db_entry->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_general_sw_info.m_lid,
					   p_remote_db_entry->m_p_df_data->m_df_group_number);

				setup_data.m_group_stack.push(p_remote_db_entry);

			} else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
					p_db_entry->m_p_df_data->m_df_group_number) {

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C21: Unexpected spine GUID 0x%" PRIx64 \
					   " LID:%u from DF group %d" \
					   " connected to line GUID 0x%" PRIx64 \
					   " LID:%u from group %d\n",
					   p_db_entry->m_general_sw_info.m_guid,
					   p_db_entry->m_general_sw_info.m_lid,
					   p_db_entry->m_p_df_data->m_df_group_number,
					   p_remote_db_entry->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_general_sw_info.m_lid,
					   p_remote_db_entry->m_p_df_data->m_df_group_number);

				rc = -1;
				break;
			}
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::SetLeaf(AnalizeDFSetupData &setup_data,
				       SwDbEntryPrtList &leafs_list,
				       osm_node_t *p_node)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;
	ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
		m_df_sw_setup[NEW_DATA];

	if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
		curr_sw_setup.m_sw_type = SW_TYPE_LEAF;
		setup_data.m_leafs.push_back(p_db_entry);
		leafs_list.push_back(p_db_entry);
	} else {
		//!!! TODO check if else is valid
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C22: SetLeaf Node (GUID 0x%" PRIx64 ") with type: %s.\n",
			   p_node->node_info.node_guid,
			   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::SetSpine(AnalizeDFSetupData &setup_data,
					osm_node_t *p_node)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;
	ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

	if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
		curr_sw_setup.m_sw_type = SW_TYPE_SPINE;
		setup_data.m_spines.push_back(p_db_entry);
	} else if (curr_sw_setup.m_sw_type != SW_TYPE_SPINE){
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C23: SetLeaf Node (GUID 0x%" PRIx64 ") with type: %s.\n",
			   p_node->node_info.node_guid,
			   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::MarkLeafsByCasNumber(AnalizeDFSetupData &setup_data,
						    SwDbEntryPrtList &leafs_list)
{
	int rc = 0;
	leafs_list.clear();

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Mark leafs by CAs Number. "
		   "Each switch with more than %u CAs is considered as leaf.\n",
		   this->m_master_db.m_max_cas_on_spine);

	osm_node_t *p_osm_node;
	// Set Leaf SW
	for (p_osm_node = (osm_node_t *) cl_qmap_head(&m_p_osm_subn->node_guid_tbl);
	     p_osm_node != (osm_node_t *) cl_qmap_end(&m_p_osm_subn->node_guid_tbl);
	     p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

		if (p_osm_node->node_info.node_type != IB_NODE_TYPE_SWITCH) {
			continue;
		}

		//Count the number of child CAs
		u_int8_t ca_count = 0;
		for (u_int8_t port_num = 1;
		     port_num <= p_osm_node->node_info.num_ports;
		     port_num++) {

			osm_node_t *p_remote_node = osm_node_get_remote_node(
							p_osm_node, port_num, NULL);
			if (!p_remote_node)
				continue;

			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

			if (remote_node_type != IB_NODE_TYPE_CA)
				continue;

			ca_count++;

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			rc = SetHcaToSwLidMapping(p_remote_physp, p_osm_node,
						  setup_data.hca_to_sw_lid_mapping);
			if (rc) {
				OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
			}
		}

		if (ca_count > this->m_master_db.m_max_cas_on_spine) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Discover LEAF (connected to %u cas)."
				   " GUID: 0x%" PRIx64 " LID: %u\n", ca_count,
				   cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
				   cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)));

			rc = SetLeaf(setup_data, leafs_list, p_osm_node);
		}
	}
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::MarkLeafsByGroupsNumber(AnalizeDFSetupData &setup_data,
						       SwDbEntryPrtList &leafs_list)
{
	int rc = 0;
	leafs_list.clear();

	if (!m_group_discovered) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C24: MarkLeafsByGroupsNumber No group was found.\n");
		return -1;
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
		if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER ||
				p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_UNKNOWN)
			continue;

		uint16_t min_group = m_max_df_group_number;
		uint16_t max_group = 0;

		osm_node_t *p_osm_node =
			sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

		for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
				port_num++)
		{
			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;
			osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

			if (remote_node_type != IB_NODE_TYPE_SWITCH)
				continue;

			ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

			//if connected to LEAf this is not a LEAF
			if (p_remote_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type
					== SW_TYPE_LEAF) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Discover SPINE (connected to LEAF)."
					   " GUID: 0x%" PRIx64 " LID: %u\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);

				min_group = 0;
				max_group = m_max_df_group_number;
				break;
			}

			min_group = min(min_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
			max_group = max(max_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
		} //ports loop

		if (min_group == max_group) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Discover LEAF (connected to single group: %u)."
				   " GUID: 0x%" PRIx64 " LID: %u\n", min_group,
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			rc = SetLeaf(setup_data, leafs_list, p_osm_node);

			SetGroupNumber(&sw_node_it->second, min_group);

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   p_df_data->m_df_group_number);
		} else
			rc = SetSpine(setup_data, p_osm_node);

		if (rc) {
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		}

	} //switches loop

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

void OSMAdaptiveRoutingManager::SetGroupNumber(ARSWDataBaseEntry* p_sw_entry,
					       u_int16_t group_number)
{
	DfSwData *p_df_data = p_sw_entry->m_p_df_data;

	if (p_df_data->m_df_prev_group_number != group_number) {
		pair <GuidToGroupMapIter, bool> insert_iter =
			m_guid_to_dfp_group.insert(
				make_pair(p_sw_entry->m_general_sw_info.m_guid, group_number));
		insert_iter.first->second = group_number;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber insert: 0x%" PRIx64 " -> %u.\n",
			   p_sw_entry->m_general_sw_info.m_guid, group_number);
	}
	p_df_data->m_df_group_number = group_number;
}

//Set previous group number (from df_data or from guid_to_dfp_group map)
//On success return 0
//On failure to set previous group number return 1
int OSMAdaptiveRoutingManager::SetPrevGroupNumber(ARSWDataBaseEntry* p_sw_entry,
						  BoolVec &used_group_numbers)
{
	DfSwData * p_df_data = p_sw_entry->m_p_df_data;

	u_int16_t prev_group_number = p_df_data->m_df_prev_group_number;
	if (prev_group_number == INVALID_DFP_GROUP_NUMBER) {
		GuidToGroupMapIter iter =
			m_guid_to_dfp_group.find(p_sw_entry->m_general_sw_info.m_guid);
		if (iter != m_guid_to_dfp_group.end())
			prev_group_number = iter->second;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber map_prev:%u\n", prev_group_number);

	} else {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber prev:%u\n", prev_group_number);
	}

	if ((prev_group_number != INVALID_DFP_GROUP_NUMBER) &&
			(used_group_numbers[prev_group_number] == false)) {

		p_df_data->m_df_group_number = prev_group_number;
		used_group_numbers[prev_group_number] = true;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber set_prev\n");
		m_group_discovered = true;
		return 0;
	}

	return 1;
}

//On the first cycle we assigne previous group numbers for previous discovered switches
//On the second cycle we generates new groups numbers for switches in new groups
//We might lose group numbers on cases in wich groups are divided and rejoined
int OSMAdaptiveRoutingManager::DiscoverGroups(AnalizeDFSetupData &setup_data,
					      SwDbEntryPrtList &leafs_list,
					      BoolVec &used_group_numbers,
					      int cycle)
{
	int rc = 0;

	if (cycle > 2) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C25: -A- DiscoverGroups: Invalid cycle number: %u\n",
			   cycle);
	}

	bool generate_new_group_number = (cycle != 1) || (m_max_df_group_number == 0);
	SwDbEntryPrtList delayed_leafs_list;

	for (SwDbEntryPrtList::iterator iter = leafs_list.begin();
	     iter !=  leafs_list.end(); ++iter)
	{
		DfSwData * p_df_data = (*iter)->m_p_df_data;
		if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER)
			continue;

		if (p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_LEAF) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C26: DiscoverGroups: SW GUID 0x%" PRIx64 "LID %u"
				   "Unexpected type %s.\n",
				   (*iter)->m_general_sw_info.m_guid,
				   (*iter)->m_general_sw_info.m_lid,
				   SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type));
			continue;
		}

		if (SetPrevGroupNumber(*iter, used_group_numbers)) {
			if (generate_new_group_number) {
				m_max_df_group_number++;
				SetGroupNumber(*iter, m_max_df_group_number);
				m_group_discovered = true;
			} else {
				delayed_leafs_list.push_back(*iter);
				continue;
			}
		}
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
			   (*iter)->m_general_sw_info.m_guid,
			   (*iter)->m_general_sw_info.m_lid,
			   p_df_data->m_df_group_number);

		setup_data.m_group_stack.push(*iter);

		while (!setup_data.m_group_stack.empty()) {
			ARSWDataBaseEntry* p_local_sw = setup_data.m_group_stack.top();
			setup_data.m_group_stack.pop();
			switch (p_local_sw->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type) {
				case SW_TYPE_LEAF:
				{
					rc = DiscoverLeaf(setup_data, p_local_sw);
					break;
				}
				case SW_TYPE_SPINE:
				{
					rc = DiscoverSpine(setup_data, p_local_sw);
					break;
				}
				default:
				{
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 8C27: Unexpected SW type.\n");
					OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
				}
			}

			if (rc)
				OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		} //same group

	}// all leafs in leafs_list

	if (!delayed_leafs_list.empty())
		rc = DiscoverGroups(setup_data, delayed_leafs_list, used_group_numbers, ++cycle);

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::SetPortsDirection()
{
	int rc = 0;
	int unknown_type_count = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

	for (sw_node_it = m_sw_db.m_sw_map.begin();
	     sw_node_it != this->m_sw_db.m_sw_map.end();
	     ++sw_node_it)
	{
		DfSwData *p_df_data = sw_node_it->second.m_p_df_data;
		DfSwSetup &curr_sw_setup = p_df_data->m_df_sw_setup[NEW_DATA];

		if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
				   "Discover Unknown SW type. GUID 0x%"
				   PRIx64 " LID %u\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			unknown_type_count++;
			continue;
		}

		osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

		for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
		     port_num++)
		{
			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;
			osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);

			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);


			if (remote_node_type == IB_NODE_TYPE_SWITCH) {
				ARSWDataBaseEntry *p_remote_db_entry =
					(ARSWDataBaseEntry*)(p_remote_node->sw->priv);
				DfSwSetup &remote_sw_setup =
					p_remote_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

				if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {
					if (curr_sw_setup.m_sw_type == SW_TYPE_LEAF) {
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
							   "ERR 8C28: Invalid LEAF to LEAF connection. GUID 0x%"
							   PRIx64 " LID %u to GUID 0x%"
							   PRIx64 " LID %u\n",
							   sw_node_it->second.m_general_sw_info.m_guid,
							   sw_node_it->second.m_general_sw_info.m_lid,
							   p_remote_db_entry->m_general_sw_info.m_guid,
							   p_remote_db_entry->m_general_sw_info.m_lid);
						OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
					} else if (curr_sw_setup.m_sw_type == SW_TYPE_SPINE) {
						curr_sw_setup.m_down_ports.set(port_num);
					}
				} else if (remote_sw_setup.m_sw_type == SW_TYPE_SPINE) {
					curr_sw_setup.m_up_ports.set(port_num);
				}
			} else if (remote_node_type == IB_NODE_TYPE_CA) {
				curr_sw_setup.m_down_ports.set(port_num);
			}
		} //ports
	}//switches

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void OSMAdaptiveRoutingManager::UpdateSmDbSwInfo()
{
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
		   "UpdateSmDbSwInfo.\n");
	m_sw_info_configured = true;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
		osm_switch_t *p_osm_switch =
			sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		if (p_df_data->m_df_group_number == INVALID_DFP_GROUP_NUMBER) {
			p_osm_switch->coord = OSM_SW_NO_COORD;
			p_osm_switch->rank = OSM_SW_NO_RANK;
			continue;
		}

		p_osm_switch->coord = p_df_data->m_df_group_number;

		switch (p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type) {
			case SW_TYPE_LEAF:
				p_osm_switch->rank = 1;
				break;
			case SW_TYPE_SPINE:
				p_osm_switch->rank = 0;
				break;
			default:
				p_osm_switch->rank = OSM_SW_NO_RANK;
				p_osm_switch->coord = OSM_SW_NO_COORD;
		}
	}
}

int OSMAdaptiveRoutingManager::AnalizeDragonFlyPlusSetup(AnalizeDFSetupData &setup_data)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	SwDbEntryPrtList leafs_list;

	int rc = MarkLeafsByCasNumber(setup_data, leafs_list);
	if (rc) {
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	BoolVec used_group_numbers(m_max_df_group_number, false);
	rc = DiscoverGroups(setup_data, leafs_list, used_group_numbers);
	if (rc) {
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = MarkLeafsByGroupsNumber(setup_data, leafs_list);
	if (rc) {
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = DiscoverGroups(setup_data, leafs_list, used_group_numbers);
	if (rc) {
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = SetPortsDirection();
	if (rc) {
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	//check if port moved from up to down or sw_type was changed
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

	for (sw_node_it = m_sw_db.m_sw_map.begin(); sw_node_it != this->m_sw_db.m_sw_map.end();
	     ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

		if ((p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type !=
					p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type) &&
		    (p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type != SW_TYPE_UNKNOWN)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
				   "Switch type was changed for GUID 0x%" PRIx64 \
				   ", LID %u: from %s to %s\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type),
						  SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type));

			p_df_data->m_df_sw_setup[OLD_DATA].Clear();
			continue;
		}

		//clear ports that change orientation
		//from up to down or from down to up
		p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports &=
			~p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports;
		p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports &=
			~p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports;

	}

	this->ARDumpDFAnalizedSetup();

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//enforce routing between HCA to each node and back
static bool leaf_enforce_complete_route[LEAF_PLFT_NUMBER] = {false, true};

static PathDescription leaf_plft_paths_description[LEAF_PLFT_NUMBER][DLID_LOCATION_NUMBER][PATHS_NUMBER] = {
    /* Leaf Up (from Spine) */
    {
        /*Leaf Same Group*/
        {{2, 2, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Leaf Remote Group*/
        {{3, 3, PATH_VALIDATION_NON, TRUE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Same Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Remote Group*/
        {{2, 2, PATH_VALIDATION_NON, TRUE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} }
    },
    /* Leaf Down (from HCA) */
    {
        /*Leaf Same Group*/
        {{2, 2, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Leaf Remote Group*/
        {{3, 3, PATH_VALIDATION_NON, FALSE}, {4, 4, PATH_VALIDATION_REMOTE_DF, FALSE} , {6, 5, PATH_VALIDATION_REMOTE, TRUE} },
        /*Spine Same Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {3, 3, PATH_VALIDATION_REMOTE, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Remote Group*/
        {{2, 2, PATH_VALIDATION_NON, FALSE}, {3, 3, PATH_VALIDATION_REMOTE_DF, FALSE} , {5, 4, PATH_VALIDATION_REMOTE, TRUE} }
    }
};

//enforce routing between HCA to each node and back
static bool spine_enforce_complete_route[SPINE_PLFT_NUMBER] = {false, true/*, false*/};

static PathDescription spine_plft_paths_description[SPINE_PLFT_NUMBER][DLID_LOCATION_NUMBER][PATHS_NUMBER] = {
    /* Spine Up (from Spine) */
    {
        /*Leaf Same Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Leaf Remote Group*/
        {{2, 2, PATH_VALIDATION_NON, FALSE}, {4, 4, PATH_VALIDATION_HOPS_DOWN, TRUE}, {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Same Group*/
        {{2, 2, PATH_VALIDATION_DOWN, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Remote Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {3, 3, PATH_VALIDATION_HOPS_DOWN, TRUE}, {0, 0, PATH_VALIDATION_NON, FALSE} }
    },
    /* Spine Down (from Leaf) VL0*/
    {
        /*Leaf Same Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {3, 3, PATH_VALIDATION_HOPS_DOWN, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Leaf Remote Group*/
        {{2, 2, PATH_VALIDATION_NON, FALSE}, {3, 3, PATH_VALIDATION_REMOTE_DF, FALSE} , {5, 4, PATH_VALIDATION_REMOTE_UP, TRUE} },
        /*Spine Same Group*/
        {{2, 2, PATH_VALIDATION_DOWN, FALSE}, {0, 0, PATH_VALIDATION_NON, FALSE} , {0, 0, PATH_VALIDATION_NON, FALSE} },
        /*Spine Remote Group*/
        {{1, 1, PATH_VALIDATION_NON, FALSE}, {2, 2, PATH_VALIDATION_REMOTE_DF, FALSE} , {4, 3, PATH_VALIDATION_REMOTE_UP, TRUE} }
    }
};

int OSMAdaptiveRoutingManager::ARCalculatePortGroupsDF(AnalizeDFSetupData &setup_data)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;

	for (int plft_id = 0 ; plft_id < SPINE_PLFT_NUMBER ; ++plft_id) {
		rc = ARCalculatePLFTPortGroups(spine_plft_paths_description[plft_id],
					       spine_enforce_complete_route[plft_id],
					       setup_data.m_spines,
					       plft_id,
					       setup_data);
		if (rc)
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	for (int plft_id = 0 ; plft_id < LEAF_PLFT_NUMBER ; ++plft_id) {
		rc = ARCalculatePLFTPortGroups(leaf_plft_paths_description[plft_id],
					       leaf_enforce_complete_route[plft_id],
					       setup_data.m_leafs,
					       plft_id,
					       setup_data);
		if (rc)
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	SavePortGroupsAndDump();

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void OSMAdaptiveRoutingManager::ARCalculatePortGroupsDFCleanup()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();

	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

		if (sw_node_it->second.in_temporary_error)
			continue;

		sw_node_it->second.m_p_df_data->m_algorithm_data.Clear();

		for (int plft_id = 0 ; plft_id < MAX_PLFT_NUMBER ; ++plft_id)
			ClearNoValidDFRouteToLid(sw_node_it->second, plft_id);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//[is Spine][is Remote]
static ar_dlid_location_t ar_dlid_location[2][2] = {
	{LEAF_SAME_GROUP,LEAF_REMOTE_GROUP},
	{SPINE_SAME_GROUP, SPINE_REMOTE_GROUP}
};

int OSMAdaptiveRoutingManager::ARCalculatePLFTPortGroups(
		PathDescription paths_description[DLID_LOCATION_NUMBER][PATHS_NUMBER],
		bool enforce_complete_route,
		list <ARSWDataBaseEntry*> &switches,
		int plft_id,
		AnalizeDFSetupData &setup_data)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	// some variables for calculation
	unsigned calculated_max_lid;

	//On new devices groups are set globally and not per pLFT
	//there for algorithm_data moved to m_p_df_data
	LidToPSGroupDataMap lid_to_ps_group_map;

	uint16_t max_lid = 0;
	uint8_t **hops = NULL;
	PortsBitset group_bitmask[PATHS_NUMBER];

	//will be used to select the static bind port
	uint16_t ports_load[MAX_SUPPORTED_PORTS];

	PSGroupData ps_group_data;
	osm_advanced_routing_data_t *p_ar_data;

	//Rewrite paths_description in more convenient data structure
	PathDescriptionEntry path_desc[DLID_LOCATION_NUMBER][DF_MAX_HOPS+1];
	memset(path_desc, 0, sizeof(path_desc));
	uint8_t max_hops_to_location[DLID_LOCATION_NUMBER];
	memset(max_hops_to_location, 0, sizeof(max_hops_to_location));

	int desc_max_hops_number = 0;
	int desc_min_hops_number = 0;
	osm_dfp_intermediate_down_path dfp_down_up_turns_mode =
		this->m_p_osm_subn->opt.dfp_down_up_turns_mode;

	for (int location_iter = 0 ; location_iter < DLID_LOCATION_NUMBER ; ++location_iter) {
		for (int path_iter = 0 ; path_iter < PATHS_NUMBER ; ++path_iter) {
			desc_max_hops_number = paths_description[location_iter][path_iter].m_max_hops_number;
			desc_min_hops_number = paths_description[location_iter][path_iter].m_min_hops_number;


			if (desc_max_hops_number == 0)
				break;

			if (desc_max_hops_number > DF_MAX_HOPS) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C29: Invalid paths_description[%d][%d]"
					   " min_hops:%d-max_hops:%d\n",
					   location_iter, path_iter,
					   desc_min_hops_number, desc_max_hops_number);
				OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
			}

			for (int i = desc_min_hops_number; i <= desc_max_hops_number; ++i) {
				path_desc[location_iter][i].m_p_desk =
					&paths_description[location_iter][path_iter];
				path_desc[location_iter][i].m_path_index = path_iter;
			}

			max_hops_to_location[location_iter] =
				max(desc_max_hops_number, (int)max_hops_to_location[location_iter]);
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "max_hops_to_location[%d] = %d\n",location_iter,
			   max_hops_to_location[location_iter]);
	}


	/*
	 * Go through all the switches and create their port groups.
	 * In this algorithm ports groups are created out of minimal hops
	 * links - each lid will recieve a group containing all the ports
	 * with minimal hops to this lid.
	 * So here's what we're doing here:
	 *   - for each switch
	 *       - for each port of the switch (port_num)
	 *           - calculate group bitmask for this lid
	 *           - asisign groups acording to device limitations
	 */
	uint16_t group_cap = 0;

	list <ARSWDataBaseEntry*>::iterator sw_entry_it = switches.begin();
	for (; sw_entry_it != switches.end() ; ++sw_entry_it) {
		if ((*sw_entry_it)->in_temporary_error)
			continue;

		if(!IsDFActive(*(*sw_entry_it))) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "DF/AR not supported or not enabled, " \
				   "CalculatePLFTPortGroups skipped.\n",
				   (*sw_entry_it)->m_general_sw_info.m_guid,
				   (*sw_entry_it)->m_general_sw_info.m_lid);
			continue;
		}

		/*
		 * use_only_minhop_path is true on intermediate spine
		 * if OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH is configured
		 */
		bool use_only_minhop_path =
		    (dfp_down_up_turns_mode == OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH) &&
		    ((*sw_entry_it)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
		    plft_id == 0; /* it means the traffic comes from spine, i.e intermediate */

		p_ar_data = &(*sw_entry_it)->m_general_sw_info.m_p_osm_sw->ar_data;

		// (&&pLFT number support&&) We are currently support max of 2 pLFT
		//
		// For Spines for traffic coming from 'down' ports with odd VLs, instead of using
		// additional PLFT, we will use the same PLFT for even VLs, but with static
		// routing.

		group_cap = (*sw_entry_it)->GetGroupCap();
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "group_cap for GUID 0x%" PRIx64 \
			   ", LID %u: is:%u\n",
			   (*sw_entry_it)->m_general_sw_info.m_guid,
			   (*sw_entry_it)->m_general_sw_info.m_lid,
			   group_cap);
		/*
		 * Step 1 -
		 *  Set default values for calculations
		 */
		calculated_max_lid = 0;

		lid_to_ps_group_map.clear();
		DFAlgorithmData &algorithm_data =
			(*sw_entry_it)->m_p_df_data->m_algorithm_data;
		unsigned &calculated_groups_number =
			algorithm_data.m_calculated_groups_number;

		ClearNoValidDFRouteToLid(*(*sw_entry_it), plft_id);
		memset(ports_load, 0, sizeof(ports_load));

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate group bitmask for GUID 0x%" PRIx64 \
			   ", LID %u: pLFT_ID:%u\n",
			   (*sw_entry_it)->m_general_sw_info.m_guid,
			   (*sw_entry_it)->m_general_sw_info.m_lid,
			   plft_id);

		/*
		 * Step 2.1 -
		 *  Iterate through all lids and calculate group bitmask for each lid.
		 */
		max_lid = (*sw_entry_it)->m_general_sw_info.m_p_osm_sw->max_lid_ho;
		hops = (*sw_entry_it)->m_general_sw_info.m_p_osm_sw->hops;

		uint16_t sw_lid_num;
		int max_hops = 0;
		for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {
			if (!hops[lid_num]) {
				//this is not sw lid. get the sw lid connected to this lid
				sw_lid_num = setup_data.hca_to_sw_lid_mapping[lid_num];
				if (sw_lid_num == 0)
					continue;

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "hops[lid:%u] = hops[sw_lid:%u]+1 \n",
					   lid_num , sw_lid_num);
			} else
				sw_lid_num = lid_num;

			ps_group_data.Clear();

			if ((*sw_entry_it)->m_general_sw_info.m_lid == sw_lid_num)
				continue; //Same Switch

			if (!hops[sw_lid_num] || hops[sw_lid_num][0] < 1 ||
			    hops[sw_lid_num][0] == OSM_NO_PATH) {
				continue;
			}

			if (sw_lid_num != lid_num)
				continue; //it is enough to go over switches and skip HCAs

			if (setup_data.sw_lid_to_df_data_mapping[sw_lid_num] == NULL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8C30: No DF Data for SW lid::%u\n", sw_lid_num);
				OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
			}

			// Calculate DLID Location
			int type_idx = (int)(setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
					m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE);
			int dist_idx = (int)(setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
					m_df_group_number != (*sw_entry_it)->m_p_df_data->m_df_group_number);

			ar_dlid_location_t dlid_location = ar_dlid_location[type_idx][dist_idx];
			max_hops = max_hops_to_location[dlid_location];
			if (dist_idx == 0)  //on same group use only primary route
				max_hops = hops[sw_lid_num][0];

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Calculate group bitmask for LID:%u->LID:%u"
				   " min_hops:%u max_hops:%d dlid_location:%d\n",
				   (*sw_entry_it)->m_general_sw_info.m_lid, sw_lid_num,
				   hops[sw_lid_num][0], max_hops, dlid_location);

			/*
			 *  Calculate group bitmask.
			 */
			u_int8_t last_port_num = 0;
			int curr_hops = 0;

			for (u_int8_t port_num = 1;
			     port_num <= (*sw_entry_it)->m_general_sw_info.m_num_ports;
			     ++port_num) {
				curr_hops = hops[sw_lid_num][port_num];

				if(max_hops < curr_hops)
					continue;

				PathDescriptionEntry & path_desc_entry =
					path_desc[dlid_location][curr_hops];

				if( path_desc_entry.m_p_desk == NULL)
					continue;

				/* strict option, don't allow long paths at all */
				if (dfp_down_up_turns_mode == OSM_DFP_DOWN_PATH_DISABLED &&
				    path_desc_entry.m_p_desk->m_is_intermediate_down_path) {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "Long paths are disabled. Skip it for "
						   "switch GUID 0x%" PRIx64
						   " for port %u, dest lid %u\n",
						   (*sw_entry_it)->m_general_sw_info.m_guid,
						   port_num, lid_num);
					continue;
				}

				switch (path_desc_entry.m_p_desk->m_path_validation_t) {
					case PATH_VALIDATION_NON:
						break;
					case PATH_VALIDATION_REMOTE_DF:
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
							   "PATH_VALIDATION_REMOTE_DF hops[%u][%u]:%u \n",
							   sw_lid_num, port_num ,curr_hops);

						if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
							continue;
						}
						break;
					case PATH_VALIDATION_REMOTE_UP:
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
							   "PATH_VALIDATION_REMOTE_UP hops[%u][%u]:%u \n",
							   sw_lid_num, port_num ,curr_hops);
						if(! (*sw_entry_it)->m_p_df_data->
								m_df_sw_setup[NEW_DATA].m_up_ports.test(port_num))
							continue;

						if(!IsRouteOnRemote(*(*sw_entry_it), plft_id, port_num, sw_lid_num))
							continue;

						break;
					case PATH_VALIDATION_REMOTE:
						if(!IsRouteOnRemote(*(*sw_entry_it), plft_id, port_num, sw_lid_num))
							continue;

						break;
					case PATH_VALIDATION_DOWN:
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
							   "PATH_VALIDATION_DOWN hops[%u][%u]:%u \n",
							   sw_lid_num, port_num ,curr_hops);
						if(! (*sw_entry_it)->m_p_df_data->
								m_df_sw_setup[NEW_DATA].m_down_ports.test(port_num)) {
							continue;
						}

						if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
							continue;
						}

						break;
					case PATH_VALIDATION_HOPS_DOWN:
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
							   "PATH_VALIDATION_HOPS_DOWN hops[%u][%u]:%u \n",
							   sw_lid_num, port_num ,curr_hops);
						if(! (*sw_entry_it)->m_p_df_data->
								m_df_sw_setup[NEW_DATA].m_down_ports.test(port_num))
							continue;

						if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
							continue;
						}

						if ((hops[sw_lid_num][0]+2 == curr_hops) &&
								!IsTrueHopsOnRemote(
									*(*sw_entry_it), port_num,
									sw_lid_num,
									path_desc_entry.m_p_desk->m_max_hops_number-1))
							continue;

						break;
					default:
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
							   "ERR 8C31: Unexpected path_validation::%u\n",
							   path_desc_entry.m_p_desk->m_path_validation_t);
						OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
				}

				group_bitmask[path_desc_entry.m_path_index].set(port_num);
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Add port:%d to bitset:%d \n",
					   port_num, path_desc_entry.m_path_index);

				last_port_num = port_num;

				ps_group_data.m_group_size++;
			}

			if (ps_group_data.m_group_size != 0) {

				//set ps_group_data
				bool set_pri = true;

				for (int i = 0 ; i < PATHS_NUMBER ; ++i) {
					if(group_bitmask[i].any()){
						if (set_pri) {
							ps_group_data.m_ps_group_bitmask.set_pri(group_bitmask[i]);
							/*
							 * Here we are at medium path, which is actually
							 * long path in intermediate dfp island.
							 * This happened because no shorter path exist.
							 */
							if (use_only_minhop_path && i) {
								OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
									   "Take long path for "
									   "switch GUID 0x%" PRIx64
									   ", path type %u, group value: %s\n",
									   (*sw_entry_it)->m_general_sw_info.m_guid,
									   i, group_bitmask[i].to_string().c_str());

							}
							set_pri = false;
						} else {
							/*
							 * Don't allow long paths on intermediate spine,
							 * if shorter exist
							 * and we on the intermediate spine switch.
							 * Here we already have shorter paths
							 * available, so don't add longer paths
							 */
							if (use_only_minhop_path) {
								OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
									   "Don't take long path for "
									   "switch GUID 0x%" PRIx64
									   ", path type %u, for dlid %u group value: %s\n",
									   (*sw_entry_it)->m_general_sw_info.m_guid,
									   i, lid_num, group_bitmask[i].to_string().c_str());
							} else {
								ps_group_data.m_ps_group_bitmask.add_sec(group_bitmask[i]);
								OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
									   "Add secondary bitset index:%d value:%s \n",
									   i, group_bitmask[i].to_string().c_str());
							}
						}

						group_bitmask[i].reset();
					}
				}

				if (!ps_group_data.m_ps_group_bitmask.any()) {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 8C32: Failed to set group_bitmask for GUID 0x%"
						   PRIx64 ", LID %u: \n",
						   (*sw_entry_it)->m_general_sw_info.m_guid,
						   (*sw_entry_it)->m_general_sw_info.m_lid);

					OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
				}

				PSGroupBitmaskToDataMapInsertRes res =
					algorithm_data.m_ps_groups_map.insert(make_pair(
								ps_group_data.m_ps_group_bitmask,ps_group_data));

				if (ps_group_data.m_group_size > 1 &&
						res.second)
				{ //if new group set group number)
					if (calculated_groups_number >= group_cap ) {
						OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
							   "Group number exceeded. "
							   "Use static port:%u for for LID:%u->LID:%u\n",
							   res.first->second.m_port_num,
							   (*sw_entry_it)->m_general_sw_info.m_lid, sw_lid_num);

						//use only static
						res.first->second.m_group_size = 1;

					} else {
						res.first->second.m_group_number =
							calculated_groups_number;
						calculated_groups_number++;
					}

				}
				else if (ps_group_data.m_group_size == 1)
					res.first->second.m_port_num =  last_port_num;

				lid_to_ps_group_map.insert(
						make_pair(sw_lid_num,&(res.first->second)));

				PrintPSGroupData("Add/Update Group Data: ", res.first->second);

			} else {
				if (enforce_complete_route &&
						(type_idx == 0 ||// dest sw is leaf
						 (*sw_entry_it)->m_p_df_data->
						 m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF)) {

					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 8C33: No Valid Route from leaf switch (or) to leaf switch "
						   " pLFT:%u GUID 0x%" PRIx64 " LID:%u->LID:%u. "
						   "Probably some links are missing.\n",
						   plft_id,
						   (*sw_entry_it)->m_general_sw_info.m_guid,
						   (*sw_entry_it)->m_general_sw_info.m_lid,
						   sw_lid_num);
					if ((*sw_entry_it)->m_p_df_data->
					    m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
						OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
					} else {
						SetNoValidDFRouteToLid(*(*sw_entry_it), plft_id,
								       sw_lid_num);
					}
				} else {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "No Valid Route pLFT:%u LID:%u->LID:%u\n",
						   plft_id,
						   (*sw_entry_it)->m_general_sw_info.m_lid,
						   sw_lid_num);

					SetNoValidDFRouteToLid(*(*sw_entry_it), plft_id, sw_lid_num);
				}
			}
		}

		/*
		 * Step 3 -
		 *  Now let's configure AR LFTs.
		 */
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Configure AR LFTs for switch guid 0x%" PRIx64 "\n",
			   (*sw_entry_it)->m_general_sw_info.m_guid);

		uint8_t port, state;
		uint16_t group_number;

		for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {
			// Find out which port was used to route to lid_num in the usual unicast routing
			u_int8_t ucast_lft_port =
				OsmSwitchGetPortByLidUseNewLft((*sw_entry_it)->m_general_sw_info.m_p_osm_sw,
							       dest_lid);
			if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
				continue;

			if (!hops[dest_lid]) {
				//this is not sw lid. get the sw lid connected to this lid
				sw_lid_num = setup_data.hca_to_sw_lid_mapping[dest_lid];
				if (sw_lid_num == 0) {
					continue;
				}
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "dest_lid:%u ,  sw_lid:%u \n",dest_lid , sw_lid_num);
			} else
				sw_lid_num = dest_lid;


			LidToPSGroupDataMapIter iter = lid_to_ps_group_map.end();

			//if not same switch
			if ((*sw_entry_it)->m_general_sw_info.m_lid != sw_lid_num)
				iter = lid_to_ps_group_map.find(sw_lid_num);

			if (iter == lid_to_ps_group_map.end()) {

				if ((*sw_entry_it)->m_general_sw_info.m_lid == sw_lid_num)
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "LID %u dest_lid:%u - Set route on same switch static port:%u\n",
						   (*sw_entry_it)->m_general_sw_info.m_lid,
						   dest_lid,
						   ucast_lft_port)
				else if ((setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
					 m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
					 ((*sw_entry_it)->m_p_df_data->
					 m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
					 (m_master_db.m_op_mode &
					 AR_OP_MODE_ENABLE_SPINE_TO_SPINE_NONE_DF_ROUTE)) {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "LID %u dest_lid:%u - Set none DF route "
						   "(Spine to Spine) static port:%u plft_id:%u\n",
						   (*sw_entry_it)->m_general_sw_info.m_lid,
						   dest_lid,
						   ucast_lft_port,
						   plft_id);
				} else {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "No route from LID %u to dest_lid:%u plft_id: %u\n",
						   (*sw_entry_it)->m_general_sw_info.m_lid,
						   dest_lid, plft_id);
					continue; //no route for this lid
				}
				osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port, OSM_AR_IB_LID_STATE_STATIC, 0);
				port = ucast_lft_port;
				state = OSM_AR_IB_LID_STATE_STATIC;
				group_number = 0;

			} else if (iter->second->m_group_size == 1) {
				u_int8_t static_port = iter->second->m_port_num;

				if (iter->second->m_port_num == 0) {
					//group number exceeded maximum
					if (iter->second->m_ps_group_bitmask.test_pri(ucast_lft_port))
						static_port = ucast_lft_port;
					else
						static_port = GetNextStaticPort(
								ports_load,
								iter->second->m_ps_group_bitmask,
								sw_lid_num != dest_lid, //is HCA
								(*sw_entry_it)->m_general_sw_info.m_num_ports);
				}

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u dest_lid: %u - " \
					   "Set lid sate Static static_port: %u (p_group: %p)\n",
					   (*sw_entry_it)->m_general_sw_info.m_lid,
					   dest_lid,
					   static_port,
					   iter->second);

				osm_ar_lft_set(p_ar_data, dest_lid, plft_id, static_port,
					       OSM_AR_IB_LID_STATE_STATIC, 0);
				port = static_port;
				state = OSM_AR_IB_LID_STATE_STATIC;
				group_number = 0;
			} else {
				u_int8_t static_port =0;

				if (iter->second->m_ps_group_bitmask.test_pri(ucast_lft_port))
					static_port = ucast_lft_port;
				else
					static_port = GetNextStaticPort(
							ports_load,
							iter->second->m_ps_group_bitmask,
							sw_lid_num != dest_lid, //is HCA
							(*sw_entry_it)->m_general_sw_info.m_num_ports);

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u dest_lid:%u - Set lid state %u Group:%u static_port:%u (p_group:%p)\n",
					   (*sw_entry_it)->m_general_sw_info.m_lid, dest_lid,
					   this->m_master_db.m_ar_mode,
					   iter->second->m_group_number,
					   static_port,
					   iter->second);

				osm_ar_lft_set(p_ar_data, dest_lid, plft_id, static_port,
					       this->m_master_db.m_ar_mode,
					       iter->second->m_group_number);
				port = static_port;
				state = this->m_master_db.m_ar_mode;
				group_number = iter->second->m_group_number;
			}

			calculated_max_lid = dest_lid;

			if (this->m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u plft id %u dest_lid: %u - " \
					   " GroupNumber: %u LidState: %u Static Port %u\n",
					   (*sw_entry_it)->m_general_sw_info.m_lid,
					   plft_id,
					   dest_lid,
					   group_number, state, port);
			}

		}      /* End of loop destination lids */

		/*
		 * Step 4 -
		 *  Now we will decide if the calculated settings are different from old,
		 *  or if we need to ignore old settings and use the new ones
		 */
		osm_ar_set_required_plft_top(p_ar_data, plft_id, calculated_max_lid);
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

u_int8_t OSMAdaptiveRoutingManager::GetNextStaticPort(uint16_t *ports_load,
						      PSPortsBitset &ps_group_bitmask,
						      bool isHCA,
						      u_int8_t num_ports)
{
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "GetNextStaticPort\n");

	bool set_load = isHCA || m_p_osm_subn->opt.port_profile_switch_nodes;

	uint16_t minLoad = 0xFFFF;
	u_int8_t static_port = 0;

	for (u_int8_t port_num = 1 ; port_num <= num_ports ; port_num++) {
		if (!ps_group_bitmask.test_pri(port_num))
			continue;
		if (!set_load)
			return port_num;
		if (ports_load[port_num] < minLoad) {
			minLoad = ports_load[port_num] ;
			static_port = port_num;
		}
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "GetNextStaticPort returned:%u\n", static_port);

	ports_load[static_port]++;
	return static_port;
}

void OSMAdaptiveRoutingManager::PrintPSGroupData(const char * str,PSGroupData &group_data)
{
	if(this->m_p_osm_log->level & OSM_LOG_DEBUG) {

		stringstream sstr;

		if (group_data.m_group_size == 1)
			sstr << " port:" << (int)group_data.m_port_num;
		else
			sstr << " ps bitset:" << group_data.m_ps_group_bitmask;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "%s Group:%p size:%d %s g_umber:%u\n",
			   str,
			   &group_data,
			   group_data.m_group_size,
			   sstr.str().c_str(),
			   group_data.m_group_number);
	}
}

bool OSMAdaptiveRoutingManager::IsRemoteSupportsDF(ARSWDataBaseEntry &sw_db_entry,
		u_int8_t port_num)
{
	//get remote sw
	osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
	osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
	if(p_remote_node == NULL ||
			p_remote_node->sw == NULL) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C34: IsRemoteSupportsDFP: Invalid remote node to Switch GUID 0x%"
			   PRIx64 ", LID %u though port: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_num);
		return false;
	}

	ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

	if(!IsDFActive(*p_remote_db_entry)){
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Next switche from Switch GUID 0x%"
			   PRIx64 ", LID %u though port: %u do not support DF\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_num);

		return false;
	}
	return true;
}

bool OSMAdaptiveRoutingManager::IsRouteOnRemote(ARSWDataBaseEntry &sw_db_entry,
		int plft_id, u_int8_t port_num, uint16_t lid_num)
{
	//get remote sw
	osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
	osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
	if (p_remote_node == NULL ||
	    p_remote_node->sw == NULL) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C35: Invalid remote node in IsRouteOnRemote"\
			   " from LID %u: plft:%d port:%u to LID:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid,
			   plft_id, port_num, lid_num);
		return false;
	}

	ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

	if (!IsDFActive(*p_remote_db_entry)) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Next switch from LID %u: plft:%d though port:%u to LID:%u "
			   "do not support DF\n",
			   sw_db_entry.m_general_sw_info.m_lid,
			   plft_id, port_num, lid_num);
		return false;
	}

	//get remote plft_id according to the HDL for DF algorithm
	int remote_plft_number = 0;
	int rc =0;
	if (sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
		if (plft_id != 1)
			rc = -1;
		remote_plft_number = 1;
	} else {
		if (plft_id != 1)
			rc = -1;
		remote_plft_number = 0;
	}
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C36: Invalid call to IsRouteOnRemote"\
			   " from LID %u: plft:%d port:%u to LID:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid,
			   plft_id, port_num, lid_num);
		return false;
	}

	set <u_int16_t> & no_df_valid_route =
		p_remote_db_entry->m_p_df_data->m_plft[remote_plft_number].m_no_df_valid_route;

	if (no_df_valid_route.find(lid_num) != no_df_valid_route.end()) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "No route from LID %u: plft:%d though port:%u to LID:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid,
			   plft_id, port_num, lid_num);
		return false;
	}
	return true;
}

bool OSMAdaptiveRoutingManager::IsTrueHopsOnRemote(ARSWDataBaseEntry &sw_db_entry,
						   u_int8_t local_port_num,
						   uint16_t lid_num,
						   int hops_num)
{
	//get remote sw
	uint8_t remote_port_num;
	osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
	osm_node_t *p_remote_node = osm_node_get_remote_node(
			p_osm_node, local_port_num, &remote_port_num);

	if (p_remote_node == NULL ||
	    p_remote_node->sw == NULL ) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C37: Invalid remote node in IsTrueHopsOnRemote from LID %u: to LID:%u\n",
			   sw_db_entry.m_general_sw_info.m_lid,
			   lid_num);
		return false;
	}


	if (!p_remote_node->sw->hops[lid_num] ||
	    p_remote_node->sw->hops[lid_num][0] > hops_num)
		return false;

	for (u_int8_t port_num = 1;
	     port_num <= p_remote_node->node_info.num_ports;
	     ++port_num) {
		if (p_remote_node->sw->hops[lid_num][port_num] != hops_num)
			continue;

		if (port_num == remote_port_num)
			continue;

		return true;
	}

	return false;
}

void OSMAdaptiveRoutingManager::SavePortGroupsAndDump()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	osm_advanced_routing_data_t *p_ar_data;

	for (; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "SavePortGroups for Switch GUID 0x%" PRIx64 ", LID %u:\n",
			   sw_node_it->second.m_general_sw_info.m_guid,
			   sw_node_it->second.m_general_sw_info.m_lid);

		DFAlgorithmData &algorithm_data =
			sw_node_it->second.m_p_df_data->m_algorithm_data;

		p_ar_data = &sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data;
		/*
		 *  Update calculated_ar_group_table:
		 *  At this point we have all assigned groups in algorithm_data.m_assign_groups list
		 */
		if (!algorithm_data.m_ps_groups_map.empty()) {

			PSGroupBitmaskToDataMapIter groups_iter = algorithm_data.m_ps_groups_map.begin();
			osm_ar_subgroup_t *p_subgroup;

			for (; groups_iter != algorithm_data.m_ps_groups_map.end();
			     ++groups_iter) {
				if (groups_iter->second.m_group_size == 1)
					continue; //this is Static port not a group

				p_subgroup = osm_ar_subgroup_get(p_ar_data, groups_iter->second.m_group_number, 0);
				PortsBitset &primary_bitset = groups_iter->second.m_ps_group_bitmask.get_primary();
				PortsBitset &secondary_bitset = groups_iter->second.m_ps_group_bitmask.get_secondary();

				primary_bitset.get_data(*p_subgroup);

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Set Group:%u primary:0x%" PRIx64 "\n",
					   groups_iter->second.m_group_number,
					   cl_hton64(p_subgroup->mask[24]));

				if (secondary_bitset.any()) {
					p_subgroup = osm_ar_subgroup_get(
							p_ar_data,
							groups_iter->second.m_group_number, 1);

					secondary_bitset.get_data(*p_subgroup);
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "Set Group:%u secondary:0x%" PRIx64 "\n",
						   groups_iter->second.m_group_number,
						   cl_hton64(p_subgroup->mask[24]));
				}
			}
		}

		/*
		 *  Now we will decide if the calculated settings are different from old,
		 *  or if we need to ignore old settings and use the new ones
		 */
		if (!algorithm_data.m_ps_groups_map.empty()) {
			this->ARUpdateSWGroupTableTop(sw_node_it->second,
					sw_node_it->second.m_p_df_data->m_algorithm_data.
					m_calculated_groups_number);
		}

     	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ARDumpDFAnalizedSetup()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (!osm_log_is_active(this->m_p_osm_log,OSM_LOG_ROUTING))
		return;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
		this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

		char buff[1024];
		string str = "---------------\n";
		sprintf(buff,
			"\t\t\t\tDragonFly Analized Setup::\n" \
			"\t\t\t\tSwitch GUID............0x%" PRIx64 "\n" \
			"\t\t\t\tLID....................%u\n"   \
			"\t\t\t\tType...................%s\n"   \
			"\t\t\t\tGroup..................%d\n"   \
			"\t\t\t\tUp Ports...............%s\n"   \
			"\t\t\t\tDown Ports.............%s\n",
			sw_node_it->second.m_general_sw_info.m_guid,
			sw_node_it->second.m_general_sw_info.m_lid,
			SW_TYPE_TO_STR(p_df_data->
				       m_df_sw_setup[NEW_DATA].m_sw_type),
				       p_df_data->m_df_group_number,
				       p_df_data->m_df_sw_setup[NEW_DATA].
				       m_up_ports.to_string().c_str(),
				       p_df_data->m_df_sw_setup[NEW_DATA].
				       m_down_ports.to_string().c_str());
		str += buff;
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

int OSMAdaptiveRoutingManager::ARMapPLFTsAndSetSL2VLAct()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;

	PortsBitset    new_up_ports;
	PortsBitset    new_down_ports;

	list<ARSWDataBaseEntry*> update_list;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
		this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

		if (sw_node_it->second.in_temporary_error)
			continue;

		if(!IsDFActive(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "DF/AR not supported or not enabled, " \
				   "ARMapPLFTsAndSetSL2VLAct skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			continue;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Map pLFT and VL2VL on Switch GUID 0x%" PRIx64 ", LID %u" \
			   " Old up:%s down:%s New up:%s down:%s \n",
			   sw_node_it->second.m_general_sw_info.m_guid,
			   sw_node_it->second.m_general_sw_info.m_lid,
			   p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports.to_string().c_str(),
			   p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports.to_string().c_str(),
			   p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports.to_string().c_str(),
			   p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports.to_string().c_str());

		new_up_ports =
			p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports &
			~p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports;

		new_down_ports =
			p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports &
			~p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

		if (new_up_ports.none() && new_down_ports.none())
			continue;

		PortsBitset new_ports = new_up_ports | new_down_ports;
		update_list.push_back(&sw_node_it->second);

		osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;
		p_osm_node->new_ext_node_info.sl_2_vl_act = 1;

		u_int8_t port_block;
		u_int8_t prev_port_block = (u_int8_t)-1;
		for (u_int8_t port_num = 0;
		     port_num <= sw_node_it->second.m_general_sw_info.m_num_ports;
		     ++port_num) {
			if (!new_ports.test(port_num))
				continue;

			port_block = port_num / IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

			if (port_block != prev_port_block) {
				ARMapPLFTs(sw_node_it->second, port_block);
				prev_port_block = port_block;
			}

		}
	}

	list<ARSWDataBaseEntry*>::iterator update_list_iter = update_list.begin();
	DfSwData *p_df_data;
	for (; update_list_iter != update_list.end(); ++update_list_iter) {
		if ((*update_list_iter)->in_temporary_error)
			continue;

		(*update_list_iter)->m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports |=
			(*update_list_iter)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports;

		(*update_list_iter)->m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports |=
			(*update_list_iter)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports;
		p_df_data = (*update_list_iter)->m_p_df_data;
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID 0x%" PRIx64 ", LID %u" \
			   " Old up: %s down: %s\n",
			   (*update_list_iter)->m_general_sw_info.m_guid,
			   (*update_list_iter)->m_general_sw_info.m_lid,
			   p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports.to_string().c_str(),
			   p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports.to_string().c_str());
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//VL: 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
static uint8_t slvl_mapping[][IB_NUMBER_OF_SLS] =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //port 0 SL2VL
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls ???
    {0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1}, // 2 vls
    {0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3}, // 4 vls
    {0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4, 5, 6, 7}, // 8 vls
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}}; // 16 vls


static uint8_t inc_slvl_mapping[][IB_NUMBER_OF_SLS] =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //reserved
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls ???
    {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}, // 2 vls
    {1, 1, 3, 3, 1, 1, 3, 3, 1, 1, 3, 3, 1, 1, 3, 3}, // 4 vls
    {1, 1, 3, 3, 5, 5, 7, 7, 1, 1, 3, 3, 5, 5, 7, 7}, // 8 vls
    {1, 1, 3, 3, 5, 5, 7, 7, 9, 9, 11, 11, 13, 13, 15, 15}}; // 16 vls


//validate that op_vl >1 or that out port is not switch
//p_physp is a osm_physp_t that suits the given sw_db_entry
int OSMAdaptiveRoutingManager::GetOpVlForVL2VL(ARSWDataBaseEntry &sw_db_entry,
					       osm_physp_t *p_physp,
					       u_int8_t out_port,
					       u_int8_t min_op_vls,
					       u_int8_t &op_vls)
{
	op_vls = 0;

	OSM_ASSERT(p_physp);

	if (!osm_link_is_healthy(p_physp)) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8C38: Switch GUID 0x%" PRIx64 ", LID %u, " \
			   "out port: %u, the link is not healthy.\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   out_port);
		return -1;
	}

	//nuber of vls opcodes
	op_vls = ib_port_info_get_op_vls(&p_physp->port_info);

	if (op_vls < min_op_vls) {
		//no problem if remote is CA
		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C39: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "Fail to get remote_physp: %u.\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   out_port);

			return -1;
		}
		u_int8_t remote_node_type = osm_node_get_type(p_remote_physp->p_node);

		if (remote_node_type == IB_NODE_TYPE_SWITCH) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8C40: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "invalid op_vl: %u < %u (minimum required by algorithm) on port: %u. "
				   "Skip setting SL2VL.\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   op_vls, min_op_vls, out_port);
			return -1;
		}
	}
	return 0;
}

static uint8_t op_vls_to_slvl_host_mapping[][IB_NUMBER_OF_SLS] =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //reserved
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 2 vls
    {0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2}, // 4 vls
    {0, 2, 4, 6, 0, 2, 4, 6, 0, 2, 4, 6, 0, 2, 4, 6}, // 8 vls
    {0, 2, 4, 6, 8, 10, 12, 14, 0, 2, 4, 6, 8, 10, 12, 14}}; // 16 vls

int OSMAdaptiveRoutingManager::GetSL2VLDF(osm_physp_t *osm_phys_port,
					uint8_t iport_num, uint8_t oport_num,
					uint8_t (*sl2vl)[IB_NUMBER_OF_SLS])
{
	int rc = 0;
	u_int8_t op_vls = 0;
	uint8_t (*p_slvl) [IB_NUMBER_OF_SLS] = NULL;
	bool is_turn = false;

	if (!m_df_configured)
		return -1;
	osm_node_t *p_node = osm_physp_get_node_ptr(osm_phys_port);
	uint64_t guid = cl_ntoh64(osm_node_get_node_guid(p_node));

	if (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
		if (oport_num == 0 && !ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info))
			return -1;

		GuidToSWDataBaseEntryIter sw_it = m_sw_db.m_sw_map.find(guid);

		if (sw_it == m_sw_db.m_sw_map.end())
			return -1;

		PortsBitset up_ports = sw_it->second.m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports;
		PortsBitset down_ports = sw_it->second.m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

		bool is_up_port_out_port = up_ports.test(oport_num);

		/* for enhanced port0 leave op vls number 0 */
		if (oport_num != 0)
			rc = GetOpVlForVL2VL(
				sw_it->second, osm_phys_port, oport_num, MIN_VL_CAP, op_vls);

		if (rc)
			return rc;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "GetSLVL sw 0x%" PRIx64 " is_Turn %s op vls %u iport %u oport %u "
			   "up: %s down: %s\n",
			   guid, is_turn ? "TRUE" : "FALSE", op_vls, iport_num, oport_num,
			   up_ports.to_string().c_str(),
			   down_ports.to_string().c_str());

		if (up_ports.test(iport_num)) {
			if (is_up_port_out_port)
				/* this is turn */
				is_turn = true;
		} else if (!down_ports.test(iport_num))
			/* the port is not connected, configure with default */
			return -1;
		p_slvl = is_turn ? &(inc_slvl_mapping[op_vls]) : &(slvl_mapping[op_vls]);

		memcpy(sl2vl, p_slvl, sizeof(*sl2vl));

	} else if (p_node->node_info.node_type == IB_NODE_TYPE_CA) {
		op_vls = ib_port_info_get_op_vls(&osm_phys_port->port_info);

		if ((m_master_db.m_en_sl_mask == OSM_ALL_SL_VL_AR_ENABLED) ||
		    (op_vls <= 2)) {
			memcpy(sl2vl, &op_vls_to_slvl_host_mapping[op_vls], sizeof(*sl2vl));
		} else {
			u_int8_t slvl_array[IB_NUMBER_OF_SLS] = {0};

			//Disabled SLs will be mapped (round robin) to VLs 0,1
			//Enabled SLs will be mapped (round robin) to VLs 2, .. max op vl
			u_int8_t ar_enabled_vl = 2;
			u_int8_t num_vls = (1 << (op_vls - 1));
			u_int8_t max_ar_enabled_vl = num_vls - 2;
			u_int8_t ar_disabled_vl = 0;

			for (u_int8_t sl = 0; sl < IB_NUMBER_OF_SLS; ++sl) {
				if ((1 << sl) & m_master_db.m_en_sl_mask) {
					//sl is AR enabled
					slvl_array[sl] = ar_enabled_vl;

					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "slvl_mapping[%u]=%u AR enabled\n",
						   sl, slvl_array[sl]);

					ar_enabled_vl += 2;
					if (ar_enabled_vl > max_ar_enabled_vl)
						ar_enabled_vl = 2;
				} else {
					slvl_array[sl] = ar_disabled_vl % 2;
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "slvl_mapping[%u]=%u AR enabled\n",
						   sl, slvl_array[sl]);
					ar_disabled_vl++;
				}
			}

			memcpy(sl2vl, &slvl_array, sizeof(*sl2vl));
		}
	}
	return rc;
}

int OSMAdaptiveRoutingManager::ARDefinePLFTs()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	osm_advanced_routing_data_t *p_ar_data;

	//Set pLFT Info
	for (; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
		if (sw_node_it->second.in_temporary_error)
			continue;

		if(!IsDFActive(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "DF/AR not supported or not enabled, " \
				   "Set pLFT Info and Def are skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);

			// clear AR configuration from sw - SM should configure this sw
			sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;

			continue;
		}

		p_ar_data = &sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data;

		osm_ar_plft_def_set(p_ar_data, 0, 0, PLFT_LID_SPACE, 0, 0);
		osm_ar_plft_def_set(p_ar_data, 1, 0, PLFT_LID_SPACE, 1, 0);
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//	PLFTMap for vl15 should be equal to vl0 settings
/* plft ids for following SLs: sl0,sl1,sl2,sl3,sl4,sl5,sl6,sl7,sl8,sl9,sl10,sl11,sl12,sl13,sl14,sl15 */
static ib_port_sl_to_plft_map_entry_t port_vl2plft_leaf_down      = {1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1};

static ib_port_sl_to_plft_map_entry_t port_vl2plft_spine_down     = {1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1};
static ib_port_sl_to_plft_map_entry_t port_vl2plft_spine_down_vl  = {2,1,2,1,2,1,2,1,0,0,0,0,0,0,0,1};

void OSMAdaptiveRoutingManager::ARMapPLFTs(ARSWDataBaseEntry &sw_db_entry,
                                        u_int8_t port_block)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	PortsBitset total_down_ports =
		sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports |
		sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

	ib_port_sl_to_plft_map_entry_t *p_port_vl2plft_down = NULL;
	if (sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type ==
			SW_TYPE_LEAF) {
		p_port_vl2plft_down = &port_vl2plft_leaf_down;
	} else if (sw_db_entry.m_p_df_data->plft_number == 2) {
		p_port_vl2plft_down = &port_vl2plft_spine_down;
	} else
		p_port_vl2plft_down = &port_vl2plft_spine_down_vl;


	//Set pLFT Map
	ib_port_sl_to_plft_map_t port_sl_to_plft_map = {0};

	u_int8_t port_num = port_block * IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;

	for (int port_idx = 0; port_idx < IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE; ++port_num, ++port_idx)
	{
		if (total_down_ports.test(port_num))
			port_sl_to_plft_map.port[port_idx] = *p_port_vl2plft_down;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Map pLFT on Switch GUID 0x%" PRIx64 ", LID %u, " \
			   "port_block:%u port_num:%u idx:%u is_down:%u total_down:%s\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   port_block, port_num, port_idx,
			   total_down_ports.test(port_num),
			   total_down_ports.to_string().c_str());
	}

	osm_ar_port_sl_to_plft_set_block(sw_db_entry.m_general_sw_info.m_p_osm_sw,
			port_block, FALSE,
			&port_sl_to_plft_map);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

